#!/usr/bin/python

# Copyright (c) Members of the EGEE Collaboration. 2004. 
# See http://www.eu-egee.org/partners/ for details on the copyright
# holders.  
# 
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at 
# 
#     http://www.apache.org/licenses/LICENSE-2.0 
# 
# Unless required by applicable law or agreed to in writing, software 
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License.

import amgaMonitor
import sys
import time
from datetime import datetime

class ServerMiner:
    """
    Used to extract information from logs.
    """
    def __init__(self):
        """ Opens all the files required to save the results
            - One main file
            - One file for each capture process
            - One file for each queue
            - One file for each propagation process
        """
        self.prevLog = None
        self.startTime = datetime.now()
        fileName = self.startTime.strftime("%Y%m%d%H%M%S") +"-amga.dat"
        self.dataFile = file(fileName, "w")
        self.dataFile.write("# time elapsed | write delta\n")

    def processLog(self, log):
        pass


class RepDaemonMiner:
    def __getTimeStamp(self):
        #return datetime.now().strftime("%Y%m%d%H%M%S")
#        return datetime.now().strftime("%H%M%S")
        return ""

    """
    Used to extract information from logs.
    """
    def __init__(self):
        """ Opens all the files required to save the results
            - One main file
            - One file for each capture process
            - One file for each queue
            - One file for each propagation process
        """
        self.prevLog = None
        self.sessionFiles = {}
        fileName = "repdaemon.dat"
        self.dataFile = file(fileName, "w")
        self.dataFile.write("%12s%12s%12s%12s%12s\n" % 
                            ("time", "Write rate", "Pending", "Send rate", "Sent KB/S"))

    def __getSessionFile(self, pID):
        if pID not in self.sessionFiles:
            print "Opening data file for session: ", pID
            fileName = "session_" + pID + ".dat"
            f = file(fileName, "w")
            f.write("%12s%12s%12s%12s%12s\n" % 
                    ("time", "Send rate", "Sent KB/S", "Sent xid", "Ack xid"))
            self.sessionFiles[pID] = f
        return self.sessionFiles[pID]

    def processLog(self, log):
        # Extract relevant information
        if self.prevLog == None:
            self.startTime = log.uptime
        else:
            totalTime = log.uptime - self.startTime            
            sampleTime = float(log.uptime - self.prevLog.uptime)
            print "Sample time: ", sampleTime
            
            writeRate = (log.writeXid - self.prevLog.writeXid)/sampleTime
            sendRate = (log.logsSentTotal - self.prevLog.logsSentTotal)/sampleTime
            dataRate = ((log.dataSentTotal - self.prevLog.dataSentTotal)/1024)/sampleTime
#            self.dataFile.write("%12d%12d%12d%12d%12.1f\n" % 
#                                (totalTime, log.logsPending, readRate, sendRate, dataRate))
            self.dataFile.write("%12d%12d%12d%12d%12.1f\n" % 
                                (totalTime, writeRate, log.logsPending, sendRate, dataRate))
            self.dataFile.flush()
            # find the capture process of the previous log with the same id
            for sessionKey in log.sessions:
                # Is this a new session? If so, skip it since we don't
                # have a previous log to use for extracting values
                if sessionKey not in self.prevLog.sessions:
                    continue

                prevSession = self.prevLog.sessions[sessionKey]
                f = self.__getSessionFile(sessionKey)
                session = log.sessions[sessionKey]
                # Disconnected
                if session.logsSent == -1:
                    sendRate = 0
                    dataRate = 0
                elif session.sessionID == prevSession.sessionID:
                    sendRate = (session.logsSent - prevSession.logsSent)/sampleTime
                    dataRate = ((session.dataSent - prevSession.dataSent)/1024)/sampleTime
                else:
                    # It's a new session. We lost some values, but that's not critical
                    sendRate = session.logsSent/ float(session.uptime)
                    dataRate = (session.dataSent/1024)/float(session.uptime)
                f.write("%12d%12d%12.1f%12d%12d\n" % 
                        (totalTime, sendRate, dataRate, session.sentXid, session.ackedXid))
                f.flush()

        # Update the values
        self.prevLog = log


#############################
# Script starts here
#############################
def main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-p", "--port", dest="port", type="int", default=8822,
                      help="AMGA instance port")
    parser.add_option("-i", "--interval", dest="interval", type="int", default=1,
                      help="Poll interval")

    (opts, args) = parser.parse_args()
    if len(args) == 0:
        host = "localhost"
    else:
        host = args[0]

    print "Querying: " + host + ":" + str(opts.port)

    parser = amgaMonitor.Parser()
    log = None
    (serverType, xmlData) = amgaMonitor.pollAMGA(host, opts.port)
    if serverType == amgaMonitor.ServerType.MDSERVER:
        miner = ServerMiner()
    elif serverType == amgaMonitor.ServerType.MDREPDAEMON:
        miner = RepDaemonMiner()
    else:
        raise Exception("Bad id: " + str(serverType))

    while True:
#        print xmlData
        log = parser.parse(serverType, xmlData)        
        print log
        miner.processLog(log)
        time.sleep(opts.interval)
        (serverType, xmlData) = amgaMonitor.pollAMGA(host, opts.port)


if __name__ == "__main__":
    sys.exit(main())
