#!/usr/bin/python

# Copyright (c) Members of the EGEE Collaboration. 2004. 
# See http://www.eu-egee.org/partners/ for details on the copyright
# holders.  
# 
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at 
# 
#     http://www.apache.org/licenses/LICENSE-2.0 
# 
# Unless required by applicable law or agreed to in writing, software 
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
# See the License for the specific language governing permissions and 
# limitations under the License.

from xml import sax
import socket
import sys
import time

# Idiom for Python enums
class ServerType:
    MDSERVER, MDREPDAEMON = range(2)

class RepDaemonSession:
    def __init__(self, node, logDict):
        self.node = node
        self.logsSent = int(logDict["repdaemon.Session." + node + ".LogsSent"])
        self.dataSent = int(logDict["repdaemon.Session." + node + ".DataSent"])
        self.ackedXid = long(logDict["repdaemon.Session." + node + ".AckedXid"])
        self.sentXid = long(logDict["repdaemon.Session." + node + ".SentXid"])
        self.uptime = int(logDict["repdaemon.Session." + node + ".Uptime"])
        self.sessionID = int(logDict["repdaemon.Session." + node + ".SessionID"])        
        
    def __str__(self):
        return str(self.__dict__)


class RepDaemonLog:
    def __init__(self, logDict):
        self.logsPending = int(logDict["repdaemon.LogsPending"])
        self.logsSentTotal = int(logDict["repdaemon.LogsSent.Total"])
        self.dataSentTotal = int(logDict["repdaemon.DataSent.Total"])        
        self.writeXid = long(logDict["repdaemon.WriteXid"])
        self.sessionsCurrent = int(logDict["repdaemon.Sessions.Current"])
        self.sessionsTotal = int(logDict["repdaemon.Sessions.Total"])
        self.uptime = int(logDict["repdaemon.Uptime"])
        self.nodeName = logDict["NodeName"]
        self.sessions = {}
        for x in logDict:
            if x.startswith("repdaemon.Session."):
                # Extract the node name
                tokens = x.split(".")
                node = tokens[2]
                # Add this node if it wasn't already found
                if (node not in self.sessions):
                    self.sessions[node] = RepDaemonSession(node, logDict)

    def __str__(self):
        lines = [x + ":" + str(self.__dict__[x]) for x in self.__dict__]
        lines = filter(lambda x: not x.startswith("sessions:"), lines)
        lines.sort()
        out = "\n".join(lines)
        out += "\nActive Sessions\n  "
        out += "\n  ".join([str(x) for x in self.sessions.values()])
        return out


class MDServerLog:
    def __init__(self, logDict):
        self.preparedConnections = int(logDict["PreparedConnections"])
        self.maxSessions = int(logDict["MaxSessions"])
        self.usedSessions = int(logDict["UsedSessions"])
        self.usedConnections = int(logDict["UsedConnections"])
        self.maxConnections = int(logDict["MaxConnections"])        
        self.sessionStorage = logDict["SessionStorage"]        
        self.uptime = int(logDict["Uptime"])
        self.nodeName = logDict["NodeName"]
#        self.logsWritten = int(logDict["repmaster.LogsWritten"])
#        self.updates = int(logDict["mdserver.Updates"])
#        self.commands = int(logDict["mdserver.Commands"])

    def __str__(self):
        lines = [x + ":" + str(self.__dict__[x]) for x in self.__dict__]
        return "\n".join(lines)


class ContentGenerator(sax.handler.ContentHandler):
    """
    Parses the XML answer obtained from the server
    """
    def __init__(self):
        sax.handler.ContentHandler.__init__(self)
        self.dict = {}
        self.key = None
        self.nextKey = False

    # ContentHandler methods
    def startDocument(self):
        pass

    def startElement(self, name, attrs):
        self.nextKey = (name == "Key")

    def endElement(self, name):
        pass

    def characters(self, content):
        content = content.strip()
        if len(content) == 0:
            return
        if self.nextKey:
            self.key = content
            # Make sure the map has an entry for this key, even if
            # the server sends an empty value string.
            self.dict[self.key] = ""
            self.nextKey = False
        else:
            if self.key != None:
                self.dict[self.key] = content
                self.key = None

    def ignorableWhitespace(self, content):
        pass

    def processingInstruction(self, target, data):
        pass


class Parser:
    def __init__(self):
        self.handler = ContentGenerator()
        
    def parse(self, serverType, xmlData):
        self.handler.dict.clear()
        sax.parseString(xmlData, self.handler)
        if serverType == ServerType.MDSERVER:
            return MDServerLog(self.handler.dict)
        elif serverType == ServerType.MDREPDAEMON:
            return RepDaemonLog(self.handler.dict)
        else:
            return self.handler.dict


def pollAMGA(host, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    data = ""
    # Read the three header lines
    while True:
        data += s.recv(1024)
        if data.count("\n") == 3:
            break

    if data.startswith("ARDA Metadata Server"):
        instance = ServerType.MDSERVER
    elif data.startswith("ARDA Metadata Replication Daemon"):
        instance = ServerType.MDREPDAEMON
    else:
        raise Exception("Unknown server identifier: " + data)
        
    # Send the command
    s.send('statistics\n\n')
    data = ""
    while True:
        newData = s.recv(4096)
        if newData == "":
            break
        data += newData
    s.close()
    return (instance, data)

        
def main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-p", "--port", dest="port", type="int", default=8822,
                      help="AMGA instance port")
    parser.add_option("-i", "--interval", dest="interval", type="int", default=1,
                      help="Poll interval")

    (opts, args) = parser.parse_args()    
    if len(args) == 0:
        host = "localhost"
    else:
        host = args[0]

    print "Querying: " + host + ":" + str(opts.port)
    parser = Parser()
    log = None
#    count = 0
    while True:
        (serverType, xmlData) = pollAMGA(host, opts.port)
        #print xmlData
        oldLog = log
        log = parser.parse(serverType, xmlData)
        print log
#        count += 1
#        print "Queries:",count
        time.sleep(opts.interval)

#    while True:
#        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#        s.connect((host, opts.port))
#        while True:
#            data = s.recv(1024)
#            if data == "":
#                break            
#        count += 1
#        print "Queries:",count
    
if __name__ == "__main__":
    sys.exit(main())
   
