/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


/*
 * AMGAFrame.cpp
 *
 *  Created on: Mar 23, 2009
 *      Author: root
 */

#include "AMGAFrame.h"

AMGAFrame* AMGAFrame::pAMGAFrame = NULL;
// signal handler variables
sighandler_t AMGAFrame::oldSIGINT;
sighandler_t AMGAFrame::oldSIGCHILD;

bool AMGAFrame::debug = false;

std::ofstream AMGAFrame::logStream;
std::string AMGAFrame::logFile;



AMGAFrame::AMGAFrame() {
	port = -1;
	logFile = "";
	runningMode = ""; // KIMNG090323 the mode of AMGA - Process Mode or Thread Mode.  the default value is "process"
	daemon = false;
	terminateRequest = false;
	debug = true;

	forceSessions = false;
	allowSessions = true;
	maxConnects = 1000000;
	sessionTimeout = 300;
	execUserID = -1;
	execUserGroup = -1;
	execRoot = "";
	maxConnectsPerUser = -1;
	reservedConnections = -1;
}

AMGAFrame::~AMGAFrame() {
}

void AMGAFrame::init(const ConfigParser &config, string strRunningMode,
		bool bDebug, bool bDaemon, int nPort, string strLogFile) {

	// the static point variable to AMGAFrame
	AMGAFrame::pAMGAFrame = this;

	AMGAFrame::debug = bDebug;
	port = nPort;
	logFile = strLogFile;
	configParser = config;
	runningMode = strRunningMode;
	daemon = bDaemon;
	DMESG("RunningMode, debug, bDaemon, nPort, logFile, runningMode : " << strRunningMode << " , " << bDebug << " , " << bDaemon << " , " << nPort << " , " << strLogFile << endl);
}

int AMGAFrame::initSessions(SSL_CTX *ctx, char *appData, int sessionTimeout) {

	// Turn off internal session cache
	SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_SERVER | SSL_SESS_CACHE_NO_INTERNAL);

	// Set callbacks for session caching in DB file
	SSL_CTX_sess_set_new_cb(ctx, AMGAFrame::ssl_callback_NewSessionCacheEntry);
	SSL_CTX_sess_set_get_cb(ctx, AMGAFrame::ssl_callback_GetSessionCacheEntry);
	SSL_CTX_sess_set_remove_cb(ctx, AMGAFrame::ssl_callback_RemoveSessionCacheEntry);

	// Set application data
	SSL_CTX_set_app_data(ctx, appData);

	// Set session context
	if (!SSL_CTX_set_session_id_context(ctx,
			(const unsigned char *) "MDServer", 8))
		abort();

	// Session timeouts
	SSL_CTX_set_timeout(ctx, sessionTimeout);

	return 0;
}

int AMGAFrame::run() {

	// Now let's become a daemon, if wanted
	if (daemon) {
		// Our process ID and Session ID
		pid_t pid, sid;

		// Fork off the parent process
		pid = fork();
		if (pid < 0) {
			exit(EXIT_FAILURE);
		}

		// We are the parent, and can die
		if (pid > 0) {
			exit(EXIT_SUCCESS);
		}

		// Change the file mode mask: full access to our own files
		umask(0);

		// Create a new session id (SID) for the child process
		sid = setsid();
		if (sid < 0) {
			ERRLOG("Could not get new session ID for daemon proces" << std::endl);
			exit(EXIT_FAILURE);
		}

		// change working dir to a safe place
		if ((chdir("/")) < 0) {
			ERRLOG("Could not chdir /");
			exit(EXIT_FAILURE);
		}

		// Close stdin, should be doing it also with stdout, but only if we have
		// a logfile set up
		close(STDIN_FILENO);
		//  close(STDOUT_FILENO);
		//  close(STDERR_FILENO);
	}

	signal(SIGHUP, hup_sig_handler);

	if (configLoad())
		return -1;

	if (initialize())
		return -1;

	// Main Loop
	MainLoop();

	LOG("Shutting down server." << std::endl);

	delete conMan;

	delete[] children;
	delete[] deadChildren;

	if (logStream.is_open())
		logStream.close();

	return 0;
}

void AMGAFrame::MainLoop() {

	/* Change Architecture */
	conMan->start(30, idleTimeout); // Look at connections every 30 sec

	LOG("Listening on port: " << port << std::endl);
	// Open the ServerSocket on the master so it can be shared by users.
	TCPServerSocket serverSocket(port);
	LOG("Starting up slaves" << std::endl);
	// Install signal handler
	AMGAFrame::oldSIGCHILD = signal(SIGCHLD, AMGAFrame::child_sig_handler);
	AMGAFrame::oldSIGINT = signal(SIGINT, AMGAFrame::int_sig_handler);
	signal(SIGTERM, AMGAFrame::int_sig_handler);


	time_t lastTime = time(NULL);
	int numCalls = 0;


        try {
          // Initialize replication
          ReplicationManager::initManager(*ConfigParser::configParser, debug);
        } catch (runtime_error &ex) {
          ERRLOG(ex.what() << endl);
          return ;
        }


	while (1) {
		// We don't wait the fist time we come, because conMan is unlocked
		/* Change Architecture */
		conMan->waitMaster(); // Wait until: timeout, interrupt, notify from child
#if 0           // just for debug
		time_t endTime = time(NULL);
		if ((endTime - lastTime)> 5)
		{
			terminateRequest = true;
		}
#endif
		if (terminateRequest) {
			LOG("Shutting down all slaves" << std::endl);
			/* Change Architecture */
			conMan->reapAll();
			LogReceiverManager::cleanup();
			while (sleep(2)) {
			};
			removeDead();
			/* Change Architecture */
			conMan->killLeftovers(true);
			/* Change Architecture */
			// Restore the old signal handler, in case someone comes late
			signal(SIGCHLD, AMGAFrame::oldSIGCHILD);
			break; // End master
		} else {
			/* Change Architecture */
			conMan->reap(idleTimeout); // Reap all idle process
			conMan->killLeftovers(false);

			// Run session expiration every 300 seconds or after 1000 connects
			time_t now = time(NULL);
			numCalls++;
			if ((now - lastTime > 300) || (numCalls >= maxSessions / 2 - 2)) {
				numCalls = 0;
				lastTime = now;
				/* Change Architecture */
				conMan->expireSessions(sessionTimeout);
			}
		}
		removeDead();

		/* Change Architecture */
		if (conMan->getPrepared() < minProcesses) {
			DMESG("->> Getting new processes: require " << minProcesses
					<< std::endl);
			int ret;
			std::vector<int> processSlots;
			processSlots.resize(minProcesses);
			/* Change Architecture */
			if (conMan->prepareProcs(processSlots)) {
				ERRLOG("Error creating slave processes" << std::endl);
			}

			for (size_t i = 0; i < processSlots.size(); i++) {
				if (processSlots[i] < 0) {
					ERRLOG("Could not spawn enough slaves" << std::endl);
					break;
				}

				DMESG("->> Forking " << i << " slot "
						<< processSlots[i]<< std::endl);
				int pid;
				switch (pid = fork()) {

				case -1: /* error */
					PERRLOG("Could not fork new server process") ;
					/* Change Architecture */
					conMan->cleanupSlot(processSlots[i]);
					break;

				case 0: /* child process */
					DMESG("->> Forking child " << i << std::endl);
					// The following is necessary because we actually have a copy of the
					// conMan object
					/* Change Architecture */

					if (!conMan->setSlave(processSlots[i])) {
						ret = slaveProcess(serverSocket);
					} else {
						ERRLOG("Unable to start process");
					}

					exit(ret);

				default: /* parent process */
					children[processSlots[i]] = pid;
				}
			}
		}
	}
}

void AMGAFrame::removeDead() {
	// Remove dead children from connection manager
	// DMESG("REMOVING dead children" << std::endl);
	for (int i = 0; i < numSlots; i++) {
		if (!deadChildren[i])
			continue;
		conMan->remove(deadChildren[i]);DMESG("REMOVED DEAD CHILD: " << deadChildren[i] << std::endl);
		deadChildren[i] = 0;
	}
}

int AMGAFrame::initialize() {

	if (!conMan->init(maxProcesses)) {
		ERRLOG("Could not initialize connection manager!"
				<< std::endl);
		return -1;
	}

	assert(conMan);
	StatsCollector::registerStatsSource(*conMan);

	if (conMan->initSessionCache(maxSessions, maxSessionSize)) {
		ERRLOG("Could not initialize session cache!"
				<< std::endl);
		return -1;
	}

	// Initialize debugging
	authCtx.setDebug(debug);

	// Init SSL: Session handling + VOMS Attribute parsing
	authCtx.init(configParser);
	if (authCtx.useSSL) {
		initSessions(authCtx.ctx, (char *) conMan, sessionTimeout);
		// Load VOMS extensions to certificates into SSL
		VOMSInfo::initASN1Parser();
	}

	numSlots = maxProcesses;
	children = new int[numSlots];
	deadChildren = new int[numSlots];
	for (int i = 0; i < numSlots; i++) {
		children[i] = deadChildren[i] = 0;
	}
	return 0;
}

int AMGAFrame::configLoad() {

	DatabaseConnection::initialize(configParser);

	try {
		if (port<0) { // Not set on command line
			port = configParser.getInt("Port");
			LOG("PORT : " + port);
		}
	} catch (std::string &err) {
		port = 8822; // Neither set on command line nor in configuration
			LOG("PORT : DEFAULT");
	}

	try {
		if (!logFile.size()) { // Not set on command line
			logFile = configParser.getString("Logfile");
		}
	} catch (std::string &err) {
		// Neither set on command line nor in configuration
		logFile = "";
	}

	if (logFile.size()) {
		logStream.open(logFile.c_str(), ios_base::app);
		Display::setOutput(logStream.rdbuf());
	}

	// Logging configuration
	try {
		int maxLogLen = configParser.getInt("MaxLogLen");
		Display::setMaxLogLen(maxLogLen);
	} catch (std::string err) {
		// Default is as many as you like
	}

	// Connection configuration
	try {
		maxConnects = configParser.getInt("MaxConnectsPerProcess");
	} catch (std::string err) {
		// Default is as many as you like
	}

	try {
		sessionTimeout = configParser.getInt("SessionTimeout");
	} catch (std::string err) {
		// Default is as many as you like
	}

	try {
		idleTimeout = configParser.getInt("IdleTimeout");
	} catch (std::string err) {
		idleTimeout = 20 * 60; // Default is 20 minutes
	}

	// User executable scripts
	try {
		execRoot = configParser.getString("ExecRoot");
		LOG("Allowing execution of scripts at " << execRoot << std::endl);
		std::string execUser = configParser.getString("ExecUser");

		struct passwd *pw = getpwnam(execUser.c_str());
		if (pw) {
			execUserID = pw->pw_uid;
			execUserGroup = pw->pw_gid;
		}
		LOG("UserID of scripts is " << execUserID
				<< ", GID is: " << execUserGroup << std::endl);
	} catch (std::string err) {
	}

	try {
		if (runningMode.compare("thread") == 0) {
			threadNum = configParser.getInt("initThreadNumber");
			LOG("The initial number of Threads : " << threadNum << std::endl);
		}
	} catch (std::string err) {
		threadNum = 4;
		LOG("The default initial number of Threads : " << threadNum << std::endl);
	}

	// Configure session handling
	try {
		std::string s = configParser.getString("Sessions");
		ConfigParser::toLower(s);
		if (s == "force") {
			forceSessions = true;
			allowSessions = true;
		}

		if (s == "allow" || s == "yes" || s == "1") {
			allowSessions = true;
		}

		if (!(s == "yes" || s == "1" || s == "force" || s == "allow" || s
				== "no" || s == "0")) {
			ERRLOG("Unknown value for option UseSSL\n");
			abort();
		}

	} catch (std::string err) {
	}

/*
	try {
		ReplicationManager::initManager(*ConfigParser::configParser, debug);
	} catch (runtime_error &ex) {
		ERRLOG(ex.what() << endl);
		return -1;
	}
*/

	try {
		maxProcesses = configParser.getInt("MaxProcesses");
	} catch (std::string err) {
		maxProcesses = 50;
	}

	// Initialize connection manager see which one we will need
	bool useDBMSessionCache;

	try {
		useDBMSessionCache = (bool) configParser.getInt("UseDBMSessionCache");
	} catch (std::string err) {
		useDBMSessionCache = false;
	}

	if (useDBMSessionCache) {
		conMan = new MDConManDB();
		LOG("Connection caching done in DBM file\n");
	} else {
		conMan = new MDConManMem();
		LOG("Connection caching done in shared memory\n");
	}

	try {
		minProcesses = configParser.getInt("MinProcesses");
	} catch (std::string err) {
		minProcesses = 2;
	}

	try {
		maxSessions = configParser.getInt("MaxSessions");
	} catch (std::string err) {
		maxSessions = 1024;
	}

	try {
		maxSessionSize = configParser.getInt("MaxSessionSize");
	} catch (std::string err) {
		maxSessionSize = 3000;
	}

	// Set connection number limitations
	try {
		maxConnectsPerUser = configParser.getInt("MaxConnectsPerUser");
	} catch (std::string err) {
	}

	try {
		reservedConnections = configParser.getInt("ReservedConnections");
	} catch (std::string err) {
	}

        try {
          MDServer::allowFedConnFail = configParser.getInt("AllowFedConnFail");
        } catch (std::string err) {
          MDServer::allowFedConnFail = 1;
        }

        try {
          MDServer::serverSideFed = configParser.getInt("ServerSideFed");
        } catch (std::string err) {
          MDServer::serverSideFed = 1;
        }


        try {
          MDServer::allowFedExecFail = configParser.getInt("AllowFedExecFail");
        } catch (std::string err) {
          MDServer::allowFedExecFail = 1;
        }

        if( MDServer::allowFedExecFail == 0 ) 
          MDServer::allowFedConnFail = 0;

        DMESG("Fed conn Fail Allow : " << MDServer::allowFedConnFail << std::endl); 
        DMESG("Fed Exec Fail Allow : " << MDServer::allowFedExecFail << std::endl); 

	return 0;
}

long AMGAFrame::calclock(struct timespec *myclock) {
	long nano_time, temp, temp_n;
	if (myclock[1].tv_nsec >= myclock[0].tv_nsec) {
		temp = myclock[1].tv_sec - myclock[0].tv_sec;
		temp_n = myclock[1].tv_nsec - myclock[0].tv_nsec;
		nano_time = 1000000000 * temp + temp_n;
	} else {
		temp = myclock[1].tv_sec - myclock[0].tv_sec - 1;
		temp_n = 1000000000 + myclock[1].tv_nsec - myclock[0].tv_nsec;
		nano_time = 1000000000 * temp + temp_n;
	}
	return nano_time;
}

/* Change Architecture */

int AMGAFrame::initConnection(TCPSocket &socket, MDServer *connection,
		UserContext &userCtx) {
	// Set socket options to TCP_NODELAY during setup
	socket.setTCPNoDelay(true);

	// Initialize client connection, check whether SSL is demanded,
	// Resume session if requested
	bool useSSL;
	// Check whether we need to resume a session
	uint64_t sessionID = 0;
	//string greeting = "ARDA Metadata Server " VERSION;
	string greeting = "ARDA Metadata Server ";

	try {
		// Plain text. Send greeting and ask user if SSL is requested.
		// call initConnection in protocolHandling.cc
		int ret = ::initConnection(socket, authCtx.useSSL, greeting, useSSL,
				sessionID);
		DMESG("===== ret : " << ret << endl);
		if (ret == 1) {
			DMESG(" returnStatistics is sent to client \n");DMESG(" ----------------------------------\n");
			/* Change Architecture */
			if (returnStatistics(socket)) {
				DMESG("return : -1" << endl);
				return -1;
			} else {
				DMESG("return : 1" << endl);
				return 1;
			}
		}
	} catch (SocketException &ex) {
		LOG(ex.what() << std::endl);
		return -1;
	}

	DMESG("Connection established from " << socket.getForeignAddress() << "\n");

	if (useSSL) {
		// The client requested SSL. Switch socket to SSL
		try {
			/* Change Architecture */
			socket.initSSLServer(authCtx.ctx, (char *) conMan);
		} catch (SocketException &ex) {
			LOG(ex.what() << std::endl);
			return -1;
		}DMESG("SSL_accepted" << std::endl);
	}
	// If the user requested SSL, this pointer is non-null
	SSL* ssl = socket.getSSLCtx();

/*
        if ( connection->dbConn->connAlive == false ) {
	  socket.send("901 DB Connection is broken\n\004\004");
          return -1;
        }
*/

	// Initialize the user context. Either by reading it from the client
	// or from the stored session.
	if (sessionID) {
		DMESG("Resuming session " << sessionID << "\n");
		/* Change Architecture */
		if (conMan->getSessionUser(sessionID, userCtx)) {
			LOG("Attempt to resume " << sessionID << " failed.\n");
			std::string response = "200 No such session\n\004\004";
			socket.send(response);
			return -1;
		}
		// TODO:
		// We would now need to make sure that the certificate with which
		// that session was established is the same as the one used now,
		// otherwise there is a tiny chance that an attacker can guess a
		// session ID and reestablish the SSL connection with a different
		// certificate, that is not checked.
		// This can be tested with a Python client using sessions but with
		// native SSL, that does not reestablish an SSL session.
		// Solution: Store certificate fingerprint in user session and
		// compare with current one.
	} else {
		// Read user name, permissions and related context information from
		// client.
		DMESG("Establishing context\n");
		try {
			establishUserContext(socket, userCtx);DMESG("Context established\n");
		} catch (SocketException &ex) {
			LOG("Error establishing context: " << ex.what() << endl);
			return -1;
		}

		DMESG("Authentication required: " << authCtx.requireAuthentication << std::endl);
		bool userAuthenticated = authenticateUser(ssl, connection, authCtx,
				userCtx);
		if (authCtx.requireAuthentication && !userAuthenticated) {
			LOG("Authentication error for user: " << userCtx.login
					<< " from " << socket.getForeignAddress() << std::endl);

                   if ( connection->dbConn->connAlive ) {

			socket.send("201 Could not authenticate user " + userCtx.login
					+ "\n\004\004");
			return -1;
                   } 
 /*
                   else {
			socket.send("901 DB Connection is broken\n\004\004");
			return -1;
                   }
 */
                       
		}
	}
	// Check whether maximum login per user is exceeded
	if (maxConnectsPerUser >= 1 && userCtx.login != "root") {
		/* Change Architecture */
		if (conMan->getUserConnections(userCtx.login) >= maxConnectsPerUser) {
			socket.send("202 Maximum connections exceeded for " + userCtx.login
					+ "\n\004\004");
			return -1;
		}
	}

	if (reservedConnections > 0 && userCtx.login != "root") {
		/* Change Architecture */
		if (conMan->getUsedConnections() + reservedConnections
				> conMan->getMaxConnections()) {
			socket.send("202 No more connections for unprivileged user "
					+ userCtx.login + "\n\004\004");
			return -1;
		}
	}


	// Set connection options
	connection->setUser(userCtx.login);
	connection->setDir(userCtx.currentDir);
	connection->setLogging(true);
	connection->setExecUser(execUserID);
	connection->setExecUserGroup(execUserGroup);
	connection->setExecRoot(execRoot);
	MDStandalone *sServer = dynamic_cast<MDStandalone *> (connection);
	if (sServer) {
		sServer->setPermissionMask(userCtx.permissionMask);
		sServer->setGroupRightsMask(userCtx.groupMask);
	}

	if (ssl) {
		X509 *cert = SSL_get_certificate(ssl);
		EVP_PKEY *pkey = X509_get_pubkey(cert);
		if (sServer) {
			sServer->setPublicKey(pkey);
			pkey = SSL_get_privatekey(ssl);
			sServer->setPrivateKey(pkey);
		}
		cert = SSL_get_peer_certificate(ssl);
		if (cert)
			connection->setDN(getCertSubjectName(cert, true));
	}
	return 0;
}

// SSL session callback funtions
int AMGAFrame::ssl_callback_NewSessionCacheEntry(SSL *ssl, SSL_SESSION *session) {

	DMESG("ssl_callback_NewSessionCacheEntry called" << std::endl);

	//MDConMan *conMan = (MDConMan *) SSL_get_app_data(ssl);

	unsigned char *sessionID = session->session_id;
	unsigned int idLen = session->session_id_length;

	int size = i2d_SSL_SESSION(session, NULL);
	DMESG("SSL Session size is: " << size << std::endl);

	const char *buf = new char[size];
	unsigned char *pp = (unsigned char *) buf;
	size = i2d_SSL_SESSION(session, &pp);

	LOG("Saving SSL session " << sessionToPrintable(sessionID, idLen)
			<< std::endl);

	/* Change Architecture */
	if (!AMGAFrame::pAMGAFrame->conMan->saveSessionSSL(buf, size, sessionID,
			idLen)) {
		ERRLOG("Could not save SSL session " << sessionToPrintable(sessionID, idLen) << std::endl);
	}

	delete[] buf;

	// Session is still valid
	return 0;
}

SSL_SESSION *AMGAFrame::ssl_callback_GetSessionCacheEntry(SSL *ssl,
		unsigned char *sessionID, int idLen, int *doCopy) {

	LOG("Resuming SSL session " << sessionToPrintable(sessionID, idLen) << std::endl);

	DMESG("ssl_callback_GetSessionCacheEntry called" << std::endl);


	/* Change Architecture */
	//MDConMan *conMan = (MDConMan *) SSL_get_app_data(ssl);

	char buf[AMGAFrame::pAMGAFrame->conMan->getMaxSessionSize()];
	int size;

	if (AMGAFrame::pAMGAFrame->conMan->getSessionSSL(sessionID, idLen, buf,
			size)) {
		ERRLOG("Session not found: " << sessionToPrintable(sessionID, idLen) << std::endl);
		return NULL;
	}

#if OPENSSL_VERSION_NUMBER >= 0x00908000L
	const unsigned char *data = (const unsigned char *) buf;
#else
	unsigned char *data = (unsigned char *) buf;
#endif
	SSL_SESSION *session = d2i_SSL_SESSION(NULL, &data, size);

	// We do not hold a further copy of the session, but OpenSSL needs
	// to increase reference count (*doCopy =0)
	*doCopy = 0;

	return session;
}

void AMGAFrame::ssl_callback_RemoveSessionCacheEntry(SSL_CTX *ctx,
		SSL_SESSION *session) {

	DMESG("ssl_callback_RemoveSessionCacheEntry called" << std::endl);

	/* Change Architecture */
	//MDConMan *conMan = (MDConMan *) SSL_CTX_get_app_data(ctx);

	unsigned char *sessionID = session->session_id;
	unsigned int idLen = session->session_id_length;

	LOG("Removing SSL session " << sessionToPrintable(sessionID, idLen) << std::endl);

	/* Change Architecture */
	if (AMGAFrame::pAMGAFrame->conMan->deleteSession(sessionID, idLen)) {
		DMESG("Removing session, session not found: " << sessionToPrintable(sessionID, idLen) << std::endl);
		return;
	}

	return;
}

//#ifdef ENABLE_MULTITHREAD
// KIMNG080415 print the ID and signal number of the thread to be called
void AMGAFrame::sig_handler(int signo) {
	DMESG("SIGNAL thid %d : %d\n" << pthread_self() << ", " << signo);
}

/**
 * Signal handler for SIGCHLD (child process changes state).
 *
 * Queries return status of terminated child processes,
 * otherwise zombie processes will occure.
 *
 * Remember that you cannot call a anything in MDConMan that
 * tries to get a lock from within a signal handler!
 */
void AMGAFrame::child_sig_handler(int sig) {

	/* Change Architecture */
	for (int i = 0; i < AMGAFrame::pAMGAFrame->numSlots; i++) {
		if (!AMGAFrame::pAMGAFrame->children[i])
			continue;
		if (waitpid(AMGAFrame::pAMGAFrame->children[i], NULL, WNOHANG)) {
			// The child will be removed laterfrom the list of processes
			// by calling conMan.remove() from the master process
			AMGAFrame::pAMGAFrame->deadChildren[i]
					= AMGAFrame::pAMGAFrame->children[i];
			// The following is not safe in a signal handler!
			//      DMESG("RECEIVED SIGCHILD: " << children[i] << std::endl);
			AMGAFrame::pAMGAFrame->children[i] = 0;
		}
	}
}

/**
 * Signal handler for SIGINT
 *
 */
void AMGAFrame::int_sig_handler(int sig) {
	/* Change Architecture */
	if (AMGAFrame::pAMGAFrame->conMan->isMaster()) {
		//    LOG("SIGINT received: please wait for shutdown..."<< std::endl);
		AMGAFrame::pAMGAFrame->terminateRequest = true;
	}
}

void AMGAFrame::hup_sig_handler(int sig) {
  if (AMGAFrame::pAMGAFrame->conMan->isMaster()) {
    if (AMGAFrame::logFile.size()) {
      // AMGAFrame::logStream.seekp(0);
      // Display::setOutput(logStream.rdbuf());
      Display::reset(AMGAFrame::logStream);
      LOG("SIGHUP : log rotating. "<< std::endl);
      DMESG("SIGHUP : log rotating. "<< std::endl);
    }
  }
}




/**
 * Signal handler for SIGUSR1
 *
 */
void AMGAFrame::child_term_handler(int sig) {
	/* Change Architecture */
	DMESG("SIGUSR1 received: " << getpid() << std::endl);
	AMGAFrame::pAMGAFrame->terminateRequest = true;
}
