/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "BDIIClient.h"
#include <stdlib.h>
#include "debug.h"


BDIIClient::LDAPException::LDAPException(const std::string &message, int e)
  : errString(message), err(e)
{
  if( err ) {
    char *p = ldap_err2string(err);
    if( p )
      errString.append(": ").append(p);
    else
      errString.append(": Unknown LDAP error");
  }
}


std::string BDIIClient::LDAPException::getError() const
{
  return errString;
}


/// Constructor of the BDIIClient class
BDIIClient::BDIIClient(const std::string &host, int port, bool dbg)
  throw(BDIIClient::LDAPException)
  : ldap(NULL), debug(dbg)
{
  who = "";
  cred = "";
  DMESG("BDII server: " << host << "  port: " << port << std::endl);
  reset(host, port);
}


BDIIClient::~BDIIClient()
{


}


int BDIIClient::reset(const std::string &host, int port)
  throw(BDIIClient::LDAPException)
{
  ldap = ldap_init(host.c_str(), port);
  if( !ldap )
    throw LDAPException("Cannot initialize LDAP connection", errno);
  return 0;
}


int BDIIClient::search(std::list< AMGA::Service > &services,
		       const std::string &baseDN,
		       const std::string &filter)
  throw(BDIIClient::LDAPException)
{
  DMESG("BDIIClient::search: baseDN:>" << baseDN << "<  filter >"
	<< filter << "<\n");

  if( !ldap )
    throw LDAPException("LDAP connection not initialized");
  
  int err;
  if( (err = ldap_simple_bind_s(ldap, who.c_str(), cred.c_str())) )
    throw LDAPException("Could not bind", err);

  // Do search
  LDAPMessage* res;
  if ( (err = ldap_search_s(ldap, baseDN.c_str(), LDAP_SCOPE_SUBTREE,
			   filter.c_str(), NULL, 0, &res)) ) {
    ldap_unbind_s(ldap);

    throw LDAPException("Error searching LDAP", err);
  }

  int n;
  // Check for entries
  if( !(n = ldap_count_entries(ldap, res)) ) {
    ldap_unbind_s(ldap);
    DMESG("No entries returned\n");
    return -1;
  }
  DMESG("Entries returned:" << n << std::endl);

  LDAPMessage *entry = ldap_first_entry(ldap, res);
  if( !entry )
    throw LDAPException("Error searching LDAP", err);
 
  do {
    char *ldapDN = ldap_get_dn(ldap, entry);
    AMGA::Service service(ldapDN);
    ldap_memfree(ldapDN);

    // Read attributes
    BerElement* ber = NULL;
    char *attr = ldap_first_attribute(ldap, entry, &ber);
    while(attr != NULL) {
      DMESG("     ATTR: " << attr);
      
      // Get attribute value
      char **values = ldap_get_values(ldap, entry, attr);

      // Print status if no values are returned (NULL ptr)
      if(values) {
	int n = ldap_count_values(values);
	if(!n)  {
	  DMESG("      : [BAD VALUE LIST]");
	} else {
	  // Output the first attribute value
	  DMESG("Value " << (*values) );
	  
	  if( !strcmp(attr, "GlueSEAccessProtocolEndpoint") )
	    service.host = *values;

	  if( !strcmp(attr, "GlueSEAccessProtocolPort") )
	    service.port = atoi(*values);

	  // Output more values if available
	  for (int i = 1; i < n; i++)
	    DMESG("More values: " << values[i]);
	}
      } else {
	DMESG("       [NO ATTRIBUTE VALUE RETURNED]");
      }

      if( values )
	ldap_value_free(values);

      ldap_memfree(attr);
      DMESG(std::endl);
      attr = ldap_next_attribute(ldap, entry, ber);

    } // Loop over attributes
    
    if(ber)
      ber_free(ber,0);

    services.push_back(service);

  } while( (entry = ldap_next_entry(ldap, entry)) );
  
  ldap_msgfree(res);

  if( (err=ldap_unbind_s(ldap)) )
    throw  LDAPException("Could not unbind", err);

  return 0;
}
