/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDClient.h"
#include "BDIIClient.h"

#include "configparser/ConfigParser.h"
#include "util/SecurityUtils.h"
#include "util/amgaProtocol.h"
#include "util/Socket.h"
#include "client/MDFed.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <errno.h>

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>

#include <fcntl.h>
#include <dlfcn.h>

#include <iostream>
#include <sstream>

using namespace std;

#include "debug.h"

#ifdef HAVE_SERVICEDISCOVERY_H
#include "ServiceDiscovery.h"
#endif

MDClient *MDClient::clientRegisteredForInt = NULL;
MDClient::MDClientLock MDClient::sslLock;

char MDClient::certKey[256];

/**
 * Signal handler for SIGINT
 *
 */
void int_sig_handler(int sig)
{
  // Re-register default handler
  signal(SIGINT, SIG_DFL);
//  std::cerr << "SIGINT received: please wait for shutdown..."
// 	    << std::endl;
  MDClient::clientRegisteredForInt->cancellationRequested=true;
}

MDClient *registerClient(MDClient *client)
{
  if( !client ) {
    signal(SIGINT, SIG_DFL);
    MDClient::clientRegisteredForInt = NULL;
    return NULL;
  }
  MDClient *old=MDClient::clientRegisteredForInt;
  MDClient::clientRegisteredForInt=client;
  // sighandler_t oldSIGINT=  
//  signal(SIGINT, int_sig_handler);

  static struct sigaction sig;
  memset(&sig, 0, sizeof(sig));
  sig.sa_handler = int_sig_handler;
  sigaction(SIGINT, &sig, NULL);
  
  return old;
}


int sslPasswdCallback(char *buf, int size, int rwflag, void *password)
{
  //  int debug = 1;
  //  DMESG("Returning password: >" << (char *)password << "<\n");
  strncpy(buf, (char *)(password), size);
  buf[size - 1] = '\0';
  return(strlen(buf));
}


MDClient::MDClientLock::MDClientLock()
{
  // Initialize OpenSSL library
  SSL_library_init();

  // Load error messages
  SSL_load_error_strings();

  // Load crypto algorithms
  OpenSSL_add_all_algorithms();

  locks = new pthread_mutex_t[CRYPTO_num_locks()];
  for(int i=0; i<CRYPTO_num_locks(); i++)
    pthread_mutex_init(&(locks[i]), NULL);

}


MDClient::MDClientLock::~MDClientLock()
{
  for(int i=0; i<CRYPTO_num_locks(); i++)
    pthread_mutex_destroy(&(locks[i]));
  delete [] locks;
}


void pthreads_locking_callback(int mode, int type, const char *file,
			       int line)
{
  if (mode & CRYPTO_LOCK) {
//    std::cout << "Locking: " << file << ":" << line << std::endl;
    pthread_mutex_lock(&(MDClient::sslLock.locks[type]));
  } else {
//    std::cout << "UNLocking: " << file << ":" << line << std::endl;
    pthread_mutex_unlock(&(MDClient::sslLock.locks[type]));
  }
}


/// Constructor of the MDClient class
/** Construct an MDClient which is configured to connect to \param
    host at \param port . If \param persistent is true, a peristent
    session will be created. The configuration file is read from the
    current directory, ~/.mdclient.config or the file on \param
    configFile if given.
    The constructor does not itself connect to the server, instead you
    need to call connectToServer() afterwards which also makes it
    possible to return error codes.
    MDClient is not thread-safe. Instead instantiate an
    MDClient per thread. The client will ensure thread-safe access
    to all global resources like the openssl library.
*/
MDClient::MDClient(const std::string &host, int port,
		   bool persistent, const std::string &configFile,
		   bool dbg)

  : debug(dbg), sock(NULL), ctx(NULL),  useSSL(false),
    connected(false), sendBufferMax(0),
    config(NULL), cancellationRequested(false), wasCancelled(false),
    sslSessionData(NULL), sessionID(0), sessionHost(""),
    sessionPort(0), sessionFile(-1), persistentSessions(false),
    sessionNumber(0)
{
  // Read configuration file
  char *c = getenv("HOME");
  std::string configHome;
  if(c) {
    configHome = c;
  } else {
    std::cerr << "HOME environment variable not set!" << std::endl;
    abort();
  } 
  configHome.append("/.mdclient.config");

  c = getenv("GLITE_LOCATION");
  std::string configDefault;
  if (c)
    configDefault = c;
  else
    configDefault = "";
  configDefault.append("/etc/mdclient.config");
  
  config=new ConfigParser();
  if(configFile.size()) {
    DMESG("Trying to read: " << configFile << std::endl);
    if(!config->init(configFile)) {
      ERRLOG("Could not read configuration: exiting" << std::endl);
      exit(5);
    }
  } else {
    DMESG("Trying to open ./mdclient.config\n");
    if(!config->init("mdclient.config")) {
      DMESG("Trying to open " << configHome << std::endl);
      if(!config->init(configHome)) {
	DMESG("Trying to open " << configDefault << std::endl);
	if(!config->init(configDefault)) {
	  ERRLOG("Could not read configuration: exiting" << std::endl);
	  exit(5);
	}
      }
    }
  }

  // config->print();

  persistentSessions = persistent;

  getConfiguredServer();

  getBDIIInfo();

  doServiceDiscovery();

  currentService = services.begin();

  mdfed = new MDFed(debug);
  mdfed->proxyState = false;
  mdfed->configF = configFile;
  mdfed->setParentMDClient(this);

  // Set session server
  setServer(host, port);

  // Initialize SSL if necessary
  requireSSL = false;
  bool trySSL = false;
  try {
    std::string s = config->getString("UseSSL");
    ConfigParser::toLower(s);
    if(s == "yes" || s == "require" || s == "1") {
      trySSL = true;
      requireSSL = true;
    } else if( s == "try") {
      trySSL = true;
    }

    if(! (s == "yes" || s == "1" || s=="require" || s == "try"
	  || s == "no" || s == "0")) {
      ERRLOG("Unknown value for option UseSSL\n");
      abort();
    }

    if(trySSL) {
      if(initSSL() && requireSSL) {
        ERRLOG("Could not initialise SSL\n");
        sslErrors();
        ERRLOG(error);
        abort();
      }
    }
  } catch (std::string err) { }
}


// Try configuration via BDII
int MDClient::getBDIIInfo()
{
  try {
    std::string bdiiServer = config->getString("BDIIServer");
    int bdiiPort;
    std::string baseDN;
    std::string filter;

    if( bdiiServer.length() == 0 ) return 0;

    try {
      bdiiPort=config->getInt("BDIIServerPort");
    } catch (std::string e) {
      bdiiPort = BDII_DEFAULT_PORT;
    }
    
    try {
      baseDN = config->getString("BDIIBaseDN");
    } catch (std::string e) { };

    try {
      filter = config->getString("BDIIFilter");
    } catch (std::string e) { };

    try {
      BDIIClient bdii(bdiiServer, bdiiPort, debug);
      DMESG("BDII initialized\n");
      bdii.search(services, baseDN.c_str(), filter.c_str());
      DMESG("BDII search done\n");
      if( debug ) {
	std::list< AMGA::Service >::const_iterator I = services.begin();
	DMESG("Servers found:\n");
	while( I!= services.end() ) {
	  DMESG("   " << I->dn << ": "
		<< I->host << "  port " << I->port << std::endl);
	  I++;
	}
      }
      
    } catch (BDIIClient::LDAPException e) {
      ERRLOG("BDII search failed: " << e.getError() << std::endl);
      return -1;
    }
  } catch (std::string e) { };

  return 0;
}


int MDClient::doServiceDiscovery()
{
# ifdef HAVE_SERVICEDISCOVERY_H

  // Open libglite-sd-c.so
  DMESG("Trying to load libglite-sd-c.so\n");
  void *handle = dlopen("libglite-sd-c.so", RTLD_LAZY);
  
  if( !handle ) {
    DMESG("  Failed: " << dlerror() << "\n");
    return 1;
  }

  // Load the symbols
  dlerror(); // Resets error
  typedef SDServiceList * (* listServices_t)(const char *type,
					     const char *site,
					     const SDVOList *vos,
					     SDException *exception);
  listServices_t sd_listServices = (listServices_t) dlsym(handle,
							  "SD_listServices");
  const char *dlsymErr = dlerror();
  if(dlsymErr) {
    DMESG("Could not load symbol SD_listServices: " << dlsymErr << '\n');
    dlclose(handle);
    return 1;
  }
  
  typedef void (* freeServiceList_t)(SDServiceList *serviceList);
  freeServiceList_t sd_freeServiceList 
    = (freeServiceList_t)dlsym(handle, "SD_freeServiceList");
  
  dlsymErr = dlerror();
  if(dlsymErr) {
    DMESG("Could not load symbol freeServiceList_t: " << dlsymErr << '\n');
    dlclose(handle);
    return 1;
  }

  SDServiceList *sl;
  // char *names[] = {"atlas", "cms"};
  char *names[1];
  names[0] = getenv("GLITE_SD_VO");

  SDVOList vos = {1, names}; // first param: size of list
  SDException exception;

  sl = sd_listServices("org.glite.AMGA", NULL, NULL, &exception);
  if( !sl ) {
    ERRLOG("No matching services found: " << exception.reason << '\n');
    dlclose(handle);
    return 2;
  }

  sd_freeServiceList(sl);

  // Close the library
  DMESG("Closing libglite-sd-c.so\n");
  dlclose(handle);
  
  return 0;

# else           // HAVE_SERVICEDISCOVERY_H
  return -1;
# endif          // HAVE_SERVICEDISCOVERY_H
}


int MDClient::getConfiguredServer()
{
  AMGA::Service service("ConfigurationFileDefault");
  try {
    service.host = config->getString("Host");
    try {
      service.port=config->getInt("Port");
    } catch (std::string e) {
      service.port = MD_DEFAULT_PORT;
    }
    
    services.push_back(service);
    
  } catch (std::string e) {
    return -1;
  }
  
  return 0;
}


int MDClient::initSSL()
{
  DMESG("initSSL called\n");

  // Initialize locking callbacks
  CRYPTO_set_id_callback((unsigned long (*)())pthread_self);
  CRYPTO_set_locking_callback(pthreads_locking_callback);
 
  OpenSSL_add_all_digests();
  // Creates client instance
  // The SSLv23_client_method() function's return value in Openssl 1.0 or higher is changed SSL_METHOD* to const SSL_METHOD*
  //SSL_METHOD *method= SSLv23_client_method();
  SSL_METHOD *method= (SSL_METHOD *)SSLv23_client_method();

  // Create SSL context
  ctx = SSL_CTX_new(method);

  if(!ctx)
    return -1;

  //  SSL_set_connect_state(ssl);
  // Disable buggy v2, enable all compatibility measures
  SSL_CTX_set_options(ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);

  if( strlen(certKey) ) {
    SSL_CTX_set_default_passwd_cb(ctx, sslPasswdCallback);
      
    SSL_CTX_set_default_passwd_cb_userdata(ctx, (void *)certKey);

    DMESG("Using certificate with a predefined password\n");
  }
  
  return 0;
}


int MDClient::initSSLSession()
{
  DMESG("initSSLSession\n");

  // If we were able to load a session to resume, we do not
  // load a certificate
  if(!sslSessionData) {
    try {
      if(config->getInt("AuthenticateWithCertificate")){ 
	int err = loadCertificate();
	if(err) {
	  ERRLOG("Could not load certificate\n");
	  return err;
	}
      }      
    } catch (std::string &err) {} ;
        
    // Verify server certificate
    try {
      verifyServerCert = config->getInt("VerifyServerCert");
    } catch (std::string &err) { verifyServerCert = true; }

    DMESG("Server verification set to " << verifyServerCert << '\n');

    if( verifyServerCert ) {
      SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, serverCertVerifyCallback);
      try {
	DMESG("Loading local CA certificates" << std::endl);
	if(!SSL_CTX_load_verify_locations(ctx, NULL, config->getString("TrustedCertDir").c_str())){
	  sslErrors();
	  ERRLOG("Error loading trusted certificates" << std::endl);
	}
      } catch (std::string &err) {
	ERRLOG("Could not load trusted certificates" << std::endl);
	return -1;
      }
    }
  }
  
  return 0;
}


int MDClient::setServer(const std::string &ipAddr, int port)
{
  std::string ipAddress = ipAddr;
  if(!ipAddr.size()) {
    if( currentService != services.end() ){
      DMESG("Using server " << currentService->dn << '\n');
      ipAddress = currentService->host;
      if( !port )
	port = currentService->port;
    } else {
      ipAddress = "localhost";
      DMESG("Setting default host: localhost\n");
    }
  }

  if( !port ){
    port = MD_DEFAULT_PORT;
    DMESG("Setting default port: " << port << '\n');
  }

  DMESG("Using server " << ipAddress << ":" << port << '\n');
  sessionHost = ipAddress;
  sessionPort = port;

  return 0;
}


/** Picks the next server from the list of servers in via BDII or the
    configuration file.
    Returns 0 if wraparound to first server occourred
    Returns 1 another server was found
    Return -1 if there are no servers
*/
int MDClient::nextServer()
{
  int ret = 1;

  currentService++;
  if( currentService == services.end() ) {
    ret = 0;
    currentService = services.begin();
  }

  setServer();

  if( currentService == services.end() )
    ret = -1;
  
  return ret;
}


void MDClient::useServer(const std::string &ipAddr, int port)
{  
  if( !port )
    port = MD_DEFAULT_PORT;

  sessionHost = ipAddr;
  sessionPort = port;
}


/// Establishes the actual connection to the server
/** This method can be called to establish the connection to the AMGA
    server for the first time or restablish it after a disconnect.
    Returns 0 on success or -1 if something went wrong. A description
    of the error is returned by the getError() method.
*/
int MDClient::connectToServer()
{  
  if( persistentSessions && !sessionID ) {
    if( restoreSessionData() )
      return -1;
  }

  DMESG("Connecting to server\n");
  if(ctx) {
    if(initSSLSession())
      return -1;
  }

  connected = false; // Whether we are able to talk to the server

  try {
    DMESG("Creating new socket\n");
    sock = new TCPSocket(sessionHost, sessionPort);
  } catch (SocketException &e) {
    std::stringstream s;
    s << "Could not open socket to " << sessionHost 
      << " port " << sessionPort << ": " << e.what();
    setError(s.str());
    return -1;
  }

  try { 
    sock->setTCPNoDelay();
  } catch (SocketException &e) { };

  bool resumeSession = false;
  if(initConnection(resumeSession)){
    setError("Could not initialize connection");
    return -1;
  }

  // Use SSL only if configured for it and server asks
  if(requireSSL && !useSSL) {
    setError("Could not initialize required SSL session");
    return -1;
  }

  if(useSSL && initSSLConnection(resumeSession) )
    return -1;

  if(!resumeSession && establishContext() ) {
    setError("Could not initialize Context");
    return -1;
  }

  connected=true;
  return 0;
}


/// Destructor of MDClient
/** This destructor releases all resources held by the MDClient class
    after disconnecting from the server and saving the session if so
    configured. You do not need to do this manually.
*/
MDClient::~MDClient()
{
  DMESG("~MDClient called\n");
  if(!connected && sslSessionData) { // We have an open session
    DMESG("Not connected, so freeing session data\n");
    SSL_SESSION_free(sslSessionData);
    sslSessionData = NULL;
  }

  if(!sessionID && persistentSessions)
    execute("close");

  mdfed->disconnectFed(persistentSessions);
  disconnect(persistentSessions);

  if(ctx)
    SSL_CTX_free(ctx);
  ctx = NULL;

  if(sessionFile >-1)
    close(sessionFile);

  delete config;

  delete mdfed;
}


int MDClient::saveSessionData()
{
  DMESG("In saveSessionData\n");
  if(!sessionID) {
    DMESG("SessionID is 0, not saving session data to file\n");
    return 0;
  }

  // Try to open (and if needed create) a session File
  // if not yet opened
  if(sessionFile < 0) {
    std::stringstream name;
    name << "/tmp/md_" << getuid();
    name << "_" << sessionHost << "_" << sessionPort
	 << "_" << sessionNumber;
    DMESG("Saving session data to file " << name.str() << "\n");

    sessionFile = open(name.str().c_str(), O_RDWR|O_CREAT,
		       S_IRUSR | S_IWUSR);
    if(sessionFile >= 0) {
      if ( !flock(sessionFile, LOCK_EX)) {
      //if( !lockf(sessionFile, F_LOCK, 0) ) {
	DMESG("Locked: " << name.str() << std::endl);
	sessionFileName = name.str();
      } else {
	DMESG("Failed to lock " << name.str() << std::endl);
	close(sessionFile);
	sessionFile=-1;
      }
    }
  }

  if(sessionFile < 0)
    return sessionFile;

  int32_t size = 0;
  char *buf = NULL;
  if( sock->isSSLSocket() ) {
    DMESG("Saving SSL session\n");
    size = i2d_SSL_SESSION(sslSessionData, NULL);
    buf = (char *)malloc(size*sizeof(char));
    unsigned char *pp=(unsigned char *)buf;
    size=i2d_SSL_SESSION(sslSessionData, &pp);
  }

  if(write(sessionFile, &sessionID, 8) != 8)
    goto cleanup_save;

  if(write(sessionFile, &size, 4) != 4)
    goto cleanup_save;

  if( size ) {
    if(write(sessionFile, buf, size) != size)
      goto cleanup_save;
  }

  if(close(sessionFile))
    goto cleanup_save;

  free(buf);

  close(sessionFile);
  sessionFile = -1;

  return 0;

 cleanup_save:
  delete [] buf;
  close(sessionFile);
  sessionFile = -1;
  unlink(sessionFileName.c_str());

  return -1;
}


int MDClient::restoreSessionData()
{
  DMESG("restoringSessionData called\n");

  sessionID = 0;
  sslSessionData = NULL;
  std::stringstream name;
  name << "/tmp/md_" << getuid();
  name << "_" << sessionHost << "_" << sessionPort
       << "_" << sessionNumber;
  if( sessionFile <0 ) {// Else:File is already opened and locked
    DMESG("Opening file " << name.str() << std::endl);
    sessionFile = open(name.str().c_str(), O_RDWR);
  } else {
    DMESG("Reusing open session file\n");
  }
  if(sessionFile >= 0) {
    DMESG("Opened " << name.str() << std::endl);
    if ( !flock(sessionFile, LOCK_EX)) {
      // lockf is not BSD compliant, therefore it does not compile
      // on CYGWIN
      //if( !lockf(sessionFile, F_LOCK, 0) ) {
      DMESG("Locked: " << name.str() << std::endl);
      sessionFileName = name.str();
    } else {
      DMESG("Failed to lock " << name.str() << std::endl);
      close(sessionFile);
      sessionFile=-1;
    }
  }

  // It is not an error to not be able to restore a session
  if(sessionFile < 0)
    return 0;
  
  DMESG("Restoring session\n");

  int err = -1;

  uint32_t size;
  unsigned char *data;
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
  const unsigned char *buff;
#else
  unsigned char *buff;
#endif

  if(read(sessionFile, &sessionID, 8) != 8) {
    DMESG("No session found!" << std::endl);
    err = 0;
    goto read_failed;
  }

  DMESG("Restoring session ID " << sessionID << std::endl);

  if(read(sessionFile, &size, 4) != 4)
    goto read_failed;
  
  err = 0;

  DMESG("SSL Session size " << size << std::endl);

  // Also check whether SSL enabled, the SSL session could be
  // a leftover from when it was enabled
  if( size && ctx ) { 
    data = (unsigned char *)malloc(size);

    if(read(sessionFile, data, size) != size){
      free(data);
      goto read_failed;
    }
    
    buff = data;
    sslSessionData = d2i_SSL_SESSION(NULL, &buff, size);
    if(sslSessionData) {
      err = 0;
      DMESG("Successfully restored session " << sessionID
	    << " from file\n");
    } else {
      DMESG("Failed to restore session: File unreadable\n");
      setError("Session File corrupted: ");
      sslErrors();
      err = -1;
    }
    free(data);
  }

  ftruncate(sessionFile, 0);
  lseek(sessionFile, 0, SEEK_SET);

  return err;

 read_failed:
  if( err )
    setError("Session file unreadable");
  sessionID = 0;
  sslSessionData = NULL;
  ftruncate(sessionFile, 0);
  close(sessionFile);
  sessionFile = -1;
  return err;
}


/// Disconnects from the AMGA server
/** Disconnects from the AMGA server and stores all data necessary to
    resume the session in a session file
*/
int MDClient::disconnect(bool saveSession)
{
  DMESG("MDClient::disconnect\n");

  if(!connected)
    return 0;

  connected=false;

  if(saveSession) {
    DMESG("Saving session, persistent: " 
	  << persistentSessions << std::endl);
    if( sock->isSSLSocket() ){
      sslSessionData = sock->getSSLSession();
      if(persistentSessions)
	saveSessionData();
      sock->sslShutdown();
    } else {
      if(persistentSessions)
	saveSessionData();
    }
  }

  if( sock ){
    delete sock;
    sock = NULL;
    DMESG("Socket deleted\n");
  }

  return 0;
}

/// This does not work!
int MDClient::setTimeout(int seconds)
{
  timeout=seconds;
  // Propagate timeout to server

  return 0;
}


/// Execute an AMGA command without waiting for the server error status
/** Executes an AMGA command without waiting for a server response
    which contains the error status. This is important for commands
    which stream data to the server where a response is only sent at
    the time when the command is commited like the upload command.
    The method returns 0 on success, a positive AMGA error code when
    the command execution failed or a negative error code if the
    MDClient had problems contacting the server or something went
    wrong internally.
    The command is given as a string in \param command .
*/
int MDClient::execNoWait(const std::string &command)
{
  if(!connected)
    connectToServer();

  DMESG("execNoWait >" << command << "<" << std::endl);
  if(prepareExec())
    return -1;

  std::string c(command);
  AMGA::encodeLine(c);
  c.append("\n");
  if(c.size()){
    if(sendLine(c, !sendBufferMax)) {
      return -1;
    }
  }
  if(sock->isReadyToRead())
    return retrieveResult();

  return 0;
}


/** Retrieves the result of the remote call. The result is the first
    line sent by the server. If an error occurred, the rest of the
    data sent by the server is read until EOT is found.
*/
int MDClient::retrieveResult()
{
  // Get return code in first response row
  std::string returnCode;
  endOfTransmission=false;
  response = "";
  int res=fetchRow(returnCode);
  DMESG("res " << res << "  returnCode " << returnCode << std::endl);
  if (errno == EINTR) {
    return 12;
  }
  if(res < 0) {// Connection closed because of quit
    return -1;
  }

  char *eptr;
  int ret=strtol(returnCode.c_str(), &eptr, 10);
  if(strlen(eptr) > 0)
    error=eptr+1;
  else
    error = "";

  if(eptr == returnCode.c_str()) {
    setError("Illegal server response: " + returnCode);
    ret = -1;
  }

  /* Proxy & Federation Handling */
  if( ret == PROXY_CODE ) {
    std::string errmsg;

    mdfed->clearFedHosts() ;
    return mdfed->initFed(errmsg, false); 
  } 

  // Be sure to read until the end of transmission, in case of error
  // do this as long as endOfTransmission is not set, otherwise get at
  // least a byte to make sure EOT is really read.
  if(res || ret) { // Be sure to read until the end of transmission
    DMESG("IN ERROR TESTING"<< std::endl);
    while(!endOfTransmission) {
      if(fetchRow(returnCode)<0) 
        break;
    }
  } else {
    if(!endOfTransmission)
      if(fetchData() < 0)
        ret = -1;
  }

  // Return errors from the execution
  if(ret)
    return ret;

  char *eptr2;
  resultCols=strtol(eptr, &eptr2, 10);
  if(eptr2==eptr)
    resultCols=-1;

  if(resultCols>-1)
    DMESG("Result Columns: " << resultCols << std::endl);

  DMESG("OK executing" << std::endl);
  return res;
}


int MDClient::prepareExec()
{
  assert(!response.size());
  if(wasCancelled)
    error.clear();
  wasCancelled=false;
  cancellationRequested=false;
  if( !isConnected() )
    return -1;
  return 0;
}


/// Executes an AMGA command
/** Executes an AMGA command \param command and returns the error code
    from the AMGA server (positive values). A negative return code
    means problems with the connection or internal problems.
*/
int MDClient::execute(const std::string &command)
{
  if(!connected)
    connectToServer();

  mdfed->lastCommand = command;
  mdfed->proxyState = false;

  std::string c(command);
  if(command=="fed_check_failure"){
    ostringstream query;
    query << " " << mdfed->failureCode ;
    c.append(query.str());
  }

  DMESG("execute >" << c << "<" << std::endl);
  if(prepareExec())
    return -1;
  AMGA::encodeLine(c);
  c.append("\n");
  if(c.size()){
    if(sendLine(c)) {
      return -1;
    }
  }

  if(command=="quit")
    return 0;

  return retrieveResult();
}


int MDClient::sendLine(const std::string &c, bool flush)
{
  const char *buff=c.c_str();
  unsigned int size=c.size();

  if(!flush) {
    sendBuffer.append(c);
    if(sendBuffer.size()<sendBufferMax)
      return 0;
    buff=sendBuffer.c_str();
    size=sendBuffer.size();
  } else {
    if(sendBuffer.size()) {
      sendBuffer.append(c);
      buff=sendBuffer.c_str();
      size=sendBuffer.size();
    } else {
      buff=c.c_str();
      size=c.size();
    }
  }

  try {
    sock->send(buff);
  } catch (SocketException e) {
    setError( ((std::string)"Could not send command: ") + e.what());
    return -1;
  }

  sendBuffer.clear();

  return 0;
}


//! Reads a row from the buffer, if necessary the buffer is first filled
//! by reading from the server
/** The line is a reference to a string which gets filled with the row
    read from the buffer or if necessary from the network. The
    readAhead option which defaults to true makes sure that a
    subsequent call to eot() is meaningful. Only if the caller knows
    that no EOT character can be at the end of the line read this
    option can be set to false. This is done internally during the
    handshaking where empty lines are EOT delimiters.
**/
int MDClient::fetchRow(std::string &line, bool readAhead)
{
  int res = 0;
  size_t pos=response.find_first_of("\n");
  if(pos!=std::string::npos){
    line=response.substr(0, pos);
    AMGA::decodeLine(line);
    response=response.substr(pos+1);
    // Read ahead to know if we have got the last line so that
    // eot() can tell the client
    if(readAhead && !response.size() && !endOfTransmission){
      if(fetchData() < 0)
        return -1;
    }
    // Check whether an additional error has occurred
    size_t canpos = line.find_first_of("\030");
    if( canpos != std::string::npos ) {
      char *eptr;
      std::string e = line.substr(canpos+1);
      res=strtol(e.c_str(), &eptr, 10);
      if(strlen(eptr) > 0)
        error=eptr+1;
      else
        error = "";
    }

    return res;
  }

  if( mdfed->proxyState ) { 
    return mdfed->fetchRow(line, readAhead);
  }

  if(endOfTransmission){
    if( wasCancelled )
      return 0;
    error="No more data.";
    return -1;
  }

  int bytes=fetchData();
  if( bytes < 0 || (!bytes && !wasCancelled) )
    return -1;

  if( !bytes && wasCancelled )
    return 0;

  return fetchRow(line, readAhead);
}


void MDClient::setError(std::string err){
  if ( error.size() )
    error = err + ": " + error;
  else
    error = err;
}


/** Does the actual reading operation from the socket or from the SSL
    connection appends to the response buffer
*/
int MDClient::doRead()
{
  char buff[1001];
  int bytes;

  try {
    DMESG("Before read" << std::endl);
    bytes = sock->recv(buff, 1000);
    if (bytes <=0)
      return bytes;

    buff[bytes]='\0';
    response.append(buff);
  } catch (SocketException& e) {
    if (e.errorNumber() == EINTR) {
      return -2;
    } else {
      return -1;
    }
    setError(((std::string)"Read failed: ") + e.what());
  }
  return bytes;
}


/** Fetches more data from the server until a full line is in the
    buffer or an EOT is detected.
*/
int MDClient::fetchData()
{
  DMESG("fetchData 1 >" << response << "<" << std::endl);
  int bytes = response.size();
  size_t pos = std::string::npos;
  while(1){
    if(cancellationRequested) {
      DMESG("Got cancellation request"<< std::endl);
      cancel();
      setError("Request cancelled");
      return 0;
    }
    // Look for EOT
    if( !endOfTransmission && (pos=response.find('\004')) != std::string::npos) {
      DMESG("Found EOT\n");
      if(protocolVersion > 1) {
        // Need 2 EOTs, make sure that we do read also the second one
        while(pos + 1 >= response.size() or response.find('\004', pos + 1) == std::string::npos){
          bytes = doRead();
          DMESG("Got: " << bytes << endl);
          cout << errno << endl;
          if(bytes==-2)
            continue;
          if (bytes <=0)
            break;
        }
      }
      break;
    }
    // Look for newline
    if( (pos=response.find_first_of("\n")) != std::string::npos)
      break;
    // Read more data
    bytes = doRead();
    if(bytes==-2)
      continue;
    DMESG("fetchData fetched " << bytes << std::endl);
    if(bytes <= 0)  // Read error or connection closed(0)
      break;
  }
  // Check whether we are at the end of the transmission
  // Read session handle if provided and shutdown connection if there
  if(pos != std::string::npos && response[pos] == '\004') {
    DMESG("EOT found >"<< response << "<" << std::endl);
    sessionID = 0;
    if( pos<response.size()-8 && response.substr(pos+1, 7) =="session" ) {
      sessionID=strtoull(response.c_str()+pos+8, NULL, 10);
      DMESG("Session ID " << sessionID << std::endl);
      disconnect(true);
    }
    response.resize(pos);
    endOfTransmission=true;
    if(bytes>0)
      bytes--;
    return bytes;
  }

  if(bytes < 0)
    return -1; // Read error. Error set by doRead();

  if(bytes==0){
    setError("Connection closed by server");
    return -1;
  }
  return bytes;
}


/// Checks whether the reply to a query was recieved fully
/** Returns true if the entire reply to a quey issued by execute() was
    recieved. As long as eot() returns false you need to call
    fetchRow(), even after a cancel() has been sent to flush the
    send-buffer of the server.
*/
bool MDClient::eot()
{
  // std::cout << "EOT >" << response << "<" << std::endl;
  if(response.size()) return false;

  if ( mdfed->proxyState ) {
    return mdfed->eot();
  } 

  if(endOfTransmission) return true;
  if(fetchData() <=0 )
    return true;
  return !response.size();
}


std::string printBytes(unsigned char *s, unsigned int len)
{
  std::stringstream asc;
  for (size_t i=0; i<len; i++){
    asc << " " << s[i] << " ";
    asc.width(2); asc.fill('0');
    asc << std::hex << ((unsigned int)s[i]);
  }
  return asc.str();
}


/// Sends an out-of-band cancel message to the server
/** Send a cancel request to the server. Note that this can be
    honoured by the server if results are streamed back to the
    client. You need to continue to call fetchRow() until eot()
    returns true.
*/
int MDClient::cancel()
{
  assert(!wasCancelled);
  cancellationRequested = false;
  wasCancelled = true;
  DMESG("SENDING CANCEL\n");
  char abortChar='\030'; // CAN byte = 0x18
  try {
    sock->sendOutOfBand(&abortChar, 1);
  } catch (SocketException e) {
    setError(((std::string)"Could not send cancel") + e.what());
    return -1;
  }
  // Flush buffer
  DMESG("Flushing buffer\n");
  while(!endOfTransmission){
    if( fetchData()<=0 )
      break;
    response.clear();
  } 
  response.clear();
  DMESG("Leaving cancel\n");
  return 0;
}


/** Initializes a connection with the server
    Needs an already establshed TCP connection and will read the
    greeting from the server, then requesting the authentication and
    security method which was selected by the user. If the server
    answers with OK the connection is established. If a session
    context exists, the client will only read the greeting and then
    send a resume command. @resumeSession will be set to true.
*/
int MDClient::initConnection(bool &resumeSession)
{
  DMESG("initConnection called" << std::endl);

  std::string greetings;
  endOfTransmission=false;
  useSSL = false;

  int res=fetchRow(greetings, false);
  if(res < 0) // Connection closed because of quit
    return -1;
  greeting.append(greetings);

  std::string version;
  res=fetchRow(version, false);
  if(res < 0) // Connection closed because of quit
    return -1;
  protocolVersion = strtoul(version.c_str()+9, NULL, 10);

  std::string options;
  res=fetchRow(options, false);
  if(res < 0) // Connection closed because of quit
    return -1;

  bool requestSSL=false;
  if(sessionID) { // We have a session to resume
    resumeSession = true;
    char sBuffer[60];
    if(sslSessionData) {
      DMESG("Resuming SSL session\n");
      sprintf(sBuffer, "resumeSSL%llu\n\n", sessionID);
      if(sendLine(sBuffer)) {
	setError("Could not request SSL reconnection");
	return -1;
      }
      requestSSL = true;
    } else {
      DMESG("Resuming non-SSL session\n");
      sprintf(sBuffer, "resume%llu\n\n", sessionID);
      sendBuffer.append(sBuffer);
      return 0;
    }
  } else { // Initiate new connection
    DMESG("No session to resume\n");
    // Should do parsing of options, here
    if(ctx) { // Check for SSL requested
      if(options.find("ssl")!=std::string::npos)
	requestSSL=true;
    }
    if(requestSSL) {// Request SSL connection
      if(sendLine("ssl\n\n")) {
	setError("Could not request SSL connection");
	return -1;
      }
    } else {
      if(sendLine("plain\n\n")) {
	setError("Could not request plain connection");
	return -1;
      }
    }
  }

  // Wait for "OK\n"
  std::string ok;
  res=fetchRow(ok, false);
  if(res < 0) // Connection closed because of quit
    return -1;

  useSSL = requestSSL;
  DMESG("useSSL: " << useSSL <<  std::endl);

  return 0;
}


//! Returns the greeting string of the server
const std::string &MDClient::getGreeting()
{
  return greeting;
}


/** Initializes a SSL connection.
    Needs an existing TCP connection as established with connect() and
    initConnection().
    If @resumeSession is false a new ssl session is started.
*/
int MDClient::initSSLConnection(bool resumeSession)
{
  DMESG("Initializing SSL Connection" << std::endl);

  try {

    if (resumeSession) {
      sock->initSSLClient(ctx, sslSessionData);
    } else {
      sock->initSSLClient(ctx);

      if( verifyServerCert ) {
	DMESG("Verifying server certificate\n");
	bool ignoreNameMismatch = false;
	try {
	  ignoreNameMismatch 
	    = config->getInt("IgnoreCertificateNameMismatch");
	} catch (std::string &err) {};
	if( sock->verifyServerCertificate(ignoreNameMismatch) ){
	  std::stringstream s;
	  s << "Server name " << sock->getForeignName()
	    << " does not match certificate DN " << sock->getPeerDN();
	  setError(s.str());
	  return -1;
	}
      }
    }

  } catch (SocketException e) {
    std::stringstream s;
    s << "Could not initSSLConnection: " << e.what();
    setError(s.str());
  }

  DMESG("Server name " << sock->getForeignAddress()
	<< " uses certificate with DN " << sock->getPeerDN() << '\n');

  return 0;
}


int MDClient::establishContext()
{
  std::string login("NULL");
  try {
    login=config->getString("Login");
  } catch (std::string err) {} ;

  std::string permMask("rw-");
  try {
    permMask=config->getString("PermissionMask");
  } catch (std::string err) {} ;

  std::string groupMask("r--");
  try {
    groupMask=config->getString("GroupMask");
  } catch (std::string err) {} ;

  std::string home("/");
  try {
    home=config->getString("Home");
  } catch (std::string err) {} ;

  std::string name("Anonymous");
  try {
    name=config->getString("Name");
  } catch (std::string err) {} ;

  std::string pass("");
  try {
    pass=config->getString("Password");
  } catch (std::string err) {} ;

  std::string line="0 ";
  line.append(login);
  line.append("\n1 ").append(name);
  line.append("\n2 ").append(permMask);
  line.append("\n3 ").append(groupMask);
  line.append("\n4 ").append(home);
  if(pass.size())
    line.append("\n5 ").append(pass);
  line.append("\n\n");

  if(sendLine(line)){
    setError("Could not set context");
    return -1;
  }

  return 0;
}


int MDClient::readConfiguration(std::list<std::string> &locations)
{
  DMESG("Reading configuration file" << std::endl);
  std::list<std::string>::iterator I=locations.begin();
  while(I!=locations.end()){
    if(config->init(*I))
      break;
    I++;
  }
  if(I==locations.end())
    return -1;

  return 0;
}


int MDClient::loadCertificate()
{
  DMESG("Loading certificates" << std::endl);
  std::string certFile;
  std::string keyFile;

  try {
    if(config->getInt("UseGridProxy")) {
      certFile=getProxyCertName();
      keyFile=certFile;
    }
  } catch (std::string err) { };

  if(!certFile.size()) {
    try{
      certFile=config->getString("CertFile");
      keyFile=config->getString("KeyFile");
    } catch (std::string err) { };
  }

  if(certFile.size()) {
    // Load local certificate from file
    if(SSL_CTX_use_certificate_chain_file(ctx, certFile.c_str())!=1){
      setError("Could not load client certificate");
      sslErrors();
      return -1;
    }
    // Load private key from local file
    if(SSL_CTX_use_PrivateKey_file(ctx, keyFile.c_str(), SSL_FILETYPE_PEM)!=1){
      setError("Could not load private key");
      sslErrors();
      return -1;
    }
  }

  // Verify private key
  if ( !SSL_CTX_check_private_key(ctx) ) {
    setError("Could not verify private key");
    sslErrors();
    return -1;
  }
  return 0;
}


void MDClient::sslErrors()
{
  char buff[201];
  BIO *mem=BIO_new(BIO_s_mem());
  ERR_print_errors(mem);
  while (!BIO_eof(mem)) {
    BIO_gets(mem, buff, 200);
    //    error.append("\n");
    error.append(buff);
  }
  BIO_free(mem);
}
