/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: md_api.cc,v 1.10 2010/04/14 07:15:30 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "md_api.h"

#include "MDClient.h"

#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
// #define _XOPEN_SOURCE /* glibc2 needs this */
#include <time.h>
#include <errno.h>

#if HAVE_XATTR_H
#include <attr/xattr.h>
#endif

#define MD_ERR_OK 0
#define MD_ERR_NOENT 1
#define MD_ERR_ILCMD 3
#define MD_ERR_PERM 4
#define MD_ERR_NOKEY 7
#define MD_ERR_INTERNAL 9
#include <iostream>

int translateError(int amga)
{
  switch(amga) {
  case MD_ERR_OK: return 0;
  case MD_ERR_PERM: return EPERM;
  case MD_ERR_NOENT: return ENOENT;
  case MD_ERR_ILCMD: return EIO;
  case MD_ERR_INTERNAL: return EIO;
#if HAVE_XATTR_H  
  case MD_ERR_NOKEY: 
    return ENOATTR;
#endif
  default: 
    return EIO;
  }
}


//! A simple interface to access the attributes of an entry
/** This command sends the getattr metadata interface command to a
    metadata server and returns the value of a single key for a
    given files.
    \param file Name of the file
    \param key The key name
    \param value The value of the file which is filled
    \return Error value (including errors of the \ref getattr interface command):
    - 0 in case of everything went OK,
    - -1 if something went wrong with the client-server connection
    - MD_ERR_NOENT: No such file or directory
    - MD_ERR_PERM:  Permission denied. You need read permission on the
                    file
 */
int getAttr(const std::string &file, const std::string &key,
	    std::string &value)
{
  MDClient client;

  return getAttr(client, file, key, value);
}


//! getAttr with external handling of the connection
int getAttr(MDClient &client, const std::string &file,
	    const std::string &key, std::string &value)
{
  std::string command="getattr ";
  command.append(file).append(" ").append(key);
  int res;
  if((res=client.execute(command)))
    return res;
  
  client.fetchRow(value); // name of file

  return client.fetchRow(value);
}


static int permsToInt(const std::string &perms)
{
  int p = 0;
  if( perms.size() < 3)
    return p;

  if( perms[0] == 'r' )
    p |= 4;
  if( perms[1] == 'w' )
    p |= 2;
  if( perms[2] == 'x' )
    p |= 1;

  return p;
}


//! Sets one ore more attributes of a file
/** The attributes are given by the caller as a list of strings,
    which needs to match the length of the values to be assigned. The
    keys need to be defined with addAttr() first.
    \param file The file name of the entry
    \param keys A list of keys
    \param values The list of values which are assigend to the keys
    \return Error code (including errors returned by the \ref setattr
    metadata interface command):
    - 0 OK
    - -1 Client-Server connection problems
    - MD_ERR_NOENT: No such file or directory
    - MD_ERR_ILCMD: Illegal command: keys and values size mismatch
    - MD_ERR_PERM:  Permission denied. You need write permission on the
                    entry
*/
int amgaStat(MDClient &client, const std::string &file,
	     struct stat *stbuf)
{
  std::string command = "stat ";
  command.append(file);

  int res = 0;
  if((res=client.execute(command))){
    errno = translateError(res);
    return -1;
  }
  std::string v;
  bool isEntry = false;

  // Name
  res = client.fetchRow(v);
  if( res ){
    errno = translateError(res);
    return -1;
  }

  // Type
  res = client.fetchRow(v);
  if( res ){
    errno = translateError(res);
    return -1;
  }
  if( v == "entry" ) {
    isEntry = true;
    stbuf->st_mode = S_IFREG;
    stbuf->st_nlink = 1;
  } else {
    stbuf->st_mode = S_IFDIR;
    stbuf->st_nlink = 2;
  }

  // Owner permissions
  res = client.fetchRow(v);
  if( res ) return res;
  //  std::cout << "Owner perms: >" << v << "< :" << permsToInt(v) << "\n";
  stbuf->st_mode |= permsToInt(v) << 6;

  if( isEntry ){
    // Group permissions
    res = client.fetchRow(v);
    if( res ){
      errno = translateError(res);
      return -1;
    }
    stbuf->st_mode |= permsToInt(v) << 3;
    //    std::cout << "Group perms: >" << v << "< :" << permsToInt(v) << "\n";
  }

  // Owner
  res = client.fetchRow(v);
  if( res ){
    errno = translateError(res);
    return -1;
  }
  // We need to translate the UID here, somehow
  char buff[1025];
  struct passwd pwbuf;
  struct passwd *pwbufp;
  if( !getpwnam_r(v.c_str(), &pwbuf, buff, 1024, &pwbufp) && pwbufp){
    stbuf->st_uid = pwbufp->pw_uid;
  }
  
  if( isEntry ) {
    // File size
    res = client.fetchRow(v);
    if( res ){
      errno = translateError(res);
      return -1;
    }
    stbuf->st_size = atoi(v.c_str());

    // Creation time
    res = client.fetchRow(v);
    if( res ){
      errno = translateError(res);
      return -1;
    }
    // Format: 1999-12-22 04:05:06
    struct tm t;
    strptime(v.c_str(), "%F %T", &t);

    stbuf->st_atime = stbuf->st_mtime = stbuf->st_ctime = mktime(&t);

    // Guid
    res = client.fetchRow(v);
    if( res ){
      errno = translateError(res);
      return -1;
    }

    // Expires
    res = client.fetchRow(v);
    if( res ){
      errno = translateError(res);
      return -1;
    }

    
    // Link
    res = client.fetchRow(v);
    if( res ){
      errno = translateError(res);
      return -1;
    }
    if( v.size() ) {
      stbuf->st_mode |= S_IFLNK;
      
    }
  }      

  while( !client.eot() && !( res=client.fetchRow(v) ) ) {};

  return res;
}


int amgaReadlink(MDClient &client,
		 const char *path, char *buf, size_t size)
{
  std::string command = "stat ";
  command.append(path);

  int res = 0;
  if((res=client.execute(command))){
    errno = translateError(res);
    return -1;
  }

  std::string v;

  buf[0] = '\0';
  for( int i = 0; i< 9; i++) {
    res = client.fetchRow(v);
    if (res){
      errno = translateError(res);
      return -1;
    }
  }

  // Finally read the link
  res = client.fetchRow(v);
  if( res ) {
    errno = translateError(res);
    return -1;
  }
  strncpy(buf, v.c_str(), size);

  int resSize = v.size();

  while( !client.eot() && !( res=client.fetchRow(v) ) ) {};
  
  if( res ) {
    errno = translateError(res);
    return -1;
  }
  return resSize;
}


//! Sets one ore more attributes of a file
/** The attributes are given by the caller as a list of strings,
    which needs to match the length of the values to be assigned. The
    keys need to be defined with addAttr() first.
    \param file The file name of the entry
    \param keys A list of keys
    \param values The list of values which are assigend to the keys
    \return Error code (including errors returned by the \ref setattr
    metadata interface command):
    - 0 OK
    - -1 Client-Server connection problems
    - MD_ERR_NOENT: No such file or directory
    - MD_ERR_ILCMD: Illegal command: keys and values size mismatch
    - MD_ERR_PERM:  Permission denied. You need write permission on the
                    entry
*/
int setAttr(const std::string &file,
	    const std::vector< std::string> &keys,
	    const std::vector< std::string> &values)
{
  MDClient client;
  
  return setAttr(client, file, keys, values);
}


//! Version of setAttr with external client management
int setAttr(MDClient &client, const std::string &file,
	    const std::vector< std::string> &keys,
	    const std::vector< std::string> &values)
{
  if(keys.size() != values.size())
    return 3;

  std::string command="setattr ";
  command.append(file).append(" ");
  std::vector< std::string>::const_iterator I;
  std::vector< std::string>::const_iterator J=values.begin();
  for(I=keys.begin(); I!=keys.end(); I++){
    command.append(" ").append(*I).append(" '");
    command.append(*J++).append("'");
  }
  return client.execute(command);
}


//! Lists the attributes of a file
/** For a given file this functioin returns a list of attributes of a
    directory and their types. Note that also attributes which are
    undefined (assigned to NULL) for the entry are listed.
    \param file Entry name
    \param attrList Returns the attributes' names as a list of strings
    \param types Returns the list of associated types as strings
    \return Error value (including errors returned by the
    \ref listattr metadata interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: No such file or directory
    - MD_ERR_PERM: Permission denied. You need read permission for the entry.
*/
int listAttr(const std::string &file,
	     std::list<std::string> &attrList,
	     std::list<std::string> &types)
{
  MDClient client;
  
  return listAttr(client, file, attrList, types);
}

//! Version of listAttr with external client management
int listAttr(MDClient &client, const std::string &file,
	     std::list<std::string> &attrList,
	     std::list<std::string> &types)
{
  std::string command="listattr ";
  command.append(file);
  int res;
  if((res=client.execute(command)))
    return res;
  
  while(!client.eot()){
    std::string attr, type;
    res=client.fetchRow(attr);
    if(res) 
      break;
    res=client.fetchRow(type);
    if(res)
      break;

    //    std::cout << "listAttr got row" << attr << "   " << type << "\n";

    attrList.push_back(attr);
    types.push_back(type);
  }
  return res;
}


//! Removes a given attribute from the list of attributes of a
//! directory(collection)
/** The function removes a given attribute from the list of
    attributes of a directory(collection).
    \return Error code (including errors returned by the 
    \ref removeattr metadata interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: Unknown directory
    - MD_ERR_PERM:  Permission denied. You need write permission on the
                  directory.
*/
int removeAttr(MDClient &client, const std::string &dir, 
	       const std::string &key)
{
  std::string command="removeattr ";
  command.append(dir).append(" ").append(key);
  return client.execute(command);
}

//! Version of removeAttr with external management of client
int removeAttr(const std::string &dir, const std::string &key)
{
  MDClient client;

  return removeAttr(client, dir, key);
}

//! Resets the value of a given attribute of a file
/** Sets the attribute of a file to NULL. getAttr() will return this value as
    an empty string.
    \param file Name of the entry
    \param key Name of the attribute
    \return Error code (Including errors returned by the \ref clearattr metadata
    interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: Unknown directory
    - MD_ERR_PERM:  Permission denied. You need write permission on the
                    file
*/
int clearAttr(const std::string &file, const std::string &key)
{
  MDClient client;

  std::string command="clearattr ";
  command.append(file).append(" ").append(key);
  return client.execute(command);
}


//! More complex interface to getting atributes using an iterator
/** Returns an AttributeList object holding a streamed connection to
    the metadata server which allows to iteratively get the filenames
    and values of several attributes of multiple files matching a
    given pattern.
    \param pattern The pattern the filenames need to match
    \param keys The names of the attributes as a list of strings
    \param attributes The AttributeList structure returned by the call
    \return Error code (Including errors returned by the \ref getattr metadata
    interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: Unknown directory
    - MD_ERR_PERM:  Permission denied. You need read permission on the
                    file
*/
int getAttr(const std::string &pattern,
	    const std::list <std::string > &keys,
	    AttributeList &attributes)
{
  MDClient *client=new MDClient();

  std::string command="getattr ";
  command.append(pattern);
  std::list< std::string >::const_iterator I=keys.begin();
  while(I != keys.end())
    command.append(" ").append((*I++).c_str());

  int res;
  if((res=client->execute(command))){
    delete client;
    return res;
  }

  attributes.connect(client);
  return res;
}


//! Instantiates an AttributeList
AttributeList::AttributeList(int c)
  : client(NULL), columns(c)
{
  // Nothing to be done
}


void AttributeList::connect(MDClient *c)
{
  if(client){
    delete client;
  }
  client=c;
}


//! Destructor of an AttributeList
/** The destructor takes care of closing the streaming connection to
    the server.
*/
AttributeList::~AttributeList()
{
  if(client)
    delete client;
}


//! Gets the entries of the AttributeList.

/** This call interpets the data streamed by the server and returns an
    entry consisting of a filename and the values of the
    attributes. Since AttributeList keeps a buffer of
    the streamed data using the MDClient class to handle the
    connection, this call will normally not need to wait for data
    arriving from the server. You will need to call this function
    repeatedly 
    \param file Returns the name of the entry
    \param attributes Returns a list of the values of the attribtues 
    \return Error value
    - 0 OK
    - -1 Client-Server connection failed
*/
int AttributeList::getRow(std::string &file, 
			  std::vector < std::string >  &attributes)
{
  if(attributes.size() != columns)
    attributes.resize(columns);
  
  int res;
  res=client->fetchRow(file);
  if(res)
    return res;

  for(unsigned int i=0; i< columns; i++){
    res=client->fetchRow(attributes[i]);
    if(res)
      break;
  }
  return res;
}


//! Checks whether the last row has been reached  
/** Checks whether there is still an entry readeable in which case
    last row returns false.
*/
bool AttributeList::lastRow()
{
  return client->eot();
}


//! Function to create a directory (collection)
/** Creates a new directory. You will need to have write permissions
    in the parent directory and the directory should not already
    exist.
    \param dir The desired name of the directory
    \return Error code (Including errors returned by the \ref createdir metadata
    interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: Parent directory does not exist
    - MD_ERR_PERM:  Permission denied. You need write permission on the
                  parent directory
*/
int createDir(const std::string &dir)
{
  MDClient client;

  std::string command="createdir ";
  command.append(dir);
  return client.execute(command);
}


//! Adds a new attribute of a directory(collection)
/** Adds a new attribute to the list of attributes to a directory or
    collection, giving its name and type. If the type is not given,
    then currently "text" is use, which is an arbitrarily long ASCII
    text. Type is only used as the storage type in the backend. The
    communication between client and server is allways pure ASCII.
    \param dir The directory(collection) affected
    \param key The name of the attribute
    \param type The storage type of the attribute in the backend
    \return Error code (Including errors returned by the \ref addattr metadata
    interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: Directory does not exist
    - MD_ERR_PERM:  Permission denied. You need write permission on
                  the directory
*/
int addAttr(const std::string &dir, const std::string &key,
	    const std::string &type)
{
  MDClient client;

  std::string command="addattr ";
  command.append(dir).append(" ").append(key).append(" ").append(type);
  return client.execute(command);
}


//! Creates a link to another entry or an entry pointing to an external location.
/** Creates a link to another entry in the catalogue or lets the entry
    point to an external location. Note that
    metadata is associated with logical filenames and thus the
    implementation will not follow this link where metadata related
    queries of updates are concerned. Local links must have the protocol
    file: while everything else is considered to be a pointer to an
    external catalogue.
    \param file The entry affected
    \param URL The link location
    \return Error code (Including errors returned by the \ref addattr metadata
    interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: Parent directory does not exist
    - MD_ERR_PERM:  Permission denied. You need write permission on
                    the directory
*/
int link(const std::string &file, const std::string &URL)
{
  MDClient client;

  std::string command="link ";
  command.append(file).append(" ").append(URL);
  return client.execute(command);
}


//! Finds entries in the catalogue matching a query
/** Finds all files in the catalogue matching a given filename pattern
    and SQL-like query. The matching entries are returned via an
    EntryList object.
    \param pattern The pattern matching possible filenames(entries)
    \param query The SQL-Like query which winds up in the WHERE clause
           of the SQL-SELECT statement
    \param entries The returned EntryList object holding the streamed
           connection to the server (call getEntry on it to get hold
           of the entries).
    \return Error code (including errors returned by the \ref find metadata
    interface command):
    - 0 OK
    - -1 Client-Server connection failed
    - MD_ERR_NOENT: No file matches pattern
*/
int findEntries(const std::string &pattern,
		const std::string &query,
		EntryList &entries)
{
  MDClient *client=new MDClient();

  std::string command="find ";
  command.append(pattern).append(" '").append(query).append("'");

  int res;
  if((res=client->execute(command))){
    delete client;
    return res;
  }

  entries.connect(client);
  return res;
}


//! Creates an entry list
EntryList::EntryList()
  : client(NULL)
{
  // Nothing to be done
}


//! Closes streaming connection to server
EntryList::~EntryList()
{
  if(client)
    delete client;
}

//! Connectis to a new streamed connection, closes an existing old one
void EntryList::connect(MDClient *c)
{
  if(client){
    delete client;
  }
  client=c;
}

int EntryList::cancel()
{
  if(!client)
    return -1;
  return client->cancel();
}


//! Return name of an entry
/** Gets the name of an entry from the list of entries returned by the
    findFiles() function.
    \return
    - 0 OK
    - -1 Client-Server connection failed, or no more data
 */
int EntryList::getEntry(std::string &file)
{
  int res;
  res=client->fetchRow(file);
  return res;
}


//! Checks whether we have read the last entry
bool EntryList::lastEntry()
{
  return client->eot();
}


//! Instantiates an UploadHandle
UploadHandle::UploadHandle()
  : client(NULL)
{
  // Nothing to be done
}


void UploadHandle::connect(MDClient *c)
{
  if(client){
    delete client;
  }
  client=c;
}


//! Destructor of an UploadHandle
/** The destructor takes care of closing the streaming connection to
    the server.
*/
UploadHandle::~UploadHandle()
{
  if(client)
    delete client;
}


//! Commits an upload
int UploadHandle::commit()
{
  if(!client)
    return -1;
  std::string command="commit";
  return client->execute(command);
}


//! Aborts an upload
int UploadHandle::abort()
{
  if(!client)
    return -1;
  std::string command="abort";
  return client->execute(command);
}


//! Puts another entry into the collection during and upload()
/** This call sends data to the server after an upload has been
    prepared with the upload() call. 
    \param entry Denotes the name of the entry
    \param values Vector of values to set for the entry
    \return Error value
    - 0 OK
    - -1 Client-Server connection failed
    - -2 The number of values is different from what was declared in upload()
*/
int UploadHandle::put(const std::string &entry,
		      const std::vector<std::string > &values)
{
  if(!client)
    return -1;
  
  std::string command="put ";
  command.append(entry);
  for(unsigned int i=0; i<values.size(); i++)
    command.append(" ").append(values[i]);
 
  return client->execNoWait(command);
}

//! Bulk upload of entries and their attributes
/** The attributes' names  are given by the caller as a list of strings.
    They need to be defined with addAttr() first. Sets up UploadHandle
    so that it can be used to upload the entries, and finally commit
    the changes (or abort the transaction).
    \param dir The name of the collection affected
    \param keys A list of keys
    \param handle Reference to the UploadHandle which is used
    afterwards to upload the entries.
    \return Error code (including errors returned by the \ref upload
    metadata interface command):
    - 0 OK
    - -1 Client-Server connection problems
    - MD_ERR_NOENT: No such file or directory
    - MD_ERR_ILCMD: Illegal command: keys and values size mismatch
    - MD_ERR_PERM:  Permission denied. You need write permission on the
                    entry
*/
int upload(const std::string &dir,
	   const std::vector<std::string > &keys,
	   UploadHandle &handle)
{
  
  MDClient *client=new MDClient();
//  client->setSendBuffer(100000);

  std::string command="upload ";
  command.append(dir);
  std::vector< std::string >::const_iterator I=keys.begin();
  while(I != keys.end())
    command.append(" ").append((*I++).c_str());

  int res;
  if((res=client->execNoWait(command))){
    delete client;
    return res;
  }

  handle.connect(client);
  return res;
}


DIR *amgaOpendir(MDClient &client, const char *name)
{
  dirent *d = NULL;

  std::string command = "dir ";
  command.append(name);
  
  int res = 0;
  if((res=client.execute(command))) {
    errno = translateError(res);
    return NULL;
  }

  d = new(struct dirent);
  d->d_ino = 0;
  d->d_off = 0;
  d->d_reclen = 55;
  d->d_type = DT_UNKNOWN;
  d->d_name[0] ='\0';

  return (DIR *)d;
}


struct dirent *amgaReaddir(MDClient &client, DIR *directory)
{ 
  struct dirent *de = (struct dirent *)directory;

  //  std::cout << "POS: " << (void *)(&(de->d_type)) << "\n";

  if( client.eot() ) {
    errno = 0;
    return NULL;
  }

  // Name
  std::string name;
  int res = client.fetchRow(name);
  if( res ) {
    errno = translateError(res);
    return NULL;
  }

  // Type
  std::string type;
  res = client.fetchRow(type);
  if( res ) {
    errno = translateError(res);
    return NULL;
  }

  if(type == "entry")
    de->d_type = DT_REG;
  else
    de->d_type = DT_DIR;

  size_t p = name.rfind("/");
  if( p != std::string::npos )
    name = name.substr(p+1);

  strncpy( de->d_name, name.c_str(), NAME_MAX);
  
  //  std::cout << "FILLED: >" << (void *)(de->d_name) << "<\n";

  return de;
}


int amgaClosedir(MDClient &client, DIR *directory)
{
  int res = 0;
  std::string v;
  while( !client.eot() && !( res=client.fetchRow(v) ) ) {};
  if(res){
    //    std::cout << "Error occourred\n";
    errno = EIO;
    return -1;
  }
  delete (struct dirent *)directory;

  return 0;
}
