/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// 
// File ConfigParser.C
//
/*! \class ConfigParser
 * \brief A class which keeps track of the entire configuration of the
 * analysis
 *
 * $Date: 2010/04/14 07:15:30 $
 * $Author: asunil $
 * 
 * The class is initialized by calling the init() proceedure which
 * takes the command line arguments. Afterwards a configuration file
 * is opened and parsed. Subsequently, the user may ask for the values
 * of the the configuration options or whether options were provided
 * at all. The type of the values is left entirely to the user who
 * must know which datatype they contain and convert them to (lists)
 * of these dataypes by calling the various readXXXOption() or
 * readXXXOptionList() functions.
 * 
*/

#include "ConfigContext.h"
#include <cstdlib>
#include <cstring>

using namespace std;

extern int c_parse(void *);
extern int c_debug;

ConfigParser *ConfigParser::configParser = NULL;

//! Constructor which does basically nothing
ConfigParser::ConfigParser()
{  

}


//! Initialize configuration using the command line arguments
bool ConfigParser::init(const string &configFileName)
{  
  // Open configuration file and set the global file-descriptor
  std::ifstream configFile;
  configFile.open(configFileName.c_str(), ios_base::in);

  if(!configFile.is_open()) return false;

  ConfigContext scanner;
  scanner.parser = this;
  scanner.lexer.switch_streams(&configFile, NULL);
  //   scanner.lexer.set_debug(1);
  //  c_debug=1; // Set this to 1 for debugging the parser
  scanner.lastErrorLineNumber = -1;

  int r=0;

  while (!r){
    r=c_parse(&scanner);
  }
  // Close config_file
  configFile.close();

  if(r != -1) return false;
  return true;
}


//! Destructor: Deletes the entries in the list of configuration file lines
ConfigParser::~ConfigParser()
{
  // The "option" string and the list are "self deleting"
  // we do not need to do anything therefore, here.
  list<ConfigFileLine *>::iterator I=configLines.begin();
  while (I != configLines.end()) {
    delete *I;
    *I++=0;
  }
  configLines.clear();
}


//! Append the option name of another ConfigFileLine as value
void ConfigParser::appendLine(ConfigFileLine *line)
{
  configLines.push_back(line);
}


//! Deletes a line from the already parsed configuration file lines
bool ConfigParser::deleteLine(const string &name,const string &scope)
{
  string s;
  if(scope.size())
    s=string(name);
  else
    s=string(scope).append("::").append(name);

  list<ConfigFileLine *>::iterator I=configLines.begin();
  while (I != configLines.end()) {
    if((*I)->isOption(s)){
      configLines.erase(I);
      return true;
    }
    I++;
  }
  return false;
}



//! Override an option
bool ConfigParser::overrideOption(const string &name,
				  string &value,
				  const string &scope)
{
  bool result=deleteLine(name, scope);
  ConfigFileLine *line=new ConfigFileLine(name);
  if(scope.size())
    line->setScope(scope);
  line->setValue(value);
  configLines.push_back(line);
  return result;
}


//! Is this option defined?
bool ConfigParser::isDefined(const string &name) const
{
  LineIterator I=configLines.begin();
  while (I != configLines.end()) {
    if((*I)->isOption(name)) return true;
    I++;
  }
  return false;
}


//! Read integer option
int ConfigParser::getInt(const string &name) const
  throw(std::string){

  LineIterator I=configLines.begin();
  while (I != configLines.end()) {
    if((*I)->isOption(name)){
      string value = (*I)->getValue();
      char *eptr;
      int v = strtol(value.c_str(), &eptr, 10);
      if( eptr == value.c_str() || *eptr )
	throw ((string)("ConfigParser::getInt: No integer option: "))
	  .append(name);
      return v;
    }
    I++;
  }
  throw ((string)("ConfigParser::getInt: No such option: "))
    .append(name);
}


//! Read double option
double ConfigParser::getDouble(const string &name) const
  throw(std::string)
{
  LineIterator I=configLines.begin();
  while (I != configLines.end()) {
    if((*I)->isOption(name)){
      string value=(*I)->getValue();
      char *eptr;
      double v = strtod(value.c_str(), &eptr);
      if( eptr == value.c_str() || *eptr )
      	throw ((string)("ConfigParser::getDouble: No double option: "))
	  .append(name);
      return v;
    }
    I++;
  }
  throw ((string)("ConfigParser::getDouble: No such option: "))
    .append(name);
}


//! Read double list option
string ConfigParser::getString(const string &name) const
  throw(std::string)
{
  LineIterator I=configLines.begin();
  while (I != configLines.end()) {
    if((*I)->isOption(name)){
      string value = (*I)->getValue();
      if(!value.size())
	throw ((string)("ConfigParser::getString: No string option: "))
	  .append(name);
      return value;
    }
    I++;
  }
  throw ((string)("ConfigParser::getString: No such option: "))
    .append(name);
}


int ConfigParser::toList(const string &s, list<string> &values) const
  throw(std::string)
{
  std::string v(s);
  std::string::size_type p = 0;
  std::string::size_type q;
  do {
    q = v.find(",", p);
    if( q == v.size() )
      throw ("ConfigParser::toList: Illegal list" + s);

    if( q == std::string::npos )
      q = v.size();

    while( q && (v[q-1] == ' ' || v[q-1] == '\t') )
      q--;

    if( q - p ) {
      values.push_back(v.substr(p, q-p));
    }

    p = q + 1;
    while( p < v.size() && (v[p] == ' ' || v[p] == '\t') )
      p++;

  } while( p < v.size() );

  return values.size();
}


//! Read integer list option
vector<int> ConfigParser::getIntList(const string &name) const
  throw(std::string)
{
  LineIterator I=configLines.begin();

  while (I != configLines.end()) {
    if((*I)->isOption(name)){
      list<string> values;
      toList((*I)->getValue(), values);
      vector<int> v(values.size());
      list<string>::const_iterator I2=values.begin();  
      for(size_t i = 0; i < values.size(); i++) {
	char *eptr;
	v[i] = strtol((*I2).c_str(), &eptr, 10);
	if( eptr == (*I2).c_str() || *eptr )
	  throw ((string)("ConfigParser::getIntList: "
			  "Not an integer list option: "))
	    .append(name);
	I2++;
      }

      return v;
    }
    I++;
  }
  throw ((string)("ConfigParser::getIntList: No such option: "))
    .append(name);
}


//! Read double list option
vector<double> ConfigParser::getDoubleList(const string &name) const
  throw(std::string)
{
  LineIterator I=configLines.begin();

  while (I != configLines.end()) {
    if((*I)->isOption(name)){
      list<string> values;
      toList((*I)->getValue(), values);
      list<string>::const_iterator I2=values.begin();      
      vector<double> v(values.size());
      for(size_t i = 0; i < values.size(); i++) {
	char *eptr;
	v[i] = strtod((*I2).c_str(), &eptr);
	if( eptr == (*I2).c_str() || *eptr )
	  throw ((string)("ConfigParser::getDoubleList: "
			  "Not a double list option: "))
	    .append(name);
	I2++;
      }

      return v;
    }
    I++;
  }
  throw ((string)("ConfigParser::getDoubleList: No such option: "))
    .append(name);
}


//! Read string list option
vector<string> ConfigParser::getStringList(const string &name) const
  throw(std::string)
{
  LineIterator I=configLines.begin();

  while (I != configLines.end()) {
    if((*I)->isOption(name)){
      list<string> values;
      toList((*I)->getValue(), values);
      list<string>::const_iterator I2=values.begin();      
      vector<string> v(values.size());
      int i=0;
      while(I2!=values.end()){
	v[i]=*I2;
	i++; I2++;
      }

      return v;
    }
    I++;
  }
  throw ((string)("ConfigParser::getStringList: No such option: "))
    .append(name);
}


void ConfigParser::toLower(std::string &s)
{
  for(size_t i = 0; i<s.size(); i++)
    s[i] = tolower(s[i]);
}


//! Print all options to cout
void ConfigParser::print() const
{
  std::cout << "Current configuration:" << std::endl;
  LineIterator I=configLines.begin();
  while (I != configLines.end()) {
    (*I)->print();
    I++;
  }
}


std::string ConfigParser::loadFromFile(const string &filename,
				       ConfigParser &configParser) 
  throw (runtime_error) 
{
  // Try the defaults. First, current directory
  string configLocation = filename;
  if(!configParser.init(configLocation)) {
    // Now home directory.
    configLocation=getenv("HOME");
    configLocation.append("/.");
    configLocation.append(filename);
    if(!configParser.init(configLocation)) {
      // Now global configuration
      const char *gliteLocation = getenv("GLITE_LOCATION");
      if(gliteLocation) {
	configLocation = getenv("GLITE_LOCATION");
      } else {
	configLocation = "";
      }
      configLocation.append("/etc/").append(filename);
      if(!configParser.init(configLocation)) {
	throw runtime_error("Could not load configuration: " + filename);
      }
    }
  }
  return configLocation;
}


/** Removes quotes from the beginning and end of the string
 * Should replace \' or \" with " and '
 */
void ConfigParser::removeQuotes(std::string &s) const
  throw(std::string)
{
  if(s[0] == '\'') {
    s=s.substr(1);
    int p=s.size()-1;
    if( p > -1 && s[p]=='\'' )
      s=s.substr(0, p);
    else
      throw ("Unbalanced quotes: " + s);
  }

  if(s[0] == '"') {
    s=s.substr(1);
    int p=s.size()-1;
    if( p > -1 && s[p]=='"' )
      s=s.substr(0, p);
    else
      throw ("Unbalanced quotes: " + s); 
  }
}

