/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: ConfigParser.h,v 1.6 2010/04/14 07:15:30 asunil Exp $
//
// A class representing the configuration of the analysis program,
// gathered from the configuration file and the command line
//

#ifndef CONFIGPARSER_H
#define CONFIGPARSER_H

#include <stdio.h>
#include <stdexcept>
#include <vector>
#include <list>
#include <fstream>

class ConfigFileLine;

class ConfigParser {
 public:

  // Constructor which does basically nothing
  ConfigParser();

  // Initialize configuration using the command line arguments
  bool init(const std::string &fileName);

  // Destructor: Deletes the entries in the list of configuration file lines
  virtual ~ConfigParser();

  // Append the option name of another ConfigFileLine as value
  void appendLine(ConfigFileLine *line);
  
  // Deletes a line from the already parsed configuration file lines
  bool deleteLine(const std::string &name, const std::string &scope="");

  // Override an option
  bool overrideOption(const std::string &name, std::string &values,
		      const std::string &scope="");

  // Is this option defined?
  bool isDefined(const std::string &name) const;

  // Read Integer Option
  int getInt(const std::string &name) const throw(std::string);

  // Read Double Option
  double getDouble(const std::string &name) const throw(std::string);

  // Read String Option
  std::string getString(const std::string &name) 
    const throw(std::string);
  
  // Read Integer List Option
  std::vector<int> getIntList(const std::string &name) const 
    throw(std::string);

  // Read Double List Option
  std::vector<double> getDoubleList(const std::string &name) const
    throw(std::string);
  
  // Read String List Option
  std::vector<std::string> getStringList(const std::string &name) const
    throw(std::string);

  // Print all options
  void print() const;

  // Convert a string to lower case
  static void toLower(std::string &s);
  
  // Tries to read a configuration file. Searches for filename in the 
  // current directory, in ~/, and then on /etc. Returns the name of the
  // configFile that was loaded.
  static std::string loadFromFile(const std::string &filename, ConfigParser &configParser) 
    throw (std::runtime_error);

  void removeQuotes(std::string &s) const
    throw(std::string);
  
  // Single instance
  static ConfigParser *configParser;

private:
  int toList(const std::string &s,
	     std::list<std::string> &values) const
    throw(std::string);

  std::list<ConfigFileLine *> configLines; //!< A list of configuration file lines
  
  //! For convinience a ValueIterator type is defined which walks along
  //! the list of option strings
  typedef std::list<ConfigFileLine *>::const_iterator LineIterator;
};

#endif // #ifndef CONFIGPARSER_H
