/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


/** Client program for a metadata server
    Usage: mdclient [-p port] server
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include "client/MDClient.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <unistd.h>
#include <getopt.h>

#define logstream std::cout
#include "debug.h"

void printUsage()
{
  std::cout << "Usage: mdcl [-dvh] [-p port] [-s host] [-c config] [-S session] command\n"
            << "       -s host       connect to host, default: localhost\n"
	    << "       -p port       connect to given port, default: 8222\n"
	    << "       -c config     alternate configuration\n"
	    << "       -S session    use session with given number, default: 0\n"
	    << "       -d            debuging on\n"
            << "       -k key        passphrase for certificates\n"
	    << "       -v            do not use sessions\n";
}


int main(int argc, char *argv[])
{
  bool debug = false;
  int optionIndex=0;
  int port=0;
  std::string host;
  std::string configFile;
  bool persistentSessions = true;
  int sessionNumber = -1;
  std::string passKey;

  static struct option options[] = {
    {"help", 0, 0, 'h'},
    {"host", 1, 0, 's'},
    {"key", 1, 0, 'k'},
    {"server", 1, 0, 's'},
    {"config", 1, 0, 'c'},
    {"debug", 0, 0, 'd'},
    {"volatile-sessions", 0, 0, 'v'},
    {"session", 1, 0, 'S'},
    {"port", 1, 0, 'p'},
    {0, 0, 0, 0}
  };

  while(1){
    int c=getopt_long(argc, argv, "hdvk:p:s:c:S:", options, &optionIndex);
    
    if(c==-1) 
      break;
    
    switch (c) {
    case 'h':
      printUsage();
      exit(0);
    case 'p':
      port=atoi(optarg);
      break;
    case 'd':
      debug=true;
      break;
    case 'v':
      persistentSessions = false;
      break;
    case 's':
      host = optarg;
      break;
    case 'S':
      sessionNumber = atoi(optarg);
      break;
   case 'k':
      passKey = optarg;
      break;
    case 'c':
      configFile = optarg;
      break;
    default:
      std::cerr << "Unknown option" << (char)c << std::endl;
      printUsage();
      exit(5);
    }
  }

  std::string line;
  for(int i = optind; i < argc; i++) {
    line.append(argv[i]).append(" ");
  }

  if( passKey.size() ) {
    std::cout << "Found key >" << passKey << "<\n";
    MDClient::certKey[0] = '\0';
    strncpy(MDClient::certKey, passKey.c_str(), 256);
    MDClient::certKey[255] = '\0';
  }

  MDClient client(host, port, persistentSessions, configFile);
  client.setDebug(debug);
  if(sessionNumber > -1)
    client.setSessionNumber(sessionNumber);
  
  // Register CTRL-C handler:
  registerClient(&client);

  if(client.connectToServer()){
    std::cerr << client.getError() << std::endl;
    return 5;
  }
  
  if(debug) {
    std::cerr << "Connected to " << client.serverName() 
	      << ":" << client.serverPort() << std::endl;
    std::cerr << client.getGreeting() << std::endl;
  }

  int res;
  for(int i = 0; i < 2; i++) {   // Try twice
    res=client.execute(line);
  
    if(res>0){
      if(res == 200) {
	client.disconnect();
	continue;
      }
	
      std::cerr << "Error " << res << ": " << client.getError() << std::endl;
      return 5;
    }
    if(res<0){
      std::cerr << client.getError() << std::endl;
      return 5;
    }
    break;
  }
   
  std::string response;
  while (!client.eot()) {
    res=client.fetchRow(response);
    if( res )
      break;
    
    std::cout << response << std::endl;
  }
  
  if(res>0){
    std::cerr << "Error: " << res  
	      << " (" << client.getError() << ")" << std::endl;
    return 5;
  }
  
  if(res<0) {
    std::cerr << "Connection error: " << client.getError() << std::endl;
    return 5;
  }

//   if(optind!=argc-1 && optind!=argc){
//     printUsage();
//     exit(5);
//   }

  return 0;
}
