/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


/** Client program for a metadata server
    Usage: mdclient [-p port] server
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include "client/MDClient.h"

#include <cstdio>

#ifdef GPL_AMGA
#include <readline/readline.h>
#include <readline/history.h>
#else
#include <editline/readline.h>
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <unistd.h>
#include <getopt.h>

#include "util/Path.h"
#include "util/Clock.h"

#define logstream std::cout
#include "debug.h"

using namespace std;

MDClient *client = NULL;

void printUsage()
{
  std::cout << "Usage: mdclient [-dkS] [-p port] [host]" << std::endl;
  std::cout << "       -d            debug on\n";
  std::cout << "       -p port       connect to given port, default 8822\n";
  std::cout << "       -S            silent mode on\n";
  std::cout << "       -c config     alternate configuration file\n";
  std::cout << "       -k key        passphrase (key) for the certificate\n";
}

std::string currentDir;


int getCurrentDir() 
{
  int testConnection = client->execute("pwd");
  if(testConnection) { // Connection error
    std::cout << "Error: " << client->getError() << std::endl;
    return 5;
  }

  testConnection =client->fetchRow(currentDir);
  if(testConnection) {
    std::cout << "Error: " << client->getError() << std::endl;
    return 5;
  }
  return 0;
}


#ifdef GPL_AMGA
#include <vector>
#include <map>
#include <string.h>

std::vector< std::pair< std::string, std::string> > commandNames;
std::list< std::pair< std::string, bool > > entries;
std::list< std::string > users;
std::list< std::string > groups;
std::list< std::string > views;
std::list< std::pair< std::string, std::string> > attrs;
std::list< std::string > typeNames;

int fillCommandList()
{
  if( client->execute("help commands") ){ // Connection error
    std::cout << "Error: " << client->getError() << std::endl;
    return 5;
  }
  
  std::list < std::string > commands;
  
  while( !client->eot() ) {
    std::string command;
    
    if( client->fetchRow(command) ) {
      std::cout << "Error: " << client->getError() << std::endl;
      return 5;
    }
    commands.push_back(command);
  }
  commands.push_back("timing");

  commandNames.resize(commands.size());
  std::list < std::string >::iterator I = commands.begin();

  for(size_t i=0; i<commandNames.size(); i++) {
    size_t pos = (*I).find_first_of(" ");
    if( pos == std::string::npos )
      pos = (*I).size();

    commandNames[i].first = (*I).substr(0, pos);
    if( pos != (*I).size() )
      commandNames[i].second = (*I).substr(pos+1);

    I++;
  }

  return 0;
}


int fillTypeNames()
{
  typeNames.push_back("int");
  typeNames.push_back("float");
  typeNames.push_back("text");
  typeNames.push_back("varchar");
  typeNames.push_back("date");

  return 0;
}


int fillEntryList(const std::string &dir)
{
  entries.clear();

  bool hadSlash = false;
  if( dir.size() && dir[dir.size()-1] == '/') 
    hadSlash = true;

  std::string command = absolutePath(dir, currentDir);
  if(hadSlash)
    command.append("/");
  size_t pos = command.find_last_of("/");
  if (pos != command.size() )
    command = command.substr(0, pos+1);
  command = std::string("dir ") + command;

  if( client->execute(command) ){ // Connection error
    std::cout << "Error: " << client->getError() << std::endl;
    return 5;
  }
  
  while( !client->eot() ) {
    std::string entry;
    std::string type;
    
    if( client->fetchRow(entry) || client->fetchRow(type) ) {
    client->getError() ;
    //  std::cout << "Error: " << client->getError() << std::endl;
      return 5;
    }
    // std::cout << "Entry: " << entry << std::endl;
    entries.push_back(std::pair< std::string, bool>(entry, type=="entry"));
  }

  return 0;
}


int fillUserList()
{
  users.clear();

  if( client->execute("user_list") ){ // Connection error
    std::cout << "Error: " << client->getError() << std::endl;
    return 5;
  }
  
  while( !client->eot() ) {
    std::string u;
    
    if( client->fetchRow(u) ) {
      std::cout << "Error: " << client->getError() << std::endl;
      return 5;
    }
    users.push_back(u);
  }

  return 0;
}


int fillGroupList()
{
  groups.clear();

  if( client->execute("grp_list -a") ){ // Connection error
    std::cout << "Error: " << client->getError() << std::endl;
    return 5;
  }
  
  while( !client->eot() ) {
    std::string g;
    
    if( client->fetchRow(g) ) {
      std::cout << "Error: " << client->getError() << std::endl;
      return 5;
    }
    groups.push_back(g);
  }

  return 0;
}


int optionIndex(const char *text, int index)
{
  int result = 0;

  for (int i=0; i<index; i++) {
    if (text[i] == ' ' || text[i]== '\t') {
      result++;
      while (i < index && (text[i] == ' ' || text[i]== '\t') )
	i++;
    }
  }

  return result;
}

/** Identifies the current argument in the args string
    0: Unknown
    1: File/dir
    2: File
    3: dir
    4: Attribute Name
    5: User Name
    6: Group Name
    7: View
    8: Type
    9: Option
   10: Master
*/
int identifyArg(const std::string &args)
{
  std::string a(args);

  if( args.size() and args[0]=='[' )
    a = args.substr(1);

  if( !strncmp(a.c_str(), "dir", 3) ) return 3;
  if( !strncmp(a.c_str(), "path", 4) ) return 3;
  if( !strncmp(a.c_str(), "entry", 5) ) return 1;
  if( !strncmp(a.c_str(), "pattern", 7) ) return 1;
  if( !strncmp(a.c_str(), "user", 4) ) return 5;
  if( !strncmp(a.c_str(), "attr", 4) ) return 4;
  if( !strncmp(a.c_str(), "group", 5) ) return 6;
  if( !strncmp(a.c_str(), "view", 4) ) return 7;
  if( !strncmp(a.c_str(), "type", 4) ) return 8;
  if( !strncmp(a.c_str(), "master", 6) ) return 10;
  if( (  a.size()>1  &&  a[0] == '-' ) or a[1] == '-' ) return 9;
  return 0;
}


/** Returns which type of option we are searching for:

*/
int optionType(const char *text, int option)
{
  std::string t(text);

  size_t pos = t.find_first_of(" \t");
  if( pos == std::string::npos )
    return 0;
  
  // Try to identify corresponding command
  int n = -1;
  for(size_t i=0; i<commandNames.size(); i++) {
    if( commandNames[i].first == t.substr(0, pos) ) {
      n=i;
      break;
    }
  }
  std::string args(commandNames[n].second);

 
  // See whether we have an - option
  bool hasOpt = false;
  while( pos < t.size() && (t[pos]==' ' || t[pos] == '\t') ) pos++;
  if( pos < t.size() && t[pos]=='-')
    hasOpt = true;

  if( n == -1 ) // Command unknown
    return 0;

  option--;
  //  std::cout << "option: " << option << "args >" << args 
  //	    << "< hasopt " << hasOpt << "\n";

  size_t i = 0;
  if ( args[i] == '[' && args[i+1] == '-' && !hasOpt) {
    //    std::cout << "Hasopt identified\n";
    while( i<args.size() && args[i]!=' ' ) i++;
    i++;
  }

  while(option > 0 && i < args.size() ) {
    if ( args[i] == ' ' )
      option--;
    i++;
  }
  
  return identifyArg(commandNames[n].second.substr(i));
}


/** Generate a possible command which, if state==0 we start again at
   the top of the list.
*/
char *commandGenerator (const char *text, int state)
{
  static int len;
  static size_t index;
  
  if( !state ) {
    index = 0;
    len = strlen(text);
  }

  std::string matchText(text, len);

  // Return next partially matching command
  while ( index < commandNames.size() ) {
    index++;
    
    if( commandNames[index-1].first.substr(0, len) == matchText )
      return( strdup( commandNames[index-1].first.c_str() ) );
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}


/** Generate a possible command which, if state==0 we start again at
   the top of the list.
*/
char *entryGenerator (const char *text, int state)
{
  static std::list < std::pair < std::string, bool > >::iterator I;
  static int len;

  if( !state ) {
    I = entries.begin();
    len = strlen(text);
  }

  std::string matchText(text, len);
  matchText = absolutePath(matchText, currentDir);

  len = matchText.size();

  // Return next partially matching command
  while ( I != entries.end() ) {
    std::list < std::pair< std::string, bool > >::iterator J = I;
    I++;

    if( (*J).first.substr(0, len) == matchText ) {
      // If long path, only return first part
      std::string entry = (*J).first;
      if( !(*J).second && (*J).first[(*J).first.size()-1] != '/' )
	entry.append("/");
      return( strdup( entry.c_str() ) );
    }
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}


/** Generate a possible username which, if state==0 we start again at
   the top of the list.
*/
char *userGenerator (const char *text, int state)
{
  //  std::cout << "User generator >" << text << "<\n";

  static std::list < std::string >::iterator I;
  static int len;

  if( !state ) {
    I = users.begin();
    len = strlen(text);
  }

  std::string matchText(text, len);

  // Return next partially matching command
  while ( I != users.end() ) {
    std::list < std::string >::iterator J = I;
    I++;

    if( (*J).substr(0, len) == matchText ) {
      return( strdup((*J).c_str()) );
    }
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}


/** Generate a possible group name. If state==0 we start again at
   the top of the list.
*/
char *groupGenerator (const char *text, int state)
{
  //  std::cout << "Group generator >" << text << "<\n";

  static std::list < std::string >::iterator I;
  static int len;

  if( !state ) {
    I = groups.begin();
    len = strlen(text);
  }

  std::string matchText(text, len);

  // Return next partially matching command
  while ( I != groups.end() ) {
    std::list < std::string >::iterator J = I;
    I++;

    if( (*J).substr(0, len) == matchText ) {
      return( strdup((*J).c_str()) );
    }
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}


/** Generate a possible username which, if state==0 we start again at
   the top of the list.
*/
char *typeGenerator (const char *text, int state)
{
  //  std::cout << "Group generator >" << text << "<\n";

  static std::list < std::string >::iterator I;
  static int len;

  if( !state ) {
    I = typeNames.begin();
    len = strlen(text);
  }

  std::string matchText(text, len);

  // Return next partially matching command
  while ( I != typeNames.end() ) {
    std::list < std::string >::iterator J = I;
    I++;

    if( (*J).substr(0, len) == matchText ) {
      return( strdup((*J).c_str()) );
    }
  }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}


char **mdclientCompletion (const char *text, int start, int end)
{
  //  std::cout << "completion: >" << rl_line_buffer << "<  start: " << start 
  //	    << "   end: " << end << std::endl;

  char **matches;

  matches = (char **)NULL;

  /* If this word is at the start of the line, then it is a command
     to complete.  Otherwise it is the name of a file in the current
     directory. */
  if (start == 0)
    matches = rl_completion_matches (text, commandGenerator);
  
  int index = optionIndex(rl_line_buffer, start);
  
  // std::cout << "Option index: " << index << std::endl;

  if( index ) {
    int option = optionType(rl_line_buffer, index);
    //    std::cout << "Option type is: " << option << std::endl;
    
    if( option == 1 || option == 2 || option == 3) {
      getCurrentDir();
      fillEntryList(text);
      matches = rl_completion_matches (text, entryGenerator);
    }

    if( option == 5 ) {
      fillUserList();
      matches = rl_completion_matches (text, userGenerator);
    }

    if( option == 6 ) {
      fillGroupList();
      matches = rl_completion_matches (text, groupGenerator);
    }

    if( option == 8 ) {
      matches = rl_completion_matches (text, typeGenerator);
    }
  }

  rl_attempted_completion_over = 1;

  return (matches);
}


void initialize_readline ()
{
  fillCommandList();
  fillTypeNames();

  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "mdclient";

  /* Tell the completer that we want a crack first. */
  rl_attempted_completion_function = mdclientCompletion;
}

#else // ifdef GPL_AMGA

void initialize_readline ()
{
  // std::cout << "NOT DOING COMMAND COMPLETION!\n";
}
#endif // else GPL_AMGA


int main(int argc, char *argv[])
{
  bool debug = false;
  int optionIndex=0;
  std::string configLocation("");
  int port=0;
  std::string host;
  bool silent = false;
  std::string passKey;

  static struct option options[] = {
    {"help", 0, 0, 'h'},
    {"debug", 0, 0, 'd'},
    {"silent", 0, 0, 'S'},
    {"config-file", 1, 0, 'c'},
    {"key", 1, 0, 'k'},
    {"port", 1, 0, 'p'},
    {0, 0, 0, 0}
  };

  while(1){
    int c=getopt_long(argc, argv, "Shdc:p:k:", options, &optionIndex);

    if(c==-1)
      break;

    switch (c) {
    case 'h':
      printUsage();
      exit(0);
    case 'p':
      port=atoi(optarg);
      break;
    case 'd':
      debug=true;
      break;
    case 'S':
      silent=true;
      break;
    case 'c':
      configLocation=optarg;
      break;
    case 'k':
      passKey = optarg;
      break;
    default:
      std::cout << "Unknown option" << (char)c << std::endl;
      printUsage();
      exit(5);
    }
  }

  if(optind!=argc-1 && optind!=argc){
    printUsage();
    exit(5);
  }

  if(optind == argc-1)
    host=(std::string) argv[optind];

  if( passKey.size() ) {
    MDClient::certKey[0] = '\0';
    strncpy(MDClient::certKey, passKey.c_str(), 256);
    MDClient::certKey[255] = '\0';
  }

  client = new MDClient(host, port, false, configLocation, debug);

  if( !silent ) 
    std::cout << "Connecting to " << client->serverName() 
	      << ":" << client->serverPort() << "..." << std::endl;

  if(client->connectToServer()){
    std::cout << client->getError() << std::endl;
    delete client;
    return 5;
  }

  // Test connection
  if( getCurrentDir() ) {
    delete client;
    return 5;
  } 

  char *line;
  if( !silent ) {
    std::cout << client->getGreeting() << std::endl;

    // Use Readline only when not silent
    using_history();

    initialize_readline();
  } else {
    line = new char[1000000];
  }

  int i =0;
  std::string prompt = "Query> ";
  
  // Show time taken by command execution
  bool timing = false;
  Clock clk;
  
  bool dontWait = false;
  while(1) {
    i++;

    if(silent) {
      prompt = "";      
      std::cin.getline(line, 10000000);
    } else {
      line = readline(prompt.c_str());
    }

    std::string command;
    if( line ) 
      command = line;
    else               // NULL returned means EOF
      command = "quit";

    // Remove leading space
    size_t i = 0;
    while(i < command.size() && isspace(command[i]))
      i++;
    command = command.substr(i);

    int res;
    if ( line && strlen(line) && !silent )
      add_history(line);  // see man history

    if( !silent )
      free(line);

    // Register CTRL-C handler:
    registerClient(client);
        
    if (command == "timing on") {
    	std::cout << "Timing on\n"; 
    	timing = true;
    	continue;
    }
    if (command == "timing off") {
    	std::cout << "Timing off\n";
    	timing = false;
    	continue;
    }  

    if( command.substr(0, 7) == "upload "
	|| command.substr(0, 4) == "put " )
      dontWait = true;

    if( command.substr(0, 6) == "commit"
	|| command.substr(0, 5) == "abort" )
      dontWait = false;
    
    DMESG("dontWait: " << dontWait << std::endl);

    if(dontWait) {
      res = client->execNoWait(command);
      prompt = "... > ";
      continue;
    } else {      
      if (timing) {
        clk.reset();
      	clk.start();
      }
       
      res=client->execute(command);                                    
      prompt = "Query> ";
    }

    if(res>0){
      std::cout << "Error: " << res << ": " << client->getError() << std::endl;
      continue;
    }
    if(res==-1) {
      std::cout << "Error: " << client->getError() << std::endl;
      break;
    }
    if(res==-2) {
      std::cout << "Cancelled: " << client->getError() << std::endl;
      break;
    }

    std::string response;
    while (!client->eot()){
      res=client->fetchRow(response);
      if(res>0){
        std::cout << "Error " << res << ": " << client->getError() << std::endl;
        break;
      }
      if(res<0)
        break;

      if( !silent )
        std::cout << ">> ";
      std::cout << response << std::endl;
    }
    
    if (timing) {
    	clk.stop();
     	printf("Time: %2.2f\n", clk.getRealTime());
    }
    
    
    // Un-Register CTRL-C handler:
    registerClient(NULL);

    if(command=="quit") {
      break;
    }

    if(res<0){
      std::cout << "Connection error: " << client->getError() << std::endl;
      break;
    }
  }
  
  delete client;

  if( !silent ) {
    clear_history();
    std::cout << "Exiting" << std::endl;
  } else {
    delete [] line;
  }

  return 0;
}
