/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "configparser/ConfigParser.h"
#include "replicationdaemon/ReplicationDaemon.h"
#include "replicationdaemon/ReplicationManager.h"

#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

#include <getopt.h>
#include <time.h>
#include <assert.h>
#include <signal.h>
#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>

static bool debug=false;
using namespace std;
using namespace boost;

#define MD_CONFIG_FILE "mdserver.config"

//sighandler_t oldSIGINT;
//sighandler_t oldSIGCHILD;


#include "debug.h"

void printUsage() {
  std::cout << "Usage: mdrepdaemon [options]" << std::endl;
  std::cout << "Options:  -p <port>        : listen port" << std::endl;
  std::cout << "          -d               : debug on" << std::endl;
  std::cout << "          -c <configFile>  : config file" << std::endl;
  std::cout << "          -D               : daemon mode" << std::endl;
  std::cout << "          -l <logFile>     : log file" << std::endl;
}

ofstream logStream;
string logFile="";

/**
 * Signal handler for SIGUSR1
 *
 */
void sigtermhandler(int sig) {
  cout << "Exiting\n";
  exit(1);
}

void sighuphandler(int sig) {
  if ( logFile.size() ) {
    Display::reset(logStream);
    LOG("SIGHUP : log rotating. "<< std::endl);
  }
}



//static void sig_pipe(int sig) {
////  bool debug = true;
////  DMESG("Got sigpipe\n");
//}

///**
// * Signal handler for SIGCHLD (child process changes state).
// *
// * Queries return status of terminated child processes,
// * otherwise zombie processes will occure.
// */
//void child_sig_handler(int sig) {
//    
//}


///**
// * Signal handler for SIGINT
// *
// */
//void int_sig_handler(int sig) {
//
//}

/**
 * Main program
 *
 */
int main(int argc,char *argv[])
{
#if __GNUC__ >= 3
  std::set_terminate(__gnu_cxx::__verbose_terminate_handler);
#endif
  
  // Inititalize options to defaults
  string serverAddress="";
  int serverPort=0;
  string configLocation="";
  bool daemon = false;
  
  static struct option options[] = {
      {"help", 0, 0, 'h'},
      {"debug", 0, 0, 'd'},
      {"daemon", 0, 0, 'D'},
      {"port", 1, 0, 'p'},
      {"config-file", 1, 0, 'c'},
      {"log", 1, 0, 'l'},
      {0, 0, 0, 0}
  };
    int optionIndex=0;
    while(1){
      int c=getopt_long(argc, argv, "dDhp:c:l:", options, &optionIndex);
  
  if(c==-1)
    break;
  
  switch (c) {
  case 'h':
    printUsage();
    exit(0);
  case 'd':
    debug=true;
    break;
  case 'D':
    daemon=true;
    break;
  case 'l':
    logFile=optarg;
    break;
  case 'p':
    serverPort=atoi(optarg);
    break;
  case 'c':
    configLocation=optarg;
    break;
  default:
    std::cerr << "Unknown option " << (char)c << std::endl;
    printUsage();
    exit(5);
  }
  }
    
  ConfigParser configParser;
  if (configLocation != "") {
    // User provided an explicit configuration file.
    if(!configParser.init(configLocation)) {
          ERRLOG("Could not read configuration: " << configLocation << std::endl);
          exit(5);
    }
  } else {
    // Load the default configuration file
    try {
        configLocation = ConfigParser::loadFromFile(MD_CONFIG_FILE, configParser);
    } catch (runtime_error &ex) {
      ERRLOG("Configuration file not found on the default search paths."  << std::endl);
      exit(5);
    }
  }
  LOG("Loaded configuration: " << configLocation << endl);
  
  DatabaseConnection::initialize(configParser);    
  
  try {
    if (serverPort == 0) { // Not set on command line
        serverPort=configParser.getInt("Replication::ReplicationDaemonPort");
    }
  } catch (string &err) {
    LOG("Could not read listen port. Using default\n");
    serverPort=8833;  // Neither set on command line nor in configuration
  }
      
  // Initialized signal handling  
  // Need SIGPIPE handler on Unix/Linux systems to catch 
  // broken pipes when the socket is closed
//    signal(SIGPIPE, sig_pipe);    
  signal(SIGINT, sigtermhandler);
  signal(SIGSTOP, sigtermhandler);  
  signal(SIGQUIT, sigtermhandler);    
  signal(SIGHUP, sighuphandler);    
  signal(SIGPIPE, SIG_IGN);

  try {
    if(!logFile.size()) { // Not set on command line
      logFile=configParser.getString("Logfile");
    }
  } catch (std::string &err) {
    // Neither set on command line nor in configuration
    logFile = "";
  }
  if(logFile.size()) {
    logStream.open (logFile.c_str() ,ios_base::app);
    Display::setOutput(logStream.rdbuf());
  }

  // Now let's become a daemon, if wanted
  if(daemon) {
    // Our process ID and Session ID
    pid_t pid, sid;

    // Fork off the parent process
    pid = fork();
    if (pid < 0) {
      exit(EXIT_FAILURE);
    }

    // We are the parent, and can die
    if (pid > 0) {
      exit(EXIT_SUCCESS);
    }

    // Change the file mode mask: full access to our own files
    umask(0);

    // Create a new session id (SID) for the child process
    sid = setsid();
    if (sid < 0) {
      ERRLOG("Could not get new session ID for daemon proces" << std::endl);
	exit(EXIT_FAILURE);
    }

    // change working dir to a safe place
    if ((chdir("/")) < 0) {
      ERRLOG("Could not chdir /");
      exit(EXIT_FAILURE);
    }

    // Close stdin, should be doing it also with stdout, but only if we have
    // a logfile set up
    close(STDIN_FILENO);
    //  close(STDOUT_FILENO);
    //  close(STDERR_FILENO);
  }

  LOG("Listening on port: " << serverPort << endl);
  // All replication classes get the debug flag from the ReplicationManager
  ReplicationManager::initManager(configParser, debug, true);
  ReplicationDaemon repDaemon(serverPort, configParser);
  repDaemon.run();    
}
