/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "server/MDInterpreter.h"
#include "server/MDStandalone.h"
#include "server/MDServerFactory.h"
#include "server/MDAsyncBuffer.h"
#include "server/MDConManMem.h"
#include "server/MDConManDB.h"
#include "server/UserContext.h"
#include "server/VOMSInfo.h"
#include "server/protocolHandling.h"
#include "replicationdaemon/MDLogWriter.h"
#include "replicationdaemon/LogReceiver.h"
#include "replicationdaemon/MountManager.h"
#include "replicationdaemon/ReplicationManager.h"

#include "util/DatabaseConnection.h"
#include "util/SecurityUtils.h"
#include "util/Socket.h"
#include "util/sslHelpers.h"
#include "util/amgaProtocol.h"

#include "configparser/ConfigParser.h"

#include <cstdio>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <getopt.h>
#include <time.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <openssl/ssl.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/safestack.h>

#include <iostream>
#include <list>
#include <sstream>

static int debug=0;

sighandler_t oldSIGINT;
sighandler_t oldSIGCHILD;

static volatile bool terminateRequest = false;

using namespace std;

#include "debug.h"

MDConMan *conMan; // Connection and session manager

// Connection management: Used from within interrupt, no STL!
int numSlots;
int *children;
int *deadChildren;

static AuthContext authCtx;
//static bool requireAuthentication = true;
//static bool allowPasswordAuthentication = false;
//static bool allowCertAuth=false;
//static bool allowMyProxyHack = false;
static bool forceSessions = false;
static bool allowSessions = true;
static int maxConnects = 1000000;
static long sessionTimeout = 300;

void printUsage()
{
  std::cout << "Usage: mdservermt [options]" << std::endl;
  std::cout << "Options:  -p <port>        : server port" << std::endl;
  std::cout << "          -d               : debug on" << std::endl;
  std::cout << "          -D               : deamon mode" << std::endl;
  std::cout << "          -c <configFile>  : config file" << std::endl;
  std::cout << "          -l <logFile>     : log file" << std::endl;
  std::cout << "          -H               : Allow holding connections" << std::endl;
}

struct rwSync {
  MDInterpreter::CommandToken command;
  std::vector< std::string > commandString;
  MDServer *connection;
  MDAsyncBuffer *buffer;
  MDInterpreter *interpreter;
  bool read;
  bool done;
  bool streaming;
  pthread_mutex_t mutex;
  pthread_cond_t cond;
};


int ssl_callback_NewSessionCacheEntry(SSL *ssl, SSL_SESSION *session)
{
  DMESG("ssl_callback_NewSessionCacheEntry called" << std::endl);
  MDConMan *conMan = (MDConMan *) SSL_get_app_data(ssl);

  unsigned char *sessionID = session->session_id;
  unsigned int idLen = session->session_id_length;

  int size= i2d_SSL_SESSION(session, NULL);
  DMESG("SSL Session size is: " << size << std::endl);

  const char *buf=new char[size];
  unsigned char *pp=(unsigned char *)buf;
  size=i2d_SSL_SESSION(session, &pp);

  LOG("Saving SSL session " 
      << sessionToPrintable(sessionID, idLen) << std::endl);

  if(!conMan->saveSessionSSL(buf, size, sessionID, idLen))
    ERRLOG("Could not save SSL session " 
       << sessionToPrintable(sessionID, idLen) << std::endl);
  
  delete [] buf;

  // Session is still valid
  return 0;
}


SSL_SESSION *ssl_callback_GetSessionCacheEntry(SSL *ssl, unsigned char *sessionID,
                           int idLen, int *doCopy)
{
  LOG("Resuming SSL session " 
      << sessionToPrintable(sessionID, idLen) << std::endl);

  DMESG("ssl_callback_GetSessionCacheEntry called" << std::endl);
  MDConMan *conMan = (MDConMan *)SSL_get_app_data(ssl);

  char buf[conMan->getMaxSessionSize()];
  int size;

  if(conMan->getSessionSSL(sessionID, idLen, buf, size)){
    ERRLOG("Session not found: " 
       << sessionToPrintable(sessionID, idLen) << std::endl);
    return NULL;
  }

#if OPENSSL_VERSION_NUMBER >= 0x00908000L
  const unsigned char *data = (const unsigned char *) buf;
#else
  unsigned char *data = (unsigned char *) buf;
#endif
  SSL_SESSION *session = d2i_SSL_SESSION(NULL, &data, size);

  // We do not hold a further copy of the session, but OpenSSL needs
  // to increase reference count (*doCopy =0)
  *doCopy = 0;
  return session;
}


void ssl_callback_RemoveSessionCacheEntry(SSL_CTX *ctx, SSL_SESSION *session)
{
  DMESG("ssl_callback_RemoveSessionCacheEntry called" << std::endl);
  MDConMan *conMan = (MDConMan *)SSL_CTX_get_app_data(ctx);
  
  unsigned char *sessionID = session->session_id;
  unsigned int idLen = session->session_id_length;

  LOG("Removing SSL session " 
      << sessionToPrintable(sessionID, idLen) << std::endl);

  if(conMan->deleteSession(sessionID, idLen)){
    ERRLOG("Removing session, session not found: "
       << sessionToPrintable(sessionID, idLen) << std::endl);
    return;
  }

  return;
}

int initSessions(SSL_CTX *ctx, char *appData, int sessionTimeout) {
  // Turn off internal session cache
  SSL_CTX_set_session_cache_mode(ctx,
                 SSL_SESS_CACHE_SERVER | SSL_SESS_CACHE_NO_INTERNAL);

  // Set callbacks for session caching in DB file
  SSL_CTX_sess_set_new_cb(ctx, ssl_callback_NewSessionCacheEntry);
  SSL_CTX_sess_set_get_cb(ctx, ssl_callback_GetSessionCacheEntry);
  SSL_CTX_sess_set_remove_cb(ctx, ssl_callback_RemoveSessionCacheEntry);

  // Set application data
  SSL_CTX_set_app_data(ctx, appData);

  // Set session context
  if(!SSL_CTX_set_session_id_context(ctx, (const unsigned char *)"MDServer", 8))
    abort();

  // Session timeouts
  SSL_CTX_set_timeout(ctx, sessionTimeout);

  return 0;
}

void *dbThreadStarter(void *arg)
{
  intptr_t res=0;

  struct rwSync *sync=(struct rwSync *)arg;
  MDInterpreter::CommandToken c=sync->command;
  MDServer *connection=sync->connection;
  MDAsyncBuffer *rwBuffer=sync->buffer;

  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  DMESG("Writer thread starting\n");

  try { // Catch BufferCancelledException
    do {
      if(sync->streaming) {
		pthread_mutex_lock(&sync->mutex);
		while(sync->done)
	  		pthread_cond_wait(&sync->cond, &sync->mutex);
      }
      c=sync->command;
      if(sync->interpreter->exec(c, sync->commandString, *connection)) {
		DMESG("3 Invalid command: " 
			<< (sync->commandString.size()?sync->commandString[0]:"")
	      	<< std::endl);
		res=3;
		if(sync->streaming) {
	  		LOG("Streaming failed" << std::endl);
	  		sync->done=true;
	  		sync->streaming=false;
	  		pthread_cond_signal(&sync->cond);
	  		pthread_mutex_unlock(&sync->mutex);
		}
		break;
      }
      if(c==MDInterpreter::cAbort || c==MDInterpreter::cCommit)
	continue;
      if(sync->streaming) {
	sync->done=true;
	pthread_cond_signal(&sync->cond);
	pthread_mutex_unlock(&sync->mutex);
      }
    } while(sync->streaming
	    && (c != MDInterpreter::cAbort && c!=MDInterpreter::cCommit));

    rwBuffer->finish();
  } catch (BufferCanceledException ex) {
    res = 1;
    DMESG("Buffer canceled. Returning " << res << std::endl);
  }
  // If we catch and rethrow we will not be able to use dbg to
  // locate the source of the initial exception!
    //  } catch (...) {
    //    DMESG("Unknown exception. Rethrowing." << std::endl);
    //    throw;
    //  }
  DMESG("Returning error code " << res << std::endl);
  return (void *)res;
}


int returnStatistics(TCPSocket &socket, MDConMan *conMan)
{
  std::ostringstream statistics;
  StatsCollector::dumpAsXml(statistics);
  try {
    socket.send(statistics.str());
  } catch (SocketException &ex) {
    PERRLOG("Cannot send response\n");
    return -1;
  }
  DMESG("Statistics sent" << std::endl);
  return 0;
}

int initConnection(TCPSocket &socket, MDConMan *conMan,
		   MDServer *connection, UserContext &userCtx)
{
  // Set socket options to TCP_NODELAY during setup
  socket.setTCPNoDelay(true);

  // Initialize client connection, check whether SSL is demanded,
  // Resume session if requested
  bool useSSL;
  // Check whether we need to resume a session
  uint64_t sessionID = 0;
  string greeting = "ARDA Metadata Server " VERSION;
  try {
    // Plain text. Send greeting and ask user if SSL is requested.
    // call initConnection in protocolHandling.cc
    int ret = initConnection(socket, authCtx.useSSL, greeting, useSSL, sessionID);
    if (ret == 1) {
      if (returnStatistics(socket, conMan)) {
        return -1;
      } else {
        return 1;
      }
    }
  } catch (SocketException &ex) {
    LOG(ex.what() << std::endl);
    return -1;
  }
  
  DMESG("Connection established" << std::endl);

  if(useSSL) {
    // The client requested SSL. Switch socket to SSL
    try {
      socket.initSSLServer(authCtx.ctx, (char *)conMan);
    } catch (SocketException &ex) {
      LOG(ex.what() << std::endl);
      return -1;
    }
    DMESG("SSL_accepted" << std::endl);
  }
  // If the user requested SSL, this pointer is non-null
  SSL* ssl = socket.getSSLCtx();

  // Initialize the user context. Either by reading it from the client
  // or from the stored session.

  if(sessionID) {
    DMESG("Resuming session " << sessionID << std::endl);
    if(conMan->getSessionUser(sessionID, userCtx)) {
      LOG("Attempt to resume " << sessionID << " failed.\n");
      std::string response = "200 No such session\n\004\004";
      socket.send(response);
      return -1;
    }
    
  } else {
    // Read user name, permissions and related context information from
    // client.
    DMESG("Establishing context\n");
    try {
      establishUserContext(socket, userCtx);
      DMESG("Context established\n");
    } catch (SocketException &ex) {
      LOG("Error establishing context: " << ex.what() << endl);
      return -1;
    }

    DMESG("Authentication required: " << authCtx.requireAuthentication << std::endl);
    bool userAuthenticated = authenticateUser(ssl, connection, authCtx, userCtx);
    if (authCtx.requireAuthentication && !userAuthenticated) {
      LOG("Authentication error for user: " << userCtx.login << std::endl);
      socket.send("201 Could not authenticate user " + userCtx.login +"\n\004\004");
      return -1;
    }
  }

  // Set connection options
  connection->setUser(userCtx.login);
  connection->setDir(userCtx.currentDir);
  connection->setLogging(true);
  MDStandalone *sServer = dynamic_cast<MDStandalone *>(connection);
  if(sServer) {
    sServer->setPermissionMask(userCtx.permissionMask);
    sServer->setGroupRightsMask(userCtx.groupMask);
  }

  if(ssl) {
    X509 *cert = SSL_get_certificate(ssl);
    EVP_PKEY *pkey = X509_get_pubkey(cert);
    if(sServer) {
      sServer->setPublicKey(pkey);
      pkey = SSL_get_privatekey(ssl);
      sServer->setPrivateKey(pkey);
    }
  }
  return 0;
}


int handleConnection(TCPSocket &socket, MDConMan *conMan,
		     MDServer *connection, MDAsyncBuffer &rwBuffer)
{
  UserContext userCtx;
  int ret = initConnection(socket, conMan, connection, userCtx);
  if (ret == -1) {
    // Error
    return -1;
  } else if (ret == 1) {
    // Ok
    return 0;
  }

  MDInterpreter interpreter;

  // From now on, use TCP_CORK ing
  //  sockflag=0;
  //  setsockopt(ipConnection, IPPROTO_TCP, TCP_NODELAY, &sockflag, sizeof(sockflag));
  //  sockflag=1;
  //  setsockopt(ipConnection, IPPROTO_TCP, TCP_CORK, &sockflag, sizeof(sockflag));

  // Loops as long as commands are sent until quit
  std::string response;
  // A priory, the session is not to be stored, unless suspendConnections is
  // true and the sessionID has been sent to the client
  bool storeSession=false;
  while(1) {
    bool transaction = false;
    DMESG("Listening..." << std::endl);
    response = "";

    std::string line;
    try {
      line = socket.readLine();
      AMGA::decodeLine(line);
    } catch (SocketException &ex) {
      DMESG("Client disconnect: " << ex.what() << endl);
      break;
    }
  
    conMan->timestamp(5);

    DMESG("Got >" << line  <<"<" << std::endl);
    MDInterpreter::CommandToken c;
    c = interpreter.parseCommand(line);

    // Special commands
    // Exit immediately
    if (c==MDInterpreter::cExit)
      break;
    if (c == MDInterpreter::cClose) {
      if(allowSessions) {
        storeSession = true;
        break;
      } else {
        break;  // handle as if exit
      }
    }
   
    // Ignore commands which are out of context
    if(c==MDInterpreter::cPut)
      continue;

    std::vector< std::string > command=interpreter.getCommand();

    // Prepare thread creation
    pthread_attr_t threadAttrs;
    pthread_attr_init(&threadAttrs);
    // pthread_attr_setdetachstate(&threadAttrs, PTHREAD_CREATE_DETACHED);    
    
    // Set the next command to be executed.
    // This is ignored if replication is disabled.
    connection->setNextCommand(line);
    
    struct rwSync sync;
    sync.command=c;    
    sync.connection=connection;
    sync.commandString=command;
    sync.streaming=false;
    sync.done=false;
    if(c==MDInterpreter::cUpload) {
      DMESG("IDENTIFIED STREAMING START" << std::endl);
      sync.streaming=true;
      pthread_mutex_init(&sync.mutex, NULL);
      pthread_cond_init(&sync.cond, NULL);
    }

    sync.buffer=&rwBuffer;
    sync.interpreter=&interpreter;
    rwBuffer.start();

    pthread_t thread;
    int ret=pthread_create(&thread, &threadAttrs, dbThreadStarter, (void *)&sync);
    if(ret) {
      ERRLOG("Could not create thread to handle DB connection"
  	     << std::endl);
      break;
    }

    if(c==MDInterpreter::cUpload) {
      while(c!=MDInterpreter::cAbort && c!=MDInterpreter::cCommit) {
        try {
          line = socket.readLine();
          AMGA::decodeLine(line);
        } catch (SocketException &ex) {
          DMESG("Client disconnect: " << ex.what() << endl);
          break;
        }
        DMESG("Streaming command: >" << line << "<" << std::endl);
        c=interpreter.parseCommand(line);
        if(c==MDInterpreter::cExit)
          break;
        command=interpreter.getCommand();
        pthread_mutex_lock(&sync.mutex);
        while ( !sync.done )
          pthread_cond_wait(&sync.cond, &sync.mutex);
        if(!sync.streaming) // An error occurred
          break;
        // Must wait for the previous command to be fully executed
        // before setting the values on these structures, since they 
        // are shared.
        connection->setNextCommand(line);

        sync.commandString=command;
        sync.command=c;
        sync.done=false;
        pthread_mutex_unlock(&sync.mutex);
        pthread_cond_signal(&sync.cond);
      }
      DMESG("Streaming master exits loop" <<  std::endl);
      if(c==MDInterpreter::cExit)
        break;
      DMESG("Streaming master FINISH" <<  std::endl);
    }

    // Loop reading all MDServer responses, sending them to the client
    bool transmitError;
    while(1) {
      transmitError = false;
      rwBuffer.read(response, 1000);  // Check for timeout if cancelled
      // DMESG("Response >" << response << "<" << std::endl);
//      cout << "Timeout\n";

      // Check for any out-of-band messages
      char controlBuf;
      if (socket.recvOutOfBand(&controlBuf, 1) > 0) {
        DMESG("Out of band\n");
        // We cannot safely cancel a thread, so we only signal
        // our request
        LOG("Request CANCELLED" << std::endl);
        rwBuffer.cancel(); 
        break;
      }

      // We are finished, but we will need to send the rest of the
      // response later
      if(rwBuffer.isProducerFinished()) {
        std::string rest;
        rwBuffer.read(rest);
        response.append(rest);
        break;
      }

      // Write out any data we have gotten
      if(response.size()) {
        // If the client has closed the connection, this will send SIGPIPE,
        // which kills the process (would need handler!)
        try { 
	  //	  DMESG("Sending: >" << response << "<\n");
          socket.send(response);
        } catch (SocketException &ex) {
          ERRLOG("Cannot send response: " << ex.what() << std::endl);
          transmitError=true;
          break;
        }
      }
    }

    // Join thread
    void *threadReturnCode;
    DMESG("Joining thread" << std::endl);
    int threadStatus=pthread_join(thread, &threadReturnCode);
    if(threadStatus) {
      ERRLOG("Could not join reader thread" << std::endl);
      break;
    }
    if(rwBuffer.isCancelled()) {
      DMESG("Request was cancelled\n");
      response.append("\030");
      response.append("12 Request cancelled\n");
    } else {
      if(threadReturnCode){
    	char buf[30];
    	DMESG("Sending error code " << (intptr_t) threadReturnCode
    	      << std::endl);
    	switch((intptr_t)threadReturnCode){
    	case 3:
    	  DMESG("Sending 3 Invalid command\n");
    	  sprintf(buf, "%d Invalid command.\n", (intptr_t) threadReturnCode);
    	  break;
    	case 12:
    	  sprintf(buf, "%d Aborted.\n", (intptr_t) threadReturnCode);
    	}
    	response.append(buf);
      }
    }

    if(sync.streaming) {
      pthread_mutex_destroy(&sync.mutex);
      pthread_cond_destroy(&sync.cond);
    }

    if(transmitError) // Don't send anything more if transmit error
      break;

    if( connection->transactionInProgress()
	|| (!forceSessions && !storeSession) ){
      // Send EOT to signal end of response
      DMESG("Sending EOT" << std::endl);
      response.append("\004\004");  // Append final EOT char
      try {
        socket.send(response);
      } catch (SocketException &ex) {
        ERRLOG("Cannot send response: " << ex.what() << endl);
        sslPrintErrors();
        transmitError=true;
        break;
      }
    } else {
      storeSession = true;
      break;
    }
  } // End of loop over all client commands

  DMESG("Closing connection (pid " << getpid() << ")\n");
  if(storeSession) {
    // Retrieve settings from connection and store them in context
    userCtx.currentDir = connection->getDir();

    DMESG("Saving session" << std::endl);
    uint64_t sessionID = conMan->saveSessionUser(userCtx);
    if(!sessionID)
      ERRLOG("Could not save session " << sessionID << std::endl);

    DMESG("Suspending session " << sessionID << std::endl);
    char sBuffer[60];
    sprintf(sBuffer, "\004session%llu\004", sessionID);
    response.append(sBuffer);
    storeSession=true;

    // Send EOT to signal end of response
    //    DMESG("Sending EOT: " << response << std::endl);
    DMESG("Sending EOT:\n");
    try {
      socket.send(response);
    } catch (SocketException &ex) {
      ERRLOG("Cannot send response");
      sslPrintErrors();
    }
  }
  
  int count = 1000;
  while( connection->transactionInProgress() && count-- ) 
    connection->transactionAbort();

  return 0;
}


/**
 * Signal handler for SIGUSR1
 *
 */
void child_term_handler(int sig) {
  DMESG("SIGUSR1 received: " << getpid() << std::endl);
  terminateRequest=true;
}


int slaveProcess(TCPServerSocket &serverSocket, MDConMan *conMan)
{
  struct sigaction act;

  memset(&act, 0, sizeof(act));
  act.sa_handler = SIG_IGN;
  sigaction(SIGINT, &act, NULL);

  signal(SIGTERM, child_term_handler);
  siginterrupt(SIGTERM, 1);

  // Tell that we are up and running, but not yet finished
  conMan->timestamp(2);

  // Initialize connection to backend
  MDAsyncBuffer rwBuffer;
  MDServer *connection=NULL;
  try {
    connection=MDServerFactory::mdServer(rwBuffer);
    if(!connection || !connection->isConnected()){
        ERRLOG("Could not connect to DB" << std::endl);
        return -1;
    }
  } catch (SQLException &ex) {
    ERRLOG("Could not connect to DB" << std::endl);
    return -1;
  }

  connection->setDebug(debug);
  connection->printODBCDriverVersion(); // Debug printout
  
  // Initialize User Manager
  std::list<MDUserManager *> userManagers;
  createUserManagers(*(ConfigParser::configParser), userManagers);
  connection->initializeUserManager(userManagers);

  // We will handle up to maxConnects connections, then we will die
  for (int i=0; i< maxConnects && !terminateRequest; i++){
    DMESG("Process does accept" << std::endl);
    conMan->timestamp(3);
    // We will handle up to maxConnects connections, then we will die
    
    try {
        auto_ptr<TCPSocket> socket(serverSocket.accept());
        DMESG("ACCEPT" << std::endl);
        conMan->timestamp(4);
        conMan->notifyMaster();
        handleConnection(*socket, conMan, connection, rwBuffer);
    } catch (SocketException &ex) {
      if(errno!=EINTR) {
        PERRLOG("Error accepting connection: " << ex.what() << endl);
      }
      break;
    }
  }
  LOG("Finished serving, going away" << std::endl);
  delete connection;
  return 0;
}


/**
 * Signal handler for SIGCHLD (child process changes state).
 *
 * Queries return status of terminated child processes,
 * otherwise zombie processes will occure.
 *
 * Remember that you cannot call a anything in MDConMan that
 * tries to get a lock from within a signal handler!
 */
void child_sig_handler(int sig) {
  for(int i=0; i<numSlots; i++){
    if(!children[i])
      continue;
    if(waitpid(children[i], NULL, WNOHANG)){
      // The child will be removed laterfrom the list of processes
      // by calling conMan.remove() from the master process
      deadChildren[i]=children[i];
      // The following is not safe in a signal handler!
      //      DMESG("RECEIVED SIGCHILD: " << children[i] << std::endl);
      children[i]=0;
    }
  }
}


/**
 * Signal handler for SIGINT
 *
 */
void int_sig_handler(int sig) {
  if(conMan->isMaster()){
    //    LOG("SIGINT received: please wait for shutdown..."<< std::endl);
    terminateRequest=true;
  }
}


void removeDead()
{
  // Remove dead children from connection manager
  // DMESG("REMOVING dead children" << std::endl);
  for(int i=0; i<numSlots; i++){
    if(!deadChildren[i])
      continue;
    conMan->remove(deadChildren[i]);
    DMESG("REMOVED DEAD CHILD: " << deadChildren[i] << std::endl);
    deadChildren[i]=0;
  }
}


/**
 * Main program
 *
 */
int amgad(int argc,char *argv[])
{
#if __GNUC__ >= 3
    std::set_terminate(__gnu_cxx::__verbose_terminate_handler);
#endif

  static struct option options[] = {
    {"help", 0, 0, 'h'},
    {"debug", 0, 0, 'd'},
    {"daemon", 0, 0, 'D'},
    {"port", 1, 0, 'p'},
    {"config-file", 1, 0, 'c'},
    {"log", 1, 0, 'l'},
    {0, 0, 0, 0}
  };

  // Inititalize options to defaults
  int optionIndex=0;
  int port=0;
  std::string logFile="";
  int daemon=0;
  std::string configLocation="";
  std::ofstream logStream;

  while(1){
    int c=getopt_long(argc, argv, "hp:dc:l:D", options, &optionIndex);

    if(c==-1)
      break;

    switch (c) {
    case 'h':
      printUsage();
      exit(0);
    case 'd':
      debug=1;
      break;
    case 'D':
      daemon=1;
      break;
    case 'p':
      port=atoi(optarg);
      break;
    case 'l':
      logFile=optarg;
      break;
    case 'c':
      configLocation=optarg;
      break;
    default:
      std::cerr << "Unknown option" << (char)c << std::endl;
      printUsage();
      exit(5);
    }
  }

  if(optind!=argc){
    std::cerr << "Too many parameters!" << std::endl;
    printUsage();
    exit(5);
  }  

  // Initialize debugging
  authCtx.setDebug(debug);
  ConfigParser::configParser = new ConfigParser();
  if (configLocation != "") {
    // User provided an explicit configuration file.
    if(!ConfigParser::configParser->init(configLocation)) {
      ERRLOG("Could not read configuration: " << configLocation << std::endl);
      exit(5);
    }
  } else {
    // Load the default configuration file
    try {
      configLocation = ConfigParser::loadFromFile("mdserver.config", *ConfigParser::configParser);        
    } catch (runtime_error &ex) {
      ERRLOG("Configuration file not found on the default search paths."  << std::endl);
      exit(5);            
    }
  }
  LOG("Loaded configuration: " << configLocation << endl);
    
  DatabaseConnection::initialize(*ConfigParser::configParser);

  try {
    if(!port) { // Not set on command line
      port=ConfigParser::configParser->getInt("Port");
    }
  } catch (std::string &err) {
    port=8822;  // Neither set on command line nor in configuration
  }
  try {
    if(!logFile.size()) { // Not set on command line
      logFile=ConfigParser::configParser->getString("Logfile");
    }
  } catch (std::string &err) {
    // Neither set on command line nor in configuration
    logFile = "";
  }
  if(logFile.size()) {
    logStream.open (logFile.c_str());
    Display::setOutput(logStream.rdbuf());
  }

  // Connection configuration
  try {
    maxConnects = ConfigParser::configParser->getInt("MaxConnectsPerProcess");
  } catch (std::string err) {
    // Default is as many as you like
  }
  try {
    sessionTimeout = ConfigParser::configParser->getInt("SessionTimeout");
  } catch (std::string err) {
    // Default is as many as you like
  }
  int idleTimeout;
  try {
    idleTimeout = ConfigParser::configParser->getInt("IdleTimeout");
  } catch (std::string err) {
    idleTimeout = 20 *60; // Default is 20 minutes
  }

  // Now let's become a daemon, if wanted
  if(daemon) {
    // Our process ID and Session ID
    pid_t pid, sid;

    // Fork off the parent process
    pid = fork();
    if (pid < 0) {
      exit(EXIT_FAILURE);
    }

    // We are the parent, and can die
    if (pid > 0) {
      exit(EXIT_SUCCESS);
    }

    // Change the file mode mask: full access to our own files
    umask(0);

    // Create a new session id (SID) for the child process
    sid = setsid();
    if (sid < 0) {
      ERRLOG("Could not get new session ID for daemon proces" << std::endl);
	exit(EXIT_FAILURE);
    }

    // change working dir to a safe place
    if ((chdir("/")) < 0) {
      ERRLOG("Could not chdir /");
      exit(EXIT_FAILURE);
    }

    // Close stdin, should be doing it also with stdout, but only if we have
    // a logfile set up
    close(STDIN_FILENO);
    //  close(STDOUT_FILENO);
    //  close(STDERR_FILENO);
  }


  // Configure session handling
  try {
    std::string s = ConfigParser::configParser->getString("Sessions");
    ConfigParser::toLower(s);
    if( s == "force" ) {
      forceSessions = true;
      allowSessions = true;
    }

    if( s == "allow" || s == "yes" || s == "1") {
      allowSessions = true;
    }

    if(! (s == "yes" || s == "1" || s=="force" || s == "allow"
	  || s == "no" || s == "0")) {
      ERRLOG("Unknown value for option UseSSL\n");
      abort();
    }

  } catch (std::string err) { }
   
  try {
    // Initialize replication
    ReplicationManager::initManager(*ConfigParser::configParser, debug);
  } catch (runtime_error &ex) {
    ERRLOG(ex.what() << endl);
    exit(1);
  }
    
  // Initialize connection management
  unsigned int maxProcesses;
  try{
    maxProcesses=ConfigParser::configParser->getInt("MaxProcesses");
  } catch (std::string err) {
    maxProcesses=50;
  }


  // Initialize connection manager see which one we will need
  bool useDBMSessionCache;
  try{
    useDBMSessionCache = (bool) ConfigParser::configParser->getInt("UseDBMSessionCache");
  } catch (std::string err) {
    useDBMSessionCache = false;
  }
  
  if(useDBMSessionCache) {
    conMan = new MDConManDB();
    LOG("Connection caching done in DBM file\n");
  } else {
    conMan = new MDConManMem();
    LOG("Connection caching done in shared memory\n");
  }
  
  assert(conMan);
  
    
  if(!conMan->init(maxProcesses)){
    ERRLOG("Could not initialize connection manager!"
	   << std::endl);
    exit(1);
  }
  StatsCollector::registerStatsSource(*conMan);
  
  int minProcesses;
  try{
    minProcesses=ConfigParser::configParser->getInt("MinProcesses");
  } catch (std::string err) {
    minProcesses=2;
  }
  numSlots=maxProcesses;
  children=new int[numSlots];
  deadChildren=new int[numSlots];
  for(int i=0; i<numSlots; i++)
    children[i]=deadChildren[i]=0;

  // Init session cache
  int maxSessionSize;
  int maxSessions;
  try{
    maxSessions = ConfigParser::configParser->getInt("MaxSessions");
  } catch (std::string err) {
    maxSessions = 1024;
  }
  try{
    maxSessionSize = ConfigParser::configParser->getInt("MaxSessionSize");
  } catch (std::string err) {
    maxSessionSize = 2000;
  }
  
  if(conMan->initSessionCache(maxSessions, maxSessionSize)) {
    ERRLOG("Could not initialize session cache!"
	   << std::endl);
    exit(1);
  }
  
  // Init SSL: Session handling + VOMS Attribute parsing
  authCtx.init(*ConfigParser::configParser);
  if (authCtx.useSSL) {
    initSessions(authCtx.ctx, (char *)conMan, sessionTimeout);
    // Load VOMS extensions to certificates into SSL
    VOMSInfo::initASN1Parser(); 
  }

  // Install signal handler
  oldSIGCHILD=signal(SIGCHLD, child_sig_handler);
  oldSIGINT=signal(SIGINT, int_sig_handler);
  signal(SIGTERM, int_sig_handler);
  conMan->start(30, idleTimeout); // Look at connections every 30 sec
  

  LOG("Listening on port: " << port << std::endl);
  // Open the ServerSocket on the master so it can be shared by users.
  TCPServerSocket serverSocket(port);

  LOG("Starting up slaves" << std::endl);
  time_t lastTime = time(NULL);
  int numCalls = 0;
  while(1) {
    // We don't wait the fist time we come, because conMan is unlocked
    conMan->waitMaster(); // Wait until: timeout, interrupt, notify from child
    if(terminateRequest) {
      LOG("Shutting down all slaves" << std::endl);
      conMan->reapAll();
      LogReceiverManager::cleanup();
      while(sleep(2)) { };
      removeDead();
      conMan->killLeftovers(true);
      // Restore the old signal handler, in case someone comes late
      signal(SIGCHLD, oldSIGCHILD);
      break;              // End master
    }  else {
      conMan->reap(idleTimeout);  // Reap all idle process
      conMan->killLeftovers(false);

      // Run session expiration every 300 seconds or after 1000 connects
      time_t now = time(NULL);
      numCalls++;
      if( (now - lastTime > 300) || (numCalls >= maxSessions/2-2)) {
	numCalls = 0;
	lastTime = now;
	conMan->expireSessions(sessionTimeout);
      }
    }
    removeDead();

    if(conMan->getPrepared() < minProcesses) {
      DMESG("->> Getting new processes: require " << minProcesses
	    << std::endl);
      int ret;
      std::vector<int> processSlots;
      processSlots.resize(minProcesses);
      if(conMan->prepareProcs(processSlots)){
	ERRLOG("Error creating slave processes" << std::endl);
      }

      for(size_t i=0; i<processSlots.size(); i++){
	if(processSlots[i] < 0) {
	  ERRLOG("Could not spawn enough slaves" << std::endl);
	  break;
	}

	DMESG("->> Forking " << i << " slot "
	      << processSlots[i]<< std::endl);

	int pid;
	switch (pid = fork()) {

	case -1:  /* error */
	  PERRLOG("Could not fork new server process");
	  conMan->cleanupSlot(processSlots[i]);
	  break;

	case 0:   /* child process */
	  DMESG("->> Forking child " << i << std::endl);
	  // The following is necessary because we actually have a copy of the
	  // conMan object
	  if(!conMan->setSlave(processSlots[i])) {
	    ret = slaveProcess(serverSocket, conMan);
	  } else {
	    ERRLOG("Unable to start process");
	  }
	  exit(ret);

	default:  /* parent process */
	  children[processSlots[i]]=pid;
	}
      }
    }
  }

  LOG("Shutting down server." << std::endl);

  delete [] children;
  delete [] deadChildren;

  if(logStream.is_open())
    logStream.close();

  return 0;
}

/**
 * Main program
 *
 */
int main(int argc,char *argv[])
{
#if __GNUC__ >= 3
    std::set_terminate(__gnu_cxx::__verbose_terminate_handler);
#endif

  try {
    return amgad(argc, argv);
  } catch (std::exception &ex) {
    ERRLOG("Fatal error: " << ex.what() << endl);
  } catch (...) {
    ERRLOG("Unknown fatal error.");
  }
}
