/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "soap/MDSoapServer.h"

#include "configparser/ConfigParser.h"

#include <iostream>
#include <getopt.h>
#include <stdlib.h>
#include <signal.h>
#include <sql.h>
#include <sqlext.h>

static SQLHENV envHandle;
static SQLHDBC cHandle;

#define OPT_PORT "Port"
#define MDSOAP_DEFAULT_PORT 8822

static bool debug = false;
// Activate debug and send it to cout
#define logstream std::cout
#include "debug.h"
#include "server/MDServerFactory.h"
#include "util/DatabaseConnection.h"

// for ApMon codes
#include <unistd.h>

#include <stdlib.h>
#include <time.h>
#include <netdb.h>
#include "ApMon.h"
#include "utils.h"
#include "string.h"
#include "stdio.h"
#include "proc_utils.h"
#include "mon_constants.h"
#include <iostream>
#include <fstream>

using namespace apmon_utils;

using namespace std;

static ConfigParser configParser;

void printUsage()
{
  std::cout << "Usage: mdsoapserver [options]" << std::endl;
  std::cout << "Options:  -p <port>        : server port" << std::endl;
  std::cout << "          -d               : debug on" << std::endl;
  std::cout << "          -D               : deamon mode" << std::endl;
  std::cout << "          -c <configFile>  : config file" << std::endl;
  std::cout << "          -l <logFile>     : log file" << std::endl;
}

typedef struct {
	string frequency;
	string vendor;
	string modelName;
	string cpuNums;
} CPUINFO;

typedef struct {
	string totalMem;
	string swapMem;
} MEMINFO;

int getNumCPUs() throw(procutils_error) {
#ifdef WIN32
	return 0;
#else
	int numCPUs = 0;
	char line[MAX_STRING_LEN];

	FILE *fp = fopen("/proc/stat", "r");

	if (fp == NULL)
		throw procutils_error("[ getNumCPUs() ] Could not open /proc/stat.");

	while (fgets(line, MAX_STRING_LEN, fp)) {
		if (strstr(line, "cpu") == line && isdigit(line[3]))
			numCPUs++;
	}

	fclose(fp);
	return numCPUs;
#endif
}

void getCPUInfo(CPUINFO *cpuInfo) throw(procutils_error) {
#ifndef WIN32
	double freq = 0;
	char line[MAX_STRING_LEN], s1[100], s2[100], s3[100];
	//  char buf[MAX_STRING_LEN];
	//  char *pbuf = buf;
	char *tmp, *tmp_trim;
	bool freqFound = false, bogomipsFound = false;

	FILE *fp = fopen("/proc/cpuinfo", "r");
	if (fp == NULL)
		throw procutils_error("[ getCPUInfo() ] Could not open /proc/cpuinfo");

	while (fgets(line, MAX_STRING_LEN, fp)) {
		if (strstr(line, "cpu MHz") == line) {
			sscanf(line, "%s %s %s %lf", s1, s2, s3, &freq);
			char temp[256] = { 0x00, };
			sprintf(temp, "%f", freq);
			cpuInfo->frequency = temp;
			freqFound = true;
			continue;
		}

		if (strstr(line, "vendor_id") == line) {
			tmp = strtok/*_r*/(line, ":");//, &pbuf);
			/* take everything that's after the ":" */
			tmp = strtok/*_r*/(NULL, ":");//, &pbuf);
			tmp_trim = trimString(tmp);
			cpuInfo->vendor = tmp_trim;
			// strcpy(apm.cpuVendor, tmp_trim);
			free(tmp_trim);
			continue;
		}

		if (strstr(line, "model name") == line) {
			tmp = strtok/*_r*/(line, ":");//, &pbuf);
			/* take everything that's after the ":" */
			tmp = strtok/*_r*/(NULL, ":");//, &pbuf);
			tmp_trim = trimString(tmp);
			cpuInfo->modelName = tmp_trim;
			//strcpy(apm.cpuModelName, tmp_trim);
			free(tmp_trim);
			continue;
		}
	}
	fclose(fp);
	if (!freqFound)
		throw procutils_error(
				"[ getCPUInfo() ] Could not find frequency or bogomips in /proc/cpuinfo");
#endif
}

void getSysMem(MEMINFO *memInfo) throw(procutils_error) {
#ifndef WIN32
	char s1[20], line[MAX_STRING_LEN];
	bool memFound = false, swapFound = false;
	char valMem[255] = { 0x00, }, valSwap[255] = { 0x00, };
	FILE *fp1;

	fp1 = fopen("/proc/meminfo", "r");
	if (fp1 == NULL)
		throw procutils_error("[ getSysMem() ] Could not open /proc/meminfo");

	while (fgets(line, MAX_STRING_LEN, fp1)) {
		if (strstr(line, "MemTotal:") == line) {
			sscanf(line, "%s %s", s1, valMem);
			memFound = true;
			continue;
		}

		if (strstr(line, "SwapTotal:") == line) {
			sscanf(line, "%s %s", s1, valSwap);
			swapFound = true;
			continue;
		}

	}
	fclose(fp1);

	if (!memFound || !swapFound)
		throw procutils_error(
				"[ getSysMem() ] Could not obtain memory info from /proc/meminfo");
	memInfo->totalMem = valMem;
	memInfo->swapMem = valSwap;
#endif
}

string getKernelInfo() throw(procutils_error) {
#ifndef WIN32
	string temp;
	string kernelInfo;
	ifstream fin;

	char szLine[255] = { 0 };
	bool osFound = false;
	try {
		fin.open("/proc/version");
		fin.getline(szLine, 255);

		kernelInfo = szLine;

		std::string::size_type findOffsetStart = 0, findOffsetEnd = 0;
		findOffsetStart = kernelInfo.find_first_of(")");
		kernelInfo = kernelInfo.substr(0, findOffsetStart+1);
		osFound = true;
	} catch (ifstream::failure e) {
		cout << "Exception opening/reading file";
		fin.close();
		throw procutils_error("[ getKernelInfo() ] Could not open /proc/version");
	}
	fin.close();

	if (!osFound)
		throw procutils_error(
				"[ getKernelInfo() ] Could not obtain Kernel info from /proc/version");

	return kernelInfo;
#endif
}

void* sendInformation2(void* data);
int sendInformation() {
  pthread_t thread;
  int data;
  int ret = pthread_create(&thread, 0, sendInformation2, (void*) &data );
  if(ret) {
    ERRLOG("Could not create thread to handle ApMon" << std::endl);
    return -1;
  }
}

void* sendInformation2(void* data) {

	char *destList[2] = {
			(char *) "http://kenobi.kisti.re.kr/amga_monalisa.conf",
			(char *) "kenobi.kisti.re.kr kisti-amga" };

	char logmsg[100];
	int timestamp;

	int parameterLen = 3;
	string parameterName[10] = { "Version", "AMGA Server Type", "IP Address",
			"CPU Frequency", "CPU Vendor", "CPU Model", "# of CPUs",
			"Total Memory", "Swap Memory", "Kernel Info" };
	string parameterValue[10] = { "", };

	try {
		ApMon::setLogLevel((char *) "FATAL");
		ApMon *apm = new ApMon(2, destList);

		string hostName, nodeName;
		char buffer[256] = { 0x00, };
		const char* default_hostName = "Unknown HostDomain";
		const char* localHostName = "localhost";

		// Finding HostName
		int len = 256;
		int ret = gethostname(buffer, len);

		if (ret != 0) {
			strncpy(buffer, default_hostName, strlen(default_hostName));
		}

		if (strstr(buffer, localHostName) != NULL) {
			strncpy(buffer, default_hostName, strlen(default_hostName));
		}

		hostName = buffer;

		// Get IP Address
		int s;
		struct ifreq ifr = { };
		s = socket(PF_INET,SOCK_DGRAM, 0);
		strncpy(ifr.ifr_name, "eth0", sizeof(ifr.ifr_name));
		string ipAddress;
		if (ioctl(s, SIOCGIFADDR, &ifr) >= 0) {
			char *ip = inet_ntoa(
					((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr);
			ipAddress = ip;
		}

		// Get CPU Info
		CPUINFO cpuInfo;

		try {
			getCPUInfo(&cpuInfo);
			std::ostringstream osstream;
			osstream << getNumCPUs();
			cpuInfo.cpuNums = osstream.str();
		} catch (procutils_error &perr) {
			/* "permanent" error (the parameters could not be obtained) */
			logger(WARNING, perr.what());
		} catch (runtime_error &err) {
			/* temporary error (next time we might be able to get the paramerers) */
			logger(WARNING, err.what());
		}

		// Get Memory Info

		MEMINFO memInfo;
		try {
			getSysMem(&memInfo);
		} catch (procutils_error &perr) {
			/* "permanent" error (the parameters could not be obtained) */
			logger(WARNING, perr.what());
		} catch (runtime_error &err) {
			/* temporary error (next time we might be able to get the paramerers) */
			logger(WARNING, err.what());
		}
		string kernelInfo;
		try {
			kernelInfo = getKernelInfo();
		} catch (procutils_error &perr) {
			/* "permanent" error (the parameters could not be obtained) */
			logger(WARNING, perr.what());
		} catch (runtime_error &err) {
			/* temporary error (next time we might be able to get the paramerers) */
		}

		nodeName = buffer;
		nodeName += " - " + ipAddress;


		sprintf(logmsg, "NodeName : %s \n", nodeName.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "HostName : %s \n", hostName.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "VERSION : %s \n", VERSION);
		logger(INFO, logmsg);
		sprintf(logmsg, "IP Address : %s \n", ipAddress.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "CPU Frequency: %s \n", cpuInfo.frequency.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "CPU Vendor : %s \n", cpuInfo.vendor.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "CPU Model Name : %s \n", cpuInfo.modelName.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "# of CPUs : %s \n", cpuInfo.cpuNums.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "Toatal Memory : %s \n", memInfo.totalMem.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "Swap Memory : %s \n", memInfo.swapMem.c_str());
		logger(INFO, logmsg);
		sprintf(logmsg, "Kernel Info : %s \n", kernelInfo.c_str());
		logger(INFO, logmsg);

		sprintf(logmsg, "Sending Information\n");
		logger(INFO, logmsg);

		/* use the wrapper function with simplified syntax */
		try {
			/* send a datagram with timestamp */
			timestamp = time(NULL);

			parameterValue[0] = VERSION;
			parameterValue[1] = "mdsoapserver";
			parameterValue[2] = ipAddress;
			/*
			parameterValue[3] = cpuInfo.frequency + " MHz";
			parameterValue[4] = cpuInfo.vendor;
			parameterValue[5] = cpuInfo.modelName;
			parameterValue[6] = cpuInfo.cpuNums;
			parameterValue[7] = memInfo.totalMem + " KB";
			parameterValue[8] = memInfo.swapMem + " KB";
			parameterValue[9] = kernelInfo;
			*/
			for (int i = 0; i < parameterLen; i++) {
				apm -> sendTimedParameter((char *) "AMGA_SERVER",
						(char *) nodeName.c_str(),
						(char *) parameterName[i].c_str(), XDR_STRING,
						(char *) parameterValue[i].c_str(), timestamp);
			}

		} catch (runtime_error &e) {
			logger(WARNING, e.what());
		}

		delete apm;
	} catch (runtime_error &e) {
		logger(WARNING, e.what());
	}
	return 0;
}


int setupODBC()
{
  DMESG("itializing ODBC" << endl);
  long res;

  // Allocate Environment handle and register version
  res=SQLAllocHandle(SQL_HANDLE_ENV,SQL_NULL_HANDLE,&envHandle);
  if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)){
    std::cerr << "mdservermt: Cannot initialize ODBC: "
       << res << std::endl;
    return 1;
  }
  res=SQLSetEnvAttr(envHandle, SQL_ATTR_ODBC_VERSION,
        (void*)SQL_OV_ODBC3, 0);
  if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)){
    std::cerr << "mdservermt: ODBC version mismatch: "
       << res << std::endl;
    goto err_no_env;
  }

  // Allocate connection handle, set timeout
  res = SQLAllocHandle(SQL_HANDLE_DBC, envHandle, &cHandle);
  if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
    std::cerr << "mdservermt: Cannot allocate connection handle: "
        << res << std::endl;
    goto err_no_env;
  }

  return 0;

 err_no_env:
  SQLFreeHandle(SQL_HANDLE_ENV, envHandle);
  return 1;
}

int main(int argc,char *argv[])
{

  static struct option options[] = {
    {"config", 1, 0, 'c'},
    {"help", 0, 0, 'h'},
    {"debug", 0, 0, 'd'},
    {"daemon", 0, 0, 'D'},
    {"port", 1, 0, 'p'},
    {"log", 1, 0, 'l'},
    {0, 0, 0, 0}
  };

  std::string configLocation="";
  std::string logFile="";
  int daemon = 0;
  int optionIndex=0;
  int port=0;
  std::ofstream logStream;


  while(1){
    int c=getopt_long(argc, argv, "hp:dc:l:D", options, &optionIndex);
    if(c==-1)
      break;

    switch (c) {
    case 'h':
      printUsage();
      exit(0);
    case 'd':
      debug=true;
      break;
    case 'D':
      daemon=1;
      break;
    case 'p':
      port=atoi(optarg);
      break;
    case 'c':
      configLocation=optarg;
      break;
    case 'l':
      logFile=optarg;
      break;

    default:
      std::cerr << "Unknown option" << (char)c << std::endl;
      printUsage();
      exit(5);
    }
  }

  if(optind!=argc){
    std::cout << "Too many parameters!" << std::endl;
    printUsage();
    exit(5);
  }


        ConfigParser::configParser = new ConfigParser();

        if (configLocation != "") {
                // User provided an explicit configuration file.
                if (!ConfigParser::configParser->init(configLocation)) {
                        ERRLOG("Could not read configuration: " << configLocation << std::endl);
                        exit(5);
                }
        } else {
                // Load the default configuration file
                try {
                        configLocation = ConfigParser::loadFromFile("amgad.config",
                                        *ConfigParser::configParser);
                } catch (runtime_error &ex) {
                        ERRLOG("Configuration file not found on the default search paths." << std::endl);
                        exit(5);
                }
        }

	// KIMNG090716 Check the mode of report information
	try {
		string report = ConfigParser::configParser->getString("Report");

		if (report.compare("disable") == 0) {
			// LOG("Report : " << report << std::endl);
		}
		else {
			// KIMNG090712 ApMon (Monalisa Client codes)
			sendInformation();
			// LOG("Report : " << report << std::endl);
		}
	} catch (std::string err) {
		sendInformation();
		// LOG("Report : enable " << std::endl);
	}

  // Now let's become a daemon, if wanted
  if (daemon) {
    // Our process ID and Session ID
    pid_t pid, sid;

    // Fork off the parent process
    pid = fork();
    if (pid < 0) {
            exit(EXIT_FAILURE);
    }

    // We are the parent, and can die
    if (pid > 0) {
            exit(EXIT_SUCCESS);
    }

    // Change the file mode mask: full access to our own files
    umask(0);

    // Create a new session id (SID) for the child process
    sid = setsid();
    if (sid < 0) {
            ERRLOG("Could not get new session ID for daemon proces" << std::endl);
            exit(EXIT_FAILURE);
    }

    // change working dir to a safe place
    if ((chdir("/")) < 0) {
            ERRLOG("Could not chdir /");
            exit(EXIT_FAILURE);
    }

    // Close stdin, should be doing it also with stdout, but only if we have
    // a logfile set up
    close(STDIN_FILENO);
    //  close(STDOUT_FILENO);
    //  close(STDERR_FILENO);
   }

  if (configLocation != "") {
  	// User provided an explicit configuration file.
    if(!configParser.init(configLocation)) {
      ERRLOG("Could not read configuration: " << configLocation << std::endl);
      exit(5);
    }
  } else {
  	// Try the defaults. First, current directory
  	configLocation = "mdsoapserver.config";
  	if(!configParser.init(configLocation)) {
  		// Now home directory.
		configLocation=getenv("HOME");
		configLocation.append("/.mdsoapserver.config");
  		if(!configParser.init(configLocation)) {
  			// Now global configuration
		    const char *gliteLocation = getenv("GLITE_LOCATION");
			configLocation = "";
		  	if(gliteLocation)
		    	configLocation = getenv("GLITE_LOCATION");
		  	configLocation.append("/etc/mdsoapserver.config");
			if(!configParser.init(configLocation)) {
		      ERRLOG("Could not read configuration: exiting" << std::endl);
		      exit(5);
		    }
  		}
  	}
  }

  if (logFile.size()) {
    logStream.open(logFile.c_str());
    Display::setOutput(logStream.rdbuf());
  }

  LOG("Loaded configuration: " << configLocation << std::endl);

  DatabaseConnection::initialize(configParser);

  if(port==0) { // Not set on command line
    try {
      port=configParser.getInt(OPT_PORT);
    } catch (std::string err) {
      port=MDSOAP_DEFAULT_PORT;  // Neither set on command line nor in configuration
    }
  }

  if(setupODBC())
    exit(5);

	MDSoapServer::setDebug(debug);

  // Start the server
	MDSoapServer soapServer(port, configParser);
	soapServer.startServing();


  // With processes, the traditional way to cleanup after
  // a CTRL^C is to use a signal handler to catch the signal
  // and perform the cleanup. When using threads this solution
  // is not safe since the thread receiving the signal might
  // have been interrupted inside a function that cannot be
  // safely re-entered (other thread could use that function).
  // So the correct way to handle this situation is to make the
  // main thread wait explicitly for the termination signals
  // using sigwait()
//   cout << "Sleeping\n";
	while (true) {
		sleep(1);
	}
  // Blocks the signals that will be used to terminate
  // the server

//   sigset_t set;
//   sigemptyset(&set);
//   sigaddset(&set, SIGINT);
//   sigaddset(&set, SIGHUP);
//   sigaddset(&set, SIGTERM);
//   pthread_sigmask(SIG_SETMASK, &set, NULL);
//
// //   // Wait for the termination request (CTRL^C)
//   int acceptedSignal;
//   if (sigwait(&set, &acceptedSignal)) {
// 	PERRLOG("sigwait error");
//   }
//   soapServer.stopServing();

 if (logStream.is_open())
   logStream.close();

}
