/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "QParser.h"

extern int yyparse(void *);
extern int yydebug;

#include <debug.h>
#include "Distinct.h"
#include "Join.h"

QParser::QParser(const std::string &prefix, bool dbg)
  : pref(prefix), debug(dbg)
{
  scanner.rootExpression=NULL;
  
  if(debug){
    scanner.lexer.set_debug(1);
    yydebug = 1;
  } else {
    scanner.lexer.set_debug(0);
    yydebug = 0;
  }
}
  

QParser::~QParser()
{
  delete scanner.rootExpression;

  std::list< QueryParser::Constraint *>::iterator I 
    = scanner.constraints.begin();
  while ( I != scanner.constraints.end() ){
    delete (*I);
    I = scanner.constraints.erase(I);
  }
}


int QParser::parse(const std::string &query, std::string &result,
		   bool plain)
{
  DMESG("Parsing: " << query << std::endl);
  if(parse(query))
    return -1;
  result="";
  return toSQL(result, plain);
}


int QParser::parse(const std::string &query)
{
  delete scanner.rootExpression;
  std::list< QueryParser::Constraint *>::iterator I = scanner.constraints.begin();
  while ( I != scanner.constraints.end() ) {
    delete (*I);
    I = scanner.constraints.erase(I);
  }
  scanner.rootExpression=NULL;

  std::istringstream in(query);
  scanner.lexer.switch_streams(&in, NULL);
  if(yyparse(&scanner) > 0 ){
    DMESG("Parsing failed for: " << query << std::endl);
    scanner.lexer.flush();
    return 1;
  }

  scanner.lexer.flush();
  return 0;
}

void QParser::addToTableMap(TableMap &tableMap)
{
  if(scanner.rootExpression)
    scanner.rootExpression->addToTableMap(tableMap);

  std::list< QueryParser::Constraint *>::iterator I 
    = scanner.constraints.begin();
  while ( I != scanner.constraints.end() ) {
    (*I)->addToTableMap(tableMap);
    I++;
  }
}


void QParser::replaceTableNames(TableMap &tableMap)
{
  if(scanner.rootExpression)
    scanner.rootExpression->replaceTableNames(tableMap);

  std::list< QueryParser::Constraint *>::iterator I 
    = scanner.constraints.begin();
  while ( I != scanner.constraints.end() ) {
    (*I)->replaceTableNames(tableMap);
    I++;
  }
}


int QParser::toSQL(std::string &result, bool plain)
{
  if(!scanner.rootExpression){ // Check for empty parse-result
    DMESG("Cannot convert to SLQ: empty result\n");
    return -1;
  }
  if( plain )
    scanner.rootExpression->toSQL("", result, engine);
  else
    scanner.rootExpression->toSQL(pref, result, engine);
  return 0;
}


int QParser::getConstraint(std::string &result, bool plain)
{
  std::list< QueryParser::Constraint *>::iterator I 
    = scanner.constraints.begin();
  std::string prefix;
  if(!plain) prefix = pref;
 
  while ( I != scanner.constraints.end() ) {
    if( !dynamic_cast<QueryParser::Join *> (*I) )
      (*I)->toSQL(prefix, result, engine);
    I++;
  }

  return 0;
}


int QParser::getJoin(std::string &result, bool plain)
{
  std::list< QueryParser::Constraint *>::iterator I 
    = scanner.constraints.begin();
  std::string prefix;
  if(!plain) prefix = pref;
 
  while ( I != scanner.constraints.end() ) {
    if( dynamic_cast<QueryParser::Join *> (*I) ) {
      (*I)->toSQL(prefix, result, engine);
    }
    I++;
  }
  return 0;
}


bool QParser::distinctOnly() const
{
  std::list< QueryParser::Constraint *>::const_iterator I 
    = scanner.constraints.begin();

  while ( I != scanner.constraints.end() ) {
    if( dynamic_cast<QueryParser::Distinct *> (*I) )
      return true;
    I++;
  }

  return false;
}
