/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: testParser.cc,v 1.7 2010/04/14 07:15:30 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include "ScannerContext.h"
#include "Distinct.h"
#include "Join.h"
#include <sstream>
#include <fstream>
#include <exception>

#include <getopt.h>
#include <stdlib.h>


extern int yyparse(void *);
extern int yydebug;

void printUsage()
{
  std::cout << "Usage: testParser [options]" << std::endl;
  std::cout << "Options:  -d               : debug on" << std::endl;
  std::cout << "          -v               : verbose mode" << std::endl;
  std::cout << "          -h               : help message" << std::endl;
  std::cout << "          -c <configFile>  : config file" << std::endl;
}


int main (int argc, char *argv[])
{

  // Inititalize options to defaults
  bool debug = false;
  bool verbose = false;
  std::string configFile("testparser.cfg");
  std::string command;

  std::set_terminate (__gnu_cxx::__verbose_terminate_handler);
  
  static struct option options[] = {
    {"help", 0, 0, 'h'},
    {"debug", 0, 0, 'd'},
    {"verbose", 0, 0, 'v'},
    {"config", 1, 0, 'c'},
    {"exec", 1, 0, 'e'},
    {0, 0, 0, 0}
  };

  int optionIndex=0;
  while(1){
    int c=getopt_long(argc, argv, "hdvc:e:", options, &optionIndex);

    if(c==-1)
      break;

    switch (c) {
    case 'h':
      printUsage();
      exit(0);
    case 'd':
      debug = true;
      break;
    case 'v':
      verbose = true;
      break;
    case 'e':
      command = optarg;
      break;
    case 'c':
      configFile = optarg;
      break;
    default:
      std::cerr << "Unknown option" << (char)c << std::endl;
      printUsage();
      exit(5);
    }
  }

  if(optind!=argc){
    std::cerr << "Too many parameters!" << std::endl;
    printUsage();
    exit(5);
  }  

  std::ifstream iF(configFile.c_str());
  iF.exceptions (std::ios_base::badbit);
  
  if (! iF.is_open() ) {
    std::cerr << "Error opening file" << std::endl;
    exit(5);
  }

  std::string test, result;
  
  while ( !iF.eof() ) {
    getline(iF, test);
    getline(iF, result);
    //    std::cout << "Read: >" << test << "< and >" << result <<"<\n";
  
    std::cout << "Parsing " << test << std::endl;

    ScannerContext scanner;
    scanner.rootExpression = NULL;
    
    std::istringstream in(test);
    scanner.lexer.switch_streams(&in, NULL);
    if( debug )
      scanner.lexer.set_debug(1);
    if( verbose )
      yydebug = 1;
    
    if(yyparse(&scanner)>0) {
      std::cout << "  ++++ Error parsing! ";
      if( result == "--" )
	std::cout << "\033[32m(OK)\033[0m\n";
      else
	std::cout << "\033[31m(FAILED)\033[0m\n";
    } else {
      std::string query;
      if( scanner.rootExpression )
	scanner.rootExpression->toSQL("user:", query);
      
      std::list< QueryParser::Constraint *>::iterator I = scanner.constraints.begin();
      while ( I != scanner.constraints.end() ) {
	(*I)->toSQL("user:", query);
	if( dynamic_cast<QueryParser::Distinct *> (*I) ) 
	  std::cout << "DISTINCT DETECTED!\n";
	if( dynamic_cast<QueryParser::Join *> (*I) ) 
	  std::cout << "JOIN DETECTED!\n";
	I++;
      }

      std::cout << "  >>>> " << query << " ";
      if( result == query )
	std::cout << "\033[32m(OK)\033[0m\n";
      else {
	std::cout << "\033[31m(FAILED)\033[0m\n";
	std::cout << "Should be: >" << result << "<\n";
      }
      delete scanner.rootExpression;
      I = scanner.constraints.begin();
      while ( I != scanner.constraints.end() )
	delete (*I++);
    }
  }

  return 0;
}
