/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "CommandParser.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include "ReplicationManager.h"

#include <iostream>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/classification.hpp>

using namespace std;
using namespace boost;

#include "debug.h"


CommandParser::CommandParser(TCPSocket &socket) 
: _socket(socket)
{
  debug = ReplicationManager::getInstance()->getDebug();
  _commands["subscribe"] = Command::cSubscribe;
  _commands["unsubscribe"] = Command::cUnsubscribe;
  _commands["subscribe_users"] = Command::cSubscribeUsers;
  _commands["unsubscribe_users"] = Command::cUnsubscribeUsers;
  _commands["unsubscribe_all"] = Command::cUnsubscribeAll;
  _commands["synchronize"] = Command::cSynchronize;
  _commands["receive"] = Command::cReceive;
  _commands["quit"] = Command::cQuit;
}

CommandParser::~CommandParser()
{
}

/**
 * To be used in case of error.
 * @throws SocketException
 */
void CommandParser::skipToEndOfCommand() throw (SocketException) {
    string line;
    while ((line = _socket.readLine()) != "") {
        DMESG(">" << line << "<\n");
    }
}
    
/**
 * throws SocketException if peer disconnects
 */
void CommandParser::nextCommand(Command &cmd) throw (SocketException, CommandParseException) {
    string line = _socket.readLine();
    if (line == "") {
        throw CommandParseException("Command terminated unexpectadely");
    }

    // Split the string into tokens ( use '-' and '*' as delimiters )
    // We need copies of the input only, and adjacent tokens are compressed
    vector<string> tokens;
    split(tokens, line, is_any_of(" "), token_compress_on);
    
    // Is it a known command?
    if (_commands.count(tokens[0]) != 1) {
        DMESG("Unrecognized token. Skipping command\n");
        // Unrecognized command
        skipToEndOfCommand();
        throw CommandParseException("Invalid command: " + tokens[0]);
    }
    DMESG("Command: " << tokens[0] << endl);

    // It is a known command
    cmd.cId = _commands[tokens[0]];
    // The mandatory arguments follow the command on the same line
    cmd.args.resize(tokens.size()-1);
    copy(tokens.begin()+1, tokens.end(), cmd.args.begin());

    // Tokens
    if (debug) {
      ostream_iterator<string> ofile(cout, ",");    
      DMESG("Parameters: ");
      copy(cmd.args.begin(), cmd.args.end(), ofile); DMESG(endl);
    }

    // The optional arguments are each on a separate line and
    // end with an empty line. key, value pairs.
    while (true) {
        line = _socket.readLine();
        DMESG("Got line >" << line << "<\n");
        if (line == "") {
            // End of command
            return;
        }
        tokens.clear();
        split(tokens, line, is_any_of("="), token_compress_on);
        if (tokens.size() != 2) {
            skipToEndOfCommand();
            throw CommandParseException("Invalid optional parameter: " + line);
        }
        trim(tokens[0]);
        trim(tokens[1]);
        cmd.optionalArgs[tokens[0]] = tokens[1]; 
    }
}


ostream& operator<<(ostream& out, const Command &cmd) {
    out << cmd.cId << " ";
    ostream_iterator<string> ofile(out, " ");
    copy(cmd.args.begin(), cmd.args.end(), ofile);
    out << endl;
    
    map<string, string>::const_iterator it;
    for (it = cmd.optionalArgs.begin(); 
        it != cmd.optionalArgs.end(); it++) 
    {
        out << (*it).first << " = " << (*it).second << endl;
    }
    return out;
}
