/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "LogMonitorThread.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "debug.h"
#include "server/MDServer.h"
#include "util/Statement.h"
#include "util/DatabaseConnection.h"
#include "util/Utils.h"
#include "util/Socket.h"
#include "ReplicationDaemon.h"
#include "ReplicationManager.h"
#include "MDLogWriter.h"
#include "util/StatsCollector.h"

#include <limits>
#include <iosfwd>
#include <ctime>
#include <set>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/xtime.hpp>



const char LOG_SEP = '\x1E';

/* Duration of sleep intervals for different 
 * activities that use polling. In seconds
 */
// How long to wait between cleaning old logs
const int LOG_GC_SLEEP_INTERVAL = 10;
// Pause between checking for new logs if no new logs
// were found in the previous check
const int LOG_CHECK_MIN_SLEEP_INTERVAL = 2;
// Pause between checking for new logs if the previous 
// query found new logs
const int LOG_CHECK_MAX_SLEEP_INTERVAL = 8;

using namespace std;
using namespace boost;

const string LOG_TEST_CONNECTION("@@ping@@");
const string COUNTER_WRITE_XID("repdaemon.WriteXid");
const string COUNTER_SESSIONS_TOTAL("repdaemon.Sessions.Total");
const string COUNTER_SESSIONS_CURRENT("repdaemon.Sessions.Current");
const string COUNTER_LOGS_SENT("repdaemon.LogsSent.Total");
const string COUNTER_DATA_SENT("repdaemon.DataSent.Total"); 
const string COUNTER_SESSION("repdaemon.Session.");
const string COUNTER_LOGS_PENDING("repdaemon.LogsPending");

const static string LOG_PERMISSIONS[] = {
	"chown", "chmod", 
	"acl_remove", "acl_add", 
	"rep_allow", "rep_disallow"
};
const static int LOG_PERMISSIONS_SIZE 
	= sizeof(LOG_PERMISSIONS)/sizeof(string*);

const static string LOG_USERS[] = { 
	"user_create", "user_remove", 
	"user_password_change",
	"user_subject_add",	"user_subject_remove",
	"grp_create", "grp_delete",
	"grp_adduser", "grp_removeuser"
};
const static int LOG_USERS_SIZE 
	= sizeof(LOG_USERS)/sizeof(string*);


// This should be executed at startup.
static time_t startupTime = time(NULL);

/*
 * Since this thread will run forever, don't delete it on exit.
 */
LogMonitorThread::LogMonitorThread(int sleepInterval, const ConfigParser &configParser) :
Runnable(false), _sleepInterval(sleepInterval)
{
  debug = ReplicationManager::getInstance()->getDebug();
  // The DB connection used for all queries by this object
  _dbConn.reset(DatabaseConnection::openConnection());
  // Uses the same DB connection
  _subMan.reset(new SubscriptionManager(*_dbConn));
  
  // Initialize the counters exported by this module
  StatsCollector::registerStatsSource(*this);
  if (!StatsCollector::hasCounter(COUNTER_SESSIONS_CURRENT)) {
    StatsCollector::setCounter(COUNTER_SESSIONS_CURRENT, 0);
    StatsCollector::setCounter(COUNTER_SESSIONS_TOTAL, 0);    
    StatsCollector::setCounter(COUNTER_LOGS_SENT, 0);
    StatsCollector::setCounter(COUNTER_DATA_SENT, 0);
  }
  setupTableNames();
  setConfig(configParser);
}

LogMonitorThread::~LogMonitorThread() {
  StatsCollector::unregisterStatsSource(*this);
}


void LogMonitorThread::setupTableNames()
{
  std::string prefix;
  if(DatabaseConnection::dbSchema.size())
    prefix = DatabaseConnection::dbSchema + ".";

  amgaTable = prefix + "amga";
  sequencesTable = prefix + "sequences";
  transactionsTable = prefix + "transactions";
  logsTable = prefix + "logs";
}


void LogMonitorThread::collectCounters(std::map<std::string,
				       std::string> &counters)
{
  time_t now = time(NULL);
  
  counters["repdaemon.Uptime"] = toString((int)(now - startupTime));
  counters["NodeName"] = ReplicationManager::getInstance()->getNodeName();

  // Get the counters from the active subscribers
  mutex::scoped_lock lk(_activeSubscribersMutex);
  for (map<string, shared_ptr<ActiveSubscription> >::const_iterator it = activeSubscribers.begin();
    it != activeSubscribers.end();    
    ++it)
  {
    ActiveSubscription *aSub = it->second.get();
    string subscriber = aSub->subscriber->subscriberID;
    string aux("repdaemon.Session." + subscriber);
    counters[aux + ".LogsSent"] = toString(aSub->logsSent);
    counters[aux + ".DataSent"] = toString(aSub->dataSent);
    counters[aux + ".AckedXid"] = toString(aSub->ackedXid);
    counters[aux + ".SentXid"] = toString(aSub->sentXid);
    counters[aux + ".SessionID"] = toString(aSub->sessionID);
    int uptime = (int)(now - aSub->startupTime);
    counters[aux + ".Uptime"] = uptime == 0? "1" : toString(uptime);
  }
  
  auto_ptr<DatabaseConnection> dbConn(DatabaseConnection::openConnection());
  // Uses the same DB connection
  SubscriptionManager subMan(*dbConn);
  // Add the ones that are disconnected
  map<string, Subscriber> subs = subMan.getAllSubscribers();
  for (map<string, Subscriber>::const_iterator it = subs.begin();
    it != subs.end();
    ++it) 
  {
    if (activeSubscribers.count(it->first) == 0) {
      // Not active
      string subscriber = it->first;
      string aux("repdaemon.Session." + subscriber);
      // -1 means disconnected.
      counters[aux + ".SentXid"] = counters[aux + ".AckedXid"] = toString(it->second.lastAckedXid);
      counters[aux + ".Uptime"] = "-1";
      counters[aux + ".LogsSent"] = "-1";
      counters[aux + ".DataSent"] = "-1";
      counters[aux + ".SessionID"] = "-1";
    }
  }
  try {
    counters[COUNTER_LOGS_PENDING] = toString(getLogCount(*dbConn));
  } catch (SQLException &ex) {
    DMESG("WARNING: " << ex.what() << endl);
  }
  
  Statement st(*dbConn);
  // Only works with PostgreSQL
  string query("SELECT \"last_log_xid\" FROM " + sequencesTable + ";");
  DMESG("SQL: " << query << endl);
  if (st.exec(query)) {
    DMESG("Error getting last log xid: " << st.getError() << endl);
    counters[COUNTER_WRITE_XID] = -1;
  } else {
    if (st.fetch()) {
      DMESG("Error getting last log xid: " << st.getError() << endl);
    }
    uint64_t xid;
    if (st.getData(1, xid)) {
      DMESG("Error getting last log xid: " << st.getError() << endl);
    }
    counters[COUNTER_WRITE_XID] = toString(xid);
  }
}

void LogMonitorThread::run() {
  LOG("Cleanup thread starting." << endl);
  while (!isCancelled()) {
    try {
        doLogCleanup();
    } catch (SQLException &ex) {
        LOG("Error deleting logs: " << ex.what() << endl);
    }
    try {
      sleep(LOG_GC_SLEEP_INTERVAL);    
    } catch (runtime_error &ex) {
      DMESG("Error sleeping: " << ex.what() << endl);
      break;
    }
  }
  DMESG("[LogMonitorThread] Finished\n");
}


//vector<Subscription> generateTestCase(){
//    vector<Subscription> subs;
////    subs.resize(5);
////    subs[0].directory = "/";
////    subs[0].curXid = 10;
////    subs[1].directory = "/l1_a";
////    subs[1].curXid = 20;
////    subs[2].directory = "/l1_b";
////    subs[2].curXid = 20;
////    subs[3].directory = "/l1_a/l2_a";
////    subs[3].curXid = 15;
////    subs[4].directory = "/l1_a";
////    subs[4].curXid = 12;
//    subs.resize(5);
//    subs[0].directory = "/";
//    subs[0].curXid = 50;
//    subs[1].directory = "/l1_a";
//    subs[1].curXid = 30;
//    subs[2].directory = "/l1_a/l1_b";
//    subs[2].curXid = 100;
//    subs[3].directory = "/l1_a/l1_a";
//    subs[3].curXid = 50;
//    subs[4].directory = "/l1_a/l1_b";
//    subs[4].curXid= 32;
//
//
//    random_shuffle(subs.begin(), subs.end());
//    return subs;
//}

int LogMonitorThread::execUpdate(Statement &st, const std::string &update) const 
  throw (SQLException) {
  DMESG("Query:" << update << endl);
  if (st.exec(update)) {
    if (st.getErrorCode() != SQL_NO_DATA) {
      throw SQLException(st);
    }
  }
  int rowsDeleted = st.numRows();
  st.close();
  DMESG("Rows deleted: " << rowsDeleted << endl);
  return rowsDeleted;
}

uint64_t LogMonitorThread::getMinXidForDir(const std::map<std::string, uint64_t> &minXids, std::string dir) const 
{
  while (true) {
//    cout << "Trying: " << dir << endl;
    std::map<std::string, uint64_t>::const_iterator it = minXids.find(dir);
    if (it != minXids.end()) {
      return it->second;
    }
    if (dir == "/") {
      // We are at the top. 
      return 0;
    }
    // Go up one level.
    size_t index = dir.rfind('/');
    dir.erase(index);
    if (dir.empty()) {
      dir = "/";
    }
  }
}

void LogMonitorThread::doLogCleanup() throw (SQLException) {
  DMESG("Purging old logs\n");
  // 1. Get list of all subscribers
  Statement st(*_dbConn);
  if (st.beginTransaction()) {
    throw SQLException(st);
  }
  vector<Subscription> subs = _subMan->getSubscriptions();
  map<string, Subscriber> subscribers = _subMan->getAllSubscribers();
//    vector<Subscription> subs = generateTestCase();

  numeric_limits<uint64_t> uint64_info;
  uint64_t globalMinXid = uint64_info.max();
    
  // First check what is the latest log that a subscriber 
  // of users/groups still needs.
  map<string, uint64_t> minUserMountsXids;
  for (map<string, Subscriber>::const_iterator it = subscribers.begin();
  	it != subscribers.end();
  	++it) 
  {
  	if (it->second.repUsers) {
  	  globalMinXid = it->second.lastAckedXid;
  	}
  }

  if (subs.empty() && globalMinXid == uint64_info.max()) {
    // No subscription and no slave is replicating users.
    // delete the logs and the transctions tables
    // TODO: Must have a lock on the subscriptions so that
    // no new subscriptions are added after checking the list of subscriptions
    // and before deleting the full tables.
    execUpdate(st, "DELETE FROM logs;");
    
  } else {
    int rowsDeleted = 0;
    bool oldDebug = debug;
    bool debug = false;
    
    // For each directory, keeps track of the minimum xid 
    // that all the subscriber have already received
    map<string, uint64_t> minXids;
    // 2. For each directory subscribed by some slave, get 
    // the minimum xid that was acknowledged by all subscribers. This
    // step is not enough since each subscription also includes all
    // subdirectories, so if a top level subscription only has xids
    // up to 5, then all the sub-directories should also preserve
    // any log with xid greater than 5.
    for (vector<Subscription>::const_iterator it = subs.begin(); it != subs.end(); ++it) {
      uint64_t lastAckedXid = subscribers[it->subscriberID].lastAckedXid;
      DMESG(it->directory << ": " << lastAckedXid << endl);
      globalMinXid = min(globalMinXid, lastAckedXid);
      if (minXids.count(it->directory) == 0) {
        // The directory is not on the map. Insert it
        minXids[it->directory] = lastAckedXid;
      } else {
        // It's already on the map. Update the value with the minimum
        // of the two values
        minXids[it->directory] = min(lastAckedXid, minXids[it->directory]);
      }
    }

    DMESG("minXids evaluated. Global: " << globalMinXid << endl);
    for (map<string, uint64_t>::const_iterator it = minXids.begin(); it != minXids.end(); ++it) {
      DMESG(it->first << ": " << it->second << endl);
    }

    // The map sorts the keys automatically, so if we iterate over it
    // we will see it ordered.    
    for (map<string, uint64_t>::iterator it = minXids.begin(); it != minXids.end(); ++it) {
      // Add the / so that directory /abc does not match directory /abcd
      string directory = it->first;
      if (!ends_with(directory, "/")) {
        directory += "/";
      }
      uint64_t xid = it->second;
      DMESG("Propagating " << directory << ": " << xid << endl);
      map<string, uint64_t>::iterator itAux = it;
      itAux++;
      while (itAux != minXids.end() && starts_with(itAux->first, directory)) {
        DMESG("  Comparing with " << itAux->first << ":" << itAux->second << endl);
        if (itAux->second > xid) {
          DMESG("  Updating\n");
          itAux->second = xid;
        }
        itAux++;
      }
    }
    
    DMESG("minXids propagated\n");
    for (map<string, uint64_t>::const_iterator it = minXids.begin(); it != minXids.end(); ++it) {
      DMESG(it->first << ": " << it->second << endl);
    }

    // Start deleting stuff
    // Get all the transactions 
    debug = oldDebug;
    
    // All logs bellow or equal globalMinXid can be deleted. Logs with
    // xids greater than that may be deleted depending on 
    // the minXid of their directory. We start by deleting everything
    // older than globalMinXid since it is a simpler query that in the
    // most common case will delete most of the logs that need to be delete,
    // thereby reducing the size of the table.
    
    ostringstream oss;
    oss << "DELETE FROM logs WHERE \"log_xid\" <= " << globalMinXid << ";";
    rowsDeleted += execUpdate(st, oss.str());
    // Transactions may span several directories. To delete all the possible
    // logs, it would be necessary to read all the individual logs of each
    // transaction to evaluate the minXid of all the directories touched
    // by the transaction. That would be too expensive, so we just delete
    // transactions if their xid is older than the global minimum. That
    // was done by the previous query.

    // Uploads have their directory indicated on the command, so we can
    // delete them together with normal logs.
    
    // Keep a list of lower level directories that were already 
    // processed and therefore shouldn't be affected by deletions on
    // of higher-level directories.
    vector<string> processedDirs;
    string dirTree;
    // Start on the bottom of the directories.
    map<string, uint64_t>::reverse_iterator rit;
    for (rit = minXids.rbegin(); rit != minXids.rend(); ++rit) {
//      cout << rit->first << ", " << rit->second << endl;
      string directory = rit->first;
      int xid = rit->second;
      ostringstream oss;
      oss << "DELETE FROM logs WHERE (\"log_xid\" <= " << xid << ") AND ";  
      if (directory == "/") {
        oss << "(\"log_directory\" like '/%')";
      } else {
        dirTree = "(\"log_directory\" = '";
        dirTree += directory;
        dirTree += "' OR \"log_directory\" LIKE '";
        dirTree += directory;
        dirTree += "/%')";
        oss << dirTree;
      }
      // Exclude all directories that were already processed
      for (size_t i=0; i<processedDirs.size(); i++) {
        oss << " AND (NOT " << processedDirs[i] << ")";
      }
      processedDirs.push_back(dirTree);
      oss << ";";
      rowsDeleted += execUpdate(st, oss.str());
    }

    dropOldSubscribers(st, subscribers);
  }
  if (st.commitTransaction()) {
    throw SQLException(st);
  }
}

void LogMonitorThread::dropOldSubscribers(Statement &st, 
					  std::map<string, Subscriber> &subscribers) 
  throw (SQLException) 
{
  int maxPendingLogs = ReplicationManager::getInstance()->getMaxPendingLogs();
  DMESG("MaxPendingLogs: " << maxPendingLogs << "\n");
  if (maxPendingLogs <= 0) {
    // There are no limits to the size of the logs table.
    return;
  }

  DMESG("Checking if log table exceeded the maximum size\n");
  int logCount = getLogCount(st.getDatabaseConnection());
  DMESG("Current size: " << logCount << "/" << maxPendingLogs << "\n"); 
  if (logCount <= maxPendingLogs) {
    // We are bellow the limit
    return;
  }
  DMESG("Maximum size exceeded. Dropping subscriber that has been disconnected for longer.\n");
  numeric_limits<uint64_t> uint64_info;
  uint64_t minXid = uint64_info.max();  
  string oldestSubscriber;
  // Must drop old subscribers. Drop the oldest subscriber
  for (map<string, Subscriber>::const_iterator it = subscribers.begin(); 
    it != subscribers.end(); 
    ++it) 
  {
    mutex::scoped_lock lk(_activeSubscribersMutex);
    
    DMESG("Testing: " << it->second.subscriberID << ":" << it->second.lastAckedXid << "\n");
    // Is this subscriber connected? if so, ignore it.
    if (activeSubscribers.count(it->second.subscriberID) != 0) {
      continue;
    }
    if (it->second.lastAckedXid < minXid) {
      minXid = it->second.lastAckedXid;
      oldestSubscriber = it->second.subscriberID;
    }
  }
  LOG("Dropping subscriber " << oldestSubscriber << "\n");
  _subMan->unsubscribeAll(oldestSubscriber);
}

void LogMonitorThread::shipUploadLogs(const Log &log, 
  ActiveSubscription &aSub, DatabaseConnection &conn) 
  throw (SQLException, SocketException) 
{
  DMESG("Shipping upload: " << log.xid << endl);

  // The first log is the upload command. Send it together with all
  // the context information extrated from the meta-log that signalled
  // the start of this upload  
  ostringstream oss;
  // Send the initial log to open the upload.
  oss << log.xid << LOG_SEP 
    << log.directory << LOG_SEP   // directory
    << log.user << LOG_SEP   // user
    << log.cwd << LOG_SEP   // cwd directory.
    << log.permissions << LOG_SEP   // permission mask
    << log.groupRights << LOG_SEP   // group rights mask
    << log.command << "\n";   // command
  string buffer;
  buffer = oss.str();
  DMESG("Shipping: " << buffer);    
  aSub.send(buffer);
  
  Statement st(conn);
  oss.str("");
  oss << "SELECT \"tr_log_command\""
      << " FROM " << transactionsTable << " WHERE \"tr_transaction_id\" = " 
      << log.xid << " ORDER BY \"tr_id\";";
  string query = oss.str();  
  DMESG("SQL: " << query << endl);
  if (st.exec(query)) {
    throw SQLException(st);
  }

  while (!st.fetch()){
    if (st.getData(1, buffer)) {
      throw SQLException("Failed to read command field from logs table", st);
    }
//      DMESG("Out> " << buffer << endl);
    buffer += "\n";
    aSub.send(buffer);
  }
  if (st.getErrorCode() != SQL_NO_DATA) {
    throw SQLException(st);
  }
}

/**
 * Checks if a directory is part of a set of subscriptions. This method treats
 * subscriptions recursive, so that subdirectories of subscribed directories 
 * are also considered subscribed.
 */
bool LogMonitorThread::isInSubscriptions(const string &directory, const vector<Subscription> &subs) {
//  DMESG("  isInSubscriptions? " << directory << endl);
  for (vector<Subscription>::const_iterator it = subs.begin(); it != subs.end(); ++it) {
    DMESG(" Testing subscription: " << it->directory << endl);
    if (directory == it->directory) {
      return true;
    }
    // So it is not for the same directory. Check if the log is for
    // a subdirectory. Must add the / at the end of the directory
    // so that /a does not match /abc, but only /a/...
    // If it is the root directory (/), then don't add a /            
    string aux = it->directory;
    if (!ends_with(aux, "/")) {
        aux += "/";
    }
//    cout << "starts_with(" << directory << ", " << aux << endl;
    if (starts_with(directory, aux)) {
      // This subscription does not need this log. The log is 
      // neither for the subscribed directory, nor for one of its
      // subdirectories.
      return true;
    }
  }
  return false;
}

// This method can be called either from the main monitoring thread or
// from the threads used to perform the initial synchronization. It should
// use different database connections depending on the caller. That's why
// it takes as an argument the database connection that will be used. If
// NULL, then the connection from the main thread is used.
void LogMonitorThread::updateSubscriberXid(
    ActiveSubscription &aSub, 
    DatabaseConnection &dbConn)
  throw (SocketException, runtime_error, SQLException) 
{
  // Poll the socket for acknowledgments from the slave. 
  // Non blocking, check first it the socket is ready for
  // being read. This must be done periodically to detect 
  // disconnections from a subscriber. The alternative is to
  // catch SIG_PIPE, but it requires signal handling.
  DMESG("[" << aSub.subscriber->subscriberID << "] Testing for acks\n");
  if (aSub.socket->isReadyToRead()) {
    string line = aSub.socket->readLine();
    string acks = line;
    while (aSub.socket->isReadyToRead()) {
      line = aSub.socket->readLine();
    }
    if (line != acks) {
      acks += "..";
      acks += line;
    }
    DMESG("[" << aSub.subscriber->subscriberID << "] Acks received: " << acks << endl);
    // We are just interested on the last ack. Assuming ordered delivery
    aSub.ackedXid = toUint64(line);
    // Update the xid.
    SubscriptionManager subMan(dbConn);
    subMan.updateXID(aSub.subscriber->subscriberID, aSub.ackedXid);
    DMESG(aSub.subscriber->subscriberID << ": " << aSub.ackedXid << endl);
  }
}

void LogMonitorThread::startStreamingLogs(ConnectionHandler &cHandler, 
  uint64_t xid, std::auto_ptr<Subscriber> subscriber)
        throw (SQLException, SocketException)
{
  // TODO: Assume that there is never more than one connection from
  // this subscriber. Otherwise, it would be necessary to lock
  // the database to prevent the deletion of logs needed by this 
  // subscriber
  vector<Subscription> subs = cHandler._subMan->getSubscriptions(subscriber->subscriberID);
  
  shared_ptr<ActiveSubscription> aSub(
    new ActiveSubscription(
              cHandler._socket, subscriber, subs));
  aSub->sentXid = xid;
  aSub->ackedXid = xid;
  {
    mutex::scoped_lock lk(_activeSubscribersMutex);
    activeSubscribers[cHandler._subscriberID] = aSub;
  }

  // There must be at least one subscription
  assert(!aSub->subs.empty() || aSub->subscriber->repUsers);

  // Prepare the statement. For efficiency we are going to 
  // use bind variables, both for input and output
  Statement statement(*cHandler._dbConn);  
  string query = "SELECT \"log_xid\", \"log_directory\", "\
        "\"log_user\", \"log_cwd\", \"log_permissions\", "\
        "\"log_group_rights\", \"log_command\" "\
        "FROM " + logsTable + " WHERE ( ? <\"log_xid\") AND (";
  string sep = "";
  if (!aSub->subs.empty()) {
  	for (vector<Subscription>::const_iterator it = aSub->subs.begin(); it != aSub->subs.end(); ++it) {
  	  query += sep + "\"log_directory\" LIKE '" + it->directory + "\%'";
  	  query += " OR \"log_directory\" = '" + it->directory + "'";
  	  sep = " OR ";
  	}
    query += " OR \"log_directory\" = '*'";
  }
  if (aSub->subscriber->repUsers) {
    query += sep;
    query += "\"log_directory\" = '";
    query += USER_CMD_META_DIR;
    query += "'";
  }
  query += ") ORDER BY \"log_xid\";";
  // TODO: we must select also logs for user groups: log_directory = usercmd. 
  // All directories must start with /, so any string that does not start like
  // that can be differentiated.  

  DMESG("SQL: Preparing: " << query << std::endl);  
  if(statement.prepare(query)) {
    throw SQLException("Failed to prepare statement", statement);
  }

  SQLINTEGER queryXid = 0;
  SQLLEN queryXidInd = 0;

  if (statement.bindParam(1, &queryXid, &queryXidInd)) {
    throw SQLException("Could not bind parameter for prepared statement", statement);
  }
  // Now the output variables
  Log log;
  statement.bind(2, log.directory, sizeof(log.directory));
  statement.bind(3, log.user, sizeof(log.user));
  statement.bind(4, log.cwd, sizeof(log.cwd));
  statement.bind(5, log.permissions, sizeof(log.permissions));
  statement.bind(6, log.groupRights, sizeof(log.groupRights));
	
  try {
    int sleepInterval = logCheckMinSleepInterval;
    auto_ptr<DatabaseConnection> conn(DatabaseConnection::openConnection());
    boost::xtime timestamp;
    while (true) {
      // Save the starting time of the query.
      boost::xtime_get(&timestamp, boost::TIME_UTC);
      
      // Update the bind variable
      queryXid = aSub->sentXid;
      DMESG("[" << aSub->subscriber->subscriberID << "] Checking for logs\n");
      // Use the database connection of the ConnectionHandler, since this
      // is still running on the same thread as the connection handler.
      if(statement.execPrepared()) {
        throw SQLException("Error executing statement", statement);
      }
      // Auxiliar connection for performing DB operations while the 
      // main connection is being used. Created lazily
      int logsRead = 0;
      while (!statement.fetch()) {
        // .fetch() reads the first 6 fields into the variables that were
        // bound to them. The last field contains the command, which is
        // of variable size so must be read manually.
        if (statement.getData(1, log.xid) || statement.getData(7, log.command)) {
          // There was an error.
          throw SQLException("Failed to read command field from logs table", statement);
        }
        
        if (boost::starts_with(log.command, "transaction")) {
          shipTransaction(log, *aSub, *conn);

        } else if (boost::starts_with(log.command, "upload")) {
          shipUploadLogs(log, *aSub, *conn);

        } else {
        	// It's a simple metadata command.
          if (!mustShip(log.command, log.directory, *aSub)) {
          	DMESG("Skipping log: " << log.command << endl);
          	continue;
          }
          ostringstream oss;
          oss << log.xid << LOG_SEP << log.directory << LOG_SEP << log.user 
              << LOG_SEP << log.cwd << LOG_SEP << log.permissions << LOG_SEP
              << log.groupRights << LOG_SEP << log.command << "\n";
          string cmd = oss.str();
//          DMESG("Shipping: " << cmd << "\n");
          aSub->send(cmd);
        }
        onLogSent(log, *aSub);

        // Check for acks only periodically. This is useful only to
        // allow the cleanup thread to delete the logs that were
        // already shipped
        logsRead++;
        if (logsRead % 128 == 0) {
          updateSubscriberXid(*aSub, *conn);
        }
      }
      if (statement.getErrorCode() != SQL_NO_DATA) {
        throw SQLException(statement);
      }
      DMESG("DEBUG: Fetching finished with: " << statement.getError() << endl);
      
      // Must close the cursor before the next query. But don't close
      // the statement or else we lose the bind parameters.
      statement.closeCursor();
      
      if (logsRead == 0) {
        // Exponential backoff, limited by LOG_CHECK_MAX_SLEEP_INTERVAL
        sleepInterval = min(sleepInterval*2, logCheckMaxSleepInterval);
        
        DMESG("[" << aSub->subscriber->subscriberID  << "] No logs found." << endl);
        
        //Jerome: Even if there is nothing new, we should test if the slave is already connected.
        // It could happen that the master doesn't detect the disconnection of the slave if this later shutdown
        // because of i.e. a power cut
        string log = "[" + aSub->subscriber->subscriberID + "] Testing if the client is already connected...";
    		try{
  			struct timeval timeout;
  			timeout.tv_sec = 5;
  		//	aSub->socket->setSNDTimeout(timeout);
  			// This should raise an exception if the client is no more connected..
  			aSub->socket->send(LOG_TEST_CONNECTION + "\n");
	    } catch (SocketException &ex) {
	    	DMESG(log << "[ERROR]\n");
	    	throw ex;
	    }
  		DMESG(log << "[OK]\n");
      } else {
        // Wait for all acks before sending more logs to the slave.
        // There is no need to send more logs to the slave if it 
        // hasn't finished processing the logs already sent.
        // This minimizes the number of queries performed on the local
        // database, and makes each query return more data at a time.
        DMESG("[" << aSub->subscriber->subscriberID << "] Done shipping. Waiting for acks: " << aSub->subscriber->lastAckedXid << "..." << endl);
        while (aSub->ackedXid != aSub->sentXid) {
          string line = aSub->socket->readLine();
          aSub->ackedXid = toUint64(line);
        }
        DMESG("Synchronized up to " << aSub->sentXid << endl);
        // Update the xid.
        SubscriptionManager subMan(*conn);
        subMan.updateXID(aSub->subscriber->subscriberID, aSub->ackedXid);
        // TODO: the slave is currently idle, waiting for new logs.
        // Should we pause anyway? It's a tradeoff between size of
        // the queries and the propagation delay. 
        // IDEA: enforce a minimum wait time between queries, already
        // including the time it takes to ship the logs.
        sleepInterval = logCheckMinSleepInterval;                
      }
      // Compute when the next cycle should start
      timestamp.sec += sleepInterval;
//      DMESG("[" << aSub->subscriberID << "] Sleeping for " << sleepInterval << " seconds." << endl);      

      boost::xtime now;
      boost::xtime_get(&now, boost::TIME_UTC);
      int sleepMs = (timestamp.nsec - now.nsec)/1000000;
      int sleepSec = timestamp.sec - now.sec;
      if (sleepMs < 0) {
        sleepMs += 1000;
        sleepSec--;
      }
      if (sleepSec < 0) {
        sleepSec = 0;
        sleepMs = 0;
      }

      DMESG("[" << aSub->subscriber->subscriberID << "] Sleeping for " << sleepSec << "." << sleepMs << " seconds " << endl);
      // Wait for the time remaining up to sleepInterval. Block on 
      // socket, to receive a notification in case the client breaks 
      // the connection
      if (aSub->socket->isReadyToRead(sleepSec, sleepMs)) {
        DMESG("Connection failed\n");
        // This should raise an exception.
        string line = aSub->socket->readLine();
        throw runtime_error("Unexpected data received from slave: " + line);
      }
    }
  } catch (std::exception &ex) {
    LOG(ex.what() << endl);
    mutex::scoped_lock lk(_activeSubscribersMutex);
    activeSubscribers.erase(cHandler._subscriberID);
    throw;
  } catch (...) {
    mutex::scoped_lock lk(_activeSubscribersMutex);
    activeSubscribers.erase(cHandler._subscriberID);
    throw;
  }
}

bool LogMonitorThread::isConnected(const string& subscriberID) {
  mutex::scoped_lock lk(_activeSubscribersMutex);
  return activeSubscribers.count(subscriberID) != 0;
}

void LogMonitorThread::shipTransaction(const Log &log, 
  ActiveSubscription &aSub, DatabaseConnection &conn) 
throw (SQLException, SocketException)
{
  DMESG("Shipping transaction: " << log.xid << endl);
  
  // Send the initial log to open a transaction. Being a special log, only the 
  // following fields are used: xid, user, permission and group masks.
  ostringstream oss;
  oss << log.xid << LOG_SEP 
      << "-" << LOG_SEP   // directory. dummy value. The slave will ignore it.
      << log.user << LOG_SEP   // user
      << "/" << LOG_SEP   // Also a dummy value. Transactions are independent of the cwd directory.
      << log.permissions << LOG_SEP   // permission mask
      << log.groupRights << LOG_SEP   // group rights mask
      << "transaction\n";   // command
  string reply = oss.str();
  DMESG("Shipping: " << reply << "\n");
  aSub.send(reply);

  Statement st(conn);
  // Filter query based on the directories subscribed by the slave
  // TODO: a prepared statement would help performance.
  oss.str("");
  oss << "SELECT \"tr_log_directory\", \"tr_log_cwd\", \"tr_log_command\""
    " FROM " << transactionsTable + " WHERE \"tr_transaction_id\" = " << log.xid << " AND (";

  // SIAHN : user replication bug fixing in a transaction
  oss << "\"tr_log_directory\" = '" << USER_CMD_META_DIR << "' OR " ;

  string sep("");
  for (vector<Subscription>::const_iterator it = aSub.subs.begin(); it != aSub.subs.end(); ++it) {
    oss << sep << "\"tr_log_directory\" like '" << it->directory << "\%'"
      << " OR \"tr_log_directory\" = '" << it->directory << "'";
    sep = " OR ";
  }
  oss << ") ORDER BY \"tr_id\";";
  string query(oss.str());
  DMESG("SQL: " << query << endl);
  if (st.exec(query)) {
    throw SQLException(st);
  }
  char directory[DIRNAME_MAX_SIZE];
  char cwd[DIRNAME_MAX_SIZE];
  string command;
  st.bind(1, directory, sizeof(directory));
  st.bind(2, cwd, sizeof(cwd));
  while (!st.fetch()) {
    // .fetch() reads the first 2 fields into the variables that were
    // bound to them. The last field contains the command, which is
    // of variable size so must be read manually.
    if (st.getData(3, command)) {
      throw SQLException("Failed to read command field from logs table", st);
    }

    if (!mustShip(command, directory, aSub)) {    	
	  DMESG("Skipping log: " << log.command << endl);
	  continue;
	}
    // Ship the log
    // The subscriber is interested on the log.
    oss.str("");
    // The logs for each individual command do not need to contain the context information
    // that was sent on the log that opened the transaction. They only need these fields:
    // <directory> <cwd> <command>
    oss << directory << LOG_SEP  // directory to which the log applies
          << cwd << LOG_SEP
          << command << "\n";
    reply = oss.str();
    DMESG("Shipping " << reply << ".\n");
    aSub.send(reply);
  }
  if (st.getErrorCode() != SQL_NO_DATA) {
    throw SQLException(st);
  }
  DMESG("Shipping commit.\n");
  aSub.send("commit\n");
}

bool LogMonitorThread::mustShip(const string &command, 
		const string &logDir, 
		ActiveSubscription &aSub) 
{
	DMESG("mustShip: " << command << ", " << logDir << endl);
	// Get the type of this log
	LogMonitorThread::CommandType logType = commandType(command);
	DMESG("Command type: " << logType << endl);
		
	if (logType == LogMonitorThread::cUsers) {
		return aSub.subscriber->repUsers;
	}	

	// It's not a basic metadata command. Search for the subscription 
	// corresponding to this directory to investigate 
	for (vector<Subscription>::const_iterator it = aSub.subs.begin(); it != aSub.subs.end(); ++it) {
		// The settings to replicate users and permissions are per subscription.
		// so we must find the right subscription.
		if (starts_with(logDir, it->directory)) {
			DMESG("Subscription: " << it->directory << endl);
			// It's this subscription
			switch (logType) {
				case LogMonitorThread::cPermissions:
					return it->repPermissions;
				case LogMonitorThread::cMetadata:
					return true;
				default:
					throw logic_error("Invalid logType: " + logType);
			}
		}
	}
	throw logic_error("Could not find subscription for log");
}

LogMonitorThread::CommandType LogMonitorThread::commandType(const string &command) {
	// Extract the command without the arguments.
	DMESG("commandType: " << command << endl);
	string cmd;
	size_t spaceIdx = command.find(" ");
	if (spaceIdx == string::npos) {
		cmd = command;
	} else {
		cmd = command.substr(0, spaceIdx);
	}
//	cout << "Searching: " << cmd << endl;
	const string *found = find(
		&LOG_PERMISSIONS[0], 
		&LOG_PERMISSIONS[LOG_PERMISSIONS_SIZE], 
		cmd);
	if (found != &LOG_PERMISSIONS[LOG_PERMISSIONS_SIZE]) {
		return LogMonitorThread::cPermissions;
	}
	found = find(
		&LOG_USERS[0],
		&LOG_USERS[LOG_USERS_SIZE],
		cmd);
	if (found != &LOG_USERS[LOG_USERS_SIZE]) {
		return LogMonitorThread::cUsers;
	}
//	cout << "Found: " << found << endl;
	// Everything else is a metadata command
	return LogMonitorThread::cMetadata;
}

void LogMonitorThread::onLogSent(const Log &log, ActiveSubscription &aSub) {
  try {
    StatsCollector::incCounter(COUNTER_LOGS_SENT);
    aSub.sentXid = log.xid;
    aSub.logsSent++;        
  } catch (runtime_error &ex) {
    LOG("WARNING: " << ex.what() << endl);
  }
}

int LogMonitorThread::getLogCount(DatabaseConnection &dbConn) 
throw (SQLException)
{
  string query = "SELECT \"log_table_tuples\" FROM " + amgaTable +";";
  Statement st(dbConn);
  DMESG("SQL: " << query << endl);
  if (st.exec(query)) {
    throw SQLException("Failed to get row count.", st);
  } else {
    int rowCount = -1;
    st.bind(1, &rowCount);
    if (st.fetch()) {
      throw SQLException("Failed to get row count.", st);
    }
    return rowCount;
  }
}

LogMonitorThread::ActiveSubscription::ActiveSubscription(
        boost::shared_ptr<TCPSocket> socketInit,
        std::auto_ptr<Subscriber> subscriberInit,
        std::vector<Subscription> &subsInit) 
: socket(socketInit), subscriber(subscriberInit), subs(subsInit), 
  logsSent(0), dataSent(0), sentXid(0), ackedXid(0), startupTime(time(NULL))
{
  StatsCollector::incCounter(COUNTER_SESSIONS_CURRENT);
  sessionID = StatsCollector::incCounter(COUNTER_SESSIONS_TOTAL);
}

void LogMonitorThread::ActiveSubscription::send(const string& data) {
  socket->send(data);
  dataSent += data.size();    
  StatsCollector::addToCounter(COUNTER_DATA_SENT, data.size());
}

LogMonitorThread::ActiveSubscription::~ActiveSubscription() {
  StatsCollector::decCounter(COUNTER_SESSIONS_CURRENT);
}

void LogMonitorThread::setConfig(const ConfigParser &configParser) {
	  
		try {
			logCheckMinSleepInterval = configParser.getInt("Replication::logCheckMinSleepInterval");
		} catch (std::string &err) {
			logCheckMinSleepInterval = LOG_CHECK_MIN_SLEEP_INTERVAL;
		}

		try {
			logCheckMaxSleepInterval = configParser.getInt("Replication::logCheckMaxSleepInterval");
		} catch (std::string &err) {
			logCheckMaxSleepInterval = LOG_CHECK_MAX_SLEEP_INTERVAL;
		}
}

Log::Log() : xid(0) {
  directory[0] = '\0';
  user[0] = '\0';
  cwd[0] = '\0';
  permissions[0] = '\0';
  groupRights[0] = '\0';
}

ostream& operator<<(ostream& out, const Log &log) {
    out << log.xid << ", " << log.directory << ", " 
        << log.user << ", " << log.cwd << ", " << log.permissions << ", " 
        << log.groupRights << ", " << log.command;
    return out;
}

//void Log::fillOut(const LogAux &logAux) {
//  directory = logAux.directory;
//  user = logAux.user;
//  cwd = logAux.cwd;
//  permissions = logAux.permissions;
//  groupRights = logAux.groupRights;
//}
