/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef LOGREADERTHREAD_H_
#define LOGREADERTHREAD_H_

#include "util/ThreadUtil.h"
#include "util/Socket.h"
#include "util/DatabaseConnection.h"
#include "util/Debuggable.h"
#include "util/StatsCollector.h"
#include "server/Constants.h"
#include "MountManager.h"
#include "SubscriptionManager.h"
#include "configparser/ConfigParser.h"

#include <vector>
#include <map>
#include <memory>
#include <boost/shared_ptr.hpp>

class ConnectionHandler;
class DatabaseConnection;

/* Used to read from the database. We need char arrays of fixed
 * size to bind with the DB variables, so we can't use strings.
 */
struct Log {
  uint64_t xid;  
  char directory[DIRNAME_MAX_SIZE];
  char user[USERNAME_MAX_SIZE];
  char cwd[DIRNAME_MAX_SIZE];
  char permissions[PERMISSION_SIZE];
  char groupRights[PERMISSION_SIZE];
  std::string command;
  
  Log();
  friend std::ostream& operator<<(std::ostream& out, const Log &log); 
};

///**
// * A replication log. 
// */
//struct Log {
//  uint64_t xid;
//  std::string directory;
//  std::string user;
//  std::string cwd;
//  std::string permissions;
//  std::string groupRights;
//  std::string command;
//
//  void fillOut(const LogAux &logAux);
//};

/**
 * Manages a thread that checks for new logs periocally and sends 
 * then to active subscribers.
 */
class LogMonitorThread : public Runnable, Debuggable, StatsSource
{
public:
	/**
   * 
   */
  LogMonitorThread(int sleepInterval, const ConfigParser &configParser);
  virtual ~LogMonitorThread();
  
  /**
   * Adds the given subscriber
   */
  void startStreamingLogs(
      ConnectionHandler &cHandler,
      uint64_t xid,
      std::auto_ptr<Subscriber> subscriber) throw (SQLException, SocketException);

  bool isConnected(const std::string& subscriberID);

  virtual void run();
  virtual void collectCounters(std::map<std::string, std::string> &counters);

private:
  /**
   * Holds all the relevant information about an active subscriber,
   * that is, a subscriber that is connected to the server.
   * Created and destroyed automatically when a subcription is created
   * and destroyed, respectively. Performs initialization and cleanup 
   * operatons
   */
  struct ActiveSubscription {
    ActiveSubscription(
        boost::shared_ptr<TCPSocket> socketInit,
        std::auto_ptr<Subscriber> subscriberInit,
        std::vector<Subscription> &subsInit);
    ~ActiveSubscription();

    boost::shared_ptr<TCPSocket> socket;
    // The subscriptions. They shouldn't change while 
    // the subscriber is active.
    std::auto_ptr<Subscriber> subscriber;
    std::vector<Subscription> subs;
    
    void send(const std::string& data);
    
    // The last xid sent to this subscriber.
    // The last xid acknowledged by this subscriber.
    // Counts the number of logs sent during this session
    int logsSent;
    int dataSent;
    uint64_t sentXid;
    uint64_t ackedXid;
    uint32_t sessionID;
    const time_t startupTime;
  };

  /** How long to sleep between checking for logs */
  unsigned int _sleepInterval;
  /** The database connection to the database with the logs */
  std::auto_ptr<DatabaseConnection> _dbConn;
    
  /** ?? */
  std::auto_ptr<SubscriptionManager> _subMan;

  /** Map between the subscriber id and the structure 
   * with its information */ 
  std::map<std::string, boost::shared_ptr<ActiveSubscription> > activeSubscribers;
  boost::mutex _activeSubscribersMutex;
  
  bool isInSubscriptions(const std::string &directory, const std::vector<Subscription> &subs);

  void shipTransaction(const Log &log, ActiveSubscription &aSub, DatabaseConnection &conn) 
    throw (SQLException, SocketException);
  void shipUploadLogs(const Log &log, ActiveSubscription &aSub, DatabaseConnection &conn) 
    throw (SQLException, SocketException);
  /** Cleanup old logs */
  void doLogCleanup() 
    throw (SQLException);
  void dropOldSubscribers(Statement &st, std::map<std::string, Subscriber> &subscribers) 
    throw (SQLException);
  /** Returns the number of logs stored on the log table */
  int getLogCount(DatabaseConnection &dbConn) 
    throw (SQLException);
  
  uint64_t getMinXidForDir(const std::map<std::string, uint64_t> &minXids, std::string dir) const;

  int execUpdate(Statement &st, const std::string &update) const
    throw (SQLException);

  //
  void updateSubscriberXid(ActiveSubscription &aSub, 
    DatabaseConnection &dbConn)
    throw (SocketException, std::runtime_error, SQLException);

  /** 
   * Called whenever a log is sent to a subscriber.
   * Used for statistics gathering
   */
  void onLogSent(const Log &log, ActiveSubscription &aSub);
  
  /**
   * Describes the type of log:
   * cMetadata - metadata command
   * cPermissions - entry/directories permissions or acls
   * cUsers - user/group management.
   */
  enum CommandType {
  	cMetadata, cPermissions, cUsers
  };    
  CommandType commandType(const std::string &command);
  bool mustShip(const std::string &command, const std::string &logDir, ActiveSubscription &aSub);
    
  /** Helper for collectCounters. Must be called while holding the lock for
   * the map */
  void addCounters(
    const std::map<std::string, boost::shared_ptr<ActiveSubscription> > &subscribers,
    std::map<std::string, std::string> &counters);

  void setConfig(const ConfigParser &configParser);
  int logCheckMinSleepInterval;
  int logCheckMaxSleepInterval;
  
  void setupTableNames();
  std::string sequencesTable;
  std::string transactionsTable;
  std::string logsTable;
  std::string amgaTable;
};

#endif /*LOGREADERTHREAD_H_*/
