/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "LogReceiver.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "util/Socket.h"
#include "util/DatabaseConnection.h"
#include "util/Statement.h"
#include "server/MDInterpreter.h"
#include "server/MDStandalone.h"
#include "server/MDBuffer.h"
#include "server/MDServerFactory.h"
#include "server/Constants.h"
#include "server/SiteManager.h"
#include "server/protocolHandling.h"
#include "configparser/ConfigParser.h"
#include "MountManager.h"
#include "ReplicationDaemonConnection.h"
#include "ReplicationManager.h"

#include "debug.h"

#include <fstream>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <boost/algorithm/string/predicate.hpp>

/* Constants */
static const char LOG_SEP = '\x1E';
static const std::string LOG_TEST_CONNECTION("@@ping@@");

static const int MQ_PERM = 0600;
// Separate the messages sent and received by the master process
static const int MSG_ID_QUERY = 1;
static const int MSG_ID_REPLY = 2;    
// How long should the master wait until checking if there 
// are dead slaves.
static const int MASTER_REAP_INTERVAL = 10;

/** Controls reconnections from master to slave */
static const int RECONNECT_DELAY_MIN = 2;
static const int RECONNECT_DELAY_MAX = 30;
static const int RECONNECT_MAX_ATTEMPTS = -1;
static const int RECEIVE_TIMEOUT= 30;

static const int MAX_ERROR_MESSAGE_SIZE = 128;

using namespace std;

/************************************************
 * STRUCTURES USED WITH MESSAGE QUEUES
 ***********************************************/ 
struct Request {
    Site site;
    bool isReconnecting;
    uint64_t curXid;
};

struct RequestMessage {
    RequestMessage() :msgType(MSG_ID_QUERY) {};
    long msgType;    
    Request req;
};

/**
 * Empty message, just to signal to the master process that the
 * mdserver process needs a new slave process for handling a new 
 * connections
 */
enum MasterOperation {
  noop, getSlave, die
};

struct MasterRequestMessage {    
    MasterRequestMessage() : msgType(MSG_ID_QUERY), op(noop) {}
    long msgType;
    MasterOperation op;
};

/**
 * Message to send as a reply through a message queue
 */
struct Reply {
    int returnValue;
    char errorMsg[MAX_ERROR_MESSAGE_SIZE];
};

/** 
 * Represents a slave process
 */
struct Process {
    // The message queue used to communicate with the process
    int mqID;
    pid_t pid;
    LogReceiverManager::ConnectionState state;
    // The master being handled by the slave
    char master[NODE_ID_MAX_SIZE];
};

struct ProcessMessage {
    ProcessMessage() :msgType(MSG_ID_REPLY) {};
    long msgType;
    int processSlot;
};


struct ReplyMessage {
    ReplyMessage() :msgType(MSG_ID_REPLY) {};

    ReplyMessage(int errorValue, const string &errorMsg) 
    : msgType(MSG_ID_REPLY) {
        reply.returnValue = errorValue;
        strncpy(reply.errorMsg, errorMsg.c_str(), MAX_ERROR_MESSAGE_SIZE);
        // strncpy does not put the null terminator on the string.
        reply.errorMsg[MAX_ERROR_MESSAGE_SIZE-1] = '\0';
    };
    
    long msgType;
    Reply reply;
};

/************************************************
 * ERROR HANDLING
 ***********************************************/
void ec(int retValue, bool throwException = false) throw (ReplicationException) {
    if (retValue == -1) {
        char logbuff[256]; 
        strcpy(logbuff, strerror_r(errno, logbuff, 255));
        LOG("Error: " << logbuff << endl);
        if (throwException) {
            throw ReplicationException(logbuff);  
        }
    }
}

void LogReceiverManager::displayStatus(pid_t pid, int status) {
    if (WIFEXITED(status)) {
        DMESG("Process: " << pid << " - Normal exit " << WEXITSTATUS(status) << endl);
        
    } else if (WIFSIGNALED(status)) {
        DMESG("Process: " << pid << " - Abnormal exit: " << WTERMSIG(status) << endl);
    } else {
        DMESG("Process: " << pid << " - Unknown status\n");
    }
}

// 
//boost::once_flag LogReceiverManager::once = BOOST_ONCE_INIT;
//LogReceiverManager *LogReceiverManager::instance = NULL;
//
//void LogReceiverManager::init() {
//    LogReceiverManager::instance = new LogReceiverManager();
//}
/************************************************
 * LOGRECEIVERMANAGER
 ***********************************************/
LogReceiverManager *LogReceiverManager::instance = NULL;

LogReceiverManager* LogReceiverManager::getInstance() {
//    if (instance == NULL) {
//        throw logic_error("LogReceiverManager not initialized");
//    }
    return LogReceiverManager::instance;
}

void LogReceiverManager::init(const ConfigParser& configParser) throw (ReplicationException) {
  if (instance != NULL) {
      throw ReplicationException("LogReceiverManager already initialized");
  }
  instance = new LogReceiverManager(configParser);
}

void LogReceiverManager::cleanup() {
    delete instance;
    instance = NULL;
}

LogReceiverManager::LogReceiverManager(const ConfigParser& configParser) throw (ReplicationException) {
  debug = ReplicationManager::getInstance()->getDebug();
  DMESG("Initializing\n");

  // Ignore the CTRL+C signal. That is handled by the main thread.
  // The signal mask is inherited by the slaves, so they will also
  // ignore SIGINT
  struct sigaction act;    
  memset(&act, 0, sizeof(act));
  act.sa_handler = SIG_IGN;
  sigaction(SIGINT, &act, NULL);

  // Create the message queues used to communicate between the main
  // server process and the master    
  if ((_masterMQ = msgget(IPC_PRIVATE, MQ_PERM)) == -1) {
    PERRLOG("Could not create message queue");
    throw ReplicationException("Could not create message queue");
  }
  
  // Create shared memory
  pid_t pid = getpid();
  ostringstream tmp;
  tmp << "LogReceiverShmLock" << pid;
  shmLockName = tmp.str();

  tmp.str("");
  tmp << "LogReceiverShm" << pid;
  shmName = tmp.str();

//    DMESG("Creating " << shmLockName << std::endl);
  if((shmLock = sem_open(shmLockName.c_str(), O_CREAT, S_IRWXU, 1)) == SEM_FAILED){
    PERRLOG("Could not create " << shmLockName);
    throw ReplicationException("Could not create shm lock");
  }

//    DMESG("Creating " << shmName << std::endl);
  if((shm = shm_open(shmName.c_str(), O_CREAT | O_RDWR, S_IRWXU))==-1){
    PERRLOG("Could not create "<< shmName);
    throw ReplicationException("Could not create shared memory");
  }

//    DMESG("Resizing shm area " << shmName << std::endl);
  int size = MAX_SLAVE_PROCESSES * (sizeof(Process));

  if (ftruncate(shm, size) == -1) {
    PERRLOG("Could not reserve space: " << size << std::endl);
    throw ReplicationException("Could not reserve space");
  }

  mem = (Process *)mmap(NULL, size, PROT_READ|PROT_WRITE,
                  MAP_SHARED, shm, 0);
  if (mem == MAP_FAILED) {
    PERRLOG("Could not map memory");
    throw ReplicationException("Could not reserve space");
  }

  setConfig(configParser);
  
  // Initialize process table
  memset(mem, 0, sizeof(Process)*MAX_SLAVE_PROCESSES);


  // We are done allocating the IPC structures. Now fork the master        
  pid = fork();
  switch (pid) {
    case -1: //Failed
      throw ReplicationException("Could not fork process");

    case 0: // Child
      master();
      exit(EXIT_SUCCESS);
      break;

    default: // parent
      _master = pid;
      break;
  }
}


LogReceiverManager::~LogReceiverManager() {
  DMESG("Closing resources\n");
  MasterRequestMessage reqMsg;
  reqMsg.op = die;
  if (msgsnd(_masterMQ, (void *)&reqMsg, sizeof(MasterOperation) , 0) == -1) {
    PERRLOG("Could not send message");
    kill(_master, SIGTERM);
  }
  
  // Wait for the process to die.
  int status;
  if (waitpid(_master, &status, 0) == -1) {
      PERRLOG("waitpid failed");
  } else {
      displayStatus(_master, status);    
      DMESG("Master died\n");
  }
  
  
  // Delete the message queue 
  if (msgctl(_masterMQ, IPC_RMID, NULL) == -1) {
     PERRLOG("Error removing msg queue");
  }

  // Delete the semaphores
  if (sem_close(shmLock) == -1) {
    PERRLOG("Error closing semaphores");   
  }
  if (sem_unlink(shmLockName.c_str()) == -1) {
    PERRLOG("Error unlink semaphore");
  }

  // Delete the shared memory
  if (munmap(mem, MAX_SLAVE_PROCESSES * (sizeof(Process))) == -1) {
    PERRLOG("Error unmapping shared memory");
  }
  if (close(shm) == -1) {
    PERRLOG("Error closing shared memory");
  }
  if (shm_unlink(shmName.c_str()) == -1) {
    PERRLOG("Error unlinking shared memory");
  }
}

void LogReceiverManager::releaseSlot(int index) {
  DMESG("releasing slot: " << index << endl);
  // Delete the message queue, if it wasn't already deleted 
  if (mem[index].mqID != 0) {
    if (msgctl(mem[index].mqID, IPC_RMID, NULL) == -1)  {
      PERRLOG("Error releasing message queue");
    }
  }
  // Empty the process structure
  memset(&mem[index], 0, sizeof(Process));    
}

/**
 *  Must be called holding the lock to the shared memory
 */
void LogReceiverManager::reapDeadSlaves() {
//    DMESG("Reaping dead slaves\n");
  // Search for slaves that died.
  for(int i=0; i< MAX_SLAVE_PROCESSES; i++) {
    if (mem[i].pid != 0) {
      // Found a child
      int status;
      int ret = waitpid(mem[i].pid, &status, WNOHANG);
      if (ret > 0) {
        DMESG("Found dead child: " << i << endl);
        // This slave died. Cleanup its structure.
        displayStatus(mem[i].pid, status);
        // Release the message queue connecting to this slave
        // and clean up the memory of the slot
        releaseSlot(i);
          
      } else if (ret == -1) {
        ec(ret);

      } else {
        // ret should be 0. Process is still running
      }
    }
  }
}
/**
 * Empty stubs, necessary only to intercept the signals 
 * and force the master process to be interrupted.
 */
void alarmHandler(int sig) {}
void slaveDeathHandler(int sig) {}

void LogReceiverManager::master(){
  DMESG("Master process created. Waiting for requests\n");
  
  // Signal handlers to force the master to be interrupted when
  // receiving a signal. Otherwie, SIGCHLD is ignored and SIGALRM 
  // kills the master process
  signal(SIGCHLD, slaveDeathHandler);
  signal(SIGALRM, alarmHandler);

  while (true) {
    // Wait for requests sent by the method startReceiveLogs.
    MasterRequestMessage req;
    // Set an alarm so that msgrcv is interrupted up periodically 
    // to purge dead children.
    alarm(MASTER_REAP_INTERVAL);
    // This method can exit by: receiving a message, interrupted when 
    // a child dies or by the alarm
    int ret = msgrcv(_masterMQ, (void *)&req, sizeof(MasterOperation), MSG_ID_QUERY, 0);
    // Disable the alarm so that we are not interrupted while 
    // handling the request.
    int err = errno;
    alarm(0);
    if (ret == -1) {
      // Failed. Was it interrupted? Perhaps because a child died.
      if (err == EINTR) {
        SharedMemoryLock lock(this);
        reapDeadSlaves();
      } else {
        PERRLOG("LogReceiver master - error in msgrcv:");
      }
      continue;
    }
    LOG("[Rep Slave] Msg received from slave: " << req.op << endl);
    // msgrcv returned successfully. We have a new message
//    DMESG("Received operation: " << req.op << endl);
    if (req.op == die) {
      DMESG("[Rep Slave] I was ordered to terminate" << endl);
      {
        // Kill all active slaves and unlink the message queues.
        SharedMemoryLock lock(this);
        for (int i = 0; i<MAX_SLAVE_PROCESSES; i++) {
          if (mem[i].pid != 0) {
            // kills the child and release its resources
            pid_t pid = mem[i].pid;
            DMESG("[Rep Slave] Killing slave: " << pid << endl);
            if (kill(pid, SIGTERM) == -1) {
              PERRLOG("Error killing slave");
            } else {
              int status;
              if (waitpid(pid, &status, 0) == -1) {
                PERRLOG("waitpid failed");
              } else {
                displayStatus(pid, status);
              }
            }
            releaseSlot(i);
          }
        }
      }
      exit(EXIT_SUCCESS);
    }
    
    assert(req.op == getSlave);
    DMESG("[Rep Slave] Received request for new slave" << endl);
    // Prepare the answer message 
    ProcessMessage pMsg;    
    
    // Find an empty slot for the new process
    // First locks access to the shared memory
    SharedMemoryLock lock(this);
    int slot = -1;
    for (int i=0; i<MAX_SLAVE_PROCESSES; i++) {
      if (mem[i].pid == 0) {
          slot = i;
          break;
      }
    }
    if (slot == -1) {
      LOG("Maximum number of slaves reached");
      pMsg.processSlot = -1;
      if (msgsnd(_masterMQ, (void *)&pMsg, sizeof(int), 0) == -1) {
        PERRLOG("Error sending message");
      }

    } else {
      DMESG("Found slot: " << slot << endl);
      pMsg.processSlot = slot;
      // Create a new msg queue for communicating with the new slave        
      mem[slot].mqID = msgget(IPC_PRIVATE, MQ_PERM);
      if (mem[slot].mqID == -1) {
          // Could not allocate the message queue
          ec(mem[slot].mqID);
          pMsg.processSlot = -1;
          if (msgsnd(_masterMQ, (void *)&pMsg, sizeof(int), 0) == -1) {
            PERRLOG("Error sending message");
          }
      } else {    
        // Process the request by forking a new slave process that
        // will establish the connection.
        pid_t pid = fork();
        switch (pid) {
          case -1: //Failed
            ec(pid);
            pMsg.processSlot = -1;
            if (msgsnd(_masterMQ, (void *)&pMsg, sizeof(int), 0) == -1) {
              PERRLOG("Error sending message");
            }
            break;

          case 0: // Child process
            slave(slot);
            exit(EXIT_SUCCESS);
            break;

          default: // parent
            // Save the pid and the message queue
            DMESG("Sending process info to MD process" << endl);
            mem[slot].pid = pid;
            if (msgsnd(_masterMQ, (void *)&pMsg, sizeof(int) , 0) == -1) {
              PERRLOG("Error sending message");          
            }
            break;
        }
      }
    }
  }
}

/**
 * Finds the slot of the process receiving logs for the given mount point.
 * Returns -1 if the connection is not active.
 */
int LogReceiverManager::findConnection(const string &master) {
  DMESG("Searching for master: " << master << endl);
  for(int i=0; i<MAX_SLAVE_PROCESSES; i++) {
    if (mem[i].pid != 0 && master == mem[i].master) {
      // Found it
      return i;
    }
  }
  // Didn't find it
  return -1;
}

void LogReceiverManager::startReceiveLogs(const Master& master, DatabaseConnection &dbConn)
    throw (ReplicationException)
{
  DMESG("Start receive logs from: " << master.id << "\n");
  {
    // This block is necessary to force the destructor of 
    // SharedMemoryLock to be executed.
    SharedMemoryLock lock(this);
    // First check if there is already a connection to this server
    if (findConnection(master.id) != -1) {
        throw ReplicationException(string("Connection to ") + master.id + " already active");
    }
  }

	
  // Request a new slave
  DMESG("Requesting a new slave\n");
  MasterRequestMessage msg;
  msg.op = getSlave;
  ec(msgsnd(_masterMQ, (void *)&msg, sizeof(MasterOperation), 0), true);
  
  DMESG("Waiting for reply\n");
  // Now wait for the info about the new process
  ProcessMessage procMsg;
  ec(msgrcv(_masterMQ, (void *)&procMsg, sizeof(int), MSG_ID_REPLY, 0), true);
  
  if (procMsg.processSlot == -1) {
    // Failed to create a new slave.
    throw ReplicationException("Could not fork slave process");
  }
  int mqID = mem[procMsg.processSlot].mqID;
  DMESG("Got a slave. Sending it a request: " << master.id << endl);
  
  // Prepare the request to the new process that will open the connection
  RequestMessage reqMsg;
  // Read the configuration of the master from the database 
  SiteManager sMan(dbConn);
  try {
    sMan.getSite(master.id, reqMsg.req.site);
  } catch (SQLException &ex) {
    throw ReplicationException(ex.what());
  }
//  strcpy(reqMsg.req.host, n->host.c_str());
//  strcpy(reqMsg.req.nodeId, n->name.c_str());
//  reqMsg.req.port = n->port;
  reqMsg.req.curXid = master.masterXid;
  reqMsg.req.isReconnecting = master.active;

  // Lock shared memory to prevent the master process from reaping
  // the slave and the message queue before we have time to 
  // extract the answer from the queue. This can happen if the slave
  // dies after answering, thereby raising a signal with the master
  // who will clean its data structures.
  SharedMemoryLock lock(this);
  // Send the request to the slave
  ec(msgsnd(mqID, (void *)&reqMsg, sizeof(Request), 0), true);

  DMESG("Request sent. Waiting for reply.\n");
  // Wait for the answer
  ReplyMessage repMsg;
  int err;
  if ((err=msgrcv(mqID, (void *)&repMsg, sizeof(Reply), MSG_ID_REPLY, 0)) == -1) {
  	PERRLOG("Error receiving message");
  	ec(err);
  }

  // We don't need the message queue anymore
  if (msgctl(mem[procMsg.processSlot].mqID, IPC_RMID, NULL) == -1) {
    PERRLOG("Error deleting message queue");
  }
  mem[procMsg.processSlot].mqID = 0;

  // Check if the connection was established successfully.
  if (repMsg.reply.returnValue == -1) {
    DMESG("Got reply, slave failed.\n");
    // Slave structures will be released by the master process
    // when we release the lock on shared memory.
    throw ReplicationException(repMsg.reply.errorMsg);
  } else {
    DMESG("Got reply, slave running sucessfully.\n");
    // Succeeded. Just return silently. 
    strncpy(mem[procMsg.processSlot].master, master.id.c_str(), 
      sizeof(mem[procMsg.processSlot].master));
  }
}

void LogReceiverManager::closeConnection(const std::string &master) {
  SharedMemoryLock lock(this);
  int slot = findConnection(master);
  if (slot == -1) {
    throw ReplicationException("No connection open for master: " + master);
  }
  pid_t pid = mem[slot].pid;
  LOG("[Rep Client] Disconnecting from " << master << endl);
  DMESG("Killing slave: " << pid << endl);
  ec(kill(pid, SIGTERM));
  // Since this method is called by md server proccess, which is not
  // the parent of the slave, we cannot release the process structures 
  // here. The master will take care of that.    
}

bool LogReceiverManager::isConnectionOpen(const std::string &master) {
  SharedMemoryLock lock(this);
  return findConnection(master) != -1;
}

LogReceiverManager::ConnectionState LogReceiverManager::getConnectionState(const std::string &master) {
  SharedMemoryLock lock(this);
  int slot = findConnection(master);
  return (slot == -1) ? 
    LogReceiverManager::Inactive : 
    mem[slot].state;
}

std::vector<std::string> LogReceiverManager::getOpenConnections() {
    throw logic_error("Not yet implemented");
}

// Send the reply message to the master process if one wasn't already sent
void sendReply(int mqID, const char *msg, bool &replySent) throw (ReplicationException) {
  if (!replySent) {
    ReplyMessage rMsg;
    if (msg == NULL) {
      rMsg.reply.returnValue = 0;      
      rMsg.reply.errorMsg[0] = '\0';
    } else {
      rMsg.reply.returnValue = -1;      
      snprintf(rMsg.reply.errorMsg, sizeof(rMsg.reply.errorMsg), 
      	"Error on LogReceiver slave: %s", msg);
    }
    ec(msgsnd(mqID, (void *)&rMsg, sizeof(Reply), 0), true);
    replySent = true;
  }
}

// For automatic cleanup of the SSL_CTX object
struct SSLContextWrapper {
  SSL_CTX *ctx;
  SSLContextWrapper(SSL_CTX *ctxInit = NULL) : ctx(ctxInit) {
  }
  ~SSLContextWrapper() {
    if (ctx != NULL) {
      SSL_CTX_free(ctx);
    }
  }
};

void LogReceiverManager::parseLog(const string& line, int nTokens, vector<string> &tokens) {
  unsigned int oldLoc = 0;
  unsigned int loc;
  tokens.clear();
  for (int i=0; i<nTokens; i++) {
    loc = line.find(LOG_SEP, oldLoc);
    tokens.push_back(line.substr(oldLoc, loc-oldLoc));
    oldLoc = loc+1;
  }
}

void executeLog(MDStandalone &server, const string &command) 
throw (runtime_error) 
{
  static MDInterpreter interpreter;
  interpreter.parseCommand(command);
  if (interpreter.execLast(server)) {
    // Cannot continue, the command failed to execute.
    throw runtime_error("Invalid command received: " + command);
  }
  // Check the data on the buffer to see if the command failed.
  string result;
  server.getBuffer().read(result);
//  cout << "Result: " << result << endl;
  // Check if the command executed correctly and interrupt the transfer
  // if it didn't. During an upload command no error code is returned
  // until the final commit, so "" means no error. Otherwise, a return
  // of 0 represents a single shot command that executed correctly.
  if (result != "" && result[0] != '0') {
    throw runtime_error("Command execution failed: " + result);
  }
}

void checkCommandResult(MDStandalone &server) throw (runtime_error) {
  // Check the data on the buffer to see if the command failed.
  string result;
  server.getBuffer().read(result);
//  cout << "Result: " << result << endl;
  // Check if the command executed correctly and interrupt the transfer
  // if it didn't. During an upload command no error code is returned
  // until the final commit, so "" means no error. Otherwise, a return
  // of 0 represents a single shot command that executed correctly.
  if (result != "" && result[0] != '0') {
    throw runtime_error("Command execution failed: " + result);
  }
}

/* Automatic resetting of the state of the Master on exit from slave() method */ 
struct SlaveCleanup {
  MountManager &mManager;
  string siteName;
  
  SlaveCleanup(MountManager &mManagerInit, const string& siteNameInit) 
  : mManager(mManagerInit), siteName(siteNameInit) 
  {}  
  ~SlaveCleanup() {
    try {
      mManager.updateMasterState(siteName, false);
    } catch (exception &ex) {
      LOG("Error resetting master state: " << ex.what() << endl);
    }
  }
};

void LogReceiverManager::slave(int slot) {
  // Open the message queue that will be used to communicate with 
  // the master or with other processes
  signal(SIGPIPE, SIG_IGN);
  signal(SIGTERM, SIG_DFL);
  
//  debug = false;
  
  DMESG("Slave slot " << slot << endl);
  int mqID = mem[slot].mqID;
//    {
//        SharedMemoryLock lock(this);
//        mqID = mem[slot].mqID;
//        mem[slot].state = Inactive;
//    }
  DMESG("Slave message queue: " << mqID << ", slot: " << slot << endl);
  
  // Wait for the request with the info about the PC to where it
  // should connect
  RequestMessage reqMsg;

  if (msgrcv(mqID, (void *)&reqMsg, sizeof(Request), MSG_ID_QUERY, 0) == -1) {        
    PERRLOG("slave msgrcv failed");
    exit(EXIT_SUCCESS);
  }
      
  DMESG("Received connection request. Node: " << reqMsg.req.site.hostname << ":" << reqMsg.req.site.port << endl);
  

  // Whether if we have already answered the master process
  bool replySent = false;
  try {
    // This try is for all the initializations that don't require contacting
    // the server. If they fail, we abort.
    MDBuffer buffer;
    auto_ptr<MDStandalone> server(dynamic_cast<MDStandalone*>(MDServerFactory::mdServer(buffer)));
    // The Mount manager should use the same DB conn as the server object,
    // so we can have transactions around all the operations performed while
    // executing a log.
    DatabaseConnection *dbConn = server->getDBConnection();
    MountManager mManager(*dbConn);
    // Allocate on stack, the dctor will reset the state 
    // of the master to inactive.
    SlaveCleanup sCleanup(mManager, reqMsg.req.site.name);

    server->setReplicating(true);
    server->setMaster(reqMsg.req.site.name);
    server->setDebug(debug);   
    
    std::list<MDUserManager *> userManagers;
    createUserManagers(*ConfigParser::configParser, userManagers);
    server->initializeUserManager(userManagers);     

    // Last log received.
    uint64_t lastLogCommitted = reqMsg.req.curXid;

    if (reqMsg.req.isReconnecting) {
      // It is a reconnection attempt, so even if the initial connection fails
      // we will keep trying on the background. So send a success message.
      sendReply(mqID, NULL, replySent);
    }

    // KIM090202
    int reconnectDelay = reconnectDelayMin;
    // Now wait for logs. One per line
    // This object will hold the SSL_CTX and release it automatically when destroyed.
    SSLContextWrapper ctxWrapper; 
    while (true) {
      auto_ptr<TCPSocket> socket;
      try {
        // Connect to the server and start receiving the logs.
        ReplicationDaemonConnection conn(reqMsg.req.site, ctxWrapper.ctx);
        ctxWrapper.ctx = conn.getSSLCtx();
        try {
          // Send the number of the last xid we have received
          socket = conn.receive(lastLogCommitted);

          struct timeval tv; 
          tv.tv_sec = receiveTimeout;
          tv.tv_usec = 0;

          socket->setRcvTimeOut(tv);
		
        } catch (exception &ex) {
          // If it fails on connection receive, the problem is not transient
          LOG("[Rep Client] Fatal error connecting to master: " << ex.what() << ", aborting\n");
          sendReply(mqID, ex.what(), replySent);
          exit(EXIT_SUCCESS);
        }

        LOG("[Rep Client] Receiving logs from " << socket->getForeignAddress() 
        	<< ":" << socket->getForeignPort() << ". XID: " 
        	<< lastLogCommitted << endl);
        reconnectDelay = reconnectDelayMin;
        
        // Connection successful
        // It was the initial attempt and we are not trying to reconnect. Either way, return 
        // success to the master process and start listening for logs.
        sendReply(mqID, NULL, replySent);
        {
          SharedMemoryLock lock(this);
          mem[slot].state = Receiving;
        }
        socket->setTCPReceiveWindow(128*1024);
//        ifstream bufferSizeFile("buffer_size");
//        if (!bufferSizeFile) {
//            cout << "[PERF_DEBUG] Could not find \"buffer_size\" file. Using default value\n"; 
//        } else {
//            int bufSize;
//            bufferSizeFile >> bufSize;        
//            cout << "[PERF_DEBUG] Setting receive buffer size to " << bufSize << endl;
//            socket->setTCPReceiveWindow(bufSize);
//        }
//        cout << "[PERF_DEBUG] Receive buffer size: " << socket->getTCPReceiveWindow() << endl;

        // Connected. Start receiving logs
        while (true) {
//        DMESG("Waiting for input\n");
          string line = socket->readLine();
          DMESG("In> " << line << "\n");
          
          //Jerome: Don't take into account the ping message of the master.
          if(line == LOG_TEST_CONNECTION){
        	  DMESG("Master " << socket->getForeignAddress() << " ping\n");
          }else{
        	  DMESG("This is not a ping from Master " << socket->getForeignAddress() << "\n");
        	  
	          vector<string> tokens;
	          parseLog(line, 7, tokens);
	
	          // Open a transaction
	          Statement st(*dbConn);
	          if (st.beginTransaction()) {
	            throw SQLException(st);
	          }
	
	          // TODO: Currently we are not checking if the directory is
	          // replicated. What should the slave do if it receives a 
	          // log for a directory that is not replicated? It is a bug
	          // in the master, it should raise an error and close the 
	          // connection       
	          // Check if we have already applied this log         
	          uint64_t logXid = toUint64(tokens[0]);
	          uint64_t mountXid;
	          string logDirectory;
	          string command = tokens[6];
	          
	          // Save the next command to be executed. for replication.
	          server->setNextCommand(command);
	          
	          bool isTransaction = boost::starts_with(command, "transaction");
	          bool isUpload = boost::starts_with(command, "upload");
	          if (isTransaction) {
	          	DMESG("Transaction xid: " << logXid << endl);
	          } else {
	            // non-transaction commands have information about the directory
	            // they update
	            // A transaction can affect several directories, so we
	            // have to analyse log by log.
	            logDirectory = tokens[1];
	            mountXid = mManager.getMountXid(logDirectory);
	            DMESG("Received XID: " << logXid << ", mount xid " << mountXid << endl);
	          }
	          if (!isTransaction && mountXid >= logXid) {
	            DMESG("Discarding old log. Log xid: " << logXid
	                << ", mount xid: " << mountXid << endl);
	          } else {
	            DMESG("Applying log\n");
	            // We don't have this log. Apply it.
	            // Use the remote security context
	            server->setDir(tokens[3]);
	            
	            // TODO: check if we are replicating permissions and if not,
	            // don't set these fields.
	            server->setUser(tokens[2]);
	            server->setPermissionMask(tokens[4]);
	            server->setGroupRightsMask(tokens[5]);
	            
	            if (isTransaction) {
	              DMESG("Handling transaction\n");
	              set<string> updatedDirs;
	              // Inform the server that a transaction is starting. No need
	              // to go through the MDInterpreter
	              server->transactionStart();
	              checkCommandResult(*server);
	              
	              // Read the commands to be executed inside the transaction.
	              while (true) {
	                line = socket->readLine();
	                // <directory> <cwd> <command>
	                DMESG("Received: " << line << endl);
	                if (line == "commit") {
	                  // Save the transaction command
	                  server->setNextCommand(line);
	                  server->transactionCommit();
	                  DMESG("Transaction commited\n");
	                  checkCommandResult(*server);
	                  break;
	                }
	                parseLog(line, 3, tokens);
	                logDirectory = tokens[0];
	                command = tokens[2];
	
	                // Save the transaction command
	                server->setNextCommand(command);                
	                // Does this directory already has this log?
	             	  mountXid = mManager.getMountXid(logDirectory);
	                if (mountXid >= logXid) {
	                  DMESG("Discarding old log. Log xid: " << logXid
	                    << ", directory xid: " << mountXid << endl);
	                } else {
	                  server->setDir(tokens[1]); // Cwd of the command
	                  executeLog(*server, command);
	                  // If we update the directory xid before appying all logs,
	                  // the subsequent logs of this transaction will fail since 
	                  // they will read the updated value and think they should
	                  // be discarded.
	                  updatedDirs.insert(logDirectory);
	                }
	              }
	              DMESG("Updating directories\n");
	              updatedDirs.insert(logDirectory);
	              
	              for (set<string>::const_iterator it = updatedDirs.begin();
	                it != updatedDirs.end(); ++it)
	              {
	                DMESG(*it << endl);
	           	    mManager.updateMountXid(*it, logXid, true);
	              }
	              
	            } else if (isUpload) {
	              DMESG("Upload: " << line << endl);
	              // Handle an upload. The first log is already part of the 
	              // sequence of upload commands so it must be executed. It
	              // will start a transaction and initiate the upload
	              executeLog(*server, command);
	              while (true) {
	                line = socket->readLine();
	                // <directory> <cwd> <command>
	                DMESG("Received: " << line << endl);
	                server->setNextCommand(line);
	                executeLog(*server, line);
	                if (line == "commit") {
	                  break;
	                } else if (line == "abort") {
	                  LOG("WARNING: abort received as a log. Should not happen\n");
	                  break;
	                }
	              }
	              DMESG("Updating directory\n");
	              mManager.updateMountXid(logDirectory, logXid, true);
	              
	            } else { // no transaction or upload
	              executeLog(*server, command);
	              // The command executed correctly. Update the local XID.
	              // This should be done in the same transaction as the table was
	              // updated.
	              mManager.updateMountXid(logDirectory, logXid, true);
	            }
	          }
	          mManager.updateMasterXid(reqMsg.req.site.name, logXid);
	          DMESG("Done applying log: " << logXid << ". Committing..." << endl);
	          // The log is correctly updated.
	          if (st.commitTransaction()) {
	            throw runtime_error("Error committing transaction: " + st.getError());
	          }
	          lastLogCommitted = logXid;
	          DMESG("Acknowledging: " << logXid << endl);
	          // Acknowledge this log.
	          ostringstream oss; oss << logXid << "\n";
	          socket->send(oss.str());
	        }
        }
      } catch (SocketException &ex) {
        LOG("Connection to master failed: " << ex.what() << endl);
        // Release the old socket        
        socket.reset();
        if (!replySent) {
          // If no reply was sent, then this is not a background reconnection
          // attempt. We must give up and inform the master
          throw ex;
//          mManager->updateMasterState(reqMsg.req.site.name, false);
//          sendReply(mqID, ex.what(), replySent);
//          exit(EXIT_SUCCESS);
        }
        LOG("Reconnecting in " << reconnectDelay << " seconds\n");
        {
            SharedMemoryLock lock(this);
            mem[slot].state = Reconnecting;
        }
        sleep(reconnectDelay);
        reconnectDelay *= 2;
        if (reconnectDelay > reconnectDelayMax) {
            reconnectDelay = reconnectDelayMax;
        }
      }
    }
  } catch (exception &ex) {
    // Fatal error (not a socket exception)
    LOG("[Rep Slave] Aborting: " << ex.what() << "\n");
    sendReply(mqID, ex.what(), replySent);

  } catch (...) {
    // Fatal error (not a socket exception)
    LOG("[Rep Slave] Aborting due to unknown exception\n");
    sendReply(mqID, "Unknown exception", replySent);
    throw;
  }
  exit(EXIT_SUCCESS);
}

void LogReceiverManager::setConfig(const ConfigParser &configParser) {
	    
		try {
			reconnectDelayMin = configParser.getInt("Replication::reconnectDelayMin");
		} catch (std::string &err) {
			reconnectDelayMin = RECONNECT_DELAY_MIN;
		}

		try {
			reconnectDelayMax = configParser.getInt("Replication::reconnectDelayMax");
		} catch (std::string &err) {
			reconnectDelayMax = RECONNECT_DELAY_MAX;
		}

		try {
			reconnectMaxAttempts = configParser.getInt("Replication::reconnectMaxAttempts");
		} catch (std::string &err) {
			reconnectMaxAttempts = RECONNECT_MAX_ATTEMPTS;
		}

		try {
			receiveTimeout = configParser.getInt("Replication::receiveTimeout");
		} catch (std::string &err) {
			receiveTimeout = RECEIVE_TIMEOUT;
		}
		
}

LogReceiverManager::SharedMemoryLock::SharedMemoryLock(LogReceiverManager *manager) 
: _manager(manager) {
//  DMESG("TRY LOCK\n");
  while (sem_wait(_manager->shmLock)) {
    if(errno == EINTR)
      continue;
    PERRLOG("Error waiting for semaphore " << _manager->shmLockName << std::endl);
  }
//  DMESG("GOT LOCK\n");
}


LogReceiverManager::SharedMemoryLock::~SharedMemoryLock() {
//  DMESG("UNLOCK\n");
  if(sem_post(_manager->shmLock)) {
    LOG("Error posting semaphore " << _manager->shmLockName << std::endl);
  }
}
