/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef LOGRECEIVER_H_
#define LOGRECEIVER_H_

#include <vector>
#include <semaphore.h>
#include <signal.h> 
#include <memory>

#include "Exceptions.h"
#include "util/Debuggable.h"
#include "configparser/ConfigParser.h"

class TCPSocket;
class MDStandalone;
class MDBuffer;
class DatabaseConnection;
struct Mount;
struct Master;
struct Process;


/** 
 * Manages the connections to the masters.
 */
class LogReceiverManager : public Debuggable
{
public:
    // Constants
    static const int MAX_SLAVE_PROCESSES = 5;
    
public:
    enum ConnectionState {
        Inactive=0, Connecting, Receiving, Reconnecting
    };

    static void addConnection(const std::string &master, 
        std::auto_ptr<TCPSocket> socket, std::auto_ptr<MDStandalone> server, 
        std::auto_ptr<MDBuffer> buffer)
    throw (ReplicationException);

    void closeConnection(const std::string &master);

    bool isConnectionOpen(const std::string &master);
    ConnectionState getConnectionState(const std::string &master);
    
    std::vector<std::string> getOpenConnections();

    void startReceiveLogs(const Master& master, DatabaseConnection &dbConn) 
        throw (ReplicationException);
        
    static void init(const ConfigParser &configParser) throw (ReplicationException);
    static void cleanup();
   
    static LogReceiverManager* getInstance();

private:
    // Singleton pattern
    static LogReceiverManager *instance;
    LogReceiverManager()
        throw (ReplicationException);
    LogReceiverManager(const ConfigParser& configParser)
        throw (ReplicationException);    
    ~LogReceiverManager();

    /**
     * Data fields
     */
    int _masterMQ;
    pid_t _master;
    
    // Shared memory
    std::string shmName;
    int shm; // File descriptor of the shared memory file
    // Pointer to the shared memory
    Process *mem;

    /**
     * The following class acquires the lock on shared memory
     * in the constructor and releases it in the destructor
     */
    class SharedMemoryLock {
    public:
        SharedMemoryLock(LogReceiverManager *manager);
        ~SharedMemoryLock();
    private:
        LogReceiverManager *_manager;
    };
    
    // Semaphores to control shared memory
    std::string shmLockName;
    sem_t *shmLock;
    
    int reconnectDelayMin;
    int reconnectDelayMax;
    int reconnectMaxAttempts;
    int receiveTimeout;
    
    // These access shared memory without acquiring the lock.
    // The methods calling them must acquire the lock themselves.
    int findConnection(const std::string &master);
    void reapDeadSlaves();
    void releaseSlot(int index);
    
    // Implements the master process
    void master();
    // Slave process. Receives logs
    void slave(int mqID);
    
    void parseLog(const std::string& line, int nTokens, std::vector<std::string> &tokens);
    void displayStatus(pid_t pid, int status); 
    
    void setConfig(const ConfigParser &configParser);
};

#endif /*LOGRECEIVER_H_*/
