/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "MDLogWriter.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "debug.h"
#include "MountManager.h"
#include "server/sqlHelpers.h"
#include "util/Statement.h"
#include "util/StatsCollector.h"
#include "util/DatabaseConnection.h"

#include "replicationdaemon/SubscriptionManager.h"

#include <ostream>
#include <list>
#include <boost/algorithm/string.hpp>

using namespace std;

bool MDLogWriter::_saveLogs = false;
bool MDLogWriter::debug = true;


// Names of the counters exposed by this class. 
// Must find a way of transmitting this information across processes.
//static const string COUNTER_LOGS_WRITTEN("repmaster.LogsWritten");
//static const string COUNTER_UPDATES("mdserver.Updates");
//static const string COUNTER_COMMANDS("mdserver.Commands");

void MDLogWriter::setSaveLogs(bool saveLogs) {
	DMESG("Save logs: " << saveLogs << endl);
	MDLogWriter::_saveLogs = saveLogs;
}

bool MDLogWriter::getSaveLogs() {
	return MDLogWriter::_saveLogs;
}	

void MDLogWriter::setDebug(bool debug) {
  MDLogWriter::debug = debug;  
}

bool MDLogWriter::getDebug() {
  return MDLogWriter::debug;
}
  
  
MDLogWriter::MDLogWriter() : _xid(0), _inTransaction(false), lastXid(0)
{
  setupTableNames();
//  if (!StatsCollector::hasCounter(COUNTER_LOGS_WRITTEN)) {
//    StatsCollector::setCounter(COUNTER_LOGS_WRITTEN, 0);
//    StatsCollector::setCounter(COUNTER_UPDATES, 0);    
//    StatsCollector::setCounter(COUNTER_COMMANDS, 0);
//  }
}

MDLogWriter::~MDLogWriter()
{
//  StatsCollector::dropCounter(COUNTER_LOGS_WRITTEN);
//  StatsCollector::dropCounter(COUNTER_UPDATES);
//  StatsCollector::dropCounter(COUNTER_COMMANDS);
}


void MDLogWriter::setupTableNames()
{
  std::string prefix;
  if(DatabaseConnection::dbSchema.size())
    prefix = DatabaseConnection::dbSchema + ".";
  
  transactionsTable = prefix + "transactions";
  logsTable = prefix + "logs";
  subscriptionsTable = prefix + "subscriptions";
  subscribersTable = prefix + "subscribers";
}


void MDLogWriter::setNextCommand(const std::string& nextCmd) {
//  StatsCollector::incCounter(COUNTER_COMMANDS);
	if (MDLogWriter::_saveLogs) {
		DMESG("Setting command: " << nextCmd << endl);
		_nextCmd = nextCmd;
	}
}

bool MDLogWriter::isSubscribed(
    Statement &statement, 
    const string &absDirectory) throw (SQLException)
{
  string query;
  if (absDirectory == USER_CMD_META_DIR) {
    // It's a user management command. 
    query = "SELECT 1 FROM " + subscribersTable + " WHERE \"sub_rep_users\" = 1";
  } else {
    // It's a metadata manipulation command. 
    query = "SELECT 1 FROM " + subscriptionsTable;
    if (absDirectory != "") {    
      list<string> parts;
      pathPieces(absDirectory, parts);
      // Check if there is any subscriber for this directory
      query += " WHERE "; 
      list<string>::iterator it=parts.begin(); 
      while (it != parts.end()) {
	query += "(\"sub_directory\" = '";
	query += *it;
	query += "')";
	it++;
	if(it != parts.end()) {
	  query += " OR ";
	}
      }
    }
  }
  // LIMIT is not accepted by Oracle.
  //  query += " LIMIT 1;";
  query += ";";
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
    throw SQLException(statement);
  }
  int dummy;
  statement.bind(1, &dummy);
  // If fetch executes correctly, them the query returned at least one result
  bool results = (statement.fetch() == 0);
  statement.close();
  return results;
}


// Needed by addentries/removeentries.
void MDLogWriter::saveLogCommand(Statement &statement, 
  const string &command, // The command to be saved
  const string &absDirectory, // The directory updated
  const string &user, // The user that executed the command
  const string &currentDir, // The context directory
  const string &permissionMask, // The user that executed the command
  const string &groupRightsMask) throw (SQLException) 
{
  if (MDLogWriter::_saveLogs) {    
    // Make a mutable copy
    string cmd = command;
    // Must quote the command as was received from the user. Done
    // differently for Oracle, Postgres and MySQL.
    statement.quoteDataField(cmd);

    // Save log.
    string query;
    if (transactionInProgress()) {
      // we are in the middle of a transaction. Save the log only if 
      // this transaction should be saved.
      if (_xid == 0) {
        DMESG("Transaction is not being saved. Discarding log\n");
        return;
      }
      // Must save the log to the transaction table, not to the main log table.
      ostringstream oss;
      oss << "INSERT INTO " << transactionsTable << 
          " (\"tr_transaction_id\", \"tr_log_directory\", \"tr_log_cwd\", \"tr_log_command\") "
          "VALUES ("
          << _xid << ", "
          << "'" << absDirectory << "', "
          << "'" << currentDir << "', "
          << "'" << cmd << "');";
      query = oss.str();
    } else {
      if (!isSubscribed(statement, absDirectory)) {
        DMESG("No subscription. not saving log\n");
        return;
      }
      // There is at least one subscriber for this directory 
      query = "INSERT INTO " + logsTable + 
          " (\"log_directory\", \"log_user\", \"log_cwd\"," 
          "\"log_permissions\", \"log_group_rights\", \"log_command\") "
          "VALUES ("
          "'" + absDirectory + "', "
          "'" + user + "', "
          "'" + currentDir + "', "
          "'" + permissionMask + "', "
          "'" + groupRightsMask + "', "
          "'" + cmd + "');";
    }
    DMESG("[LOGWRITER] " << query << endl);
    if (statement.exec(query)) {
        throw SQLException("Failed to write log.", statement);
    }
    
    // DEBUG
    SubscriptionManager sm(statement.getDatabaseConnection());
    uint64_t xid = sm.getCurrentXid();
    DMESG("Insertion XID: " << xid << endl);    
    if (xid < lastXid) {
      ostringstream oss;
      oss << "xid sequence resetted. Old xid: " << lastXid << " , newXid: " << xid << ", log: " << query;
      throw runtime_error(oss.str()); 
    }
    lastXid = xid;
    
//    StatsCollector::incCounter(COUNTER_LOGS_WRITTEN);
  }
  // Success. Either saved log correctly or saving of logs is disabled.          
}

// Saves a log of an update to the database.
/** 
 * Must be called from within a transaction.
 */
void MDLogWriter::saveStoredLog(Statement &statement, 
  const string &absDirectory, // The directory updated
  const string &user, // The user that executed the command
  const string &currentDir, // The context directory
  const string &permissionMask, // The user that executed the command
  const string &groupRightsMask) throw (SQLException)
{
  // The server object calls this method for all update operations
//  StatsCollector::incCounter(COUNTER_UPDATES);
  if (MDLogWriter::_saveLogs) {
    saveLogCommand(statement, 
      _nextCmd, absDirectory, user, currentDir, 
      permissionMask, groupRightsMask);
    // Clear the command, since each log should only be saved once.
    _nextCmd = "";
  }
}

uint64_t MDLogWriter::getLastLogXid(Statement &st) const
throw (SQLException) 
{ 
  switch (st.getDatabaseConnection().getDriverType()) {
  case DatabaseConnection::dtPSQL :
    return st.sequenceGetCurrentValue("logs_log_xid_seq");
  case DatabaseConnection::dtOracle :
    return st.sequenceGetCurrentValue("logs_seq");
  case DatabaseConnection::dtMySQL5 :  
  {
    // Not tested
    string query = "SELECT LAST_INSERT_ID();";
    DMESG("[LOGWRITER] " << query << endl);
    if (st.exec(query)) {
      throw SQLException(st);
    }
    if (st.fetch()) {
      throw SQLException(st);
    }  
    uint64_t v;
    st.getData(1, v);
    st.closeCursor();
    return v;
  }
  default:
    throw SQLException("Not supported");
  }
}

void MDLogWriter::beginUpload(Statement &statement,
      const string &absDirectory, // The directory updated
      const std::string &user, // The user that executed the command
      const string &cwd, // The directory updated
      const std::string &permissionMask,
      const std::string &groupRightsMask) throw (SQLException)
{
  // An upload and a transaction count as a single log.
  // StatsCollector::incCounter(COUNTER_LOGS_WRITTEN);
  if (MDLogWriter::_saveLogs) {
    if (isSubscribed(statement, absDirectory)) {
      ostringstream oss;
      oss << "INSERT INTO " << logsTable << 
        " (\"log_directory\", \"log_user\", \"log_cwd\"," 
        "\"log_permissions\", \"log_group_rights\", \"log_command\") "
        "VALUES ('" 
        << absDirectory << "', "
        "'" << user << "', "
        "'" << cwd << "', "
        "'" << permissionMask << "', "
        "'" << groupRightsMask << "', "
        "'" << _nextCmd << "');";
      string query = oss.str();
      DMESG("[LOGWRITER] " << query << endl);
      if (statement.exec(query)) {
        throw SQLException("Failed to write log.", statement);
      }
      DMESG("Wrote log\n");
      _xid = getLastLogXid(statement); 
      DMESG("XID: " << _xid << endl);  
    }
    _inTransaction = true;
  }
}

void MDLogWriter::beginTransaction(
      Statement &statement,
      const std::string &user, // The user that executed the command
      const std::string &permissionMask,
      const std::string &groupRightsMask) throw (SQLException)
{
  // Uploads and transactions count as a single log.
//  StatsCollector::incCounter(COUNTER_LOGS_WRITTEN);

  // Check if there is at least one subscriber.
  if (MDLogWriter::_saveLogs) {
    if (isSubscribed(statement, "")) {
      ostringstream oss;
      oss << "INSERT INTO " << logsTable << 
        " (\"log_directory\", \"log_user\", \"log_cwd\"," 
        "\"log_permissions\", \"log_group_rights\", \"log_command\") "
        "VALUES ("
        "'*', "
        "'" << user << "', "
        "NULL, "
        "'" << permissionMask << "', "
        "'" << groupRightsMask << "', "
        "'transaction');";
      string query = oss.str();
      DMESG("[LOGWRITER] " << query << endl);
      if (statement.exec(query)) {
        DMESG("Error: " << statement.getError() << endl);
        throw SQLException("Failed to write log.", statement);
      }
      // Get the xid
      _xid = getLastLogXid(statement);
    }
    _inTransaction = true;
  }
}

void MDLogWriter::commit() {
  _inTransaction = false;
  _xid = 0;
}

void MDLogWriter::abort() {
  _inTransaction = false;  
  _xid = 0;
}

bool MDLogWriter::transactionInProgress() {
  return  _inTransaction == true;
}
