/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "MountManager.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "util/Statement.h"
#include "util/DatabaseConnection.h"
#include "util/Utils.h"
#include "server/MDServer.h"
#include "server/MDStandalone.h"
#include "ReplicationManager.h"
#include "MDLogWriter.h"
#include "MountManager.h"

#include <assert.h>
#include <iosfwd>
#include <iterator>

#include "debug.h"

using namespace std;

const string Mount::MOUNTED("Mounted");
const string Mount::SYNCHRONIZING("Synchronizing");
const string Mount::SYNCHRONIZED("Synchronized");
const string Mount::UNKNOWN("Unknown");

/************************************
 * MountManager
 ************************************/

MountManager::MountManager(DatabaseConnection &dbConnection) 
: _dbConn(dbConnection) 
{
  debug = ReplicationManager::getInstance()->getDebug();
  DMESG("Reusing connection\n");
  setupTableNames();
}

MountManager::~MountManager() {
//    DMESG("~MountManager\n");
}


void MountManager::setupTableNames()
{
  std::string prefix;
  if(DatabaseConnection::dbSchema.size())
    prefix = DatabaseConnection::dbSchema + ".";

  mastersTable = prefix + "masters";
  mountsTable = prefix + "mounts";
  masterindexTable = prefix + "masterindex";
  amgaTable = prefix + "amga";
}


void MountManager::getAllMounts(vector<Mount> &mounts) 
throw (SQLException) 
{
  getMounts(mounts, "");
}


void MountManager::getMountsOfMaster(
	const std::string &master, std::vector<Mount> &mounts) 
    throw (SQLException)
{
  getMounts(mounts, string("\"mnt_master\" = '") + master + "'");
}

void MountManager::getActiveMasters(vector<Master> &masters) 
throw (SQLException) 
{
  getMasters(masters, " \"mst_active\" = 1");
}

//void MountManager::getAllMasters(vector<Master> &masters) 
//throw (SQLException) 
//{
//    getMasters(masters, "");
//}

auto_ptr<Mount> MountManager::getMountInfo(const string &directory) 
throw (SQLException) 
{
  vector<Mount> v;    
  getMounts(v, string("\"mnt_directory\" = '") + directory + "'");
  if (v.empty()) {
    DMESG("Directory is not mounted\n");
    // Return unassigned auto_ptr
    return auto_ptr<Mount>();
  } else {
    if (v.size() != 1) {
      // Sanity check
      ostringstream oss;
      oss << "More than one mount in the same mount point:\n";
      ostream_iterator<Mount> out(oss, "\n");
      copy(v.begin(), v.end(), out);
      oss << "\n";
      LOG(oss.str());
      assert(false);
    }
    return auto_ptr<Mount>(new Mount(v[0]));
  }    
}

auto_ptr<Master> MountManager::getMaster(const string &master) throw (SQLException) {
  vector<Master> v;    
  getMasters(v, string("\"mst_id\" = '") + master + "'");
  if (v.empty()) {
    // Return unassigned auto_ptr
    return auto_ptr<Master>();
  } else {
    assert(v.size() == 1);
    return auto_ptr<Master>(new Master(v[0]));
  }
}


///**
// * Gets the latest xid that was applied to the given directory
// */
//uint64_t MountManager::getDirectoryXid(const std::string &directory) 
//    throw (SQLException)
//{
//  Statement statement(_dbConn);
//  string query = "SELECT \"rep_xid\" FROM masterindex WHERE \"directory\" = '" 
//      + directory + "';";
//  DMESG("SQL: >" + query + "<\n");        
//  if(statement.exec(query)) {
//    throw SQLException(statement);
//  }
//  uint64_t xid;
//  if (statement.fetch()){
//    throw SQLException("Failed to retrieve most recent xid", statement);
//  }
//  if (statement.getData(1, xid)) {
//    // Failed
//    return 0;
//  } else {
//    DMESG("XID: " << xid << endl);    
//    return xid;
//  }
//}

/**
 * Gets the latest xid that was received by 
 */
//uint64_t MountManager::getMasterXid(const std::string &mount) 
//    throw (SQLException)
//{
//  Statement statement(_dbConn);
//  string query = "SELECT \"mst_last_xid\" FROM masters, mounts "
//    " WHERE mounts.\"mnt_master\" = masters.\"mst_id\" "
//    " AND \"mnt_directory\" = '" + mount + "';";
//  DMESG("SQL: >" + query + "<\n");
//  if(statement.exec(query)) {
//    throw SQLException(statement);
//  }
//  
//  uint64_t xid;
//  if (statement.fetch()){
//    throw SQLException("Failed to retrieve most recent xid", statement);
//  }
//  statement.getData(1, xid);
//  return xid;
//}



void MountManager::umountAll()  throw (SQLException) {
  // Update both the masters and the mounts table inside a transaction
//  Statement statement(cHandle, debug);
//  statement.beginTransaction(true);
  //TODO: Must delete all the tables that are mounted and update the masterindex


//    // TODO: if the master table is empty, this will return an error
//    string query("DELETE FROM masters;");
//    if(statement.exec(query)){
//        if (statement.getError() != "") {
//            throw SQLException("Cannot delete from masters table", statement);
//        }
//    }
//  statement.commitTransaction();  
}

Mount::MountState Mount::mountStateToEnum(const string &stateStr) throw (SQLException) {
  if (stateStr == MOUNTED) {
    return Mount::Mounted;
  } else if (stateStr == SYNCHRONIZING) {
  	return Mount::Synchronizing;
  } else if (stateStr == SYNCHRONIZED) {  	
    return Mount::Synchronized;
  } else if (stateStr == UNKNOWN) {
    return Mount::Unknown;     
  } else {
    // Unknown state. Die!
    throw SQLException("Invalid state string: " + stateStr);
  }
}
    
void MountManager::getMasters(vector<Master> &masters, 
        const string &whereClause)  throw (SQLException) 
{
  masters.clear();
  string query = "SELECT \"mst_id\", \"mst_active\", \"mst_last_xid\" FROM "
    + mastersTable;
  if (whereClause.size() != 0) {
    query += " WHERE " + whereClause;
  }
  query += ";";
  DMESG("SQL: " << query << endl);
  Statement statement(_dbConn);
  if(statement.exec(query)){
    throw SQLException(statement);
  }
  Master master;
  char id[ADDRESS_MAX_SIZE];        
  statement.bind(1, id, sizeof(id));
  short int active;
  statement.bind(2, &active);      

  while (!statement.fetch() ){
    // Read the script from the result of the query.
    statement.getData(3, master.masterXid);
    master.id = id;
    master.active = (active != 0);
    DMESG("Fetched: " << master << endl);
    masters.push_back(master);
  }
}

void MountManager::getMounts(vector<Mount> &mounts, const string &whereClause)  throw (SQLException)
{
  mounts.clear();
  string query = "SELECT \"mst_id\", \"mst_active\", \"mst_last_xid\", \"mnt_state\", "
  	  "\"mnt_directory\", \"mnt_rep_permissions\" "
      "FROM " + mountsTable + ", " + mastersTable
      + " WHERE mounts.\"mnt_master\" = masters.\"mst_id\"";

  if (!whereClause.empty()) {
      query += " AND " + whereClause;
  }
  query += ";";
  DMESG("SQL: " << query << endl);
  Statement statement(_dbConn);
  if (statement.exec(query)) {
    // This method returns non-zero either when no rows are updated
    // or when the connection fails. We check the error to distinguish
    // between the two cases.
    if (statement.getErrorCode() != SQL_NO_DATA) {
      throw SQLException(statement);
    }
  }

  Mount mount;
  char id[ADDRESS_MAX_SIZE];
  statement.bind(1, id, sizeof(id));
  short int mstActive;
  statement.bind(2, &mstActive);      
  char state[MOUNT_STATE_MAX_SIZE];
  statement.bind(4, state, sizeof(state));
  char directory[DIRNAME_MAX_SIZE];
  statement.bind(5, directory, sizeof(directory));
  short int repPermissions;
  statement.bind(6, &repPermissions);

  while (!statement.fetch() ){
    statement.getData(3, mount.master.masterXid);
    mount.master.active = mstActive;
    mount.state = Mount::mountStateToEnum(state);
    mount.repPermissions = repPermissions;
    mount.directory = directory;
    mount.master.id = id;
//    mount.repUsers = repUsers;
    DMESG("Fetched: " << mount << endl);
    // Read the script from the result of the query.
    mounts.push_back(mount);
  }

  if (statement.getErrorCode() != SQL_NO_DATA) {
    throw SQLException("Error in query: ", statement);
  }
}

void MountManager::testMount(
    const string &masterID,
    const string &directory,
    MDStandalone &server)  throw (SQLException, ReplicationException)
{    
  DMESG("preMounting: " << masterID << ":" << directory << endl);
          
  // Sanity checks.
  // 1. Is this mount point already in use?
  auto_ptr<Mount> mountInfo = getMountInfo(directory);
  if (mountInfo.get() != NULL) {
      ostringstream oss;
      oss << "Directory already mounted: " << *mountInfo;
      throw ReplicationException(oss.str());
  }
   
  // 2. Directory must not exist, unless it is the root directory, in each case it should be empty.
  if (directory == "/") {
    vector<string> results;
    server.listAttr(directory);
    try {
      server.getResult(results);
    } catch (runtime_error &ex) {
      throw ReplicationException(string("Mount target does not exist (") + ex.what() + ")");
    }
    if (!results.empty()) {
      throw ReplicationException("Mount target not empty. Contains attributes.");
    }   
    server.dir(directory);
    server.getResult(results);     
    if (!results.empty()) {
      throw ReplicationException("Mount target not empty. Contains entries.");
    }
  } else {
    // It is a normal directory. It must not exist.
    server.dir(directory);
    try {
      vector<string> results;
      server.getResult(results);
      throw ReplicationException("Mount target already exist. Delete it first.");
    } catch (runtime_error &ex) {
      // It's ok.
    }
  }
}

void MountManager::insertMasterIfNotExists(const string &masterID) throw (SQLException) {	
	auto_ptr<Master> master = getMaster(masterID);
  	if (master.get() == NULL) {
      // 0 XID means that the directory is not yet synchronized.
      string query =  "INSERT INTO masters (\"mst_id\", \"mst_active\", \"mst_last_xid\" ) VALUES ('";
      query += masterID;
      query += "', 0, 0);";
      DMESG("SQL: " << query << endl);
      Statement st(_dbConn);
      if(st.exec(query)) {
        throw SQLException("Cannot update masters table", st);
      }
    }	
}

void MountManager::mountUsers(const string &masterID)
	throw (SQLException, ReplicationException)
{
	string usersMaster = getUsersMaster(); 
	if (usersMaster != "") {
		throw ReplicationException("Users are already being replicated from " + usersMaster); 
	}
	// It's fine, we are not replicating from nowhere.
	insertMasterIfNotExists(masterID);

	// Write to the amga table to set the new master for the user/group 
	// information
	string query = "UPDATE amga SET \"rep_users_master\"='";
	query += masterID;
	query += "', \"rep_users_xid\"=0;";
	Statement statement(_dbConn);
	DMESG("SQL: " << query << endl);
	if(statement.exec(query)) {
   		throw SQLException("Cannot mount users", statement);
  	} 	
}

void MountManager::umountUsers(const string &master)  
	throw (SQLException, ReplicationException)
{
	// Even if the users are not being replicated, 
	// perform the call without raising an error.
	string query = "UPDATE amga " 
    "SET \"rep_users_master\"=NULL, \"rep_users_xid\"=NULL " 
    "WHERE \"rep_users_master\"='" + master + "';";   
	Statement statement(_dbConn);
	if(statement.exec(query)) {
    if (statement.getErrorCode() == SQL_NO_DATA) {
      throw ReplicationException("No users mounted from master");      
    } else {    
      throw SQLException("DB error", statement);
    }
  }
  // Now delete from the masters table 
  query = "DELETE FROM " + mastersTable + " WHERE \"mst_id\"='";
  query += master;
  query += "' AND NOT EXISTS (SELECT 1 FROM " + mountsTable + " WHERE \"mnt_master\" = '";
  query += master;
  query += "');"; 
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
    if (statement.getErrorCode() == SQL_NO_DATA) {
      // Ignore, this probably means that the users are mounted from this master
    } else {
      throw SQLException(statement);
    }
  }
}

void MountManager::updateMountXid(const std::string &mount, uint64_t xid, bool ignoreNoRowsUpdated) 
throw (SQLException)
{
  if (mount == USER_CMD_META_DIR) {
    updateUsersXid(xid);
    
  } else {
    Statement statement(_dbConn);
    ostringstream os;
    os << "UPDATE " << masterindexTable << " SET \"rep_xid\" = " << xid
       << " WHERE \"directory\" = '" << mount << "';";

    string query = os.str();
    DMESG("SQL: " << query << endl);
    if (statement.exec(query)) {
      if (ignoreNoRowsUpdated && statement.getErrorCode() == SQL_NO_DATA) {
        // When ignoreNoRowsUpdated is true, then we don't raise an error
        // if no rows are updated.
        return;
      }
	    throw SQLException(statement); 
    }
  }
}

uint64_t MountManager::getMountXid(const std::string &mount) 
throw (SQLException)
{
  if (mount == USER_CMD_META_DIR) {
    return getUsersXid();		
  } else {
    Statement statement(_dbConn);
    string query = "SELECT \"rep_xid\" FROM " + masterindexTable + " WHERE \"directory\" = '" 
      + mount + "';";
    DMESG("SQL: >" + query + "<\n");        
    if(statement.exec(query)) {
      throw SQLException(statement);
    }
    uint64_t xid;
    if (statement.fetch()){
      throw SQLException("Failed to retrieve most recent xid", statement);
    }
    if (statement.getData(1, xid)) {
      // Failed
      return 0;
    } else {
      DMESG("XID: " << xid << endl);    
      return xid;
    }
  }	
}


void MountManager::updateUsersXid(uint64_t xid) 
throw (SQLException)
{
	string usersMaster = getUsersMaster(); 
	if (usersMaster == "") {
		throw ReplicationException("Users are not being replicated"); 
	}
	ostringstream oss;
	oss << "UPDATE " << amgaTable << " SET \"rep_users_xid\"=" << xid << ";";
	Statement statement(_dbConn);
	DMESG("SQL: " << oss.str() << endl);
	if(statement.exec(oss.str())) {
   		throw SQLException("Cannot umount users", statement);
  	}
}

uint64_t MountManager::getUsersXid() 
throw (SQLException)
{
	return getUsersMount().xid;	    	
}

std::string MountManager::getUsersMaster()  
	throw (SQLException)
{
	return getUsersMount().id;	
}

UserMount MountManager::getUsersMount() 
throw (SQLException) 
{
	string query = "SELECT \"rep_users_master\", \"rep_users_xid\" FROM "
	  + amgaTable + ";";
	Statement statement(_dbConn, debug);
	DMESG("SQL: " << query << endl);
	if(statement.exec(query)) {
   		throw SQLException("Cannot retrieve master node of users", statement);
  	}
  	UserMount uMount;
  	char id[ADDRESS_MAX_SIZE];
    statement.bind(1, id, sizeof(id));
	if (statement.fetch()) {
        throw SQLException("Cannot retrieve master node of users", statement);        
	}
	uMount.id = id;
	// Return 1 if the value is NULL
  	if (statement.getData(2, uMount.xid)) {
  		if (statement.getErrorCode() == SQL_NULL_DATA) {
  			uMount.xid = 0;
  		} else {
  			throw SQLException("Error retrieving users mount xid", statement);
  		}
  	}
	return uMount;
}

/** 
 * Mounts a remote directory.
 *
 * Contats the remote host
 */
void MountManager::mount(
        const string &masterID, 
        const string &directory,
        bool repPermissions,
        MDStandalone &server)  throw (SQLException, ReplicationException)
{
  DMESG("Mounting: " << directory << "@" << masterID << endl);
  // Update the local tables.
  // Update both the masters and the mounts table inside a transaction
  Statement statement(_dbConn, debug);

  statement.beginTransaction();
  testMount(masterID, directory, server);
  server.setDebug(debug);

//  if (directory != "/") {
//    DMESG("Creating directory: " << directory << endl);
//    server.createDir(directory);
//    vector<string> v;
//    try {
//      // Throws an exception if it fails
//      server.getResult(v);
//    } catch (runtime_error &ex) {
//      // Either directory already exists or the parent directory does not exist
//      throw ReplicationException(string("Cannot create local directory: ") + ex.what());
//    }
//  }
  
  // Check if master is already present in the masters table.
  // 1. Insert it if not.
  insertMasterIfNotExists(masterID);  

  // Now the mounts table
  ostringstream oss;
  oss << "INSERT INTO " << mountsTable << " (\"mnt_master\", \"mnt_directory\", " 
      	"\"mnt_rep_permissions\", \"mnt_state\" ) "
      	"VALUES ('" 
      	<< masterID << "', '" 
      	<< directory << "', " 
      	<< repPermissions << ", '"
      	<< Mount::MOUNTED << "');";
  string query = oss.str(); 
  DMESG("SQL: " << query << endl);
  if(statement.exec(query)) {
    throw SQLException("Cannot update mount table", statement);
  }
  statement.commitTransaction();
}

void MountManager::umount(const string &directory) throw (SQLException, ReplicationException) {
  DMESG("Unmounting " << directory << endl);

  Statement statement(_dbConn);
  statement.beginTransaction();
  // Get the name of the master and the number of mounts 
  // from this master
  string query = "SELECT \"mnt_master\" FROM " + mountsTable
    + " WHERE \"mnt_master\" = (SELECT \"mnt_master\" FROM " + mountsTable 
    + " WHERE \"mnt_directory\" = '" + directory + "');";
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
    if (statement.getErrorCode() == SQL_NO_DATA) {
      throw ReplicationException("Subscription not found");
    } else {
      throw SQLException(statement);
    }
  }
  char master[NODE_ID_MAX_SIZE];
  statement.bind(1, master, sizeof(master));
  int count=0;
  while (!statement.fetch()) {
    count++;
  }
  statement.close();
  
  assert(count > 0);
  query = "DELETE FROM " + mountsTable + " WHERE \"mnt_directory\"='" + directory + "';";
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
    if (statement.getErrorCode() == SQL_NO_DATA) {
      throw ReplicationException("Subscription not found");
    } else {
      throw SQLException(statement);
    }
  }
  statement.close();
  
  if (count == 1) {
    // There was only one entry for this master. Delete it from the master table.
    query = "DELETE FROM " + mastersTable + " WHERE \"mst_id\"='";
    query += master;
    query += "' AND NOT EXISTS (SELECT 1 FROM " + amgaTable 
      + " WHERE \"rep_users_master\" = '";
    query += master;
    query += "');"; 
    DMESG("SQL: " << query << endl);
    if (statement.exec(query)) {
      if (statement.getErrorCode() == SQL_NO_DATA) {
        // Ignore, this probably means that the users are mounted from this master
      } else {
        throw SQLException(statement);
      }
    }
  }
  statement.close();

  query = "UPDATE " + masterindexTable + " SET \"master\" = '' WHERE \"directory\" like '" + directory + "%';";
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
    if (statement.getErrorCode() == SQL_NO_DATA) {
      // SIAHN_20110110
      // ignore the below line to fix "Synchronizing bug"
      // throw ReplicationException("Can not remove master record in masterindex");
    } else {
      throw SQLException(statement);
    }
  }
  statement.close();

  statement.commitTransaction();
}

//void MountManager::updateDirectoryXid(const std::string &directory, uint64_t newXid) 
//    throw (SQLException)
//{
//  Statement statement(_dbConn);
//  ostringstream os;
//  os << "UPDATE masterindex SET \"rep_xid\" = " << newXid
//          << " WHERE \"directory\" = '" << directory << "';";
//  string query = os.str();
//  DMESG("SQL: " << query << endl);
//  if (statement.exec(query)) {
//    throw SQLException(statement);
//  }
//}

void MountManager::getMountsInState(Mount::MountState state, vector<Mount> &mounts) 
throw (SQLException) 
{
	getMounts(mounts, "\"mnt_state\" = '" + Mount::mountStateToString(state) + "'");
}

void MountManager::updateMasterXid(const std::string &master, uint64_t newXid) 
throw (SQLException)
{
  Statement statement(_dbConn);
  ostringstream os;
  os << "UPDATE masters SET \"mst_last_xid\"=" << newXid 
    << " WHERE \"mst_id\"='" << master << "';";
  string query = os.str();
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
      throw SQLException(statement);
  }
}

void MountManager::updateMountState(const string &mount, Mount::MountState state) 
throw (SQLException) 
{
  Statement statement(_dbConn);  
  ostringstream oss;
  oss << "UPDATE mounts SET \"mnt_state\"='" << Mount::mountStateToString(state)
      << "' WHERE \"mnt_directory\"='" << mount << "';";
  string query = oss.str();
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
      throw SQLException(statement);
  }
}

void MountManager::updateMasterState(const string &master, bool active) 
throw (SQLException) 
{
  Statement statement(_dbConn);  
  ostringstream oss;
  oss << "UPDATE masters SET \"mst_active\"=" << (active ? 1 : 0)
      << " WHERE \"mst_id\"='" << master << "';";
  string query = oss.str();
  DMESG("SQL: " << query << endl);
  if (statement.exec(query)) {
      throw SQLException(statement);
  }
}

string Mount::mountStateToString(Mount::MountState state) {
  switch (state) {    
    case Mounted: return MOUNTED;
    case Synchronizing: return SYNCHRONIZING;
    case Synchronized: return SYNCHRONIZED;
    case Unknown: return UNKNOWN;
    default:
      ostringstream os;
      os << "Invalid mount state: " << state;
      throw logic_error(os.str());
  }
}

  
ostream& operator<<(ostream& out, const Mount &mnt) {
  out << mnt.directory << "/" << mnt.master 
  	<< ", repPermissions: " << mnt.repPermissions 
  	<< ", " << mnt.getStateAsString();
  return out;
}    

ostream& operator<<(ostream& out, const Master &master) {
  out << master.id << " - " << master.masterXid << ", " << (master.active? "Active":"Inactive");
  return out;
}    

std::ostream& operator<<(std::ostream& out, const UserMount &mount) {
	out << mount.id << " - " << mount.xid;
	return out;
}
