/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef REPDBCLIENT_H_
#define REPDBCLIENT_H_

#include <ostream>
#include <vector>
#include <boost/thread/once.hpp>

class MDServer;
class Statement;
class MDStandalone;
class DatabaseConnection;

#include "server/Constants.h"
#include "server/MDBuffer.h"
#include "util/DBUtil.h"
#include "util/Debuggable.h"
#include "Exceptions.h"
#include <stdint.h>
#include <memory>


struct Master {    
  std::string id;
  uint64_t masterXid;
  bool active;

  Master() : masterXid(0), active(false) {};
  friend std::ostream& operator<<(std::ostream& out, const Master &master);
};    


/**
 * Represents a subscription to a remote directory.
 */
struct Mount {
  enum MountState {
    Unknown, Mounted, Synchronizing, Synchronized
  };
  /* Subscribed to master, but no data copied. */ 
  static const std::string MOUNTED;
  /* Copying data from master. This state is only temporary and is used to detect and
   * recover from failed synchronizations after crashes of the slave */
  static const std::string SYNCHRONIZING; 
  /* The metadata was copied from the master, but we are not receiving logs. */
  static const std::string SYNCHRONIZED;
  static const std::string UNKNOWN;


  std::string directory;
  bool repPermissions;  
  MountState state;    
  Master master;

	
  static std::string mountStateToString(MountState state);
  static MountState mountStateToEnum(const std::string &stateStr) 
    throw (SQLException);	
  std::string getStateAsString() const {
    return mountStateToString(state);
  };
  Mount() : repPermissions(false), state(Unknown) {};  
  
  friend std::ostream& operator<<(std::ostream& out, const Mount &mount);
};

struct UserMount {
  uint64_t xid;
  std::string id;  
  
  UserMount() : xid(0){};
  friend std::ostream& operator<<(std::ostream& out, const UserMount &mount);
};


/** 
 * Handles all operations that contact the DB at the slave.
 */
class MountManager : public Debuggable
{
private:
    DatabaseConnection  &_dbConn;
//    bool                _ownsConnection;
    
    // Singleton pattern
//    static boost::once_flag once;
//    static MountManager *instance;
//    static void init() throw (SQLException, ReplicationException);
    
//    void getMasters(std::vector<Master> &masters, 
//        const std::string &whereClause)  throw (SQLException);    

    void insertMasterIfNotExists(const std::string &masterID) 
    	throw (SQLException); 
    void getMounts(std::vector<Mount> &mounts, 
        const std::string &whereClause)  throw (SQLException);

public:
  /**
   * Performs all operations using the given connection. This 
   * constructor is intended for callers that want to execute 
   * the operations on this class inside a transaction.
   */
  MountManager(DatabaseConnection &_dbConn);
//	MountManager() throw (ReplicationException);
  
  virtual ~MountManager();
    
//  static MountManager* getInstance() throw (SQLException, ReplicationException);
  
  void getMasters(std::vector<Master> &masters, const std::string &whereClause) 
    throw (SQLException);
  void getActiveMasters(std::vector<Master> &masters) 
    throw (SQLException);
  std::auto_ptr<Master> getMaster(const std::string &master) 
    throw (SQLException);    
  void getAllMounts(std::vector<Mount> &subs) 
    throw (SQLException);
  void getMountsOfMaster(const std::string &master, std::vector<Mount> &subs) 
    throw (SQLException);
  void getMountsInState(Mount::MountState state, std::vector<Mount> &mounts) 
  	throw (SQLException);
    
  /**
   * @return an unassigned auto_ptr if there is no mount on the given mount point
   * or the information about the mount
   */
  std::auto_ptr<Mount> getMountInfo(const std::string &directory) throw (SQLException);
//    void getAllMasters(std::vector<Master> &masters) throw (SQLException);
    
  void mount(const std::string &masterID, 
      const std::string &directory,
      bool repPermissions, MDStandalone &server)
      throw (SQLException, ReplicationException);
 
  /** 
   * Performs some of the tests that would be done while creating the
   * given mount, but does not change anything on the database. 
   * Passing this test does not ensure that the mount will succeed,
   * it only checks for the most common causes of failure for mounts.
   */
  void testMount(const std::string &masterID,
		 const std::string &directory, MDStandalone &server)  
    throw (SQLException, ReplicationException);
  
  void umount(const std::string &directory) throw (SQLException, ReplicationException);
    
  void umountAll() throw (SQLException);
  
  void mountUsers(const std::string &masterID)  
    throw (SQLException, ReplicationException);
  
  void umountUsers(const std::string &master)  
    throw (SQLException, ReplicationException);
  
  std::string getUsersMaster()  
    throw (SQLException);
  UserMount getUsersMount()
    throw (SQLException);
  
  void updateUsersXid(uint64_t xid) 
    throw (SQLException);
  uint64_t getUsersXid() 
    throw (SQLException);
  
  void updateMountXid(const std::string &mount, uint64_t xid, bool ignoreNoRowsUpdated = false) 
    throw (SQLException);	
  
  /** Get the mount xid on the mounts table. */
  uint64_t getMountXid(const std::string &mount)
    throw (SQLException);        
  void updateMountState(const std::string &mount, Mount::MountState state) 
    throw (SQLException);
  void updateMasterState(const std::string &master, bool active)
    throw (SQLException);

 /** Update the master xid on the masters table.*/
  void updateMasterXid(const std::string &master, uint64_t newXid)
    throw (SQLException);
  
  void setupTableNames();

  std::string mastersTable;
  std::string mountsTable;
  std::string masterindexTable;
  std::string amgaTable;
};

#endif /*REPDBCLIENT_H_*/
