/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "ReplicationDaemon.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "server/Constants.h"
#include "server/protocolHandling.h"
#include "server/MDStandalone.h"
#include "server/MDSocketBuffer.h"
#include "server/MDServerFactory.h"
#include "ReplicationManager.h"

#include "util/SocketUtil.h"
#include "util/Socket.h"
#include "util/Utils.h"
#include "util/Statement.h"
#include "util/Clock.h"

#include "debug.h"

#include <boost/tokenizer.hpp>
#include <boost/thread/thread.hpp>
#include <boost/algorithm/string/predicate.hpp>

#include <openssl/ssl.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/safestack.h>

boost::mutex ConnectionHandler::_mutex;
std::set<std::string> ConnectionHandler::_activeConnections;

using namespace std;
using namespace boost;

ReplicationDaemon::ReplicationDaemon(int listenPort, ConfigParser &configParser)
: Runnable(true), _listenPort(listenPort), _configParser(configParser)
{
  debug = ReplicationManager::getInstance()->getDebug();
  _logMonitor.reset(new LogMonitorThread(3, configParser));
  // Loads the ssl related configuration and initializes 
  // SSL if requested on the config file.
  DMESG("Initializing SSL\n");
  _authCtx.init(configParser);
}

ReplicationDaemon::~ReplicationDaemon()
{
    DMESG("~ReplicationDaemon\n");
}

void ReplicationDaemon::run() {
  DMESG("MASTER thread starting\n");
  // Start the thread that will check for new logs.
  _logMonitor->start();

	DMESG("Listening on port: " << _listenPort << std::endl);
	try {
    TCPServerSocket servSock(_listenPort);     // Server Socket object
    int i = 0;
    while (!isCancelled()) {
      shared_ptr<TCPSocket> sock(servSock.accept());
      try {
        // TCP client handling function
        DMESG("ACCEPT " << i << endl);
        i++;
        // Just in case start() throws an exception
        auto_ptr<ConnectionHandler> cHandler(new ConnectionHandler(sock, *this));
        cHandler->start();
        // The thread started successfully. when it finishes, it will
        // delete the ConnectionHandler object so we can release it
        // from the auto_ptr
        cHandler.release();
      } catch (SocketException &e) {
        LOG(e.what() << endl);        
      } catch (runtime_error &e) {
        LOG(e.what() << endl);
      }
    }
	} catch (SocketException &e) {
    LOG("Aborting. Failed to open server socket: " << e.what() << ". Shutting down\n");
    _logMonitor->cancel();
    _logMonitor->join();
	}
	LOG("[Replication daemon] Finished\n");
}


/*****************************************
 * ConnectionHandler class
 ****************************************/
ConnectionHandler::ConnectionHandler(shared_ptr<TCPSocket> sock, ReplicationDaemon &repDaemon)
: Runnable(true), _repDaemon(repDaemon), _authCtx(repDaemon._authCtx), _socket(sock), _subscriberID("")
{
  debug = ReplicationManager::getInstance()->getDebug();
//  DMESG("New connection handler\n");
}

ConnectionHandler::~ConnectionHandler() {  
  DMESG("~ConnectionHandler\n");
  // The connection might not have been established completly 
  // (statistics or failed negotiation). In these cases, there
  // is nothing from this connection in the _activeConnections vector,
  // so we can avoid acquiring this lock.
  if (_subscriberID != "") {
    mutex::scoped_lock lk(_mutex);
    size_t deletedElems = _activeConnections.erase(_subscriberID);
    DMESG("Deleted: " << deletedElems << ", Remaining: " << _activeConnections.size() << endl);
  }
}

void ConnectionHandler::run() {
  DMESG("Connection handler thread starting.\n");

  try {
    LOG("Connection received from: " << _socket->getForeignAddress() << ":" << _socket->getForeignPort() << std::endl);
    // Do client greeting
    // Initialize client connection, check whether SSL is demanded,
    // Resume session if requested
    bool useSSL;
    // We don't support sessions, but initConnection needs this parameter
    uint64_t sessionID = 0;
    string greeting = "ARDA Metadata Replication Daemon 0.1";
    try {
      // Plain text. Send greeting and ask user if SSL is requested.
      int ret = initConnection(*_socket, _authCtx.useSSL, greeting, useSSL, sessionID);
      if (ret == 1) {
        // Handle statistics
        LOG("Sending statistics.\n");
        sendStatistics();
        return;

      } else if (ret != 0) {
        // Failed to initialize connection
        LOG("Failed to initialize connection.\n");
        return;
      }
    } catch (SocketException &ex) {        
      LOG("Protocol negotiation failed: " << ex.what() << std::endl);
      return;
    }

    if (useSSL) {
      DMESG("Initializing SSL\n");
      // The client requested SSL. Switch socket to SSL
      try {
        _socket->initSSLServer(_authCtx.ctx, (char *)NULL);
      } catch (SocketException &ex) {
        LOG("Error initializing SSL: " << ex.what() << std::endl);
        return;
      }
      DMESG("SSL_accepted" << std::endl);
    }

    DMESG("Establishing user context\n");
    try {
      establishUserContext(*_socket, _userCtx);
      DMESG("Context established\n");
    } catch (SocketException &ex) {
      LOG("Error establishing user context: " << ex.what() << endl);
      return;
    }

    DMESG("Connecting to database." << endl);
    _dbConn.reset(DatabaseConnection::openConnection());
    
//    MDAsyncBuffer rwBuffer;
    MDSocketBuffer rwBuffer(_socket.get());
    // Open a new connection to the local backend.
    _mdServer.reset(dynamic_cast<MDStandalone *> 
            (MDServerFactory::mdServer(rwBuffer, _dbConn.get())));
    if(!_mdServer.get() || !_mdServer->isConnected()){
      // Abort connection...
      _socket->send("9 Error connecting to database.\n");
      return;
    }
    // Reuse the MDServer DB connection for the subscription manager
    _subMan.reset(new SubscriptionManager(*_dbConn));
    _mdServer->setDebug(debug);
    
    // Initialize user manager.
    std::list<MDUserManager *> userManagers;
    createUserManagers(_repDaemon._configParser, userManagers);
    _mdServer->initializeUserManager(userManagers);

    DMESG("Authentication required: " << _authCtx.requireAuthentication << std::endl);
    bool userAuthenticated = authenticateUser(
        _socket->getSSLCtx(), _mdServer.get(), _authCtx, _userCtx);
    if (_authCtx.requireAuthentication && !userAuthenticated) {
        LOG("Authentication error for user: " << _userCtx.login << std::endl);
        _socket->send("201 Authentication error\n");
        return;
    }
    
    // Set connection options
    _mdServer->setUser(_userCtx.login);
    _mdServer->setDir(_userCtx.currentDir);
    _mdServer->setPermissionMask(_userCtx.permissionMask);
    _mdServer->setGroupRightsMask(_userCtx.groupMask);
    _mdServer->setLogging(false);
    
    if (useSSL) {
      SSL *ssl = _socket->getSSLCtx();
      X509 *cert = SSL_get_certificate(ssl);
      _mdServer->setPublicKey(X509_get_pubkey(cert));
      _mdServer->setPrivateKey(SSL_get_privatekey(ssl));
    }

    // Use the login as the subscriberID
    if (_userCtx.login == "") {
      _socket->send("1 Empty login\n");
      return;
    }
    _subscriberID = _userCtx.login;
    { // Open scope to delimit duration of the lock
      DMESG("SubscriberID: " << _subscriberID << endl);
      mutex::scoped_lock lk(_mutex);
      
      // Check if this user is already connected.
      if (_activeConnections.count(_subscriberID) == 0 &&
         !_repDaemon._logMonitor->isConnected(_subscriberID))
      {
        _activeConnections.insert(_subscriberID);
      } else {
        LOG("Slave already connected: " << _subscriberID << endl);
	LOG("ActiveConnections: " << _activeConnections.count(_subscriberID)
	    << ", connections according to log-monitor: " 
	    << _repDaemon._logMonitor->isConnected(_subscriberID)
	    << "\n");
        _socket->send("1 Already connected\n");
        return;
      }
    }
    _socket->send("0\n");

    LOG("Client authenticated: " <<  _subscriberID << " (" 
        << _socket->getForeignAddress() << ":" << _socket->getForeignPort() << ")\n");
    CommandParser cParser(*_socket);

    while (true) {
      try {
        DMESG("Waiting for next command." << endl);
        Command c;
        cParser.nextCommand(c);
        DMESG("Command received from client " << c << endl);
        switch (c.cId) {
          case Command::cSubscribe:
            subscribe(c);
            break;

          case Command::cUnsubscribe:
            unsubscribe(c);
            break;
            
          case Command::cSubscribeUsers:
            subscribeUsers(c);
            break;

          case Command::cUnsubscribeUsers:
            unsubscribeUsers(c);
            break;

          case Command::cUnsubscribeAll:
            unsubscribeAll(c);
            break;

          case Command::cSynchronize:
            synchronize(c);
            break;

          case Command::cQuit:
            _socket->send("0\n");
            return;

          case Command::cReceive:
            receive(c);
            // Negotiation is over.
            return;

          default:
              _socket->send("9 Internal error.\n");
              throw logic_error("Invalid command id");
        }
      } catch (CommandParseException& ex) {
        // The client sent a bad command. non-fatal
        LOG("Invalid command: " << ex.what() << endl);
        _socket->send(string("8 Invalid command:") + ex.what() + "\n");

      } catch (runtime_error& ex) {
        // This error is received from the mdserver. It already contains 
        // the error code on the message string.
        LOG("Error processing command: " << ex.what() << endl);
        _socket->send(string(ex.what()) + "\n");
              
      } catch (SQLException& ex) {
        // There was an non-fatal error executing the command.
        LOG("1 Command failed: " << ex.what() << endl);
        _socket->send(string("1 Command failed: ") + ex.what() + "\n");
        
      } catch (SocketException& ex) {
        // Fatal error, abort the connection
        LOG("IO error: " << ex.what() << endl);
        break;
      }
    } 
  } catch (std::exception &ex) {
    DMESG("Connection handler dying: " << ex.what() << endl);
  }
}

void ConnectionHandler::sendStatistics() throw (SocketException) {
  ostringstream stats;
  StatsCollector::dumpAsXml(stats);
  _socket->send(stats.str());
}

void ConnectionHandler::subscribe(const Command &c) 
throw (SocketException, SQLException, runtime_error) 
{
  DMESG("Subscribe\n");
  // Check arguments
  if (c.args.size() != 1) {
    _socket->send("8 Invalid number of arguments.\n");
  } else {
    LOG("[" << _subscriberID << "] subscribe. Directory: " << c.args[0] << "\n");        
    _mdServer->checkReplicatePermission(c.args[0]);
    
    // Optional arguments. inherit, repUsers, repPermissions
    bool inherit = true;
//    bool repUsers = false;
    bool repPermissions = false;
    for (map<string,string>::const_iterator it = c.optionalArgs.begin();
    	it != c.optionalArgs.end(); ++it) 
   	{
   	  bool flag = (it->second == "true");
   	  if (it->first == "inherit") {
   	  	inherit = flag;
//   	  } else if (it->first == "repUsers") {
//   	  	repUsers = flag;
   	  } else if (it->first == "repPermissions") {
   	  	repPermissions = flag;
   	  } else {
   	  	// Unknown option. Log and ignore.
   	  	LOG("[WARNING, ReplicationDaemon] Received unknown command: " << it->first << endl); 
   	  }
   	}
    
    _subMan->subscribe(_subscriberID, c.args[0], inherit, repPermissions);
    _socket->send("0\n");
  }
}

void ConnectionHandler::unsubscribe(const Command &c) throw (SocketException, SQLException, runtime_error) {
  DMESG("unsubscribe\n");
  // Check arguments
  if (c.args.size() != 1) {
    _socket->send("8 Invalid number of arguments.\n");
  } else {
    // The user can only unsubscribe its own directories, so
    // no need to check for permissions.
    _subMan->unsubscribe(_subscriberID, c.args[0]);
    _socket->send("0\n");   
  }
}

void ConnectionHandler::unsubscribeAll(const Command &c) throw (SocketException, SQLException, runtime_error) {
  DMESG("unsubscribe_all\n");
  // Check arguments
  if (c.args.size() != 0) {
    _socket->send("8 Invalid number of arguments.\n");
  } else {
    // The user can only unsubscribe its own directories, so
    // no need to check for permissions.
    _subMan->unsubscribeAll(_subscriberID);
    _socket->send("0\n");
  }
}


void ConnectionHandler::synchronize(const Command &c) 
throw (SocketException, SQLException, runtime_error) 
{
    // Check arguments
    if (c.args.size() != 1) {
      _socket->send("8 Invalid number of arguments.\n");
      return;
    }
//    cout << "[perf]: Receive:" << _socket->getTCPReceiveWindow() << endl;
//    cout << "[perf]: Send:" << _socket->getTCPSendWindow() << endl;
  
    string directory = c.args[0];
    LOG("[" << _subscriberID << "] synchronize. Directory: " << directory << "\n");
    _mdServer->checkReplicatePermission(directory);
   
    // Verify if the subscription exists
    Subscription sub;
    if (!_subMan->getSubscription(_subscriberID, directory, sub)) {
      _socket->send("1 Unknown subscription.\n");
      return;
    }
    // Needed for the xid
    auto_ptr<Subscriber> subscriber = _subMan->getSubscriber(_subscriberID);

    LOG("Subscription: " << sub << endl);
    MDSocketBuffer *rwBuffer = dynamic_cast<MDSocketBuffer *>(&_mdServer->getBuffer());
    assert(rwBuffer != NULL);

//    // Create the thread to execute the dump on the server backend.
//    DumpTask *task = new DumpTask(*_mdServer, directory, sub.repPermissions);
//    boost::thread *thread = new boost::thread(*task);
//    // This object makes sure that the thread is cleaned up properly, 
//    // even when the method exits due to an exception
//    DumpTaskWrapper dtWrapper(task, thread);
    try {    
      uint64_t xid = _mdServer->dump(directory, sub.repPermissions, true);
      DMESG("COMMIT\n");
      rwBuffer->append("COMMIT\n\n");
      rwBuffer->flush();
      string line = _socket->readLine();
      if (line == "0") {
        if (subscriber->lastAckedXid == 0) {  
          // Save the xid of this slave only if there is none
          _subMan->updateXID(_subscriberID, xid);
        }
      } else {
         DMESG("Error synchronizing with client: " << line << endl);
      }
    } catch (...) {
      DMESG("ABORT\n");
      rwBuffer->append("COMMIT\n\n");
      rwBuffer->flush();
      return;
    }
}

void ConnectionHandler::subscribeUsers(const Command &c) 
throw (SocketException, SQLException, runtime_error) 
{
  DMESG("subscribe_users\n");
  // Check arguments
  if (c.args.size() != 0) {
    _socket->send("8 Invalid number of arguments.\n");
  } else {
    LOG("[" << _subscriberID << "] subscribe users.\n");
    // TODO: How to know if this slave should be allowed to replicate users?        
//    _mdServer->checkReplicatePermission(c.args[0]);

	// Register in the database that this slave wants to replicate the users.
    _subMan->subscribeUsers(_subscriberID);
    
    string response;
    MDSocketBuffer *rwBuffer = dynamic_cast<MDSocketBuffer *>(&_mdServer->getBuffer());    
    assert(rwBuffer != NULL);    
    // Make sure the DB is not changed while we dump both users and groups.
    DatabaseConnection *dbConn = _mdServer->getDBConnection();
    Statement st(*dbConn);
    st.beginTransaction();
    
    // Send the current xid
    SubscriptionManager subMan(*dbConn);
    ostringstream oss;
    uint64_t curXid = subMan.getCurrentXid();
    oss << "#xid.users=" << curXid << "\n";

    // Dump users and groups synchronously. The result set should 
    // be small, so no need to do it asynchronously.
    _mdServer->setDebug(true);
    // The flag tells it not to dump the root user.
    _mdServer->userDump(true);
    // This will transform the next line into a comment. Necessary to force
    // the slave to ignore the result of executing the grpDump 
    rwBuffer->append("#");
    _mdServer->grpDump(true);
    st.commitTransaction();
    rwBuffer->append(oss.str());
    rwBuffer->flush();
    _socket->send("COMMIT\n\n");
    string line = _socket->readLine();
    if (line == "0") {
      auto_ptr<Subscriber> subscriber = _subMan->getSubscriber(_subscriberID);
      if (subscriber->lastAckedXid == 0) {  
        // Save the xid of this slave only if there is none
        _subMan->updateXID(_subscriberID, curXid);
      }
    } else {
    	DMESG("Error synchronizing with client: " << line << endl);
    }
  }
}

void ConnectionHandler::unsubscribeUsers(const Command &c) throw (SocketException, SQLException, runtime_error) {
  DMESG("unsubscribe_users\n");
  // Check arguments
  if (c.args.size() != 0) {
    _socket->send("8 Invalid number of arguments.\n");
  } else {
    _subMan->unsubscribeUsers(_subscriberID);
    _socket->send("0\n");
  }
}


void ConnectionHandler::receive(const Command &c) throw (SocketException, SQLException, runtime_error) {
  DMESG("Start receiving\n");
  if (c.args.size() != 1) {
    _socket->send("8 Invalid number of arguments.\n");

  } else {
    uint64_t xid;
    try {
      xid = toUint64(c.args[0]);
    } catch (std::runtime_error &ex) {
      // Convert to one of the exceptions that are declared by this method
      throw SQLException(string("8 Invalid xid: ") + c.args[0] + "\n");
    }
    LOG("[" << _subscriberID << "] Receiving logs. Xid: " << c.args[0] << "\n");
    // The OK will be sent by the registerSubscriber method, because 
    // only in the middle of this method will we know if the operation 
    // will succeed or not. 
    
    // Get all subscriptions for this subscriber from the database
    auto_ptr<Subscriber> subscriber = _subMan->getSubscriber(_subscriberID);
    if (subscriber.get() == NULL) {
      throw SQLException("No subscriptions");
    } 
  
    // Check if the xid required by the subscriber is not older than
    // the one we have for it.
    // TODO: Each host can have several subscriptions, but they should
    // all be at the same point in the logstream. I'll have to change
    // the database scheme to reflect this
    uint64_t subXid = subscriber->lastAckedXid;
    if ((xid+1) < subXid) {
      ostringstream oss;
      oss << "Subscriber requested a log that it had already acknowledged. Requested " 
          << xid << ", last acked: " << subXid;
      throw SQLException(oss.str());
    }
    // Send ok to the client, the operation succeeded
    _socket->send("0\n");
	if (xid != subXid) {
		// The slave is saying that it has all the logs up to xid.
		// The local information about the slave xid is incorrect. Update it.
		_subMan->updateXID(_subscriberID, xid);
	}
    _repDaemon._logMonitor->startStreamingLogs(*this, xid, subscriber);
  }
}
