/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef LISTENERTHREAD_H_
#define LISTENERTHREAD_H_

#include "util/ThreadUtil.h"
#include "util/Socket.h"
#include "util/Debuggable.h"
#include "util/SecurityUtils.h"
#include "server/UserContext.h"
#include "CommandParser.h"
#include "SubscriptionManager.h"
#include "LogMonitorThread.h"

#include <boost/shared_ptr.hpp>
#include <stdexcept>
#include <memory>
#include <map>

class ConfigParser;

/**
 * Accepts connections from slaves and spans a thread to handle them
 * using the ConnectionHandler class.
 */ 
class ReplicationDaemon : public Debuggable, Runnable {
private:
    int _listenPort;
    /** Maximum size of logs table before dropping subscribers */
    int _maxPendingLogs;
    // The following state is accessed by individual connection handlers
    ConfigParser &_configParser;
    std::auto_ptr<LogMonitorThread> _logMonitor;
    AuthContext _authCtx;
    friend class ConnectionHandler;

public:
	ReplicationDaemon(int listenPort, ConfigParser &configParser);
	virtual ~ReplicationDaemon();
  
    int getMaxPendingLogs() { return _maxPendingLogs; };
	  void run();
};

/**
 * Handles a single connection
 */
class ConnectionHandler : public Debuggable, public Runnable {
private:    
  // Static objects to keep track of all active connections to this
  // replication daemon instance.
  static std::set<std::string> _activeConnections;
  // Protects access to the above map
  static boost::mutex _mutex;

  void sendStatistics() throw (SocketException);
  // Command handlers
  void subscribe(const Command &c) throw (SocketException, SQLException, std::runtime_error);
  void unsubscribe(const Command &c) throw (SocketException, SQLException, std::runtime_error);
  void unsubscribeAll(const Command &c) throw (SocketException, SQLException, std::runtime_error);
  void synchronize(const Command &c) throw (SocketException, SQLException, std::runtime_error);
  void receive(const Command &c) throw (SocketException, SQLException, std::runtime_error);
  void unsubscribeUsers(const Command &c) throw (SocketException, SQLException, std::runtime_error);
  void subscribeUsers(const Command &c) throw (SocketException, SQLException, std::runtime_error);
  
public:
  ConnectionHandler(boost::shared_ptr<TCPSocket> sock, ReplicationDaemon &repDaemon);
  virtual ~ConnectionHandler();

  virtual void run();
//  void start();
  
  ReplicationDaemon &_repDaemon;
  // These are references to the objects owned by ReplicationDaemon
  AuthContext &_authCtx;

  // Local objects
  boost::shared_ptr<TCPSocket> _socket;
  std::string _subscriberID;
  std::auto_ptr<DatabaseConnection> _dbConn;
  std::auto_ptr<SubscriptionManager> _subMan;
  std::auto_ptr<MDStandalone> _mdServer;
  UserContext _userCtx;
};


#endif /*LISTENERTHREAD_H_*/
