/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "ReplicationDaemonConnection.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "debug.h"
#include "server/MDInterpreter.h"
#include "server/MDStandalone.h"
#include "server/MDBuffer.h"
#include "server/SiteManager.h"
#include "server/MDServerFactory.h"
#include "util/Utils.h"
#include "util/Path.h"
#include "util/openssl_mt.h"
#include "util/SecurityUtils.h"
#include "util/sslHelpers.h"
#include "util/DatabaseConnection.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "util/Clock.h"
#include "MountManager.h"
#include "ReplicationManager.h"

#include <openssl/ssl.h>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <iosfwd>
#include <algorithm>
#include <limits>

using namespace std;
using namespace boost;


int sslPasswdCallback(char *buf, int size, int rwflag, void *password)
{
//   int debug = 1;
//   DMESG("Returning password: >" << (char *)password << "<\n");
  strncpy(buf, (char *)(password), size);
  buf[size - 1] = '\0';
  return(strlen(buf));
}


SSL_CTX* ReplicationDaemonConnection::getSSLContext(const Site &site) 
throw (std::runtime_error) 
{
  DMESG("Initializing SSL context\n");
  DMESG("Site: " << site << endl);
  // Global initialization. Doesn't do anything if already initialized
  SSLManager::init();

  // Creates client instance
  // For SSL 1.0 Higher : SSL 1.0 change SSL_METHOD to const SSL_METHOD
#if OPENSSL_VERSION_NUMBER >= 0x10000000L
  const SSL_METHOD *method= SSLv23_client_method();
#else
  SSL_METHOD *method= SSLv23_client_method();
#endif
  
  // Create SSL context
  SSL_CTX *ctx = SSL_CTX_new(method);
  if(ctx == NULL) {
      throw runtime_error("Could not create SSL context");
  }

  SSL_CTX_set_default_passwd_cb(ctx, sslPasswdCallback);
  
  SSL_CTX_set_default_passwd_cb_userdata(ctx, (void *)certPassword);
  certPassword[0] = '\0';
  if( strlen(site.certPassword) ){
    DMESG("Using certificate with a password\n");
    strncpy(certPassword, site.certPassword, 256);
    certPassword[255] = '\0';
  }

  //  SSL_set_connect_state(ssl);
  // Disable buggy v2, enable all compatibility measures
  SSL_CTX_set_options(ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);

  if (site.authenticateWithCertificate) {
    // Grid proxies take precedence over normal user certs.
    string certFile, keyFile;  
    if(site.useGridProxy) {
      certFile = getProxyCertName();
      keyFile = certFile;
    }

    if (certFile.empty()){
      // No grid proxy files, so try to use normal proxies
      certFile = site.certFile;
      keyFile = site.keyFile;
    }

    // Load certificates    
    if(!certFile.empty()) {
      DMESG("Loading certificate: " << certFile << endl);
      // Load local certificate from file
      if(SSL_CTX_use_certificate_chain_file(ctx, certFile.c_str())!=1){
        throw runtime_error("Could not load client certificates");
      }
      DMESG("Loading private key file: " << keyFile << endl);
      // Load private key from local file
      if(SSL_CTX_use_PrivateKey_file(ctx, keyFile.c_str(), SSL_FILETYPE_PEM)!=1){
        throw runtime_error("Could not load private key");
      }
    }

    // Verify private key
    if (!SSL_CTX_check_private_key(ctx)) {
        throw runtime_error("Could not verify private key");
    }
  }
  // Verify server certificate
  if(site.verifyServerCert) {
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, serverCertVerifyCallback);
    DMESG("Loading local certificates" << std::endl);
    if(!SSL_CTX_load_verify_locations(ctx, NULL, site.trustedCertDir)){
        throw runtime_error("Error loading trusted certificates");
    }
  }
  return ctx;
}

ReplicationDaemonConnection::ReplicationDaemonConnection(const string &siteName, DatabaseConnection &dbConn, SSL_CTX *ctx)
throw (SocketException, ReplicationException, SQLException)
: _ctx(ctx), _masterID(siteName)
{    
  debug = ReplicationManager::getInstance()->getDebug();

  SiteManager sMan(dbConn);
  Site site;
  sMan.getSite(siteName, site);
  connect(site);
}

ReplicationDaemonConnection::ReplicationDaemonConnection(const Site &site, SSL_CTX *ctx)
throw (SocketException) 
: _ctx(ctx), _masterID(site.name)
{
  debug = ReplicationManager::getInstance()->getDebug();  
  connect(site);
}

void ReplicationDaemonConnection::connect(const Site &site)
    throw (SocketException)
{
    LOG("[Rep Client] Connecting to: " << site.hostname << ":" << site.port << " as " <<  site.login <<  endl);
    _socket.reset(new TCPSocket(site.hostname, site.port));

    // Do the greeting. Discard the first two lines.
    string line = _socket->readLine(); // ARDA metadata replication...
    if (!starts_with(line, "ARDA Metadata Replication Daemon")) {
        throw SocketException(string("Invalid server greeting: ") + line);
    }
    
    _socket->readLine(); // protokol ...
    line = _socket->readLine(); // offered protocols: plain ssl
    
    // The socket is connected. Now depending on the configuration initialize
    // SSL and authenticate.
    if (site.useSSL) {
        if (line.find("ssl") == string::npos) {
            // Server does not offer ssl
            throw SocketException("Server does not accept SSL connections");
        } else {
            _socket->send("ssl\n\n");
        }
    } else {
        // client wants to use plain text connections
        if (line.find("plain") == string::npos) {
            // Server does not offer plain text connections
            throw SocketException("Server does not accept plain text connections");
        } else {
            _socket->send("plain\n\n");
        }
    }
    line = _socket->readLine();
    if (line != "OK") {
        throw SocketException(string("Server refused connection method: ") + line);
    }
    
  if (site.useSSL) {
    // Create and initialize an SSL_CTX for this client. 
    // TODO: cache sessions
      try {
        if (_ctx == NULL) {
          _ctx = getSSLContext(site);
        }
        _socket->initSSLClient(_ctx, NULL);
      } catch (runtime_error &ex) {
        LOG("[Rep Client] [ERROR]" << ex.what() << sslErrors() << endl);
        throw SocketException(string("Error connecting to master node: ") + ex.what());
    }
  }

  DMESG("Establishing user context\n");
  line="0 ";
  line.append(site.login).append("\n1 ").append(site.name);
  if(strlen(site.password) != 0) {
    line.append("\n5 ").append(site.password);
  }
  line.append("\n\n");
  _socket->send(line);

  line = _socket->readLine();
  if (line != "0") {
    throw SocketException("Authentication failed: " + line);
  }
  // Now authenticate
  DMESG("Connected to: " << site.hostname << ":" << site.port << " as " <<  site.login << endl);
}

ReplicationDaemonConnection::~ReplicationDaemonConnection()
{
}

void ReplicationDaemonConnection::subscribe(const std::string &directory, 
	bool repPermissions, bool inherit)	 
    throw (SocketException, ReplicationException) 
{
    DMESG("Subscribing: " << directory << endl);
    ostringstream oss;
    oss << "subscribe " << directory 
    << "\ninherit=" << (inherit ? "true" : "false")
    << "\nrepPermissions=" << (repPermissions ? "true" : "false")
    << "\n\n";    
    _socket->send(oss.str());
    
    string line = _socket->readLine();        
    // Should be 0
    if (line != "0") {
        throw ReplicationException(string("Subscribe failed! Remote error: ") + line);
    }
}
 
void ReplicationDaemonConnection::unsubscribe(const std::string &directory) 
    throw (SocketException, ReplicationException)
{
    _socket->send("unsubscribe " + directory + "\n\n");
    string line = _socket->readLine();
    if (line != "0") {
        throw ReplicationException(string("Unsubscribe failed: ") + line);
    }
}

void ReplicationDaemonConnection::unsubscribeAll() 
    throw (SocketException, ReplicationException)
{
    _socket->send("unsubscribe_all\n\n");
    string line = _socket->readLine();
    if (line != "0") {
        throw ReplicationException(string("Unsubscribe failed: ") + line);
    }
}

void ReplicationDaemonConnection::synchronize(
    const std::string &remoteDir,
    MountManager &mManager, 
    MDStandalone &mdServer,
    MDBuffer *cancelBuffer)
    throw (SocketException, ReplicationException)
{ 
  mdServer.setDebug(debug);  
//  mdServer.setDebug(false);  
  LOG("[Rep Client] Synchronizing " << remoteDir << endl);
  
  // Used for error recovery. All the operations that change the
  // backend state are done inside this try block. If any error
  // occurs, an exception is thrown an intercepted by the catch 
  // block, which will clean up any partial state that may have
  // been left.
  _socket->send("synchronize " + remoteDir + "\n\n");    
  string line = _socket->readLine();
  if (line != "0") {
    // No side effects yet, no need to clean up.
    throw ReplicationException("Synchronize failed. Remote error: " + line);
  }
  executeDump(remoteDir, mManager, mdServer, cancelBuffer);
  LOG("[Rep Client] Synchronize done " << remoteDir << endl);
}

void ReplicationDaemonConnection::executeDump(
    const std::string &remoteDir,
    MountManager &mManager, 
    MDStandalone &mdServer,
    MDBuffer *cancelBuffer)
throw (SocketException, ReplicationException) 
{
//  cout << "[perf]: Receive:" << _socket->getTCPReceiveWindow() << endl;
//  cout << "[perf]: Send:" << _socket->getTCPSendWindow() << endl;  

  // This method is used both for dumping users/groups and metadata.
  // The former have no directory associated with them.
  bool dumpUsers = remoteDir == "";
  try {
    // Start synchronization. The roles between master and slave
    // are reversed temporarily, with the master sending commands
    // and client executing them. The client can at any time return
    // an error to the server, to stop the transfer.
    
    // The masterXid is the minimum of all xid's received
    numeric_limits<uint64_t> uint64_info;
    uint64_t masterXid = uint64_info.max();

    // For timing    
    Clock total;
    Clock sockRead;
    Clock intClk;
    Clock execClk;
    
    MDInterpreter interpreter;  
    string result;
    string line;
    total.start();
    do {
      sockRead.resume();
      line = _socket->readLine();
      sockRead.stop();
      DMESG("In> " << line.substr(0, 40) << "...\n");
      if (line == "COMMIT") {
        // Read the NULL line that the server always sends
        _socket->readLine();
        mManager.updateMasterXid(_masterID, masterXid);
        // Transmission ended successfully
        DMESG("Done synchronizing" << endl);
        _socket->send("0\n");
        
        total.stop();
        DMESG(total.print("[PERF] Total"));
        DMESG(sockRead.print("[PERF] Socket read"));
        DMESG(intClk.print("[PERF] MDInterpreter"));
        DMESG(execClk.print("[PERF] exec"));
        return;
      }
      if (line == "ABORT") {
        // Read the NULL line that the server always sends
        _socket->readLine();
        _socket->send("0\n");
        throw ReplicationException("Synchronization aborted by master: " +line.substr(5));
      }
      if (starts_with(line, "#xid.")) {
        // it's a comment
        int index = line.find_first_of('=');
        string directory = line.substr(5, index-5);
        trim(directory);
        uint64_t xid = toUint64(line.substr(index+1));
        try {
          if (directory == "masterindex") {
          	masterXid = min(masterXid, xid);
          } else if (directory == "users") {          	
            mManager.updateUsersXid(xid);
            masterXid = min(masterXid, xid);
          } else {
          	// It's a directory
            if (directory == ".") {
                directory = remoteDir;
            } else  {   
                // absolutePath requires that the second directory
                // always ends with "/"
              directory = absolutePath(directory, remoteDir + "/");
            }
            mManager.updateMountXid(directory, xid);
            masterXid = min(masterXid, xid);
          }
        } catch (SQLException &ex) {
          _socket->send(string("1 Internal error: ") + ex.what() + "\n");
          while (_socket->readLine() != "") {}
          throw ReplicationException(string("Internal error updating database: ") + ex.what());
        }
        // It's a comment, so don't try to execute it.
        continue;
      }
      intClk.resume();

      AMGA::decodeLine(line);
      interpreter.parseCommand(line.c_str());
      intClk.stop();
      execClk.resume();
      if(interpreter.execLast(mdServer)) {
          // Invalid command
          _socket->send("1 Invalid command: " + line + "\n");
          // SIAHN_20110111, remove the below line 
          //   in order to cut connection immideately 
          // while (_socket->readLine() != "") {}
          throw ReplicationException(string("Local executiong failed. Invalid command:") + line);
      }
      execClk.stop();
      // Check the data on the buffer to see if the command failed.
      mdServer.getBuffer().read(result);
      // Check if the command executed correctly and interrupt the transfer 
      // if it didn't. During an upload command no error code is returned
      // until the final commit, so "" means no error. Otherwise, a return
      // of 0 represents a single shot command that executed correctly.
      if (result != "" && result[0] != '0') {
        // Command failed to execute
        LOG("[Rep Client] Command failed. Error: " << result << endl);
        _socket->send(result);
        // SIAHN_20110111, remove the below line
        //   in order to cut connection immideately 
        // while ((line = _socket->readLine()) != "");
        throw ReplicationException("Local execution failed. " + result.substr(0, result.size()-1));
      }
      
//      cout << "(" << (cancelBuffer != NULL) << ", " << (cancelBuffer->isCancelled()) << ")";
      if (cancelBuffer != NULL && cancelBuffer->isCancelled()) {
        throw ReplicationException("User cancelled");
      } 
    } while (true);
    
  } catch (...) {
    // Something went wrong while replicating. Any open transaction is
    // automatically rolled back, but since each directory is copied on 
    // its own transaction, we need to clean the partial changes 
    // from previous directories.
    if (!dumpUsers) {
	    LOG("[Rep Client] Cleaning up from failed synchronization\n");
	    try {
	      mdServer.deleteDirectoryContents(remoteDir, true);
	    } catch (std::exception &ex) {
	      LOG("[Rep Client] [ERROR] Failed to delete directory: " << ex.what() << endl);
	    }
    }
    // Retrow the exception to inform the caller of the error
  	throw;
  }
}

std::string ReplicationDaemonConnection::getHost() {
  return _socket->getForeignAddress();
}

int ReplicationDaemonConnection::getPort() {
  return _socket->getForeignPort();
}
    
    
auto_ptr<TCPSocket> ReplicationDaemonConnection::receive(int xid)
    throw (SocketException, ReplicationException)
{
    ostringstream os;
    os << "receive " << xid << "\n\n";
    _socket->send(os.str());
    string line = _socket->readLine();
    if (line != "0") {
        throw ReplicationException("Failed to initiate subscription: " + line);
    }
    // Hand over the socket. This instance of RepDaemonClient can no longer be used,
    // since it is no longer possible to send commands to the repdaemon
    return _socket;
}

// Client side replication of users/groups.
void ReplicationDaemonConnection::subscribeUsers(
	MountManager &mManager, MDStandalone &mdServer)
    throw (SocketException, ReplicationException)
{
  DMESG("Subscribing users and groups\n");
  string command = "subscribe_users\n\n";
  _socket->send(command);

  string line = _socket->readLine();
  // Should be 0
  if (line != "0") {
    throw ReplicationException("Error subscribing users: " + line);
  }
  executeDump("", mManager, mdServer);
}

void ReplicationDaemonConnection::unsubscribeUsers()	 
    throw (SocketException, ReplicationException) 
{
	DMESG("Unsubscribing users and groups\n");
    string command = "unsubscribe_users\n\n";
    _socket->send(command);

    string line = _socket->readLine();        
    // Should be 0
    if (line != "0") {
        throw ReplicationException("Remote error: " + line);
    }
}
