/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef REPDAEMONCLIENT_H_
#define REPDAEMONCLIENT_H_

#include "util/Socket.h"
#include "util/DBUtil.h"
#include "util/Debuggable.h"
#include "Exceptions.h"

#include <memory>
#include <stdexcept>

class MDStandalone;
class MountManager;
class DatabaseConnection;
class MDBuffer;
struct Site;

/**
 * Represents a connection to a replication daemon.
 */
class ReplicationDaemonConnection : protected Debuggable
{
public:
  ReplicationDaemonConnection(const Site &site, SSL_CTX *ctx = NULL)
    throw (SocketException);
    
  ReplicationDaemonConnection(const std::string &siteName, 
    DatabaseConnection &dbConn, 
    SSL_CTX *ctx = NULL)
    throw (SocketException, ReplicationException, SQLException);

  virtual ~ReplicationDaemonConnection();

  void subscribe(const std::string &remoteDirectory, 
  	  bool repPermissions, bool inherit)
      throw (SocketException, ReplicationException);

  void unsubscribe(const std::string &remoteDirectory) 
      throw (SocketException, ReplicationException);
  
  void subscribeUsers(MountManager &mManager, MDStandalone &mdServer)
    throw (SocketException, ReplicationException);
        
  void unsubscribeUsers()
    throw (SocketException, ReplicationException);

  void unsubscribeAll() throw (SocketException, ReplicationException);

	/**
	 * Copies the data from the remote host. If it fails, it cleans
	 * everything that might have been copied into the local database.
	 */
  void synchronize(const std::string &remoteDirectory, 
      MountManager &mManager, 
      MDStandalone &mdServer,
      MDBuffer *cancelBuffer = NULL)
      throw (SocketException, ReplicationException);

  std::auto_ptr<TCPSocket> receive(int xid)
      throw (SocketException, ReplicationException);

  std::string getHost();    
  int getPort();
  
  SSL_CTX *getSSLCtx() {
    return _ctx;
  };

  
private:
  void connect(const Site &site)
      throw (SocketException); 
      
  SSL_CTX* getSSLContext(const Site &site) 
    throw (std::runtime_error);
  
  void executeDump(const std::string &remoteDir,
    MountManager &mManager, 
    MDStandalone &mdServer,
    MDBuffer *cancelBuffer = NULL) throw (SocketException, ReplicationException);

  // Ownership of this connection will be retained until the client
  // calls resume. After that, the socket is given to the caller
  // so it can use it for receiving logs. The auto_ptr is used to
  // hand over the socket safely
  std::auto_ptr<TCPSocket> _socket;
  // Does not own this object
  SSL_CTX *_ctx;
  
  std::string _masterID;
  char certPassword[256];
};

#endif /*REPDAEMONCLIENT_H_*/
