/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "ReplicationManager.h"

#include "debug.h"

#include "replicationdaemon/MDLogWriter.h"
#include "replicationdaemon/LogReceiver.h"
#include "replicationdaemon/MountManager.h"
#include "util/Utils.h"
#include "util/DatabaseConnection.h"
#include "server/MDServerFactory.h"
#include "server/MDStandalone.h"

#include <memory>
#include <sstream>

using namespace std;

ReplicationManager* ReplicationManager::instance = NULL;

/*****************************************
 * ReplicationManager
 *****************************************/
ReplicationManager::ReplicationManager(const ConfigParser &configParser, bool debugInit, bool asRepDaemon)
    throw (std::runtime_error, SQLException)
: Debuggable(debugInit), enableSlave(false), 
  enableMaster(false), enableRepDaemon(asRepDaemon), maxPendingLogs(0)
{
  if (enableRepDaemon) {
    initAsRepDaemon(configParser);
  } else {
    StatsCollector::registerStatsSource(*this);
    try {
      enableSlave = configParser.getInt("Replication::EnableSlave");
    } catch (std::string &err) { }
    try {
      enableMaster = configParser.getInt("Replication::EnableMaster");
    } catch (std::string &err) { }
    
    DMESG("Slave: " << enableSlave << ", Master: " << enableMaster << endl);
    if (enableMaster) {
      initAsMaster(configParser);
    }
    if (enableSlave) {
      initAsSlave(configParser);
    }
  }
}

ReplicationManager::~ReplicationManager()
{
  StatsCollector::unregisterStatsSource(*this);
  DMESG("~ReplicationManager\n");
}

void ReplicationManager::initAsSlave(const ConfigParser &configParser) {
  LOG("Enabling replication as a slave\n");
}

void ReplicationManager::initAsMaster(const ConfigParser &configParser) {
  LOG("Enabling replication as a master\n");
  MDLogWriter::setDebug(debug);
  // Initialize master-side replication
  MDLogWriter::setSaveLogs(true);
}

void ReplicationManager::initAsRepDaemon(const ConfigParser &configParser) {
  LOG("Initializing as replication daemon\n");    
  try {
    maxPendingLogs = configParser.getInt("Replication::MaxPendingLogs");
  } catch (std::string &err) { 
  }
}

void ReplicationManager::initManager(const ConfigParser &configParser, 
      bool debug, bool asRepDaemon) 
throw (std::runtime_error, SQLException) 
{
  if (instance == NULL) {
    instance = new ReplicationManager(configParser, debug, asRepDaemon);
    
    // These objects access the replication manager so they must
    // be initialized after the global instance variable is created.
    if (instance->isSlave()) {
    	// KIMNG090202
    	// Initialize slave-side replication
      LogReceiverManager::init(configParser);
      
      // Startup all connections that are marked as being on the active state
      auto_ptr<DatabaseConnection> dbConn(DatabaseConnection::openConnection());
      MountManager mManager(*dbConn);
      vector<Master> masters;
      mManager.getActiveMasters(masters);
            
      for (size_t i = 0; i<masters.size(); i++) {
        try {
          LogReceiverManager::getInstance()->startReceiveLogs(masters[i], *dbConn);
        } catch (ReplicationException &ex) {
          // This error could be caused by a local system problem
          // or due to connection problems. Either way, ignore it.
          LOG("Failed to connect to master: " << masters[i].id
            << ". Reason: " << ex.what() << endl);
        }
      }
    }
    
    // If the slave process dies while synchronizing, the mount is left
    // on a partially synchronized state. We must remove it. Or resume it,
    // if that were possible.
    try {
    	std::auto_ptr<DatabaseConnection> dbConn(DatabaseConnection::openConnection());
    	MDBuffer buffer;
    	std::auto_ptr<MDStandalone> mdServer(dynamic_cast<MDStandalone*>(
    		MDServerFactory::mdServer(buffer, dbConn.get())));
    	vector<Mount> mounts;
    	MountManager mManager(*dbConn);
    	mManager.getMountsInState(Mount::Synchronizing, mounts);
		for (vector<Mount>::const_iterator it = mounts.begin();
			it != mounts.end();
			++it)
		{
			DMESG("Cleaning inconsistent mount: " << it->directory << endl);	
			try {
				mdServer->umountImpl(*it, mManager);
			} catch (SQLException &ex) {
				DMESG("Failed to clean mount " << it->directory << ". Error: " << ex.what() << endl);
			}
		}
    } catch (exception &ex) {
    	DMESG("Failed to clean failed mounts. Database may be inconsistent. Error: " << ex.what() << endl);
    }
  } else {
    throw runtime_error("Already initialized");
  }
}

void ReplicationManager::destroy() {
    delete instance;
    instance = NULL;
}

void ReplicationManager::collectCounters(std::map<std::string, std::string> &counters) {
}
