/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "SubscriptionManager.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "debug.h"
#include "util/Path.h"
#include "util/Statement.h"
#include "util/DatabaseConnection.h"
#include "server/Constants.h"
#include "ReplicationManager.h"

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <iosfwd>

using namespace std;
using namespace boost;

SubscriptionManager::SubscriptionManager(DatabaseConnection &dbConn)
: _dbConn(dbConn)
{
  debug = ReplicationManager::getInstance()->getDebug();
  setupTableNames();
}

SubscriptionManager::~SubscriptionManager() {
}


void SubscriptionManager::setupTableNames()
{
  std::string prefix;
  if(DatabaseConnection::dbSchema.size())
    prefix = DatabaseConnection::dbSchema + ".";

  masterindexTable = prefix + "masterindex";
  subscribersTable = prefix + "subscribers";
  subscriptionsTable = prefix + "subscriptions";
  logsTable = prefix + "logs";
  sequencesTable = prefix + "sequences";
}


vector<Subscription> SubscriptionManager::getSubscriptions(const string& subscriber) throw (SQLException) {
  return _getSubscriptions("\"sub_subscriber\" = '" + subscriber + "'");
}

map<string, Subscriber> SubscriptionManager::getAllSubscribers()
throw(SQLException) 
{
	return _getSubscribers("");
}

std::set<std::string> SubscriptionManager::getSubscribedDirs()
  throw(SQLException) 
{
  string query = "SELECT DISTINCT \"sub_directory\" FROM " 
    + subscriptionsTable + ";";
  Statement statement(_dbConn, false);
  DMESG("Query: " << query << endl);
  if(statement.exec(query)){
    throw SQLException(statement);
  }

  // The data types used by the struct Subscription are not compatible
  // with the ones that we need to use with ODBC. So we need to convert
  // then later.
  char directory[DIRNAME_MAX_SIZE];
  statement.bind(1, directory, sizeof(directory));
  
  set<string> subscribedDirs;
  while (!statement.fetch()) {
    // Read the script from the result of the query.
    DMESG("Fetched: " << directory << endl);
  	subscribedDirs.insert(directory);
  }
  statement.unBind();
  statement.close();
  return subscribedDirs;
}  

vector<Subscription> SubscriptionManager::_getSubscriptions(const string& whereClause) 
throw (SQLException) 
{
  string query = "SELECT \"sub_subscriber\", \"sub_directory\", "
    "\"sub_rep_permissions\" FROM " + subscriptionsTable;
  if (!whereClause.empty()) {
      query += " WHERE " + whereClause;
  }
  query += ";";
  DMESG("SQL: " << query << endl);
  Statement statement(_dbConn);
  if (statement.exec(query)) {
    // This method returns non-zero either when no rows are updated
    // or when the connection fails. We check the error to distinguish
    // between the two cases.
    if (statement.getErrorCode() != SQL_NO_DATA) {
      throw SQLException(statement);
    }
  }

  // The data types used by the struct Subscription are not compatible
  // with the ones that we need to use with ODBC. So we need to convert
  // then later.
  char subscriberID[NODE_ID_MAX_SIZE];
  statement.bind(1, subscriberID, sizeof(subscriberID));
  char directory[DIRNAME_MAX_SIZE];
  statement.bind(2, directory, sizeof(directory));
  short int repPermissions;
  statement.bind(3, &repPermissions);
//  short int repUsers;
//  statement.bind(4, &repUsers);
//    statement.bind(5, sub.leaselog.cwd, sizeof(log.cwd));
  Subscription sub;
  
  vector<Subscription> subs;
  while (!statement.fetch()) {
    // Convert the data types to the ones used on the struct Subscription
//    statement.getData(6, sub.subscriber.lastAckedXid);
    sub.subscriberID = subscriberID;
//    sub.inherit = inherit;
    sub.repPermissions = repPermissions;
    sub.directory = directory;
    // Read the script from the result of the query.
    DMESG("Fetched: " << sub << endl);
    // Hopefully, it makes a copy of the log.
    subs.push_back(sub);
  }
  statement.unBind();
  statement.close();
  return subs;
}

map<string, Subscriber> SubscriptionManager::_getSubscribers(const string& whereClause) 
throw (SQLException) 
{
  string query = "SELECT \"sub_id\", \"sub_rep_users\", \"sub_lease_duration\", "
  	"\"sub_cur_xid\" FROM " + subscribersTable;
  if (!whereClause.empty()) {
      query += " WHERE " + whereClause;
  }
  query += ";";
  DMESG("SQL: " << query << endl);
  Statement statement(_dbConn);
  if (statement.exec(query)) {
    // This method returns non-zero either when no rows are updated
    // or when the connection fails. We check the error to distinguish
    // between the two cases.
    if (statement.getErrorCode() != SQL_NO_DATA) {
      throw SQLException(statement);
    }
  }

  // The data types used by the struct Subscription are not compatible
  // with the ones that we need to use with ODBC. So we need to convert
  // then later.
  char subscriberID[NODE_ID_MAX_SIZE];
  statement.bind(1, subscriberID, sizeof(subscriberID));
  short int repUsers;
  statement.bind(2, &repUsers);
//    statement.bind(3, sub.leaselog.cwd, sizeof(log.cwd));
  Subscriber sub;
  
  map<string, Subscriber> subscribers;
  while (!statement.fetch()) {
    // Convert the data types to the ones used on the struct Subscription
    sub.subscriberID = subscriberID;
    sub.repUsers = repUsers;        
    statement.getData(4, sub.lastAckedXid);    
    // Read the script from the result of the query.
    DMESG("Fetched: " << sub << endl);
    // Hopefully, it makes a copy of the log.
    subscribers[sub.subscriberID] = sub;
  }
  statement.unBind();
  statement.close();
  return subscribers;
}

std::auto_ptr<Subscriber> SubscriptionManager::getSubscriber(const std::string& subId) 
  throw(SQLException) 
{
	map<string, Subscriber> subs = _getSubscribers("\"sub_id\" = '" + subId + "'");
	if (subs.size() == 0) {
		return auto_ptr<Subscriber>(NULL);
	} else {
		return auto_ptr<Subscriber>(new Subscriber(subs[subId]));
	}
}

/**
 * Get all subscribers of this directory, including the
 * ones that are subscribed to a parent directory with the
 * inherit flag set to true.
 * 
 * @param directory
 * @return
 * @throw(SQLException)
 */
vector<Subscription> SubscriptionManager::getDirectorySubscribers(const string& directory) 
throw(SQLException) {
  string query = "\"sub_directory\"='" + directory + "'";    
  vector<std::string> tokens;
  split(tokens, directory, is_any_of("/"), token_compress_on);    
  for (size_t i = 0; i < tokens.size(); ++i) {
    query += " OR (\"sub_directory\"='/" + tokens[i] + "' AND \"sub_inherit\"=1)";
  }
  return _getSubscriptions(query);
}

/**
 * Returns all subscriptions.
 * @return
 * @throw(SQLException) 
 */
vector<Subscription> SubscriptionManager::getSubscriptions() throw(SQLException) {
    return _getSubscriptions("");
}

void SubscriptionManager::subscribe(
    const string& subscriber, const string& directory, 
    bool inherit, bool repPermissions) throw(SQLException)
{
    string dir = absolutePath(directory, "/");    
    DMESG("Subscribing " << subscriber << ", " << dir << ", " << inherit << endl);
    // Does the directory exist? 
    string query = "SELECT 1 FROM " + masterindexTable 
      + " WHERE \"directory\"='" + dir + "';";

    Statement statement(_dbConn);
    statement.beginTransaction();
    if(statement.exec(query)){
        throw SQLException("Could not open transaction on DB", statement);
    }    
    int exists;
    statement.bind(1, &exists);
    if(statement.fetch()){
        // Does not exist
        throw SQLException("Directory " + dir + " does not exist", statement);
    }
    // Prepare the statement for being reused.
    statement.unBind();
    statement.close();
    // Directory exists. Create the subscription
 
    // If there is any previous subscription of the same directory from 
    // the same user , overwrite it.
    try {
        unsubscribe(subscriber, dir);
        LOG("Discarded previous subscription\n");
    } catch (ReplicationException &ex) {
        // It's ok, the subscription did not exist.
    }

	// If there is no entry for this subscriber, create a new one      
    if (getSubscriber(subscriber).get() == NULL) {
    	// The subscriber must be inserted 
    	query = "INSERT INTO " + subscribersTable + " ("
        	"\"sub_id\", \"sub_rep_users\", \"sub_cur_xid\") "
        	"VALUES ('" + subscriber +  "', 0, 0);";
    	DMESG("SQL: >" << query << "<\n");
    	if(statement.exec(query)){
        	throw SQLException(statement);
    	}
    }
    // Now insert the subscription    
    ostringstream oss;    
    oss << "INSERT INTO " << subscriptionsTable << " ("\
        "\"sub_subscriber\", \"sub_directory\", \"sub_rep_permissions\") "
        "VALUES ('"
        << subscriber <<  "', '" << dir << "', " << (repPermissions ? "1" : "0") << ");";
    query = oss.str();
    DMESG("SQL: >" << query << "<\n");
    if(statement.exec(query)){
        throw SQLException(statement);
    }
    statement.commitTransaction();
}

/**
 * Unsubscribes from a directory. An user can only unsubscribe from the
 * directories it owns.
 */
void SubscriptionManager::unsubscribe(
    const string& subscriber, 
    const string& directory) throw(SQLException, ReplicationException)
{
    string dir = absolutePath(directory, "/");
    DMESG("Unsubscribing " << subscriber << ", " << dir << "\n");

    // First remove from database
    string query = "DELETE FROM " + subscriptionsTable 
      + " WHERE \"sub_subscriber\" = '"
      + subscriber + "' AND \"sub_directory\" = '" + dir + "';";
    DMESG("SQL: " << query << endl);
    Statement statement(_dbConn);
    if(statement.exec(query)){
        if (statement.getErrorCode() == SQL_NO_DATA) {
            throw ReplicationException("No subscription found");
        } else {
            throw SQLException("Error deleting previous subscription", statement);
        }
    }
    if (statement.numRows() <= 0) {
        throw ReplicationException("No subscription found");
    }
    
    // Clean up the subscribers table if there are no more subscriptions
    // for this subscriber and if it is not replicating users.
    query = "DELETE FROM " + subscribersTable 
      +" WHERE (\"sub_id\" = '" + subscriber
      + "') AND (\"sub_rep_users\" = 0) AND (0 = (SELECT COUNT(*) FROM "
      + subscriptionsTable + " WHERE \"sub_subscriber\" = '"
      + subscriber + "'));";

    DMESG("SQL: " << query << endl);
    if(statement.exec(query)){
    	if (statement.getErrorCode() != SQL_NO_DATA) {
        	throw SQLException("Error deleting previous subscription", statement);
      	}
    }
}

void SubscriptionManager::dropSubscriber(const std::string& subscriber) 
    throw(SQLException, ReplicationException) 
{
  string query = "DELETE FROM " + subscribersTable 
    + " WHERE (\"sub_id\" = '" + subscriber + "');";
  DMESG("SQL: " << query << endl);
  Statement st(_dbConn);
  if(st.exec(query)){
    if (st.getErrorCode() == SQL_NO_DATA) {
      throw ReplicationException("Subscriber not found: " + subscriber);
    } else {
      throw SQLException("Error deleting previous subscription", st);
    }
  }
}    
    
void SubscriptionManager::unsubscribeAll(
    const string& subscriber) throw(SQLException)
{
    DMESG("Unsubscribing " << subscriber << "\n");
    throw logic_error("Not implemented exception");
//    // First remove from database
//    string query = "DELETE FROM subscribers WHERE \"sub_id\" = '" + subscriber + "';";
//    DMESG("SQL: " << query << endl);
//    Statement statement(_dbConn);
//    if(statement.exec(query)){
//        throw SQLException("Error unsubscribing: ", statement);
//    }
//    if (statement.numRows() <= 0) {
//        throw SQLException("Could not unsubscribe user: ", statement);
//    }
}

void SubscriptionManager::subscribeUsers(const string& subscriber) 
	throw(SQLException)
{
  DMESG("Subscribing users " << subscriber << endl);
	string query;
	// If there is no entry for this subscriber, create a new one      
  if (getSubscriber(subscriber).get() == NULL) {
   	// The subscriber must be inserted 
   	query = "INSERT INTO " + subscribersTable + " ("
       	"\"sub_id\", \"sub_rep_users\", \"sub_cur_xid\") "
      	"VALUES ('" + subscriber +  "', 1, 0);";
  } else {
  	// An entry already exists. Update it.
  	query = "UPDATE subscribers SET \"sub_rep_users\" = 1 WHERE \"sub_id\" = '" 
  		+ subscriber + "';";  
  }    
  DMESG("SQL: >" << query << "<\n");
  Statement statement(_dbConn);
	if(statement.exec(query)){
    	throw SQLException(statement);
	}
}

void SubscriptionManager::unsubscribeUsers(const string& subscriber) 
	throw(SQLException)
{
    DMESG("Unsubscribing users" << subscriber << endl);
    // If this subscriber has other subscriptions, just update its line 
    // on the subscribers table. Otherwise, delete it.
    string query;
    // TODO: a hasSubscriptions(subscriber) method could speed things up
    vector<Subscription> subs = getSubscriptions(subscriber);
    if (subs.empty()) {
      query = "DELETE FROM " + subscribersTable 
	+" WHERE \"sub_id\" = '" + subscriber + "';";
    } else {
      // There are still subscriptions
      query = "UPDATE " + subscribersTable 
	+ " SET \"sub_rep_users\" = 0 WHERE \"sub_id\" = '" + subscriber + "';";
    }
    Statement statement(_dbConn);
    if(statement.exec(query)){
      throw SQLException("Error deleting previous subscription", statement);
    }
}

void SubscriptionManager::updateXID(const std::string& subscriber, uint64_t xid) 
throw (SQLException) 
{
  DMESG("Updating xid. Subscriber: " << subscriber << ", xid: " << xid << endl);
  // First remove from database
  ostringstream oss;
  oss << "UPDATE subscribers SET \"sub_cur_xid\"=" 
    << xid << " WHERE \"sub_id\"='" << subscriber << "';";    
  
  Statement statement(_dbConn);
  string query = oss.str();
  DMESG("SQL: >" << query << endl);
  if(statement.exec(query) || statement.numRows() <= 0) {
    throw SQLException("Error updating subscriber information", statement);
  }
}

/**
 * Drops all subscriptions from the database.
 * @throw(SQLException)
 */
void SubscriptionManager::clearAllSubscriptions() throw(SQLException) {
    Statement statement(_dbConn, debug);
    if(statement.exec("DELETE FROM " + subscribersTable + ";")){
      throw SQLException("Could not delete subscriptions", statement);
    }
}


bool SubscriptionManager::getSubscription(
    const string& user, const string& directory,
    Subscription &sub) throw(SQLException)
{	
  vector<Subscription> subs = _getSubscriptions("\"sub_subscriber\" = '"  + user + "' AND \"sub_directory\" = '" + directory + "'");
  if (subs.empty()) {
    DMESG("Subscription not found: " << user << ", dir: " << directory << endl);
  	return false;
  } else {
  	sub = subs[0];  	
    return true;
  }
}

/**
 * Retrieves the xid of the oldest log still available on the database.
 * Returns 0 if no logs are available.
 */
uint64_t SubscriptionManager::getOldestXid() throw (SQLException) {
  Statement statement(_dbConn);
  if(statement.exec("SELECT min(\"log_xid\") FROM " + logsTable + ";")){
    throw SQLException("Error executing statement: ", statement);
  }

  uint64_t xid;
  if (statement.fetch()){
    statement.getData(1, xid);
    throw SQLException(statement);
  }
  return (statement.getColumnLen(1) == SQL_NULL_DATA)? 0 : xid;
}

///**
// * Retrieves the xid of the most recent log. Returns 0 if no 
// * logs are available.
// */
//uint64_t SubscriptionManager::getLatestXid() throw (SQLException) {
//  Statement statement(_dbConn);
//  if(statement.exec("SELECT max(\"log_xid\") FROM logs;")){
//    throw SQLException("Error executing statement: ", statement);
//  }
//
//  uint64_t xid;
//  if (statement.fetch()){
//    statement.getData(1, xid);
//    throw SQLException(statement);
//  }
//  return (statement.getColumnLen(1) == SQL_NULL_DATA)? 0 : xid;
//}

/**
 * Return the value that will be used for the next xid. This queries
 * the sequence object on the database
 */
uint64_t SubscriptionManager::getCurrentXid() throw (SQLException) 
{
  Statement statement(_dbConn);
  std::string query;
  int isCalled = 1;
  switch (_dbConn.getDriverType()) {
  	case DatabaseConnection::dtOracle: {
	  // CURRVAL is not defined in a session before the first call to
	  // the sequence, so it cannot be used. Instead, we use NEXTVAL which
	  // is always defined, but that advances the sequence even though
	  // this method shouldn't do that. I believe there is no problem
	  // in advancing the sequence, apart from using the sequence faster.  
	  query = "SELECT logs_seq.NEXTVAL FROM dual;";
	  DMESG("SQL: " << query << endl);
	  if(statement.exec(query)){
	    throw SQLException("Error executing statement: ", statement);
	  }
	  break;
  	}
  	
  	case DatabaseConnection::dtPSQL: {
	  // This code works only with PostgreSQL
	  query = "SELECT \"last_value\", \"is_called\" from logs_log_xid_seq;";
	  DMESG("SQL: " << query << endl);
	  if(statement.exec(query)){
	    throw SQLException("Error executing statement: ", statement);
	  }
	  statement.bind(2, &isCalled);
	  break;
  	}
  	
  	case DatabaseConnection::dtMySQL5: {
	  query = "SELECT \"last_log_xid\" FROM " + sequencesTable + ";";
	  if(statement.exec(query)){
	    throw SQLException("Error executing statement: ", statement);
	  }
	  break;
  	}
  	
  	default: 
	  throw SQLException("Function not supported by backend");
  }

  if (statement.fetch()){
    throw SQLException(statement);
  }
  uint64_t xid;
  statement.getData(1, xid);
  DMESG("xid: " << xid << ", isCalled: " << isCalled << endl);
  
  // On PostgreSQL when a sequence is created the last_value field
  // is pointing to the first value that will be returned, so the first
  // use of the sequence will not increment this value. After that, 
  // the last_value field actually contains the last value to have 
  // been returned. To handle the special case of a sequence that wasn't
  // accessed yet, we check the is_called field that indicates if the
  // sequence was called at least once.
  if (!isCalled) {
    xid--;
  }
  return xid;
}


//ostream& operator<<(ostream& out, const Subscription &sub) {
//  out << sub.subscriberID << ", " << sub.directory << 
//  	", repPermissions: " << sub.repPermissions << sub.inherit << ", " << sub.curXid << ", " << sub.expiry;
//  return out;
//}

ostream& operator<<(ostream& out, const Subscription &sub) {
  out << sub.subscriberID << ", " << sub.directory << 
  	", Permissions: " << sub.repPermissions;
  return out;
}

ostream& operator<<(ostream& out, const Subscriber &subscriber) {
  out << subscriber.subscriberID << ", repUsers:" << subscriber.repUsers << 
  	", xid: " << subscriber.lastAckedXid;
  return out;
}

//bool Subscription::operator==(const Subscription &other) const {
//  return subscriberID == other.subscriberID && directory == other.directory;
//}
//
//bool Subscription::operator<(const Subscription &other) const {
//  // Order first based on the subscriberID and then on the directory
//  return subscriberID == other.subscriberID ? 
//    directory < other.directory :
//    subscriberID < other.subscriberID;
//}
