/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef SUBSCRIPTIONMANAGER_H_
#define SUBSCRIPTIONMANAGER_H_

#include <vector>
#include <string>
#include <map>
#include <set>
#include <memory>

class DatabaseConnection;

#include "MountManager.h"
#include "Exceptions.h"
#include "util/Debuggable.h"

struct Subscriber {
  std::string subscriberID;
  bool repUsers;
  uint64_t lastAckedXid;
    
//  std::vector<std::string> directories;
//  uint64_t lastSentXid;

  Subscriber() : lastAckedXid(0) {};
  ~Subscriber() {};
  
  friend std::ostream& operator<<(std::ostream& out, const Subscriber &sub);
};

struct Subscription {
  std::string subscriberID;
  std::string directory;
  bool repPermissions;
  
//  std::string subscriberID;
//  bool inherit;
//  bool repUsers;
//  uint64_t curXid;
  // TODO: must use a date/time
//  std::string expiry;

//  Subscription() : inherit(true), repPermissions(false), curXid(0)
  Subscription() : repPermissions(false) {};
  ~Subscription() {};
  
//  bool operator==(const Subscription &other) const;
//  bool operator<(const Subscription &other) const;  

  friend std::ostream& operator<<(std::ostream& out, const Subscription &sub);
};

class SubscriptionManager : public Debuggable
{
public:
//	SubscriptionManager(DatabaseConnection *dbConn = NULL);
  SubscriptionManager(DatabaseConnection &dbConn);
	virtual ~SubscriptionManager();

  /**
   * Get all subscribers of this directory, including the
   * ones that are subscribed to a parent directory with the
   * inherit flag set to true.
   *
   * @param directory
   * @return
   * @throw(SQLException) 
   */
  std::vector<Subscription> getDirectorySubscribers(const std::string& subscriber) 
  throw (SQLException);
      
  std::auto_ptr<Subscriber> getSubscriber(const std::string& subId) 
  throw(SQLException);
  
  std::vector<Subscription> getSubscriptions(const std::string& subscriberID) 
  throw (SQLException);
  
  std::map<std::string, Subscriber> getAllSubscribers()
  throw(SQLException);

  std::vector<Subscription> getSubscriptions() 
  throw(SQLException);
  
  std::set<std::string> getSubscribedDirs()
  throw(SQLException);

  void subscribe(
      const std::string& subscriber, 
      const std::string& directory, 
      bool inherit, 
      bool repPermissions) throw(SQLException);

  void unsubscribe(const std::string& subscriber, const std::string& directory) 
  throw(SQLException, ReplicationException);
  
  void dropSubscriber(const std::string& subscriber) 
    throw(SQLException, ReplicationException);
  /** Drop all subscriptions from a subscriber. */
  void unsubscribeAll(const std::string& subscriber) 
  throw(SQLException);
  
  void subscribeUsers(const std::string& subscriber) 
	throw(SQLException);
  void unsubscribeUsers(const std::string& subscriber) 
	throw(SQLException);
	
  /**
   * Drops all subscriptions from the database.
   * @throw(SQLException)
   */
  void clearAllSubscriptions() throw(SQLException);
  
  /**
   * @return true if a subscription was found, in which case the
   * sub parameter is filled out. Otherwise returns false and
   * does not change the sub parameter.
   */
    bool getSubscription(
        const std::string& user, const std::string& directory,
        Subscription &sub) throw(SQLException);
                
    void updateXID(const std::string& subscriber, uint64_t xid) throw (SQLException);
    
//    uint64_t getLatestXid() throw (SQLException);
    uint64_t getOldestXid() throw (SQLException);
    uint64_t getCurrentXid() throw (SQLException);
    
private:
    std::vector<Subscription> _getSubscriptions(const std::string& whereClause) throw (SQLException);
    std::map<std::string, Subscriber> _getSubscribers(const std::string& whereClause) throw (SQLException);

    DatabaseConnection &_dbConn;

    
  void setupTableNames();
  std::string masterindexTable;
  std::string subscribersTable;
  std::string subscriptionsTable;
  std::string logsTable;
  std::string sequencesTable;
};

#endif /*SUBSCRIPTIONMANAGER_H_*/
