/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: DBUserManager.cc,v 1.24 2011/01/26 05:57:13 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "DBUserManager.h"
#include "MDServerFactory.h"
#include "util/Statement.h"
#include "util/DatabaseConnection.h"
#include "util/sslHelpers.h"
#include "Constants.h"

#include <fstream>
#include <iostream>
#include <sstream>



static bool debug=false;
#include "debug.h"

#define MAX_SUBJECT_LENGTH 257
#define MAX_PASS_LENGTH 65


DBUserManager::DBUserManager()
  : ownsConnection(false), dbConn(NULL), initialized(false)
{
  DMESG("Initializing DBUserManager\n");
  userManagerName = "DBUserManager";
  setupTableNames();
}


DBUserManager::DBUserManager(DatabaseConnection *dbConnection)
  : ownsConnection(false), dbConn(dbConnection), initialized(false)
{
  DMESG("Initializing DBUserManager\n");
  userManagerName = "DBUserManager";
}


void DBUserManager::setupTableNames()
{
  std::string prefix;
  if(DatabaseConnection::dbSchema.size())
    prefix = DatabaseConnection::dbSchema + ".";

  usersTable = prefix + "users";
  certificatesTable = prefix + "certificates";
}


DBUserManager::~DBUserManager()
{
  if (dbConn != NULL)
    disconnect();
}


void DBUserManager::setConnection(DatabaseConnection *dbConnection)
{
  dbConn = dbConnection;
}


int DBUserManager::connectDataSource(const std::string &dataSource,
				     const std::string &dbUser,
				     const std::string &dbPass)
{  
    try {
        dbConn = new DatabaseConnection(dataSource, dbUser, dbPass);
        ownsConnection = true;
        return 0;
    } catch (SQLException &ex) {
        LOG(ex.what());
    }
    return -1;
}


int DBUserManager::disconnect()
{
    if (ownsConnection) {
        delete dbConn;
        ownsConnection = false;
    }
    return 0;
}


/** Checks whether a certain subject can be used as authorization for
    a given user (role). Note that several subjects can be used as a
    credential for a given role.
*/
bool DBUserManager::checkSubject(const std::string &user,
				 const std::string &subject) const
{
  DMESG("checkSubject: got >" << subject << "<\n");

  if(!dbConn)
    return -1;

  Statement statement(*dbConn);
  std::string tmp_user = user ;
  std::string tmp_subject = subject ;
//   statement.quoteDataField(tmp_user);
//   statement.quoteDataField(tmp_subject);

  std::string query="SELECT \"subject\", \"name\" FROM " + certificatesTable
    + " WHERE \"name\" = '" + tmp_user + "' AND \"subject\" = '" + tmp_subject + "';";

  DMESG("SQL: >" << query << "<" << std::endl);
  if(statement.exec(query)){
    ERRLOG("Unable to read certificates table!\n");
    return false;
  }

  char usr[USERNAME_MAX_SIZE+1];
  char sub[MAX_SUBJECT_LENGTH+1];
  
  statement.bind(1, sub, USERNAME_MAX_SIZE);
  statement.bind(2, usr, MAX_SUBJECT_LENGTH);
  
  if( !statement.fetch() ) {
    DMESG("checkSubject: allowing role " << user << std::endl);
    return true;
  }

  return false;
}


/** Checks whether the password of the user (role) is
    correct.
*/
bool DBUserManager::checkPass(const std::string &user,
			      const std::string &pass) const
{
  DMESG("Checking password\n");

  if(!dbConn)
    return -1;

  std::string digest = getDigest(pass);

  std::string query="SELECT \"name\", \"password\" FROM " + usersTable +
    " WHERE \"name\" = '";
  query.append(user).append("' AND \"password\" ='");
  query.append(digest);
  query.append("';");

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    ERRLOG("Unable to read users table!\n");
    return false;
  }

  char usr[USERNAME_MAX_SIZE+1];
  char password[MAX_PASS_LENGTH+1];

  statement.bind(1, usr, USERNAME_MAX_SIZE);
  statement.bind(2, password, MAX_PASS_LENGTH);
  
  if(!statement.fetch()) {
    return true;
  };
  
  return false;
}


/** Checks whether the subject is known to the system
 */
bool DBUserManager::isKnownSubject(const std::string &subject) const
{
  if(!dbConn)
    return -1;

  Statement statement(*dbConn);
  std::string tmp_subject = subject ;
//   statement.quoteDataField(tmp_subject);

  std::string query="SELECT \"subject\", \"name\" FROM " + certificatesTable
    + " WHERE \"subject\" = '" + tmp_subject +"';";

  DMESG("SQL: >" << query << "<" << std::endl);
  if(statement.exec(query)){
    return -2;
  }

  char sub[MAX_SUBJECT_LENGTH+1];
  char usr[USERNAME_MAX_SIZE+1];
  
  statement.bind(1, sub, MAX_SUBJECT_LENGTH);
  statement.bind(2, usr, USERNAME_MAX_SIZE);
  
  if( !statement.fetch() ) {
    DMESG("isKnownSubject: subject exists " << usr << std::endl);
    return true;
  }

  return false;
}


/** Checks whether the user is known to the system
 */
bool DBUserManager::isKnownUser(const std::string &user) const
{
  if(!dbConn)
    return -1;

  std::string query="SELECT \"name\", \"password\" FROM " + usersTable +
    " WHERE \"name\" = '";
  query.append(user).append("';");

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    return -2;
  }

  char usr[USERNAME_MAX_SIZE];
  char password[MAX_PASS_LENGTH];

  statement.bind(1, usr, sizeof(usr));
  statement.bind(2, password, sizeof(password));
  
  if(!statement.fetch()) {
    return true;
  };
  
  return false;
}


/** Adds a user to the database
    Returns 101 if user exists
    Returns -1 if connection to DB failed
*/
int DBUserManager::addUser(const std::string &user,
			   const std::string &pass, bool passIsHash)
{
  std::string digest;
  // if it is an hash, don't hash it again. This is for user replication
  if(pass.size()) {
  	if (passIsHash) {
  	  digest = pass;
  	} else { 
      digest = getDigest(pass);
  	} 
  }

  if(dbConn) {
    std::string query="INSERT INTO " + usersTable +
      " (\"name\", \"password\") VALUES ('";
    query.append(user).append("', ");
    if(digest.size())
      query.append("'").append(digest).append("');");
    else 
      query.append("NULL);");
    
    DMESG("SQL: >" << query << "<" << std::endl);
    Statement statement(*dbConn);
    if(statement.exec(query)){
      return 101;
    }
  }

  return 0;
}


/** Adds a new user-subject pair
    Returns 100 if user does not exist
    Returns -1 if connection to DB failed
*/
int DBUserManager::addUserSubject(const std::string &user,
				  const std::string &subject)
{
  if(dbConn) {

    Statement statement(*dbConn);
    std::string tmp_user = user ; 
    std::string tmp_subject = subject ; 
    // statement.quoteDataField(tmp_user);
    // statement.quoteDataField(tmp_subject);
    
    std::string query="INSERT INTO " + certificatesTable
      + " (\"subject\", \"name\") VALUES ('" + tmp_subject + "', '"
      + tmp_user + "');";
    
    DMESG("SQL: >" << query << "<" << std::endl);
    if(statement.exec(query)){
      return 100;
    }
  }

  return 0;
}


/** Removes a user-subject pair
    Returns 100 if user/certificate does not exist
    Returns -1 if connection to DB failed
*/
int DBUserManager::removeUserSubject(const std::string &user,
				     const std::string &subject)
{
  if(dbConn) {

    Statement statement(*dbConn);
    std::string tmp_user = user ;
    std::string tmp_subject = subject ;
//     statement.quoteDataField(tmp_user);
//     statement.quoteDataField(tmp_subject);

    std::string query="DELETE FROM " + certificatesTable
      + " WHERE \"subject\" = '" + tmp_subject + "' AND \"name\" = '"
      + tmp_user + "';";
    
    DMESG("SQL: >" << query << "<" << std::endl);
    if(statement.exec(query)){
      return 100;
    }
  }

  return 0;
}


/** Changes User Password
    Returns 100 if user does not exist
    Returns -1 if connection to DB failed
*/
int DBUserManager::changeUserPass(const std::string &user,
				  const std::string &pass)
{
  std::string digest = getDigest(pass);

  DMESG("pass:" << pass << std::endl);
  DMESG("digest: " << digest << std::endl);

  if(dbConn) {
    std::string query="UPDATE users SET \"password\" = '";
    query.append(digest).append("' WHERE \"name\" = '");
    query.append(user).append("';");
    
    DMESG("SQL: >" << query << "<" << std::endl);
    Statement statement(*dbConn);
    if(statement.exec(query)){
      return 100;
    }
  }

  return 0;
}


/** Removes a user
    Returns 100 if user does not exist
    Returns -1 if connection to DB failed
*/
int DBUserManager::removeUser(const std::string &user)
{
  if(dbConn) {
    std::string query="DELETE FROM " + usersTable + " WHERE \"name\" = '"
      + user + "';";
    
    DMESG("SQL: >" << query << "<" << std::endl);
    Statement statement(*dbConn);
    if(statement.exec(query)){
      return 100;
    }
  }

  return 0;
}


/** Returns a list of all known users
    Returns 0 on success and 9 on internal DB error.
 */
int DBUserManager::listUsers(std::list < std::string > &users)
{
  if(!dbConn)
    return 21;

  std::string query="SELECT \"name\" FROM " + usersTable + ";";
  
  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    return 9;
  }
  
  char usr[USERNAME_MAX_SIZE+1];
  
  statement.bind(1, usr, USERNAME_MAX_SIZE);
  
  while (!statement.fetch()) {
    users.push_back(usr);
  }
  
  return 0;
}


/** List users
    Lists all credentials that a user has. If user is an empty string,
    then all users are listed
    Returns
    100 : User does not exist
    21 : DB functionality not available
    9: Trouble with DB
*/
int DBUserManager::listUserCred(const std::string &user,
				MDUserManager::UserCred &cred)
{
  if(!dbConn)
    return 21;
  
  std::string query="SELECT " + usersTable + ".\"name\", " 
    + usersTable + ".\"password\", " 
    + certificatesTable + ".\"subject\" FROM " 
    + usersTable + " LEFT OUTER JOIN " + certificatesTable +
    " ON (" + usersTable + ".\"name\" = " + certificatesTable + ".\"name\")";
  if(user.size())
    query += " WHERE " + usersTable + ".\"name\" = '" + user + "';";
  else
    query += " ORDER BY " + usersTable + ".\"name\";";
  
  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    return 9;
  }

  char usr[USERNAME_MAX_SIZE+1];
  char password[MAX_PASS_LENGTH+1];
  char subject[MAX_SUBJECT_LENGTH+1];
  
  statement.bind(1, usr, USERNAME_MAX_SIZE);
  statement.bind(2, password, MAX_PASS_LENGTH);
  statement.bind(3, subject, MAX_SUBJECT_LENGTH);

  int res = 100;

  while (!statement.fetch()) {
    DMESG("Pushing back: >" << subject << "<\n");
    if(user == usr) {
      res = 0;
      if( strlen(subject) )
	cred.dbDNs.push_back(subject);
      if(strlen(password)) cred.pass = password;
    }
  }

  return res;
}
