/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: GridMapUserManager.cc,v 1.11 2011/01/26 05:57:13 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "GridMapUserManager.h"
#include "util/sslHelpers.h"

#include <fstream>
#include <iostream>
#include <sstream>

static bool debug=false;
#include "debug.h"

#define MAX_SUBJECT_LENGTH 256
#define MAX_PASS_LENGTH 64


GridMapUserManager::GridMapUserManager()
{
  DMESG("Initializing GridMapUserManager\n");
  userManagerName="GridMapUserManager";
}


GridMapUserManager::~GridMapUserManager()
{
  // We don't hold any resource...
}


int GridMapUserManager::initFromGridMap(const std::string &file)
{
  DMESG("INIT FORM GRID MAP" << std::endl);

  std::ifstream mapFile(file.c_str());

  if(!mapFile.is_open())
    return -1;

  char buff[513];
  int lineNum = 0;
  while(!mapFile.eof()) {
    lineNum++;
    mapFile.getline(buff, 512);
    std::string line(buff);
    std::string subject;
    std::string user;
    if(parseLine(line, subject, user)){
      LOG("Error parsing grid map file, line " << lineNum << std::endl);
      LOG("  : " << line << std::endl);
      return -2;
    }
    DMESG("Parsed: subject >" << subject 
	  << "<  user >" << user << "<" << std::endl);
    if(subject.size()) {
      DMESG("Putting entry into mapping" << std::endl);
      certificates[user]=subject;
    }
  }
  initialized = true;
  
  return 0;
}
 

/** Checks whether a certain subject can be used as authorization for
    a given user (role).
    Note that several subjects can be used as a credential for a given
    role.
*/
bool GridMapUserManager::checkSubject(const std::string &user,
				      const std::string &subject) const
{
  DMESG("checkSubject: got >" << subject << "<\n");

  std::map< std::string, std::string >::const_iterator I 
    = certificates.begin();
  if((I=certificates.find(subject)) != certificates.end()) {
    DMESG("Matching to role >" <<  user<< "<\n");
    if(I->second == user)
      return true;
  }
  DMESG("checkSubject: not allowing role " << user << std::endl);
  
  return false;
}


/** Checks whether the password of the user (role) is
    correct.
*/
bool GridMapUserManager::checkPass(const std::string &user,
				   const std::string &pass) const
{
  return false;
}


int GridMapUserManager::parseLine(const std::string &line,
				  std::string &subject, 
				  std::string &role)
{
  subject="";

  if(!line.size())
    return 0;

  // Comment line
  if(line[0] == '#')
    return 0;

  size_t subjectStart=line.find("\"");
  if(subjectStart==std::string::npos){
    // No subject on line
    if(line.find_first_not_of(" \t")) // But line not empty
      return -1;
    return 0; // Empty line: OK
  }

  if(subjectStart >= line.size()-1)
    return -1;
  
  size_t subjectEnd=line.find("\"", subjectStart+1);
  if(subjectEnd == std::string::npos)
    return -1;

  subject=line.substr(subjectStart+1, subjectEnd-subjectStart-1);
  role="";

  if(subjectEnd == line.size())
    return 0;

  size_t roleStart=line.find_first_not_of(" \t", subjectEnd+1);
  if(roleStart == std::string::npos)
    return 0;

  size_t roleEnd=line.find_first_of(" \t", roleStart+1);
  if(roleEnd == std::string::npos)
    roleEnd=line.size();

  role=line.substr(roleStart, roleEnd-roleStart);

  return 0;
}


/** Checks whether the subject is known to the system
 */
bool GridMapUserManager::isKnownSubject(const std::string &subject) const
{
  std::map< std::string, std::string >::const_iterator I;
  if((I=certificates.find(subject)) != certificates.end())
    return true;

  return false;
}



/** Lists all users known via GridMapFile
    Returns 0
*/
int GridMapUserManager::listUsers(std::list< std::string > &users)
{
  std::map< std::string, std::string >::const_iterator I;
  I = certificates.begin();
  while( I != certificates.end() )
    users.push_back(I++->second);
    
  return 0;
}


/** Lists user credentials
    Lists all credentials that a user has. If user is an empty string,
    then all users are listed
    Returns
    100 : User does not exist
*/
int GridMapUserManager::listUserCred(const std::string &user,
				     MDUserManager::UserCred &cred)
{
  int res = 100;

  std::map< std::string, std::string >::const_iterator I;
  I = certificates.begin();
  
  while( I != certificates.end() ) {
    if( user == I->first ) {
      cred.fixedDNs.push_back(I->second);
      res = 0;
    }
    I++;
  }

  return res;
}
