/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDAsyncBuffer.h"

#include <stdexcept>
#include <iostream>
#include <assert.h>
#include <boost/thread/xtime.hpp>

static bool debug = false;
#include "debug.h"


using namespace std;
using namespace boost;

// #include <iostream>

MDAsyncBuffer::MDAsyncBuffer()
  : maxSize(DEFAULT_MAX_SIZE), readThreshold(DEFAULT_READ_THRESHOLD), flushRequested(false)
{
}

MDAsyncBuffer::~MDAsyncBuffer()
{
}

bool MDAsyncBuffer::isAsynchronous() const {
  return true;
}

void MDAsyncBuffer::cancel() {
  DMESG("Cancel" << endl);
  // The mutex is released at the end of this method, by the
  // Guard object destructor.
  mutex::scoped_lock lk(monitor);
  DMESG("Cancel locked" << endl);
  if (cancelled) {
    DMESG("[Warning]: Buffer already cancelled\n");
  }

  cancelled = true;
  producerFinished = true;
  buffer = "";

  // Must always send a signal since the variable
  // that was just updated is checked in the condition
  // loop of read and append.
  buffer_not_empty.notify_one();
  buffer_not_full.notify_one();    
}

bool MDAsyncBuffer::isCanceled() {
  bool debug = true;
  return cancelled;
}

bool MDAsyncBuffer::isProducerFinished() {
  return producerFinished;
}

bool MDAsyncBuffer::isEmpty() {
  return buffer.empty();
}

void MDAsyncBuffer::start()
{
  DMESG("Start" << endl);
  mutex::scoped_lock lk(monitor);
  DMESG("[start] Acquired lock" << endl);
  if (!buffer.empty()) {
    LOG("Buffer is not empty!" << endl);
  }
  assert(buffer.empty());
  // The following operations taken as a whole are not atomic.
  // So, they must be protected.
  
  //  DMESG("[start] locked " << endl);
  producerFinished=false;
  cancelled=false;
  //  DMESG("[start] Unlocked " << endl);
}

void MDAsyncBuffer::finish()
{
	DMESG("[finish]" << endl);
	mutex::scoped_lock lk(monitor);
	DMESG("[finish] Acquired lock" << endl);
	if (producerFinished || cancelled) {
		DMESG("[finish] WARNING - producerFinished == " << producerFinished
			<< ", Cancelled: " << cancelled << endl);
	}

	//  DMESG("[finish] Locked " << endl);
	producerFinished = true;
	cancelled = false;
    buffer_not_empty.notify_one();
    buffer_not_full.notify_one();	
//  DMESG("[finish] Unlocked " << endl);
}

void MDAsyncBuffer::flush()
{
	DMESG("[flush]" << endl);
	mutex::scoped_lock lk(monitor);
	DMESG("[finish] Acquired lock" << endl);
	if (producerFinished || cancelled) {
		DMESG("[finish] WARNING - producerFinished == " << producerFinished
			<< ", Cancelled: " << cancelled << endl);
	}

	//  DMESG("[finish] Locked " << endl);
	flushRequested = true;
    buffer_not_empty.notify_one();
}

// WARNING: This method will kill any thread that calls it
// if the cancel flag is set to true.
void MDAsyncBuffer::append(const std::string &text) throw(BufferCanceledException)
{
  DMESG("[append]" << endl);
  //   DMESG("[append]: >" << text << "<" << endl);
  mutex::scoped_lock lk(monitor);
  DMESG("[append] Locked" <<endl);
  //  assert(!done && !cancelled);
  // TODO: This is not clean. The buffer should never exceed
  // the max size.
  while (buffer.size() > (size_t)maxSize && !producerFinished) {
    DMESG("[append]: waiting" << endl);
    // Wait for the reader to free some space on the buffer
    buffer_not_full.wait(lk);
    DMESG("[append]: awake" << endl);
  }
  if (producerFinished || cancelled) {
    DMESG("[append] No more writing allowed. " <<
    (cancelled ? "Buffer cancelled." : "Producer finished.") <<
    " Discarding data." << endl);
    throw BufferCanceledException("No further writing allowed. Data discarded");
  } else {
   // Ok, we have space to put more stuff on the buffer
   buffer.append(text);

   // Do we have enough data for a read? If the buffer is full
   // we should also force the client to read, regardless of
   // the readThreshold, or else the server will block the
   // next time append() is called without signaling the client.
   DMESG("[append]: buffer size: " << buffer.size() << endl);
   if (buffer.size() >= readThreshold ||
       buffer.size() > maxSize) {
     // Yes. Inform the reader. He might or might not be blocked
     // on the condition variable, but we must signal it anyway.
     buffer_not_empty.notify_one();
   }
  }
  //  DMESG("[append] Unlocked" <<endl);
}


void MDAsyncBuffer::read(std::string &buf, int timeout) throw(BufferCanceledException) {
  DMESG("[Read]" << endl);
  mutex::scoped_lock lk(monitor);
  DMESG("[Read] locked" << endl);
//  assert(!done && !cancelled);
  // Do we have data available to return to the reader?
//  DMESG("Done: " << done << endl);
    
  while (buffer.size() < readThreshold && !producerFinished && !flushRequested) {
    DMESG("[Read]: waiting" << endl);
    // Wait for the reader to free some space on the buffer
    if (timeout < 0) {
      buffer_not_empty.wait(lk);
    } else {
      boost::xtime xt;
      boost::xtime_get(&xt, boost::TIME_UTC);
      xt.nsec += 1000*1000*timeout;
      if (!buffer_not_empty.timed_wait(lk, xt)) {
        // Timeout was reached
        DMESG("Read timeout\n");
        return;
      }
    }
    DMESG("[Read]: awake" << endl);
  }
  if (cancelled) {
  	throw BufferCanceledException("Query cancelled. Time limit exceeded.");
  }
  flushRequested = false;
  buf="";
  buffer.swap(buf);
  // Wake up the write, if it is waiting for available space.
  buffer_not_full.notify_one();
  // Ok, we are done.
//   DMESG("[read] " << buf << endl);
}

void MDAsyncBuffer::setBufferSize(unsigned int  s) {
	mutex::scoped_lock lk(monitor);
	maxSize=s;
}

unsigned int MDAsyncBuffer::getBufferSize() {
  return maxSize;
}

/** How much data should be on the buffer before reading
 * is allowed. The class will make an effort to return to
 * the reader blocks of at least this given size.
 */
void MDAsyncBuffer::setReadThreshold(unsigned int readThreshold) {
  this->readThreshold = readThreshold;
}

unsigned int MDAsyncBuffer::getReadThreshold() {
  return readThreshold;
}
