/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef MDASYNBUFFER_H
#define MDASYNBUFFER_H

#include <exception>
#include <boost/thread/thread.hpp>
#include <boost/thread/condition.hpp>

#define DEFAULT_MAX_SIZE 16*1024
#define DEFAULT_READ_THRESHOLD 4*1024

#include "MDBuffer.h"


class BufferCanceledException : public std::exception {
	std::string _msg;
public:
	BufferCanceledException(const std::string &msg)
	: _msg(msg){}

	~BufferCanceledException() throw() {} ;
	virtual const char* what() const throw() {
		return _msg.c_str();
	}
};


class MDAsyncBuffer : public MDBuffer{
public:

  MDAsyncBuffer();
  virtual ~MDAsyncBuffer();

  virtual void append(const std::string &text) throw(BufferCanceledException);


  virtual bool isCanceled();

  /**
	Puts the buffer in a state where it can be written and read.
	This method will fail if the buffer is already started.
   */
  virtual void start();
	/**
	* Terminates the current operation. This method should be used
	* by the producer to inform the client that there is no more data
	* available and that the operation terminated normally. In this
	* state read is allowed but writes will fail.
	*/
  void finish();
  /**
   * Terminates current operation, clears buffer and sets the
   * cancel flag to true. This method should be used when the
   * operation is terminated abruptily.
   */
  virtual void cancel();

  bool isProducerFinished();


  /**
   * Blocks until data is available. If the buffer is canceled
   * while readers are waiting, an exception is thrown. An exception
   * is also thrown if this method is called while the buffer is canceled.
   *
   * @param buf
   * @param timeout how long to block, in milliseconds.
   */
  virtual void read(std::string &buf, int timeout = -1) throw(BufferCanceledException);
  virtual bool isAsynchronous() const;

  /** Buffer max size */
  void setBufferSize(unsigned int s);
  unsigned int getBufferSize();

  /** How much data should be on the buffer before reading
   * is allowed. The class will make an effort to return to
   * the reader blocks of at least this given size.
   */
  void setReadThreshold(unsigned int s);
  unsigned int getReadThreshold();

  bool isEmpty();
  
  virtual void flush();

private:
  unsigned int maxSize;
  unsigned int readThreshold;
  
  bool flushRequested;

  boost::condition buffer_not_full, buffer_not_empty;
  boost::mutex monitor;
};

#endif // MDAsyncBuffer_H
