/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDConMan.h"
#include "util/Utils.h"
#include "replicationdaemon/ReplicationManager.h"

#include <ctime>
#include <sys/mman.h>
#include <signal.h>
#include <fcntl.h>

#include <iostream>
#include <list>

#include <errno.h>
#include <assert.h>
#include <stdio.h>

#include "debug.h"

static bool debug = false;


// This should be executed at startup.
static time_t startupTime = time(NULL);
static MDConMan *masterConMan;

int MDConMan::interval;
int MDConMan::idleTime;
int MDConMan::size;
int MDConMan::entries;
int MDConMan::urandom;
std::string MDConMan::lockName;
std::string MDConMan::mapName;
std::string MDConMan::masterSemName;
std::string MDConMan::USlockName;

unsigned int MDConMan::maxSessions=0;
unsigned int MDConMan::maxSessionSize=0;

static void timer_handler(int sig)
{
  DMESG("Timer called" << std::endl);

  masterConMan->reap(masterConMan->getMaxIdleTime());
  DMESG("Interval2: " << masterConMan->getInterval() << std::endl);
  alarm(masterConMan->getInterval());
}


MDConMan::MDConMan()
  : master(false), useSessionCache(false), sem(NULL), mem(NULL)
{
  // Nothing to do...
}


MDConMan::MDConMan(int nprocs)
  : master(nprocs), useSessionCache(false), sem(NULL), mem(NULL)
{
  DMESG("Con Man 2 " << master << std::endl);
  init(nprocs);
}


MDConMan::~MDConMan()
{
  DMESG("Master: " << master << std::endl);
  // Semaphore must be there, to call remove
  if(sem)
    closeConMan();
  if(master)
    close(urandom);

  DMESG("DONE " << master << std::endl);
}


int MDConMan::closeConMan()
{
  DMESG("closeConMan: " << master << std::endl);
  if(sem_close(sem)){
    PERRLOG("Could not close semaphore " << lockName);
  }

  if(sem_close(masterSem)){
    PERRLOG("Could not close semaphore " << masterSemName);
  }

  if(sem_close(USsem)){
    PERRLOG("Could not close semaphore " << USlockName);
  }


  if(master){
    if(sem_unlink(lockName.c_str())){
      PERRLOG("Could not unlink semaphore " << lockName);
    }

    if(sem_unlink(masterSemName.c_str())){
      PERRLOG("Could not unlink semaphore " << masterSemName);
    }

    if(sem_unlink(USlockName.c_str())){
      PERRLOG("Could not unlink semaphore US" );
    }

  }

  if(mem){
    if(munmap(mem, size)){
      PERRLOG("Could not munmap"<< mapName);
    }
  }

  if(close(fd)){
    PERRLOG("Could not close shared memory " << mapName);
  }

  if(master){
    if(shm_unlink(mapName.c_str())){
      PERRLOG("Could not unlink"<< mapName);
    }
  }
  return 0;
}

bool MDConMan::init(int nprocs)
{
  if(nprocs) {  // Master process
    master=true;
    entries=nprocs;
    char buff[10];
    sprintf(buff, "%d", getpid());
    lockName="MDConManLock";
    lockName.append(buff);

    mapName="MDConManShm";
    mapName.append(buff);

    masterSemName="MDMasterSem";
    masterSemName.append(buff);

    USlockName="USLock";
    USlockName.append(buff);


    DMESG("Creating " << lockName << std::endl);
    if((sem=sem_open(lockName.c_str(), O_CREAT, S_IRWXU, 0)) == SEM_FAILED){
      PERRLOG("Could not create " << lockName);
      return false;
    }


    DMESG("Creating US lock "  << USlockName << std::endl);
    if((USsem=sem_open(USlockName.c_str(), O_CREAT, S_IRWXU, 1)) == SEM_FAILED){
      PERRLOG("Could not create " << USlockName);
      return false;
    }
    int sval = -1;
    sem_getvalue(USsem, &sval);
    DMESG("Creating US lock "  << sval << std::endl);


    DMESG("Creating " << masterSemName << std::endl);
    if((masterSem=sem_open(masterSemName.c_str(), O_CREAT, S_IRWXU, 1))
       ==SEM_FAILED){
      PERRLOG("Could not create " << masterSemName);
      return false;
    }

    DMESG("Creating " << mapName << std::endl);
    if((fd=shm_open(mapName.c_str(), O_CREAT | O_RDWR, S_IRWXU))==-1){
      PERRLOG("Could not create "<< mapName);
      return false;
    }

    DMESG("Resizing shm area " << mapName << std::endl);
    size=entries*(sizeof(struct Process));

    if(ftruncate(fd, size) ==-1){
      PERRLOG("Could not reserve space: " << size << std::endl);
      return false;
    }
    DMESG("Opening /dev/urandom" << std::endl);
    urandom = open("/dev/urandom", O_RDONLY);
    assert(urandom >= 0);

  } else {   // Slave processes
    DMESG("Slave opening " << lockName << std::endl);
    if( (sem=sem_open(lockName.c_str(), O_RDWR)) == SEM_FAILED){
      PERRLOG("Could not open " << lockName);
      return false;
    }


    DMESG("Slave opening US lock "   << USlockName  << std::endl);
    if( (USsem=sem_open(USlockName.c_str(), O_RDWR)) == SEM_FAILED){
      PERRLOG("Could not open " << USlockName);
      return false;
    }

    int sval = -1;
    sem_getvalue(USsem, &sval);
    DMESG("Slave US lock "  << sval << std::endl);
    DMESG("Slave US "  << USsem << std::endl);


    DMESG("Slave opening " << masterSemName << std::endl);
    if( (masterSem=sem_open(masterSemName.c_str(), O_RDWR)) == SEM_FAILED){
      PERRLOG("Could not open " << masterSemName);
      return false;
    }

    DMESG("Slave opening " << mapName << std::endl);
    if((fd=shm_open(mapName.c_str(), O_RDWR, S_IRWXU))==-1 ){
      PERRLOG("Could not open " << mapName);
      return false;
    }

  }
  mem=(struct Process *)mmap(NULL, size,
			     PROT_READ|PROT_WRITE,
			     MAP_SHARED, fd, 0);
  if(mem==MAP_FAILED) {
    PERRLOG("Could not map memory");
    return false;
  }


  // Initialize process table
  if(master) {
    for(int i=0; i<nprocs; i++) {
    	mem[i].user[0]=mem[i].pid=mem[i].status=0;
    	// KIMNG080416 add a thread_num & a pThreadData to initialize values
    	mem[i].thread_num=0;
    	mem[i].pThreadData = NULL;
    }
  }


  // Initialize session cache, if enabled by calling
  // initSessionCache() manually, first
  if(useSessionCache)
    return !initSessionCache();


  return true;
}


//! Transforms a MDConMan object into a slave
/** This is called after a fork so that the child can transform its
    copy of MDConMan into a slave.
*/
int MDConMan::setSlave(int slot)
{
  master=false;

  closeConMan();
  if(useSessionCache)
    closeSessionCache();

  init();

  return takeSlot(slot);
}


int MDConMan::takeSlot(int slot)
{
  DMESG("takeSlot " << slot << std::endl);
  if(lock())
    return -1;

  if(mem[slot].status != 1 || mem[slot].pid != -1){
    unLock();
    return -1;
  }

  DMESG("takeSlot: taking " << slot << std::endl);
  mem[slot].timestamp=time(NULL);
  mem[slot].status=1;
  mem[slot].pid=getpid();
  mem[slot].user[0]='\0';

  return unLock();
}


//! Updates the timestamp of a process
/** Updates the timestamp of a process to mark that it is not idle.
    \param pid The process id of the process to mark as not idle, or 0
    for the current process.
    \param slot Slot number of process. If left to default 0, then
    correct slot is searched.
    \param status New status of the process in \param slot
*/
void MDConMan::timestamp(int status, const std::string &user,
			 pid_t pid, int slot)
{
  DMESG("Timestamping: pid " << pid << "(is " << ((int)getpid()) << ")"
	<< "  status " << status << std::endl);

  if(lock())
    return;

  if(!pid) pid=getpid();

  int i;
  for(i=slot; i<entries; i++){ // Updates existing entry
    if(mem[i].pid == pid) {
      mem[i].timestamp=time(NULL);
      mem[i].status=status;
      strncpy(mem[i].user, user.c_str(), USERNAME_MAX_SIZE);
      break;
    }
  }

  if(i==entries){          // Creates new entry
    for(i=0; i<entries; i++){
      if(mem[i].pid == 0) {
	DMESG("Taking new slot " << i << std::endl);
	mem[i].pid=pid;
	mem[i].timestamp=time(NULL);
	mem[i].status=status;
	strncpy(mem[i].user, user.c_str(), USERNAME_MAX_SIZE);
	break;
      }
    }
  }

  if(i==entries){         // Ran out of process table space
    LOG("Not enough processes!" << std::endl);
  }

  unLock();
}


//! Identifies stale processes
/** Identifies processes which were idle for a too long time.
    @param secs The number of seconds after which processes are
                considered to be idle.
    @return Returns the first process identified to be idle
*/
pid_t MDConMan::identifyStale(int secs)
{
  pid_t pid=0;

  DMESG("identifyStale() called at " << (int)time(NULL) << " stale time "
	<< secs << std::endl);

  if(lock())
    return pid;

  nPrepared =0;

  for(int i=0; i<entries; i++){
    DMESG("Slot: " << i << "  pid: " << mem[i].pid << "  status: "
	  << mem[i].status << std::endl);
    if(mem[i].pid) {
      switch (mem[i].status) {
      case 0:  // Not assigned
	continue;
      case 1:  // Assigned, not yet running
	if(mem[i].timestamp <= time(NULL)-20){
	  // Master reserved slot but did not use it
	  mem[i].status=0;
	  mem[i].pid=0;
	  mem[i].timestamp=0;
	  mem[i].user[0]='\0';
	  LOG("FOUND a prepared, abandoned slot!" << std::endl);
	  continue;
	}
      case 2:  // Preparing db connection [60 sec timeout]
	if(mem[i].timestamp <= time(NULL)-60){
	  pid=mem[i].pid;
	  DMESG("identified: " << pid << std::endl);
	  break;
	}
	nPrepared++;
	continue;
      case 3:  // Waiting for connect [10 min timeout]
	if(mem[i].timestamp <= time(NULL)-600){
	  pid=mem[i].pid;
	  DMESG("identified: " << pid << std::endl);
	  break;
	}
	nPrepared++;
	continue;
      case 6:  // Termination signal sent
	continue;
      case 4:  // Connection established
      case 5:  // Processing request
      default:
	if(mem[i].timestamp <= time(NULL)-secs){
	  pid=mem[i].pid;
	  DMESG("identified: " << pid << std::endl);
	  break;
	}
      }
    }
  }

  if(unLock())
    return pid;

  return pid;
}


//! Removes a process from the list of kept processes
/** Removes a process from the list of processes taken care of.
    @param pid The process id to remove
    @return Returns 0 in case of error, else a positive value
*/
int MDConMan::remove(pid_t pid)
{
  if(lock())
    return -1;

  int i;
  for(i=0; i<entries; i++){
    if(mem[i].pid == pid) {
      mem[i].pid=0;
      mem[i].timestamp=0;
      mem[i].status=0;
      mem[i].user[0]='\0';
    }
  }

  if(i==entries)
    i=0;

  if(unLock())
    return 0;

  return i;
}


//! Cleans up a slot if a child could not be spawned
int MDConMan::cleanupSlot(int slot)
{
  if(slot<0 || slot>entries) {
    ERRLOG("Calling Error!" << std::endl);
    abort();
  }
  if(lock())
    return -1;

  mem[slot].pid=0;
  mem[slot].status=0;

  if(unLock())
    return 0;

  return 0;
}


//! Prepare new slots to house new processes
/** Fills a list of slots with the number of the free slot which is
    assigned and sets up the corresponding process slot
*/
int MDConMan::prepareProcs(std::vector<int> &slots)
{
  if(lock())
    return -1;

  DMESG("Starting" << std::endl);

  int i;
  std::vector<int>::iterator I=slots.begin();
  for(i=0; i<entries; i++) {
    DMESG("Slot: " << i << "  pid: " << mem[i].pid << "  status: "
	  << mem[i].status << std::endl);
    if(mem[i].pid==1)
      LOG("Slot already prepared: THIS MUST NOT HAPPEN!" << std::endl);

    if(!mem[i].pid && !mem[i].status) { // Unused slot
      DMESG("Found slot: " << i << std::endl);
      mem[i].status=1;
      mem[i].pid=-1;
      mem[i].timestamp=time(NULL);
      *I++=i;
      if(I==slots.end())
	break;
    }
  }

  if(unLock())
    return -1;

  while(I!=slots.end())
    *I++=-1;

  return 0;
}


int MDConMan::waitMaster(int sec)
{
  DMESG("Wait Master!" << std::endl);

  if(sec==-1)
    sec=interval;

  if(!sec) {
    if(sem_wait(masterSem)) {
      PERRLOG("Error waiting for " << masterSemName  << std::endl);
      return -1;
    }
    return 0;
  }

  struct timespec t;
  time(&t.tv_sec);
  t.tv_sec+=sec;
  t.tv_nsec=0;

  if(sem_timedwait(masterSem, &t)) {
    if(errno==ETIMEDOUT){
      DMESG("Master wakes up because timeout" << std::endl);
    } else {
      if(errno==EINTR){
	DMESG("Master wakes up because of interrrupt" << std::endl);
	return 0;
      }
      PERRLOG("Error timed waiting for " << masterSemName  << std::endl);
      return -1;
    }
  }

  return 0;
}


int MDConMan::notifyMaster()
{
  DMESG("Notify Master!" << std::endl);

  if(sem_post(masterSem)) {
    PERRLOG("Error posting " << masterSemName  << std::endl);
    return -1;
  }
  return 0;
}


//! Locks the semaphore that guards the process table
int MDConMan::lock()
{
  DMESG("LOCK\n");
  int err;
  while( (err=sem_wait(sem)) ) {
    if(errno == EINTR)
      continue;
    PERRLOG("Error waiting for semaphore MDConManLock" << std::endl);
    return -1;
  }

  return 0;
}


//! Unlocks the semaphore that guards the process table
int MDConMan::unLock()
{
  DMESG("UNLOCK\n");
  if(sem_post(sem)){
    LOG("Error posting semaphore " << lockName<< std::endl);
    return -1;
  }
  return 0;
}


int MDConMan::reap(int idletime)
{
  DMESG("Reap called" << std::endl);

  int pid;
  while( (pid=identifyStale(idletime)) ){
    DMESG("Reaping process: " << pid << std::endl);
    kill(pid, SIGTERM);
    timestamp(6, "", pid);
  }
  DMESG("Leaving reap" << std::endl);
  return 0;
}


int MDConMan::reapAll()
{
  DMESG("Reap all called" << std::endl);

  for(int i=0; i<entries; i++){
    if(mem[i].pid && mem[i].status){
      int pid=mem[i].pid;
      DMESG("Asking to terminate: " << pid << std::endl);
      kill(pid, SIGTERM);
      timestamp(6, "", pid, i);
    }
  }
  DMESG("Leaving reap all" << std::endl);
  return 0;
}


int MDConMan::killLeftovers(bool immediately)
{
  DMESG("KillLeftovers called, immediately is " << immediately
	<< std::endl);

  for(int i=0; i<entries; i++){
    if(mem[i].pid && mem[i].status==6){
      int pid=mem[i].pid;
      if(immediately || mem[i].timestamp <= time(NULL)-5){
	DMESG("KILLING: " << pid << std::endl);
	kill(pid, SIGKILL);
      }
    }
  }
  DMESG("Leaving reap all" << std::endl);
  return 0;
}


int MDConMan::start(int sec, int maxIdleTime)
{
  interval=sec;
  idleTime=maxIdleTime;
  masterConMan=this;
  unLock();
  return 0;
}


bool MDConMan::startTimer(int secs, int maxIdleTime)
{
  DMESG("Starting timer: " << secs << "   " << maxIdleTime << std::endl);

  interval=secs;
  idleTime=maxIdleTime;
  masterConMan=this;
  signal(SIGALRM, timer_handler);
  DMESG("Interval1: " << interval << std::endl);
  alarm(interval);
  unLock();
  return true;
}


/** Returns a deterministic session ID based on the processID and the
    slot number in which this process runs in the connection manager.
*/
uint64_t MDConMan::mySessionID()
{
  pid_t pid=getpid();
  int slot;
  for(slot=0; slot<entries; slot++) {
    if(mem[slot].pid==pid)
      break;
  }
  if(slot>=entries)
    return 0;

  uint64_t session=slot;
  session<<=16;
  session+=pid;
  session<<=16;
  session+=time(NULL);

  return session;
}

/** Returns a random 64bit session ID
 */
uint64_t MDConMan::newSessionID()
{
  uint64_t sessionID;
  if(read(urandom, &sessionID, 8)<=0)
    ERRLOG("Problem reading random numbers");

  if(!sessionID)
    sessionID++;

  return sessionID;
}

/** Return the maximum size a session can have
 */
size_t MDConMan::getMaxSessionSize() const
{
  return maxSessionSize;
}


unsigned char *MDConMan::saveSessionSSL(const char *data, int size,
					unsigned char *sID,
					unsigned int sIDLen)
{
  ERRLOG("MDConMan::saveSessionSSL not available!\n");
  return NULL;
}


uint64_t MDConMan::saveSessionUser(const UserContext &context,
				   uint64_t sID)
{
  ERRLOG("MDConMan::saveSessionUser not available!\n");
  return 0;
}


int MDConMan::getSessionSSL(unsigned char *sID, unsigned int sIDLen,
			    char *data, int &size)
{
  ERRLOG("MDConMan::getSessionSSL not available!\n");
  return -1;
}


int MDConMan::getSessionUser(const uint64_t &sID,
			     UserContext &context)
{
  ERRLOG("MDConMan::saveSessionUser not available!\n");
  return -1;
}


int MDConMan::deleteSession(unsigned char *sID,
			    unsigned int sIDLen)
{
  ERRLOG("MDConMan::delete not available!\n");
  return -1;
}


int MDConMan::initSessionCache(int sessions, int maxSize,
			       const std::string &file)
{
  ERRLOG("MDConMan::saveSessionUser not available!\n");
  return -1;
}


int MDConMan::closeSessionCache()
{
  ERRLOG("MDConMan::closeSessionCache not available!\n");
  return -1;
}


int MDConMan::expireSessions(int age)
{
  ERRLOG("MDConMan::saveSessionUser not available!\n");
  return -1;
}


void MDConMan::marshalUserContext(const UserContext &context,
				  char *buffer, int &offset)
{
  strncpy(buffer+offset, context.login.c_str(), UserContext::maxLoginSize);
  offset += UserContext::maxLoginSize;
  strncpy(buffer+offset, context.name.c_str(), UserContext::maxNameSize);
  offset += UserContext::maxNameSize;
  strncpy(buffer+offset, context.permissionMask.c_str(),
	  UserContext::maxPermissionSize);
  offset += UserContext::maxPermissionSize;
  strncpy(buffer+offset, context.groupMask.c_str(),
	  UserContext::maxGroupMaskSize);
  offset += UserContext::maxGroupMaskSize;
  strncpy(buffer+offset, context.currentDir.c_str(),
	  UserContext::maxPathSize);
  offset += UserContext::maxPathSize;
}


void MDConMan::unmarshalUserContext(UserContext &context,
				    char *buffer, int &offset)
{
  context.login = std::string(buffer + offset);
  offset += UserContext::maxLoginSize;
  context.name = std::string(buffer + offset);
  offset += UserContext::maxNameSize;
  context.permissionMask = std::string(buffer + offset);
  offset += UserContext::maxPermissionSize;
  context.groupMask=std::string(buffer + offset);
  offset += UserContext::maxGroupMaskSize;
  context.currentDir = std::string(buffer + offset);
  offset += UserContext::maxPathSize;
}


int MDConMan::getMaxConnections() const
{
  return entries;
}


int MDConMan::getPreparedConnections() const
{
  int prepared =0;

  for(int i=0; i<entries; i++){
    if( mem[i].pid && mem[i].status == 3 )
      prepared++;
  }  return prepared;
}


int MDConMan::getUsedConnections() const
{
  int nUsed =0;

  for(int i=0; i<entries; i++){
    if(mem[i].pid && (mem[i].status == 4 || mem[i].status == 5))
      nUsed++;
  }

  return nUsed;
}


int MDConMan::getMaxSessions() const
{
  return maxSessions;
}


int MDConMan::getUserStatistics(std::map<std::string, int> &stats)
{
  if(lock())
    return -1;

  DMESG("In stats:\n");

  for(int i=0; i<entries; i++){
    if(mem[i].pid && mem[i].status == 5) {
    std::string user = mem[i].user;
      if( !user.size() )
	user = "<unknown>";
      DMESG("User: " << user << "\n");
      if( stats.find(user) != stats.end() )
	stats[user]++;
      else
	stats[user]=1;
    }
  }

  if(unLock())
    return -1;

  return stats.size();
}


int MDConMan::getUserConnections(const std::string &user)
{
  int nConnections = 0;

  if(lock())
    return -1;

  for(int i=0; i<entries; i++){
    if(mem[i].pid && mem[i].status == 5) {
    if( user == mem[i].user )
    	nConnections++;
    }
  }

  if(unLock())
    return -1;

  return nConnections;
}


void MDConMan::collectCounters(std::map<std::string, std::string> &counters)
{
  DMESG("Collecting counters\n");

  counters["MaxConnections"] = toString(getMaxConnections());
  counters["PreparedConnections"] = toString(getPreparedConnections());
  counters["UsedConnections"] = toString(getUsedConnections());
  counters["MaxSessions"] = toString(getMaxSessions());
  counters["UsedSessions"] = toString(getUsedSessions());
  counters["SessionStorage"] = getSessionStorage();
  counters["Uptime"] = toString((int)(time(NULL) - startupTime));
  counters["NodeName"] = ReplicationManager::getInstance()->getNodeName();

  std::map<std::string, int> stats;
  getUserStatistics(stats);
  DMESG("Stat size is " << stats.size() << "\n");
  std::map<std::string, int>::const_iterator I = stats.begin();
  while( I != stats.end() ) {
    DMESG("Looking at " << I->first << "\n");
    counters["User::" + I->first] = toString(I->second);
    I++;
  }
}


int MDConMan::USlock()
{
  int err;
  while( (err = sem_wait(USsem)) ) {
    if(errno == EINTR)
      continue;
    PERRLOG("Error waiting for semaphore MDConManLock" << std::endl);
    DMESG("ERR\n");
    return -1;
  }

  return 0;
}

int MDConMan::USunlock()
{
  if(sem_post(USsem)){
    LOG("Error posting semaphore US" << std::endl);
    return -1;
  }
  return 0;
}

