/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


/** -*- Mode:c++ -*-
    $Id: MDConMan.h,v 1.11 2011/01/26 05:57:13 asunil Exp $

    MDConMan is a class which implements an asynchronous buffer which
    sits in between MDServer and the server program//
**/

#ifndef MDCONMAN_H
#define MDCONMAN_H

#include <vector>
#include <semaphore.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdint.h>

#include "UserContext.h"
#include "util/StatsCollector.h"
#include "Constants.h"

class MDConMan : public StatsSource {
public:
  MDConMan();
  MDConMan(int nprocs);
  virtual ~MDConMan();

  bool init(int nprocs=0);
  int setSlave(int slot);
  void timestamp(int status, const std::string &user="",
		 pid_t pid=0, int slot=0);

  pid_t identifyStale(int secs);
  int reap(int idletime);
  int killLeftovers(bool immediately);
  int reapAll();

  int USlock();
  int USunlock();

  // Start with timer, or without
  bool startTimer(int secs, int maxIdleTime);
  int start(int sec, int maxIdleTime);

  static int getInterval() { return interval; };

  static int getMaxIdleTime() { return idleTime; };
  int closeConMan();
  int remove(pid_t pid);

  uint64_t mySessionID();

  virtual unsigned char *saveSessionSSL(const char *data, int size,
					unsigned char *sID,
					unsigned int sIDLen);

  virtual uint64_t saveSessionUser(const UserContext &context,
				   uint64_t sID = 0);

  virtual int getSessionSSL(unsigned char *sID, unsigned int sIDLen,
			    char *data, int &size);

  virtual int getSessionUser(const uint64_t &sID,
			     UserContext &context);

  virtual int deleteSession(unsigned char *sID,
			    unsigned int sIDLen);

  virtual int initSessionCache(int sessions=1024, int maxSize=3000,
			       const std::string &file="");

  virtual int closeSessionCache();
  virtual int expireSessions(int age = 1000);

  int cleanupSlot(int slot);
  int prepareProcs(std::vector<int> &slots);

  int waitMaster(int time=-1);
  int notifyMaster();

  int getPrepared() const { return nPrepared; }

  bool isMaster() const { return master; }
  uint64_t newSessionID();
  size_t getMaxSessionSize() const;

  // Statistics
  int getMaxConnections() const;
  int getPreparedConnections() const;
  int getUsedConnections() const;
  int getMaxSessions() const;
  virtual int getUsedSessions() const = 0;
  virtual std::string getSessionStorage() const = 0;
  int getUserConnections(const std::string &user);
  int getUserStatistics(std::map<std::string, int> &stats);

  // Statistics gathering
  virtual void collectCounters(std::map<std::string, std::string> &counters);

protected:
  static unsigned int maxSessions;
  static unsigned int maxSessionSize;
  void marshalUserContext(const UserContext &context,
			  char *buffer, int &offset);
  void unmarshalUserContext(UserContext &context,
			    char *buffer, int &offset);

  int lock();
  int unLock();

  bool master;
  bool useSessionCache;

private:
  int takeSlot(int slot);

  struct Process {
    pid_t pid;
    time_t timestamp;
    int16_t status;
    // uint64_t sessionID; KIMNG090331 This is not used on the code.
    char user[USERNAME_MAX_SIZE];

    // KIMNG080416 the change of the Process struct for a multi-threaded process
    /* 0 : before fork
     * 1 : after fork(creating a process) & assigned a index of the mem array
     * 2 : before connecting to DB
     * 3 : after connecting DB and before creating threads(?)
     * 4 : after creating threads and before running threads
     * 5 : after running threads
     * 6 : after sening a termination signal to all threadss
     */
    uint64_t thread_num; // the # of thread which a process has
    pthread_t *pThreadData; // the pointer of thread struct

  };

  static int interval;
  static int idleTime;

  sem_t *sem;
  sem_t *masterSem;
  struct Process *mem;
  int fd; // File descriptor of the shared memory file

  sem_t *USsem;
  static std::string USlockName;


  static int size;
  static int entries;
  static std::string lockName;
  static std::string mapName;
  static std::string masterSemName;
  static int urandom;


  int nPrepared; // Number of prepared processes
};

#endif
