/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDConManDB.h"

#include <list>

#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "util/sslHelpers.h"

static int debug = 0;
#include "debug.h"

std::string MDConManDB::dbFile;


MDConManDB::MDConManDB()
{
  // Nothing to do...
}


MDConManDB::MDConManDB(int nprocs)
  : MDConMan(nprocs)
{
  // Nothing more to do, here
}


MDConManDB::~MDConManDB()
{
  DMESG("~MDConManDB, master: " << master << std::endl);
  if(useSessionCache)
    closeSessionCache();
}


unsigned char *MDConManDB::saveSessionSSL(const char *data, int size,
					  unsigned char *sID, unsigned int sIDLen)
{
  DMESG("saveSessionSSL: " << sessionToPrintable(sID, sIDLen) 
	<< "  size: " << size << "  maximum session size is "
	<< maxSessionSize << std::endl);

  if(!useSessionCache)
    return 0;

  unsigned int actualSize=size
    + sizeof(time_t);

  if(actualSize > maxSessionSize) {
    ERRLOG("Could not save session: Session too large, size is "
	   << actualSize << " but maximum allowed size is " 
	   << maxSessionSize << "\n");
    return NULL;
  }

  char *buffer=new char[actualSize];
  int offset=0;
  *((time_t *)(buffer+offset)) = time(NULL);
  offset += sizeof(time_t);

  memcpy(buffer+offset, data, size);

  datum key, value;
  key.dptr = (char *)sID;
  key.dsize = sIDLen;
  value.dptr = buffer;
  value.dsize = actualSize;

  if( lock() )
    return NULL;

  GDBM_FILE db= gdbm_open((char *)(dbFile.c_str()), 0, GDBM_WRITER, 0600, NULL);
  if(!db) {
    PERRLOG("Could not open session db (" << gdbm_errno << ")  ");
    unLock();
    return NULL;
  }

  int ret = gdbm_store(db, key, value, GDBM_INSERT);

  if( ret ){
    ERRLOG("Error inserting session into db: " << gdbm_strerror(ret) << std::endl);
    delete buffer;
    gdbm_close(db);
    unLock();
    return NULL;
  }

  DMESG("session stored" << std::endl);
  delete [] buffer;
  gdbm_close(db);
  unLock();

  return sID;
}


uint64_t MDConManDB::saveSessionUser(const UserContext &context, uint64_t sID)
{
  DMESG("In saveSessionUser " << sID << std::endl);
  if(!useSessionCache)
    return 0;

  if(!sID)
    sID = newSessionID();

  DMESG("sessionID " << sID << std::endl);
  unsigned int actualSize = sizeof(time_t)
    + UserContext::maxLoginSize
    + UserContext::maxNameSize
    + UserContext::maxPermissionSize
    + UserContext::maxGroupMaskSize
    + UserContext::maxPathSize;

  if(actualSize > maxSessionSize)
    return 0;

  char *buffer=new char[actualSize];
  int offset=0;
  *((time_t *)(buffer+offset)) = time(NULL);
  offset += sizeof(time_t);

  marshalUserContext(context, buffer, offset);

  datum key, value;
  key.dptr = (char *)&sID;
  key.dsize = sizeof(sID);
  value.dptr = buffer;
  value.dsize = actualSize;

  DMESG("storing session ... " << std::endl);
  if( lock() )
    return 0;

  GDBM_FILE db= gdbm_open((char *)(dbFile.c_str()), 0, GDBM_WRITER, 0600, NULL);
  if(!db) {
    PERRLOG("Could not open session db (" << gdbm_errno << ")  ");
    unLock();
    return 0;
  }

  int ret = gdbm_store(db, key, value, GDBM_INSERT);

  if( ret ){
    ERRLOG("Error inserting session into db: " << gdbm_strerror(ret) << std::endl);
    delete buffer;
    gdbm_close(db);
    unLock();
    return 0;
  }

  DMESG("session stored" << std::endl);
  delete [] buffer;
  gdbm_close(db);
  unLock();

  return sID;
}


int MDConManDB::getSessionSSL(unsigned char *sID, unsigned int sIDLen,
			      char *data, int &size)
{  
  DMESG("getSessionSSL: " << sessionToPrintable(sID, sIDLen) << std::endl);

  if(!useSessionCache)
    return 0;

  datum key, value;
  key.dptr = (char *)sID;
  key.dsize = sIDLen;

  if( lock() )
    return -1;

  GDBM_FILE db= gdbm_open((char *)(dbFile.c_str()), 0, GDBM_WRITER, 0, NULL);
  if(!db) {
    PERRLOG("Could not create session db (" << gdbm_strerror(gdbm_errno) << ")  ");
    unLock();
    return -1;
  }

  value = gdbm_fetch(db, key);

  gdbm_close(db);
  unLock();

  if(!value.dptr) {
    PERRLOG("Error retrieving session from db (" << gdbm_strerror(gdbm_errno) << ")  ");
    return -1;
  }

  DMESG("Retrieved session " << std::endl);

  if(value.dsize > maxSessionSize){
    free(value.dptr);
    return -1;
  }

  int offset = 0;
  time_t timestamp = *(time_t *)(value.dptr + offset);
  offset += sizeof(time_t);

  memcpy(data, value.dptr+offset, value.dsize-offset);

  size=value.dsize-offset;

  free(value.dptr);
  DMESG("getSessionData exiting " << std::endl);
  return 0;
}


int MDConManDB::getSessionUser(const uint64_t &sID, UserContext &context)
{
  if(!useSessionCache)
    return 0;

  DMESG("getSessionData for " << sID << std::endl);

  datum key, value;
  key.dptr = (char *) &sID;
  key.dsize = sizeof(sID);

  if( lock() )
    return 0;

  GDBM_FILE db= gdbm_open((char *)(dbFile.c_str()), 0, GDBM_WRITER, 0, NULL);
  if(!db) {
    PERRLOG("Could not create session db (" << gdbm_strerror(gdbm_errno) << ")  ");
    unLock();
    return -1;
  }

  value = gdbm_fetch(db, key);

  if(value.dptr)
    gdbm_delete(db, key);

  gdbm_close(db);
  unLock();

  if(!value.dptr) {
    PERRLOG("Erro retrieving session from db (" << gdbm_strerror(gdbm_errno) << ")  ");
    return -1;
  }

  DMESG("Retrieved session " << sID << std::endl);

  if(value.dsize > maxSessionSize){
    free(value.dptr);
    return -1;
  }

  int offset = 0;
  time_t timestamp = *(time_t *)(value.dptr + offset);
  offset += sizeof(time_t);
  unmarshalUserContext(context, value.dptr, offset);

  free(value.dptr);
  return 0;
}


int MDConManDB::deleteSession(unsigned char *sID, unsigned int sIDLen)
{
  if(!useSessionCache)
    return 0;

  DMESG("deleteSessionData for " << sID << std::endl);

  datum key;
  key.dptr = (char *)sID;
  key.dsize = sIDLen;

  if( lock() )
    return -1;

  GDBM_FILE db= gdbm_open((char *)(dbFile.c_str()), 0, GDBM_WRITER, 0, NULL);
  if(!db) {
    PERRLOG("Could not open session db (" 
	    << gdbm_strerror(gdbm_errno) << ")  ");
    unLock();
    return -1;
  }

  if(gdbm_delete(db, key)){
    PERRLOG("Error deleting session from db (" 
	    << gdbm_strerror(gdbm_errno) << ")  ");
    gdbm_close(db);
    unLock();
    return -1;
  }

  gdbm_close(db);
  unLock();

  return 0;
}


/** Initialize the session cache. Needs to be called by user application before
    daughter processes are being spawned. No locking is done.
    After successful creation of the database file, the useSessionCache member is
    set to true.
*/
int MDConManDB::initSessionCache(int sessions, int maxSize,
				 const std::string &file)
{
  if (!master)
    return 0;

  DMESG("Initializing session cache" << std::endl);

  maxSessions=sessions;
  maxSessionSize=maxSize;

  dbFile=file;
  if(!file.size()){
    dbFile="/tmp/mdConMan";
    char buff[10];
    sprintf(buff, "%d", getpid());
    dbFile.append(buff).append(".db");
  }

  unlink(dbFile.c_str());
  GDBM_FILE db= gdbm_open((char *)(dbFile.c_str()), 8192, GDBM_WRCREAT,
			  S_IWUSR | S_IRUSR, NULL);
  if(!db) {
    PERRLOG("Could not create session db " << gdbm_strerror(gdbm_errno) << "  ");
    unLock();
    return -1;
  }

  gdbm_close(db);

  useSessionCache = true;

  return 0;
}


int MDConManDB::closeSessionCache()
{
  if(master){
    DMESG("Unlinking db" << std::endl);
    unlink(dbFile.c_str());
  }

  return 0;
}


int MDConManDB::expireSessions(int age)
{
  if(!useSessionCache)
    return -1;

  LOG("Expiring sessions from session DB" << std::endl);

  time_t now = time(NULL);

  datum key, value;
  std::list<std::pair <char *, size_t> > expiredSessions;

  if( lock() )
    return -1;

  GDBM_FILE db= gdbm_open((char *)(dbFile.c_str()), 0, GDBM_WRITER, 0600, NULL);
  if(!db) {
    PERRLOG("Could not open session db for scanning (" << gdbm_errno << ")  ");
    unLock();
    return -1;
  }

  key = gdbm_firstkey(db);
  while( key.dptr ) {
    value = gdbm_fetch(db, key);
    if(value.dptr) {
      if(*(time_t *)(value.dptr) < now-age) {
	std::pair <char *, size_t> p(key.dptr, key.dsize);
	expiredSessions.push_back(p);
	DMESG("Marking for expiery " << std::endl);
	free(value.dptr);
      }
    } else {
      PERRLOG("Could not access session db for scanning (" << gdbm_errno << ")  ");
      unLock();
      return -1;
    }
    key = gdbm_nextkey(db, key);
  }

  std::list<std::pair <char *, size_t> >::iterator I;
  for (I=expiredSessions.begin(); I != expiredSessions.end(); I++) {
    key.dptr = I->first;
    key.dsize = I->second;
    if(gdbm_delete(db, key))
      PERRLOG("Could not delete session "
	      << gdbm_strerror(gdbm_errno) << std::endl);
    free(key.dptr);
  }

  // Finally, make available free space
  // ret = gdbm_reorganize(db);

  gdbm_close(db);
  unLock();

  return 0;
}


int MDConManDB::getUsedSessions() const
{
  int used = 0;

  return used;
}


std::string MDConManDB::getSessionStorage() const
{
  return "GDBM";
}
