/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDConManMem.h"

#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#include "util/sslHelpers.h"

static int debug = 0;
#include "debug.h"

std::string MDConManMem::hashMapName;

MDConManMem::MDConManMem()
  : hashMem((Hash *)MAP_FAILED)
{
  // Nothing to do...
}


MDConManMem::MDConManMem(int nprocs)
  : MDConMan(nprocs), hashMem((Hash *) MAP_FAILED)
{
  //session cache
}


MDConManMem::~MDConManMem()
{
  DMESG("~MDConManDB, master: " << master << std::endl);
  if(useSessionCache)
    closeSessionCache();
}


uint32_t MDConManMem::hash(unsigned char *key, unsigned int len)
{
  uint32_t h = 0;

  for ( unsigned int i=0; i <= len/4; i+=4 ) {
    for ( unsigned int j=0; j< 4; j++ ) {
      if(i*4+j >= len)
	break;
      h <<=5;
      h ^= key[i*4+j];
    }
  }

  if(debug) {
    std::stringstream asc;
    for (unsigned int i=0; i<len; i++){
      asc.width(2); asc.fill('0');
      asc << std::hex << ((unsigned int)key[i]);
    }
    DMESG("Key " << asc << " -> " << (h&hashBits) << std::endl);
  }

  return h & hashBits;
}


/** Gets a free hash slot. This function must be called with lock()
    taken!
*/
int MDConManMem::getNewHashSlot(unsigned char *key, unsigned int len)
{
  uint32_t h = hash(key, len);
  unsigned int tries=0;
  DMESG("getNewHashSlot h " << h << std::endl);
  while(hashMem[h].keyLen && tries++ < maxSessions)
    h = (h+1) % maxSessions;
 
  if(tries >= maxSessions)
    return -1;

  DMESG("h now " << h << std::endl);

  return h;
}


/** Gets the hash slot associated with the given key.
    This function must be called with lock() taken!
*/
int MDConManMem::getHashSlot(unsigned char *key, unsigned int len)
{
  uint32_t h = hash(key, len);
  unsigned int tries=0;
  DMESG("getHashSlot h " << h << "  len " << len << std::endl);
  while(tries++ < maxSessions) {
    if(len == (unsigned int) hashMem[h].keyLen) {
      if(!strncmp((const char *)key, hashMem[h].key, len))
	break;
    }
    h = (h+1) % maxSessions;
  }
 
  if(tries >= maxSessions)
    return -1;

  DMESG("hash slot now " << h << std::endl);

  return h;
}


unsigned char *MDConManMem::saveSessionSSL(const char *data, int size,
					   unsigned char *sID,
					   unsigned int sIDLen)
{
  DMESG("saveSessionSSL: " << sessionToPrintable(sID, sIDLen) 
	<< "  size: " << size << "  maximum session size is "
	<< maxSessionSize << std::endl);

  if (!useSessionCache)
    return NULL;

  if(sIDLen > MAX_KEYLEN)
    return NULL;

  unsigned int actualSize=size
    + sizeof(time_t);

  if(actualSize > maxSessionSize) {
    ERRLOG("Could not save session: Session too large, size is "
	   << actualSize << " but maximum allowed size is " 
	   << maxSessionSize << "\n");
    return NULL;
  }

  if(lock())
    return NULL;

  int s = getNewHashSlot(sID, sIDLen);
  if (s<0) {
    ERRLOG("Could not save session: Too many open sessions\n");
    unLock();
    return NULL;
  }

  char *buffer = ((char *)&hashMem[maxSessions])+maxSessionSize*s;
  int offset=0;
  *((time_t *)(buffer+offset)) = time(NULL);
  offset += sizeof(time_t);
  
  memcpy(buffer+offset, data, size);
  memcpy(hashMem[s].key, sID, sIDLen);
  hashMem[s].keyLen = sIDLen;
  hashMem[s].dSize = size + offset;

  DMESG("session stored" << std::endl);

  unLock();

  return sID;
}


uint64_t MDConManMem::saveSessionUser(const UserContext &context, uint64_t sID)
{
  if (!useSessionCache)
    return 0;

  if(!sID)
    sID = newSessionID();

  DMESG("sessionID " << sID << std::endl);
  unsigned int actualSize = sizeof(time_t)
    + UserContext::maxLoginSize
    + UserContext::maxNameSize
    + UserContext::maxPermissionSize
    + UserContext::maxGroupMaskSize
    + UserContext::maxPathSize;

  if(actualSize > maxSessionSize)
    return 0;

  if(lock())
    return 0;

  int s = getNewHashSlot((unsigned char *) &sID, 8);
  if (s<0) {
    ERRLOG("Could not save session: No available slot!\n");
    unLock();
    return 0;
  }
  
  char *buffer=((char *)&hashMem[maxSessions])+maxSessionSize*s;
  int offset=0;
  *((time_t *)(buffer+offset)) = time(NULL);
  offset += sizeof(time_t);

  marshalUserContext(context, buffer, offset);
  *(uint64_t *)( &(hashMem[s].key) ) = sID;
  hashMem[s].keyLen = 8;
  hashMem[s].dSize = offset;

  unLock();

  return sID;
}


int MDConManMem::getSessionSSL(unsigned char *sID, unsigned int sIDLen,
			       char *data, int &size)
{
  DMESG("getSessionSSL: " << sessionToPrintable(sID, sIDLen) << std::endl);
  if (!useSessionCache)
    return 0;

  if(lock())
    return -1;

  int s = getHashSlot(sID, sIDLen);
  if (s<0) {
    ERRLOG("Could not retrieve SSL session: Session not found\n");
    unLock();
    return -1;
  }

  char *value =((char *)&hashMem[maxSessions]) +maxSessionSize*s;

  int offset = 0;
  time_t timestamp = *(time_t *)(value + offset);
  offset += sizeof(time_t);

  size = hashMem[s].dSize-offset;
  memcpy(data, value+offset, size);
  
  unLock();

  DMESG("getSessionData exiting " << std::endl);
  return 0;
}


/** Returns a user session context for a given sessionID.
    Removes the sessionID from the session storage.
*/
int MDConManMem::getSessionUser(const uint64_t &sID, UserContext &context)
{
  
  DMESG("getSessionUser "  << sID << std::endl);
  if (!useSessionCache)
    return 0;

  if(lock())
    return -1;

  int s = getHashSlot((unsigned char *) &sID, 8);
  if (s<0) {
    ERRLOG("Could not retrieve user session: Session not found\n");
    unLock();
    return -1;
  }

  char *value = ((char *)&hashMem[maxSessions])+maxSessionSize*s;

  int offset = 0;
  time_t timestamp = *(time_t *)(value + offset);
  offset += sizeof(time_t);

  unmarshalUserContext(context, value, offset);
  hashMem[s].keyLen = 0;
  unLock();

  return 0;
}


int MDConManMem::deleteSession(unsigned char *sID, unsigned int sIDLen)
{
  DMESG("deleting Session with size: "<< sIDLen << std::endl);

  if (!useSessionCache)
    return 0;

  if( lock() )
    return -1;

  int s = getHashSlot(sID, sIDLen);
  DMESG("slot : " << s << std::endl);
  if (s<0) {
    ERRLOG("Could not delete session: Session not found\n");
    unLock();
    return -1;
  }

  hashMem[s].keyLen=0;

  unLock();

  DMESG("Session deleted fine\n");

  return 0;
}


/** Initialize the session cache. Needs to be called by user application before
    daughter processes are being spawned. No locking is done.
    After successful creation of the database file, the useSessionCache member is
    set to true.
*/
int MDConManMem::initSessionCache(int sessions, int maxSize,
				  const std::string &file)
{
  DMESG("Initializing shared memory session cache" << std::endl);

  if(master) {
    maxSessions = sessions;
    maxSessionSize =maxSize;
    hashBits = ffs(sessions);

    assert(!(sessions ^ (1 << (hashBits-1))));
    hashBits = (1 << (hashBits-1) )-1;

    hashMapName="MDSessionShm";
    char buff[10];
    sprintf(buff, "%d", getpid());
    hashMapName.append(buff);

    DMESG("Creating " << hashMapName << std::endl);
    if((hashFD=shm_open(hashMapName.c_str(), O_CREAT | O_RDWR, S_IRWXU))==-1){
      PERRLOG("Could not create "<< hashMapName);
      return -1;
    }

    hashSize=sessions*(sizeof(struct Hash)+maxSessionSize);
    DMESG("Resizing shm area " << hashMapName << " to " << hashSize << std::endl);
    // hashSize = 1000000;

    if(ftruncate(hashFD, hashSize) ==-1){
      PERRLOG("Could not reserve hash space for " << hashMapName
	      << " (" << hashSize << ") needed");
      return -1;
    }
  } else {
    DMESG("Slave opening " << hashMapName << std::endl);
    if((hashFD=shm_open(hashMapName.c_str(), O_RDWR, S_IRWXU))==-1 ){
      PERRLOG("Could not open " << hashMapName);
      return false;
    }
  }

  // Needs to be done by all processes
  hashMem=(struct Hash *)mmap(NULL, hashSize,
			      PROT_READ|PROT_WRITE,
			      MAP_SHARED, hashFD, 0);
  if(hashMem==MAP_FAILED) {
    PERRLOG("Could not map hash memory");
    return -1;
  }

  if(master) { // Initialize the session cache
    for(unsigned int i=0; i<maxSessions; i++)
      hashMem[i].keyLen=0;
  }
  useSessionCache = true;
  return 0;
}


int MDConManMem::closeSessionCache()
{
  DMESG("closeSessionCache with master=" << master << std::endl); 
  if(hashMem != MAP_FAILED){
    if(munmap(hashMem, hashSize)){
      PERRLOG("Could not munmap "<< hashMapName);
    }
  }

  if(close(hashFD)){
    PERRLOG("Could not close shared memory " << hashMapName);
  }

  if(master){
    if(shm_unlink(hashMapName.c_str())) {
      PERRLOG("Could not unlink"<< hashMapName);
    }
  }
  return 0;
}


int MDConManMem::expireSessions(int age)
{
  LOG("Expiring sessions in shared memory\n");
  int expired = 0;

  if(!useSessionCache)
    return -1;

  time_t now = time(NULL);

  if(lock())
    return -1;

  for (unsigned int i=0; i<maxSessions; i++) {
    if(!hashMem[i].keyLen)
      continue;
    char *value = ((char *)&hashMem[maxSessions])+maxSessionSize*i;
    time_t timestamp = *(time_t *)value;
    if(*(time_t *)value < now-age) {
      DMESG("Expiring slot " << i << std::endl);
      hashMem[i].keyLen = 0;
      expired++;
    }
  }

  unLock();
  LOG("... " << expired << " sessions expired\n");

  return 0;
}


int MDConManMem::getUsedSessions() const
{
  int used = 0;

  for (unsigned int i=0; i<maxSessions; i++) {
    if(hashMem[i].keyLen)
      used++;
  }

  return used;
}


std::string MDConManMem::getSessionStorage() const
{
  return "shm";
}
