/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDFCMySQLServer.cc,v 1.20 2010/04/14 07:15:30 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDFCMySQLServer.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"
#include "sqlHelpers.h"
#include "MySQLHelper.h"

#include <sys/stat.h>
#include <assert.h>
#include <string>
#include <sstream>
#include <sqlext.h>

#include "debug.h"

MDFCMySQLServer::MDFCMySQLServer(MDBuffer &outbuf)
  : MDLFCServer(outbuf)
{
  dbHelper = new MySQLHelper();
  parser->setEngine(std::string("LFC ") + dbHelper->backend());
}


MDFCMySQLServer::~MDFCMySQLServer()
{
  delete dbHelper;
}


bool MDFCMySQLServer::alterTableColumns(const std::string &tableName,
					const AttrList &keyList,
					bool remove)
{
  
  DMESG("List-size: " << keyList.size() << std::endl);

  std::string query="ALTER TABLE ";
  query.append(tableName);
  if(remove) 
    addAttributeList(keyList, query, " DROP COLUMN ");
  else
    addTableDescription(keyList, query, " ADD ");
 
  query.append(";");
  
  DMESG("SQL: >" << query << "<" << std::endl);
  
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Could not alter table", statement);
    return false;
  }
  return true;
}


//! Creates a table with a given schema
/** Creates a table with a given schema, for which a directory must already
    exist. 
    \param id The id of the assciated directory
    \param indexTable Name of entry indexing table
    \param parent The name of the table which is inherited
    \return Returns the name of the table which was created
*/
std::string MDFCMySQLServer::createTable(const EntryProps &p,
					 AttrList &attrs,
					 Statement &statement)
{
  DMESG("In createTable\n");
  std::stringstream query;

  query << "CREATE TABLE dir" << p.id << " ("
	<< "\"entry\" BIGINT UNSIGNED PRIMARY KEY REFERENCES "
	<< p.indexTable << "(fileid) ON DELETE CASCADE";
  
  AttrList::const_iterator I=attrs.begin();    
  while(I!=attrs.end()) {
    std::string key;
    if(parser->parse(I->first, key,p.flags & EP_PLAIN)){
      out.append("7 Illegal Key\n");
      return "";
    }
    query << ", " << key << " "
	  << (dbHelper->mdToDBType((*I).second));
    I++;
  }

  query << ") TYPE = InnoDB;";

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if(statement.exec(query.str())){
    printError("9 Could not create table", statement);
    return "";
  }

  query.str("");
  query << "dir" << p.id;

  return query.str();
}


int MDFCMySQLServer::getAttributes(const std::string &tableName,
				   AttrList &keys,
				   bool all)
{
  Statement statement(*dbConn);
  if(statement.columns(tableName)){
    printError("9 Cannot read table description", statement);
    return 9;
  }

  char column[65];
  statement.bind(4, column, 64);
  char type[11];
  statement.bind(5, type, 10);
  int size;
  short int precision;
  statement.bind(7, &size);
  statement.bind(9, &precision);

  if(statement.fetch()){
    return 1;
  }

  while(!statement.fetch()) {
    std::string key(column);
    // Only user arguments if all is wrong
    if(!all && key.substr(0, parser->prefix().size()) !=  parser->prefix())
      continue;

    if(!all) // Remove prefix for user arguments
      key=key.substr(parser->prefix().size());

    keys.push_back(Attribute(key, sqlType(atoi(type), size, precision)));
  }

  return 0;
}


void MDFCMySQLServer::doBulkRead(Statement &statement, std::string &query,
				 size_t fields, bool &foundSomething)
{
  char buff[1001];
  if(statement.exec(query)){
    printError("9 Internal error", statement);
    foundSomething=true;
    return;
  }

  DMESG("EXEC done" << std::endl);
  while(1) {
    if(statement.fetch()){
      DMESG("Fetch failed: Last data read" << std::endl);
      break;
    }

    if(!foundSomething) {
      /* Fix 20090728 : Return #column instead of #rows */
      out.append("0");
      char buff[20];
      sprintf(buff, " %d\n", fields);
      out.append(buff);
    }
    foundSomething = true;

    SQLHSTMT s=statement.getStatement();
    SQLLEN err;

    for(size_t i=0; i<fields; i++){
      DMESG("Getting Field: " << i << std::endl);
      while(1) {
        int res=SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err);
        DMESG("Res: " << res << "   err: " << err << std::endl);
        if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO)
          break;
        if(err==SQL_NULL_DATA) break;
        std::string line(buff);
        AMGA::encodeLine(line);
        out.append(line);
      }
     out.append("\n");
    }
  }
  statement.close();

}


//! Sets attributes of all file matching pattern
/** Sets a list of attributes to a list of values for a file.
    The dimensions of the input vectors for the attributes and values
    need to be the same.
    \param path The relative or absolute path to the file
    \param attributes The vector of attributes to set
    \param values The vector of values

    ERRORS:
    - 1 File or directory not found
    - 3 Illegal command
      The attributes and values vectors don't match
    - 4 Permission Denied
*/
void MDFCMySQLServer::setAttr(const std::string &path,
			      std::vector< std::string > &attributes,
			      std::vector< std::string > &values)
{
  if(attributes.size() != values.size()) {
    out.append("3 Illegal command\n");
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string pattern;
  bool isPattern=sqlPattern(p.name, pattern);
  if( p.type() ) {
    isPattern = true;
    pattern="%";
  }

  // Permissions are checked on a per-entry basis only in the
  // update statement

  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::string query;
  Statement statement(*dbConn);
  query="UPDATE ";
  query.append(p.directoryTable).append(", ");
  query.append(p.indexTable);
  query.append(" SET ");
  for(unsigned int i=0; i<attributes.size(); i++){
    if(parser->parse(attributes[i], keys[i])){
      out.append("7 Illegal Key\n");
      return;
    }
    query.append(p.directoryTable).append(".");
    query.append(keys[i]);
    query.append("='");
    query.append(values[i]).append("'");
    if(i<attributes.size()-1)
      query.append(", ");
  }
  query.append(" WHERE ");
  if(! (isPattern && pattern == "%")  ){
    query.append(p.indexTable).append(".name");
    if(isPattern)
      query += " LIKE '";
    else
      query += " = '";
    query += pattern + "' AND ";
  }
  query += p.directoryTable + ".\"entry\" = " + p.indexTable + ".fileid";
  query += ";";

  DMESG("SQL: >" << query << "<" << std::endl);

  if(statement.exec(query)){
    int err = statement.getErrorCode();
    std::string msg = statement.getError();
    std::string sqlErr = statement.getSqlState();
    DMESG(err << "msg >" << msg << "< sqlErr: " << sqlErr
	  << std::endl);
    if(statement.numRows()==0) {
      printError("1 No such file", statement);
      return;
    }
    out.append("10 No such key: ");
    out.append(msg);
    out.append("\n");
    return;
  }

  out.append("0\n");
}


/** Clears an \param attribute of files matching \param path .
    The output is streamed into the out class member. For a
    description see the protocol description.
*/
void MDFCMySQLServer::clearAttr(const std::string &path,
				const std::string &attribute)
{
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string key;
  if(parser->parse(attribute, key)){
    out.append("7 Illegal Key\n");
    return;
  }

  std::string pattern;
  bool isPattern = sqlPattern(p.name, pattern);
  if( p.type() ) {
    isPattern = true;
    pattern="%";
  }

  // Permissions are checked on a per-entry basis only in the
  // update statement

  std::string query="UPDATE ";
  query += p.directoryTable + ", ";
  query += p.indexTable;
  query += " SET ";
  query += p.directoryTable + ".";
  query += key + "= NULL";
  query += " WHERE ";
  if(! (isPattern && pattern == "%")  ){
    query += p.indexTable + ".name";
    if(isPattern)
      query += " LIKE '";
    else
      query += " = '";
    query += pattern + "' AND";
  }
  query += p.directoryTable + "\".entry\" = " + p.indexTable + ".fileid";
  query.append(";");

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("2", statement);
    return;
  }

  out.append("0\n");
}



std::string MDFCMySQLServer::updateAttrQuery(const EntryProps &p,
					     const std::string &entries,
					     bool isPattern,
					     std::vector< std::string > &attributes,
					     std::vector< std::string > &values,
					     const std::string &condition,
					     bool eval)
{
  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::string query="UPDATE ";
  query.append(p.directoryTable).append(", ");
  query.append(p.indexTable);
  query.append(" SET ");
  for(unsigned int i=0; i<attributes.size(); i++){
    if(parser->parse(attributes[i], keys[i])){
      out.append("7 Illegal Key\n");
      return "";
    }
    std::string value;
    if(parser->parse(values[i], value)){
      out.append("9 Illegal value\n");
      return "";
    }
    query.append(p.directoryTable).append(".");
    query.append(keys[i]);
    query.append("=");
    query.append(value).append("");
    if(i<attributes.size()-1)
      query.append(", ");
  }
  query.append(" WHERE ");
  if(! (isPattern && entries == "%")  ){
    query.append(p.indexTable).append(".");
    query.append("name");
    if(isPattern)
      query.append(" LIKE '");
    else
      query.append(" = '");
    query.append(entries).append("'");
    query.append(" AND ");
  }
  query.append(p.indexTable);
  query.append(".fileid = ").append(p.directoryTable);
  query.append(".\"entry\"");

  if(condition.size())
    query.append(" AND (").append(condition).append(")");

  query.append(";");

  return query;
}

