/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDFCOracleServer.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"
#include "sqlHelpers.h"
#include "OracleHelper.h"

#include <sqlext.h>

#include <iostream>
#include "debug.h"

MDFCOracleServer::MDFCOracleServer(MDBuffer &outbuf)
  : MDLFCServer(outbuf)
{
  dbHelper = new OracleHelper();
  parser->setEngine(std::string("LFC ") + dbHelper->backend());
}


MDFCOracleServer::~MDFCOracleServer()
{
  dbHelper = new OracleHelper();
}


//! Creates a table with a given schema
/** Creates a table with a given schema, for which a directory must already
    exist. 
    \param id The id of the assciated directory
    \param indexTable Name of entry indexing table
    \param parent The name of the table which is inherited
    \return Returns the name of the table which was created
*/
std::string MDFCOracleServer::createTable(const EntryProps &p,
					  AttrList &attrs,
					  Statement &statement)
{
  DMESG("In createTable\n");
  std::stringstream query;

  query << "CREATE TABLE dir" << p.id << " ("
	<< "\"entry\" INTEGER PRIMARY KEY REFERENCES "
	<< p.indexTable << "(fileid) ON DELETE CASCADE";
  
  AttrList::const_iterator I=attrs.begin();    
  while(I!=attrs.end()) {
    std::string key;
    if(parser->parse(I->first, key)){
      out.append("7 Illegal Key\n");
      return "";
    }
    query << ", " <<  key << " "
	  << (dbHelper->mdToDBType((*I).second));
    I++;
  }
  query << ");";
  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if(statement.exec(query.str())){
    printError("9 Could not create table", statement);
    return "";
  }

  query.str("");
  query << "dir" << p.id;

  return query.str();
}


bool MDFCOracleServer::alterTableColumns(const std::string &tableName,
				       const AttrList &keyList,
				       bool remove)
{
  DMESG("List-size: " << keyList.size() << std::endl);

  std::string query="ALTER TABLE ";
  query.append(tableName);
  if(remove) 
    addAttributeList(keyList, query, " DROP COLUMN ", false);
  else
    addTableDescription(keyList, query, " ADD ", false);
 
  query.append(";");
  
  DMESG("SQL: >" << query << "<" << std::endl);
  
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Could not alter table", statement);
    return false;
  }
  return true;
}


void MDFCOracleServer::doBulkRead(Statement &statement, std::string &query,
				size_t fields, bool &foundSomething)
{
  char buff[1001];
  if(statement.exec(query)){
    printError("9 Internal error", statement);
    foundSomething=true;
    return;
  }

  DMESG("EXEC done" << std::endl);
  while(1) {
    if(statement.fetch()){
      DMESG("Fetch failed: Last data read" << std::endl);
      break;
    }

    if(!foundSomething) { 
      out.append("0");

      /* Fix 20090728 : Return #column instead of #rows */
      char buff[20];
      sprintf(buff, " %d\n", fields);
      out.append(buff);

      /*
      int rows=statement.numRows();
      if(rows>-1) {
	char buff[20];
	sprintf(buff, " %d\n", rows);
	out.append(buff);
      } else {
	out.append("\n");
      }
      */
    }
    foundSomething = true;

    SQLHSTMT s=statement.getStatement();
    SQLLEN err;

    for(size_t i=0; i<fields; i++){
      DMESG("Getting Field: " << i << std::endl);
      while(1) {
	int res=SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err);
	DMESG("Res: " << res << "   err: " << err << std::endl);
	if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO)
	  break;
	DMESG("HERE 1" << std::endl);
	if(err==SQL_NULL_DATA) break;
	DMESG("GET got: >" << buff << "<" << std::endl);
	std::string line(buff);
	AMGA::encodeLine(line);
	out.append(line);
      }
     out.append("\n");
    }
  }
  statement.close();
}


int MDFCOracleServer::getAttributes(const std::string &table,
				  AttrList &keys,
				  bool all)
{
  std::string tableName=stripSchema(table);
  for(unsigned int i=0; i<tableName.size(); i++)
    tableName[i]=toupper(tableName[i]);

  std::string query="select column_name, data_type, data_length, "
    "data_precision, data_scale from user_tab_columns where table_name='";
  query.append(tableName);
  query.append("';");


  char key[1024]; 
  char data_type[64];
  char data_length[64];
  char data_precision[64];
  char data_scale[64];

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);

  if(statement.exec(query)){
    printError("9 Cannot read master description table", statement);
    return -1;
  }
  statement.bind(1, key, 1024);
  statement.bind(2, data_type, 64);
  statement.bind(3, data_length, 64);
  statement.bind(4, data_precision, 64);
  statement.bind(5, data_scale, 64);
  
  while(!statement.fetch()){
    DMESG("Row: >" << key << "<   type >" << data_type
	  << "<   length >" << data_length << "<   precision >"
	  << data_precision << "<   scale >" << data_scale << "<" << std::endl);
    
    if(!all 
       && strncmp(key, parser->prefix().c_str(),
		  parser->prefix().size())
       )
      continue;
    std::stringstream type;

    int length = ((OracleHelper *)dbHelper)->toInt(data_length);
    int precision = ((OracleHelper *)dbHelper)->toInt(data_precision);
    int scale=((OracleHelper *)dbHelper)->toInt(data_scale);

    DMESG("Or  type >" << data_type
	  << "<   length " << length << "  precision "
	  << precision << "   scale " << scale << std::endl);
    
    if(!strcmp(data_type, "NUMBER")) {
      if(scale<0 && precision < 0)
	type<<"number";
      if(scale==0)
	type<<"integer";
      if(precision>0 && scale>0) {
	if(!precision)
	  type << "number(" << scale << ")";
	else
	  type << "number(" << precision << "," << scale << ")";
      }
    }
    if(!strcmp(data_type, "VARCHAR2"))
      type << "varchar(" << length << ")";
    if(!strcmp(data_type, "CHAR"))
      type << "char(" << length << ")";
    if(!strcmp(data_type, "DATE"))
      type << "date";
    if(!strncmp(data_type, "TIMESTAMP", 9)){
      type << "timestamp";
      type << "(" << scale << ")";
    }
    if(!strcmp(data_type, "FLOAT"))
      type << "float";
    if(!strcmp(data_type, "LONG"))
      type << "long";

    if(!type.str().size())
      type.str()="unknown";

    if(!all)
      keys.push_back(Attribute(key+parser->prefix().size(), type.str()));
    else
      keys.push_back(Attribute(key, type.str()));
  }

  statement.unBind();
  statement.close();

  return 0;
}


//! Sets attributes of all file matching pattern
/** Sets a list of attributes to a list of values for a file.
    The dimensions of the input vectors for the attributes and values
    need to be the same.
    \param path The relative or absolute path to the file
    \param attributes The vector of attributes to set
    \param values The vector of values

    ERRORS:
    - 1 File or directory not found
    - 3 Illegal command
      The attributes and values vectors don't match
    - 4 Permission Denied
*/
void MDFCOracleServer::setAttr(const std::string &path,
			       std::vector< std::string > &attributes,
			       std::vector< std::string > &values)
{
  if(attributes.size() != values.size()) {
    out.append("3 Illegal command\n");
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string pattern;
  bool isPattern=sqlPattern(p.name, pattern);
  if( p.type() ) {
    isPattern = true;
    pattern="%";
  }

  std::list< std::string > groups;
  int perms=checkPermissions(p, 'w', groups);
  if(perms <= 0){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::string query;
  Statement statement(*dbConn);
  query = "UPDATE ";
  query += p.directoryTable;
  query.append(" SET ");
  for(unsigned int i=0; i<attributes.size(); i++){
    if(parser->parse(attributes[i], keys[i])){
      out.append("7 Illegal Key\n");
      return;
    }
    query.append(p.directoryTable).append(".");
    query.append(keys[i]);
    query.append("='");
    query.append(values[i]).append("'");
    if(i<attributes.size()-1)
      query.append(", ");
  }
  query += " WHERE EXISTS (SELECT ";
  query += p.indexTable + ".name FROM ";
  query += p.indexTable + " WHERE ";
  if(! (isPattern && pattern == "%")  ){
    query += p.indexTable + ".name";
    if(isPattern)
      query += " LIKE '";
    else
      query += " = '";
    query += pattern + "'";
    query += " AND ";
  }
  query += p.indexTable + ".fileid = ";
  query += p.directoryTable + ".\"entry\");";

  DMESG("SQL: >" << query << "<" << std::endl);

  if(statement.exec(query)){
    int err = statement.getErrorCode();
    std::string msg = statement.getError();
    std::string sqlErr = statement.getSqlState();    
    DMESG(err << "msg >" << msg << "< sqlErr: " << sqlErr
	  << std::endl);
    if(statement.numRows()==0) {
      printError("1 No such file", statement);
      return;
    }
    out.append("10 No such key: ");
    out.append(msg);
    out.append("\n");
    return;
  }

  out.append("0\n");
}


/** Clears an \param attribute of files matching \param path .
    The output is streamed into the out class member. For a
    description see the protocol description.
*/
void MDFCOracleServer::clearAttr(const std::string &path,
				 const std::string &attribute)
{
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string key;
  if(parser->parse(attribute, key)){
    out.append("7 Illegal Key\n");
    return;
  }

  std::string pattern;
  bool isPattern = sqlPattern(p.name, pattern);
  if( p.type() ) {
    isPattern = true;
    pattern="%";
  }

  std::list< std::string > groups;
  int perms=checkPermissions(p, 'w', groups);
  if(perms <= 0){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  std::string query="UPDATE ";
  query.append(p.directoryTable);
  query.append(" SET ").append(p.directoryTable).append(".").append(key);
  query.append("= NULL");
  query += " WHERE EXISTS (SELECT ";
  query += p.indexTable + ".name FROM ";
  query += p.indexTable + " WHERE ";
  if(! (isPattern && pattern == "%")  ){
    query += p.indexTable + ".name";
    if(isPattern)
      query += " LIKE '";
    else
      query += " = '";
    query += pattern + "'";
    query += " AND ";
  }
  query += p.indexTable + ".fileid = ";
  query += p.directoryTable + ".\"entry\");";

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("2", statement);
    return;
  }

  out.append("0\n");
}


std::string MDFCOracleServer::updateAttrQuery(const EntryProps &p,
					      const std::string &entries,
					      bool isPattern,
					      std::vector< std::string > &attributes,
					      std::vector< std::string > &values,
					      const std::string &condition,
					      bool eval)
{
  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::string query="UPDATE ";
  query.append(p.directoryTable);
  query.append(" SET ");
  for(unsigned int i=0; i<attributes.size(); i++){
    if(parser->parse(attributes[i], keys[i])){
      out.append("7 Illegal Key\n");
      return "";
    }
    std::string value;
    if(parser->parse(values[i], value)){
      out.append("9 Illegal value\n");
      return "";
    }
    query.append(p.directoryTable).append(".");
    query.append(keys[i]);
    query.append("=");
    query.append(value).append("");
    if(i<attributes.size()-1)
      query.append(", ");
  }
  query += " WHERE EXISTS (SELECT ";
  query += p.indexTable + ".name FROM ";
  query += p.indexTable + " WHERE ";
  if(! (isPattern && entries == "%")  ){
    query += p.indexTable + ".name";
    if(isPattern)
      query += " LIKE '";
    else
      query += " = '";
    query += entries + "'";
    query += " AND ";
  }
  query += p.indexTable + ".fileid = ";
  query += p.directoryTable + ".\"entry\"";
  if(condition.size())
    query += ((std::string)" AND (") + condition + ")";
  query += ");";

  return query;
}

