/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDInterpreter.cc,v 1.100 2011/02/07 02:00:43 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDInterpreter.h"
#include "MDStandalone.h"
#include "protocolHandling.h"

#include <cstdlib>
#include <climits>
#include <ctype.h>
#include <cassert>
#include <typeinfo>

using namespace std;

static bool debug = false;
#include "debug.h"

const int MDInterpreter::CommandDesc::MAX_ARGS = SHRT_MAX;

/**
 * If minArgs and maxArgs are not initialized, then this object will allow
 * any number of arguments when hasValidArgsSize is called. If only minArgs
 * is provided it will make maxArgs equal to that value, so that only
 * that number of args is accepted. If maxArgs is also given, then
 * the number of args must be between the two values. 
 */
MDInterpreter::CommandDesc::CommandDesc(MDInterpreter::CommandToken tokenInit, short int minArgsInit, short int maxArgsInit)
:  token(tokenInit), minArgs(minArgsInit), maxArgs(maxArgsInit) 
{
	// If only minArgs is provided, then it is assumed that it is also
	// the maximum number of arguments
	if (minArgs != -1 && maxArgs == -1) {
		maxArgs = minArgs;
	}
	assert(minArgs <= maxArgs);
}
	
bool MDInterpreter::CommandDesc::hasValidArgsSize(std::vector<std::string> &cmdTokens) 
const 
{
	if (minArgs == -1) {
		// Was not initialized, so accept any number of arguments.
		return true;
	}
	short int argsSize = cmdTokens.size() - 1;
	if (argsSize < minArgs) {
		return false;
	}
	// Now check the maximum number of arguments
	return maxArgs == -1 || argsSize <= maxArgs;  
}

MDInterpreter::MDInterpreter()
  : uploadHandle(NULL)
{
  // Init the command map
  typedef map< string, CommandDesc >::value_type valType;
  // Replication  
  cmdMap.insert(valType("rep_list_mounts", CommandDesc(cRepListMounts, 0)));
  cmdMap.insert(valType("rep_list_subscribers", CommandDesc(cRepListSubscribers, 0)));
  cmdMap.insert(valType("rep_show_permissions", CommandDesc(cRepShowPermissions, 0)));
  cmdMap.insert(valType("rep_umount_users", CommandDesc(cRepUmountUsers, 0)));
  cmdMap.insert(valType("rep_umount", CommandDesc(cRepUmount, 1,2)));
  cmdMap.insert(valType("rep_start_receive", CommandDesc(cRepStartReceive, 1)));
  cmdMap.insert(valType("rep_stop_receive", CommandDesc(cRepStopReceive, 1)));
  cmdMap.insert(valType("rep_mount_users", CommandDesc(cRepMountUsers, 1)));
  cmdMap.insert(valType("rep_start_receive", CommandDesc(cRepStartReceive, 1)));
  cmdMap.insert(valType("rep_allow", CommandDesc(cRepAllow, 2)));
  cmdMap.insert(valType("rep_disallow", CommandDesc(cRepDisallow, 2)));
  cmdMap.insert(valType("rep_users_allow", CommandDesc(cRepUsersAllow, 1)));
  cmdMap.insert(valType("rep_users_disallow", CommandDesc(cRepUsersDisallow, 1)));

  cmdMap.insert(valType("rep_mount", CommandDesc(cRepMount, 2, 3)));
  cmdMap.insert(valType("rep_drop_subscriber", CommandDesc(cRepDropSubscriber, 1)));
  cmdMap.insert(valType("fed_mount", CommandDesc(cProxyMount, 2, 3)));
  cmdMap.insert(valType("fed_umount", CommandDesc(cProxyUmount, 1)));
  cmdMap.insert(valType("fed_list_mounts", CommandDesc(cProxyListMounts, 0)));
  cmdMap.insert(valType("fed_policy", CommandDesc(cProxyPolicy, 0)));
  cmdMap.insert(valType("fed_check_failure", CommandDesc(cProxyCheckFailure, 1)));

  cmdMap.insert(valType("abort", CommandDesc(cAbort, 0)));
  cmdMap.insert(valType("acl_add", CommandDesc(cAclAdd, 3)));
  cmdMap.insert(valType("acl_remove", CommandDesc(cAclRemove, 2)));
  cmdMap.insert(valType("acl_show", CommandDesc(cAclShow, 1)));
  cmdMap.insert(valType("addattr", CommandDesc(cAddAttr, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("addentries", CommandDesc(cAddEntries, 1, CommandDesc::CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("addentry", CommandDesc(cAddEntry, 3, CommandDesc::MAX_ARGS)));
  
  cmdMap.insert(valType("backend", CommandDesc(cBackend, 0)));

  cmdMap.insert(valType("cd", CommandDesc(cCd, 1)));
  cmdMap.insert(valType("chmod", CommandDesc(cChmod, 2)));
  cmdMap.insert(valType("chown", CommandDesc(cChown, 2, 3)));
  cmdMap.insert(valType("clearattr", CommandDesc(cClearAttr, 2)));
  cmdMap.insert(valType("close", CommandDesc(cClose, 0)));
  cmdMap.insert(valType("commit", CommandDesc(cCommit, 0)));

  cmdMap.insert(valType("constraint_add_check", CommandDesc(cConstraintAddCheck,3 )));
  cmdMap.insert(valType("constraint_add_not_null", CommandDesc(cConstraintAddNotNull, 3)));
  cmdMap.insert(valType("constraint_add_reference", CommandDesc(cConstraintAddReference, 4)));
  cmdMap.insert(valType("constraint_add_unique", CommandDesc(cConstraintAddUnique, 3)));
  cmdMap.insert(valType("constraint_add_primary_key", CommandDesc(cConstraintAddPrimaryKey, 2)));
  cmdMap.insert(valType("constraint_drop", CommandDesc(cConstraintDrop, 2)));
  cmdMap.insert(valType("constraint_list", CommandDesc(cConstraintList, 1)));
 
  cmdMap.insert(valType("createdir", CommandDesc(cCreateDir, 1, 2)));
  cmdMap.insert(valType("makedir", CommandDesc(cMkDir, 1, 2)));
  cmdMap.insert(valType("dir", CommandDesc(cDir, 0, 1)));
  cmdMap.insert(valType("dump", CommandDesc(cDump, 1, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("execute", CommandDesc(cExecute, 1, 10)));
  cmdMap.insert(valType("exit", CommandDesc(cExit, 0)));
  cmdMap.insert(valType("find", CommandDesc(cFind, 1, 2)));
  cmdMap.insert(valType("getattr", CommandDesc(cGetAttr, 2, CommandDesc::MAX_ARGS)));
 
  cmdMap.insert(valType("grp_adduser", CommandDesc(cGrpAddUser, 2)));
  cmdMap.insert(valType("grp_create", CommandDesc(cGrpCreate, 1)));
  cmdMap.insert(valType("grp_delete", CommandDesc(cGrpDelete, 1)));
  cmdMap.insert(valType("grp_dump", CommandDesc(cGrpDump, 0)));
  cmdMap.insert(valType("grp_list", CommandDesc(cGrpList, 0, 1)));
  cmdMap.insert(valType("grp_member", CommandDesc(cGrpMember, 0, 1)));
  cmdMap.insert(valType("grp_removeuser", CommandDesc(cGrpRemoveUser, 2)));
  cmdMap.insert(valType("grp_show", CommandDesc(cGrpShow, 1)));
  cmdMap.insert(valType("grp_voms_add", CommandDesc(cGrpVomsAdd, 2)));
  cmdMap.insert(valType("grp_voms_remove", CommandDesc(cGrpVomsRemove, 1)));
  cmdMap.insert(valType("grp_voms_list", CommandDesc(cGrpVomsList, 0)));
 
  cmdMap.insert(valType("capabilities_list", CommandDesc(cCapabilitiesList, 0, 1)));
  cmdMap.insert(valType("capabilities_add", CommandDesc(cCapabilitiesAdd, 2)));
  cmdMap.insert(valType("capabilities_remove", CommandDesc(cCapabilitiesRemove, 2)));

  cmdMap.insert(valType("help", CommandDesc(cHelp, 0, 1)));
  cmdMap.insert(valType("index_create", CommandDesc(cIndexCreate, 3, 4)));
  cmdMap.insert(valType("index_remove", CommandDesc(cIndexRemove, 1)));
  cmdMap.insert(valType("insert", CommandDesc(cInsert, 3, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("import", CommandDesc(cImport, 2, 3)));
  cmdMap.insert(valType("lfn_lookup", CommandDesc(cLFNLookup, 1)));
  cmdMap.insert(valType("link", CommandDesc(cLink, 2)));
  cmdMap.insert(valType("listattr", CommandDesc(cListAttr, 1)));
  cmdMap.insert(valType("listentries", CommandDesc(cListEntries, 0, 1)));
  cmdMap.insert(valType("ls", CommandDesc(cLS, 0, 2)));
  cmdMap.insert(valType("name", CommandDesc(cName, 0)));
  cmdMap.insert(valType("put", CommandDesc(cPut, 1, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("pwd", CommandDesc(cPwd, 0)));
  cmdMap.insert(valType("quit", CommandDesc(cExit, 0)));
  cmdMap.insert(valType("removeattr", CommandDesc(cRemoveAttr, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("renameattr", CommandDesc(cRenameAttr, 3)));

  cmdMap.insert(valType("removeentries", CommandDesc(cRemoveEntries, 1, CommandDesc::MAX_ARGS))); 

  cmdMap.insert(valType("replica_add", CommandDesc(cReplicaAdd, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("replica_get", CommandDesc(cReplicaGet, 1, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("replica_list", CommandDesc(cReplicaList, 1, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("replica_register", CommandDesc(cReplicaRegister, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("replica_remove", CommandDesc(cReplicaRemove, 1, 2)));
  cmdMap.insert(valType("replica_unregister", CommandDesc(cReplicaUnRegister, 2, CommandDesc::MAX_ARGS)));
 
  cmdMap.insert(valType("rm", CommandDesc(cRm, 1, 2)));
  cmdMap.insert(valType("rmdir", CommandDesc(cRmDir, 1)));
  cmdMap.insert(valType("schema_cp", CommandDesc(cSchemaCp, 2)));
  cmdMap.insert(valType("schema_create", CommandDesc(cSchemaCreate, 1, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("schema_remove", CommandDesc(cSchemaRemove, 1, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("selectattr", CommandDesc(cSelectAttr, 2, CommandDesc::MAX_ARGS)));
 
  cmdMap.insert(valType("sequence_create", CommandDesc(cSequenceCreate, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("sequence_current", CommandDesc(cSequenceCurrent, 1)));
  cmdMap.insert(valType("sequence_next", CommandDesc(cSequenceNext, 1)));
  cmdMap.insert(valType("sequence_remove", CommandDesc(cSequenceRemove, 1)));

  // Site management
  cmdMap.insert(valType("site_add", CommandDesc(cSiteAdd, 1, 2)));
  cmdMap.insert(valType("site_remove", CommandDesc(cSiteRemove, 1)));
  cmdMap.insert(valType("site_list", CommandDesc(cSiteList, 0)));
  cmdMap.insert(valType("site_dump", CommandDesc(cSiteDump, 0, CommandDesc::MAX_ARGS)));  
  cmdMap.insert(valType("site_set_properties", CommandDesc(cSiteSetProperties, 3, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("site_get_properties", CommandDesc(cSiteGetProperties, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("site_list_properties", CommandDesc(cSiteListProperties, 1)));

  cmdMap.insert(valType("setattr", CommandDesc(cSetAttr, 3, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("stat", CommandDesc(cStat, 1)));
  cmdMap.insert(valType("ticket_check", CommandDesc(cTicketCheck, 1)));
  cmdMap.insert(valType("ticket_get", CommandDesc(cTicketGet, 0)));
  cmdMap.insert(valType("transaction", CommandDesc(cTransaction)));
  cmdMap.insert(valType("updateattr", CommandDesc(cUpdateAttr, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("update", CommandDesc(cUpdate, 2, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("upload", CommandDesc(cUpload, 1, CommandDesc::MAX_ARGS)));

  cmdMap.insert(valType("user_create", CommandDesc(cUserCreate, 1, 3)));
  cmdMap.insert(valType("user_dump", CommandDesc(cUserDump, 0)));
  cmdMap.insert(valType("user_list", CommandDesc(cUserList, 0)));
  cmdMap.insert(valType("user_listcred", CommandDesc(cUserCredList, 1)));
  cmdMap.insert(valType("user_password_change", CommandDesc(cUserPassChange, 2)));
  cmdMap.insert(valType("user_remove", CommandDesc(cUserRemove, 1)));
  cmdMap.insert(valType("user_subject_add", CommandDesc(cUserSubjectAdd, 2)));
  cmdMap.insert(valType("user_subject_remove", CommandDesc(cUserSubjectRemove, 2)));
  cmdMap.insert(valType("user_voms_add", CommandDesc(cUserVomsAdd, 2)));
  cmdMap.insert(valType("user_voms_remove", CommandDesc(cUserVomsRemove, 1)));
  cmdMap.insert(valType("user_voms_list", CommandDesc(cUserVomsList, 0)));

  cmdMap.insert(valType("vacuum", CommandDesc(cVacuum, 0)));
  cmdMap.insert(valType("version", CommandDesc(cVersion, 0)));
  cmdMap.insert(valType("view_create", CommandDesc(cViewCreate, 4, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("view_create_union", CommandDesc(cViewCreateUnion, 4, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("view_remove", CommandDesc(cViewRemove, 1)));
  cmdMap.insert(valType("whoami", CommandDesc(cWhoAmI, 0)));
  cmdMap.insert(valType("sudo", CommandDesc(cSudo, 1)));

  cmdMap.insert(valType("SELECT", CommandDesc(cSELECT,0,CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("INSERT", CommandDesc(cINSERT,0,CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("DELETE", CommandDesc(cDELETE, 0, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("UPDATE", CommandDesc(cUPDATE, 0, CommandDesc::MAX_ARGS)));
  cmdMap.insert(valType("updateattr_single", CommandDesc(cUpdateSingle, 2, CommandDesc::MAX_ARGS)));

}


MDInterpreter::~MDInterpreter()
{
  // Nothing to be deconstructed
}


/** Removes the quotes from the beginning and end of the string.
 * It does not remove the \' 
 * 
 */
void MDInterpreter::removeQuotes(std::string &s)
{
  if(s[0] == '\'')
    s.erase(0, 1);
  int p=s.size()-1;
  if( p > -1 && s[p]=='\'' ){
      s.erase(p);
  }
}


/** Handle escape sequences for single quotes
    Replaces all appearances of "\'" in the string by "'"
*/
void MDInterpreter::handleOctalEscapes(std::string &s)
{
  size_t pos = 0;
  while (1) {
    pos = s.find("\\", pos);
    if(pos == std::string::npos)
      return;
    if( pos < s.size()-3 
	&& isdigit(s[pos+1]) && isdigit(s[pos+2]) && isdigit(s[pos+3]) ){
      unsigned char buf[2];
      buf[0]='\0'; buf[1]='\0'; 
      for( size_t i=1; i < 4; i++) {
	buf[0] *= 8;
	buf[0] += s[pos+i]-'0';
      }
      s.replace(pos, 4, (const char *)buf);
    }
    pos++;
  }
}


/** Returns the first part of a command string and cuts this part off.
    \param in Input command string. First part gets cut off.
    \param part Returns the identified part with any quotes removed
    and escape sequences handled
    \return Returns the size of the part found or -1 if there was a
    parsing error.
*/
int MDInterpreter::parsePart(const std::string &in, unsigned int &pos, std::string &part)
{  
  int escapeLevel=0;
  part="";

  // Skip whitespace
  size_t size = in.size();
  while(pos<size && (in[pos]==' ' || in[pos]=='\t' || in[pos]==13))
    pos++;

  unsigned int begin=pos;
  while(pos<size){
    if( (in[pos]==' ' || in[pos]=='\t' || in[pos]==13) && !escapeLevel)
      break;
    if( in[pos]=='\''){
      if( escapeLevel == 0 ) {
        escapeLevel = 1;
      } else {
        if( in[pos+1] == '\'' ){
          pos++;
        } else {
          escapeLevel = 0;
          pos++;
          break;
        }
      }
    }
    pos++;
  }

  if(escapeLevel){
    ERRLOG("Unbalanced quotes: " << in << "  " << pos << std::endl);
    return -1;
  }

  part=in.substr(begin, pos-begin);

  // Consider '' not to be an empty string, even if it becomes one
  int partSize=part.size();
  removeQuotes(part);
  handleOctalEscapes(part);
  //unescapeString(part);

  return partSize;
}


int MDInterpreter::parseAll(const std::string &line,
			    std::vector< std::string > &parts)
{
  parts.clear();
  int error;
  unsigned int pos = 0;
  string part;
  while( (error = parsePart(line, pos, part)) >0){
    parts.push_back(part);
    DMESG("Part : " << part << " \n");
  }

  return error;
}


MDInterpreter::CommandToken MDInterpreter::parseCommand (const std::string &line)
{
  if( !line.size() )
    return cInvalid;

  inputLine = line ;


  // Check for a comment
  if (!line.empty() && line.at(0) == '#') {
    // It's a comment
    command.resize(1);
    command[0]="";
    cmdToken = cComment;

  } else if (parseAll(line, command)) {
    cmdToken = cInvalid;

  } else {     
    map<string, CommandDesc>::const_iterator it = cmdMap.find(command[0]);
    if (it == cmdMap.end()) {
      // Command not found
      cmdToken = cInvalid;
    } else {
      if (!it->second.hasValidArgsSize(command)) {
        // Invalid number of arguments
        cmdToken = cInvalidNumberOfArgs;
      } else {  		
      	cmdToken = it->second.token; 
      }
    }
  }
  return cmdToken; 
}


int MDInterpreter::getOptions(std::vector< std::string > &cString,
			      std::string &options) const
{
  if( cString.size() ) {
    cString.erase(cString.begin());
    if( cString[0].size() && cString[0][0]=='-') {
      options = cString[0];
      cString.erase(cString.begin());
    }
  }
  return options.size() > 0? 1 : 0;
}

void splitAttrsValues(vector<string>::const_iterator start, vector<string>::const_iterator end, 
    vector<string> &attrs, vector<string> &values) 
{
  if ((end-start) % 2 != 0) {
    throw runtime_error("Invalid number of arguments");
  }
    size_t len = (end - start)/2;
    attrs.clear();
    values.clear();    
    for(vector<string>::const_iterator it=start; it != end; ){
        attrs.push_back(*it);
        ++it;
        values.push_back(*it);
        ++it;
    }
}

int MDInterpreter::execLast(MDServer &connection) {
	return exec(cmdToken, command, connection);
}

/**
 * Returns 0 if the command is valid and its execution was attempted,
 * 1 if the command is not valid. A return of 0 means only that the
 * command is valid, not that it executed correctly. For that, check
 * the data stored on the MDBuffer.
 */
int MDInterpreter::exec(MDInterpreter::CommandToken c,
			std::vector< std::string > &cString,
			MDServer &connection)
{  
  attributes.clear();
  values.clear(); 
  attrList.clear();
  
//  for (int i=0; i<cString.size(); i++) {
//    cout << cString[i] << endl;
//  }

  // for federation use 2010.06
  connection.cmdToken = c;
    
  try {
    std::string options;
    bool eval = false;
    
//    cout << "Command: " << c << " " << cString[0] << endl;  
  
    switch(c) {
    case MDInterpreter::cInvalid:
      cout << "Sending invalid command\n";
      // Send a descriptive error message to the client.
      connection.sendToClient("9 Invalid command.\n");
      break;
      
    case MDInterpreter::cInvalidNumberOfArgs:
      // Send a descriptive error message to the client.
      connection.sendToClient("9 Invalid command. Wrong number of arguments\n");
      break;
      
    case MDInterpreter::cComment:
      // We must answer to the client with an OK.
      connection.sendToClient("0\n");
      break;
      
    case MDInterpreter::cHelp:
      if( cString.size() == 2 )
        connection.help(cString[1]);
      else
        connection.help();
      break;
      
    case MDInterpreter::cUpload: {
      uploadHandle.reset(new UploadHandle());
    	vector<string> attrs(cString.begin()+2, cString.end());
      if (connection.upload(*uploadHandle, cString[1], attrs)) {
      	// Failed
        uploadHandle->aborted = true;        
        return 1;
      }
      break;
    }
  
    case MDInterpreter::cPut: {
      if(uploadHandle.get() == NULL) {
        connection.sendToClient("9 Invalid command. No upload in progress.\n");
        return 1;
      }
      if (uploadHandle->aborted) {
        connection.sendToClient("9 Invalid command. Upload was aborted.\n");
        return 1;
      }
      if (connection.put(*uploadHandle, cString.begin()+1, cString.end())) {
      	// Failed
        DMESG("Upload failed\n");
        uploadHandle->aborted = true;
        return 1;
      }
      break;
    }
  
    case MDInterpreter::cAbort:
      if(uploadHandle.get() != NULL) {
      	if (!uploadHandle->aborted) // Upload
	  connection.abort(*uploadHandle);
        uploadHandle.reset(NULL);
      } else { // Transactions
        connection.transactionAbort();
      }
      break;
  
    case MDInterpreter::cCommit:
      if(uploadHandle.get() != NULL) {
      	if (!uploadHandle->aborted) // Upload
	  connection.commit(*uploadHandle);
      	uploadHandle.reset(NULL);
      } else { // Transactions
	connection.transactionCommit();
      }
      break;
  
    case MDInterpreter::cGetAttr: {
    	vector<string> attrs(cString.begin()+2, cString.end());  	
      connection.getAttr(cString[1], attrs);
      break;
    }
  
    case MDInterpreter::cSetAttr: {
      splitAttrsValues(cString.begin()+2, cString.end(), attributes, values);    
      connection.setAttr(cString[1], attributes, values);
      break;
    }
  
    case MDInterpreter::cInsert:
      eval = true;
    case MDInterpreter::cAddEntry:
      splitAttrsValues(cString.begin()+2, cString.end(), attributes, values);
      connection.addEntry(cString[1], attributes, values, eval);
      break;
  
    case MDInterpreter::cUpdateAttr:
      splitAttrsValues(cString.begin()+2, cString.end()-1, attributes, values);
      connection.updateAttr(cString[1], attributes, values,
  			  cString[cString.size()-1]);
      break;

    case MDInterpreter::cUpdateSingle:
      splitAttrsValues(cString.begin()+2, cString.end()-1, attributes, values);
      connection.updateSingle(cString[1], attributes, values,
                          cString[cString.size()-1]);
      break;

  
    case MDInterpreter::cUpdate:
      splitAttrsValues(cString.begin()+2, cString.end()-1, attributes, values);
      connection.update(cString[1], attributes, values, cString[cString.size()-1]);
      break;
  
    case MDInterpreter::cListAttr:
      connection.listAttr(cString[1]);
      break;
  
    case MDInterpreter::cAddAttr:
      if(cString.size()%2) {  // Allow only pairs of attribute and type
        connection.sendToClient("9 Invalid command. Bad number of attributes.\n");
        return 1;
      }
      for(unsigned int i=2; i<cString.size()-1; i+=2)
        attrList.push_back(Attribute(cString[i], cString[i+1]));
      connection.addAttr(cString[1], attrList);
      break;
  
    case MDInterpreter::cClearAttr:
      connection.clearAttr(cString[1], cString[2]);
      break;
  
    case MDInterpreter::cRemoveAttr:
      for(unsigned int i=2; i<cString.size(); i++)
        attrList.push_back(Attribute(cString[i],""));
      connection.removeAttr(cString[1], attrList);
      break;
  
    case MDInterpreter::cRenameAttr:
      connection.renameAttr(cString[1], cString[2], cString[3]);
      break;

    case MDInterpreter::cFind:
      if(cString.size() >2)
        connection.find(cString[1], cString[2]);
      else
        connection.find(cString[1]);
      break;
  
    case MDInterpreter::cSelectAttr:  
      attributes.resize((cString.size()-2));
      for(unsigned int i=0; i<attributes.size(); i++)
        attributes[i]=cString[i+1];
      connection.selectAttr(attributes, cString[cString.size()-1]);
      break;
  
    case MDInterpreter::cAddEntries:
      values.resize((cString.size()-1));
      for(unsigned int i=0; i<values.size(); i++)
        values[i]=cString[i+1];
      connection.addEntries(values);
      break;
  
    case MDInterpreter::cRemoveEntries:
      values.resize((cString.size()-1));
      for(unsigned int i=0; i<values.size(); i++)
        values[i]=cString[i+1];
      connection.removeEntries(values);
      break;
  
    case MDInterpreter::cPwd:
      connection.pwd();
      break;
  
    case MDInterpreter::cWhoAmI:
      connection.whoAmI();
      break;
  
    case MDInterpreter::cBackend:
      connection.backend();
      break;

    case MDInterpreter::cIndexCreate: {
      string part=cString[3];
      removeQuotes(part);
      parseAll(part, attributes);
  
      if(cString.size() == 5)
        connection.createIndex(cString[1], cString[2], attributes, cString[4]);
      else 
        connection.createIndex(cString[1], cString[2], attributes);
      break;
    }
    case MDInterpreter::cIndexRemove:
      connection.indexRemove(cString[1]);
      break;
  
    case MDInterpreter::cTransaction:
      connection.transactionStart();
      break;
  
    case MDInterpreter::cViewCreate:
      attributes.resize((cString.size()-4));
      for(unsigned int i=0; i<attributes.size(); i++)
        attributes[i]=cString[i+3];
      connection.createView(cString[1], cString[2], attributes,
  			  cString[cString.size()-1]);
      break;
  
    case MDInterpreter::cViewCreateUnion:
      attributes.resize((cString.size()-3));
      for(unsigned int i=0; i<attributes.size(); i++)
        attributes[i]=cString[i+2];
      connection.createViewUnion(cString[1], attributes,
				 cString[cString.size()-1]);
      break;

    case MDInterpreter::cViewRemove:
      connection.viewRemove(cString[1]);
      break;
  
    case MDInterpreter::cSequenceRemove:
      connection.sequenceRemove(cString[1]);
      break;
      
    case MDInterpreter::cSequenceNext:
      connection.sequenceNext(cString[1]);
      break;
      
    case MDInterpreter::cSequenceCurrent:
      connection.sequenceCurrent(cString[1]);
      break;
  
    case MDInterpreter::cSequenceCreate:
      if(cString.size()==3)
        connection.createSequence(cString[1], cString[2], "1", "1");
      if(cString.size()==4)
        connection.createSequence(cString[1], cString[2], cString[3], "1");
      if(cString.size()==5)
        connection.createSequence(cString[1], cString[2], cString[3], cString[4]);
      break;
  
    case MDInterpreter::cSchemaCreate:
      for(unsigned int i=2; i<cString.size()-1; i+=2)
        attrList.push_back(Attribute(cString[i], cString[i+1]));
      if(cString.size()%2)
        connection.createSchema(cString[1], attrList, cString[cString.size()-1]);
      else
        connection.createSchema(cString[1], attrList);
      break;
  
    case MDInterpreter::cSchemaRemove:
      connection.removeSchema(cString[1]);
      break;
  
    case MDInterpreter::cDir:
      if(cString.size() ==2)
        connection.dir(cString[1]);
      else
        connection.dir();
      break;
  	
    case MDInterpreter::cListEntries:
      if(cString.size() ==2)
        connection.listEntries(cString[1]);
      else
        connection.listEntries();
      break;
    
    case MDInterpreter::cCd:
      connection.cd(cString[1]);
      break;
  
    case MDInterpreter::cConstraintAddUnique:
      connection.constraintAddUnique(cString[1], cString[2], cString[3]);
      break;
  
    case MDInterpreter::cConstraintAddNotNull:
      connection.constraintAddNotNull(cString[1], cString[2], cString[3]);
      break;
  
    case MDInterpreter::cConstraintAddCheck:
      connection.constraintAddCheck(cString[1], cString[2], cString[3]);
      break;
  
    case MDInterpreter::cConstraintAddReference:
      connection.constraintAddReference(cString[1], cString[2],
  				      cString[3], cString[4]);
      break;

    case MDInterpreter::cConstraintAddPrimaryKey:
      connection.constraintAddPrimaryKey(cString[1], cString[2]);
      break;

    case MDInterpreter::cConstraintDrop:
      connection.constraintDrop(cString[1], cString[2]);
      break;
  
    case MDInterpreter::cConstraintList:
      connection.constraintList(cString[1]);
      break;
  	
    default: // Now come the commands of a standalone metadata server
      try {
        MDStandalone &standalone = dynamic_cast<MDStandalone &>(connection);
  
        switch (c) {
        case MDInterpreter::cMkDir:
  	if(cString.size() == 2) {
  	  cString.resize(3);
  	}
  	cString[2] += ",files";
        case MDInterpreter::cCreateDir:
  	if(cString.size()>2)
  	  standalone.createDir(cString[1], cString[2]);
  	else
  	  standalone.createDir(cString[1]);
  	break;
  	
        case MDInterpreter::cLS:
  	getOptions(cString, options);
  	if( cString.size() )
  	  standalone.ls(options, cString[0]);
  	else
  	  standalone.ls(options);
  	break;
  
        case MDInterpreter::cStat:
  	standalone.stat(cString[1]);
  	break;
  	
        case MDInterpreter::cRm:
  	getOptions(cString, options);
  	
  	if( options.size() ) {
  	  if ( cString.size() == 2)
  	    standalone.rm(cString[0], cString[1], options);
  	  else
  	     standalone.rm(cString[0], "", options);
  	} else {
  	  if( cString.size() == 2 )
  	    standalone.rm(cString[0], cString[1]);
  	  else
  	    standalone.rm(cString[0]);
  	}
  	break;
  
  	
        case MDInterpreter::cRmDir:
	  standalone.rmDir(cString[1]);
	  break;
	  
        case MDInterpreter::cLink:
	  standalone.link(cString[1], cString[2]);
	  break;
	  
	case MDInterpreter::cImport:
	  if( cString.size() > 3 ) 
	    standalone.import(cString[1], cString[2], cString[3]);
	  else
	    standalone.import(cString[1], cString[2]);
	  break;
	  
	case MDInterpreter::cSudo:
	  standalone.sudo(cString[1]);
	  break;  
	  
        case MDInterpreter::cChown:
        {
        	bool rOption = false;
        	int argIndex = 1;
        	if( ( cString.size() > 3 ) && (cString[1][0] == '-') )
        		if( (cString[1].size() == 2) && (cString[1][1] == 'R') )  
        		{
        			rOption = true;
        			argIndex = 2;
        		}
        		else {
        			connection.sendToClient("9 Invalid option. recursive option is -R.\n");
        			return 1;
        		}
        	standalone.chown(cString[argIndex], cString[argIndex+1], rOption);	
	  	break;
		}
        case MDInterpreter::cChmod:
	  standalone.chmod(cString[1], cString[2]);
	  break;
	  
        case MDInterpreter::cGrpCreate:
	  standalone.createGroup(cString[1]);
	  break;
  	
        case MDInterpreter::cGrpDelete:
	  standalone.deleteGroup(cString[1]);
	  break;
  	
        case MDInterpreter::cGrpAddUser:
	  standalone.grpAddUser(cString[1], cString[2]);
	  break;
  	
        case MDInterpreter::cGrpRemoveUser:
	  standalone.grpRemoveUser(cString[1], cString[2]);
	  break;
  	
        case MDInterpreter::cGrpMember:
  	if(cString.size() == 2)
  	  standalone.grpMember(cString[1]);
  	else
  	  standalone.grpMember();
  	break;
  	
	case MDInterpreter::cGrpList:
	  getOptions(cString, options);
	  
	  if( options.size() ) {
	    bool all = false;
	    if( options.find("a") != std::string::npos )
	      all = true;
	    standalone.grpList("", all);
	  } else {
	    if(cString.size() == 1)
	      standalone.grpList(cString[0]);
	    else
	      standalone.grpList();
	  }
	  break;
  
        case MDInterpreter::cGrpShow:
  	standalone.grpShow(cString[1]);
  	break;

        case MDInterpreter:: cGrpVomsAdd:
        standalone.userVomsAdd(cString[1], cString[2],2);
        break;

        case MDInterpreter::cGrpVomsRemove:
        standalone.userVomsRemove(cString[1],2);
        break;

        case MDInterpreter::cGrpVomsList:
        standalone.userVomsList(2);
        break;

  	
        case MDInterpreter::cAclRemove:
  	standalone.aclRemove(cString[1], cString[2]);
  	break;
  	
        case MDInterpreter::cAclAdd:
  	standalone.aclAdd(cString[1], cString[2], cString[3]);
  	break;
  	
        case MDInterpreter::cAclShow:
	  standalone.aclShow(cString[1]);
	  break;
 	  
        case MDInterpreter::cCapabilitiesList:
	  if(cString.size() == 2)
	    standalone.capabilitiesList(cString[1]);
	  else
	    standalone.capabilitiesList();
	  break;
	  
        case MDInterpreter::cCapabilitiesAdd:
	  standalone.capabilitiesAdd(cString[1], cString[2]);
	  break;
     
        case MDInterpreter::cCapabilitiesRemove:
	  standalone.capabilitiesRemove(cString[1], cString[2]);
	  break;
  	
        case MDInterpreter::cSchemaCp:
	  standalone.schemaCp(cString[1], cString[2]);
	  break;
  	
        case MDInterpreter::cVacuum:
  	if(cString.size()==2)
  	  standalone.vacuum(cString[1]);
  	else
  	  standalone.vacuum("");
  	break;
  
        case MDInterpreter::cName:
  	standalone.getName();
  	break;
  
        case MDInterpreter::cUserCreate:
        { 
	  // Can have an -h option
	  bool passIsHash = false;
	  size_t argIndex;
	  if (cString[1] == "-h") {
	    passIsHash = true;
	    argIndex = 2;
	  } else {
	    // Password is not an hash
	    passIsHash = false;
	    argIndex = 1;
	  }
	  string user = cString[argIndex];
	  string password;
	  if (cString.size() > argIndex+1) {
	    password = cString[argIndex+1];
	  }
  	  standalone.userCreate(user, password, passIsHash);
  	break;
        }
        
        case MDInterpreter::cUserRemove:
  	standalone.userRemove(cString[1]);
  	break;
        
        case MDInterpreter::cUserPassChange:
  	standalone.userPassChange(cString[1], cString[2]);
  	break;
        
        case MDInterpreter:: cUserSubjectAdd:
  	standalone.userSubjectAdd(cString[1], cString[2]);
  	break;
        
        case MDInterpreter::cUserSubjectRemove:
  	standalone.userSubjectRemove(cString[1], cString[2]);
  	break;
  		
        case MDInterpreter:: cUserVomsAdd:
        standalone.userVomsAdd(cString[1], cString[2],1);
        break;

        case MDInterpreter::cUserVomsRemove:
        standalone.userVomsRemove(cString[1],1);
        break;

        case MDInterpreter::cUserVomsList:
        standalone.userVomsList(1);
        break;

        case MDInterpreter::cUserCredList:
  	if(cString.size() == 2)
  	  standalone.userListCred(cString[1]);
  	else
  	  standalone.userListCred();
  	break;
  
        case MDInterpreter::cUserList:
  	standalone.userList();
  	break;
  
        case MDInterpreter::cTicketGet:
  	standalone.ticketGet();
  	break;
  
        case MDInterpreter::cTicketCheck:
  	standalone.ticketCheck(cString[1]);
  	break;
   
	case MDInterpreter::cExecute:
	standalone.execute(cString);
        break;
	  
        case MDInterpreter::cDump: 
        {
	  string secopt;
	  size_t argIndex;
	  if (cString.size() >= 2 && cString[1][0] == '-') {
	    // There's an option. Get the rest of the value.
	    secopt = cString[1].substr(1);
	    argIndex = 2;
	  } else {
	    argIndex = 1;
	  }
	  // The option was removed. Check if there is an argument. Without
	  // arguments, dump dumps everything, starting at the root directory.
	  string dir = cString.size() == argIndex+1 ? cString[argIndex] : "/";
	  standalone.dump(dir, secopt);
	  break;
  	}
        case MDInterpreter::cGrpDump:
  	standalone.grpDump();
  	break;
          
        case MDInterpreter::cUserDump:
  	standalone.userDump();
  	break;

      // Federation commands
        case MDInterpreter::cProxyMount: 
          standalone.proxyMount(cString[1], cString[2]);
          break;

        case MDInterpreter::cProxyUmount:
          standalone.proxyUmount(cString[1]);
          break;

        case MDInterpreter::cProxyListMounts:
          standalone.proxyListMounts();
          break;

        case MDInterpreter::cProxyPolicy:
          standalone.proxyPolicy();
          break;

        case MDInterpreter::cProxyCheckFailure:
          standalone.proxyCheckFailure(cString[1]);
          break;

      //{Replication commands
        case MDInterpreter::cRepListMounts:
          standalone.listMounts();
          break;
  
        case MDInterpreter::cRepListSubscribers: {
        	standalone.listSubscribers();
          break;
        }
          
        case MDInterpreter::cRepMount: {
          size_t idx;
          // To preserve the old behaviour when no option is given
          string secOpt;
          if (cString[1][0] == '-') {
            secOpt = cString[1].substr(1);
            idx = 2;
          } else {
            idx = 1;
          }
          if (cString.size() != idx+2) {
            connection.sendToClient("9 Invalid command.\n");
          	return 1;
          }
          string master = cString[idx];
          string dir = cString[idx+1];
          standalone.mount(master, dir, secOpt);
          break;
        }
        case MDInterpreter::cRepUmountAll: 
          standalone.umountAll();
          break;       
          
        case MDInterpreter::cRepUmount: {
          size_t idx;
          // To preserve the old behaviour when no option is given
          string secOpt;
          if (cString[1][0] == '-') {
            secOpt = cString[1].substr(1);
            idx = 1;
          } else {
            idx = 0;
          }
          if (cString.size() != idx+2) {
            connection.sendToClient("9 Invalid command.\n");
                return 1;
          }
          string dir = cString[idx+1];
          if ( secOpt == "noerase" ) 
            standalone.umount(dir, false);
          else  
            standalone.umount(dir, true);
          break;
        }

        case MDInterpreter::cRepMountUsers:
          standalone.mountUsers(cString[1]);
          break;
          
        case MDInterpreter::cRepUmountUsers:
          standalone.umountUsers();
          break;
          
  
        case MDInterpreter::cRepSynchronize:
          standalone.synchronize(cString[1]);
          break;
          
        case MDInterpreter::cRepStartReceive:
          standalone.startReceiveLogs(cString[1]);
          break;
  
        case MDInterpreter::cRepStopReceive:
          standalone.stopReceiveLogs(cString[1]);
          break;
  
        case MDInterpreter::cRepAllow: 
          standalone.repAllow(cString[1], cString[2]);
          break;
  
        case MDInterpreter::cRepDisallow:
          standalone.repDisallow(cString[1], cString[2]);
          break;
  
	case MDInterpreter::cRepUsersAllow: 
          standalone.repUsersAllow(cString[1]);
          break;
  
        case MDInterpreter::cRepUsersDisallow:
          standalone.repUsersDisallow(cString[1]);
          break;

        case MDInterpreter::cRepShowPermissions:
          standalone.repShowPermissions();
          break;
          
        case MDInterpreter::cRepDropSubscriber:
          standalone.repDropSubscriber(cString[1]);
          break;
  
        case MDInterpreter::cReplicaAdd:
  	if( cString.size() == 3 )
  	  standalone.replicaAdd(cString[1], cString[2]);
  	else
  	  standalone.replicaAdd(cString[1], cString[2], cString[3]);
  	break;
  
        case MDInterpreter::cReplicaRemove:
        if( cString.size() == 3 )
    	  standalone.replicaRemove(cString[1], cString[2]);
        else
          standalone.replicaRemove(cString[1]);
  	break;
  	
        case MDInterpreter::cReplicaRegister: {
	  getOptions(cString, options);
	  std::vector< std::string > guids, sites;
	  guids.resize(cString.size()/2);
	  sites.resize(cString.size()/2);
	  for(unsigned int i=0; i<guids.size(); i++) {
	    guids[i]=cString[i*2];
	    sites[i]=cString[i*2+1];
	  }
	  standalone.replicaRegister(options, guids, sites);
	  break;
	}

  	case MDInterpreter::cReplicaUnRegister: {
	  getOptions(cString, options);
	  std::vector< std::string > guids, sites;
	  guids.resize(cString.size()/2);
	  sites.resize(cString.size()/2);
	  for(unsigned int i=0; i<guids.size(); i++) {
	    guids[i]=cString[i*2];
	    sites[i]=cString[i*2+1];
	  }
	  standalone.replicaUnregister(options, guids, sites);
	  break;
	}
  	
        case MDInterpreter::cReplicaList:
  	getOptions(cString, options);
  	standalone.replicaList(options, cString);
  	break;	
  
        case MDInterpreter::cReplicaGet:
  	getOptions(cString, options);
  	standalone.replicaGet(options, cString);
  	break;	
  	
        case MDInterpreter::cLFNLookup:
  	getOptions(cString, options);
  	standalone.lfnLookup(cString);
  	break;	
  	
        // Site Management
        case MDInterpreter::cSiteAdd:
	  if( cString.size()==3 )
	    standalone.siteAdd(cString[1], cString[2]);
	  else
	    standalone.siteAdd(cString[1], "");
  	break;
  
        case MDInterpreter::cSiteRemove:
  	standalone.siteRemove(cString[1]);
  	break;	
  	
	case MDInterpreter::cSiteList:
	  standalone.siteList();
	  break;	
      
	case MDInterpreter::cSiteDump: {
          vector<string> sites(cString.begin()+1, cString.end());
          standalone.siteDump(sites);       
          break;
	}
        case MDInterpreter::cSiteListProperties: {
          standalone.siteListProperties(cString[1]);
          break;
        }
        case MDInterpreter::cSiteSetProperties: {
          // args: site, property names, property values
          splitAttrsValues(cString.begin()+2, cString.end(), attributes, values);
          standalone.siteSetProperties(cString[1], attributes, values);
          break;
        }
        case MDInterpreter::cSiteGetProperties: {
          // args: site, property names
          vector<string> attrs(cString.begin()+2, cString.end());
          standalone.siteGetProperties(cString[1], attrs);
          break;
        }
  	
      case MDInterpreter::cSELECT: {
          standalone.sqlExec(inputLine, MDInterpreter::cSELECT);
          break;
      }

      case MDInterpreter::cINSERT: {
          standalone.sqlExec(inputLine, MDInterpreter::cINSERT);
          break;
      }

      case MDInterpreter::cDELETE: {
          standalone.sqlExec(inputLine, MDInterpreter::cDELETE);
          break;
      }

      case MDInterpreter::cUPDATE: {
          standalone.sqlExec(inputLine, MDInterpreter::cUPDATE);
          break;
      }

      case MDInterpreter::cVersion: {
        standalone.getVersion();
        break;
      }

      // Replication commands}
  
        default: {
          connection.sendToClient("9 Unknown command.\n");
  	      return 1;
        }
        }
        return 0; // Was a standalone server command
      }
      catch (std::bad_cast &ex) {
        connection.sendToClient(string("9 Internal error: ") + ex.what() + "\n");
        return 1; // Was not a standalone server command
      }
    }
    return 0; // Was normal metadata server command
  } catch (runtime_error &ex) {
    connection.sendToClient(string("9 Internal error: ") + ex.what() + "\n");
    return 1;
  }
}

/**
 * Cancels the current upload. Useful for when an upload must be
 * interrupted due to some other error.
 */
int MDInterpreter::abortUpload(MDServer &connection) 
{
	// Delete the current upload handle, if any.
	uploadHandle.reset(NULL);
	return 0;
}


#define FED_NONE   0
#define FED_ALL    1
#define FED_RANDOM 2
#define FED_ONCE   4
#define FED_SELE   8

int MDInterpreter::findType(int cmdToken) 
{
  switch (cmdToken) {
    case cChmod:
    case cChown:
    case cRemoveEntries:
    case cUpdateAttr:
    case cUpdate:
    case cSetAttr:
    case cDELETE:
    case cUPDATE:
    case cAclRemove: 
    case cAclAdd:
    case cRm:
      return FED_ALL + FED_ONCE;
    case cGetAttr:
    case cFind:
    case cListEntries:
    case cSelectAttr:
    case cLS:
    case cDir:
    case cAclShow:
      return FED_ALL ;
    case cSELECT:
      return FED_ALL + FED_SELE ;
    case cAddEntry:
    case cAddEntries:
    case cInsert:
    case cINSERT:
      return FED_RANDOM + FED_ONCE;
    case cListAttr:
    case cStat:
      return FED_RANDOM;
    case cProxyUmount:
      return -1;
    default : 
      return FED_NONE;
  }
}
