/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
// $Id: MDInterpreter.h,v 1.49 2010/11/02 02:10:01 asunil Exp $
//
#ifndef MDINTERPRETER_H
#define MDINTERPRETER_H

#include <vector>
#include <string>
#include <memory>
#include <map>
#include "util/Utils.h"
#include "server/MDServer.h"

class MDServer;
class UploadHandle;

class MDInterpreter {
 public:
  enum CommandToken {cExit, cClose, cHelp,
		     cSchemaCreate, cSchemaRemove,
		     cGetAttr, cSetAttr, cListAttr, cClearAttr, cRenameAttr,
		     cRemoveAttr, cAddAttr, cFind, cCreateDir, cAddEntries,
		     cRemoveEntries, cDir, cRm, cRmDir, cLink, cStat,
		     cListEntries, cMkDir, cLS,
		     cAddEntry, cInsert, cChmod, cChown, cCd, cPwd, cName,
		     cVersion, cWhoAmI,
		     cViewCreate, cViewCreateUnion, cIndexCreate, cIndexRemove, cViewRemove,
		     cSequenceCreate, cSequenceRemove,
		     cSequenceNext, cSequenceCurrent,

		     cConstraintAddUnique, cConstraintAddNotNull,
		     cConstraintAddCheck, cConstraintAddReference,
		     cConstraintList, cConstraintDrop,
		     cConstraintAddPrimaryKey,

		     cGrpCreate, cGrpDelete, cGrpAddUser, cGrpRemoveUser,
		     cGrpMember, cGrpShow, cGrpList,
		     cAclShow, cAclRemove, cAclAdd,
		     cSchemaCp, cSelectAttr, cUpdateAttr, cUpdate,
		     cUpload, cPut, cCommit, cAbort, cVacuum,
		     cUserCreate, cUserRemove, cUserPassChange,
		     cUserSubjectAdd, cUserSubjectRemove, cUserList,
		     cUserVomsAdd, cUserVomsRemove, cUserVomsList,
		     cGrpVomsAdd, cGrpVomsRemove, cGrpVomsList,
		     cUserCredList,
		     cCapabilitiesList, cCapabilitiesAdd,
		     cCapabilitiesRemove,
		     cTicketGet, cTicketCheck,
		     cTransaction,
		     cDump, cGrpDump, cUserDump, cBackend,
		     cImport,
		     cExecute,
		     cSudo,
		     
                     cSQL, cSELECT, cINSERT, cUPDATE, cDELETE, cUpdateSingle,
         
		     // Replication
		     cRepListMounts, cRepMount,
		     cRepUmount, cRepUmountAll, cRepSynchronize,
		     cRepStartReceive, cRepStopReceive, cRepAllow,
		     cRepDisallow, cRepShowPermissions,
		     cRepMountUsers, cRepUmountUsers,
		     cRepListSubscribers, cRepDropSubscriber,
		     cRepUsersAllow, cRepUsersDisallow,

 		     // Federation
		     cProxyListMounts, cProxyMount, cProxyUmount, 
		     cProxyPolicy, cProxyCheckFailure, 
         
		     // Replica management commands 
		     cReplicaAdd, cReplicaRemove,
		     cReplicaRegister, cReplicaUnRegister,
		     cReplicaList, cReplicaGet, cLFNLookup,

		     // Site management		     		     
		     cSiteAdd, cSiteRemove, cSiteList, cSiteDump, 
		     cSiteSetProperties, cSiteGetProperties, cSiteListProperties,
		     // Others
		     cInvalid,
		     cInvalidNumberOfArgs,
		     cComment // It's not a command, but a comment
  };

  MDInterpreter();
  ~MDInterpreter();

  CommandToken parseCommand (const std::string &line);
  std::vector< std::string > getCommand() const { return command; };

  int execLast(MDServer &connection); 
  int exec(MDInterpreter::CommandToken c,
	   std::vector< std::string > &command,
	   MDServer &connection);

  int abortUpload(MDServer &connection);
  
  // Public visibility so it can be used by MDSoapServer
  static int parseAll(const std::string &line, std::vector< std::string > &parts);

  // federation
  static int findType(int cmdToken);


 private:
 
  struct CommandDesc {  	
	CommandDesc(MDInterpreter::CommandToken tokenInit,
		    short int minArgsInit=-1, short int maxArgsInit=-1);
		
	bool hasValidArgsSize(std::vector<std::string> &cmdTokens) const;
	
	MDInterpreter::CommandToken token;
	short int minArgs;
	short int maxArgs;
	
	const static int MAX_ARGS;
  };

  static int parsePart(const std::string &in, unsigned int &pos, std::string &part);
  static void removeQuotes(std::string &s);
  static void handleOctalEscapes(std::string &s);

  int getOptions(std::vector< std::string > &cString, 
		 std::string &options) const;

  std::vector< std::string > command;
  CommandToken cmdToken;
  
  std::string inputLine;

  // Should be static
  std::map< std::string, CommandDesc > cmdMap;
  
  std::auto_ptr<UploadHandle> uploadHandle;
  
  // Auxiliar, used by exec()
  std::vector< std::string > attributes;
  std::vector< std::string > values;
  AttrList attrList;
};

#endif // MDINTERPRETER_H
