/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDLFCServer.cc,v 1.34 2010/10/26 07:27:47 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDLFCServer.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"
#include "sqlHelpers.h"
#include "DBHelper.h"
#include "util/Path.h"

#include <sys/stat.h>
#include <assert.h>
#include <string>
#include <sstream>
#include <stdexcept>

#include "debug.h"

MDLFCServer::MDLFCServer(MDBuffer &outbuf)
  : MDServer(outbuf)
{
  masterindexTable = "Cns_file_metadata";
}


MDLFCServer::~MDLFCServer()
{
  // Disconnection handled by MDServer
}


void MDLFCServer::setupTableNames()
{
  MDServer::setupTableNames();

  std::string prefix;
  if(dbSchema.size())
    prefix = dbSchema + ".";
  masterindexTable = prefix + "masterindex";
}


/** Prints out all commands
    help
*/
void MDLFCServer::help()
{
  MDServer::help();
  
  out.append("Indices:\n");
  out.append("index_create name dir 'key_1 ... key_n' [type]\n");
  out.append("index_remove path\n");
}


int MDLFCServer::memberOf(const std::string usr,
			  std::list< std::string > &groups) const
{
  return 0;
}


std::string MDLFCServer::permsToString(int filemode, int mode)
{
  std::string m = "-------";
  if(filemode & S_IFDIR)
    m[0]= 'd';
  if(filemode & S_IFLNK)
    m[0]= 'l';
  
  if(filemode &S_IRUSR)
    m[1]='r';
  if(filemode &S_IWUSR)
    m[2]='w';
  if(filemode &S_IXUSR)
    m[3]='x';

  if(filemode &S_IRGRP)
    m[4]='r';
  if(filemode &S_IWGRP)
    m[5]='w';
  if(filemode &S_IXGRP)
    m[6]='x';
  
  if(filemode &S_IROTH)
    m[7]='r';
  if(filemode &S_IWOTH)
    m[8]='w';
  if(filemode &S_IXOTH)
    m[9]='x';
  
  // Now we have the full unix-perms, but we will only return what we
  // were asked for by mode
  switch(mode) {
  case 1: // File
    return m.substr(1, 3);
  case 2: // Directory
    return m.substr(1, 6);
  case 0:
    return m;
  }
  return m;
}


//! Discovers properties of an entry
/** This function fills out the EntryProps structure.
    For directories props.name is the full path of the directory,
    for entries it contains the entry name. In both cases
    parent is set to the parent directory.
*/
int MDLFCServer::getEntryProps(const std::string &entry,
			       EntryPropList &entryProps,
			       const std::string &indexTable,
			       const bool proxy)
{
  std::string aPath;
  aPath = absolutePath(entry, currentDir);

  std::list< std::string > parts;
  pathPieces(aPath, parts);

  int error = 0;

  std::list< std::string >::iterator I=parts.begin();
  int parentid = 0;
  int oldSize = 0;
  EntryProps props;
  while( I !=parts.end() ) {
    std::stringstream query;
    std::string part(*I++, oldSize);
    if(!oldSize)
      oldSize += part.size();
    else
      oldSize += part.size()+1;

    DMESG("part: " << part << std::endl);

    std::string pattern;
    bool isPattern=sqlPattern(part, pattern);

    query << "SELECT name, filemode, owner_uid, "
	  << "gid, fileid FROM "
	  << masterindexTable << " WHERE parent_fileid = "
	  << parentid << " AND name ";
    if(isPattern)
      query << "LIKE";
    else
      query << "=";
    query << " '" << pattern << "';";

    DMESG("SQL: >" << query.str() << "<" << std::endl);
    Statement statement(*dbConn);
    if(statement.exec(query.str())) {
      return 9;
    }
    char name[255];
    int filemode, owner_uid, gid, fileid;
    statement.bind(1, name, sizeof(name));
    statement.bind(2, &filemode);
    statement.bind(3, &owner_uid);
    statement.bind(4, &gid);
    statement.bind(5, &fileid);

    
    if(statement.fetch())
      return 1;

    // Check for directory and access rights
    std::string permissions = permsToString(filemode, 0);
    props.flags = 0;
    DMESG("LOOKing at " << name << std::endl);
    if(permissions[0] == 'd') {
      props.flags |= EPT_DIR;
      DMESG("Dir type set to dtDir\n");
    }
    if(I != parts.end() &&  props.type() == EPT_NONE)
      return -1;

    // Fill properties
    // Need to setup fake group with ACLs for gid!
    props.parentID = parentid;
    char buff[10];
    snprintf(buff, 9, "%d", owner_uid);
    if(parentid == 0) {
      props.parent = "/";
      props.parentID = parentid;
      props.parentPerms = permissions.substr(1, 6);
      props.parentACLs = "";
      snprintf(buff, 9, "%d", owner_uid);
      props.parentOwner = buff;
    } else {
      if(props.parent != "/")
	props.parent.append("/");
      if(props.name != "/")
	props.parent.append(props.name);
      props.parentID = props.id;
      props.parentPerms = props.perms;
      props.parentACLs = props.ACLs;
      props.parentOwner = props.owner;
    }
    props.name = part;
    props.id = fileid;
    props.owner = buff;
    props.perms = permissions.substr(1, 6);
    props.directoryTable = "x";
    props.indexTable = masterindexTable;

    parentid = fileid;

    props.print();
  }

  // Get associated table name, if table exists
  char buff[12];
  if(props.type() == EPT_NONE){
    DMESG("NOT dir\n");
    snprintf(buff, 11, "dir%ld", props.parentID);
  } else {
    DMESG("IS dir\n");
    snprintf(buff, 11, "dir%ld", props.id);
    std::string name = props.name;
    props.name = props.parent;
    if(props.name != "/")
      props.name.append("/");
    if(name != "/")
      props.name.append(name);
  }
  DMESG("Trying to get table information for " << buff << std::endl);
  Statement statement(*dbConn);
  if(statement.tables(buff)){
    printError("9 Could not get table information", statement);
    return 9;
  }
  if(statement.fetch()){
    DMESG("Table not found\n");
  } else {
    props.directoryTable = buff;
  }
    
  // Store entry properties in list
  entryProps.push_back(props);

  return error;
}


//
// Metadata commands
//

void MDLFCServer::createSchema(const std::string &name,
			       AttrList &attrs,
			       const std::string &parent)
{
  
  EntryPropList propList;
  int error=getEntryProps(name, propList);
  if(handleEntryPropErrors(error, name))
    return;
  EntryProps p(*propList.begin());

  std::list< std::string > groups;
  if(p.type() == EPT_DIR) {
    if(checkPermissions(p, 'w', groups)<=0){
      DMESG("Permission denied" << std::endl);
      out.append("4 Permission denied\n");
      return;
    }
  } else {
    if(checkDirPermissions(p, 'w', groups)<=0){
      DMESG("Permission denied" << std::endl);
      out.append("4 Permission denied\n");
      return;
    }
  }   

  if( !p.type() )
    p.id = p.parentID;

  Statement statement(*dbConn);
  std::string table = createTable(p, attrs, statement);
  if(!table.size()) {
    printError("9 Could not create table", statement);
    return;
  }
  
  out.append("0\n");
  
  return;
}


void MDLFCServer::removeSchema(const std::string &name)
{  
  EntryPropList propList;
  int error=getEntryProps(name, propList);
  if(handleEntryPropErrors(error, name))
    return;
  EntryProps p(*propList.begin());

  if( p.type() != EPT_DIR ) {
    out.append("23 Not a directory\n");
    return;
  }

  std::list< std::string > groups;
  if(checkPermissions(p, 'w', groups)<=0){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  // There is not really a schema ...
  if(p.directoryTable == "x")
    return;
    
  std::string query="DROP TABLE ";
  query.append(p.directoryTable);
  query.append(";");
  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Could not remove schema", statement);
    return;
  }

  out.append("0\n");
}


//! Adds an entry with it's attributes to a directory
/** Inserts an entry with a list of attributes and a list of
    correspondin values.
    The dimensions of the input vectors for the attributes and values
    need to be the same.
    \param path The relative or absolute path to the file
    \param attributes The vector of attributes to set
    \param values The vector of values

    ERRORS:
    - 1 File or directory not found
    - 3 Illegal command
      The attributes and values vectors don't match
    - 4 Permission Denied
*/
void MDLFCServer::addEntry(const std::string &path,
			   std::vector< std::string > &attributes,
			   std::vector< std::string > &values, bool eval)
{
  if(attributes.size() != values.size()) {
    out.append("3 Illegal command\n");
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string dir=p.parent;

  std::list< std::string > groups;
  if(checkDirPermissions(p, 'w', groups) <= 0){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::string query="INSERT INTO ";
  query.append(p.directoryTable);
  query.append(" (\"entry\"");
  for(unsigned int i=0; i<attributes.size(); i++){
    if(parser->parse(attributes[i], keys[i])){
	out.append("7 Illegal Key\n");
	return;
      }
    query.append(", ").append(keys[i]);
  }
  query.append(") VALUES (");
  char buff[21];
  snprintf(buff, 20, "%ld", p.id);
  query.append(buff);  if(eval) {
    for(unsigned int i=0; i<values.size(); i++) {
      std::string value;
      if(parser->parse(values[i], value)){
	out.append("9 Illegal value: ");
	out.append(values[i]);
	out.append("\n");
	return;
      }
      query.append(", ").append(value);
    }
  } else {
    for(unsigned int i=0; i<values.size(); i++)
      query.append(",'").append(values[i]).append("'");
  }
  query.append(");");

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
//    int err = statement.getErrorCode();
    std::string msg = statement.getError();
//    std::string sqlErr = statement.getSqlState();
    if(msg.find("unique")!=std::string::npos) {  // in postgres/sqlite
      printError("15 Entry exists", statement);
      return;
    }
    out.append("10 No such key: ");
    out.append(msg);
    out.append("\n");
    return;
  }

  out.append("0\n");
}


/** Adds a list of entries to several directories, the entries must not
    exist before
*/
void MDLFCServer::addEntries(std::vector<std::string> &paths)
{
  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;
  }

  for(size_t entry=0;  entry<paths.size(); entry++) {

    EntryPropList propList;
    int error=getEntryProps(paths[entry], propList);
    if(handleEntryPropErrors(error, paths[entry]))
      return;
    EntryProps p(*propList.begin());

    DMESG("Entry props: name >" << p.name
	  << "<  parent >"  << p.parent << "<"  << std::endl);

    if(error && error!=1 && handleEntryPropErrors(error, paths[entry]))
      return;

    std::list< std::string > groups;
    if(checkPermissions(p, 'w', groups) <=0 ) {
      out.append("4 Permission denied\n");
      return;
    }

    if(inSubdir(p.name) || p.type() == EPT_DIR) {
      out.append("1 No such file or directory: ");
      out.append(paths[entry]);
      out.append("\n");
      return;
    }

    std::string tableName=p.directoryTable;
    if(tableName == "x") {
      AttrList attrs;
      p.id = p.parentID;
      tableName = p.directoryTable = createTable(p, attrs, statement);
    }
    if(!tableName.size())
      return;

    std::stringstream query;
    query << "INSERT INTO " << tableName << " (\"entry\") VALUES ("
	  << p.id << ");";

    DMESG("SQL: >" << query.str() << "<" << std::endl);

    if(statement.exec(query.str())){
      printError("15 Entry exists", statement);
      return;
    }
  }

  statement.commitTransaction();
  out.append("0\n");
}


/** Removes a list of entries from several directories
    This operation is handled as a transaction, so either all entries
    get removed, or none in case something goes wrong with an entry
    \param paths : list of paths of entries to remove
*/
void MDLFCServer::removeEntries(std::vector<std::string> &paths)
{
  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;
  }
  std::list< std::string > groups;
  for(size_t entry=0;  entry<paths.size(); entry++) {

    EntryPropList propList;
    int error=getEntryProps(paths[entry], propList);
    if(handleEntryPropErrors(error, paths[entry]))
      return;
    EntryProps p(*propList.begin());

    DMESG("Entry props: name >" << p.name
	  << "<  parent >"  << p.parent << "<"  << std::endl);

    if(error && error!=1 && handleEntryPropErrors(error, paths[entry]))
      return;

    if(checkPermissions(p, 'w', groups) <=0 ) {
      out.append("4 Permission denied\n");
      return;
    }

    if(inSubdir(p.name) || p.type() == EPT_DIR) {
      out.append("1 No such file or directory: ");
      out.append(paths[entry]);
      out.append("\n");
      return;
    }


    std::string tableName=p.directoryTable;
    if( !tableName.size() ) {
      AttrList attrs;
      p.id = p.parentID;
      tableName = p.directoryTable = createTable(p, attrs, statement);
    }
    if(!tableName.size())
      return;

    std::stringstream query;
    query << "DELETE FROM " << tableName << " WHERE \"entry\" = "
	  << p.id << ";";

    DMESG("SQL: >" << query.str() << "<" << std::endl);

    if(statement.exec(query.str())){
      printError("1 Entry does not exist", statement);
      return;
    }
  }

  statement.commitTransaction();
  out.append("0\n");
}


/** Gets attribute(s) of a number of files
    path is the path to the files where wildcards are only allowed in
    the file part of the path
    attributes is the list of attributes to which need to be returned
    The output including all errors is streamed into the out class
    member. For a description see the protocol description.
*/
void MDLFCServer::getAttr(const std::string &path,
			  std::vector< std::string > &attributes)
{
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  if(!tableExists(p.directoryTable, p.type()))
    return;

  std::string pattern;
  bool isPattern = sqlPattern(p.name, pattern);
  if( p.type() ) {
    isPattern = true;
    pattern="%";
  }

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }

  std::string query="SELECT ";
  query.append(p.indexTable).append(".name, ");

  for(size_t i=0; i<attributes.size(); i++){
    std::string key;
    if(parser->parse(attributes[i], key)){
      out.append("7 Illegal Key\n");
      return;
    }
    query.append(p.directoryTable).append(".");
    query.append(key);
    if(i<attributes.size()-1)
      query.append(", ");
  }
  query.append(" FROM ").append(p.directoryTable);
  query.append(", ").append(p.indexTable);
  query.append(" WHERE ").append(p.indexTable).append(".name");
  if(isPattern)
    query.append(" LIKE '");
  else
    query.append(" = '");
  query.append(pattern);
  query.append("' AND ").append(p.indexTable).append(".fileid = ");
  query.append(p.directoryTable).append(".\"entry\";");

  DMESG("SQL: >" << query << "<" << std::endl);

  bool foundSomething=false;
  doBulkRead(statement, query, attributes.size()+1, foundSomething);

  if(!foundSomething){
    printError("1 No such file or directory", statement);
    return;
  }
  statement.commitTransaction();
}


/** Finds files matching path and the query which is in the query
    language described in QParser.
    The output including all errors is streamed into the out class
    member. For a description see the protocol description.
    \param path Pattern for entries to match
    \param query Query which entries need to match
*/
void MDLFCServer::find(const std::string &path,
		       const std::string &query)
{
  char buff[1024];

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  if(!tableExists(p.directoryTable, p.type()))
    return;

  if(parser->parse(query)){
    out.append("8 Illegal Query\n");
    return;
  }
  TableMap tableMap;
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap

  std::string q;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q);

  std::string pattern;
  bool isPattern=sqlPattern(p.name, pattern);
  if( p.type() ) {
    isPattern = true;
    pattern = "%";
  }

  std::string queryString="SELECT ";
  queryString.append(p.indexTable);
  queryString.append(".name FROM ");
  queryString.append(p.indexTable);
  queryString.append(", ");
  queryString.append(p.directoryTable);
  queryString.append(" WHERE ");
  queryString.append(p.indexTable);
  queryString.append(".name");
  if(isPattern)
    queryString.append(" LIKE '");
  else
    queryString.append(" = '");
  queryString.append(pattern).append("'");
  queryString.append(" AND ");
  queryString.append(p.indexTable).append(".fileid = ");
  queryString.append(p.directoryTable).append(".\"entry\"");
  if(q.size()) {
    queryString.append(" AND (");
    queryString.append(q).append(")");
  }
  queryString.append(";");

  DMESG("SQL: >" << queryString << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(queryString)){
    printError("2 Connection failed", statement);
    return;
  }
  statement.bind(1, buff, 1024);

  out.append("0");
  int rows=statement.numRows();
  if(rows>-1) {
    char buff[20];
    sprintf(buff, " %d\n", rows);
    out.append(buff);
  } else {
    out.append("\n");
  }

  while(1){
    if(statement.fetch())
      break;

    out.append(buff);
    out.append("\n");
  }
}


void MDLFCServer::listEntries(const std::string &directory)
{ 
  return find(directory, "");
}


//
// Metadata optional
//


/** Creates an index
    createindex name dir 'key1 key2' algo
    Creates an index name on dir for key using the indexing algorithm algo
*/
void MDLFCServer::createIndex(const std::string &name,
			      const std::string &directory,
			      const std::vector<std::string> &attributes,
			      const std::string &algo)
{
  for(unsigned int i=0; i<name.size(); i++){
    if(!isalnum(name[i]) || isupper(name[i])) {
      out.append("17 Illegal name (for index, view)\n");
      return;
    }
  }

  std::list< std::string >keys;
  std::vector<std::string>::const_iterator I=attributes.begin();
  while(I!=attributes.end()){
    std::string key;
    if(parser->parse(*I, key)){
      out.append("7 Illegal Key ");
      out.append(*I);
      out.append("\n");
      return;
    }
    keys.push_back(key);
    I++;
  }

  if(!keys.size()){
    out.append("7 No key\n");
    return;
  }
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return;
  EntryProps p(*propList.begin());

  if( p.type() != EPT_DIR ) {
    out.append("17 No such directory ");
    out.append(directory);
    out.append("\n");
    return;
  }

  Statement statement(*dbConn);
  std::string table=p.directoryTable;
  if(table=="x") {
    AttrList attrs;
    p.directoryTable=table=createTable(p, attrs, statement);
  }
  if(!table.size())
    return;

  std::string entry="index_";
  entry.append(table).append("_").append(name);

  std::string query = "CREATE INDEX ";
  query.append(entry).append(" ON ");
  query.append(table);
  // query.append(" USING ");
  //  query.append(algo);
  query.append(" (");
  std::list< std::string >::const_iterator J=keys.begin();
  while(J!=keys.end()){
    query.append(*J);
    J++;
    if(J!=keys.end())
      query.append(", ");
  }
  query.append(");");

  DMESG("SQL: >" << query << "<" << std::endl);
  if(statement.exec(query)){
    printError("9 Internal DB error", statement);
    return;
  }
  out.append("0\n");
}


void MDLFCServer::indexRemove(const std::string &path)
{
  size_t pos = path.rfind("/");
  if(pos == std::string::npos || pos >= path.size() -1) {
    out.append("1 No such index\n");
    return;
  }
  std::string name = path.substr(pos+1);
  std::string newpath = path.substr(0, pos);
  
  EntryPropList propList;
  int error=getEntryProps(newpath, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  if( p.type() != EPT_DIR ) {
    out.append("17 No such directory ");
    out.append(path);
    out.append("\n");
    return;
  }

  p.name = name;
  std::string query = "DROP INDEX index_";
  char buff[21];
  snprintf(buff, 20, "dir%ld_", p.id);
  query.append(buff);
  query.append(p.name);
  snprintf(buff, 20, " ON dir%ld", p.id);
  query.append(buff);
  query.append(";");
  
  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Internal DB error", statement);
    return;
  }

  out.append("0\n");
  return;
}



void MDLFCServer::createView(const std::string &name,
			     const std::string &directory,
			     const std::vector<std::string> &attributes,
			     const std::string &restriction)
{
  ERRLOG("MDLFCServer::createView not implemented\n");
  out.append("21 Call not implemented\n");
}


void MDLFCServer::createViewUnion(const std::string &name,
				  const std::vector<std::string> &dirs,
				  const std::string &restriction)
{
  ERRLOG("MDLFCServer::createViewUnion not implemented\n");
  out.append("21 Call not implemented\n");
}


void MDLFCServer::createSequence(const std::string &name,
				 const std::string &dir,
				 const std::string &increment,
				 const std::string &start)
{
  ERRLOG("MDLFCServer::createSequence not implemented\n");
  out.append("21 Call not implemented\n");
}


void MDLFCServer::viewRemove(const std::string &path)
{
  ERRLOG("MDLFCServer::viewRemove not implemented\n");
  out.append("21 Call not implemented\n");
}


void MDLFCServer::sequenceRemove(const std::string &path)
{
  ERRLOG("MDLFCServer::sequenceRemove not implemented\n");
  out.append("21 Call not implemented\n");
}


int MDLFCServer::upload(UploadHandle &h,
				const std::string &dir,
				std::vector< std::string > &attributes)
{	
  ERRLOG("MDLFCServer::upload not implemented\n");
  out.append("21 Call not implemented\n");
  return 0;
}


int MDLFCServer::put(UploadHandle &h,
		const std::vector<std::string>::iterator valStart,
    const std::vector<std::string>::iterator valEnd)
{
  ERRLOG("MDLFCServer::put not implemented\n");
  out.append("21 Call not implemented\n");
  return 0;
}


void MDLFCServer::abort(UploadHandle &h)
{
  ERRLOG("MDLFCServer::abort not implemented\n");
  out.append("21 Call not implemented\n");
}


void MDLFCServer::commit(UploadHandle &h)
{
  ERRLOG("MDLFCServer::commit not implemented\n");
  out.append("21 Call not implemented\n");
}


void MDLFCServer::sequenceNext(const std::string &name)
{
  ERRLOG("MDLFCServer::squenceNext not implemented\n");
  out.append("21 Call not implemented\n");
}


void MDLFCServer::sequenceCurrent(const std::string &name)
{
  ERRLOG("MDLFCServer::sequenceCurrent not implemented\n");
  out.append("21 Call not implemented\n");
}


/** Selects attribute(s) of a number of files based on a condition
    path is the path to the files where wildcards are only allowed in
    the file part of the path
    attributes is the list of attributes to which need to be returned
    The output including all errors is streamed into the out class
    member. For a description see the protocol description.
*/
void MDLFCServer::selectAttr(std::vector<std::string> &attributes,
			     const std::string &q)
{
  TableMap tableMap;
  if(createTableMap(attributes, tableMap))
    return; // Error handled by createTableMap

  if(parser->parse(q)){
    out.append("8 Illegal Query\n");
    return;
  }
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap

  std::string query="SELECT ";

  std::string q2;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q2);

  std::vector<std::string>::iterator I=attributes.begin();
  while(I!=attributes.end()){
    std::string key;
    parser->parse(*I++);
    parser->replaceTableNames(tableMap);
    parser->toSQL(key);
    query.append(key);
    if(I!=attributes.end())
      query.append(", ");
  }

  std::list< std::string > distinctTables;
  getDistinctTables(tableMap, distinctTables);

  std::list< std:: string >:: iterator J=distinctTables.begin();
  query.append(" FROM ");
  while(J != distinctTables.end()) {
    query.append(*J++);
    if(J!=distinctTables.end())
      query.append(", ");
  }
  query.append(", ").append(masterindexTable);

  if(!q2.size())
    q2="true";

  query.append(" WHERE ").append(q2);
//   J=distinctTables.begin();
//   while(J != distinctTables.end()) {
//     query.append(" AND (").append(*J);
//     query.append(".\"owner\"='").append(user).append("'");
//     query.append(" OR ").append(*J++);
//     query.append(".\"group_rights\" LIKE 'r%')");
//   }

  query.append(" AND (");
  J=distinctTables.begin();
  while(J != distinctTables.end()) {
    query.append(masterindexTable).append(".fileid = ");
    query.append(*J++).append(".\"entry\"");
    if(J!=distinctTables.end())
      query.append(" OR ");
  }
  query.append(");");


  // begin transaction
  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  DMESG("SQL: >" << query << "<" << std::endl);

  bool dummy=false;
  doBulkRead(statement, query, attributes.size(), dummy);

  if(!dummy){
    out.append("0\n");
  }
  statement.commitTransaction();
}


//
// File catalogue support
//

void MDLFCServer::dir(const std::string &directory)
{
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return;
  EntryProps p(*propList.begin());

  std::string pattern;
  bool isPattern = sqlPattern(directory, pattern);

  std::stringstream query;
  if( p.type() == EPT_DIR ) {
    isPattern = true;
    pattern = "%";
    out.append("0\n");
    query << "SELECT name, filemode FROM " << masterindexTable
	  << " WHERE parent_fileid = " << p.id;
    if( isPattern ){
      query << " AND name LIKE '" << pattern << "';";
    } else {
      query << " AND name = '" << pattern << "';";
    }

  } else {
    out.append("1 Currently not implemented\n");
  }

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query.str())) {
    printError("9 cannot read LFC file table", statement);
    return;
  }

  char name[255];
  int filemode;
  statement.bind(1, name, sizeof(name));
  statement.bind(2, &filemode);

  while( !statement.fetch() ) {
    out.append(name);
    if(filemode & S_IFDIR)
      out.append("\ncollection\n");
    else
      out.append("\nentry\n");
  }
}

//uint64_t MDLFCServer::sequenceNextValue(const std::string &sequence) throw (SQLException) {
//  throw SQLException("Not implemented");
//}
//
//
//uint64_t MDLFCServer::getLastGeneratedXid() throw (SQLException) {
//  throw SQLException("Not yet implemented");
//}
