/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDMySQL5Server.cc,v 1.40 2011/01/26 05:57:13 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDMySQL5Server.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"
#include "MySQLHelper.h"

#include "sqlHelpers.h"

#include <sqlext.h>

#include <iostream>
#include "debug.h"

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string.hpp>

using namespace std;

MDMySQL5Server::MDMySQL5Server(MDBuffer &outbuf)
  : MDStandalone(outbuf)
{
  dbHelper = new MySQLHelper();
  parser->setEngine(dbHelper->backend());
}


MDMySQL5Server::~MDMySQL5Server()
{
  delete dbHelper;
}


//! Creates a table for a directory
/** Creates a table for a directory which must already exist in an
    index table. The created table is inserted into the index table.
    If parent is given, the created table inherits the
    schema of the parent table.
    \param id The id number of the directory
    \param indexTable Name of directory index for table
    \param parent The name of the table which is inherited
    \return Returns the name of the table which was created
    Any error printout is done by this function
*/
std::string MDMySQL5Server::createTable(const EntryProps &p,
					AttrList &attrs,
					Statement &statement)
{
  DMESG("In createTable\n");
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return "";
  }
  
  std::string mainType = "varchar(256)";
  if( typeFromFlags(p.flags) )
    mainType = typeFromFlags(p.flags);
  if( p.mainType.size() )
    mainType = p.mainType;

  std::string tableType="InnoDB";
  if(p.tableType.size())
    tableType = p.tableType;

  std::ostringstream query;
  std::string sep = "";
  
  query << "CREATE TABLE dir" << p.id << " (";
  if( !(p.flags & EP_PLAIN) ) {
	  query << "\"file\" " << dbHelper->mdToDBType(mainType);
	  sep = ", ";

	  if( p.flags & EP_PERMS ) {
		  query << sep << "\"owner\" VARCHAR(" << USERNAME_MAX_SIZE-1 << "),"
				  " \"permissions\" char(3),"
				  " \"group_rights\" char(3)";
		  sep = ", ";
	  }

	  if (p.flags & EP_SHARED) {
		  query << sep << "\"dir\" integer REFERENCES "
				  << p.indexTable << "(\"id\")"
				  << " ON DELETE CASCADE";
		  sep = ", ";
	  }

	  if( p.flags & EP_FCAT ) {
		  query << sep << "\"link\" VARCHAR(1025),"
				  " \"created\" DATETIME,"
				  " \"size\" BIGINT,"
				  " \"guid\" VARCHAR(32),"
				  " \"md5\" VARCHAR(32)";
		  sep = ", ";
	  }

	  if( p.flags & EP_ACLS ) {
		  query << sep << "\"acls\" VARCHAR(2048)";
		  sep = ", ";
	  }
  }
  
  AttrList::const_iterator I=attrs.begin();    
  while(I!=attrs.end()) { 
    std::string key;
    if(parser->parse(I->first, key, p.flags & EP_PLAIN)){
      out.append("7 Illegal Key\n");
      return "";
    }
    query << sep << key << " " << (dbHelper->mdToDBType((*I).second));
    sep = ", ";
    I++;
  }

  std::string length;
  if( mainType.size()> 7 && mainType.substr(0, 7) == "varchar" ) {
    length = mainType.substr(7);
    mainType = mainType.substr(0,7);
  }

  if( !(p.flags & EP_PLAIN) ) {
    if (p.flags & EP_SHARED) {
      if( length.size() )
	query << ", PRIMARY KEY(\"" << p.mainColumn << "\""  << length << ", \"dir\")";
      else
	query << ", PRIMARY KEY(\"" << p.mainColumn << "\", \"dir\")";
    }
    else {
      if( length.size() )
	query << ", PRIMARY KEY(\"" << p.mainColumn << "\"" << length << ")";
      else
	query << ", PRIMARY KEY(\"" << p.mainColumn << "\")";
    }
  }

  query << ") DEFAULT CHARACTER SET latin1 COLLATE latin1_general_cs"
	<< ", ENGINE = " << tableType << ";";
  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if( statement.exec(query.str()) ) {
    printError("9 Could not create table: ", statement);
    return "";
  }

  // If FC table, we need some triggers
  if( p.flags & EP_FCAT ) {
    query.str("");
    query << "CREATE TRIGGER tnewguid" << p.id << " AFTER INSERT ON dir" << p.id 
	  << " FOR EACH ROW CALL newguid(";
    if (p.flags & EP_SHARED)
      query << "concat('dir" << p.id << "_s', NEW.dir), NEW.guid";
    else
      query << "'dir" << p.id << "', NEW.guid";
    query << ");";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Could not alter table", statement);
      return "";
    }
    
    query.str("");
    query << "CREATE TRIGGER tdelguid" << p.id << " AFTER DELETE ON dir" << p.id 
	  << " FOR EACH ROW "
	  << "   BEGIN"
	  << "     DELETE FROM guids WHERE \"guid\" = OLD.\"guid\";"
	  << "   END;";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Could not alter table", statement);
      return "";
    }
  }

  if(commitDirectoryToIndex(p.id, p.indexTable, statement, p.flags)<0){
    return "";
  }
  if( statement.commitTransaction() ) {
    printError("9 Cannot commit transaction", statement);
    return "";
  }

  query.str("");
  query << "dir" << p.id;
  
  return query.str();
}


int MDMySQL5Server::lockTable(Statement &statement, 
			      const std::string &indexTable)
{
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return -1;
  }

  DMESG("SQL: >" << "LOCK" << "<" << std::endl);
  std::string query="LOCK TABLES ";
  query.append(indexTable).append(" WRITE;");
  if(statement.exec(query)){
    printError("9 Error acquiring lock", statement);
    return -1;
  }
  DMESG("SQL: >" << "ACQUIRED LOCK" << "<" << std::endl);
  return 0;
}


void MDMySQL5Server::unlockTables(Statement &statement)
{
  DMESG("UNLOCKING MySQL TABLES\n");
  if(statement.exec("UNLOCK TABLES;")) {
    ERRLOG("Unlocking of MySQL tables failed\n");
  }
}


bool MDMySQL5Server::dirPattern(const std::string &in, std::string &out) const
{
  return dirRegexpPattern(in, out);
}


std::string MDMySQL5Server::matchSubdirs(const std::string &dir) const
{
  std::string query="\"directory\" REGEXP '^";
  if(dir.size()>1)
    query.append(dir);
  query.append("/[^/]+$'");
  return query;
}

int MDMySQL5Server::getAttributes(const std::string &table,
				  AttrList &keys,
				  bool all)
{
  std::string query = "SELECT column_name, column_type "
    "FROM INFORMATION_SCHEMA.columns where table_name = '"
    + stripSchema(table) + "'  AND table_schema = ";

  if( getSchema(table).size() )
    query += "'" + getSchema(table) +"'";
  else
    query += "database()";
  query += " ORDER BY column_name;";

  char key[1024];
  char type[1024];

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Cannot read master description table", statement);
    return -1;
  }
  statement.bind(1, key, 1024);
  statement.bind(2, type, 1024);

  //Mandatory for plain tables
  strncpy(type, dbHelper->dbToMDType(type).c_str(), 1024);

  while(!statement.fetch()){
    if(!all
       && strncmp(key, parser->prefix().c_str(),
		  parser->prefix().size())
       )
      continue;

    DMESG(">>> " << key << " " << type << std::endl);

    if(!all)
      keys.push_back(Attribute(key+parser->prefix().size(), type));
    else
      keys.push_back(Attribute(key, type));
  }

  statement.unBind();
  statement.close();

  return 0;
}


bool MDMySQL5Server::alterTableColumns(const std::string &tableName,
				       const AttrList &keyList,
				       bool remove)
{
  DMESG("List-size: " << keyList.size() << std::endl);

  std::string query="ALTER TABLE ";
  query.append(tableName);
  if(remove) 
    addAttributeList(keyList, query, " DROP COLUMN ");
  else
    addTableDescription(keyList, query, " ADD ");
 
  query.append(";");
  
  DMESG("SQL: >" << query << "<" << std::endl);
  
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Could not alter table", statement);
    return false;
  }
  return true;
}


void MDMySQL5Server::doBulkRead(Statement &statement, std::string &query,
				size_t fields, bool &foundSomething)
{
  char buff[1001];
  if(statement.exec(query)){
    printError("9 Internal error", statement);
    foundSomething=true;
    return;
  }

  DMESG("EXEC done" << std::endl);
  while(1) {
    if(statement.fetch()){
      DMESG("Fetch failed: Last data read" << std::endl);
      break;
    }

    if(!foundSomething) {
      /* Fix 20090728 : Return #column instead of #rows */
      out.append("0");
      char buff[20];
      sprintf(buff, " %d\n", fields);
      out.append(buff);
    }
    foundSomething = true;

    SQLHSTMT s=statement.getStatement();
    SQLLEN err;

    for(size_t i=0; i<fields; i++){
      DMESG("Getting Field: " << i << std::endl);
      while(1) {
        int res=SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err);
        DMESG("Res: " << res << "   err: " << err << std::endl);
        if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO)
          break;
        if(err==SQL_NULL_DATA) break;
        std::string line(buff);
        AMGA::encodeLine(line);
        out.append(line);
      }
     out.append("\n");
    }
  }
  statement.close();
}


/** vacuums a database
 */
void MDMySQL5Server::vacuum(const std::string &options)
{
  out.append("21 Function not implemented\n");
}

int MDMySQL5Server::dumpIndexes(Statement &statement, const string &directory, const string &tablename) {

  // TODO: dump indexes
  DMESG("dumpIndexes: " << tablename << endl);

  std::vector< std::string > result;
  boost::algorithm::split(result, tablename, boost::is_any_of("."));

  std::string dir2;

  if( result.size() == 1) dir2 = tablename;
  else dir2 = result.at(1);

  // Build the select query
  ostringstream oss;

  oss << "select table_name, index_name,"
      << " group_concat(column_name SEPARATOR ' ')"
      << " from information_schema.statistics where ";

  if( result.size() != 1 ) 
    oss << "table_schema = '" << result.at(0) << "' and " ;

  oss << "table_name = '" << dir2 << "' group by index_name";
  
  string query = oss.str();
  DMESG("SQL: >" << query << "<" << std::endl);
  if (statement.exec(query)) {
    ERRLOG("Could not dump indexes from " << tablename << std::endl);
    return 1;
  }

  char table_name[1024];
  char index_name[1024];
  char column_name[1024];
  statement.bind(1, table_name, 1024);
  statement.bind(2, index_name, 1024);
  statement.bind(3, column_name, 1024);
  
  try {
    while(!statement.fetch()){
      if(strcmp(column_name, "file") == 0){ 
        DMESG("skip dumping index file>" << table_name << std::endl);
        continue;
      }
      char str[4096];
      string index_name_str(index_name);
      
      boost::algorithm::split(result, index_name_str, boost::is_any_of("_"));
      index_name_str = result.at(result.size()-1);
      boost::to_lower(index_name_str);
      for(unsigned int i=0; i<index_name_str.size(); i++){
        if(!isalnum(index_name_str[i])) 
          index_name_str[i] = 'x';
      }

      
      string column_name_str(column_name);
      boost::erase_all(column_name_str, "user:");
      
      out.append("index_create ");
      out.append(index_name_str);
      out.append(" ");
      out.append(directory);
      out.append(" '");
      out.append(column_name_str);
      out.append("' ");
      out.append("\n");
      DMESG("dump index: >" << table_name << " : " << index_name_str << std::endl);
    }
  } catch (SQLException &ex) {
    DMESG("Error dumping indexes: " << ex.what() << endl);
    return 1;
  }
  statement.unBind();
  statement.close();
  return 0;
}

//uint64_t MDMySQL5Server::sequenceNextValue(const std::string &sequence)
//  throw (SQLException) 
//{ 
//  MySQLHelper* helper = dynamic_cast<MySQLHelper*>(dbHelper);
//  if (helper == NULL) {
//    LOG("Could not downcast to MySQLHelper");
//    throw SQLException("Internal error");
//  } else { 
//    return helper->sequenceNextValue(dbConn, sequence);
//  }
//}
//
//
//uint64_t MDMySQL5Server::sequenceCurrentValue(const std::string &sequence)
//  throw (SQLException)
//{ 
//  MySQLHelper* helper = dynamic_cast<MySQLHelper*>(dbHelper);
//  if (helper == NULL) {
//    LOG("Could not downcast to MySQLHelper");
//    throw SQLException("Internal error");
//  } else { 
//    return helper->sequenceCurrentValue(dbConn, sequence);
//  }
//}
