/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDOracleServer.cc,v 1.36 2011/01/26 05:57:13 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDOracleServer.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"
#include "OracleHelper.h"
#include "sqlHelpers.h"

#include <sqlext.h>

#include <iostream>
#include "debug.h"

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string.hpp>


using namespace std;


MDOracleServer::MDOracleServer(MDBuffer &outbuf)
  : MDStandalone(outbuf)
{
  dbHelper = new OracleHelper();
  parser->setEngine(dbHelper->backend());
}


MDOracleServer::~MDOracleServer()
{
  delete dbHelper;
}


int MDOracleServer::lockTable(Statement &statement, 
			      const std::string &indexTable)
{
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return -1;
  }

  DMESG("SQL: >" << "LOCK" << "<" << std::endl);
  std::string query="LOCK TABLE ";
  query.append(indexTable).append(" IN EXCLUSIVE MODE;");
  if(statement.exec(query)){
    printError("9 Error acquiring lock", statement);
    return -1;
  }
  DMESG("SQL: >" << "ACQUIRED LOCK" << "<" << std::endl);
  return 0;
}


bool MDOracleServer::dirPattern(const std::string &in, std::string &out) const
{
  return dirRegexpPattern(in, out);
}


std::string MDOracleServer::matchSubdirs(const std::string &dir) const
{
  std::string query="REGEXP_LIKE(\"directory\", '^";
  if(dir.size()>1)
    query.append(dir);
  query.append("/[^/]+$')");
  return query;
}


bool MDOracleServer::alterTableColumns(const std::string &tableName,
				       const AttrList &keyList,
				       bool remove)
{
  DMESG("List-size: " << keyList.size() << std::endl);

  std::string query="ALTER TABLE ";
  query.append(tableName);
  if(remove) 
    addAttributeList(keyList, query, " DROP COLUMN ", false);
  else
    addTableDescription(keyList, query, " ADD ", false);
 
  query.append(";");
  
  DMESG("SQL: >" << query << "<" << std::endl);
  
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Could not alter table", statement);
    return false;
  }
  return true;
}


void MDOracleServer::doBulkRead(Statement &statement, std::string &query,
				size_t fields, bool &foundSomething)
{
  char buff[1001];
  if(statement.exec(query)){
    printError("9 Internal error", statement);
    foundSomething=true;
    return;
  }

  DMESG("EXEC done" << std::endl);
  while(1) {
    if(statement.fetch()){
      DMESG("Fetch failed: Last data read" << std::endl);
      break;
    }

    if(!foundSomething) { 
      /* Fix 20090728 : Return #column instead of #rows */
      out.append("0");
      char buff[20];
      sprintf(buff, " %d\n", fields);
      out.append(buff);
    }
    foundSomething = true;

    SQLHSTMT s=statement.getStatement();
    SQLLEN err;

    for(size_t i=0; i<fields; i++){
      DMESG("Getting Field: " << i << std::endl);
      while(1) {
	int res=SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err);
	DMESG("Res: " << res << "   err: " << err << std::endl);
	if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO)
	  break;
	if(err==SQL_NULL_DATA) break;
	std::string line(buff);
	AMGA::encodeLine(line);
	out.append(line);
      }
     out.append("\n");
    }
  }
  statement.close();
}


int MDOracleServer::getAttributes(const std::string &table,
				  AttrList &keys,
				  bool all)
{
  std::string tableName=stripSchema(table);
  for(unsigned int i=0; i<tableName.size(); i++)
    tableName[i]=toupper(tableName[i]);
  
  std::string owner = dbSchema; // Call this a hack!
  for(unsigned int i=0; i<owner.size(); i++)
    owner[i]=toupper(owner[i]);

  // Use all_tab_columns instead.
  std::string query = "SELECT column_name, data_type, data_length, "
    "data_precision, data_scale FROM all_tab_columns WHERE "
    "owner ='" + owner +"' AND " 
    "table_name='" + tableName + "';";

//   std::string query="SELECT column_name, data_type, data_length, "
//     "data_precision, data_scale FROM " 
//     + defaultSchema + ".user_tab_columns WHERE table_name='";
//   query.append(tableName);
//   query.append("';");


  char key[1025]; 
  char data_type[65];
  char data_length[65];
  char data_precision[65];
  char data_scale[65];

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);

  if(statement.exec(query)){
    printError("9 Cannot read master description table", statement);
    return -1;
  }
  
  DMESG("Statement executed\n");

  statement.bind(1, key, 1024);
  statement.bind(2, data_type, 64);
  statement.bind(3, data_length, 64);
  statement.bind(4, data_precision, 64);
  statement.bind(5, data_scale, 64);
  
  DMESG("Fetching!\n");

  while(!statement.fetch()){
//    DMESG("Row: >" << key << "<   type >" << data_type
//	  << "<   length >" << data_length << "<   precision >"
//	  << data_precision << "<   scale >" << data_scale << "<" << std::endl);
    
    if(!all 
       && strncmp(key, parser->prefix().c_str(),
		  parser->prefix().size())
       )
      continue;
    std::stringstream type;

    int length = ((OracleHelper *)dbHelper)->toInt(data_length);
    int precision = ((OracleHelper *)dbHelper)->toInt(data_precision);
    int scale = ((OracleHelper *)dbHelper)->toInt(data_scale);

//    DMESG("Or  type >" << data_type
//	  << "<   length " << length << "  precision "
//	  << precision << "   scale " << scale << std::endl);
    
    if(!strcmp(data_type, "NUMBER")) {
      if(scale<0 && precision < 0)
	type<<"number";
      if(scale==0)
	type<<"integer";
      if(precision>0 && scale>0) {
	if(!precision)
	  type << "number(" << scale << ")";
	else
	  type << "number(" << precision << "," << scale << ")";
      }
    }
    if(!strcmp(data_type, "VARCHAR2"))
      type << "varchar(" << length << ")";
    if(!strcmp(data_type, "CHAR"))
      type << "char(" << length << ")";
    if(!strcmp(data_type, "DATE"))
      type << "date";
    if(!strncmp(data_type, "TIMESTAMP", 9)){
      type << "timestamp";
      type << "(" << scale << ")";
    }
    if(!strcmp(data_type, "FLOAT"))
      type << "float";

    if(!strcmp(data_type, "LONG"))
      type << "long";

    if(!strcmp(data_type, "CLOB"))
      type << "clob";

    if(!type.str().size())
      type.str()="unknown";

    if(!all)
      keys.push_back(Attribute(key+parser->prefix().size(), type.str()));
    else
      keys.push_back(Attribute(key, type.str()));
  }

  statement.unBind();
  statement.close();

  return 0;
}


/** vacuums a database
 */
void MDOracleServer::vacuum(const std::string &options)
{
  out.append("0\n");
}


//uint64_t MDOracleServer::sequenceNextValue(const std::string &sequence) 
//    throw (SQLException)
//{
//  std::string query=std::string("SELECT ") + sequence + ".nextval FROM dual;";
//  Statement statement(*dbConn);
//  if(statement.exec(query)){
//    throw SQLException(statement);
//  }  
//  if(statement.fetch()){
//    throw SQLException(statement);
//  }
//  uint64_t v;
//  statement.getData(1, v);  
//  return v;  
//}
//
//
//uint64_t MDOracleServer::sequenceCurrentValue(const std::string &sequence) 
//    throw (SQLException)
//{
//  std::string query=std::string("SELECT ") + sequence + ".currval FROM dual;";
//  Statement statement(*dbConn);
//  if(statement.exec(query)){
//    throw SQLException(statement);
//  }  
//  if(statement.fetch()){
//    throw SQLException(statement);
//  }
//  uint64_t v;
//  statement.getData(1, v);
//  return v;  
//}
//
//uint64_t MDOracleServer::getLastGeneratedXid() throw (SQLException) {
//  return sequenceCurrentValue("logs_seq");
//}

//! Creates a table for a directory with a given schema
/** Creates a table for a directory which must already exist in an
    index table. The created table is inserted into the index table.
    Attributes are created according to the given schema
    \param id The id number of the directory
    \param indexTable Name of directory index for table
    \param attrs The list of attributes to create
    \param statement A handle to the statement which already has
    a transaction open
    \return Returns the name of the table which was created
    Any error printout is done by this function
*/
std::string MDOracleServer::createTable(const EntryProps &p,
					AttrList &attrs,
					Statement &statement)
{
  DMESG("In MDOracleServer::createTable, id: " << p.id 
	<< ", indexTable: " << p.indexTable 
	<< ", flags: " << p.flags << ", type >" << p.mainType << "<\n");

  if( statement.beginTransaction() ) {
    printError("9 Cannot commit transaction", statement);
    return "";
  }

  std::string mainType = typeFromFlags(p.flags);
  if( !p.mainType.empty() ) {
    mainType = p.mainType;
    DMESG("Overriding main Type");
  }
  std::stringstream query;

  query << "CREATE TABLE dir" << p.id << " (";
  if ( !(p.flags & EP_PLAIN) ) 
    query << "\"" << p.mainColumn << "\" " << dbHelper->mdToDBType(mainType);

  if (p.flags & EP_SHARED) {
    query << ", \"dir\" INTEGER REFERENCES " << p.indexTable << "(\"id\")"
	  << " ON DELETE CASCADE";
  }

  if( p.flags & EP_PERMS ) {
    query << ", \"owner\" VARCHAR(" << USERNAME_MAX_SIZE-1 << "),"
	  << " \"permissions\" CHAR(3),"
	  << " \"group_rights\" CHAR(3)";
  }

  if( p.flags & EP_FCAT ) {
    query << ", \"link\" VARCHAR(1024),"
	  << " \"created\" TIMESTAMP,"
	  << " \"size\" NUMBER(38),"
	  << " \"guid\" VARCHAR(32),"
	  << " \"md5\" VARCHAR(32)";
  }
  if (p.flags & EP_ACLS) {
    query << ", \"acls\" VARCHAR(2048)";
  }

  AttrList::const_iterator I=attrs.begin();    
  while(I!=attrs.end()) { 
    std::string key;
    if(parser->parse(I->first, key,p.flags & EP_PLAIN)){
      out.append("7 Illegal Key\n");
      return "";
    }
    query << ", " << key << " " << (dbHelper->mdToDBType((*I).second));
    I++;
  }
  
  std::stringstream pkey;
  pkey << "p_key_dir" << p.id;

  if( !(p.flags & EP_PLAIN ) ) {
    if( p.flags & EP_SHARED )
      query << ", CONSTRAINT " << pkey.str() 
	    << "PRIMARY KEY (\"" << p.mainColumn << "\", \"dir\")";
    else {
      query << ", CONSTRAINT " << pkey.str() 
	    << " PRIMARY KEY (\"" << p.mainColumn << "\")";
    }    
  }

  query << ");";

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if(statement.exec(query.str())){
    printError("9 Could not alter table", statement);
    return "";
  }

  // If FC table, we need some triggers
  if( p.flags & EP_FCAT ) {
    query.str("");
    query << "CREATE TRIGGER tnewguid" << p.id << " AFTER INSERT ON dir" << p.id 
	  << " REFERENCING NEW AS newRow FOR EACH ROW CALL newguid(";
    if (p.flags & EP_SHARED)
      query << "concat('dir" << p.id << "_s', :newRow.\"dir\"), :newRow.\"guid\"";
    else
      query << "'dir" << p.id << "', :newRow.\"guid\"";
    query << ");";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Could not alter table", statement);
      return "";
    }
    
    query.str("");
    query << "CREATE TRIGGER tdelguid" << p.id << " AFTER DELETE ON dir" << p.id 
	  << " REFERENCING OLD AS oldRow FOR EACH ROW "
	  << " BEGIN "
	  << "   DELETE FROM " << guidsTable << " WHERE \"guid\"=:oldRow.\"guid\";"
	  << " END;";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Could not alter table", statement);
      return "";
    }
  }

  if(commitDirectoryToIndex(p.id, p.indexTable, statement, p.flags)<0){
    return "";
  }
  
  if( statement.commitTransaction() ) {
    printError("9 Cannot commit transaction", statement);
    return "";
  }
  
  query.str("");
  query << "dir" << p.id;

  return query.str();
}

int MDOracleServer::dumpIndexes(Statement &statement, const std::string &directory, const std::string &tablename) {

  // TODO: dump indexes
  DMESG("dumpIndexes: " << tablename << endl);

  std::vector< std::string > result;
  boost::algorithm::split(result, tablename, boost::is_any_of("."));

  std::string dir2;
  if( result.size() == 1) dir2 = tablename;
  else dir2 = result.at(1);
  boost::to_upper(dir2);

  // Build the select query
  ostringstream oss;

  oss << "SELECT INDEX_NAME, COLUMN_NAME FROM user_ind_columns "
      << " WHERE TABLE_NAME = '" << dir2 << "' ORDER BY INDEX_NAME";

  string query = oss.str();
  DMESG("SQL: >" << query << "<" << std::endl);
  if (statement.exec(query)) {
    ERRLOG("Could not dump indexes from " << tablename << std::endl);
    return 1;
  }

  char index_name[1024];
  char column_name[1024];
  std::string new_index_name;
  std::string new_column_name;
  statement.bind(1, index_name, 1024);
  statement.bind(2, column_name, 1024);

  try {
    while(!statement.fetch()){
      if(strcmp(column_name, "file") == 0){
        DMESG("skip dumping index file>" << tablename << std::endl);
        continue;
      }
      char str[4096];
      string index_name_str(index_name);
  
      boost::algorithm::split(result, index_name_str, boost::is_any_of
("_"));
      index_name_str = result.at(result.size()-1);
      boost::to_lower(index_name_str);
      for(unsigned int i=0; i<index_name_str.size(); i++){
        if(!isalnum(index_name_str[i])) 
          index_name_str[i] = 'x';
      }

      string column_name_str(column_name);
      boost::erase_all(column_name_str, "user:");

      if( (index_name_str == new_index_name) || 
          (new_index_name.size() == 0 ) ) {
        new_index_name = index_name_str;
        if (new_column_name.size()) 
          new_column_name.append(" ");
        new_column_name.append(column_name_str);
        continue;
      } 

      out.append("index_create ");
      out.append(new_index_name);
      out.append(" ");
      out.append(directory);
      out.append(" '");
      out.append(new_column_name);
      out.append("' ");
      out.append("\n");
      DMESG("dump index: >" << tablename << " : " << index_name_str << std::endl); 

      new_index_name = index_name_str;
      new_column_name = column_name_str;
    } 
  } catch (SQLException &ex) {
    DMESG("Error dumping indexes: " << ex.what() << endl);    return 1;
  }

  if ( new_index_name.size() != 0 ) {
    out.append("index_create ");
    out.append(new_index_name);
    out.append(" ");
    out.append(directory);
    out.append(" '");
    out.append(new_column_name);
    out.append("' ");
    out.append("\n");
    DMESG("dump index: >" << tablename << " : " << new_index_name << std::endl);
  }
  statement.unBind();
  statement.close();
  return 0;
}

