/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDPgServer.cc,v 1.59 2011/01/26 05:57:13 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDPgServer.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"

#include "PgHelper.h"
#include "sqlHelpers.h"

#include <sqlext.h>

#include <sstream>
#include <iostream>
#include "debug.h"

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string.hpp>


using namespace std;

MDPgServer::MDPgServer(MDBuffer &outbuf)
  : MDStandalone(outbuf)
{
  dbHelper = new PgHelper();
  parser->setEngine(dbHelper->backend());
}


MDPgServer::~MDPgServer()
{
  delete dbHelper;
}


int MDPgServer::lockTable(Statement &statement,
			  const std::string &indexTable)
{
  DMESG("Locking table\n");
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return -1;
  }

  DMESG("SQL: >" << "LOCK" << "<" << std::endl);
  std::string query="LOCK TABLE ";
  query.append(indexTable).append(" IN EXCLUSIVE MODE;");
  if(statement.exec(query)){
    printError("9 Error acquiring lock", statement);
    return -1;
  }
  DMESG("SQL: >" << "ACQUIRED LOCK" << "<" << std::endl);
  return 0;
}


std::string MDPgServer::matchSubdirs(const std::string &dir) const
{
  std::string query="\"directory\" ~ '^";
  if(dir.size()>1)
    query.append(dir);
  query.append("/[^/]+$'");
  return query;
}


bool MDPgServer::dirPattern(const std::string &in, std::string &out) const
{
  return dirRegexpPattern(in, out);
}


//! Creates a table for a directory
/** Creates a table for a directory which must already exist in an
    index table. The created table is inserted into the index table.
    If parent is given, the created table inherits the
    schema of the parent table.
    \param p EntryProps properties of the table. The following fields must be filled:
          id The id number of the directory
          indexTable Name of directory index for table
          parent The name of the table which is inherited
          mainColumn Name of the main table column
          flags      Flags of the table.
    \param attrs A list of initial attributes of the list
    \param statement The current statement, so that we can continue a transaction.
    \return Returns the name of the table which was created
    Any error printout is done by this function
*/
std::string MDPgServer::createTable(const EntryProps &p,
				    AttrList &attrs,
				    Statement &statement)
{
  DMESG("In createTable, id: " << p.id << ", indexTable: " << p.indexTable 
	<< ", flags: " << p.flags << ", type >" << p.mainType << "<\n");
  if( statement.beginTransaction() ) {
    printError("9 Cannot begin transaction", statement);
    return "";
  }

  std::string mainType = typeFromFlags(p.flags);
  if( !p.mainType.empty() ) {
    mainType = p.mainType;
    DMESG("Overriding main Type");
  }
  std::stringstream query;

  if( !(p.flags & EP_INHERIT) ) {
    query << "CREATE TABLE dir" << p.id << " (";
    if( !(p.flags & EP_PLAIN) ) {
      query << "\"" << p.mainColumn << "\" " << dbHelper->mdToDBType(mainType);
      if (p.flags & EP_SHARED) {
        query << ", \"dir\" INTEGER REFERENCES " 
	    << p.indexTable << "(\"id\")"
	    << " ON DELETE CASCADE";
      }
      if( p.flags & EP_PERMS ) {
        query << ", \"owner\" VARCHAR(" << USERNAME_MAX_SIZE-1 << "),"
  	    << " \"permissions\" CHAR(3),"
	    << " \"group_rights\" CHAR(3)";
      }
      if( p.flags & EP_FCAT ) {
        query << ", \"link\" VARCHAR(1025),"
	    << " \"created\" TIMESTAMP WITHOUT TIME ZONE,"
	    << " \"size\" BIGINT,"
	    << " \"guid\" BYTEA,"
	    << " \"md5\" VARCHAR(32)";
      }
      if( p.flags & EP_ACLS )
        query << ", \"acls\" VARCHAR(2048)";
    }

    AttrList::const_iterator I=attrs.begin();    
    while(I!=attrs.end()) { 
      std::string key;
      if(parser->parse(I->first, key,p.flags & EP_PLAIN)){
	out.append("7 Illegal Key\n");
	return "";
      }
      if( !(p.flags & EP_PLAIN ) || I!=attrs.begin() )
	query << ", ";
      query << key << " " << (dbHelper->mdToDBType((*I).second));
      I++;
    }
    
    if( !(p.flags & EP_PLAIN) ) {
      if( p.flags & EP_SHARED )
	query << ", PRIMARY KEY (\"" << p.mainColumn << "\", \"dir\")";
      else {
	query << ", PRIMARY KEY (\"" << p.mainColumn << "\")";
      }
    }

    query  << ");";

  } else {
    query << "CREATE TABLE dir" << p.id << " () INHERITS (" << p.inheritsFrom << ");";
  }

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Could not alter table", statement);
    return "";
  }

  // If FC table, we need some triggers
  if( p.flags & EP_FCAT ) {
    query.str("");
    query << "CREATE TRIGGER tnewguid AFTER INSERT ON dir" << p.id 
	  << " FOR EACH ROW EXECUTE PROCEDURE newguid(";
    if (p.flags & EP_SHARED)
      query << "'_s'";
    else
      query << "''";
    query << ");";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Could not alter table", statement);
      return "";
    }
    
    query.str("");
    query << "CREATE TRIGGER tdelguid AFTER DELETE ON dir" << p.id 
	  << " FOR EACH ROW EXECUTE PROCEDURE delguid();";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Could not alter table", statement);
      return "";
    }
  }
  if(commitDirectoryToIndex(p.id, p.indexTable, statement, p.flags)<0){
    return "";
  }
  if( statement.commitTransaction() ) {
    printError("9 Cannot commit transaction", statement);
    return "";
  }
  query.str("");
  query << "dir" << p.id;

  return query.str();
}


bool MDPgServer::alterTableColumns(const std::string &tableName,
				   const AttrList &keyList,
				   bool remove)
{
  DMESG("List-size: " << keyList.size() << std::endl);

  std::string query="ALTER TABLE ";
  query.append(tableName);
  if(remove)
    addAttributeList(keyList, query, " DROP COLUMN ");
  else
    addTableDescription(keyList, query, " ADD COLUMN ");
  query.append(";");

  Statement statement(*dbConn);
  DMESG("SQL: " << query << std::endl);
  if(statement.exec(query)){
    printError("9 Could not alter table", statement);
    return false;
  }
  return true;
}


void MDPgServer::doBulkRead(Statement &statement, std::string &query,
			    size_t fields, bool &foundSomething)
{
  DMESG("doBulkRead\n");
  char buff[1001];
  if( statement.beginTransaction(true) ) {
    printError("9 Internal error", statement);
    foundSomething=true;
    return;
  }

  if(statement.exec("DECLARE getattr_c CURSOR FOR "+ query)){
    printError("9 Internal error", statement);
    foundSomething=true;
    return;
  }

  DMESG("EXEC done" << std::endl);
//  sleep(50);

  while(1){
    DMESG("GOING to FETCH"<< std::endl);
    query="FETCH 100 FROM getattr_c;";
    if(statement.exec(query)){
      printError("9 TERRIBLE ERROR fetching from cursor", statement);
      DMESG("Error FETCH" << std::endl);
      foundSomething=true;
      break;
    }

    if(statement.fetch()){
      DMESG("Fetch failed: Last data read" << std::endl);
      break;
    }
    DMESG("FETCH done" << std::endl);
    if(!foundSomething) {
      out.append("0");
      char buff[20];
      sprintf(buff, " %d\n", fields);
      out.append(buff);
    }
    foundSomething = true;

    SQLHSTMT s=statement.getStatement();

    SQLLEN err;

    do {
      for(size_t i=0; i<fields; i++){
        while(SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err) != SQL_NO_DATA){
	  if(err==SQL_NULL_DATA) break;
	  std::string line(buff);
	  AMGA::encodeLine(line);
	  out.append(line);
	}
	out.append("\n");
      }
    } while(!statement.fetch());
    statement.close();
  }
  statement.close();
  if(statement.exec("CLOSE getattr_c;")){
    printError("9 CLOSE does not work", statement);
    return;
  }

  statement.commitTransaction();
}


/** Returns the attributes of a table, if 
    \param all is true, then also the system attributes are returned
    in \param keys , otherwise only the user defined.
    Returns 0 on success, a non-null value otherwise.
*/
int MDPgServer::getAttributes(const std::string &table,
			      AttrList &keys,
			      bool all)
{
  // How to get the schema description SQL:
  // Copy schema: psql -E
  // \d <table>

  std::string schema = getSchema(table);

  std::string query="SELECT a.attnum, a.attname AS field, "
    "t.typname AS type, a.attnotnull, format_type(a.atttypid, a.atttypmod) "
    "FROM pg_class c, pg_attribute a, pg_type t, pg_namespace n "
    "WHERE c.relname = '" + stripSchema(table) 
    + "' and a.attnum > 0 and a.attrelid = c.oid AND a.atttypid = t.oid "
    "AND c.relnamespace = n.oid ";
  if( schema.size() )
    query += "and n.nspname = '" + getSchema(table) + "' ";
  query += "ORDER BY a.attnum;";

  char key[1024];
  char type[1024];

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Cannot read master description table", statement);
    return -1;
  }
  statement.bind(2, key, 1024);
  statement.bind(5, type, 1024);

  while(!statement.fetch()){
    if(!all
       && strncmp(key, parser->prefix().c_str(),
		  parser->prefix().size())
       )
      continue;

    DMESG(">>> " << key << " " << type << std::endl);

    if(!all)
      keys.push_back(Attribute(key+parser->prefix().size(), type));
    else
      keys.push_back(Attribute(key, type));
  }

  statement.unBind();
  statement.close();

  return 0;
}


/** vacuums a database
 */
void MDPgServer::vacuum(const std::string &options)
{
  std::string query="VACUUM ";
  query.append(options).append(";");

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Internal DB error", statement);
    return;
  }
  out.append("0\n");
}

//uint64_t MDPgServer::sequenceNextValue(Statement &st, const std::string &sequence)
//  throw (SQLException)
//{
//  std::string query="SELECT nextval('" + sequence + "');";
////  DMESG("SQL: >" << query << "<" << std::endl);
//  uint64_t v;
//  st.bind(1, &v);
//  if(st.exec(query)){
//    throw SQLException(statement);
//  }
//  if(st.fetch()){
//    throw SQLException(statement);
//  }
//  return v;
//}
//
//uint64_t MDPgServer::sequenceCurrentValue(Statement &st, const std::string &sequence)
//  throw (SQLException)
//{
//  std::string query="SELECT currval('" + sequence + "');";
////  DMESG("SQL: >" << query << "<" << std::endl);
//  uint64_t v;
//  st.bind(1, &v);
//  if(st.exec(query)){
//    throw SQLException(statement);
//  }
//  if(st.fetch()){
//    throw SQLException(statement);
//  }
//  return v;
//}
//
//uint64_t MDPgServer::sequenceCurrentValue(const std::string &sequence)
//  throw (SQLException) 
//{
//  return sequenceCurrentValue(Statement(*dbConn), sequence);
//}
//
//uint64_t MDPgServer::sequenceNextValue(const std::string &sequence)
//  throw (SQLException) 
//{
//  return sequenceNextValue(Statement(*dbConn), sequence);
//}

int MDPgServer::dumpIndexes(Statement &statement, const string &directory, const string &tablename) {
  // TODO: dump indexes
  DMESG("dumpIndexes: " << tablename << endl);

  std::vector< std::string > result;
  boost::algorithm::split(result, tablename, boost::is_any_of("."));

  std::string dir2;
  if( result.size() == 1) dir2 = tablename;
  else dir2 = result.at(1);

  // Build the select query
  ostringstream oss;

  oss << "select ";
  oss << "    gd.table_name2,  ";
  oss << "    gd.index_name2,  ";
  oss << "array_to_string ( array (  ";
  oss << "select a.attname as column_names ";
  oss << "from pg_class t, pg_class i, pg_index ix, pg_attribute a ";
  oss << "where ";
  oss << "t.oid = ix.indrelid ";
  oss << "and i.oid = ix.indexrelid ";
  oss << "and a.attrelid = t.oid ";
  oss << "and a.attnum = ANY(ix.indkey) ";
  oss << "and t.relkind = 'r' ";
  oss << "and t.relname = '" << dir2 << "' ";
  oss << "and i.relname = gd.index_name2), ' ') ";
  oss << "from ";
  oss << "( select t2.relname as table_name2, i2.relname as index_name2 ";
  oss << "from pg_class t2, pg_class i2, pg_index ix2 ";
  oss << "where ";
  oss << "t2.oid = ix2.indrelid ";
  oss << "and i2.oid = ix2.indexrelid ";
  oss << "and t2.relkind = 'r' ";
  oss << "and t2.relname = '" << dir2 << "' ";
  oss << ") gd  ";
  oss << "group by gd.table_name2, gd.index_name2; ";

  string query = oss.str();
  DMESG("SQL: >" << query << "<" << std::endl);
  if (statement.exec(query)) {
    ERRLOG("Could not dump indexes from " << tablename << std::endl);
    return 1;
  }

  char table_name[1024];
  char index_name[1024];
  char column_name[1024];
  statement.bind(1, table_name, 1024);
  statement.bind(2, index_name, 1024);
  statement.bind(3, column_name, 1024);

  try {
    while(!statement.fetch()){
      if(strcmp(column_name, "file") == 0){
        DMESG("skip dumping index file>" << table_name << std::endl);
        continue;
      }
      char str[4096];
      string index_name_str(index_name);

      boost::algorithm::split(result, index_name_str, boost::is_any_of("_"));
      index_name_str = result.at(result.size()-1);
      boost::to_lower(index_name_str);
      for(unsigned int i=0; i<index_name_str.size(); i++){
        if(!isalnum(index_name_str[i])) 
          index_name_str[i] = 'x';
      }

      string column_name_str(column_name);
      boost::erase_all(column_name_str, "user:");

      out.append("index_create ");
      out.append(index_name_str);
      out.append(" ");
      out.append(directory);
      out.append(" '");
      out.append(column_name_str);
      out.append("' ");
      out.append("\n");
      DMESG("dump index: >" << table_name << " : " << index_name_str << std::endl);
    }
  } catch (SQLException &ex) {
    DMESG("Error dumping indexes: " << ex.what() << endl);
    return 1;
  }
  statement.unBind();
  statement.close();
  return 0;
}

