/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDSQLiteServer.cc,v 1.17 2010/04/14 07:15:30 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDSQLiteServer.h"
#include "util/Statement.h"
#include "util/amgaProtocol.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"
#include "SQLiteHelper.h"

#include "sqlHelpers.h"
#include <sqlext.h>

#include <iostream>
#include "debug.h"

MDSQLiteServer::MDSQLiteServer(MDBuffer &outbuf)
  : MDStandalone(outbuf)
{
  dbHelper = new SQLiteHelper();
  parser->setEngine(dbHelper->backend());
}


MDSQLiteServer::~MDSQLiteServer()
{
  delete dbHelper;
}


int MDSQLiteServer::lockTable(Statement &statement, 
			      const std::string &indexTable)
{
  if(statement.beginTransaction(true)){
    printError("9 Internal error", statement);
    return -1;
  }
  return 0;
}


bool MDSQLiteServer::dirPattern(const std::string &in, std::string &out) const
{
  out = in;
//  bool pattern = false;
  size_t p;
  if( (p=in.find_first_of("*?"))!=std::string::npos) {
    if(!p || (in[p-1] != '\\') || !(p-1) || ( in[p-2] !='\\') )
      return true;
  }
  return false;
}


std::string MDSQLiteServer::matchSubdirs(const std::string &name) const
{
  std::string query="directory GLOB '";
  query.append(name);
  if( name == "/" )
    query.append("*'");
  else 
    query.append("/*'");
  return query;
}


/** This function returns the attributes of a directory.
    Make sure that the directory does exist. */
int MDSQLiteServer::getAttributes(const std::string &table,
				  AttrList &keys,
				  bool all)
{
  if (table == "x")
    return 0;

  std::string query="SELECT sql, name, type FROM SQLITE_MASTER WHERE name='";
  query.append(table).append("' ORDER BY type DESC;");
  
  char sql[1024];
  char name[128];
  char type[128];

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Cannot read master description table", statement);
    return -1;
  }
  statement.bind(1, sql, 1024);
  statement.bind(2, name, 128);
  statement.bind(3, type, 128);
  
  if(statement.fetch()){
    printError("9 Cannot read master description table", statement);
    statement.unBind();
    statement.close();
    return -1;
  }

  std::string m(sql);
  
  DMESG("Table created with >" << m << "<" << std::endl);
  size_t pos=m.find_first_of("(");
  DMESG("Pos" << pos << std::endl);
  while(pos<m.size() -1 ){  // the ")"
    DMESG("Pos: " << pos << std::endl);
    size_t newpos=m.find(", ", pos);
    pos++;
    if(newpos==std::string::npos || newpos-pos==0)
      newpos=m.size()-1;

    // Skip PRIMARY KEY Def
    if( m.find("PRIMARY KEY") == pos ) {
      pos = newpos +1;
      continue;
    }
    
    size_t midpos=m.find_first_of(" ", pos);
    std::string key=m.substr(pos, midpos-pos);
    std::string type= m.substr(midpos+1, newpos-midpos-1);
    if(!all && key[0]=='"')
      key=key.substr(1, key.size()-2);
    pos = newpos + 1;
    if(!all && key.substr(0, parser->prefix().size()) !=  parser->prefix())
      continue;

    if(!all)
      key=key.substr(parser->prefix().size());

    DMESG(">" << key << "<  >" << type << "<" << std::endl);
    keys.push_back(Attribute(key,type));
  }

  statement.unBind();
  statement.close();

  return 0;
}


bool MDSQLiteServer::alterTableColumns(const std::string &tableName,
				       const AttrList&keyList,
				       bool remove)
{
  AttrList keys;

  Statement statement(*dbConn);
  if(statement.beginTransaction(true)) {
    printError("9 Internal error", statement);
    return false;
  }

  if(getAttributes(tableName, keys, true))
    return false;

  AttrList newKeys(keys);
  if(remove) {
    AttrList::const_iterator I=keyList.begin();
    while(I!=keyList.end()){
      AttrList::iterator J=newKeys.begin();
      while(J!=newKeys.end()) {
	if(J->first==I->first) {
	  newKeys.erase(J);
	  break;
	}
	J++;
      }
      I++;
    }
  } else {
    AttrList::const_iterator I=keyList.begin();
    while(I!=keyList.end())
      newKeys.push_back(*I++);
  }

  std::string query="CREATE TEMPORARY TABLE ";
  query.append(tableName).append("_ (");
  addTableDescription(newKeys, query);
  query.append(", PRIMARY KEY (\"file\")");
  query.append(");");
  DMESG("SQL: >" << query << "<" << std::endl);
  
  if(statement.exec(query)){
    printError("9 Cannot copy table (1)", statement);
    return false;
  }

  // Copy new columns into new table
  query="INSERT INTO ";
  query.append(tableName).append("_ (");
  if(remove)
    addAttributeList(newKeys, query);
  else
    addAttributeList(keys, query);
  query.append(") SELECT ");
  if(remove)
    addAttributeList(newKeys, query);
  else
    addAttributeList(keys, query);
  query.append(" FROM ").append(tableName).append(";");
  DMESG("SQL: >" << query << "<" << std::endl);
  
  if(statement.exec(query)){
    printError("9 Cannot copy table (2)", statement);
    return false;
  }

  query="DROP TABLE ";
  query.append(tableName).append(";");
  
  if(statement.exec(query)){
    printError("9 Cannot copy table (3)", statement);
    return false;
  }

  query="CREATE TABLE ";
  query.append(tableName).append(" (");
  addTableDescription(newKeys, query);
  query.append(", PRIMARY KEY (\"file\")");
  query.append(");");
  DMESG("SQL: >" << query << "<" << std::endl);
  
  if(statement.exec(query)){
    printError("9 Cannot copy table (4)", statement);
    return false;
  }

  query="INSERT INTO ";
  query.append(tableName).append(" (");
  addAttributeList(newKeys, query);
  query.append(") SELECT * FROM ").append(tableName).append("_;");
  DMESG("SQL: >" << query << "<" << std::endl);
  
  if(statement.exec(query)){
    printError("9 Cannot copy table (5)", statement);
    return false;
  }
  query="DROP TABLE ";
  query.append(tableName).append("_;");
  
  if(statement.exec(query)){
    printError("9 Cannot copy table (6)", statement);
    return false;
  }

  if(statement.commitTransaction()){
    printError("9 Cannot copy table (7)", statement);
    return false;
  }
  
  return true;
}


void MDSQLiteServer::doBulkRead(Statement &statement, std::string &query,
				size_t fields, bool &foundSomething)
{
  char buff[1001];
  if(statement.exec(query)){
    printError("9 Internal error", statement);
    foundSomething=true;
    return;
  }

  DMESG("EXEC done" << std::endl);
  while(1) {
    if(statement.fetch()){
      DMESG("Fetch failed: Last data read" << std::endl);
      break;
    }

    if(!foundSomething) {
      /* Fix 20090728 : Return #column instead of #rows */
      out.append("0");
      char buff[20];
      sprintf(buff, " %d\n", fields);
      out.append(buff);
    }
    foundSomething = true;

    SQLHSTMT s=statement.getStatement();
    SQLLEN err;

    for(size_t i=0; i<fields; i++){
      DMESG("Getting Field: " << i << std::endl);
      while(1) {
        int res=SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err);
        DMESG("Res: " << res << "   err: " << err << std::endl);
        if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO)
          break;
        if(err==SQL_NULL_DATA) break;
        std::string line(buff);
        AMGA::encodeLine(line);
        out.append(line);
      }
     out.append("\n");
    }
  }
  statement.close();

}


/** vacuums a database
 */
void MDSQLiteServer::vacuum(const std::string &options)
{
  std::string query="VACUUM ";
  query.append(options).append(";");

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Internal DB error", statement);
    return;
  }

  out.append("0\n");
}


//uint64_t MDSQLiteServer::sequenceNextValue(const std::string &sequence) 
//    throw (SQLException)
//{
//  throw SQLException("Not implemented");     
//}
