/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDServer.cc,v 1.162 2011/02/14 09:29:05 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDServer.h"
#include "util/Statement.h"
#include "util/DatabaseConnection.h"
#include "MDBuffer.h"
#include "queryparser/QParser.h"
#include "DBHelper.h"
#include "DBUserManager.h"
#include "VOMSAttrUserManager.h"
#include "GridMapUserManager.h"
#include "replicationdaemon/MDLogWriter.h"
#include "MDConMan.h"
#include "VOMSInfo.h"

#include "sqlHelpers.h"
#include <assert.h>
#include <iostream>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/shared_ptr.hpp>

#include "debug.h"

#define SQL_ERR_MSG_SIZE 1024

#define COMMAND_GROUPS 8
#define COMMANDS 57

int MDServer::serverSideFed;
int MDServer::allowFedConnFail;
int MDServer::allowFedExecFail;


using namespace std;

static const char commandGroups [COMMAND_GROUPS][3][50] = {
  {"0", "help", "Help"},
  {"1", "metadata", "Basic metadata commands"},
  {"2", "metadata-optional", "Optional metadata commands"},
  {"3", "directory", "Information on directories"},
  {"4", "replication", "Information on replication commands"},
  {"5", "constraints", "Constraints on attributes"},
  {"13", "index", "Index handling"},
  {"23", "federation", "Information on federation commands"}
};


static const char commands[COMMANDS][4][100] = {
  {"0", "help", "[topic]",
   "Displays help on a command or a topic"},
  {"1", "setattr", "entry attribute value [attribute value]...",
   "Sets given attributes to specified values for all entries matching entry"},
  {"1", "addattr", "dir attribute type",
   "Adds a new attribute to a directory"},
  {"1", "listattr", "entry",
   "Lists all attributes of an entry"},
  {"1", "removeattr", "dir attribute",
   "Removes an attribute from a directory"},
  {"1", "renameattr", "dir oldName newName",
   "Renames an attribute of a directory"},
  {"1", "clearattr", "entry attribute",
   "Sets the attribute to NULL for all entries matching entry pattern"},
  {"1", "find", "pattern 'query'",
   "Returns all files matching pattern for which query is true"},
  {"1", "getattr", "pattern attribute1 attribute2 ...",
   "Returns the values of the attributes for all files matching pattern"},
  {"1", "addentries", "entry1...",
   "Adds entries (also accross directories)"},
  {"1", "removeentries", "entry1...",
   "Removes entries (also accross directories)"},   
  {"1", "addentry", "entry1 [attr value]...",
   "Adds a new entry and initializes some attributes"},
  {"1", "listentries", "[directory/schema]",
   "Lists all entries of a directory or schema. Default is current directory"},
  {"1", "schema_create", "schema [attr-name attr-type]+ [option]",
   "Creates a schema with given attributes"},   
  {"1", "upload", "dir [attr-name]+",
   "Starts an upload of entries into the catalogue"},
  {"1", "put", "entry [values]+",
   "Inserts a new entry during upload"},   
  {"1", "abort", "",
   "Abort upload or transaction"},
  {"1", "commit", "",
   "Commit upload or transaction"},         
  {"1", "transaction", "",
   "Initiate a transaction"},
  {"2", "updateattr", "pattern attr expression [attr expression]... condition",
   "Updates attributes of all files matching pattern and condition"}, // 20
  {"2", "update", "pattern attr expression [attr expression]... condition",
   "Updates attributes of all files matching pattern and condition"},
  {"2", "selectattr", "attr... condition",
   "Returns the values of given attributes for all files matching condition"},
  {"3", "dir", "[directory]",
   "Lists all files and subdirectories of a directory. Default is current"},
  {"3", "cd", "directory",
   "Changes current working directory"},
  {"3", "pwd", "",
   "Prints current working directory"},
  {"3", "whoami", "",
   "Prints current user"},
  {"4", "rep_umount", "[-noerase] <remoteDirectory>",
   "Unmounts a replication point. Local data is deleted"},
  {"4", "rep_mount", "[-perms|-nosync] master remoteDirectory",
   "Mounts a directory from the server running at the given address"},
  {"4", "rep_mount_users", "master",
   "Replicate user and groups information from a remote node."},
  {"4", "rep_umount_users", "",
   "Drops the user and group information that was being replicated from a master"},
  {"4", "rep_list_mounts", "",
   "Lists all mount points"},   
  {"4", "rep_start_receive", "master",
   "Connects to the master and starts listening for the logs of all mount points from this master"},
  {"4", "rep_stop_receive", "master",
   "Stops listening for the logs of all mount points from this master"},   
  {"4", "rep_allow", "directory group",
   "Grants to group the right to replicate <directory> and all its subdirectories"},
  {"4", "rep_disallow", "directory group",
   "Remove from group the right to replicate directory and all its subdirectories"},
  {"4", "rep_users_allow", "user",
   "Grants to user the right to replicate login information for users"},
  {"4", "rep_users_disallow", "user",
   "Remove from user the right to replicate login information for users"},
  {"4", "rep_show_permissions", "",
   "Show the replication rights for all directories"},
  {"4", "rep_list_subscribers", "",
   "Show the current subscribers and their subscriptions"},
  {"4", "rep_drop_subscriber", "<subscriberID>",
   "Drops a subscriber and all related subscriptions"}, // 40
  {"4", "proxy_umount", "<remoteDirectory>",
   "Unmounts a proxy point."},
  {"4", "proxy_mount", "master remoteDirectory",
   "Proxy mount a directory from the server running at the given address"},
  {"4", "proxy_list_mounts", "",
   "Lists all proxy points"},   
  {"5", "constraint_add_not_null", "directory attribute name",
   "Adds a named not NULL constraint on the given attribute of a table"},
  {"5", "constraint_add_unique", "directory attribute name",
   "Adds a named UNIQUE constraint on the given attribute of a table"},
  {"5", "constraint_add_reference", "directory attribute referred_attr name",
   "Adds a named foreign key constraint on the given attribute of table"},
  {"5", "constraint_add_check", "directory check name",
   "Adds a named check constraint on a directory"},
  {"5", "constraint_add_primary_key", "directory key(s)",
   "Changes the primary key to the expression given"},
  {"5", "constraint_list", "directory",
   "Lists all constraints on the attributes of a directory"},
  {"5", "constraint_drop", "directory name",
   "Drops the constraint with the given name of a directory"}, // 50
  {"13", "index_create", "name dir 'key_1 ... key_n' [type]",
   "Creates a new index(of given type) named name for certain keys on a directory"},
  {"13", "index_remove", "path",
   "Removes an index"},
  {"23", "fed_umount", "<remoteDirectory>",
   "Unmounts a federation point"},
  {"23", "fed_mount", "<site> <remoteDirectory>",
   "Mounts a directory from the site"},
  {"23", "fed_list_mounts", "", 
   "Lists all mount points"},
  {"23", "fed_policy", "", 
   "Shows the current policy (AllowFedConnFail, AllowFedExecFail"},
  {"23", "fed_check_failure", "", 
   "Shows whether the last command had failures"}

};


MDServer::MDServer(MDBuffer &outbuf)
  : currentDir("/"), 
    parser(NULL), out(outbuf), connected(false),
    siteName(""), master(""), replicating(false), synchronizing(false),
    logging(0), ownsConnection(false), mdfed(NULL)
{
  parser=new QParser("user:", false); // Set to true for debugging of parser!
  
  logWriter.reset(new MDLogWriter());

  // Add help index entries
  unsigned int maxGroup = 0;
  for(unsigned int i = 0; i < COMMANDS; i++ ) {
    struct HelpEntry e;
    e.topic = atol(commands[i][0]);
    maxGroup = maxGroup < e.topic? e.topic : maxGroup;
    e.command = commands[i][1];
    e.args = commands[i][2];
    e.help = commands[i][3];
    helpEntries.push_back(e);
  }

  // Create help topics
  if (maxGroup + 1 > helpTopics.size() )
    helpTopics.resize(maxGroup+1);

  for(int i = 0; i < COMMAND_GROUPS; i++ )
    helpTopics[atol(commandGroups[i][0])]=commandGroups[i][1];
}


MDServer::~MDServer() {
    if (ownsConnection) {
        delete dbConn;
    }
    delete parser;
}


void MDServer::setupTableNames()
{
  std::string prefix;
  if(dbSchema.size())
    prefix = dbSchema + ".";

  amgaTable = prefix + "amga";
  constraintsTable = prefix + "constraints";
  usersTable = prefix + "users";
  groupsTable = prefix + "groups";
  masterindexTable = prefix + "masterindex";
  certificatesTable = prefix + "certificates";
}


void MDServer::flushOutputBuffer(std::string &result)
{
  out.flush(result);
}


int MDServer::initializeUserManager(std::list< MDUserManager *> &uMs)
{
  DBUserManager *dbUserManager;
  VOMSAttrUserManager *vomsUserManager;

  std::list< MDUserManager* >::const_iterator J;
  J=uMs.begin();

  while( J != uMs.end() ) {
    MDUserManager *uM = *(J);
    if((dbUserManager = dynamic_cast<DBUserManager *>(uM))) {
      DMESG("DBUser Setting Connection \n");
      dbUserManager->setConnection(dbConn);
    }
    if(vomsUserManager = dynamic_cast<VOMSAttrUserManager *>(uM)) {
      DMESG("VOMSAttrUser Setting Connection \n");
      vomsUserManager->setConnection(dbConn);
    }
    J++;
  }
  userManagers = uMs;

  return 0;
}


void MDServer::printError(std::string errnum, std::exception &ex)
{
    LOG("Error: " << ex.what() << endl);
    ostringstream os;
    os << errnum << " ";
    if (ex.what() == NULL) {
        os << "No detail available";
    } else {
        os << ex.what();
    }
    os << "\n";
    out.append(os.str());
}

  
void MDServer::printError(std::string errnum, Statement &statement) const
{
  int err = statement.getErrorCode();
  std::string msg = statement.getError();
  std::string sqlErr = statement.getSqlState();
  // Do something with error here
  if(atoi(errnum.c_str())==9) {
    LOG(errnum << "  " << err << " " << sqlErr << ": " << msg << std::endl);
  } else {
    // DMESG does not work in this function...
    LOG(errnum << "  " << err << " " << sqlErr << ": " << msg << std::endl);
  }

  for(unsigned int i =0; i<msg.size(); i++)
    if( msg[i]=='\n' ) msg[i]=' ';

  string reply(errnum);
  if(msg.size()>2) {
    reply += " ";
    reply += msg;
  }
  reply += "\n";

  /* modified by SUNIL                                */
  /* To remove previously stored messages to a client */
  out.start();
  
  out.append(reply);
}


void MDServer::setDBConnection(DatabaseConnection *connection,
			       bool ownsConnection)
{
  DMESG("Setting DB Connection. Owns? " << ownsConnection << std::endl);
  this->dbConn = connection;
  connected = true;    
  this->ownsConnection = ownsConnection;
}

/** Prints out all commands
    help
*/
void MDServer::help(const std::string &about)
{
  if(!about.size()) {
    out.append("0\n" + helpEntries.begin()->command + " " 
      + helpEntries.begin()->args + "\n" + helpEntries.begin()->help + ".\n");
  }

  std::list< struct HelpEntry >::const_iterator I;

  if(about == "commands") {
    out.append("0\n");
    for (I = helpEntries.begin(); I != helpEntries.end(); I++) {
      out.append((*I).command + " " + (*I).args + "\n");
    }
    return;
  }

  if(about.size()) {
    for (I = helpEntries.begin(); I != helpEntries.end(); I++) {
      if( (*I).command == about ) {
	out.append("0\n" + (*I).command + " " + (*I).args + "\n  " + (*I).help + ".\n");
	return;
      }
    }
    for (size_t i = 0; i< helpTopics.size(); i++) {
      if(helpTopics[i] == about) {
	out.append("0\n");
	for (I = helpEntries.begin(); I != helpEntries.end(); I++) {
	  if( (*I).topic == i ) {
	    out.append((*I).command + " " + (*I).args + "\n  " + (*I).help + ".\n");
	  }
	}
	return;
      }
    }
    out.append("22 No such help topic. ");
  }

  out.append("Valid topics are: ");

  for(size_t i = 0; i< helpTopics.size(); i++) {
    if(helpTopics[i].size()) {
      out.append(helpTopics[i] + " ");
    }
  }
  out.append("commands\n");
}



int MDServer::setUser(const std::string &u)
{
  user=u;
  return 0;
}


int MDServer::setDir(const std::string &dir)
{
  currentDir=dir;
  if(!currentDir.size())
    currentDir="/";
  if(currentDir[currentDir.size()-1]!='/')
    currentDir.append("/");
  return 0;
}


int MDServer::checkAttribute(const std::string &attr,
			     std::string &out, bool allowSystem,
			     bool plain) const
{
  if( !attr.size() || (attr.size() >= COLUMN_NAME_MAX_SIZE) )
    return -1;

  if( !isalpha(attr[0]) )
    return -1;

  for( size_t i =0; i< attr.size(); i++){
    if( !(isalnum(attr[i]) || attr[i]=='_') )
      return -1;
  }
  
  if( allowSystem ) {
    if( attr == "CREATED" ) {
      out += "\"created\"";
      return 0;
    }
    if( attr == "GUID" ) {
      out += "\"guid\"";
      return 0;
    }
    if( attr == "SIZE" ) {
      out += "\"size\"";
      return 0;
    }
    if( attr == "MD5" ) {
      out += "\"md5\"";
      return 0;
    }
    if( attr == "PERMISSIONS" ) {
      out += "\"permissions\"";
      return 0;
    }
    if( attr == "GROUP_RIGHTS" ) {
      out += "\"group_rights\"";
      return 0;
    }
    if( attr == "ACL" ) {
      out += "\"acls\"";
      return 0;
    }
    if( attr == "FILE" ) {
      out += "\"file\"";
      return 0;
    }
    if( attr == "LINK" ) {
      out += "\"link\"";
      return 0;
    }
    if( attr == "OWNER" ) {
      out += "\"owner\"";
      return 0;
    }
  }
  
  if( !plain )
    out += std::string("\"") + parser->prefix() + attr + "\"";
  else
    out += attr;

  return 0;
}


int MDServer::checkAttributes(const std::string &attrs,
			      std::string &out, bool allowSystem,
			      bool plain) const
{
  std::list< std::string > attributes;
  if (!attrs.empty())
    split(attributes, attrs, boost::is_any_of(","), boost::token_compress_on);  
 
  std::list< std::string >::const_iterator I = attributes.begin();
  while( I != attributes.end() ) {
    if( checkAttribute(*I, out, allowSystem, plain) )
      return -1;
    I++;
    if (I != attributes.end() )
      out += ", ";
  }
  
  return 0;
}


//! Handle errors of getEntryProps()
/** Writes out the errors that can possibly occour in getEntryProps()
    to out. Returns 0 if no error found, otherwise -1.
*/
int MDServer::handleEntryPropErrors(int error, const std::string &file)
{
  if(file.size()) {
    switch (error){
    case 9:
      out.append("9 Internal error processing: " + file + "\n");
      return -1;
    case 1:
      out.append("1 File or directory not found: " + file + "\n");
      return -1;
    case 4:
      out.append("4 Permission denied: " + file + "\n");
      return -1;
    case 100:
      return -1;
    }
    return 0;
  }

  switch (error){
  case 9:
    out.append("9 Internal Error\n");
    return -1;
  case 1:
    out.append("1 File or directory not found\n");
    return -1;
  case 4:
    out.append("4 Permission denied\n");
    return -1;
  case 100:
    return -1;
  }
  return 0;
}


/// Check whether the user has the given right on the entry
/** Checks whether the rights on the entry are enough to do something
    with it. Returns 2 if owner permission enough, 1 if it's the
    ACLs. In case of error negative values are returned, in
    case of permission denied 0.
*/
int MDServer::checkPermissions(const struct EntryProps &p,
			       const char right,
			       std::list< std::string > &groups)
{
  DMESG("checkPermssions for "<< p.name << " (" << right << ")\n");
  DMESG("User : "<< user << "\n");
  DMESG("Owner: "<< p.owner << "\n");


  if (replicating)
    return 1; // When replicating, everything is allowed

  // If the directory is replicated, the w permission is denied.
  // Non replicated directories have the master field set to MASTER
  if (right == 'w' && p.master != "") {
    DMESG("Master: >" << p.master << "<\n");
    DMESG("Write permission denied. Directory is replicated\n");
    return -2;
  }

  if(user=="root" || user == "0")
    return 1;

  if( right=='a' && user==p.owner )
    return 1;

  if( right=='t') {
    if(user==p.owner ) {
      return 1;
    } else {
      return 0;
    }
  }

  if ( right == 'r' || right == 'w' || right == 'x' ) {
    // Check permissions, setup:
    int offset=0;
    if (right=='w')
      offset=1;
    if (right=='x')
      offset=2;

    // Check user based permissions
    DMESG("user: >" << user << "<  owner: " << p.owner << std::endl);
    if(user==p.owner){
      DMESG(right << " vs " << p.perms[offset] << std::endl);
      if(p.perms[offset]==right) return 2;
      if( (right == 'x') && (p.perms[offset] == 't') ) return 2;
      return 0;
    }
  }

  DMESG("ACL string: " << p.ACLs << std::endl);
  std::map<std::string, std::string> aclMap;
  aclStringToMap(p.ACLs, aclMap);

  if(!groups.size()) {
    memberOf(user, groups);
    groups.push_back("system:anyuser");
  }

  std::list< std::string >::const_iterator I=groups.begin();
  while(I!=groups.end()){
    DMESG("Group perm: " << (*I) << " is " << aclMap[*I] << std::endl);
    if( aclMap[*I].find_first_of(right) != std::string::npos )
      return 1;
    I++;
  }

  return 0;
}


/// Check whether the user has the given right on the parent dir
/** Checks whether the rights on the parent dir  are enough to do
    something with it. Returns
    2 if owner permission enough,
    1 if it's the parent ACLs. 
    0 permission denied 
    -2 permission denied because entry/directory is a replica
*/
int MDServer::checkDirPermissions(const struct EntryProps &p, const char right,
				  std::list< std::string > &groups)
{
  DMESG("checkDirPermssions for "<< p.name << " (" << right << ")\n");
  if (replicating) // When replicating, everything is allowed
    return 1; 

  // If the directory is replicated, the w permission is denied.
  // Non replicated directories have the master field set to ""
  if (right == 'w' && p.master != "") {
    DMESG("Write permission denied. Directory is replicated\n");
    return -2;
  }

  if( user == "root" || user == "0" )
    return 1;

  if( right=='a' && user==p.owner )
    return 1;

  if ( right == 'r' || right == 'w' || right == 'x' ) {
    // Check permissions, setup:
    int offset=0;
    if (right=='w')
      offset=1;
    if (right=='x')
      offset=2;
    
    // Check user based permissions
    DMESG("user: >" << user << "<  parent owner: "
	  << p.parentOwner << std::endl);
    if(user==p.parentOwner){
      DMESG(right << " vs " << p.parentPerms[offset] << std::endl);
      if(p.parentPerms[offset]==right) return 2;
      return 0;
    }
  }

  DMESG("ACL string: " << p.parentACLs << std::endl);
  std::map<std::string, std::string> aclMap;
  aclStringToMap(p.parentACLs, aclMap);

  if(!groups.size()) {
    memberOf(user, groups);
    groups.push_back("system:anyuser");
  }

  std::list< std::string >::const_iterator I=groups.begin();
  while(I!=groups.end()){
    DMESG("Group perm: " << (*I) << " is " << aclMap[*I] << std::endl);
    if( aclMap[*I].find_first_of(right) != std::string::npos )
      return 1;
    I++;
  }

  return 0;
}


bool MDServer::checkPermResult(int error) {
  if (error > 0) {
      return true;
  }
  string str;
  switch (error) {
    case 0:
      str = "4 Permission denied.\n"; 
      break;      
    case -2:
      str = "4 Permission denied. Cannot update a replica.\n"; 
      break;      
    default:
      ostringstream oss;
      oss << "9 Internal error. Code: " << error << ".\n";
      str = oss.str();
      break;
  }
  DMESG(str);
  out.append(str);
  return false;
}


/** Looks up the tables by their name in the masterindex on the
    database and fills the TableMap structure with their properties
    It is checked that the 'x' permission is set for the tables.
    Users of the table map need to make sure that they check
    additional r/w permissions for the entries accessed.
*/
int MDServer::setupTableMap(TableMap &tableMap)
{
  TableMap::iterator I=tableMap.begin();
  while(I!=tableMap.end()) {
    EntryPropList propList;

    std::string directory = I->first;
    int subdirSearch = 0;
    // If the dirctory name ends in a +, then we search also shared subdirs
    if( directory[directory.size()-1] == '+' ) {
      directory.resize(directory.size()-1);
      subdirSearch = 1;
    }

    int error=getEntryProps(directory, propList);
    if(handleEntryPropErrors(error, directory))
      return -1;

    EntryProps p(*propList.begin());
    
    if( (p.type() != EPT_DIR) && (p.type() != EPT_VIEW)) {
      out.append("1 Not a directory: " + I->first + "\n");
      return -1;
    }
    
    std::list< std::string > groups;
    if(checkPermissions(p, 'x', groups)<=0) {
      out.append("4 Permission denied: " + I->first + "\n");
      return -1;
    }

    I->second = TableEntry(p, subdirSearch);
    I++;
  }

  return 0;
}


void MDServer::removeDoubleTables(TableMap &tableMap)
{
  std::map< std::string, std::string > tables;

  TableMap::iterator I=tableMap.begin();
  while( I != tableMap.end() ) {
    std::string table(I->second.props.directoryTable);
    if( tables.find(table) != tables.end() ) { // Ouch this is ugly and scales like n^2
      tableMap.erase(I);
      I = tableMap.begin();
      tables.clear(); 
    } else{
      tables[table]=table;
      I++;
    }
  }
}


/** Creates a TableMap by filling in the names of the directories
    The name of the tables and the properties of the entries need to
    be added in a second step doing setupTableMap
 */
int MDServer::createTableMap(const std::vector<std::string> &attributes,
			     TableMap &tableMap)
{
  std::vector<std::string>::const_iterator I=attributes.begin();
  while(I!=attributes.end()){
    std::string key;
    if(parser->parse(*I, key)){
      out.append("7 Illegal Key ");
      out.append(*I);
      out.append("\n");
      return -1;
    }
    parser->addToTableMap(tableMap);
    I++;
  }
  return 0;
}


/** Returns the (database) names of the distinct tables in the map
 */
void MDServer::getDistinctTables(TableMap &tableMap,
				 std::list<std::string> &distinctTables)
{
  DMESG("Distinct tables called for " << tableMap.size() << '\n');

  TableMap::iterator I=tableMap.begin();
  while(I != tableMap.end()) {
    std::list<std::string>::iterator J=distinctTables.begin();
    while(J != distinctTables.end()){
      if(*J == I->second.props.directoryTable)
	break;
      J++;
    }
    if( J== distinctTables.end()) {
      distinctTables.push_back(I->second.props.directoryTable);
    }
    I++;
  }
}


/// Helper function, mainly for alterTableColumns()
/** Creates a comma separated list of the columns (\param keys) of a
    table to the \param query string. Before every column name, the
    string in \param m is inserted.
*/
void MDServer::addAttributeList(const AttrList &keys,
				std::string &query,
				const std::string m,
				bool withComma)
  const
{
  AttrList::const_iterator I=keys.begin();
  while (I!=keys.end()) {
    query.append(m).append(I++->first);
    if(I!=keys.end() && withComma)
      query.append(", ");
  }
}


/// Helper function, mainly for alterTableColumns()
/** Creates a comma separated list of the columns (\param keys) and
    their types of a table to the \param query string. Before every
    column name, the string in \param m is inserted. The string is the
    same as used for a standard SQL CREATE TABLE statement.
*/
void MDServer::addTableDescription(const AttrList &keys,
				   std::string &query,
				   const std::string m,
				   bool withComma)
  const
{
  AttrList::const_iterator I=keys.begin();

  while (I!=keys.end()) {
    query.append(m);
    query.append(I->first);
    query.append(" ");
    query.append(I++->second);
    if(I!=keys.end() && withComma)
      query.append(", ");
  }
}


/// Checks whether the table exists
bool MDServer::tableExists(std::string &tableName, int dType) const
{
  if(!tableName.size()){
    out.append("1 Directory not found\n");
    return false;
  }
  if( dType == EPT_SEQUENCE || dType == EPT_INDEX ){
    out.append("1 File not found\n");
    return false;
  }
  return true;
}


bool MDServer::tableUsesColumn(const std::string &tableName,
			       const std::string &key)
{

  std::string query="SELECT ";
  query.append(key);
  query.append(" FROM ").append(tableName);
  query.append(" WHERE ").append(key).append(" IS NOT NULL");

  if( !strcmp(dbHelper->backend(),"Oracle")){
    /* this feature is not tested */
    query.append(" AND rownum <= 1");
  } else {
    query.append(" LIMIT 1;");
  }

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    return false;
  }

  if( statement.fetch() ){
    DMESG("Table does not use column "<< key << std::endl);
    return false;
  }

  DMESG("Table uses column "<< key << std::endl);

  statement.close();

  return true;
}


void MDServer::pwd()
{
  out.append("0\n");
  out.append(currentDir);
  out.append("\n");
}


void MDServer::whoAmI()
{
  out.append("0\n");
  out.append(user);
  out.append("\n");
}


/// Checks whether the entry is in a subdir
bool MDServer::inSubdir(const std::string &name) const
{
  size_t pos=0;
  for(;;) {
    pos=name.find_first_of("/", pos+1);
    if(pos==std::string::npos)
      return false;

    if(pos==0)
      return true;

    if(name[pos-1] !='\\')
      return true;

    if(pos == 1)
      return false;

    if(name[pos-2] != '\\')
      return true;
  }
}


/** Create an attribute for a collection of files in a dir:
    addattr dir attribute type

    The directory is created if it does not exist (createdir),
    but the attribute must not exist. The type must be supported by the
    backend.
*/
void MDServer::addAttr(const std::string &dir,
		       AttrList &attributes)
{
  DMESG("addAttr called" << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());  
  if( p.type() != EPT_DIR ) {
    out.append("23 Not a directory\n");
    return;
  }
  
  std::list< std::string > groups;

  if( p.sticky() && !checkPermResult(checkPermissions(p, 't', groups)) )
    return;

  int perms = checkPermissions(p, 'w', groups); // Can only be directory!
  if (!checkPermResult(perms)) {
    return;
  }
  
  DMESG("Printing dir properties\n");
  p.print(debug);

  int dirID = p.parentID;
  if( p.flags )
    dirID = p.id;
  
  if(!p.directoryTable.size()){
    p.directoryTable=createTable(p, attributes, statement);
    DMESG("Got >" << p.directoryTable << "<\n");
    if(p.directoryTable.size()) {
      // [rep] Sucessful operation. Write the log
      if (saveLog(statement, getEntryDirectory(p))) {
          statement.commitTransaction();
          out.append("0\n");
      }
      // saveLog writes an error message to the out buffer
      return;
    }
  }
  if(!p.directoryTable.size())
    return;

  AttrList::iterator I=attributes.begin();
  while (I!=attributes.end()) {
    std::string key;
    if(checkAttribute(I->first, key, false, p.flags & EP_PLAIN)){
      out.append("7 Illegal Key\n");
      return;
    }
    I->second = dbHelper->mdToDBType(I->second);
    I++->first=key;
  }

  if(!alterTableColumns(p.directoryTable, attributes, false)){
    // Error output done by alterTableColumns
    return;
  }
  
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }  
}


void MDServer::renameAttr(const std::string &path,
			  const std::string &oldName,
			  const std::string &newName)
{
  DMESG("renameAttr called for " << path << "  rename " << oldName 
	<< " to "<< newName << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }

  EntryPropList propList;
  if(handleEntryPropErrors(getEntryProps(path, propList), path))
    return;

  DMESG("Checking for directory\n");
  EntryProps p(*propList.begin());  
  if( p.type() != EPT_DIR ) {
    out.append("23 Not a directory\n");
    return;
  }
  
  std::list< std::string > groups;

  if( p.sticky() && !checkPermResult(checkPermissions(p, 't', groups)) )
    return;

  if( !checkPermResult(checkPermissions(p, 'w', groups) ) )
    return;
  
  if(p.directoryTable=="x") {
    out.append("10 No such attribute\n");
    return;
  }

  DMESG("Checking attributes");

  std::string newColumn;
  if(checkAttribute(newName, newColumn, false, (bool) (p.flags & EP_PLAIN))){
    out.append("7 Illegal Key ");
    out.append(newName);
    out.append("\n");
    return;
  }

  
  std::string oldColumn;
  if(checkAttribute(oldName, oldColumn, false, (bool)(p.flags & EP_PLAIN))){
    out.append("7 Illegal Key ");
    out.append(oldName);
    out.append("\n");
    return;
  }

  stringstream query;
  // This works only on PostgreSQL and Oracle:
  query << "ALTER TABLE " << p.directoryTable << " RENAME COLUMN "
    << oldColumn << " TO " << newColumn << ";";
  
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }

  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
    statement.commitTransaction();  
    out.append("0\n");
  }

}


/** Lists all attributes set for a file
    \param path is the path to the file
    The output is streamed into the out class member. For a
    description see the protocol description.
    To read the attributes of an entry, you need read permissions on
    the parent directory of the entry.
*/
void MDServer::listAttr(const std::string &path)
{
  DMESG("listAttr called" << std::endl);

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;

  EntryProps p(*propList.begin());

  if(p.directoryTable == "" && (p.type() != EPT_DIR) ) {
    out.append("1 No such file or directory\n");
    return;
  }
  
  std::list< std::string > groups;
  int perms = p.flags ? checkPermissions(p, 'x', groups)
    : checkDirPermissions(p, 'x', groups);
  if (!checkPermResult(perms))
    return;

  AttrList keys;
  int res = 1;
  if( p.directoryTable != "" )
    res = getAttributes(p.directoryTable, keys, p.flags & EP_PLAIN);
  AttrList::const_iterator I=keys.begin();
  
  ostringstream oss;
  int rows=keys.size();
  switch(res){
  case 0: {
    oss << "0";
    /* Fix 20090728 : Return #column instead of #rows 
    if(rows>-1) {
        oss << " " << rows;
    }
    */
    oss << "\n";
    while(I!=keys.end()) {
      oss << (*I).first << "\n";
      oss << dbHelper->dbToMDType((*I).second) << "\n";
      I++;
    }
    break;
  }
  case 1:
    oss << "0\n"; // No table associated to schema
    break;
  case 9:
    break; // error handled by getAttributes()
  case -1:
    break; // error handled by getAttributes()
  default:
    oss << "9 Internal error\n";
  }
  out.append(oss.str());
}


int MDServer::removeColumnConstraints(Statement &statement, int id,
				      const std::string &column)
{
  std::stringstream query;
  query << "DELETE FROM " << constraintsTable << " WHERE \"table_id\"=" << id 
	<< " AND \"column\"='" << column << "';";
  DMESG("SQL: " << query.str() << endl);
  int err;
  if ( (err = statement.exec(query.str())) ) {
    if (statement.getErrorCode() == SQL_NO_DATA) {
      return 0;
    }
  }
  return err;
}


/** Removes an attribute from a directory
    Write permission on the directory is needed. Attributes which are
    set by any of the entries in the directory cannot be removed
    (clearAttr() can be used first).

    The output is streamed into the out class member. For a
    description see the protocol description.
    ERRORS:
*/
void MDServer::removeAttr(const std::string &dir,
			  AttrList &attributes)
{
  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
      
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());

  std::list< std::string > groups;

  if( p.sticky() && !checkPermResult(checkPermissions(p, 't', groups)) )
    return;

  int perms = (p.type() == EPT_DIR) ?
    checkPermissions(p, 'w', groups):
    checkDirPermissions(p, 'w', groups);
  if( !checkPermResult(perms) )
    return;
  
  if (attributes.empty()) {
    out.append("0\n");
    return;
  }
  
  AttrList::iterator I=attributes.begin();
  while(I!=attributes.end()) {
    std::string key;
    if(checkAttribute(I->first, key, false, p.flags & EP_PLAIN)){
      out.append("7 Illegal Key: " + key + "\n");
      return;
    }

    removeColumnConstraints(statement, p.id, I->first);

    I++->first=key;

    if(tableUsesColumn(p.directoryTable, key)){
      out.append("14 Attribute in use: " + key + "\n");
      return;
    }
  }

  if(!alterTableColumns(p.directoryTable, attributes, true))
    return; // Error handled by alterTableColumn

  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


void MDServer::update(const std::string &path,
		      std::vector< std::string > &attributes,
		      std::vector< std::string > &values,
		      const std::string &condition)
{
  return updateAttr(path, attributes, values, condition, false);
}


//! Atomically updates attributes of entries depending on a condition
/** Sets a list of attributes to a list of values for a file.
    The dimensions of the input vectors for the attributes and values
    need to be the same.
    \param path The relative or absolute path to the file
    \param attributes The vector of attributes to set
    \param values The vector of values
    \param condition The update condition

    ERRORS:
    - 1 File or directory not found
    - 3 Illegal command
      The attributes and values vectors don't match
    - 4 Permission Denied
*/
void MDServer::updateAttr(const std::string &path,
			  std::vector< std::string > &attributes,
			  std::vector< std::string > &values,
			  const std::string &condition,
			  bool eval)
{
  DMESG("updateAttr called, eval = " << eval << "\n");

  if(attributes.size() != values.size()) {
    out.append("3 Illegal command\n");
    return;
  }
  EntryPropList propList;
  
  Statement statement(*dbConn, debug);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }  
  
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string file;
  bool isPattern;
  if( p.type() == EPT_DIR ) {
    file="%";
    isPattern = true;
  } else {
    isPattern = sqlPattern(p.name, file);
  }

  if(!p.directoryTable.size()){ // Directory does not exist
    out.append("1 Directory not found\n");
    return;
  }

  std::list< std::string > groups;
  int perm=checkPermissions(p, 'w', groups);
  if( perm <= 0 && !(p.flags & EP_ACLS) ){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  if(parser->parse(condition)){
    out.append("8 Illegal Query\n");
    return;
  }
  TableMap tableMap;
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap

  std::string q;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q);

  std::string query = updateAttrQuery(p, file, isPattern,
				      attributes, values, q, eval);

  if( !query.size() )
    return; // Error handled in updateAttrQuery
  
  DMESG("SQL: >" << query << "<" << std::endl);

  // Make sure the vectors have the same scope as the associated statement
  std::vector <SQLLEN> types(values.size());
  std::vector <SQLINTEGER> sizes(values.size());
  int err;
  if( !eval ) {
    for(unsigned int i=0; i<values.size(); i++) {
      DMESG("Value " << i << ": >" << values[i] << "<\n");
      
      sizes[i]=values[i].size();
      //sizes[i]=0;
      types[i]=SQL_NTS;
      // types[i] = values[i].size();
      if (statement.bindParam(i+1, values[i].c_str(), sizes[i], &types[i])) {
	printError("9 Could not bind parameters for prepared statement", statement);
	return;
      }
    }    
    err = statement.exec(query);
  } else {
    err = statement.exec(query);
  }

  if( err ){
    if (statement.getErrorCode() != SQL_NO_DATA) {
      std::string msg = statement.getError();
      std::string sqlErr = statement.getSqlState();

      DMESG("Error is >" << msg << "< -> " << sqlErr << std::endl);

      if(sqlErr=="00000") {  // in postgres 7.3/sqlite
        printError("18 Update failed (1)", statement);
        return;
      }
      out.append("10 No such key: " + msg + "\n");
      return;
    } else { 
      // In PostgreSQL 8.1 with libpsqlodbc.so, version: 08.01.0200
      // we get SQL_NO_DATA
      printError("18 Update failed. No rows updated", statement);
      return;
    }
  } else { 
    // In MySQL, if no row updated
    if( statement.numRows() == 0 ) {
      out.append("18 Update failed. No rows updated\n");
      return;
    }

  }

  updatedRows = statement.numRows() ;

  // [rep] Successful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


void MDServer::cd(const std::string &dir)
{
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());

  if( (p.type() != EPT_DIR) && (p.type() != EPT_VIEW) ) {
    out.append("1 Not a directory\n");
    return;
  }
  
  std::list< std::string > groups;
  if (!checkPermResult(checkPermissions(p, 'x', groups))) {
    return;
  }

  currentDir=p.name;
  assert(currentDir.size());
  if(currentDir[currentDir.size()-1] != '/')
    currentDir.append("/");

  DMESG("Current Dir now: " << currentDir << std::endl);

  out.append("0\n");
}


void MDServer::sendToClient(const std::string &message) {
    out.append(message);
}

void MDServer::printODBCDriverVersion() 
{
    DMESG("ODBC Driver: " << dbConn->getODBCDriverName() 
	  << ", version: " << dbConn->getODBCDriverVersion() << std::endl);
}


bool MDServer::checkSubject(const std::string &usr,
			    const std::string &subject) const
{
  std::list < MDUserManager * >::const_iterator I = userManagers.begin();
  while (I != userManagers.end() )
    if( (*I++)->checkSubject(usr, subject) )
      return true;

  return false;
}


bool MDServer::checkPass(const std::string &usr,
			 const std::string &pass) const
{
  std::list < MDUserManager * >::const_iterator I = userManagers.begin();
  while (I != userManagers.end() )
    if( (*I++)->checkPass(usr, pass) )
      return true;

  return false;
}


bool MDServer::checkCert(const std::string &usr, X509 *cert,
			 std::string &vo, std::list< std::string > &groups,
			 std::string &eUser)
{
  LOG("checkCERT CALLED IN MDSERVER, this should not happen\n");

  std::list < MDUserManager * >::const_iterator I = userManagers.begin();
  while (I != userManagers.end() ) {
    if( (*I++)->checkCert(usr, cert, vo, groups, eUser) )
      return true;
  }
  return false;
}

/********************
 * Transaction support
 *********************/
void MDServer::transactionStart()
{
  DMESG("Starting transaction\n");
  Statement statement(*dbConn, debug);
  if(statement.beginTransaction()){
    printError("9 Could not start transaction", statement);
    return;
  }
  // Inform the LogWriter that we are starting a transaction. The
  // MDLogWriter will store a flag on the DatabaseConnection and
  // will save all logs generated during this transaction to the
  // transaction table instead of log table
  if (logWriter->getSaveLogs()) {
    try {
      logWriter->beginTransaction(statement, user, permissionMask, groupRightsMask);
    } catch (SQLException &ex) {
      printError("9 Could not write replication logs", statement);
      return;
    }
  }
  statement.detach();
  assert(!out.getSize());
  out.append("0\n");
}


void MDServer::transactionCommit() 
{
  DMESG("Committing transaction\n");
  // Tell the MDLogWriter to finish the transaction and provide it
  // with the context that is global to the whole transaction.
  logWriter->commit();
  Statement statement(*dbConn, debug);

  statement.attach();
  if(statement.commitTransaction()){
    printError("9 Could not commit", statement);
    return;
  }
  out.append("0\n");
}


void MDServer::transactionAbort()
{
  DMESG("Aborting transaction\n");
  logWriter->abort();
  Statement statement(*dbConn);
  statement.attach();
  if(statement.rollbackTransaction()){
    printError("9 Abort failed", statement);
    return;
  }
  
  // the transaction on the MDLogWriter is automatically rolledback
  out.append("0\n");
}


bool MDServer::transactionInProgress() const {
  return dbConn->transactionInProgress();
}

/*********************
 * Replication
 ********************/
 
/**
 * If master is different than "", then this instance assumes that
 * it is being used to executed replication commands received from
 * the given master.
 */
void MDServer::setMaster(const std::string &master) {
  this->master = master;
}

std::string MDServer::getMaster() const {
  return master;
}

void MDServer::setNextCommand(const std::string &nextCmd) {
  nextCommand = nextCmd;
  logWriter->setNextCommand(nextCmd);
}
  
void MDServer::getResult(vector<string> &results) throw (runtime_error) {
  results.clear();
  // Read the result from the buffer
  string result;
  out.read(result);    
  
  // Checks if it is a valid result
  size_t index = result.find_first_of('\n');
  if (index == string::npos) {
    throw runtime_error("Invalid answer: " + result);
  }    
  // If it does not start with 0, then it was an error
  if (result[0] != '0') {        
    // Must remove the \n at the end of result
    throw runtime_error(result.substr(0, index));
  }       
  // The operation executed successfully. Skip the first line and split 
  // the rest of the result
  result = result.substr(index+1);
  if (result != "") {
    boost::split(results, result, boost::is_any_of("\n"), boost::token_compress_on);
  }
}


/**********************
 * Constraint handling
 *********************/

std::string MDServer::getTypeOfAttr(const EntryProps &p, 
				    const std::string &attr)
{
  std::string keyType;
  AttrList keys;
  if( p.flags & EP_PLAIN )
    getAttributes(p.directoryTable, keys, true); // all needed for plain tables
  else
    getAttributes(p.directoryTable, keys);
  AttrList::const_iterator I=keys.begin();    
  while( I!=keys.end() ) { 
    std::string key;
    DMESG(I->first << ": " << I->second << "\n");
    if( I->first == attr )
      keyType = I->second;
    I++;
  }

  return keyType;
}


void MDServer::constraintAddNotNull(const std::string &dir,
				    const std::string &attr,
				    const std::string &name)
{
  DMESG("constraintAddNotNull called: ..." << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());
  if( p.type() != EPT_DIR ) {
    out.append("1 No such directory " + dir + "\n");
    return;
  }
  
  if( p.directoryTable == "" ){
    out.append("7 Illegal key " + attr + "\n");
    return;
  }
  
  std::stringstream query;
  
  query << "INSERT INTO " << constraintsTable
	<< " (\"table_id\", \"name\", \"type\", \"column\")"
	<< " VALUES (" << p.id << ", '" << name << "', 'N', '"
	<< attr << "');";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("27 Constraint exists", statement);
    return;
  }

  query.str("");
  std::string key;
  if(checkAttribute(attr, key, true, p.flags & EP_PLAIN)){
    out.append("7 Illegal key\n");
    return;
  }

  if( !strcmp(dbHelper->backend(), "MySQL") ) {
    std::string keyType = getTypeOfAttr(p, attr);
     query << "ALTER TABLE " << p.directoryTable << " MODIFY COLUMN "
	  <<  key << " " << keyType << " NOT NULL;";
  } else if( !strcmp(dbHelper->backend(), "Oracle") ) {
    query << "ALTER TABLE " << p.directoryTable << " MODIFY "
	  <<  key << " NOT NULL;";
  } else {
    query << "ALTER TABLE " << p.directoryTable << " ALTER COLUMN "
	  <<  key << " SET NOT NULL;";
  }
  
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }
    
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


void MDServer::constraintAddUnique(const std::string &dir,
				   const std::string &attr,
				   const std::string &name)
{
  DMESG("constraintAddUnique called: ..." << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());
  if( p.type() != EPT_DIR ) {
    out.append("1 No such directory " + dir + "\n");
    return;
  }
  
  if( p.directoryTable == "" ){
    out.append("7 Illegal key " + attr + "\n");
    return;
  }
  
  std::stringstream query;
  
  query << "INSERT INTO " << constraintsTable
	<< " (\"table_id\", \"name\", \"type\", \"column\")"
	<< " VALUES (" << p.id << ", '" << name << "', 'U', '"
	<< attr << "');";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("27 Constraint exists", statement);
    return;
  }

  query.str("");
  std::string key;
  std::string keyLength;
  std::string a(attr);
  size_t pos1 = a.find_first_of("(");
  size_t pos2 = a.find_first_of(")");
  if( pos1 != std::string::npos && pos2 != std::string::npos ) {
    keyLength=a.substr(pos1, pos2-pos1+1);
    a = a.substr(0, pos1);
  }

  if(checkAttribute(a, key, true, p.flags & EP_PLAIN)){
    out.append("7 Illegal key\n");
    return;
  }

  query << "ALTER TABLE " << p.directoryTable << " ADD CONSTRAINT "
	<< name << " UNIQUE (" << key << keyLength << ");";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();
      out.append("0\n");
  }
}


void MDServer::constraintAddCheck(const std::string &dir,
				  const std::string &check,
				  const std::string &name)
{
  DMESG("constraintAddCheck called: ..." << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());
  if( p.type() != EPT_DIR ) {
    out.append("1 No such directory " + dir + "\n");
    return;
  }
  
  if( p.directoryTable == "" ){
    out.append("7 No attribute yet in table\n");
    return;
  }
  
  std::string constraint;
  if(parser->parse(check, constraint)){
    out.append("7 Illegal constraint\n");
    return;
  }

  std::stringstream query;
  query << "INSERT INTO " << constraintsTable
	<< " (\"table_id\", \"name\", \"type\", \"constraint\")"
	<< " VALUES (" << p.id << ", '" << name << "', 'C', '"
	<< constraint << "');";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("27 Constraint exists", statement);
    return;
  }

  query.str("");
  query << "ALTER TABLE " << p.directoryTable << " ADD CONSTRAINT " 
	<< name << " CHECK (" << constraint << ");";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }
    
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


void MDServer::constraintAddPrimaryKey(const std::string &dir,
				       const std::string &attr)
{
  DMESG("constraintChangePrimaryKey called: ..." << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());
  if( p.type() != EPT_DIR ) {
    out.append("1 No such directory " + dir + "\n");
    return;
  }
  
  if( !p.directoryTable.size() ){
    out.append("7 No attribute yet in table\n");
    return;
  }

  std::string a(attr);
  std::string key;
  if(checkAttributes(a, key, true, p.flags & EP_PLAIN)){
    out.append("7 Illegal key\n");
    return;
  }

  std::stringstream query;
  std::stringstream constraintName;
  constraintName << "p_key";
  if( !strcmp(dbHelper->backend(), "Oracle") ) {
    query << "ALTER TABLE " << p.directoryTable << " ADD CONSTRAINT p_key" 
	  << " PRIMARY KEY " << "(" << key << ");";
    constraintName << "_" << p.id;
  } else {
    query << "ALTER TABLE " << p.directoryTable << " ADD PRIMARY KEY " 
	  << "(" << key << ");";
  }

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }

  query.str("");
  query << "INSERT INTO " << constraintsTable
	<< " (\"table_id\", \"name\", \"type\", \"column\")"
	<< " VALUES (" << p.id << ", '" << constraintName.str() << "', 'P', '"
	<< attr << "');";
  
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("27 Constraint exists", statement);
    return;
  }
 
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


void MDServer::constraintAddReference(const std::string &dir,
                                      const std::string &attr,
                                      const std::string &ref,
                                      const std::string &name)
{
  DMESG("constraintAddReference called: ..." << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());
  if( p.type() != EPT_DIR ) {
    out.append("1 No such directory " + dir + "\n");
    return;
  }
  
  if( !p.directoryTable.size() ){
    out.append("7 Illegal key " + attr + "\n");
    return;
  }
  
  std::stringstream query;
  query << "INSERT INTO constraints" 
        << " (\"table_id\", \"name\", \"type\", \"column\", \"constraint\")"
        << " VALUES (" << p.id << ", '" << name << "', 'C', '"
        << attr << "', '" << ref << "');";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("27 Constraint exists", statement);
    return;
  }

  query.str("");
  std::string key;
  if(checkAttribute(attr, key, true, p.flags & EP_PLAIN)){
    out.append("7 Illegal key\n");
    return;
  }

  std::vector<std::string> attributes(1);
  attributes[0] = ref;

  TableMap tableMap;
  if(createTableMap(attributes, tableMap))
    return; // Error handled by createTableMap

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap
  
  std::string ref_key;
  if(parser->parse(ref)){
    out.append("7 Illegal key\n");
    return;
  }
  parser->replaceTableNames(tableMap);
  parser->toSQL(ref_key);

  std::string ref_table;
  size_t d=ref_key.find_first_of(".");
  if( d != std::string::npos ) {
    ref_table = ref_key.substr(0, d);
    ref_key = ref_key.substr(d+1);
  }
   
  query << "ALTER TABLE " << p.directoryTable 
        << " ADD CONSTRAINT " << name << " FOREIGN KEY ("
        << key << ") REFERENCES " << ref_table;
  if( ref_key.size() )
    query << '(' << ref_key << ")";
  query << ';';
  
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }
    
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


void MDServer::constraintDrop(const std::string &dir,
                              const std::string &name)
{
  DMESG("constraintDrop called: ..." << std::endl);

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());
  if(p.type() != EPT_DIR) {
    out.append("1 No such directory " + dir + "\n");
    return;
  }
  
  if( !p.directoryTable.size() ){
    out.append("28 No such constraint\n");
    return;
  }
  
  std::stringstream query;
  query << "SELECT \"column\", \"type\" FROM " << constraintsTable
        << " WHERE \"table_id\" = " 
        << p.id << " AND \"name\"='" << name << "';";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }

  char attr[COLUMN_NAME_MAX_SIZE+1];
  char type[2];
  std::string key;
  statement.bind(1, attr, COLUMN_NAME_MAX_SIZE);
  statement.bind(2, type, 2);
  
  if( statement.fetch() ) {
    if( name != "p_key"){
      out.append("28 No such constraint\n");
      return;
    }
  } else {
    if(checkAttribute(attr, key, false, p.flags & EP_PLAIN)){
      out.append("7 Illegal key\n");
      return;
    }
  }

  statement.unBind();
  statement.close();
  
  query.str("");
  query << "DELETE FROM " << constraintsTable << " WHERE \"table_id\"="
        << p.id << " AND \"name\" = '" << name << "';";

  DMESG("Name is: >" << name << "<");

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str()) && name != "p_key"){
    printError("28 Constraint does not exist", statement);
    return;
  }

  if(name == "p_key") // Even if query failed because of default PK
    type[0]='P';

  query.str("");
  switch(type[0]) {
  case 'N':
    if( !strcmp(dbHelper->backend(), "MySQL") ) {
      std::string keyType = getTypeOfAttr(p, attr);
      query << "ALTER TABLE " << p.directoryTable << " MODIFY COLUMN "
            <<  key << " " << keyType << ";";
    } else if (!strcmp(dbHelper->backend(), "Oracle") ) {
      query << "ALTER TABLE " << p.directoryTable << " MODIFY "
	    <<  key << " NULL;";
    } else {
      query << "ALTER TABLE " << p.directoryTable << " ALTER COLUMN "
            << key << " DROP NOT NULL;";
    }
  case 'P':
    if( !strcmp(dbHelper->backend(), "MySQL") ) {
      query << "ALTER TABLE "<< p.directoryTable
            << " DROP PRIMARY KEY;";
    } else if (!strcmp(dbHelper->backend(), "Oracle") ) {
      query << "ALTER TABLE " << p.directoryTable 
            << " DROP CONSTRAINT " << name;
      if(name == "p_key")
	query << "_dir" << p.id;
       query << ";";
    } else {
      query << "ALTER TABLE " << p.directoryTable // PostgreSQL
            << " DROP CONSTRAINT \"" << stripSchema(p.directoryTable) << "_pkey\";";
    }
    break;
  default:
    query << "ALTER TABLE " << p.directoryTable << " DROP CONSTRAINT "
          << name << ";";
  }

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return;
  }

  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();
      out.append("0\n");
  }    
}


int MDServer::listTableConstraints(const EntryProps &p,
				   Statement &statement,
				   std::string &buf)
{
  std::stringstream query;
  query << "SELECT \"name\", \"type\", \"column\", \"constraint\""
	<< " FROM " << constraintsTable 
	<< " WHERE \"table_id\" = " << p.id << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return -1;
  }

  char name[COLUMN_NAME_MAX_SIZE*1];
  char type[2];
  char attr[COLUMN_NAME_MAX_SIZE*1];
  char constraint[1025];
  statement.bind(1, name, COLUMN_NAME_MAX_SIZE);
  statement.bind(2, type, 2);
  statement.bind(3, attr, COLUMN_NAME_MAX_SIZE);
  statement.bind(4, constraint, 1024);
  
  ostringstream oss;  
  while( !statement.fetch() ){
    switch(type[0]) {
    case 'N':
      oss << "constraint_add_not_null " << p.name << " " << attr << " " << name << "\n";
      break;
    case 'U':
      oss << "constraint_add_unique " <<  p.name << " "<< attr << " " << name << "\n";
      break;
    case 'C':
      oss << "constraint_add_check " << p.name << " "
      << attr << " " << constraint << " " << name << "\n";
      break;
    case 'R':
      oss << "constraint_add_reference " << p.name << " "
        << attr << " " << (std::string)constraint << " "
        << name << "\n";
      break;
    case 'P':
      oss << "constraint_set_primary_key " << p.name << " "
        << attr << "\n";
      break;

    }
  }
  buf += oss.str();

  statement.unBind();
  statement.close();

  return 0;
}


 
void MDServer::constraintList(const std::string &dir)
{
  DMESG("constraintList called: ..." << std::endl);

  Statement statement(*dbConn);
  
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());
  if( p.type() != EPT_DIR ) {
    out.append("1 No such directory " +dir + "\n");
    return;
  }
  
  if( p.directoryTable == "x" ){
    out.append("0\n");
    return;
  }

  std::string buf;
  if( listTableConstraints(p, statement, buf) )
    return;
  
  out.append("0\n" + buf);
}

/**
 * The result from getEntryProps depends on whether the client provides
 * a wildcard of the form /dir/* or /dir. In both cases the directory
 * affected is /dir. Used for extracting the correct directory for saving
 * replication logs.
 */ 
std::string MDServer::getEntryDirectory(const EntryProps &props) const {
  return (props.type() == EPT_DIR) ? props.name : props.parent;
}


void MDServer::backend()
{
  out.append("0\n");
  out.append(dbHelper->backend());
  out.append("\n");
}


/* Sunil Ahn
   This file2 function is used for updateattr_single command */

int MDServer::find2(EntryProps &p, 
                    const std::string &path, 
                    const std::string &query, 
                    std::string &path2)
{
  std::list< std::string > groups; // Caching of groups user is in
  char buff[1024];


  if(parser->parse(query)){
    out.append("8 Illegal Query\n");
    return -1;
  }
  TableMap tableMap;
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return -1; // Error handled by createTableMap

  std::string q;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q);

  std::string pattern;
  bool isPattern=sqlPattern(p.name, pattern);
  if((p.flags & EP_TYPE) != EPT_NONE) {
    isPattern = true;
    pattern = "%";
  }

  std::stringstream queryString;
  queryString << "SELECT \"file\" FROM " << p.directoryTable;
  if( !(isPattern && pattern == "%") ) {
    queryString << " WHERE \"file\"";
    if(isPattern)
      queryString << " LIKE '";
    else
      queryString << " = '";
    queryString << pattern << "' AND ";
  } else {
    queryString << " WHERE ";
  }
  if(q.size()) {
    queryString << " (" << q << ") AND";
  }

  if( p.flags & EP_SHARED )
    queryString << "AND  \"dir\" = " << p.parentID ;
  queryString << "( 1 = 1 )  LIMIT 1 ;" ;


  LOG("SQL: >" << queryString.str() << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(queryString.str())){
    printError("9 Internal error:", statement);
    return -1;
  }

  if( statement.numRows() == 0 ) return 0; 

  statement.bind(1, buff, 1024);

  std::stringstream buff2 ;

  while(1){
    if(statement.fetch())
      break;

    if( (p.flags & EP_TYPE) == EPT_NONE ) {
       buff2 << p.parent << "/" << buff ;
    } else {
       buff2 << p.name << "/" << buff ;
    }
  }

  path2 = buff2.str();
  return 0;
}


/* Sunil Ahn
   update single entry which matchs given condition */

void MDServer::updateSingle(const std::string &path,
                          std::vector< std::string > &attributes,
                          std::vector< std::string > &values,
                          const std::string &condition,
                          bool eval)
{                         
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string path2 ;


  if ( lockForSingle() != 0 ) {
    out.append("99 Table Locking Error \n");
    return;
  }

  if( find2(p, path, condition, path2) != 0 ) {
    out.append("99 Finding Entry Error \n");
    unlockForSingle();
    return;
  }
 
  if (path2.size() == 0) {
    out.append("0\n");
    unlockForSingle() ;
    return;
  }

  updateAttr(path2, attributes, values, "", eval) ;
  
  unlockForSingle() ;

  std::string reponse;
  out.print(reponse);
  if ( reponse[0] == '0' ) {
    out.start();
    out.append("0\n");
    out.append(path2);
    out.append("\n");
  }

}


/* 
   For Dynamic VOMS Group management ...
   At authentication, it checks the given cert are involed which groups
*/


void MDServer::addCertGroup(X509 *cert){
  if (cert == NULL) return ;
  vomsGroupList.clear();

  VOMSInfo vomsInfo(cert);

  std::list < Capability >::const_iterator I = vomsInfo.capabilities.begin();
  while(I != vomsInfo.capabilities.end() ) {
    if (I->group.size() == 0) {I++;continue;}

    std::string query;
    query.append("SELECT name FROM voms_group WHERE subject = '");
    query.append(I->group);
    query.append("';");

    DMESG("SQL: >" << query << "<" << std::endl);

    Statement statement(*dbConn);
    if(statement.exec(query)){
      I++;
      continue;
    }

    char usr[USERNAME_MAX_SIZE+1];
    bzero(usr, USERNAME_MAX_SIZE+1);

    statement.bind(1, usr, USERNAME_MAX_SIZE);

    if(! statement.fetch()) {
      std::string tmp = usr;
      DMESG("Added Group: >" << tmp << "<" << std::endl);
      vomsGroupList.push_back(tmp);
    }

    I++;
  }

  return ;
}


int MDServer::lockForSingle()
{
  MDConMan *conn = getConMan() ;
  return conn->USlock();
}

int MDServer::unlockForSingle()
{
  MDConMan *conn = getConMan() ;
  return conn->USunlock();
}
