/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: MDServer.h,v 1.95 2010/10/26 07:27:47 asunil Exp $
//
#ifndef MDSERVER_H
#define MDSERVER_H

/** The following errors are returned:
      1 File or dir does not exist
      2 Connection to DB failed
      3 Invalid command
      4 Permission denied
      5 Group exists
      6 Group does note exist
      7 Illegal key
      8 Illegal query
      9 Internal DB error
     10 No such key/attribute  // Not yet used in getattr!!!
     11 Directory not empty
     12 Aborted
     13 Illegal type
     14 Attribute in use
     15 Entry exists
     16 Directory exists
     17 Illegal name (for index, view, sequence, group)
     18 Update failed
     19 Not an entry
     20 Session invalid
     21 Function not implemented
     22 No such help topic
     23 Not a directory
     24 User already member of that group
     25 Illegal access permission format
     26 ACL length exceeded
     27 Constraint exists
     28 Constraint does not exist
     29 Operation not permitted on plain table
     30 Illegal options
     31 Site exists
     32 Illegal GUID
     33 Illegal value
     34 User unknown
     35 Import error, missing column
*/

#include <openssl/ssl.h>

#include <vector>
#include <functional>
#include <string>
#include <list>
#include <map>
#include <stdexcept>
#include <memory>
#include <set>

#include "Constants.h"
#include "util/TableEntry.h"
#include "util/Debuggable.h"
#include "util/DBUtil.h"
#include "client/MDFed.h"

#include <stdint.h>

class QParser;
class Statement;
class MDBuffer;
class UploadHandle;
class DBHelper;
class MDUserManager;
class MDLogWriter;
class DatabaseConnection;
class MDConMan;

typedef std::pair<std::string, std::string> Attribute;
typedef std::list< Attribute > AttrList;
typedef std::set< std::string > Capabilities;

// Local constants
static const std::string GLOBAL_DB_SEQUENCER("sequencer");

class Attribute_eq : public std::unary_function<Attribute, bool>
{
 public:
  explicit Attribute_eq(const std::string &s)
    : key(s) {
  };
    bool operator() (const Attribute &a) const {return a.first==key; };
  private:
     std::string key;
};


class MDServer: public Debuggable {
 public:
  MDServer(MDBuffer &outbuf);
  virtual ~MDServer();
  /**
   * ownsConnection can be set to false to prevent the destructor 
   * of this class from releasing the environment. Useful
   * when this server instance is created over an existing connection
   */
  void setDBConnection(DatabaseConnection *connection, bool ownsConnection); 
  DatabaseConnection *getDBConnection() {
    return dbConn;
  }

  void setDefaultSchema(const std::string &schema) {
    dbSchema = schema;
    setupTableNames();
  }

  virtual void setupTableNames();

  void flushOutputBuffer(std::string &result);

  // Help
  virtual void help(const std::string &about = "");

  // Metadata commands
  virtual void createSchema(const std::string &name,
			    AttrList &attributes,
			    const std::string &parent = "") = 0;

  virtual void removeSchema(const std::string &name) = 0;

  virtual void addEntries(std::vector<std::string> &paths) = 0;

  virtual void removeEntries(std::vector<std::string> &paths) = 0;

  virtual void addEntry(const std::string &paths,
			std::vector< std::string > &attributes,
			std::vector< std::string > &values, bool eval = false) = 0;

  virtual void getAttr(const std::string &path,
		       std::vector< std::string > &attributes) = 0;

  virtual void setAttr(const std::string &path,
		       std::vector< std::string > &attributes,
		       std::vector< std::string > &values) = 0;

  virtual void listAttr(const std::string &path);

  virtual void removeAttr(const std::string &dir,
			  AttrList &attributes);

  virtual void renameAttr(const std::string &dir,
			  const std::string &oldName, const std::string &newName);

  virtual void clearAttr(const std::string &path,
			 const std::string &attribute) = 0;

  virtual void addAttr(const std::string &dir, AttrList &attributes);

  virtual void find(const std::string &path,
		    const std::string &query="") = 0;

  virtual void listEntries(const std::string &path = ".") = 0;

  // Metadata optional
  virtual void createView(const std::string &name,
			  const std::string &directory,
			  const std::vector<std::string> &attributes,
			  const std::string &restriction) = 0;

  virtual void createViewUnion(const std::string &name,
			       const std::vector<std::string> &dirs,
			       const std::string &restriction) = 0;

  virtual void createIndex(const std::string &name, const std::string &dir,
			   const std::vector< std::string > &attributes,
			   const std::string &algo="BTREE") = 0;

  virtual void createSequence(const std::string &name,
			      const std::string &dir,
			      const std::string &increment,
			      const std::string &start) = 0;

  virtual void indexRemove(const std::string &path) = 0;

  virtual void viewRemove(const std::string &path) = 0;

  virtual void sequenceRemove(const std::string &path) = 0;

  virtual void selectAttr(std::vector<std::string> &attributes,
			  const std::string &condition) = 0;

  virtual void updateAttr(const std::string &path,
			  std::vector< std::string > &attributes,
			  std::vector< std::string > &values,
			  const std::string &condition,
			  bool eval = true);

  void updateSingle(const std::string &path,
                          std::vector< std::string > &attributes,
                          std::vector< std::string > &values,
                          const std::string &condition,
                          bool eval = true);

  int find2(EntryProps &p,
                    const std::string &path,
                    const std::string &query,
                    std::string &path2);



  void setConMan(MDConMan *con ) { conMan = con; }
  MDConMan* getConMan() { return conMan ; }
  int lockForSingle();
  int unlockForSingle();



  virtual void update(const std::string &path,
		      std::vector< std::string > &attributes,
		      std::vector< std::string > &values,
		      const std::string &condition);

  virtual int upload(UploadHandle &handle,
  			  const std::string &dir,
		      std::vector< std::string > &attributes) = 0;

  virtual int put(UploadHandle &h,
		      const std::vector<std::string>::iterator valStart,
		      const std::vector<std::string>::iterator valEnd) = 0;

  virtual void abort(UploadHandle &handle) = 0;

  virtual void commit(UploadHandle &handle) = 0;  

  virtual void sequenceNext(const std::string &name) = 0;
  virtual void sequenceCurrent(const std::string &name) = 0;
      
  virtual void transactionStart();

  virtual void transactionCommit();

  virtual void transactionAbort();
  
  std::string getTypeOfAttr(const EntryProps &p, const std::string &attr);

  virtual void constraintAddNotNull(const std::string &dir,
				    const std::string &attr,
				    const std::string &name);

  virtual void constraintAddUnique(const std::string &dir,
				   const std::string &attr,
				   const std::string &name);


  virtual void constraintAddCheck(const std::string &dir,
				  const std::string &check,
				  const std::string &name);

  virtual void constraintAddReference(const std::string &dir,
				      const std::string &attr,
				      const std::string &ref,
				      const std::string &name);
  
  virtual void constraintAddPrimaryKey(const std::string &dir,
				       const std::string &key);

  virtual void constraintDrop(const std::string &dir,
			      const std::string &name);
  
  virtual void constraintList(const std::string &dir);


  /** Used to send a specific answer to the client */
  virtual void sendToClient(const std::string &message);

  // File-catalogue integration and standalone
  virtual void dir(const std::string &dir=".") = 0;
  virtual void cd(const std::string &dir);
  void pwd();

  // Authentication
  int initializeUserManager(std::list< MDUserManager *> &userManagerList);

  virtual bool checkCert(const std::string &user, X509 * cert,
			 std::string &vo, std::list< std::string > &groups,
			 std::string &eUser);
  
  bool checkSubject(const std::string &user,
		    const std::string &subject) const;

  bool checkPass(const std::string &user,
		 const std::string &pass) const;

  // Information about service
  void getName();
  void whoAmI();
  bool isConnected() const { return connected; }

  int setUser(const std::string &u);

  int setDir(const std::string &dir);

  void setExecUser( int i ){ execUID = i; }

  void setExecUserGroup( int i ){ execGID = i; }

  void setExecRoot( const std::string &d ) { execRoot = d; }

  std::string getDir() const { return currentDir; }

  std::string getUser( ) const { return user; }

  void setDN(const std::string &dn) { distinguishedName = dn; }

  void backend();

  // Logging and debugging
  void setLogging(int l) { logging = l; };
  void printODBCDriverVersion();

  MDBuffer& getBuffer() {
    return out;
  }
  /**
   * Obtains an unique value extracted from a sequence on the 
   * database backend. This is guaranteed to be unique across
   * all AMGA instances, past and future (up to 2^64 values).
   */
  uint64_t sequencerNextValue() throw (SQLException);

  // Federation
  int cmdToken;
  std::string nextCommand;
  
  // Replication
  void setNextCommand(const std::string &nextCmd);  

  void getResult(std::vector<std::string> &results) throw (std::runtime_error);
  /**
   * If master is different than "", then this instance assumes that
   * it is being used to executed replication commands received from
   * the given master. Used by the createDir command
   */
  void setMaster(const std::string &master);
  std::string getMaster() const;
  
  void setReplicating(bool replicatingInit) { replicating =  replicatingInit; }
  bool isReplicating() const { return replicating; }
  
  void setSynchronizing(bool synchronizingInit) { synchronizing =  synchronizingInit; }
  bool isSynchronizing() const { return synchronizing; }

  bool transactionInProgress() const;

  typedef std::list<EntryProps> EntryPropList;

  virtual int getAttributes(const std::string &table,
			    AttrList &keys,
			    bool all=false) = 0;

  bool checkPermResult(int error);

  virtual int getEntryProps(const std::string &entry,
			    EntryPropList &props,
			    const std::string &indexTable = "",
			    bool enableProxy = true) = 0;
  
  virtual int handleEntryPropErrors(int error,
				    const std::string &file="");

  int checkPermissions(const struct EntryProps &p, const char right,
		       std::list< std::string > &groups);

  // Update
  int getUpdatedRows() { return updatedRows; }
  void setUpdatedRows(int u) { updatedRows = u; }

  DatabaseConnection *dbConn;
  DBHelper         *dbHelper;

  void addCertGroup(X509 *cert);

  char* getPermissionMask() { return permissionMask; }
  char* getGroupRightsMask() { return groupRightsMask; }

  // config of federation 
  static int allowFedConnFail;
  static int allowFedExecFail;
  static int serverSideFed;

  MDFed* mdfed;



protected:

  void printError(std::string errnum, Statement &s) const;

  void printError(std::string errnum, std::exception &ex); 

  int checkAttribute(const std::string &attr, std::string &out,
		     bool allowSystem, bool plain) const;

  int checkAttributes(const std::string &attrs,
		      std::string &out, bool allowSystem,
		      bool plain) const;

  std::string getEntryDirectory(const EntryProps &entryProp) const;

  virtual int checkDirPermissions(const struct EntryProps &p, const char right,
				  std::list< std::string > &groups);

  virtual int memberOf(const std::string usr,
		       std::list< std::string > &groups) const = 0;

  virtual std::string createTable(const EntryProps &p,
				  AttrList &attrs,
				  Statement &statement) = 0;

  virtual bool alterTableColumns(const std::string &tableName,
				 const AttrList &attributes,
				 bool remove) = 0;

  void addAttributeList(const AttrList &keys, std::string &query,
			const std::string m="",
			bool withComma = true) const;

  void addTableDescription(const AttrList &keys,
			   std::string &query,
			   const std::string m="",
			   bool withComma = true) const;

  bool tableExists(std::string &tableName, int dType) const;

  bool inSubdir(const std::string &name) const;

  bool tableUsesColumn(const std::string &tableName,
		       const std::string &key);

  void removeDoubleTables(TableMap &tableMap);

  int setupTableMap(TableMap &tableMap);

  int createTableMap(const std::vector<std::string> &attributes,
		     TableMap &tableMap);

  void getDistinctTables(TableMap &tableMap,
			 std::list<std::string> &distinctTables);

  virtual std::string updateAttrQuery(const EntryProps &p,
				      const std::string &entries,
				      bool isPattern,
				      std::vector< std::string > &attributes,
				      std::vector< std::string > &values,
				      const std::string &condition,
				      bool eval) = 0;

  int listTableConstraints(const EntryProps &p, Statement &statement,
			   std::string &buf);

  int removeColumnConstraints(Statement &statement, int id,
			      const std::string &column);

  // Connection settings
  std::string      dbSchema;

  std::string      siteName;         /// The name of this server instance
  std::string      user;             /// User in the current session
  std::string      currentDir;
  int              execUID;
  int              execGID;
  std::string      execRoot;
  std::string      distinguishedName;

  MDConMan *conMan ;		     /// Used for UpdateSingle Command

  QParser          *parser;          /// The parser for SQL queries
  mutable MDBuffer &out;
  
  // These fields are needed here for use by the transaction methods.
  char             permissionMask[4];
  char             groupRightsMask[4];

  struct HelpEntry {
    unsigned int topic;
    std::string command;
    std::string args;
    std::string help;
  };

  std::list< struct HelpEntry > helpEntries;
  std::vector< std::string > helpTopics;
  std::list < MDUserManager * > userManagers;

  // Replication
  std::auto_ptr<MDLogWriter> logWriter;
  std::string master;
  // True when being used for writing replication logs.
  bool replicating;
  bool synchronizing;


  
  
  // Derived classes supporting replication should override this method
  virtual bool saveLog(Statement &statement, const std::string &parent) {return true;}
  
//  virtual uint64_t sequenceNextValue(const std::string &sequence) 
//    throw (SQLException) = 0;

  // Names of the tables used
  std::string amgaTable;
  std::string constraintsTable;
  std::string usersTable;
  std::string groupsTable;
  std::string masterindexTable;
  std::string certificatesTable;


  std::list<std::string> vomsGroupList;  
  
private:
  bool             connected;
  void             *userArg;
  int              logging;
  bool             ownsConnection;
  int 		   updatedRows;
};

#endif // MDSERVER_H
