/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDServerFactory.cc,v 1.25 2010/04/14 07:15:30 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDServerFactory.h"
#include "MDPgServer.h"
#include "MDSQLiteServer.h"
#include "MDOracleServer.h"
#include "MDFCMySQLServer.h"
#include "MDFCOracleServer.h"
#include "MDMySQLServer.h"
#include "MDMySQL5Server.h"
#include "util/Statement.h"
#include "util/DatabaseConnection.h"

#include <string>
#include <iostream>

bool MDServerFactory::debug = false;

#include "debug.h"

MDServer *MDServerFactory::mdServer(MDBuffer &outbuf, DatabaseConnection *dbConn)
throw (SQLException) 
{
  bool ownsConnection;
  if (dbConn == NULL) {
    DMESG("Creating a new connection\n");
    dbConn = DatabaseConnection::openConnection();
    ownsConnection = true;
  } else {
    DMESG("Reusing connection\n");
    ownsConnection = false;
  }

  DMESG("Checking driver type\n");
  DatabaseConnection::DriverType dtype = dbConn->getDriverType();

  std::string searchSchema;
  if( DatabaseConnection::dbSchema.size() ) {
    searchSchema = DatabaseConnection::dbSchema;
  }  else { 
    searchSchema = "%";

    // On Oracle we must use a schema prefix, at least to find the tables
    // of this user.
    if(dtype == DatabaseConnection::dtOracle)
      searchSchema = DatabaseConnection::dbUser;

    if(dtype == DatabaseConnection::dtMySQL5)  
      searchSchema = dbConn->getDatabaseName();
  }

  DMESG("Checking table exists in schema " << searchSchema << "\n");
  std::string lfcSchema;
  std::string defaultSchema;
  // Check for LFC table
  bool lfc = false;
  if(checkTableExists(dbConn, "Cns_file_metadata", searchSchema, lfcSchema) == 1) {
    LOG("LFC database detected\n");
    lfc = true;
  } else {
    // Check for masterindex table
    if(checkTableExists(dbConn, "masterindex", searchSchema, defaultSchema) != 1)
      ERRLOG("ERROR: No AMGA database detected!\n");
  }

  DatabaseConnection::defaultSchema = defaultSchema;

  DMESG("Creating server instance\n");
  MDServer *server = NULL;
  switch(dtype) {
  case DatabaseConnection::dtPSQL:
    DMESG("Using server MDPgServer" << std::endl);
    server=new MDPgServer(outbuf);
    break;
  case DatabaseConnection::dtSQLite:
    DMESG("Using server MDSQLiteServer" << std::endl);
    server=new MDSQLiteServer(outbuf);
    break;
  case DatabaseConnection::dtOracle:
    if(lfc) { 
      DMESG("Using server MDFCOracleServer" << std::endl);
      server=new MDFCOracleServer(outbuf);
      ((MDLFCServer *)server)->setLFCSchema(lfcSchema);
    } else {
      DMESG("Using server MDOracleServer" << std::endl);
      server=new MDOracleServer(outbuf);
    }
    break;
  case DatabaseConnection::dtMySQL:
    if(lfc) {
      DMESG("Using server MDFCMySQLServer" << std::endl);
      server=new MDFCMySQLServer(outbuf);
      ((MDLFCServer *)server)->setLFCSchema(lfcSchema);
    } else {
      DMESG("Using server MDMySQLServer" << std::endl);
      server=new MDMySQLServer(outbuf);
    }
    break;
  case DatabaseConnection::dtMySQL5:
    DMESG("Using server MDFCMySQL5Server" << std::endl);
    server=new MDMySQL5Server(outbuf);
    break;
  default:
  case DatabaseConnection::dtUnknown:
    DMESG("Unknown ODBC driver type" << std::endl);
    throw SQLException("Unknown ODBC driver type");
  }
  
  server->setDBConnection(dbConn, ownsConnection);
  server->setDefaultSchema(defaultSchema);
  DMESG("Using default schema: " << defaultSchema << "\n");

  if(dtype == DatabaseConnection::dtOracle) {
    Statement statement(*dbConn);
    statement.exec("ALTER SESSION SET nls_timestamp_format = "
		   "'YYYY-MM-DD HH24:MI:SS';");
  }
  return server;
}


int MDServerFactory::checkTableExists(DatabaseConnection *conn,
				      const std::string &name,
				      const std::string &searchSchema,
				      std::string &detectedSchema)
{
  DMESG("Searching for table " << name << " in schema " << searchSchema << "\n");

  Statement statement(*conn, debug);
  DatabaseConnection::DriverType dtype = conn->getDriverType();

  char n[1000];
  char schema[1000];

  std::string myTable = name;
  std::string mySchema = searchSchema;

  std::stringstream query;
  switch( dtype ) {
  case DatabaseConnection::dtOracle:
    for(size_t i = 0; i < myTable.size(); i++)
      myTable[i] = toupper(myTable[i]);
    for(size_t i = 0; i < mySchema.size(); i++)
      mySchema[i] = toupper(mySchema[i]);
    query << "SELECT owner, table_name FROM all_tables WHERE owner LIKE '"
	  << mySchema << "' AND table_name = '" << myTable << "';";

    break;
  case DatabaseConnection::dtPSQL:
    query << "SELECT schemaname, tablename FROM pg_tables WHERE "
	  << "schemaname LIKE '" << searchSchema << "' AND tablename = '"
	  << name << "';";
    break;
  case DatabaseConnection::dtMySQL:
  case DatabaseConnection::dtMySQL5:
    query << "SELECT table_schema, table_name FROM INFORMATION_SCHEMA.TABLES "
	  << "WHERE table_schema LIKE '" << searchSchema 
	  << "' AND table_name = '" << name << "';";
    break;
  case DatabaseConnection::dtSQLite:
    query << "SELECT TABLE_SCHEMA, TABLE_NAME FROM INFORMATION_SCHEMA_TABLES "
	  << "WHERE TABLE_SCHEMA LIKE '" << searchSchema
	  << "' AND TABLE_NAME = '" << name << "';";
    break;
  case DatabaseConnection::dtUnknown:
    ERRLOG("MDServerFactory::checkTableExists: "
           "Could not determine DB type!\n");
    return -1;
  }

  if (statement.exec(query.str())) {
    ERRLOG("MDServerFactory::checkTableExists: "
	   "Could not open tables for reading on DB\n");
    return -1;
  }
  statement.bind(1, schema, sizeof(schema));
  statement.bind(2, n, sizeof(n));

  if( !statement.fetch() ) {
    detectedSchema = schema;
    DMESG("Detected Schema is: " << detectedSchema << ", table name: " << n << "\n");
    return 1;
  }

  DMESG("Table not found, name: " << name << "\n");

  return 0;
}
