/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "MDSql.h"
#include "util/Statement.h"
#include "server/MDInterpreter.h"

#define COMMAND_GROUPS 1
#define COMMANDS 4

static const char commandGroups [COMMAND_GROUPS][3][50] = {
  {"24", "sql", "Native SQL Processing"}
};

static const char commands[COMMANDS][4][100] = {
  {"24", "SELECT", "attr... FROM dir  ...",
   "Select attributes of all matching pattens. Refer SQL-92 specification" },
  {"24", "UPDATE", "dir SET ... WHERE ...",
   "Update attributes of all matching pattens. Refer SQL-92 specification" },
  {"24", "DELETE", "FROM dir WHERE ...",
   "Remove rows having matching pattens. Refer SQL-92 specification" },
  {"24", "INSERT", "INTO dir [attr...] VALUES (...)",
   "Insert a new attribute. Refer SQL-92 specification" },
}; 


MDSql::MDSql(MDBuffer &outbuf) : MDServer(outbuf)
{
  asp = new AmgaSqlParser;
  asp->mdserver = this ;

  // Add help index entries
  unsigned int maxGroup = 0;
  for(unsigned int i = 0; i < COMMANDS; i++ ) {
    struct HelpEntry e;
    e.topic = atol(commands[i][0]);
    maxGroup = maxGroup < e.topic? e.topic : maxGroup;
    e.command = commands[i][1];
    e.args = commands[i][2];
    e.help = commands[i][3];
    helpEntries.push_back(e);
  }

  // Create help topics
  if (maxGroup + 1 > helpTopics.size() )
    helpTopics.resize(maxGroup+1);

  for(int i = 0; i < COMMAND_GROUPS; i++ )
    helpTopics[atol(commandGroups[i][0])]=commandGroups[i][1];

}

MDSql::~MDSql()
{
	delete asp;
}


void MDSql::printSQLAttr() 
{
  std::map <int, std::string> :: const_iterator uTi;
  uTi = asp->usedAttr.begin();
  while(uTi!=asp->usedAttr.end()) {
    out.append(uTi->second);
    out.append("\n");
    uTi++;
  }
}

int MDSql::sqlParse(std::string &line, int type, std::string &query)
{
  if ( asp->__sql_set_query((char*) line.c_str(), type)  == -1 ) {
      out.append("88 Too Long Input \n") ;
      return -1;
  }

  int exception = 0;


  do {
    if( sqlyyparse((void*) asp))   { exception = 1; break; }
    if( asp->error)                { exception = 1; break; }
    if( asp->parseResult == NULL ) { exception = 1; break; }
  } while(false) ;

  DMESG("Error code : " << asp->error << std::endl );
  if ( asp->parseResult != NULL ) 
    DMESG("Result : " << asp->parseResult << std::endl );

  if( exception ) {
    switch ( asp->error ) {
      case ER_NOT_VALID_TABLE_NAME2:
      case ER_PERMISSION_NOT_ALLOWED:
        return -1;
      default: break;
    }
    asp->flush();

    // std::string result_1;
    // out.flush(result_1);
    // DMESG("message >" << result_1 << "<" << std::endl );
    
    if ( asp->parseResult != NULL ) {
      if ( asp->error != 100 ) {
        out.append("88 SQL Parsing Error : ") ;
        out.append(asp->parseResult );
        out.append("\n");
      } else {
        // out.append("100\n") ;
        // out.append(asp->parseResult);
      }
      free(asp->parseResult);
      asp->parseResult = NULL;
    } else {
      out.append("88 SQL Parsing Error\n");
    }
    return -1;
  } else {
    query = asp->parseResult ;
    DMESG("Parsed SQL String : " << query << "\n");
  }

  return 0;
}


int MDSql::sqlExec(std::string &line, int type)
{
  std::string query;
  
  if( sqlParse(line, type, query) < 0 ) return -1;
   
  switch (type) {
    case MDInterpreter::cSELECT :
    {
      Statement statement(*dbConn);

      out.append("0 ");
      out.append(asp->numColumn) ;
      out.append("\n") ;
      printSQLAttr();

      bool  foundSomething = true;
      doBulkRead(statement, query, asp->numColumn, foundSomething );

      break ;
    } 
    case MDInterpreter::cINSERT :
    case MDInterpreter::cDELETE :
    case MDInterpreter::cUPDATE :
    {
      Statement statement(*dbConn);
      
      if( statement.beginTransaction() )  {
          out.append("88 SQL Handling Transaction Begin Error \n");
          break;
      }   

      
      if( statement.exec(query.c_str())) {
          if ( statement.getErrorCode() == SQL_NO_DATA ) {
            out.append("0\n");
            // out.append("No rows were modified\n");
            setUpdatedRows( 0 );
            break;
          }
          DMESG("rows 3: " << statement.numRows() << std::endl );
          out.append("88 SQL Operation Error : " );
          out.append(statement.getError());    
          out.append("\n");   
          break;
      } 

      setUpdatedRows( statement.numRows() );
      
      if ( saveLog(statement, asp->firstReferTable ) ) {
          statement.commitTransaction();
          out.append("0\n");
          // out.append(getUpdatedRows());
          // out.append("  has been modified\n");
      } else {
          out.append("88 SQL Query Replication Failed \n");
      }   
      break ;
    } 
    default:
    {
        out.append("0\n");
        out.append(asp->parseResult );
        out.append("\n");
    }
  } // end of switch


  free(asp->parseResult);
  asp->parseResult = NULL;
  return 0;

}

void MDSql::handleOutput(){
  std::string swp;
  out.print(swp);

  size_t pos = swp.find('\n');
  if( pos == std::string::npos ) {
    out.append("0\n");
    return;
  }

  std::string swp2 = swp.substr(pos+1);

  out.start();

  out.append("0\n") ;
  printSQLAttr();

  out.append(swp2);

}


