/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: MDStandalone.h,v 1.138 2011/02/14 09:29:05 asunil Exp $
//
#ifndef MDSTANDALONE_H
#define MDSTANDALONE_H

#include "sqlHelpers.h"
#include "MDServer.h"
#include "util/Clock.h"

#include <stdexcept>
#include <iostream>
#include <fstream>
#include <set>
#include <sql.h>
#include <openssl/evp.h>

#include "MDSql.h"

struct Mount;
struct Master;
class MountManager;

class UploadHandle {
public:
  friend class MDStandalone;
  
  std::vector<char *> buffers;
  std::vector<SQLLEN *> types;
  std::vector<int> bufferSizes;
  std::string directory;

  int cachedParams;
  int bulkSize;

  // Used when synchronizing
  int operationsCount;
  
  // If the GUID should be generated.
  bool genGUID;
  // How many values the client must provide
  int userProvidedAttrs;
  
  // The destructors of members are called in the reverse 
  // order of the completion of their constructor. Since
  // logsStatement depends on the database connection,
  // the destructor for the logsStatement must be called
  // before the one of the database connection.
  std::auto_ptr<DatabaseConnection> logsDBConn;
  std::auto_ptr<Statement> logsStatement;
  std::auto_ptr<Statement> statement;
  MDLogWriter* logWriter;
  
  // For the MDInterpreter to know if the upload aborted or not.
  bool aborted;
  
public:  
  UploadHandle();
  ~UploadHandle();
};

/*
 * NOTE: Make sure to set the user permission and group rights mask
 * before using this object.
 */

class MDStandalone: public MDSql {
    // To access lockTable/unlockTable
    friend class SiteManager;
public:
  MDStandalone(MDBuffer &outbuf);
  virtual ~MDStandalone();

  // Metadata commands
  virtual void createSchema(const std::string &name,
			    AttrList &attrs,
			    const std::string &parent = "");
  
  virtual void removeSchema(const std::string &name);

  void addEntry(const std::string &paths,
	       std::vector< std::string > &attributes,
	       std::vector< std::string > &values, bool eval = false);

  virtual void addEntries(std::vector<std::string> &paths);

  virtual void removeEntries(std::vector<std::string> &paths);

  virtual void getAttr(const std::string &path,
		       std::vector< std::string > &attributes);

  void setAttr(const std::string &path,
	       std::vector< std::string > &attributes,
	       std::vector< std::string > &values);

  void clearAttr(const std::string &path, const std::string &attribute);

  void find(const std::string &path, const std::string &query="");

  void listEntries(const std::string &dir=".");

  // Metadata optional
  void createView(const std::string &name,
		  const std::string &mainDirectory,
		  const std::vector<std::string> &attributes,
		  const std::string &restriction);

  void createViewUnion(const std::string &name,
		       const std::vector<std::string> &dirs,
		       const std::string &restriction);
    
  void createIndex(const std::string &name, const std::string &dir,
		   const std::vector< std::string > &attributes,
		   const std::string &algo="");

  void createSequence(const std::string &name, const std::string &dir,
		      const std::string &increment, const std::string &start);

  void indexRemove(const std::string &path);

  void viewRemove(const std::string &path);

  void sequenceRemove(const std::string &path);

  void selectAttr(std::vector<std::string> &attributes,
		  const std::string &condition);

  int upload(UploadHandle &handle,
	     const std::string &dir,
	     std::vector< std::string > &attributes);

  // Use iterators to avoid copying the values into a temporary vector
  int put(UploadHandle &h,
	  const std::vector<std::string>::iterator valStart,
	  const std::vector<std::string>::iterator valEnd);
  
  void abort(UploadHandle &handle);
  void commit(UploadHandle &handle);

  void sequenceNext(const std::string &name);
  void sequenceCurrent(const std::string &name);
  void sequenceValue(const std::string &name, bool next);

  // Schema management
  void schemaCp(const std::string &dir1,
		const std::string &dir2);

  // Information about service;
  void getName();
  void getVersion();
  
  // File-Catalogue commands
  void createDir(const std::string &dir, const std::string &option="");
  void createDir(const std::string &dir, AttrList &attributes,
		 const std::string &option="");

  void dir(const std::string &dir="." );

  void ls(const std::string &options = "", const std::string &dir=".");

  void rm(const std::string &path, const std::string &condition="",
	  const std::string &options="");

  void rmDir(const std::string &path);
  void link(const std::string &path, const std::string &URL);
  void stat(const std::string &path);
  //  void cp(const std::vector< std::string > &sources, const std::string &path);
  //  void mv(const std::vector< std::string > &sources, const std::string &path);
  void chown(const std::string &path, const std::string &user, bool recursiveOption);
  void chmod(const std::string &path, const std::string &perms);
  void ls(const std::string &options, std::vector< std::string > &paths);

  // Group commands
  void createGroup(const std::string &grp);
  void deleteGroup(const std::string &grp);
  void grpShow(const std::string &grp);
  void grpList(const std::string &grp="", bool all = false);
  void grpAddUser(const std::string &group, const std::string &usr);
  void grpRemoveUser(const std::string &group, const std::string &usr);
  void grpMember(const std::string &usr="");
  void grpDump(bool replication = false);

  // ACL commands
  void aclAdd(const std::string &entry,
	      const std::string &grp,
	      const std::string &permissions);
  void aclRemove(const std::string &entry,
		 const std::string &grp);
  void aclShow(const std::string &directory);

  // Volume and Distribution management
  void createVolume(const std::string &path);
  virtual void vacuum(const std::string &options) =0;

  // Authentication/security, used during Authentication
  int setPermissionMask(const std::string &m) {
    strncpy(permissionMask, m.c_str(), 4);
    return 0;
  };
  int setGroupRightsMask(const std::string &m) {
    strncpy(groupRightsMask, m.c_str(),4);
    return 0;
  };
  // User Management
  void userCreate(const std::string &user, const std::string &pass="", bool passIsHash = false);
  void userRemove(const std::string &user);
  void userPassChange(const std::string &user, const std::string &pass);
  void userSubjectAdd(const std::string &user, const std::string &subject);
  void userSubjectRemove(const std::string &user, const std::string &subject);

  void userVomsRemove(const std::string &subject, int type);
  void userVomsList(int type);
  void userVomsAdd(const std::string &usr, 
                   const std::string &subject, int type);



  void userListCred(const std::string &usr = "");
  void userList();
  int userDump(bool replication = false);

  void setPublicKey(EVP_PKEY *pkey);
  void setPrivateKey(EVP_PKEY *pkey);

  // Tickets
  void ticketGet();
  void ticketCheck(const std::string &ticket);

  static const std::string DUMP_NO_PERMS;
  static const std::string DUMP_NO_SYNC; 
  
  // Database dumps
  uint64_t dump(const std::string &directory, const std::string &secOpt, 
            bool replication = false);
  uint64_t dump(const std::string &topdir, bool dumpPerms, 
  	    bool replication = false);

  // Advanced Authentication with group creation
  
  virtual bool checkCert(const std::string &user, X509 * cert,
			 std::string &vo, std::list< std::string > &groups,
			 std::string &eUser);

  void deleteDirectoryContents(
    const std::string &dirName, bool deleteDirectory)
    throw(std::exception);

  void deleteDirectoryContents(
    const EntryProps &dir, bool deleteDirectory)
    throw(std::exception);

  // Capabilitites
  void capabilitiesRemove(const std::string &who, const std::string &cap);
  void capabilitiesAdd(const std::string &who, const std::string &cap);
  void capabilitiesList(const std::string &who="");
  
  // Replication 
  void listMounts();
  void listSubscribers();
  void umountImpl(const Mount &mount, MountManager &mManager, bool erase = true)
  	throw (std::exception);
  void umount(const std::string &remoteDir, bool erase = true);
  void umountAll(); 
  void mount(const std::string &nodeName, 
  	const std::string &remoteDir, 
  	const std::string &secOpt);
  	
  void mountUsers(const std::string &nodeName);
  void umountUsers();  

  // federation
  void proxyMount(const std::string &nodeName, const std::string &remoteDir);
  void proxyUmount(const std::string &remoteDir);
  void proxyListMounts();
  void proxyPolicy();
  void proxyCheckFailure(std::string code);
  int findFedFlags();
  	
  void synchronize(const std::string &remoteDir);
  void startReceiveLogs(const std::string &master);
  void stopReceiveLogs(const std::string &remoteDir);
  
  void repAllow(const std::string &directory, const std::string &user);
  void repDisallow(const std::string &directory, const std::string &user);
  void repUsersAllow(const std::string &who);
  void repUsersDisallow(const std::string &who);

  void repShowPermissions();
  void repDropSubscriber(const std::string &subscriber);
    
  // Replication as a master. Called by the ReplicationDaemon  
  void checkReplicatePermission(const std::string &directory) 
    throw (std::runtime_error);
    
  // Replica management for an FC

  typedef std::map< std::string, std::pair< std::string, std::string > > SiteMap;

  void replicaAdd(const std::string &guid, const std::string &surl,
		  const std::string &expires="");
 
  void replicaRemove(const std::string &surl, const std::string &guid="");

  void replicaRegister(const std::string &options,
		       const std::vector< std::string > &guids,
		       const std::vector< std::string > &site);

  void replicaUnregister(const std::string &options,
			 const std::vector< std::string > &guids,
			 const std::vector< std::string > &sites);

  void replicaList(const std::string &options,
		   std::vector< std::string > &guids);

  void replicaGet(const std::string &options, 
		  std::vector<std::string> &guid);
  // Site management
  bool checkSitePropertyExists(const std::string &property);
  void siteAdd(const std::string &site, const std::string &hostAndPort);
  void siteRemove(const std::string &site);
  void siteList();
  void siteDump(const std::vector<std::string> &siteNames);
  void siteListProperties(const std::string &site);
  void siteSetProperties(const std::string &site, 
        const std::vector<std::string> &properties, 
        const std::vector<std::string> &values);
  void siteGetProperties(const std::string &site, 
			 const std::vector<std::string> &property);

  int getSiteIDs(const std::vector< std::string > & sites,
		 std::map< std::string, int > &idMap);

  // Guids
  void lfnLookup(std::vector< std::string >&guid);

  
  // Foreign tables
  void import(const std::string &tablename, const std::string &dirctory,
	      const std::string &options = "");

  // Script execution
  virtual void execute(const std::vector<std::string> &args);

  // sudo
  void sudo(const std::string &new_user);

  virtual int getEntryProps(const std::string &entry,
			    EntryPropList &props,
			    const std::string &indexTable = "",
			    bool enableProxy = true);

  int addDirectoryToIndex(const std::string dir,
			  const EntryProps &initialProps,
			  Statement &statement);

  int finalizeDirectoryToIndex(int num, Statement &statement);

  virtual void doBulkRead(Statement &statement, std::string &query,
			  size_t fields, bool &foundSomething) = 0;

  int removeTableFromIndex(const std::string &dir, const std::string &index,
			   Statement &statement);
  std::string permissionClause(char op, int options,
                               std::list<std::string> &groups,
                               const std::string &table="") const ;


protected:
  virtual int prepareDirlikeEntry(const std::string &directory, 
				  EntryProps &p, std::string &name);



  virtual int dropTable(Statement &statement, const std::string &tableName);
  
  virtual std::string createTable(const EntryProps &p,
				  AttrList &attrs,
				  Statement &statement);

  virtual int dumpIndexes(Statement &statement, 
                          const std::string & dir, 
                          const std::string & table) ;

  virtual int lockTable(Statement &statement,
			const std::string &indexTable)=0;

  virtual void unlockTables(Statement &statement);

  int commitDirectoryToIndex(int id, const std::string &index,
			     Statement &statement, int flags = 0);

  

  virtual std::string opOnly() const =0;
  virtual std::string matchSubdirs(const std::string &dir) const=0;
  virtual bool dirPattern(const std::string &in, std::string &out) const = 0;
  virtual std::string opRegExpMatch() const = 0;
  virtual bool hasCursors() const =0;
  virtual bool hasExists() const =0;

  int updateCapabilities(const std::string &u, Capabilities &caps);

  int getCapabilities(const std::string &u, Capabilities &caps);

  /** Replication */
  bool saveLog(Statement &statement, const std::string &parent);
  
  bool saveUserOrGroupLog(Statement &statement);  
  bool saveUserLog(Statement &statement, const std::string &user);
  bool saveGroupLog(Statement &statement, 
  		const std::string &group, 
  		const std::string *user = NULL);

  //
  // Replica management
  //
  struct Replica {
    std::string guid;
    std::string parent;
    std::string surl;
    std::string table;
    std::string owner;
    std::string permissions;
    std::string group_rights;
    std::string directory;
    std::string dirACLs;
    std::string dirOwner;
    std::string dirPermissions;
    int tableID;
  };
  virtual void replicaGetLFN(const std::string &options, 
			     std::vector< std::string >&lfns);
  virtual void replicaGetFast(const std::string &options, 
			      std::vector< std::string >&lfns);

  virtual int entryFromGUID(std::vector< std::string >&guids,
			    std::vector< EntryProps > &entries);

  bool decodeTableRef(const std::string &tableRef, std::string &table,
		      int &id) const;
  
  void replicaListLFNs(std::vector< std::string > &lfns);

  
  virtual void setupTableNames();
  std::string guidsTable;
  std::string replicasTable;
  std::string mountsTable;
  std::string sitesTable;
  std::string subscribersTable;
  std::string transactionsTable;
  std::string masterTable;
  std::string logsTable;
  std::string usersequencesTable;	

private:

  class DirLine {
  public:
    int id;
    char directory[FILENAME_MAX_SIZE];
    char table[TABLE_NAME_MAX_SIZE];
    int flags;
    char mainColumn[TABLE_NAME_MAX_SIZE];
    char master[TABLE_NAME_MAX_SIZE];
    char owner[USERNAME_MAX_SIZE];
    char permissions[PERMISSION_SIZE];
    char acls[ACL_MAX_SIZE];
    char repGroups[ACL_MAX_SIZE];
    uint64_t rep_xid;
    char info[DIR_INFO_MAX_SIZE];

    bool operator<(const DirLine &other) const{
      // std::cout << "Comparing " << directory << " " <<  other.directory << std::endl;
      bool val= dirLineIsBefore((std::string) directory,
				(std::string) other.directory);
      // std::cout << "-> " << val << std::endl;
      return val;
    };
  };

  void deleteDirectoryContents(
    const std::string &dirName, bool deleteDirectory, bool checkperms)
    throw(std::exception);

  void deleteDirectoryContents(
    const EntryProps &dir, bool deleteDirectory, bool checkperms)
    throw(std::exception);

  friend std::ostream &operator<<(std::ostream &out, const MDStandalone::DirLine &dl);

  int getSubDirs(const std::string &directory, std::vector<DirLine> &dirs);

  bool hasSubdirs(const struct EntryProps &p);

  int removeIfEmpty(const EntryProps &p, bool forceAttributeRemoval,
		    Statement &statement);

  int removeTableConstraints(Statement &statement, int id);

  void rmObject(const std::string &path, int type);
  
  int deleteDirObjects(EntryPropList propList, int objectType,
		       bool force, Statement &statement, std::string &parent);

  int entryExists(EntryProps &p);

  int getMembers(const std::string name,
		 std::list< std::string > &members);

  virtual int memberOf(const std::string usr,
		       std::list< std::string > &groups) const;

  std::string expandGroupName(const std::string grp);

  bool doesGroupExist(const std::string grp,
		      Statement &statement);

  int grpChangeUser(const std::string &grp,
		    const std::string &usr,
		    bool add);

  int addRCapability(const std::string &directory, const std::string &grp);
  
  int removeRCapability(const std::string &directory, const std::string &grp);
  
  int fillACLMap(EntryProps &p, std::map< std::string, std::string > &acls);

  void aclChange(const std::string &entry,
		 const std::string &grp,
		 const std::string &permissions, bool add);

  std::string entriesExist(const std::string &tableName, int id=0) const;

  std::string subdirsExist(const std::string &index,
			   const std::string &parentDir) const;

  virtual std::string updateAttrQuery(const EntryProps &p,
				      const std::string &entries,
				      bool isPattern,
				      std::vector< std::string > &attributes,
				      std::vector< std::string > &values,
				      const std::string &condition,
				      bool eval);

  int doDump(const std::string& dumpRoot, 
           const EntryProps& dir,
           bool dumpPerms,
           bool replication);
  int dumpTable(Statement &statement, 
            const std::string &dumpRoot,
            const EntryProps &dirProps,
            bool dumpPerms, 
            bool replication);
  int dumpEntries(Statement &statement,
            const EntryProps &dirProps,
		        const AttrList &keys,
		        bool dumpPerms,		  
            bool replication);

  int dumpEntriesPG(Statement &statement,
                    const EntryProps &dirProps,
                    const AttrList &keys,
                    bool dumpPerms,
                    bool replication,
                    std::vector<std::string> &sysAttrs,
                    std::string &query);

  int dumpIndexes(Statement &statement, const std::string &dir);
  
  bool findAmongAttributes(std::vector< std::string > &attrs,
			  const std::string &attribute) const;
        
  std::map<std::string, std::string> createAttrDefaultValues() const;
  
  void completeUploadAttributes(std::vector< std::string > &attrs,
				const EntryProps &entryProps) const;

  bool completeAddEntryAttrs(std::vector< std::string > &attrs,
			     std::vector< std::string > &values,
			     const std::string &attribute,
			     const std::string &value);

  int statDir(EntryProps &p, const std::string &options, bool &found);

  int doFilesLS(Statement &statement, EntryProps &p,
		bool &found,
		const std::vector< std::string > &attrs,
		const std::string &options);
  
  int doDirsLS(Statement &statement, EntryProps &p, bool &found,
	       const std::string &options);

  int decodeLSOptions(std::string &options,
		      std::vector< std::string > &attrs);

  // Group management functions safe to call within authentication
  // phase for on-the-fly groups
  int doCreateGroup(const std::string &grp);

  // Replication
  void changeRepGroups(const std::string &directory, const std::string &grp, bool allow);    
  
  bool checkIfRoot();
  bool checkIfUserRepAllowed();
  bool checkIfCapabilityAllowed(const std::string &capability);
  bool checkIfSlaveReplication();
  bool checkIfMasterActive(Master &master);
  bool checkIfMasterActive(MountManager &mManager, const std::string &master);
  
  void checkEntryPropErrors(int error, const std::string &file) 
    throw (std::runtime_error);  
  std::auto_ptr<Mount> getMountPoint(MountManager &mManager, const std::string &mPoint);
  std::auto_ptr<Master> getMaster(MountManager &mManager, const std::string &master);
    
  static void csvStringToSet(const std::string &csvString, std::set<std::string> &values);
  static std::string setToCsvString(const std::set<std::string> &values);

  int replicaGetState(const std::vector< std::string > &guids,
		      SiteMap &siteMap, SiteMap &unknownGUIDs);
  
  int updateReplicas(SiteMap &siteMap);


  // federation
  int printProxy(EntryProps& p);
  
  int chownChangeOwner(Statement &statement, const std::string path, const std::string owner, bool recursiveOption=false);

};

#endif // MDSTANDALONE_H
