/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: MDUserManager.cc,v 1.16 2011/01/26 05:57:13 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDUserManager.h"
#include "MDServerFactory.h"

#include <fstream>
#include <iostream>
#include <sstream>

#include "util/sslHelpers.h"

static bool debug=false;
#include "debug.h"

#define MAX_SUBJECT_LENGTH 256
#define MAX_PASS_LENGTH 64


MDUserManager::MDUserManager()
  : privateKey(NULL), publicKey(NULL)
{
  DMESG("Initializing MDUserManager\n");
  OpenSSL_add_all_digests();
}


MDUserManager::~MDUserManager()
{
  // We don't hold any resource...
}



/** Checks whether a certain subject can be used as authorization for a given
    user (role).
    Note that several subjects can be used as a credential for a given role.
*/
bool MDUserManager::checkSubject(const std::string &user,
				 const std::string &subject) const
{
  return false;
}


/** Checks whether the password of the user (role) is
    correct.
*/
bool MDUserManager::checkPass(const std::string &user,
			      const std::string &pass) const
{
  return false;
}


bool MDUserManager::checkCert(const std::string &user, X509 *cert,
			      std::string &vo,
			      std::list< std::string > &groups,
			      std::string &eUser) const
{
  std::string subject = getCertSubjectName(cert, false);
  DMESG("Trying to authenticate user " << user
	  << " via subject >" << subject << std::endl);

  if(subject.size()) {
    if( checkSubject(user, subject) )
      return true;

    // Trying traditional subject name
    subject = getCertSubjectName(cert, true);
    DMESG("Trying to authenticate user " << user
	  << " via subject >" << subject << std::endl);
    if(subject.size()) {
      return checkSubject(user, subject);
    }
  }

  LOG("No client certificate sent: Cannot authenticate with cert!\n");
  return false;
}



/** Checks whether the subject is known to the system
 */
bool MDUserManager::isKnownSubject(const std::string &subject) const
{
  return false;
}


/** Checks whether the user is known to the system
 */
bool MDUserManager::isKnownUser(const std::string &user) const
{
  return false;
}


std::string MDUserManager::getDigest(const std::string &pass)
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  EVP_MD_CTX *ctx = EVP_MD_CTX_create();

  if(!ctx) {
    ERRLOG("Could not create digest context\n");
    return "";
  }
  
  const EVP_MD *md;
  md = EVP_get_digestbyname("sha1");
  if(!md){
    ERRLOG("Unkown digest\n");
    return "";
  }

  EVP_DigestInit_ex(ctx, md, NULL);
  EVP_DigestUpdate(ctx, pass.c_str(), pass.size());
  
  unsigned char md_value[EVP_MAX_MD_SIZE];
  unsigned int md_len;
  EVP_DigestFinal_ex(ctx, md_value, &md_len);
  
  EVP_MD_CTX_destroy(ctx);

  std::stringstream digest;
  for (size_t i=0; i<md_len; i++)
    digest << std::hex << (int)md_value[i];

  return digest.str();
#else
  return "";
#endif
}


/** Adds a user to the database
    Dummy implementation that allways returns
    21: function not implemented
*/
int MDUserManager::addUser(const std::string &user,
			   const std::string &pass, bool passIsHash)
{
  return 21;
}


/** Adds a new user-subject pair
    Dummy implementation that allways returns
    21: function not implemented
*/
int MDUserManager::addUserSubject(const std::string &user,
				  const std::string &subject)
{
  return 21;
}


/** Removes a user-subject pair
    Dummy implementation that allways returns
    21: function not implemented
*/
int MDUserManager::removeUserSubject(const std::string &user,
				     const std::string &subject)
{
  return 21;
}

/** Changes User Password
    Dummy implementation that allways returns
    21: function not implemented
*/
int MDUserManager::changeUserPass(const std::string &user,
				  const std::string &pass)
{
  return 21;
}


/** Removes a user
    Dummy implementation that allways returns
    21: function not implemented
*/
int MDUserManager::removeUser(const std::string &user)
{
  return 21;
}


/** List user credentials
    Dummy implementation that allways returns
    21: function not implemented
*/
int MDUserManager::listUserCred(const std::string &user,
				MDUserManager::UserCred &cred)
{
  return 21;
}

/** List users
    Dummy implementation that allways returns 0
*/
int MDUserManager::listUsers(std::list< std::string > &users)
{
  return 0;
}


int MDUserManager::ticketCreate(std::string &ticket, const std::string &user,
				const std::string &groups)
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  if(!privateKey)
    return 103;

  char timebuf[50];
  time_t timet;
  struct tm timetm;
  time(&timet);
  gmtime_r(&timet, &timetm);
  asctime_r(&timetm, timebuf);

  ticket.append("|notbefore ");
  ticket.append(timebuf); ticket.resize(ticket.size()-1);

  ticket.append("|notafter ");
  timet+= 60 * 60 * 24; // 24h
  gmtime_r(&timet, &timetm);
  asctime_r(&timetm, timebuf);
  ticket.append(timebuf); ticket.resize(ticket.size()-1);

  ticket.append("|user ");
  ticket.append(user);
  ticket.append("|groups ");
  ticket.append(groups);
  ticket.append("|");
  DMESG("Ticket: " << ticket << std::endl);
  
  EVP_MD_CTX *ctx = initSignMessage();
  if(!ctx)
    return -1;

  updateSignMessage(ctx, ticket);

  std::string signature = finalizeSignMessage(ctx, privateKey);

  DMESG("Signature: " << signature << std::endl);
  ticket.append(signature);

  EVP_MD_CTX_destroy(ctx);

  return 0;
#else
  return -1;
#endif
}


int MDUserManager::ticketCheck(const std::string &ticket)
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  DMESG("ticketCheck: " << ticket << std::endl);

  unsigned int pos = ticket.rfind("|");
  if(!pos || pos >= ticket.size()-1)
    return 104;

  std::string signature(ticket, pos+1);
  std::string message(ticket, 0, pos+1);
  DMESG("Message >" << message << "<\n");
  DMESG("Signature >" << signature << "<\n");
  
  EVP_MD_CTX *ctx = initVerifyMessage();
  if(!ctx)
    return -1;

  DMESG("1\n");

  updateVerifyMessage(ctx, message);

  DMESG("2\n");

  int res = finalizeVerifyMessage(ctx, publicKey, signature);

  DMESG("3 " << res << std::endl);

  EVP_MD_CTX_destroy(ctx);

  DMESG("4 " << res << std::endl);
  if(!res)
    return 105;

  return 0;
#else
  return -1;
#endif
}
