/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: MDUserManager.h,v 1.11 2010/04/14 07:15:30 asunil Exp $
//
/** MDUserManager is the base class to different classes which allow
    to users and their to be managed.
    It provides dummy implementations of functions to check whether
    users exist or their credentials are OK.
    It also provides a way to create tickets.
**/

#ifndef MD_USER_MANAGER_H
#define MD_USER_MANAGER_H

#include <string>
#include <list>
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
#include <openssl/evp.h>
#else
#include <openssl/x509.h>
#endif

class MDUserManager {
 public:
  // User Credential Member Class
  struct UserCred {
    std::string pass;
    std::string vo;
    std::list < std::string > fixedDNs;
    std::list < std::string > dbDNs;
  };

  MDUserManager();
  virtual ~MDUserManager();

  virtual bool checkSubject(const std::string &user,
			    const std::string &subject) const;

  virtual bool checkPass(const std::string &user,
			 const std::string &pass) const;

  virtual bool checkCert(const std::string &user, X509 *cert,
			 std::string &vo, std::list< std::string > &groups,
			 std::string &eUser) const;

  virtual bool isKnownUser(const std::string &user) const;
  virtual bool isKnownSubject(const std::string &subject) const;

  bool isInitialized() const {
    return initialized;
  };

  static std::string getDigest(const std::string &pass);

  virtual int addUser(const std::string &user,
		      const std::string &pass="", bool passIsHash = false);

  virtual int addUserSubject(const std::string &user,
			     const std::string &subject);

  virtual int removeUserSubject(const std::string &user,
				const std::string &subject);

  
  virtual int changeUserPass(const std::string &user,
			     const std::string &pass);

  virtual int removeUser(const std::string &user);

  virtual int listUserCred(const std::string &user, UserCred &cred);

  virtual int listUsers( std::list < std::string > &users);

  int ticketCreate(std::string &ticket, const std::string &user,
		   const std::string &groups="");

  int ticketCheck(const std::string &ticket);

  void setPublicKey(EVP_PKEY *pkey) {
    publicKey = pkey;
  };

  void setPrivateKey(EVP_PKEY *pkey) {
    privateKey = pkey;
  };


  std::string umName() const {
    return userManagerName;
  };

 protected:
  bool initialized;

  EVP_PKEY *privateKey;
  EVP_PKEY *publicKey;

  std::string userManagerName;
};

#endif //MD_USER_MANAGER_H
