/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "SiteManager.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDStandalone.h"
#include "util/Statement.h"
#include "util/DatabaseConnection.h"
#include "replicationdaemon/MountManager.h"
#include "debug.h"

#include <ostream>
#include <boost/array.hpp>

using namespace std;

/** List of configuration properties for sites */ 
static const boost::array<const char *, 13> SITE_PROPS = {
    "hostname", 
    "port",
    "login",
    "password",
    "use_ssl",
    "authenticate_with_certificate",
    "cert_file",
    "key_file",
    "use_grid_proxy",
    "verify_server_cert",
    "trusted_cert_dir",
    "require_data_encryption",
    "cert_password"
};

SiteManager::SiteManager(DatabaseConnection &dbConnInit, bool debug)
: Debuggable(debug), dbConn(dbConnInit)
{
}

SiteManager::~SiteManager()
{
}

/** Adds a site to the list of sites
*/
void SiteManager::add(const std::string &siteName, const std::string &host, int port, 
    MDStandalone &server)
    throw (SQLException)
{
  // The MDStandalone parameter is used only to lock the table. This code should
  // be on the Statement or DatabaseConnection classes. And DatabaseConnection
  // should have derived classes implementing the specific functionality for each
  // backend.     
  DMESG("siteAdd\n");
  Statement st(dbConn);

  // Gets the first unused id from the sites table
  if(server.lockTable(st, "sites")){
    throw SQLException("Could not lock sites table\n");
  }

  std::string query="SELECT \"id\" FROM sites ORDER BY \"id\";";

  int newID = 0;
  int id;

  DMESG("SQL: >" << query << "<" << std::endl);

  if(st.exec(query)){
    server.unlockTables(st);
    throw SQLException("Error reading sites table", st);
  }
  st.bind(1, &id);

  while( !st.fetch() ){
    if( id != newID )
      break;
    newID++;
  };
  
  DMESG("New site id is " << newID << std::endl);

  st.unBind();
  st.close();

  if( newID >= (SI_MAX_SIZE-1)*4+1 ) {
    server.unlockTables(st);
    throw SQLException("No more free site IDs");
  }

  ostringstream oss;
  oss << "INSERT INTO sites (\"id\", \"name\", \"hostname\", \"port\") "
     << " VALUES (" 
     << newID << ", '" << siteName << "', '" << host << "', " << port << ");";
  query = oss.str();
  DMESG("SQL: >" << query << "<" << std::endl);
  if(st.exec(query)){
    server.unlockTables(st);
    throw SQLException("Site exists", st);
  }
  // The following is only necessary for MySQL which does not limit
  // locking to transaction like any decent database
  server.unlockTables(st);  
  st.commitTransaction();
}


/** Removes a site from the list of sites
*/
void SiteManager::remove(const std::string &site) 
throw (SQLException, ReplicationException) 
{
  Statement st(dbConn);
  
  // Check if there are any mounts from this site
  MountManager mManager(dbConn);  
  auto_ptr<Master> master = mManager.getMaster(site);
  if (master.get() != NULL) {
    throw ReplicationException("Cannot remove site while it is in use by mounts.");
  }
  
  // No mounts, safe to delete.
  ostringstream query;
  query << "DELETE FROM sites WHERE \"name\"='" << site << "';";
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(st.exec(query.str())){
    if (st.getErrorCode() == SQL_NO_DATA) {    
      throw ReplicationException("Site does not exist");
    } else {
      throw SQLException("DB error", st);
    }
  }
  if (st.numRows() <= 0) {
    throw ReplicationException("Site does not exist");
  }
}


/** Lists all sites satisfying the given where clause. Can be the "" string
*/
void SiteManager::readSites(const string &whereClause, vector<Site> &sites) throw (SQLException) 
{
    ostringstream os;
    os << "SELECT \"id\", \"name\"";
    for (size_t i=0; i<SITE_PROPS.size(); i++) {    
        os << ", " << " \"" << SITE_PROPS[i] << "\"";
    }
    os << " FROM sites " << whereClause << " ORDER BY \"id\";";
    string query = os.str();
    DMESG("SQL: " << query << endl);
    Statement st(dbConn);
    if(st.exec(query)) {
        throw SQLException("Error querying sites table", st);
    }
    
    // boolean variables are stored as short int in the backend. We
    // must convert them to boolean, so we can't bind directly to
    // the Site structure.
    short int useSSL;
    short int authenticateWithCertificate;
    short int useGridProxy;
    short int verifyServerCert;
    short int requireDataEncryption;
    
    Site s;
    st.bind(1, &(s.id));
    st.bind(2, s.name, sizeof(s.name));
    st.bind(3, s.hostname, sizeof(s.hostname));
    st.bind(4, &(s.port));
    st.bind(5, s.login, sizeof(s.login));
    st.bind(6, s.password, sizeof(s.password));
    st.bind(7, &useSSL);
    st.bind(8, &authenticateWithCertificate);
    st.bind(9, s.certFile, sizeof(s.certFile));
    st.bind(10, s.keyFile, sizeof(s.keyFile));
    st.bind(11, &useGridProxy);
    st.bind(12, &verifyServerCert);
    st.bind(13, s.trustedCertDir, sizeof(s.trustedCertDir));
    st.bind(14, &requireDataEncryption);
    st.bind(15, s.certPassword, sizeof(s.certPassword));
    
    sites.clear();
    while (!st.fetch()) {
        s.useSSL = useSSL;
        s.authenticateWithCertificate = authenticateWithCertificate;
        s.useGridProxy = useGridProxy;
        s.verifyServerCert = verifyServerCert;
        s.requireDataEncryption = requireDataEncryption; 
    
        // Makes a copy of s
        sites.push_back(s);        
    }    
}

void SiteManager::getAll(vector<Site> &sites) 
throw (SQLException)
{
    readSites("", sites);
}

void SiteManager::checkPropertyExists(const string &property) 
throw (SQLException) 
{
    // Check if the property exists
    if (std::find(SITE_PROPS.begin(), SITE_PROPS.end(), property) == SITE_PROPS.end()) {
        throw SQLException("Invalid propery: " + property);
    }
}

void SiteManager::setProperties(const string &site, 
        const vector<string> &properties, 
        const vector<string> &values) 
throw (SQLException) 
{
    ostringstream os;
    os << "UPDATE sites SET";
    string sep; 
    for (size_t i=0; i<properties.size(); i++) {
        // Check if the property exists
        checkPropertyExists(properties[i]);
        os << sep << " \"" << properties[i] << "\" = '" << values[i] << "'";
        sep = ","; 
    }
    os << " WHERE \"name\" = '" << site << "';";
    string query = os.str();
    DMESG("SQL: " << query << endl);
    Statement st(dbConn);
    if(st.exec(query)) {
        if (st.getErrorCode() == SQL_NO_DATA) {
            throw SQLException("Unknown site: " + site); 
        } else {
            throw SQLException("Error updating sites table", st);
        }
    }
}

/** For automatic reclaiming of memory on the siteGetProperty method
 */
struct CharPointerContainer {
    vector<char *> v;
    ~CharPointerContainer() {
        for (size_t i=0; i<v.size(); i++) {
            delete [] v[i];
        }
    }
};

bool SiteManager::isSiteKnown(const string &siteName) 
throw (SQLException)
{
    vector<Site> v;
    readSites("WHERE \"name\" = '" + siteName + "'", v);
    return !v.empty();
}

void SiteManager::getSite(const string &siteName, Site &site) 
throw (SQLException)
{
    vector<Site> v;
    readSites("WHERE \"name\" = '" + siteName + "'", v);
    if (v.empty()) {
        throw SQLException("Site not found: " + siteName);
    }
    assert(v.size() == 1);
    site = v[0];
}

void SiteManager::getSites(const vector<string> &siteNames, vector<Site> &sites) 
throw (SQLException)
{
    ostringstream os;
    string sep = "WHERE "; 
    for (size_t i=0; i<siteNames.size(); i++) {
        os << sep << "\"name\" = '" + siteNames[i] + "'";
        sep = " OR ";
    }
    readSites(os.str(), sites);
}

void SiteManager::getProperties(const string &site, 
        const vector<string> &properties,
        vector<string> &values)
  throw (SQLException) 
{
    // Check if the property exists
    ostringstream os;
    os << "SELECT";
    string sep("");
    for (size_t i=0; i<properties.size(); i++) {        
        checkPropertyExists(properties[i]);
        os << sep << " \"" << properties[i] << "\"";
        sep = ",";
    }
    os << " FROM sites WHERE \"name\" = '" << site << "';";
    string query = os.str();
    DMESG("SQL: " << query << endl);
    Statement st(dbConn);
    if(st.exec(query)) {
        throw SQLException("Error querying sites table", st);
    }
    // This container will delete the objects pointed to automatically
    CharPointerContainer p;
    for (size_t i=0; i<properties.size(); i++) {
        char *ptr = new char[SITE_PROP_MAX_SIZE];
        st.bind(i+1, ptr, SITE_PROP_MAX_SIZE);
        p.v.push_back( ptr );
    }
    if (st.fetch()) {
        throw SQLException("Error reading from DB", st);
    }
    values.clear();    
    for (size_t i=0; i<p.v.size(); i++) {
        values.push_back(p.v[i]);
    }
}

std::ostream& operator<<(std::ostream& out, const Site& site) {
    out << "id=" << site.id 
        << "\nname=" << site.name
        << "\nhostname=" << site.hostname
        << "\nport=" << site.port
        << "\nlogin=" << site.login 
        << "\npassword=" << (strlen(site.password) == 0 ? "" : "<set>")
        << "\nuse_ssl=" << site.useSSL
        << "\nuse_grid_proxy=" << site.useGridProxy 
        << "\nauthenticate_with_certificate=" << site.authenticateWithCertificate
        << "\nverify_server_cert=" << site.verifyServerCert 
        << "\ncert_file=" << site.certFile 
        << "\nkey_file=" << site.keyFile
        << "\ntrusted_cert_dir=" << site.trustedCertDir
        << "\nrequire_data_encryption=" << site.requireDataEncryption
        << "\ncert_password=" << site.certPassword;
    return out;
}
