/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: VOMSAttrUserManager.cc,v 1.17 2011/01/26 05:57:13 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "VOMSAttrUserManager.h"
#include "VOMSInfo.h"
#include "util/Statement.h"

static bool debug=false;
#include "debug.h"
#include "Constants.h"

#define MAX_SUBJECT_LENGTH 256
#define MAX_PASS_LENGTH 64

#define VOMS_USER_TABLE "voms_user"
#define VOMS_GRP_TABLE "voms_group"

VOMSAttrUserManager::VOMSAttrUserManager()
{
  DMESG("Initializing VOMSAttrUserManager\n");
  userManagerName="VOMSAttrUserManager";
}

VOMSAttrUserManager::VOMSAttrUserManager(DatabaseConnection *dbConnection)
   : dbConn(dbConnection)
{
  DMESG("Initializing VOMSAttrUserManager\n");
  userManagerName="VOMSAttrUserManager";
}

void VOMSAttrUserManager::setConnection(DatabaseConnection *dbConnection)
{
  DMESG("VOMSAttrUserManager : Setting DB connection\n");
  dbConn = dbConnection;
}

VOMSAttrUserManager::~VOMSAttrUserManager()
{
  // We don't hold any resource...
}

bool VOMSAttrUserManager::checkSubject(const std::string &name,
                     const std::string &subject, int type) const
{
  DMESG("checkSubject: got >" << subject << "<\n");

  if(!dbConn)
    return false;

  std::string target_table; 

  if( type == TUSER ) {
    target_table = VOMS_USER_TABLE ;
  } else {
    target_table = VOMS_GRP_TABLE;
  }

  std::string query="SELECT \"subject\", \"name\" FROM " + target_table
    + " WHERE \"subject\" = '" + subject + "';";

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    ERRLOG("Unable to read certificates table!\n");
    return false;
  }

  char usr[USERNAME_MAX_SIZE];
  char sub[MAX_SUBJECT_LENGTH];

  statement.bind(1, sub, USERNAME_MAX_SIZE);
  statement.bind(2, usr, MAX_SUBJECT_LENGTH);

  if( !statement.fetch() ) {
    DMESG("checkSubject: allowing role " << name << std::endl);
    name = usr;
    return true;
  }

  return false;
}


int VOMSAttrUserManager::parseMapping(const std::string mapping,
				      std::string &external,
				      std::string &internal) const
{
  
  std::string str(mapping);
  size_t p1 = str.find("(");
  size_t p2 = str.find(")");
  if(p1 != std::string::npos && p2 != std::string::npos) {
    unsigned int p3 = p2-1;
    p2 = p1 + 1;
    p1 -= 1;
    while (p1 >0 && isblank(str[p1])) p1--;
    while (p2 <str.size() && isblank(str[p2])) p2++;

    while (p3 >0 && isblank(str[p3])) p3--;
    if (! (0 < p1 && p1 < p2 && p2 < p3 ) )
      return 1;
    
    external = str.substr(0, p1+1);
    internal = str.substr(p2, p3-p2+1);
  } else {
    external = str;
  }

  return 0;
}


int VOMSAttrUserManager::allowVO(const std::string &vo)
{
  std::string voName;
  std::string voDefault;
  
  if(parseMapping(vo, voName, voDefault))
    return 1;
    
  DMESG("voName >" << voName << "<  voDefault >" << voDefault << "<\n"); 

  if(voDefault.size())
    defaultUserMap[voName] = voDefault;

  allowedVOs.push_back(voName);

  return 0;
}

 
int VOMSAttrUserManager::mapGroup(const std::string &groupMapping)
{
  std::string voGroup;
  std::string mdGroup;
  
  if(parseMapping(groupMapping, voGroup, mdGroup))
    return 1;
  
  if(!mdGroup.size())
    return 1;

  DMESG("voGroup >" << voGroup << "<  mdGroup >" << mdGroup << "<\n"); 
  
  groupMap[voGroup] = mdGroup;

  return 0;
}


 
int VOMSAttrUserManager::mapUser(const std::string &userMapping)
{
  std::string voRole;
  std::string user;
  
  if(parseMapping(userMapping, voRole, user))
    return 1;
  
  if(!user.size())
    return 1;

  DMESG("voRole >" << voRole << "<  user >" << user << "<\n"); 
  
  userMap[voRole] = user;

  return 0;
}


std::string VOMSAttrUserManager::getMappedGroup(const std::string voGroup) const
{
  std::map < std::string, std::string >::const_iterator I;
  I = groupMap.find(voGroup);
  if(I != groupMap.end())
    return I->second;
  return "";
}



std::string VOMSAttrUserManager::getMappedUser(const std::string voRole) const
{
  DMESG("GetMappedUser >" << voRole << "<\n"); 

  std::map < std::string, std::string >::const_iterator I;
  I = userMap.find(voRole);
  if(I != userMap.end())
    return I->second;

  std::string name ; 
  if( checkSubject(name, voRole, TUSER) ) 
    return name;
  return "";
}


std::string VOMSAttrUserManager::defaultUser(const std::string &vo) const
{
  std::map < std::string, std::string >::const_iterator I;
  I = defaultUserMap.find(vo);
  if(I != defaultUserMap.end())
    return I->second;
  return "";
}


int VOMSAttrUserManager::disallowVO(const std::string &vo)
{
  allowedVOs.remove(vo);
  return 0;
}


/** Checks whether a certain certificate can be used as authorization for
    a given user (role) by examining the attributes in the extensions
    of the VOMS certificate.
    Note that several certs can be used as a credential for a given
    role.
*/
bool VOMSAttrUserManager::checkCert(const std::string &user, X509 *cert,
				    std::string &vo,
				    std::list< std::string > &groups,
				    std::string &eUser) const
{
  DMESG("VOMSAttrUserManager CheckCert\n");
  VOMSInfo vomsInfo(cert);

  std::list< std::string >::const_iterator J;
  J=allowedVOs.begin();
  while( J != allowedVOs.end() ) {
    DMESG("VOMSAttrUserManager Allowed VO : " << *J << " : user : " << user << " : " << vo << "\n" );
    if(vomsInfo.allowedRole(user, *J, vo, groups)) {
      if(user == "NULL") {
	if(defaultUser(vo) == "") {
	  LOG("No default login for VO " << vo << std::endl);
	  return false;
	} else {
	  eUser = defaultUser(vo);
	}
      } else {
	eUser = getMappedUser(user);
	if( !eUser.size() ) {
	  LOG("Requested user " << user << " does not exist!\n");
	  return false;
	}
      }	
      DMESG("checkCert: allowing " << user << std::endl);
      return true;
    }
    J++;
  }
  return false;
}


int VOMSAttrUserManager::listUsers(std::list < std::string> &users)
{
  std::map< std::string, std::string >::const_iterator I;
  I = userMap.begin();
  while( I != userMap.end() )
    users.push_back(I++->second);

  I = defaultUserMap.begin();
  while( I != defaultUserMap.end() )
    users.push_back(I++->second);

  return 0;
}


int VOMSAttrUserManager::listUserCred(const std::string &user,
				      MDUserManager::UserCred &cred)
{
  int res = 100;

  std::map< std::string, std::string >::const_iterator I;
  I = userMap.begin();
  while( I != userMap.end() ) {
    if(I->second == user) {
      res = 0;
      if(cred.vo.size()) cred.vo.append(" ");
      cred.vo.append(I->first);
    }
    I++;
  }

  I = defaultUserMap.begin();
  while( I != defaultUserMap.end() ) {
    if(I->second == user) {
      res = 0;
      if(cred.vo.size()) cred.vo.append(" ");
      cred.vo.append(I->first);
    }
    I++;
  }
  return res;
}




int VOMSAttrUserManager::addUserVoms(const std::string &user,
                                  const std::string &subject, int type)
{
  if(dbConn) {
    std::string query="INSERT INTO " ; 
    if (type == 1 ) { 
      query.append(VOMS_USER_TABLE) ;
    } else if (type == 2 ) {
      query.append(VOMS_GRP_TABLE) ;
    } else {
      return 100;
    }
    query.append(" (\"subject\", \"name\") VALUES ('");
    query.append(subject);
    query.append("', '");
    query.append(user);
    query.append("');");

    DMESG("SQL: >" << query << "<" << std::endl);
    Statement statement(*dbConn);
    if(statement.exec(query)){
      return 100;
    }
  }
  return 0;
}


int VOMSAttrUserManager::listUserVoms(const std::string &result, int type)
{
  if(dbConn) {
    std::string query="SELECT * FROM ";
    if (type == 1 ) {
      query.append(VOMS_USER_TABLE);
    } else if (type == 2 ) {
      query.append(VOMS_GRP_TABLE) ;
    } else {
      return 100;
    }
    query.append(" ;") ;
    DMESG("SQL: >" << query << "<" << std::endl);

    Statement statement(*dbConn);
    if(statement.exec(query)){ return 100; }

    char usr[USERNAME_MAX_SIZE];
    char sub[MAX_SUBJECT_LENGTH];

    statement.bind(1, sub, USERNAME_MAX_SIZE);
    statement.bind(2, usr, MAX_SUBJECT_LENGTH);

    while( !statement.fetch() ) {
      result.append(sub);
      result.append("\n");
      result.append(usr);
      result.append("\n");
    }
  }

  std::map< std::string, std::string >::const_iterator I;
  I = userMap.begin();
  while( I != userMap.end() ) {
    result.append(I->first);
    result.append("\n");
    result.append(I->second);
    result.append("\n");
    I++;
  }

  return 0;
}

int VOMSAttrUserManager::removeUserVoms(const std::string &subject, int type)
{
  if(dbConn) {
    std::string query="DELETE FROM ";
    if (type == 1 ) {
      query.append(VOMS_USER_TABLE);
    } else if (type == 2 ) {
      query.append(VOMS_GRP_TABLE);
    } else {
      return 100;
    }
    query.append(" WHERE \"subject\" = '");
    query.append(subject);
    query.append("' ;"); ;

    DMESG("SQL: >" << query << "<" << std::endl);
    Statement statement(*dbConn);
    if(statement.exec(query)){
      return 100;
    }
  }

  return 0;
}


