/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: VOMSInfo.cc,v 1.21 2012/02/01 06:42:03 gpark Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "VOMSInfo.h"

#include <openssl/asn1.h>
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
#include <openssl/asn1t.h>
#include <openssl/ocsp.h>
#endif

#include "util/sslHelpers.h"

static bool debug = false;
#include "debug.h"

int VOMSInfo::vomsAttrNID;
X509V3_EXT_METHOD VOMSInfo::vomsAttrEXT;
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
ASN1_SEQUENCE(ACDigest) = {
  ASN1_SIMPLE(ACDigest, type, ASN1_ENUMERATED),
  ASN1_SIMPLE(ACDigest, oid, ASN1_OBJECT),
  ASN1_SIMPLE(ACDigest, algorithm, X509_ALGOR),
  ASN1_SIMPLE(ACDigest, digest, ASN1_BIT_STRING)
} ASN1_SEQUENCE_END(ACDigest)

IMPLEMENT_ASN1_FUNCTIONS(ACDigest)


ASN1_SEQUENCE(ACIssuer) = {
  ASN1_SEQUENCE_OF(ACIssuer, issuer, GENERAL_NAME),
  ASN1_SIMPLE(ACIssuer, serial, ASN1_INTEGER),
  ASN1_EXP_OPT(ACIssuer, uid, ASN1_BIT_STRING, V_ASN1_BIT_STRING)
} ASN1_SEQUENCE_END(ACIssuer)

IMPLEMENT_ASN1_FUNCTIONS(ACIssuer)


ASN1_SEQUENCE(ACForm) = {
  ASN1_SEQUENCE_OF(ACForm, names, GENERAL_NAME),
  ASN1_IMP_OPT(ACForm, issuer, ACIssuer, 0),
  ASN1_IMP_OPT(ACForm, digest, ACDigest, 1)
} ASN1_SEQUENCE_END(ACForm)

IMPLEMENT_ASN1_FUNCTIONS(ACForm)


ASN1_SEQUENCE(ACHolder) = {
  ASN1_IMP_OPT(ACHolder, baseid, ACIssuer, 0),
  ASN1_IMP_SEQUENCE_OF_OPT(ACHolder, name, GENERAL_NAMES, 1),
  ASN1_IMP_OPT(ACHolder, digest, ACDigest, 2)
} ASN1_SEQUENCE_END(ACHolder)

IMPLEMENT_ASN1_FUNCTIONS(ACHolder)


ASN1_SEQUENCE(ACValidity) = {
  ASN1_SIMPLE(ACValidity, notBefore, ASN1_GENERALIZEDTIME),
  ASN1_SIMPLE(ACValidity, notAfter, ASN1_GENERALIZEDTIME)
} ASN1_SEQUENCE_END(ACValidity)


IMPLEMENT_ASN1_FUNCTIONS(ACValidity)

ASN1_SEQUENCE(ACIETFAttribute) = {
  ASN1_IMP_SEQUENCE_OF_OPT(ACIETFAttribute, names, GENERAL_NAME, 0),
  ASN1_SEQUENCE_OF(ACIETFAttribute, values, ASN1_OCTET_STRING)
} ASN1_SEQUENCE_END(ACIETFAttribute)

IMPLEMENT_ASN1_FUNCTIONS(ACIETFAttribute)


ASN1_SEQUENCE(ACAttribute) = {
  ASN1_SIMPLE(ACAttribute, type, ASN1_OBJECT),
  ASN1_SET_OF(ACAttribute, ietfAttribute, ACIETFAttribute)
} ASN1_SEQUENCE_END(ACAttribute)

IMPLEMENT_ASN1_FUNCTIONS(ACAttribute)


ASN1_SEQUENCE(AttributeCertificateInfo) = {
  ASN1_SIMPLE(AttributeCertificateInfo, version, ASN1_INTEGER),
  ASN1_SIMPLE(AttributeCertificateInfo, holder, ACHolder),
  ASN1_IMP_OPT(AttributeCertificateInfo, form, ACForm, 0),
  ASN1_SIMPLE(AttributeCertificateInfo, signatureAlgorithm, X509_ALGOR),
  ASN1_SIMPLE(AttributeCertificateInfo, serialNumber, ASN1_INTEGER),
  ASN1_SIMPLE(AttributeCertificateInfo, validity, ACValidity),
  ASN1_SEQUENCE_OF(AttributeCertificateInfo, attributes, ACAttribute),
  ASN1_IMP_OPT(AttributeCertificateInfo, id, ASN1_INTEGER, V_ASN1_BIT_STRING),
  ASN1_SEQUENCE_OF_OPT(AttributeCertificateInfo, extensions, X509_EXTENSION)
} ASN1_SEQUENCE_END(AttributeCertificateInfo)

IMPLEMENT_ASN1_FUNCTIONS(AttributeCertificateInfo)


ASN1_SEQUENCE(AttributeCertificate) = {
  ASN1_SIMPLE(AttributeCertificate, attributeCertInfo, AttributeCertificateInfo),
  ASN1_SIMPLE(AttributeCertificate, signatureAlgorithm, X509_ALGOR),
  ASN1_SIMPLE(AttributeCertificate, signature, ASN1_BIT_STRING)
} ASN1_SEQUENCE_END(AttributeCertificate)

IMPLEMENT_ASN1_FUNCTIONS(AttributeCertificate)


ASN1_SEQUENCE(ACSequence) = {
  ASN1_SEQUENCE_OF(ACSequence, attributeCertificates, AttributeCertificate)
} ASN1_SEQUENCE_END(ACSequence)
IMPLEMENT_ASN1_FUNCTIONS(ACSequence)

#endif

VOMSInfo::VOMSInfo(X509 *cert)
  : parseOK(false), attributes(NULL)
{
  DMESG("Initializing VOMSInfo\n");

  if(!cert)
    ERRLOG("VOMSInfo: No certificate!\n");

  if(!parseCert(cert))
     parseOK = true;
}

VOMSInfo::~VOMSInfo()
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  if(attributes)
    ACSequence_free(attributes);
#endif
}


int VOMSInfo::initASN1Parser()
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  vomsAttrNID = OBJ_create("1.3.6.1.4.1.8005.100.100.5",
			   "VOMSAttrSequence", "VOMS Attributes");
  vomsAttrEXT.ext_nid = vomsAttrNID;
  vomsAttrEXT.ext_flags = X509V3_EXT_MULTILINE;
  vomsAttrEXT.ext_new = (void * (*)()) ACSequence_new;
  vomsAttrEXT.ext_free = (void (*)(void *))ACSequence_free;
  vomsAttrEXT.d2i = (void* (*)(void*, OPENSSL_CONST unsigned char**, long int))d2i_ACSequence;
  vomsAttrEXT.i2d = (int (*)(void*, unsigned char**))i2d_ACSequence;
  // For SSL 1.0 Higher : SSL 1.0 change SSL_METHOD to const SSL_METHOD
#if OPENSSL_VERSION_NUMBER >= 0x10000000L
  vomsAttrEXT.i2s = (char* (*)(const v3_ext_method*, void*))NULL;
  vomsAttrEXT.s2i = (void* (*)(const v3_ext_method*, v3_ext_ctx*, OPENSSL_CONST char*))NULL;
  vomsAttrEXT.i2v = (stack_st_CONF_VALUE* (*)(const v3_ext_method*, void*, stack_st_CONF_VALUE*))NULL;
  vomsAttrEXT.v2i = (void* (*)(const v3_ext_method*, v3_ext_ctx*, stack_st_CONF_VALUE*))NULL;
  vomsAttrEXT.r2i = (void * (*) (const v3_ext_method*, v3_ext_ctx*, OPENSSL_CONST char*)) NULL;
  vomsAttrEXT.i2r = (int (*)(const v3_ext_method*, void*, BIO*, int)) NULL;
#else
  vomsAttrEXT.i2s = (char* (*)(v3_ext_method*, void*))NULL;
  vomsAttrEXT.s2i = (void* (*)(v3_ext_method*, v3_ext_ctx*, OPENSSL_CONST char*))NULL;
  vomsAttrEXT.i2v = (STACK* (*)(v3_ext_method*, void*, STACK*))NULL;
  vomsAttrEXT.v2i = (void* (*)(v3_ext_method*, v3_ext_ctx*, STACK*))NULL;
  vomsAttrEXT.r2i = (void * (*) (v3_ext_method*, v3_ext_ctx*, OPENSSL_CONST char*)) NULL;
  vomsAttrEXT.i2r = (int (*)(v3_ext_method*, void*, BIO*, int)) NULL;
#endif

  vomsAttrEXT.usr_data = NULL;

  X509V3_EXT_add(&vomsAttrEXT);
  return 0;
#else
  return -1;
#endif
}


int VOMSInfo::parseCert(X509 *cert)
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  int extcount = X509_get_ext_count(cert);

  DMESG("Number of certificate extensions: " << extcount << std::endl);

  for(int i = 0; i < extcount; i++) {
    const char        *extstr;
    X509_EXTENSION    *ext;
    char buf[201];
    
    ext = X509_get_ext(cert, i);
    extstr = OBJ_nid2sn(OBJ_obj2nid(X509_EXTENSION_get_object(ext)));
    OBJ_obj2txt(buf, 200, ext->object, 1);
    DMESG("  Name: " << extstr << " (" << buf << ")\n");
    
    if(strcmp(extstr, "VOMSAttrSequence")) 
      continue;

    OPENSSL_CONST unsigned char  *data;
    STACK_OF(CONF_VALUE) *val;
    CONF_VALUE           *nval;

// For SSL 1.0 Higher : SSL 1.0 change SSL_METHOD to const SSL_METHOD
#if OPENSSL_VERSION_NUMBER >= 0x10000000L
    X509V3_EXT_METHOD    const *meth;
#else
    X509V3_EXT_METHOD    *meth;
#endif

    void                 *ext_str = NULL;
 
    if (!(meth = X509V3_EXT_get(ext))) {
      DMESG("    No method to read V3 extension\n");
      continue;
    }
     
    data = ext->value->data;
    long int length = ext->value->length;
    DMESG("Data : " << (void *)data << "  len: " << length << std::endl);

    if (meth->it) {
      DMESG("With it!\n");
      ext_str = ASN1_item_d2i(NULL, &data, ext->value->length,
			      ASN1_ITEM_ptr(meth->it));
    } else {
      DMESG("Without it!\n");
      ext_str = meth->d2i(NULL, &data, ext->value->length);
    }

    DMESG(" Got data structure: " << ext_str << std::endl);
    if(!ext_str) {
      sslPrintErrors();
      continue;
    }
    attributes = (ACSequence *)ext_str;

  }

  if(!attributes)
    return 1;

  int nAttrCerts = sk_num((const STACK*)(attributes->attributeCertificates));
  DMESG("Number of attribute certificates " << nAttrCerts << std::endl);

  for (int i = 0; i< nAttrCerts; i++) {
    AttributeCertificate *ac;
    ac = (AttributeCertificate *)
      sk_value((const STACK*)attributes->attributeCertificates, i);
    
    DMESG("  Version: " << ASN1_INTEGER_get(ac->attributeCertInfo->version)
	  << "  validity: " << (void *) ac->attributeCertInfo->validity 
	  << "\n" );
    
    int nAttrs = sk_num((const STACK*)ac->attributeCertInfo->attributes);

    DMESG("  Number of attrs: " << nAttrs << std::endl);

    for (int j = 0; j < nAttrs; j++) {
      ACAttribute *attr; 
      attr = (ACAttribute *)sk_value((const STACK*)ac->attributeCertInfo->attributes, j);

      int nIETFAttr = sk_num((const STACK*)attr->ietfAttribute);
      DMESG("Number of IETF Attributes: " << nIETFAttr << std::endl);

      for (int k = 0; k < nIETFAttr; k ++) {
	ACIETFAttribute *ietfAttr;
	ietfAttr = (ACIETFAttribute *)sk_value((const STACK*)attr->ietfAttribute,k);
	
	int nNames  = sk_num((const STACK*)ietfAttr->names);
	int nValues = sk_num((const STACK*)ietfAttr->values); 

	DMESG("Names: " << nNames << "  values: " << nValues << std::endl);

	if(nNames != 1) {
	  ERRLOG("ERROR: Illegal attribute certificate: More than one VO!");
	  return 1;
	}
	if(nValues <1) {
	  ERRLOG("ERROR: Illegal attribute certificate: No attributes");
	  return 1;
	}
	
	GENERAL_NAME *name;
	name = (GENERAL_NAME *)sk_value((const STACK*)ietfAttr->names, 0);
	std::string vo;
	std::string voServer;
	if(name->type == GEN_URI) {
	  std::string voString((char *)name->d.ia5->data, name->d.ia5->length);
	  DMESG("name: " << voString << std::endl);
	  if(voString.find("://") != std::string::npos) {
	    vo = voString.substr(0, voString.find("://"));
	    voServer = voString.substr(voString.find("://")+3);
	    DMESG("VO >" << vo << "<  VO-Server >" << voServer << "<\n");
	  } else {
	    vo = voString;
	  }
	} else {
	  LOG("ERROR: Attribute name is not VOMS-URL!\n");
	}

	for (int l = 0; l < nValues; l++) {
	  ASN1_OCTET_STRING *value;
	  value = (ASN1_OCTET_STRING *)sk_value((const STACK*)ietfAttr->values, l);

	  Capability c;
	  c.vo = vo;
	  c.voServer = voServer;

	  std::string aString((char *)value->data, value->length);
	  DMESG("Attribute is: " << aString << std::endl);
	  
	  parseAttribute(aString, c.group, c.role, c.capability);

	  capabilities.push_back(c);
	}
      }
    }
  }

  return 0;
#else
  return 1;
#endif
}


int VOMSInfo::parseAttribute(const std::string &attribute,
			     std::string &group,
			     std::string &role,
			     std::string &capability) const
{
  //  DMESG("Parsing attribute >" << attribute << "<\n");
  std::string aString(attribute);

  size_t voEnd = aString.find(":");
  if(voEnd != std::string::npos)
    aString = aString.substr(voEnd+1);

  size_t capStart = aString.find("/Capability=");
  if(capStart != std::string::npos) {
    capability = aString.substr(capStart + 12);
    aString = aString.substr(0, capStart);
  }
  size_t roleStart = aString.find("/Role=");
  if(roleStart != std::string::npos) {
    role = aString.substr(roleStart + 6);
    aString = aString.substr(0, roleStart);
  }
  
  if(aString.size())
    group = aString;
  
  //  DMESG("role: >" << role << "< group: >" << group << "<\n");

  return 0;
}

/** Check whether the previously parsed attributes in a certificate
    allow a user to acquire a requested role
    \param role Role string with the format GROUP/Role=ROLE/Capability=CAPABILITY
        where 
        - GROUP is e.g. /Groupname
        - ROLE is e.g. Role
	- CAPABILITY is e.g. NULL
    \param allowedVO Requested VO, can be '*'
    \param vo Assigned VO
    \param groups All groups allowed by the certificate's attribute 
    Returns true on successful authorization
*/
bool VOMSInfo::allowedRole(const std::string &role, 
			   const std::string &allowedVO,
			   std::string &vo, 
			   std::list< std::string > &groups) const
{
  std::string reqGroup, reqRole, reqCap;
  if(role != "NULL"){
    parseAttribute(role, reqGroup, reqRole, reqCap);
  } else {
    reqGroup = "*";
    reqRole = "NULL";
    reqCap = "NULL";
  }

  DMESG("Checking allowedRole, requested group >" << reqGroup 
	<< "<, role >" << reqRole << "<, cap >" << reqCap 
	<< "< in VO >" << allowedVO << "<\n");

  std::list < Capability >::const_iterator I = capabilities.begin();
  while(I != capabilities.end() )
    groups.push_back(I++->group);

  I = capabilities.begin();
  while(I != capabilities.end() ) {
    DMESG("Role >" << I->role << "< vs " << reqRole << "\n");
    if( I->role != "" ) {
      DMESG("Old style Cert\n");
      if(I->role != reqRole) {
	I++;
	continue;
      }
      
      DMESG("Group " << I->group << " vs " << reqGroup << "\n");
      if(I->group != reqGroup && reqGroup != "*") {
	I++;
	continue;
      }
    }    

    DMESG("1.\n");
    if(allowedVO == "*")
      return true;

    DMESG("2.\n");
    if(!allowedVO.size())
      return false;

    DMESG("Now VO\n");

    std::string allowedVOName;
    std::string voServer;
    if(allowedVO.find("://") != std::string::npos) {
      allowedVOName = allowedVO.substr(0, allowedVO.find("://"));
      voServer = allowedVO.substr(allowedVO.find("://")+3);
    } else {
      allowedVOName = allowedVO;
    }

    if( !voServer.size() || voServer == "*" || voServer == I->voServer) {
      if( allowedVOName == I->vo || allowedVOName == "*") {
	vo = I->vo;
	// group = I->group;
	return true;
      }
    }
    I++;
  }

  vo = "";
  groups.clear();
  return false;
}
