/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


/** -*- Mode:c++ -*-
    $Id: VOMSInfo.h,v 1.12 2012/02/06 05:41:54 gpark Exp $

    VOMSInfo takes a certificate issued by a VOMS and sent by the
    client during the SSL handshake and extracts the attributes stored
    in X509v3 certificate extensions.
    The server can then access the extracted data structures and use
    the informatio for authentication and authorization.
**/

#ifndef VOMSINFO_H
#define VOMSINFO_H

#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/safestack.h>

#include <string>
#include <list>

typedef struct {
  ASN1_ENUMERATED            *type;
  ASN1_OBJECT                *oid;
  X509_ALGOR                 *algorithm;
  ASN1_BIT_STRING            *digest;
} ACDigest;

typedef struct {
  STACK_OF(GENERAL_NAME)     *issuer;
  ASN1_INTEGER               *serial;
  ASN1_BIT_STRING            *uid;
} ACIssuer;

typedef struct {
  STACK_OF(GENERAL_NAME)     *names;
  ACIssuer                   *issuer;
  ACDigest                   *digest;
} ACForm;

typedef struct {
  ACIssuer                   *baseid;
  STACK_OF(GENERAL_NAMES)    *name;
  ACDigest                   *digest;
} ACHolder;

typedef struct {
  ASN1_GENERALIZEDTIME       *notBefore;
  ASN1_GENERALIZEDTIME       *notAfter;
} ACValidity;

typedef struct {
  STACK_OF(GENERAL_NAME)     *names;
  STACK_OF(ASN1_OCTET_STRING)      *values;
} ACIETFAttribute;

DECLARE_STACK_OF(ACIETFAttribute)
DECLARE_ASN1_SET_OF(ACIETFAttribute)

typedef struct {
  ASN1_OBJECT                *type;
  STACK_OF(ACIETFAttribute)  *ietfAttribute;
} ACAttribute;

DECLARE_STACK_OF(ACAttribute)

typedef struct {
  ASN1_INTEGER               *version;
  ACHolder                   *holder;
  ACForm                     *form;
  X509_ALGOR                 *signatureAlgorithm;
  ASN1_INTEGER               *serialNumber;
  ACValidity                 *validity;
  STACK_OF(ACAttributes)     *attributes;
  ASN1_BIT_STRING            *id;
  STACK_OF(X509_EXTENSION)   *extensions;
} AttributeCertificateInfo;

typedef struct {
  AttributeCertificateInfo   *attributeCertInfo;
  X509_ALGOR                 *signatureAlgorithm;
  ASN1_BIT_STRING            *signature;
} AttributeCertificate;

DECLARE_STACK_OF(AttributeCertificate)

typedef struct {
  STACK_OF(AttributeCertificate) *attributeCertificates;
} ACSequence;

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
DECLARE_ASN1_FUNCTIONS(ACDigest)
DECLARE_ASN1_FUNCTIONS(ACIssuer)
DECLARE_ASN1_FUNCTIONS(ACForm)
DECLARE_ASN1_FUNCTIONS(ACHolder)
DECLARE_ASN1_FUNCTIONS(ACValidity)
DECLARE_ASN1_FUNCTIONS(ACIETFAttribute)
DECLARE_ASN1_FUNCTIONS(ACAttribute)
DECLARE_ASN1_FUNCTIONS(AttributeCertificateInfo)
DECLARE_ASN1_FUNCTIONS(AttributeCertificate)
DECLARE_ASN1_FUNCTIONS(ACSequence)
#endif

  typedef struct {
    std::string vo;
    std::string voServer;
    std::string group;
    std::string role;
    std::string capability;
  } Capability;


class VOMSInfo {
public:
  VOMSInfo(X509 *cert);
  ~VOMSInfo();
  
  bool validCert() const { return parseOK; };

  static int initASN1Parser();

  bool allowedRole(const std::string &role, const std::string &allowedVO,
		   std::string &vo,
		   std::list< std::string > &groups) const;


  std::list< Capability > capabilities;

private:
  int parseCert(X509 *cert);
  int parseAttribute(const std::string &attribute, std::string &group,
		     std::string &role, std::string &capability) const;

  static int vomsAttrNID;
  static X509V3_EXT_METHOD vomsAttrEXT;
  bool parseOK;
  ACSequence *attributes;

};

#endif // VOMSINFO_H
  
// For Openssl version 1.0 or Highter  
#if defined(OPENSSL_VERSION_NUMBER) && OPENSSL_VERSION_NUMBER >= 0x10000000L
typedef _STACK STACK;
#endif

