/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: VOMSUserManager.cc,v 1.13 2012/02/01 06:42:03 gpark Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "VOMSUserManager.h"
#include "util/sslHelpers.h"

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <fstream>
#include <iostream>
#include <sstream>
#include <sys/types.h>
#include <regex.h>
#include <libxml/xmlversion.h>
#include <libxml/globals.h>
#include <libxml/tree.h>

static bool debug = false;
#include "debug.h"

#define MAX_SUBJECT_LENGTH 256
#define MAX_PASS_LENGTH 64


VOMSUserManager::VOMSUserManager()
  : sock(0), ctx(NULL), ssl(NULL)
{
  DMESG("Initializing VOMSUserManager\n");
  userManagerName="VOMSUserManager";
}


VOMSUserManager::~VOMSUserManager()
{
  // We don't hold any resource...
}


int VOMSUserManager::init(const std::string &cert,
			  const std::string &key)
{
  hostcert = cert;
  hostkey = key;

  return 0;
}
 

int VOMSUserManager::addVOMS(const std::string &URL,
			     const std::string &name)
{
  std::string hostname;
  std::string path;
  int port;

  LOG("VOMSUserManager: adding user " << name << " as members of "
      << URL << "..." << std::endl);

  if(parseURL(URL, hostname, path, port)) {
    ERRLOG("VOMSUserManager: Illegal URL: " << URL << std::endl);
    return -1;
  }

  if(connectToServer(hostname, port)) {
    ERRLOG("VOMSUserManager: Could not connect\n");
    return -1;
  }

  if(initSSL(hostcert, hostkey)) {
    ERRLOG("VOMSUserManager: Could not initiate SSL connection\n");
    return -1;
  }

  std::string document;
  if(doGetRequest(path, document)) {
    ERRLOG("VOMSUserManager: Could not retrieve VOMS listing\n");
    return -1;
  }

  DMESG("---------------------------------------\n" << document
	<< "\n---------------------------------------\n" );

  if(disconnect()) {
    ERRLOG("Could not disconnect properly from VOMS server\n");
  }

  if(parseXML(document, name)) {
    ERRLOG("VOMSUserManager: Could not parse VOMS listing\n");
    return -1;
  }

  DMESG("VOMSUserManager: user " << name << " added\n");

  return 0;
}
 

/** Checks whether a certain subject can be used as authorization for
    a given user (role).
    Note that several subjects can be used as a credential for a given
    role.
*/
bool VOMSUserManager::checkSubject(const std::string &user,
				   const std::string &subject) const
{
  std::map< std::string, std::string >::const_iterator I 
    = certificates.begin();
  if((I=certificates.find(subject)) != certificates.end()) {
    DMESG("Matching to role >" <<  user<< "<\n");
    if(I->second == user)
      return true;
  }
  DMESG("checkSubject: not allowing role " << user << std::endl);
  
  return false;
}


/** Checks whether the subject is known to the system
 */
bool VOMSUserManager::isKnownSubject(const std::string &subject) const
{
  std::map< std::string, std::string >::const_iterator I;
  if((I=certificates.find(subject)) != certificates.end())
    return true;

  return false;
}


int VOMSUserManager::connectToServer(const std::string &hostname,
				     int port)
{
  // Use the first address.
  if( (sock=socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    ERRLOG("VOMSUserManger: Could not open socket\n");
    return -1;
  }

  struct addrinfo req, *ans;

  // Request only IPv4 sockets
  bzero(&req,sizeof(struct addrinfo));
  req.ai_family = AF_INET;
  req.ai_socktype = SOCK_STREAM;
  int err;
  if ((err = getaddrinfo(hostname.c_str(), NULL, &req, &ans)) != 0) {
    ERRLOG("VOMSUserManger: Hostname lookup failed: " 
	   << gai_strerror(err) << std::endl);
    return -1;
  }

  // Must set the port.
  ((struct sockaddr_in*)ans->ai_addr)->sin_port = htons(port);
  DMESG("Connecting to " << hostname << "  port " << port << std::endl);
  
  err = connect(sock, ans->ai_addr, ans->ai_addrlen);

  // Release the memory allocated by getaddrinfo. It's not needed anymore
  freeaddrinfo(ans);
  if (err < 0) {
    std::stringstream errstring;
    errstring << "Could not connect to server " << hostname 
	      << ":" << port;
    ERRLOG("VOMSUserManger: " << errstring << std::endl);
    return -1;
  }

  DMESG("VOMSUserManger: connected\n");

  return 0;
}


int VOMSUserManager::initSSL(const std::string &cert,
			     const std::string &key)
{
  // Creates client instance
  // For SSL 1.0 Higher : SSL 1.0 change SSL_METHOD to const SSL_METHOD
#if OPENSSL_VERSION_NUMBER >= 0x10000000L
  const SSL_METHOD *method= SSLv23_client_method();
#else
  SSL_METHOD *method= SSLv23_client_method();
#endif

  // Create SSL context
  ctx = SSL_CTX_new(method);
  //  SSL_set_connect_state(ssl);
  // Disable buggy v2, enable all compatibility measures
  SSL_CTX_set_options(ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);

  if(SSL_CTX_use_certificate_chain_file(ctx, cert.c_str())!=1){
    ERRLOG("VOMSUserManger: Could not load client certificate\n");
    sslPrintErrors();
    return -1;
  }

  // Load private key from local file
  if(SSL_CTX_use_PrivateKey_file(ctx, key.c_str(), SSL_FILETYPE_PEM)!=1){
    ERRLOG("VOMSUserManger: Could not load private key\n");
    sslPrintErrors();
    return -1;
  }
  
  // Verify private key
  if ( !SSL_CTX_check_private_key(ctx) ) {
    ERRLOG("VOMSUserManger: Could not verify private key\n");
    sslPrintErrors();
    return -1;
  }
  
  // see man BIO_f_ssl
  if(!(ssl = SSL_new(ctx))) {
    ERRLOG("VOMSUserManger: Could not create ssl object\n");
    return -1;
  }
  
  if(!SSL_set_fd(ssl, sock)) {
      ERRLOG("VOMSUserManger: Could not set socket in BIO\n");
      sslPrintErrors();
      return -1;
  }
    
  if(!SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY)) {
    ERRLOG("VOMSUserManger: Could not set SSL_MODE_AUTO_RETRY\n");
    return -1;
  }

  int res=SSL_connect(ssl);
  if(res<0){
    ERRLOG("VOMSUserManger: Could not do SSL handshake: Fatal error\n");
    sslPrintErrors();
    return -1;
  }

  if(res==0) {
    ERRLOG("VOMSUserManger: Could not do SSL handshake\n");
    sslPrintErrors();
    return -1;
  }
  
  return 0;
}


int VOMSUserManager::disconnect()
{
  if(ssl) {
    SSL_shutdown(ssl);
    SSL_free(ssl);
  }

  if(ctx)
    SSL_CTX_free(ctx);

  if(sock >=0){
    if (close(sock)) {
      ERRLOG("VOMSUserManger: Could not close socket");
      return -1;
    }
  }
  return 0;
}


int VOMSUserManager::sendRequest(const std::string &m)
{
  DMESG("Sending request: >" << m << "<\n");
  unsigned int totalbytes =0;
  int bytes;
  const char *buff=m.c_str();
  unsigned int size=m.size();

   while( (bytes=SSL_write(ssl, buff+totalbytes, size-totalbytes))>0) {
      totalbytes+=bytes;
      if(totalbytes == size)
	break;
    }

  if(bytes<=0){
    ERRLOG("VOMSUserManger: Could not send command");
    return -1;
  }

  DMESG("Request sent!\n");
  return 0;
}


int VOMSUserManager::readResponse(std::string &m)
{
  DMESG("Reading response\n");
  char buff[1001];
  int bytes = 0;
  int bodyBytes = 0;
  int bodyBytesRead = -1;
  int httpError = -1;
  bool connectionClose = false;

  while ( bodyBytesRead < bodyBytes || connectionClose ) {
    DMESG("READING " << bodyBytesRead << "  " << bodyBytes << std::endl);
    bytes=SSL_read(ssl, buff, 1000);

    if (bytes < 0)
      return bytes;

    buff[bytes] = '\0';

    //   DMESG("Got >" << buff << "<\n");

    m.append(buff);

    // See whether we have the status code
    if(httpError < 0) {
      //      DMESG("Trying to find OK\n");
      size_t pos = m.find("\r\n");
      if(pos != std::string::npos && pos >10) {
	sscanf(m.c_str() + 9, "%d", &httpError);
      }
      DMESG("HTTP status " << httpError << std::endl);
    }

    // See whether we have a full header
    if(httpError == 200) {
      if( !(bodyBytes || connectionClose) ) {
	size_t pos = m.find("Content-Length:");
	if(pos == std::string::npos) {
	  if(m.find("Connection: close") != std::string::npos) {
	    connectionClose = true;
	    DMESG("No http keepalive\n");
	  } else {
	    ERRLOG("VOMSUserManger: Illegal document\n");
	    DMESG("====\n" << m << "\n===\n");
	    return -1;
	  }
	}
	sscanf(m.c_str() + pos + 16, "%d", &bodyBytes);
      }
    }
    
    if(httpError > 0 && httpError != 200) {
      ERRLOG("VOMSUserManger: Unknown http response\n");
      return -1;
    }

    if( (bodyBytes || connectionClose) && bodyBytesRead <0) {
      size_t pos = m.find("\r\n\r\n");
      if(pos != std::string::npos && pos) {
	m = m.substr(pos + 4);
	bodyBytesRead = 0;
      }
    }

    if(bodyBytesRead >=0)
      bodyBytesRead = m.size();

    if(!bytes)
      break;
  }

  return 0;
}


int VOMSUserManager::doGetRequest(const std::string &request,
				  std::string &result)
{
  if(!ssl)
    return -1;

  std::stringstream s;
  s << "GET " << request << " HTTP/1.0\r\n"
    << "User-Agent: AMGA\r\n"
    << "Accept: */*\r\n"
    //    << "Host: " << hostname << "\r\n"
    << "Connection: Keep-Alive\r\n\r\n";

  if( sendRequest(s.str()) )
    return -1;
  
  return readResponse(result);
}


// https://lcg-voms.cern.ch:8443/voms/lhcb/services/VOMSAdmin?method=listMembers
int VOMSUserManager::parseURL(const std::string &URL,
			      std::string &hostname,
			      std::string &path,
			      int &port)
{
  regex_t reURL;
  
  if(regcomp(&reURL,
	     "https://([-a-zA-Z0-9_.]+)(:([0-9]+))?(/(.*))", REG_EXTENDED)){
     
    abort();
  }
  
  // Match URL
  regmatch_t match[6];
  int err = regexec(&reURL, URL.c_str(), 6, match, 0);

  if(!err){
    hostname=URL.substr(match[1].rm_so, match[1].rm_eo-match[1].rm_so);
    if(match[3].rm_so > -1)
      port = atol(URL.substr(match[3].rm_so,
			     match[3].rm_eo-match[3].rm_so).c_str());
    else
      port = 80;
    path = URL.substr(match[4].rm_so);
  } else {
    char buff[200];
    regerror(err, &reURL, buff, 200);
    ERRLOG("Illegal URL: " << URL << "\n  " << buff << std::endl);
    return -1;
  }
  
  DMESG("Host >" << hostname << "<  >" << port
	<< "<  >" << path << "<" << std::endl);

  regfree(&reURL);

  return 0;
}


int VOMSUserManager::parseXML(const std::string &xmlList,
			      const std::string &group)
{
  LIBXML_TEST_VERSION
  xmlKeepBlanksDefault(0);
  
  
  xmlDocPtr document = xmlParseMemory(xmlList.c_str(), xmlList.size());

  if(!document)
    return -1;

  xmlNodePtr current = xmlDocGetRootElement(document);
  // Search for listMembersReturn
  while(current) {
    DMESG("Name: " << current->name << std::endl);
    if(!strcmp((char *)current->name, "listMembersReturn"))
      break;
    current = current->xmlChildrenNode;
  }

  // Move to items
  xmlNodePtr items = current->children;
  DMESG("Item: " << items->name << std::endl);
  
  while(items) {
    current = items->children;
    DMESG("Child: " << current->name << std::endl);
    while(current && strcmp((char *)current->name, "DN")) {
      DMESG("Have " << current->name << std::endl);
      current = current->next;
    }
    if(!current)
      break;

    if( current->children && 
	!strcmp((char *)current->children->name, "text") ) {
      DMESG("Found " << ((char *)current->children->content)
	    << std::endl);
      certificates[(char *)current->children->content] = group;
    }
    items = items->next;
  }
  
  if(items)
    ERRLOG("Could not extract all subjects\n");

  if(document)
    xmlFreeDoc(document);

  return 0;
}


/** Lists user credentials
    Lists all credentials that a user has. If user is an empty string,
    then all users are listed
    Returns
    100 : User does not exist
*/
int VOMSUserManager::listUserCred(const std::string &user,
				  MDUserManager::UserCred &cred)
{
  int res = 100;

  std::map< std::string, std::string >::const_iterator I;
  I = certificates.begin();
  
  while( I != certificates.end() ) {
    if( user == I->second ) {
      cred.fixedDNs.push_back(I->first);
      res = 0;
    }
    I++;
  }

  return res;
}


/** Returns a list of all known users
    Returns 0, succeeds always
*/
int VOMSUserManager::listUsers( std::list < std::string > &users)
{
 std::map< std::string, std::string >::const_iterator I;
  I = certificates.begin();
  while( I != certificates.end() )
    users.push_back(I++->second);

  return 0;
}
