/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


//
// $Id: protocolHandling.cc,v 1.26 2010/11/08 07:40:39 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "protocolHandling.h"
#include "util/sslHelpers.h"
#include "util/Socket.h"
#include "util/SecurityUtils.h"
#include "server/MDServer.h"
#include "configparser/ConfigParser.h"
#include "server/GridMapUserManager.h"
#include "server/DBUserManager.h"
#include "server/VOMSAttrUserManager.h"
#include "server/VOMSUserManager.h"


#include <unistd.h>
#include <iostream>
#include <boost/algorithm/string.hpp>

static bool debug=false;
#include "debug.h"

using namespace std;

/** Negociates a connection between the client and the server.    
    Return value ==0: OK
    Return value 1: statistics requested
    @throw SocketException on error
*/
int initConnection(TCPSocket &socket, bool advertiseSSL, const std::string greeting,
            bool &useSSL, uint64_t &sessionID) throw (SocketException)
{
  sessionID=0;
  std::string out(greeting);
  out.append("\nProtocol 2\n");
  out.append("plain");
  if(advertiseSSL)
    out.append(" ssl");
   out.append("\n");

  socket.send(out);
  
  // Read requested connection/authentication type
  DMESG("Waiting for connection type" << std::endl);
  useSSL=false;
  std::string line;
  bool statistics = false;
  do {
    line = socket.readLine();    
    DMESG("LINE >" << line << "<" << std::endl);
    if(line == "statistics") {
      statistics = true;
    } else if(line=="ssl") {
      useSSL=true;
      DMESG("Using SSL" << std::endl);
    } else if(line.substr(0, 9) == "resumeSSL"){
      sessionID=strtoull(line.c_str()+9, NULL, 10);
      DMESG("Identified resume request for SSL " << sessionID << std::endl);
      useSSL = true;
    } else if(line.substr(0, 6) == "resume"){
      sessionID=strtoull(line.c_str()+6, NULL, 10);
      DMESG("Identified resume request for " << sessionID << std::endl);
    }

  } while (line.size());

  if( statistics )
    return 1;

  // If resuming a non-SSL session, do not send confirmation
  if(sessionID && !useSSL)
    return 0;

  DMESG("Sending OK" << std::endl);

  // Return OK
  socket.send("OK\n");
  DMESG("OK sent" << std::endl);
  return 0;
}


std::string groupList(const std::list < std::string > &groups)
{
  std::string gList;

  std::list< std::string>::const_iterator I = groups.begin();
  while( I!= groups.end()) {
    gList.append(*I++);
    if( I!= groups.end() )
      gList.append(", ");
  }
  return gList;
}


bool myproxyHack(const AuthContext &authCtx, MDServer *connection,
		 OPENSSL_CONST SSL *ssl, string &login)
{
  if(!authCtx.allowMyProxyHack)
    return false;

  STACK_OF(X509) *stack = SSL_get_peer_cert_chain(ssl);
  if(!stack)
    return false;
  
  DMESG("Doing MyProxy HACKED authentication!\n");

  for(int i = 0; i< sk_X509_num(stack); i++) {
    X509 *cert = sk_X509_value(stack, i);
    std::list< std::string > groups;
    std::string vo, eUser;
    if( connection->checkCert(login, cert, vo, groups, eUser) ){
      login = eUser;
      DMESG("Client cert authentication succeeded with MyProxy hack: user "
        << login << " in VO " << vo 
        << " with group " << groupList(groups) << std::endl);
      return true;
    }
  }

  return false;
}


bool authenticateUser(SSL *ssl, MDServer *connection,
		      const AuthContext &authCtx, UserContext &context)
{
   bool userAuthenticated = false;
  // We do role-based authorization via certs if
  // ssl is enabled and this is allowed
  if(ssl != NULL && authCtx.allowCertAuth) {
    DMESG("Trying client cert authentication...\n");
    X509 *cert=SSL_get_peer_certificate(ssl);
    if(cert) { // Check, there is a cert
      if(SSL_get_verify_result(ssl) == X509_V_OK) { // Check it's OK
        if(context.login == "") {
          context.login = "NULL";
        }
        std::string vo, eUser;
	std::list< std::string > groups;
        if(! connection->checkCert(context.login, cert, vo, groups, eUser)){
          LOG("Role " << context.login << " denied for '"
            <<  getCertSubjectName(cert) << "'\n");  
          userAuthenticated = myproxyHack(authCtx, connection, ssl, context.login);
        } else {
          if(eUser.size())
            context.login=eUser;
          LOG("Authentication succeeded for '" << getCertSubjectName(cert) << "': user "
            << context.login << " in VO '" << vo 
            << "' with groups '" << groupList(groups) << "'\n");
          userAuthenticated = true;
        }

        connection->addCertGroup(cert);
      } else {
        LOG("Certificate validation failed: " << SSL_get_verify_result(ssl) << std::endl);
      }
      X509_free(cert);
    } else {
      LOG("No certificate presented by peer!\n");
    }
  } else {  // Give reasons why no cert based authentication
     if(ssl == NULL)
       DMESG("No certificate based authentication because not an SSL session\n");
     if(!authCtx.allowCertAuth) {
       DMESG("Certificate based authentication disabled\n");
     }
  }

  // Do password based authentication if allowed
  if (authCtx.allowPasswordAuthentication && context.password.size()) {
    DMESG("Trying password authentication...\n");
    if(!connection->checkPass(context.login, context.password)){
      LOG("Password authentication for " << context.login
        << " failed because password incorrect\n");
    } else {
      LOG("Password authentication succeeded for " << context.login << "\n");
      userAuthenticated = true;
    }
  } else { // Explain why no password authentication done
     if(!authCtx.allowPasswordAuthentication) {
       DMESG("Password authentication disabled\n");
     }
     if(!context.password.size()) {
      DMESG("No password sent!\n");
     }
  }  

  return userAuthenticated;
}

void establishUserContext(TCPSocket &socket, struct UserContext &context) throw (SocketException)
{
  context.login = "nobody";
  context.name = "Unknown";
  context.permissionMask = "rwx";
  context.groupMask = "r-x";
  context.currentDir = "/";
  context.password = "";

  std::string line;
  while(1) {
    line = socket.readLine();
    DMESG("Line :>" << line << "<\n");
    if(!line.size())
      break;
    char *data;
    int field=strtol(line.c_str(), &data, 10);
    if(strlen(data) < 1){
      DMESG("Illegal field" << std::endl);
    }
    switch(field) {
    case 0:
      context.login=data+1;
      DMESG("Login = >" << context.login << "<" << std::endl);
      break;
    case 1:
      context.name=data+1;
      DMESG("Name = " << context.name << std::endl);
      break;
    case 2:
      context.permissionMask=data+1;
      DMESG("PermissionMask = " << context.permissionMask << std::endl);
      break;
    case 3:
      context.groupMask=data+1;
      DMESG("Group Mask = " << context.groupMask << std::endl);
      break;
    case 4:
      context.currentDir=data+1;
      DMESG("CurrentDir = " << context.currentDir<< std::endl);
      break;
    case 5:
      context.password=data+1;
      DMESG("Got a password\n");
      break;
    default:
      DMESG("Unknown: " << line << endl);
    }
  }
}

void removeUserManagers( list<MDUserManager *> &uMs ) 
{

  DBUserManager *dbUserManager;
  GridMapUserManager *gridmapUserManager;
  VOMSUserManager *vomsUserManager;
  VOMSAttrUserManager *vomsattrUserManager;

  std::list< MDUserManager* >::const_iterator J;
  J=uMs.begin();

  while( J != uMs.end() ) {
    MDUserManager *uM = *(J);
    if((dbUserManager = dynamic_cast<DBUserManager *>(uM))) {
      DMESG("Delete DBUser\n");
      delete dbUserManager;
    }
    if((gridmapUserManager = dynamic_cast<GridMapUserManager *>(uM))) {
      DMESG("Delete GridMapUser\n");
      delete gridmapUserManager;
    }
    if(vomsUserManager = dynamic_cast<VOMSUserManager *>(uM)) {
      DMESG("Delete VOMSUser\n");
      delete vomsUserManager;
    }
    if(vomsattrUserManager = dynamic_cast<VOMSAttrUserManager *>(uM)) {
      DMESG("Delete VOMSAttrUser\n");
      delete vomsattrUserManager;
    }
    J++;
  }
  return ;
}

void createUserManagers(const ConfigParser &configParser, list<MDUserManager *> &userManagers) {
  try {
    if(configParser.getInt("UserDB")) {
      DBUserManager *uM = new DBUserManager();
      userManagers.push_back(uM);
      DMESG("Using DBUserManager\n");
    }
  } catch (std::string e) { }

  try {
    std::string mapFileName=configParser.getString("MapFile");
    GridMapUserManager *uM = new GridMapUserManager();
    assert(uM);
    if(uM->initFromGridMap(mapFileName)) {
      ERRLOG("Could not read grid-map-file\n");
    } else {
      DMESG("Using GridMapUserManager\n");
      userManagers.push_back(uM);
    }
  }  catch (std::string e) { }

  if( configParser.isDefined("VOMSGroups") ) {
    VOMSUserManager *uM = new VOMSUserManager();
    assert(uM);
    try{
      if( uM->init(configParser.getString("CertFile"),
           configParser.getString("KeyFile")) ) {
    ERRLOG("Could not initialize VOMS User Manager\n");
      } else {
    DMESG("Using VOMSMapUserManager\n");
    std::vector< std::string > vomsList
      = configParser.getStringList("VOMSGroups");

    for(size_t i = 0; i< vomsList.size(); i+=2 ) {
      if( i < vomsList.size()-1 ) {
        if(! uM->addVOMS(vomsList[i], vomsList[i+1]) ) {
          DMESG("Added group " << vomsList[i+1]
          << " for members of " << vomsList[i] << std::endl);
        } else {
          ERRLOG("Could not add "  << vomsList[i+1]
             << " for members of " << vomsList[i] << std::endl);
        }
      }
    }
    userManagers.push_back(uM);
      }
    } catch (std::string err) {
      ERRLOG("Error initializing VOMSUserManager: "
         << err << std::endl);
    }
  }

  DMESG("Starting VOMSAttrUserManager\n");
  if( configParser.isDefined("VirtualOrganizations") ) {
    VOMSAttrUserManager *uM = new VOMSAttrUserManager();
    assert(uM);
    std::vector< std::string > voList
      = configParser.getStringList("VirtualOrganizations");
    
    for(size_t i = 0; i< voList.size(); i++ ) {
      if( !uM->allowVO(voList[i]) ) {
    DMESG("Added VO " << voList[i] << std::endl);
      } else {
    ERRLOG("Could not add "  << voList[i] << std::endl);
      }
    }

    assert(uM);
    try{
      std::vector< std::string > groupList
        = configParser.getStringList("VOGroupMap");

      for(size_t i = 0; i< groupList.size(); i++ ) {
    if( !uM -> mapGroup(groupList[i]) ) {
      DMESG("Added GroupMapping " << groupList[i] << std::endl);
    } else {
    ERRLOG("Could not add "  << groupList[i] << std::endl);
    }
      }
    } catch (std::string err) {
      DMESG("VOGroup not defined "<< err << std::endl);
    }

    try{
      std::vector< std::string > userList
    = configParser.getStringList("VOUserMap");

      for(size_t i = 0; i< userList.size(); i++ ) {
    if( !uM -> mapUser(userList[i]) ) {
      DMESG("Added user mapping " << userList[i] << std::endl);
    } else {
      ERRLOG("Could not add user map "  << userList[i] << std::endl);
    }
      }
    } catch (std::string err) {
      DMESG("VOUserMap not defined "<< err << std::endl);
    }
    
    userManagers.push_back(uM);
  }
}

std::string escapeString(char *str) {
    std::string s(str);
    escapeString(s);
    return s;
}

void escapeString(std::string &str) {
    DMESG("Escape 1 "<< str << std::endl);
    boost::replace_all(str, "\\", "\\\\");
    DMESG("Escape 2 "<< str << std::endl);
    boost::replace_all(str, "'", "''");
    DMESG("Escape 3 "<< str << std::endl);
}

void unescapeString(std::string &str) {
    boost::replace_all(str, "\\'", "'");
    boost::replace_all(str, "\\\\", "\\");
}

void unescapePreparedString(std::string &str) {
    boost::replace_all(str, "\\n", "\n");
    boost::replace_all(str, "\\r", "\r");
    boost::replace_all(str, "''", "'");
    boost::replace_all(str, "\\\\", "\\");
}
