/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sqlHelpers.h"
#include "sql.h"
#include <sstream>

#include <iostream>
static int debug = 0;
#include "debug.h"

/** Transforms shell wildcards * and ? to pattens which can be
    matched by the SQL LIKE operator. If shell wild-cards are found
    then also the SQL wild-cards are escaped.
    You must only match the resulting pattern with LIKE if also a
    pattern was found!
    returns 1 if a pattern was found, else 0;
*/
int sqlPattern(const std::string &in, std::string &out)
{
  DMESG("In " << in  << std::endl);
  int pattern=0;
  size_t i=0;
  
  // Check whether the outcome will be a pattern
  while(i<in.size()){
    size_t p = in.find_first_of("*?", i);
    if(p!=std::string::npos){
      if(!p || (p > 0 && in[p-1] != '\\') || (p > 1 && in[p-2] =='\\') ){
	pattern = 1;
	break;
      }
      i = p + 1;
      continue;
    }
    break;
  }

  i = 0;

  // Now replace * and ? by % and _ if not escaped
  // and if we have a pattern escape % and _
  while(i<in.size()){
    size_t p = in.find_first_of("*?%_", i);
    if(p!=std::string::npos){
      if(in[p]=='%'){
	out.append(in.substr(i, p-i));
	if(pattern)
	  out.append("\\%");
	else
	  out.append("%");
	i = p + 1;
	continue;
      }
      if(in[p]=='_'){
	out.append(in.substr(i, p-i));
	if(pattern)
	  out.append("\\_");
	else
	  out.append("_");
	i = p + 1;
	continue;
      }
      if(!p || (p > 0 && in[p-1] != '\\') || (p > 1 && in[p-2] =='\\') ){
	out.append(in.substr(i, p-i));
	if(in[p]=='*')
	  out.append("%");
	else
	  out.append("_");
	DMESG("Replaced: " << out << std::endl);
      } else {
	if ( p > 0 && in[p-1] == '\\') {
	  out.append(in.substr(i, p - i -1) );
	  out += in[p];
	} else {
	  out.append(in.substr(i, p-i+1));
	}
      }
      i=p+1;
    } else {
      out.append(in.substr(i));
      i=in.size();
    }
  }

    // Now we will need to take care of backslashes...

  return pattern;
}


//! Returns the SQL datatype as a string for an ODBC datatype (integer)
std::string sqlType(int type, int size, int precision)
{
  std::stringstream t;

  DMESG("sqlType: " << type << "   " << size << "   " 
	<< precision << std::endl); 

  switch(type){
  case SQL_CHAR:
    t << "char(" << size << ")";
    return t.str();
  case SQL_NUMERIC:
    t << "numeric(" << size << "," << precision << ")";
    return t.str();
  case SQL_DECIMAL:
    t << "decimal(" << size << "," << precision << ")";
    return t.str();
  case SQL_INTEGER:
    return "integer";
  case SQL_SMALLINT:
    return "smallint";
  case SQL_REAL:
    return "real";
  case SQL_FLOAT:
    return "double precision";
  case SQL_DOUBLE:
    return "double precision";
  case SQL_TYPE_TIME:
  case 10:     // Known to Postgres
    return "time";
  case SQL_DATETIME:
  case SQL_TYPE_DATE:
    return "date";
  case 11:     // Known to Postgres
  case SQL_TYPE_TIMESTAMP:
    return "timestamp";
  case SQL_VARCHAR:
    t << "varchar(" << size << ")";
    return t.str();

  case SQL_UNKNOWN_TYPE:
  case -1:
    return "UNKNOWN";
  }
  return "REALLY UNKNOWN TYPE";
}



bool dirRegexpPattern(const std::string &in, std::string &out)
{
  bool pattern = false;
  out = "^";
  size_t i = 0;
  while(i < in.size()){
    size_t p = in.find_first_of("*?[]$^", i);
    if(p!=std::string::npos){
      if(!p || (in[p-1] != '\\') || !(p-1) || ( in[p-2] !='\\') ){
	out.append(in.substr(i, p-i));
	if(in[p]=='*')
	  out.append("[^/]*");
	else if(in[p]=='?')
	  out.append("[^/]");
	else
	  out.append("\\").append(in.substr(i, p-i+1));
      } else {
	out.append(in.substr(i, p-i+1));
      }
      i=p+1;
      pattern = true;
    } else {
      out.append(in.substr(i));
      i=in.size();
    }
  }
  out.append("$");
  return pattern;
}


int pathPieces(const std::string &path, std::list< std::string > &parts)
{
  size_t pos=1;
  parts.push_back("/");
  DMESG("Path: >" << path << "<" << std::endl);
  size_t npos=path.find('/', pos+1);
  DMESG("Pos: " << npos << std::endl);

  while((pos=path.find('/', pos+1))!=std::string::npos) {
    DMESG("PART: >" << path.substr(0,pos) << "<" << std::endl);
    if(path[pos-1]!='\\')
      parts.push_back(path.substr(0, pos));
  }

  if(path!="/")
    parts.push_back(path);

  return parts.size();
}


bool dirLineIsBefore(const std::string &left, const std::string &right)
{
  size_t n = left.size() < right.size()? left.size():right.size();
  for(size_t i=0; i < n, right.size(); i++){
    if(right[i] == '/' && left[i] != '/')
      return true;
    if(right[i] > left[i])
      return true;
    if(right[i] < left[i])
      return false;
  }
  return true;
}


/** Checks whether @param left is a sub directory of @right.
    This does not mean it is a direct subdirectory! If you want
    to check this, too, use dirLevel, too.
*/
bool isSubdirOf(const std::string &left, const std::string &right)
{
  size_t n = left.size() < right.size()? left.size():right.size();
  if(left.substr(0, n) != right.substr(0, n))
    return false;
  if(left.size() <= n)
    return false;
  if(left[n-1] == '/' && right[n-1] == '/' && left[n] != '/')
    return true;
  if(left[n] != '/')
    return false;
  return true;
}


/** Return the directory depth
    Returns the depth of the directory level. / has level 0, /a has 1 etc.
*/
int dirLevel(const std::string &dir)
{
  int count=0;
  if(dir=="/")
    return count;
  for(size_t i = 0; i < dir.size(); i++){
    if(dir[i] == '/')
      count++;
  }
  return count;
}
