/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <stdlib.h>
#include <assert.h>
#include <stdexcept>
#include <memory>
#include <boost/thread/condition.hpp>

//#include "soap/soapMetadataService.h" // get server object

// #include "soap/soapH.h" // get server object
#include "DBConnection.h"
#include "DBResult.h"
#include "server/MDServerFactory.h"
#include "util/ThreadUtil.h"
#include "server/protocolHandling.h"

static bool debug = false;
#define logstream std::cout
#include "debug.h"

#define OPT_MAX_BUFFER "SoapMaxBuffer"


using namespace boost;
using namespace std;

/**
 *
 */
class AsynchCommandTask : public Runnable {
private:
	AsynchCommand *_cmd;
	MDStandalone *_server;
	MDAsyncBuffer *_rwBuffer;

public:
	AsynchCommandTask(AsynchCommand *cmd, MDStandalone *server, MDAsyncBuffer *rwBuffer)
	: _cmd(cmd), _server(server), _rwBuffer(rwBuffer) {
	}
	~AsynchCommandTask() {
// 		//DMESG("[~AsynchCommandTask]\n");
		delete _cmd;
	}

	void run() {
		try {
			// Thread cancelation in the MDServer is enabled only during
			// key points, in order to avoid the cancels crashing the program
			// from STL calls that do not retrow the cancelation exception.
			// Therefore, threads calling the MDServer must have cancelation
			// disabled.
			int oldState;
			ec(pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oldState));
			DMESG("[AsynchCommandTask]: Running\n");
			_cmd->executeCommand(_server);
			_rwBuffer->finish();
		} catch (std::logic_error &e) {
			DMESG("[AsynchCommandTask]: Interrupted by exception: " << e.what() << endl);
			_rwBuffer->cancel();
		} catch (BufferCanceledException &bce) {
			DMESG("[AsynchCommandTask]: Buffer cancelled\n");
			_rwBuffer->cancel();
		} catch (...) {
			DMESG("[AsynchCommandTask]: Unknown exception\n");
			_rwBuffer->cancel();
			throw;
		}

		DMESG("[AsynchCommandTask]: Finished\n");
	}
};

DBConnection::DBConnection(ConfigParser *config) :
  dbServer(NULL), _buffer(""), _busy(false)
{
	dbServer= (MDStandalone*)MDServerFactory::mdServer(_rwBuffer);
	if(!dbServer)
		return;

	if(!dbServer->isConnected()){
		LOG("Could not connect to DB!" << endl);
		return;
	}

        dbServer->setDebug(debug);

        
        try {
          int maxBufferSize = config->getInt(OPT_MAX_BUFFER);
          setBufferSize( maxBufferSize );
          // DMESG( "DBConnection MAX BUFFER : " << maxBufferSize << endl );
        } catch (string err) {} 


        // Initialize User Manager
        std::list<MDUserManager *> userManagers;
        createUserManagers(*(config), userManagers);
        dbServer->initializeUserManager(userManagers); 
        // Using dbConn to set a DB Connection for User Management


	// Setting user and permissions
	dbServer->setUser("root");
	dbServer->setPermissionMask("rwx");
	dbServer->setGroupRightsMask("r--");
	dbServer->setLogging(true);
	dbServer->setDebug(true);

	DMESG("Connected to DB" << endl);
}

DBConnection::~DBConnection() {
  DMESG("Destructor\n");
  delete dbServer;
}

void DBConnection::startRequest() {
	DMESG("startRequest.\n");
	dbServer->setUser("root");
	_rwBuffer.start();
}

void DBConnection::finishRequest() {
	DMESG("finishRequest.\n");
	_rwBuffer.finish();
}

bool DBConnection::getDebug() {
	return debug;
}
void DBConnection::setDebug(bool _debug) {
	debug = _debug;
}

void DBConnection::startAsynchRequest(AsynchCommand *cmd) {
	DMESG("startAsynchRequest.\n");
	assert(!_busy);
	// This object must not be created on the stack, because the thread
	// will outlive this method invocation. Since there will be only
	// one busy asynchronous request per DBconnection
	_busy = true;

	// Start the thread on the detached state. This is useful for cancelling
	// the thread, since like this we don't have to join with the thread,
	// which is a potentially time-consuming operation. Being able to cancel
	// the DB thread without blocking is especially important for the reaper
	// thread, which holds a lock on the SessionManager while cancelling
	// the expired threads.
	_rwBuffer.start();
//	ec(pthread_create(
//		&_dbThread,
//		NULL, //&thread_attr,
//		execute,
//		new AsynchCommandTask(cmd, dbServer, &_rwBuffer)));
}

void DBConnection::cancelAsynchRequest() {
	DMESG("cancelAsynchRequest\n");
	assert(_busy);

	_rwBuffer.cancel();
    /* TODO: the mdserver is not calling pthread_cancel(). Is it working here?
       If so, why didn't it work with the mdserver?
     */
	// The next time the DB thread tries to write something to the buffer,
	// it would be killed with an exception. But since that may take a while
	// if the thread is in the middle of a long running query, we can cancel
	// it explicitly.
	DMESG("Canceling DB thread: " << _dbThread << endl);
	if (pthread_cancel(_dbThread) == ESRCH) {
		DMESG("DB thread already died. Ignoring.\n");
	} else {

	}
	// The thread was created on the detached state, so no
	// need for joining with it.
	joinWithDBThread();
	_busy = false;
}


void DBConnection::checkForAsynchQueryEnd() {
  DMESG("Checking for end of asynchronous request\n");
}

void DBConnection::joinWithDBThread() {
  DMESG("Joining with DB thread\n");
  mutex::scoped_lock lk(monitor);
  assert(_busy);
  void *result;
  ec(pthread_join(_dbThread, &result));
  if (result == PTHREAD_CANCELED) {
    DMESG("DB thread was cancelled.\n");
  } else if (result == 0) {
    DMESG("DB thread exited normally.\n");
  } else {
    DMESG("DB thread exited with unknown result: " << result << endl);
  }
}

void DBConnection::getResult(DBResult& result)
throw (BufferCanceledException)
{
	DMESG("getResult." << endl);
	if (!isAsynchRequestActive()) {
		// If it is a single shot request we must explicitly tell the
		// buffer that we are done, otherwise a call to read will block.
		_rwBuffer.finish();
	}
	string str;
	_rwBuffer.read(str);
	if (str!="") {
		parseResponse(str, result);
	}
	checkForAsynchQueryEnd();
}

void DBConnection::getResult(DBResult& result, vector<string>& entries)
throw (BufferCanceledException)
{
	DMESG("getResult (and data)" << endl);
	if (!isAsynchRequestActive()) {
		_rwBuffer.finish();
	}
	string str;
	_rwBuffer.read(str);
	if (str!="") {
		parseResponse(str, result);
		parseData(str, entries);
	}
	checkForAsynchQueryEnd();
}

void DBConnection::getData(vector<string>& entries)
throw (BufferCanceledException)
{	
  DMESG("getData." << endl);
  string response;
  _rwBuffer.read(response);  // Check for timeout if cancelled
  parseData(response, entries);
  checkForAsynchQueryEnd();
}

void DBConnection::parseData(const string& str, vector<string> &data) {
//   DMESG("[parseData]" << endl);
  // We have already parsed the contents of the buffer, which
  // does not have any \n character. So continue parsing only
  // from the next character to be added to the buffer.
  int parsePosition = _buffer.size();
  int cutPosition = 0;
  _buffer.append(str);
  size_t index;
  while ((index = _buffer.find("\n", parsePosition)) != string::npos) {
    // Discard the \n
    data.push_back(_buffer.substr(cutPosition, index-cutPosition));
//     DMESG("Added line: >" << data.back() << "<" << endl);
    cutPosition = index+1;
    parsePosition = index+1;
  }
// index == string::npos.
// Discard the part of the buffer that was already consumed.
//   _buffer = _buffer.substr(cutPosition);
  _buffer.erase(0, cutPosition);
}

//void DBConnection::parseData(const string& str, vector<ns__MDEntry> &entries) {
////   DMESG("[parseData]" << endl);
//  // We have already parsed the contents of the buffer, which
//  // does not have any \n character. So continue parsing only
//  // from the next character to be added to the buffer.
//  int parsePosition = _buffer.size();
//  int cutPosition = 0;
//  _buffer.append(str);
//  size_t index;
//  int i = 0;
//  while ((index = _buffer.find("\n", parsePosition)) != string::npos) {
//	// TODO: Place the data on the entry field. Right now the server 
//	// does not provide enough information to fill out all the fields on
//	// the MDEntry object. Birger, could you fix it?
//	entries.resize(i+1);
//    // Discard the \n
//	entries[i].entry = _buffer.substr(cutPosition, index-cutPosition);
//	entries[i].attributes.resize(1);
//	entries[i].attributes[0].name = "unavailable";
//	entries[i].attributes[0].schema = "unavailable";
//	entries[i].attributes[0].type = "unavailable";
//	entries[i].attributes[0].value = "unavailable";
//
////    data.push_back(_buffer.substr(cutPosition, index-cutPosition));
////     DMESG("Added line: >" << data.back() << "<" << endl);
//    cutPosition = index+1;
//    parsePosition = index+1;
//  }
//// index == string::npos.
//// Discard the part of the buffer that was already consumed.
////   _buffer = _buffer.substr(cutPosition);
//  _buffer.erase(0, cutPosition);
//}

void DBConnection::parseResponse(string& str, DBResult& result) {
  size_t index = str.find("\n", 0);
  if (index == string::npos) {
    LOG("Error: no \\n found\n");
    return;
  }
  result.serverResult = str.substr(0, index);
  // Discard also the \n
  str.erase(0, index+1);
  const char* firstline = result.serverResult.c_str();
  char *endPtr;
  result.errorCode = strtol(firstline, &endPtr, 10);
  if (endPtr == firstline) {
    LOG("Errror! Could not parse number: " << firstline << endl);
  }
  // Removes the separator white space.
  if (*endPtr == ' ') {
    endPtr++;
  }
  result.errorMsg = endPtr;
}
