/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef DBCONNECTION_H
#define DBCONNECTION_H
#include <pthread.h>
#include <vector>
#include <string>
#include <pthread.h>

#include <boost/thread/condition.hpp>

#include "server/MDAsyncBuffer.h"
#include "MDSoapBuffer.h"
#include "server/MDStandalone.h"
#include "configparser/ConfigParser.h"
#include "DBResult.h"
// #include "soap/soapStub.h"


class AsynchCommand {
public:
	AsynchCommand() {
	}
	virtual ~AsynchCommand() {
	}
	virtual void executeCommand(MDStandalone *server) = 0;
};

class DBConnection {
public:
	DBConnection(ConfigParser *config);
	~DBConnection();

	/** For synchronous (non-streaming) operations */
	void startRequest();
	void finishRequest();

	/** For asynchronous (streaming) operations */
	void startAsynchRequest(AsynchCommand *cmd);
	/** Terminates request */
	void cancelAsynchRequest();
	/**
	* One of the getResult methods must be called once after executing
	* the operation on the dbServer field.
	* This particular overload of getResult extracts the initial result,
	* but reads no data from the buffer. Should be used on requests that
	* do not return any data.
	* @param result
	*/
	void getResult(DBResult& dbResult, std::vector<std::string>& entries)
		throw (BufferCanceledException);

	/**
	* Similar to the other getResult method, with the difference that
	* this will extract the result and try to parse any data retrieved
	* from the asynchronous buffer. The data will be stored on the
	* vector given as argument.
	* @param result
	*/
	void getResult(DBResult& dbResult)
		throw (BufferCanceledException);
	void getData(std::vector<std::string>& entries)
		throw (BufferCanceledException);

	/**
	 * @return Whether this connection is busy servicing an asynchronous request.
	 */
	bool isAsynchRequestActive() {return _busy;}
	MDStandalone *dbServer;

	static bool getDebug();
	static void setDebug(bool _debug);
 
        bool isOverFlow() { return _rwBuffer.isOverFlow(); }

        void setBufferSize( int size ) { _rwBuffer.setBufferSize(size); }

private:
	MDSoapBuffer _rwBuffer;
	bool _debug;
	/** Stores partial lines received from the AsyncBuffer */
	std::string _buffer;
	// Whether the connection is busy with a request.
	bool _busy;
	
	pthread_t _dbThread;
	boost::mutex monitor;
	

	void parseData(const std::string& response, std::vector<std::string> &entries);
	/** Removes from response the characters that were parsed. */
	void parseResponse(std::string& response, DBResult& dbResult);
	void checkForAsynchQueryEnd();
	void joinWithDBThread();
};
#endif // DBCONNECTION_H
