/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "DBConnectionPool.h"
#include "configparser/ConfigParser.h"
#include <stdexcept>

static bool debug = false;
#define logstream std::cout
#include "debug.h"

using namespace boost;
using namespace std;

DBConnectionPool::DBConnectionPool(ConfigParser *config, size_t nConnections) :
  _poolSize(nConnections), _config(config)
{
  if (nConnections <= 0) {
    throw logic_error("Invalid number of connections");
  }
  LOG("Opening " << _poolSize << " DB Connections\n");
  for (size_t i=0; i<_poolSize; i++) {
    _pool.push_back(new DBConnection(_config));
  }
}

void DBConnectionPool::createNewConnection() {
    mutex::scoped_lock lk(monitor);
	DMESG("[createNewConnection] Pool size: " << _pool.size() << endl);
	_pool.push_back(new DBConnection(_config));
}

DBConnectionPool::~DBConnectionPool()
{
  if (_pool.size() != _poolSize) {
    DMESG("Not all connections are checked-in\n");
//    throw logic_error("Not all connections are checked-in");
  }
  for (size_t i=0; i<_pool.size(); i++) {
    delete _pool[i];
  }
  delete _config;
}

/**
 * Blocks if no connection is available
 */
// DBConnection* DBConnectionPool::checkoutConnection() {
//   Guard<Mutex> g(_mutex);
//   DMESG("[checkOutConnection]\n");
//   while (_pool.empty()) {
//   	DMESG("Pool empty. Waiting.\n");
// 	_cnd.wait();
//   }
//   DBConnection* dbConn = _pool.back();
//   _pool.pop_back();
//   DMESG("[checkoutConnection] " << dbConn << endl);
//   return dbConn;
// }

/**
 * Returns null if no connection is available.
 */
DBConnection* DBConnectionPool::checkoutConnection() {
  mutex::scoped_lock lk(monitor);
  DBConnection* dbConn;
  if (_pool.empty()) {
  	dbConn = NULL;
	DMESG("[checkoutConnection] No connection available!\n");
  } else {
	dbConn = _pool.back();
	_pool.pop_back();
  }
  DMESG("[checkoutConnection] " << dbConn << endl);
  return dbConn;
}

/**
 */
void DBConnectionPool::checkinConnection(DBConnection *dbConn) {
	// TODO: Make sure the connection is still working properly
    mutex::scoped_lock lk(monitor);
	DMESG("[checkinConnection] " << dbConn << ", size: " << (_pool.size() + 1) << endl);
	_pool.push_back(dbConn);
	DMESG("[checkinConnection] Done\n");
}

/**
 * Returns the number of available connections on the
 * connection pool
 */
size_t DBConnectionPool::availableConnections() {
  return _pool.size();
}
