/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "MDSoapClient.h"
#define KEEP_ALIVE false

#include "WSDairH.h"
#include "WSDairStub.h"
#include "SQLAccess.nsmap"
#include "configparser/ConfigParser.h"
#include "util/SecurityUtils.h"

#include <signal.h>

static bool debug = false ;
#include "debug.h"

MDSoapClient *MDSoapClient::clientRegisteredForInt = NULL;

template <class T> std::string StringOf(T object)
{
  std::ostringstream os;
  os << object;
  return(os.str());
}

MDSoapClient::MDSoapClient(const std::string &host, int port, 
                           const std::string & path, 
                           const std::string &configFile, 
                           bool keep_alive, bool dbg)
  : debug(dbg), useSSL(false), authWithCert(0), certFile(""), keyFile(""),
    cancellationRequested(false)
{

   if(keep_alive)
      soap_init2(&soap, SOAP_IO_KEEPALIVE, SOAP_IO_KEEPALIVE);
   else soap_init(&soap);


  // Read configuration file
  char *c = getenv("HOME");
  std::string configHome;
  if(c) {
    configHome = c;
  } else {
    std::cerr << "HOME environment variable not set!" << std::endl;
    abort();
  } 
  configHome.append("/.mdclient.config");

  c = getenv("GLITE_LOCATION");
  std::string configDefault;
  if (c)
    configDefault = c;
  else
    configDefault = "/opt/glite";
  configDefault.append("/etc/mdclient.config");

  config = new ConfigParser();
  if(configFile.size()) {
    DMESG("Trying to read: " << configFile << std::endl);
    if(!config->init(configFile)) {
      cerr << "Could not read configuration: exiting" << std::endl;
      exit(5);
    }
  } else {
    if(!config->init("mdclient.config")) {
      DMESG("Can not read: mdclient.config " << std::endl);
      if(!config->init(configHome)) {
        DMESG("Can not read: " << configHome << std::endl);
        if(!config->init(configDefault)) {
          DMESG("Can not read: " << configDefault << std::endl);
          cerr << "Could not read configuration: exiting" << std::endl;
          exit(5);
        }
      }
    }
  }

  if( host.size() ) {
    _host = host;
    DMESG("Host : " << host << std::endl);
  } else {
    DMESG("try to read HOST Variable from File " << std::endl);
    try {
      _host = config->getString("Host");
    } catch (std::string e) {
      DMESG("CAn not read HOST Variable from File " << std::endl);
      _host = "localhost";
    }
  }
  DMESG("HOST: " << _host << std::endl);

  if( path.size() ) {
    _path = path;
  } else {
    _path = "/";
  }
  DMESG("PATH: " << _path << std::endl);

  if( port ) {
    _port = port;
  } else {
    try {
      _port=config->getInt("Port");
    } catch (std::string e) {
      _port = MDSOAP_DEFAULT_PORT;
    }
  }
  DMESG("PORT: " << _port << std::endl);

  requireSSL = false;
  bool trySSL = false;
  try {
    std::string s = config->getString("UseSSL");
    ConfigParser::toLower(s);
    if(s == "yes" || s == "require" || s == "1") {
      trySSL = true;
      requireSSL = true;
    } else if( s == "try") {
      trySSL = true;
    }

    DMESG("trySSL: " << trySSL << std::endl);
  
    if(! (s == "yes" || s == "1" || s=="require" || s == "try"
          || s == "no" || s == "0")) {
      cerr << "Unknown value for option UseSSL\n";
      exit(-1);
    }   
          
    if(trySSL) { 
      if(initsoapSSL() && requireSSL) {
        cerr << "Could not initialise SSL\n";
        exit(-1);
      } 
      access_end_point = "https://" ;
    } else {
      access_end_point = "http://" ;
    }

    access_end_point += _host + ":" + StringOf(_port) + _path ; 

  } catch (std::string err) { }

  DMESG("Access End Point : " <<  access_end_point  << endl);
 
  establishContext();

  setUser();

}

int MDSoapClient::initsoapSSL() {

  try {
      if( (authWithCert = config->getInt("AuthenticateWithCertificate")) ){
        int err = loadCertificate();
        if(err) {
          cerr << "Could not load certificate\n";
          return err;
        }
      }
  } catch (std::string &err) { authWithCert = 0 ;} ;


  soap_ssl_init();

  int ret;
  int sslmode ;
  if (authWithCert) { 
    // sslmode = SOAP_SSL_REQUIRE_SERVER_AUTHENTICATION | SOAP_SSL_SKIP_HOST_CHECK;
    sslmode = SOAP_SSL_NO_AUTHENTICATION | SOAP_SSL_SKIP_HOST_CHECK;
    DMESG("GSI Key File : " << keyFile << "\n");
    DMESG("GSI Cert File : " << certFile << "\n");
    ret = soap_ssl_client_context( &soap, sslmode, keyFile.c_str(), NULL
                , certFile.c_str(), NULL, NULL) ; 
  } else {
    sslmode = SOAP_SSL_NO_AUTHENTICATION ;
    ret = soap_ssl_client_context( &soap, sslmode, NULL, NULL, NULL, NULL, NULL) ; 
  }

  if ( ret ) { 
   cerr << "Failed to Make soap ssl Context \n";
   return -1; 
  } 

  return 0 ;
}

void MDSoapClient::setUser()
{
  soap.userid = (char*) _userid.c_str();
  soap.passwd = (char*) _passwd.c_str();
}

int MDSoapClient::establishContext()
{
  std::string login("NULL");
  try {
    login=config->getString("Login");
  } catch (std::string err) {} ;

  std::string pass("");
  try {
    pass=config->getString("Password");
  } catch (std::string err) {} ;

  _userid = login ;
  _passwd = pass ;

  return 0;
}


int MDSoapClient::loadCertificate()
{
  try {
    if(config->getInt("UseGridProxy")) {
      certFile=getProxyCertName();
      keyFile=certFile;
    }
  } catch (std::string err) { };

  if(!certFile.size()) {
    try{
      certFile=config->getString("CertFile");
      keyFile=config->getString("KeyFile");
    } catch (std::string err) { return -1;  };
  }

  return 0;
}


MDSoapClient::~MDSoapClient(){
  soap_end(&soap);
  soap_done(&soap);
}

void MDSoapClient::freeSoap()
{
  soap_destroy(&soap);
}

std::string MDSoapClient::getError()
{
  if ( soap.fault == NULL ) { 
    return "No fault message";
  } else if (soap.fault->faultstring == NULL ) {
    return "No Reason";
  } 

  return soap.fault->faultstring ;
}

/**
 * Signal handler for SIGINT
 *
 */
void int_sig_handler(int sig)
{
  signal(SIGINT, SIG_DFL);
  MDSoapClient::clientRegisteredForInt->cancellationRequested=true;
}

MDSoapClient *registerClient(MDSoapClient *client)
{
  if( !client ) {
    signal(SIGINT, SIG_DFL);
    MDSoapClient::clientRegisteredForInt = NULL;
    return NULL;
  }
  MDSoapClient *old=MDSoapClient::clientRegisteredForInt;
  MDSoapClient::clientRegisteredForInt=client;

  static struct sigaction sig;
  memset(&sig, 0, sizeof(sig));
  sig.sa_handler = int_sig_handler;
  sigaction(SIGINT, &sig, NULL);

  return old;
}


