/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <fstream>
#include <sstream>
#include <pthread.h>
#include <signal.h>
#include "stdsoap2.h"


#include "DBResult.h"
#include "util/ThreadUtil.h"
#include "server/MDStandalone.h"
#include "server/MDBuffer.h"
#include "util/sslHelpers.h"
#include "util/openssl_mt.h"

#include "SQLAccess.nsmap"
#include "WSDairH.h"

#include "server/MDInterpreter.h"
#include "server/DBHelper.h"
#include "util/Statement.h"
#include "util/TableEntry.h"
#include "server/VOMSInfo.h"

#include "WSDAIRUtil.h"
#include "MDSoapServer.h"
#include "S_SQLAccess.h"
#include "S_SQLAccessFactory.h"
#include "S_SQLResponse.h"
#include "S_SQLRowset.h"

typedef std::list<EntryProps> EntryPropList;

static bool debug = false;
// Activate debug and send it to cout
#define logstream std::cout
#include "debug.h"

// These are the keys read from the configuration file
// and the corresponding default values
#define OPT_MAX_CONNECTIONS "MaxConnections"
#define DEFAULT_MAX_CONNECTIONS 10

#define OPT_MAX_SESSIONS_PER_CLIENT "MaxSessionsPerClient"
#define DEFAULT_MAX_SESSIONS_PER_CLIENT 5

#define OPT_SOAP_KEEP_ALIVE "SoapKeepAlive"
#define DEFAULT_SOAP_KEEP_ALIVE true

#define OPT_WORKER_THREADS "WorkerThreads"
#define DEFAULT_WORKER_THREADS 10

#define OPT_SESSION_TIMEOUT "SessionTimeout"
// in seconds
#define DEFAULT_SESSION_TIMEOUT 60

#define OPT_QUERY_TIMEOUT "QueryTimeout"
// in seconds
#define DEFAULT_QUERY_TIMEOUT -1

#define OPT_INDIRECT_SERVICE_LIMIT "IndirectServiceLimit"
// in seconds
#define DEFAULT_INDIRECT_SERVICE_LIMIT -1

#define OPT_USE_SSL "UseSSL"
#define OPT_SERVER_CERT_KEY_FILE "CertKeyFile"
#define OPT_REQUIRE_AUTH "RequireAuthentication"
#define OPT_ALLOW_CERT_AUTH "AllowCertificateAuthentication"
#define OPT_ALLOW_PASS_AUTH "AllowPasswordAuthentication"
#define OPT_ALLOW_GRID_PROXY_LOGIN "AllowGridProxyLogin"
#define OPT_TRUSTED_CERT_DIR "TrustedCertDir"
#define OPT_DATA_RESOURCE_ABSTRACT_NAME "DataResourceAbstractName"
#define OPT_DATASET_FORMAT_URI "DatasetFormatURI"

#define SOAP_SOCKET_BACKLOG 50

using namespace std;

typedef BoundedQueue<struct soap*> TaskQueue;

std::string DEFAULT_DRAN = "Metadata";



string parseSourceIP(int ip) {
	ostringstream oss;
	oss << ((ip >> 24)&0xFF) << "." <<
		((ip >> 16)&0xFF) << "." <<
		((ip >> 8)&0xFF) << "." <<
		 (ip&0xFF);
	return oss.str();
}

void logCall(struct soap *soap, const string& callDesc) {
	LOG("Request from " << parseSourceIP(soap->ip) << ": " << callDesc << endl);
}

/**
 * Catch broken pipes. Taken from gSOAP examples. I'm not
 * sure why this is needed, but will use it anyway.
 */

void sigpipe_handle(int x) { }


class WorkerTask : public Runnable {
private:
  TaskQueue *_taskQueue;
  ConnectionManager _conMan;
  bool _useSSL;

public:
  WorkerTask(
    TaskQueue *taskQueue,
    DBConnectionPool *dbConnPool,
    SessionManager *sessionManager,
    bool useSSL
  ) : _taskQueue(taskQueue), _useSSL(useSSL)
  {
    _conMan.dbConnPool = dbConnPool;
    _conMan.sessionManager = sessionManager;
  }

  ~WorkerTask() {
    DMESG("~WorkerTask\n");
  }

  void run() {

    struct soap *soapCtx = NULL;

    try {
      DMESG("[Worker] " << pthread_self() << " created" << endl);
      while (true) {

        DMESG("[WORKER] Waiting for request\n");
        // Calls to MDServer disable cancellation. We must re-enable it
        // or else it is not possible to cancel the worker threads.
        ec(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL));
        try {
          // Get a request
          soapCtx = _taskQueue->next();
        } catch (boost::lock_error e) {
	  LOG("Master/Worker task queue cancelled\n");
          break;
        }

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
        if (_useSSL) {
        // Must establish the SSL context. This can be done either
        // on the master or in a slave thread. I'm not sure which
        // alternative is the best. I guess that this ssl_accept
        // involves several message exchanges with the client. If
        // so, then it's better to do it on a slave, in
        // order to free the master to accept other connections
          if (soap_ssl_accept(soapCtx)) {
            soap_print_fault(soapCtx, stderr);
            soap_done(soapCtx);
            free(soapCtx);
            continue; /* when soap_ssl_accept fails, we should just go on */
          }
        }
#endif

        DMESG("[WORKER] Received request. Going to process.\n");
        DMESG("[WORKER] Keep alive: " << soapCtx->keep_alive << endl);

        _conMan.cwd = "";

      // Pass the DBConnection pool and the session manager to the
      // SOAP server implementation. They are the ones who know if
      // DB connections should be returned to the pool after serving
      // the request or if they should be kept checked out for
      // streaming operations

        soapCtx->user = &_conMan;
        if (soap_serve(soapCtx) != SOAP_OK) {// process RPC request
          // TODO: soap_server returns an error often, even when the
          // operation succeed. Check what is the problem.
          // soap_print_fault(soapCtx, stderr); // print error*/
        }
        soap_destroy(soapCtx); // dealloc C++ data
        soap_end(soapCtx); // dealloc data and clean up
        soap_done(soapCtx); // detach soap struct
        soap_free(soapCtx);
      }
      DMESG("[Worker] Dying" << endl);
    } catch (...) {
      DMESG("[WORKER] Interrupted by unknown exception.\n");
      soap_destroy(soapCtx); // dealloc C++ data
      soap_end(soapCtx); // dealloc data and clean up
      soap_done(soapCtx); // detach soap struct
      soap_free(soapCtx);
      throw;
    }
  } /* end of run */

};

class MasterTask : public Runnable {
private:
  struct soap *_soap;
  TaskQueue *_taskQueue;

public:
  MasterTask(struct soap *soap, TaskQueue *taskQueue)
    : _soap(soap), _taskQueue(taskQueue) { }

  ~MasterTask() { }

  void run() {
    DMESG("[Master] " << pthread_self() << " created" << endl);
    for (int i = 1; ; i++) {

      DMESG("[Master] Waiting for connections...\n");
      int s = -1;	
      s = soap_accept(_soap);
      if (s < 0) {
        soap_print_fault(_soap, stderr);
        break;
      }
      DMESG("[Master] " << i << ": Request from IP=" 
                        << parseSourceIP(_soap->ip) << endl);

      // Create a copy of the context that will be handed over to a worker
      struct soap *soapCtxCopy = soap_copy(_soap); // make a safe copy
      if (soapCtxCopy == NULL) {
        throw runtime_error("Could not copy soap environment\n");
      }

      DMESG("[Master] SOAP context copied. Adding to task queue.\n");
      // Put this task on the buffer
      _taskQueue->add(soapCtxCopy);
    }
  }
};

void MDSoapServer::initSSL() {
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  LOG("Initializing SSL support\n");
  // SSL stuff. Initialize multithreading on openSSL
  CRYPTO_thread_setup();

	/**
	* Possible modes:
	* SOAP_SSL_NO_AUTHENTICATION,
	* SOAP_SSL_REQUIRE_CLIENT_AUTHENTICATION :
	* SOAP_SSL_DEFAULT;
	*/
  VOMSInfo::initASN1Parser();

  int sslMode;

  if ( _requireAuthentication ) {
    if ( _allowCertAuth ) {
      sslMode = SOAP_SSL_REQUIRE_CLIENT_AUTHENTICATION;
    } else {
      sslMode = SOAP_SSL_DEFAULT;
    }
    if ( _allowGridProxyLogin ) {
      LOG("Allowing proxy certificates" << endl);
      // During the call to soap_ssl_server_context() this function pointer will
      // be given to OpenSSL as the callback to verify certificates 
      // (SSL_CTX_set_verify())		
      _soap.fsslverify = certVerifyCallbackAllowProxy;		
    }

  } else {
    sslMode = SOAP_SSL_NO_AUTHENTICATION;
  }

  if (soap_ssl_server_context(&_soap,
		sslMode,
		_serverCertKeyFile.c_str(),
		NULL,
		NULL,
                _trustedCertDir.c_str(),
		NULL,
		NULL,
		"MDSoapServer"))
  {
    soap_print_fault(&_soap, stderr);
    throw runtime_error("Cannot initialize SSL server context\n");
  }

  if ( _requireAuthentication && _allowGridProxyLogin ) {
    // The cert store is created only after initializing the OpenSSL context
    // Install a callback to override the default verification of certs. Proxy
    // certificates are signed by end user certificates that in principle are
    // not allowed to sign other certificates. This callback will ignore this
    // validation error.
    _soap.ctx->cert_store->check_issued=checkIssuerCallback;
  }
#endif

}

AuthContext _authCtx;

#define OPT_SOAP_MAX_BUFFER "SoapMaxBuffer"

MDSoapServer::MDSoapServer(int port, ConfigParser &configParser) :
	_taskQueue(20),
	_port(port),
	_dbConnPool(NULL),
	_sessionManager(NULL),
	_soapKeepAlive(DEFAULT_SOAP_KEEP_ALIVE),
	_useSSL(false),
        _requireAuthentication(false), 
        _allowCertAuth(false), 
        _allowGridProxyLogin(false), 
        _allowPasswordAuthentication(false) 
{

  _serverCertKeyFile = "";
  _trustedCertDir = "";

  int nDBConnections = DEFAULT_MAX_CONNECTIONS;
  int nWorkers = DEFAULT_WORKER_THREADS;
  int sessionTimeout = DEFAULT_SESSION_TIMEOUT;
  int queryTimeout = DEFAULT_QUERY_TIMEOUT;
  int indirectServiceLimit = DEFAULT_INDIRECT_SERVICE_LIMIT;
  int maxSessionsPerClient = DEFAULT_MAX_SESSIONS_PER_CLIENT;

  try {
    // int nSoapMaxBuffers = configParser.getInt(OPT_SOAP_MAX_BUFFER);
    // DMESG("MAX SOAP BUFFER : " << nSoapMaxBuffers << endl );
  } catch (string err) {}

  try {
    nDBConnections = configParser.getInt(OPT_MAX_CONNECTIONS);
  } catch (string err) {}

  try {
    nWorkers = configParser.getInt(OPT_WORKER_THREADS);
  } catch (string err) {}

  try {
    _soapKeepAlive = (configParser.getInt(OPT_SOAP_KEEP_ALIVE) != 0);
  } catch (string err) {}

  try {
    sessionTimeout = configParser.getInt(OPT_SESSION_TIMEOUT);
  } catch (string err) {}

  try {
    queryTimeout = configParser.getInt(OPT_QUERY_TIMEOUT);
  } catch (string err) {}

  try {
    indirectServiceLimit = configParser.getInt(OPT_INDIRECT_SERVICE_LIMIT);
  } catch (string err) {}
  
  try {
    DEFAULT_DRAN = configParser.getString(OPT_DATA_RESOURCE_ABSTRACT_NAME);
  } catch (string err) {}

  try {
    maxSessionsPerClient = configParser.getInt(OPT_MAX_SESSIONS_PER_CLIENT);
  } catch (string err) {}

  try {
    _requireAuthentication = configParser.getInt(OPT_REQUIRE_AUTH); 
    if (_requireAuthentication )
      _allowPasswordAuthentication = configParser.getInt(OPT_ALLOW_PASS_AUTH); 
  } catch (string err) {}
      
  try {
    _useSSL = configParser.getInt(OPT_USE_SSL);
    if ( _useSSL ) {
      _allowCertAuth = configParser.getInt(OPT_ALLOW_CERT_AUTH);
      _serverCertKeyFile = configParser.getString(OPT_SERVER_CERT_KEY_FILE);
      _trustedCertDir = configParser.getString(OPT_TRUSTED_CERT_DIR);
      if ( _allowCertAuth && _requireAuthentication ) {
        _allowGridProxyLogin = configParser.getInt(OPT_ALLOW_GRID_PROXY_LOGIN); 
      }
    }
  } catch (string err) {}

  _authCtx.requireAuthentication = _requireAuthentication;
  _authCtx.useSSL = _useSSL ;
  _authCtx.allowGridProxyLogin = _allowGridProxyLogin;
  _authCtx.allowPasswordAuthentication = _allowPasswordAuthentication;
  _authCtx.allowCertAuth = _allowCertAuth;
  _authCtx.indirectServiceLimit = indirectServiceLimit;

  int maxSessions = (nDBConnections - nWorkers/2);

  // Make sure that there is at least one connection available for sessions.
  // The server will not allow more than (MaxConnections - WorkerThreads)
  // sessions to be opened, so that at any time a worker thread always has
  // an availble DBConnection.
  if (nWorkers > nDBConnections) {
    LOG("MaxConnections (" << nDBConnections << ") cannot be smaller than WorkerThreads (" << nWorkers << "). Increasing MaxConnections to " << (nWorkers) << endl);
    nDBConnections = nWorkers;
  }
  LOG("*** CONFIGURATION ****\n");
  LOG("  Listen port          = " << port << endl);
  LOG("  Worker Threads       = " << nWorkers << endl);
  LOG("  MaxConnections       = " << nDBConnections << endl);
  LOG("  MaxSessionsPerClient = " << maxSessionsPerClient << endl);
  LOG("  MaxSessions          = " << maxSessions << endl);
  LOG("  SessionTimeout       = " << sessionTimeout << endl);
  LOG("  QueryTimeout         = " << queryTimeout << endl);
  LOG("  IndirectServiceLimit = " << indirectServiceLimit << endl);
  LOG("  Soap keepalive       = " << _soapKeepAlive << endl);
  LOG("  Use SSL              = " << _useSSL << endl);
  LOG("  Require Auth         = " << _requireAuthentication << endl);
  LOG("  Allow Cert Auth      = " << _allowCertAuth << endl);
  LOG("  Allow Proxy Certs    = " << _allowGridProxyLogin << endl);
  LOG("  Allow Password Auth  = " << _allowPasswordAuthentication << endl);
  LOG("  Server Cert Key File = " << _serverCertKeyFile << endl);
  LOG("  Trusted Cert Dir     = " << _trustedCertDir << endl);

  // Initialize soap and worker threads
  if (_soapKeepAlive) {
    soap_init2(&_soap, SOAP_IO_KEEPALIVE, SOAP_IO_KEEPALIVE);
    LOG("*****Keep alive enabled*****"<<endl);
  } else {
    soap_init(&_soap);
  }

  /* Need SIGPIPE handler on Unix/Linux systems to catch broken pipes: */
  //signal(SIGPIPE, sigpipe_handle);
  signal(SIGPIPE,  SIG_IGN);

  if ( _useSSL )  initSSL();

  // Enable sockets to be reused.
  _soap.bind_flags = SO_REUSEADDR;

  int m = soap_bind(&_soap, NULL, _port, SOAP_SOCKET_BACKLOG);
  if (m < 0) {
    soap_print_fault(&_soap, stderr);
    throw runtime_error("Could not open socket");
  }
  LOG("Server socket bound to port " << port << endl);

  _soap.send_timeout = 60; // 60 seconds
  _soap.recv_timeout = 60; // 60 seconds
  _soap.accept_timeout = 0; // Server never stops accepting.
  _soap.max_keep_alive = 100; // max keep-alive sequence


  _dbConnPool = new DBConnectionPool(&configParser, nDBConnections);
  _sessionManager = new SessionManager(
		sessionTimeout,
		queryTimeout,
		maxSessions,
		maxSessionsPerClient);

  LOG("Creating " << nWorkers << " worker threads\n");
  _workers.resize(nWorkers);
  for(int i=0; i<nWorkers; i++){
    ec(pthread_create( &_workers[i], NULL, execute
      , new WorkerTask(&_taskQueue, _dbConnPool, _sessionManager, _useSSL)));
  }
}

MDSoapServer::~MDSoapServer(){
  DMESG("~MDSoapServer\n");
  DMESG("Killing workers... " << endl);
  for(size_t i=0; i<_workers.size(); i++){
    cancelAndJoinThread(_workers[i], "Worker");
  }

  DMESG("Deleting sessionManager... " << endl);
  delete _sessionManager;

  DMESG("Cleaning SOAP structures.\n");
  soap_destroy(&_soap); 
  soap_end(&_soap); 
  soap_done(&_soap); // close master socket and detach environment
  soap_free(&_soap); 

  if (_useSSL) {
    // Cleanup the multithreading support for openSSL
    CRYPTO_thread_cleanup();
  }

  DMESG("Deleting connection pool\n");
  delete _dbConnPool;
}

void MDSoapServer::stopServing() {
	cancelAndJoinThread(_master, "Master");
}

void MDSoapServer::startServing() {
  LOG("Starting master thread\n");
  ec(pthread_create(
		&_master,
		NULL,
		execute,
		new MasterTask(&_soap, &_taskQueue)));
}

void MDSoapServer::setDebug(bool _debug) {
  debug = _debug;
  DBConnection::setDebug(_debug);
  MDSoapBuffer::setDebug(_debug);
  WSDAIRSetDebug(_debug);
}

bool MDSoapServer::getDebug() {
  return debug;
}


/*****************************************************
        These classes implement WS-DAIR operations
********************************************************/


/************************************************************************
SQLAccess::SQLExecute
  Directs a SQLExpression and optional SQLExecuteRequestParameters to 
  a relational data resource. The SQLExecuteRequestParameters are used for 
  passing values to Stored Procedures and User Defined Functions.
  The SQLExecuteResponse combines statically typed and dynamically 
  typed structures. Static typing is used for the status values and 
  for single value outputs. The tabular output of the rowset
  is provided in a format dynamically determined by the input 
  DataseFormatURI.
************************************************************************/

int __ns1__SQLExecute( struct soap *soap
                     , _wsdair__SQLExecuteRequest* req
                     , _wsdair__SQLExecuteResponse* res )
{                        
  int ret;
  AutoDBConnection helper(soap->user);


  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  ret = checkDefaultDran(soap, req->wsdai__DataResourceAbstractName );
  if (ret != SOAP_OK) return ret ;

  /* Check if the request includes Expression */
  if ( req->wsdair__SQLExpression == NULL ) {
    return generateSoapError(soap, "Invalid Expression Fault" , NULL
            , SOAP_TYPE_wsdai__InvalidExpressionFaultType );
  }

  std::string query = req->wsdair__SQLExpression->Expression;

  trim2(query);
  LOG( "SQL Query >>"  << query << "<<" << endl );
  DMESG( "EndPoint >>"  << soap->endpoint << "<<" << endl );

  if (query.size() <= 0 ) {
    return generateSoapError(soap, "Invalid Expression Fault" , NULL
            , SOAP_TYPE_wsdai__InvalidExpressionFaultType );
  }



  /* initialize necessary data from _wsdair__SQLExecuteRequest*/
  S_SQLAccess sqla;
  sqla.init(soap, & helper
          , req->wsdai__DataResourceAbstractName
          , req->wsdair__SQLExpression->Expression
          , debug );

  /* check dataset format                        */
  /* It should be WebRowset, or it returns Error */
  ret = checkDatasetFormatURI( soap, &(req->wsdai__DatasetFormatURI));
  if (ret != SOAP_OK) return ret ;

  /* execute query */
  ret = sqla.handleQuery();
  if( ret != SOAP_OK ) return ret;

  /* make _wsdair__SQLExecuteResponse */
  sqla.makeResponse(res);

  return SOAP_OK;
}

/************************************************************************
SQLAccess::GetSQLPropertyDocument
  Provides a mechanism for obtaining a copy of the SQLAccessDescription.
************************************************************************/

int __ns1__GetSQLPropertyDocument( struct soap *soap
 	, _wsdai__GetDataResourcePropertyDocumentRequest* req
	,  wsdair__SQLPropertyDocumentType* doc  )
{
  int ret;
  AutoDBConnection helper(soap->user);

  DMESG ("SQLAccess::GetSQLPropertyDocument\n");
  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  ret = checkDefaultDran(soap, req->wsdai__DataResourceAbstractName );
  if (ret != SOAP_OK) return ret ;

  /* initialize necessary data from _wsdai__GetDataResourcePropertyDocumentRequest */
  S_SQLAccess sqla;
  std::string query;
  sqla.init(soap , &helper, req->wsdai__DataResourceAbstractName, query, debug);

  /* make wsdair__SQLPropertyDocumentType response */
  sqla.makePropertyDocument(doc, DEFAULT_DRAN);
  return SOAP_OK;
}

/************************************************************************
SQLAccessFactory::SQLExecuteFactory
  Create a new data access service that corresponds to the results of a SQL Query.
************************************************************************/

int __ns2__SQLExecuteFactory( struct soap *soap,
    _wsdair__SQLExecuteFactoryRequest*  req,
    wsdai__DataResourceAddressListType* res )
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  ret = checkDefaultDran(soap, req->wsdai__DataResourceAbstractName );
  if (ret != SOAP_OK) return ret ;

  /* Check if the request includes Expression */
  if ( req->wsdair__SQLExpression == NULL ) {
    return generateSoapError(soap, "Invalid Expression Fault" , NULL
            , SOAP_TYPE_wsdai__InvalidExpressionFaultType );
  }

  std::string query = req->wsdair__SQLExpression->Expression;

  trim2(query);
  LOG( "SQL Query >>"  << query << "<<" << endl );

  if (query.size() <= 0 ) {
    return generateSoapError(soap, "Invalid Expression Fault" , NULL
            , SOAP_TYPE_wsdai__InvalidExpressionFaultType );
  }

  /* initialize necessary data from _wsdair__SQLExecuteFactoryRequest */
  S_SQLAccessFactory sqlaf;
  sqlaf.init(soap, &helper, 
            req->wsdai__DataResourceAbstractName, query, debug);


  /* process query : indirect data access                    */
  ret = sqlaf.handleQuery();
  if( ret != SOAP_OK ) return ret;

  /* make wsdai__DataResourceAddressListType response */
  sqlaf.makeResponse(res);
  return SOAP_OK;
}


/************************************************************************
SQLResponse::GetSQLResponsePropertyDocument
  Allows a copy of the SQLResponsePropertyDocument document to be retrieved
************************************************************************/


int __ns3__GetSQLResponsePropertyDocument( struct soap *soap,    
     _wsdai__GetDataResourcePropertyDocumentRequest* req,
     wsdair__SQLResponsePropertyDocumentType* res)
{
  int ret;

  AutoDBConnection helper(soap->user);


  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  /* Initialize data from _wsdai__GetDataResourcePropertyDocumentRequest */
  S_SQLResponse ssr;
  ssr.init(soap, &helper, 0, NULL
         , req->wsdai__DataResourceAbstractName, NULL, 0, debug);

  int num_column;
  std::string table;
  char type;

  /* Parse input Data Resource Abstract Name */ 
  ret = ssr.parseDran(table, type, num_column);
  if( ret != SOAP_OK ) return ret;

  bool update = ssr.isUpdate();


  /* make wsdair__SQLResponsePropertyDocumentType response */
  ssr.makePropertyDocument(update, res);

  return SOAP_OK;
}


/************************************************************************
SQLResponse::GetSQLResponseItem
  Return a specified number of Items from a service that represents 
  a SQL Response. This provides an alternative way to access SQL Response 
  Items to the operations (GetSQLRowSet, GetSQLUpdateCount, etc) for 
  SQLResponseAccess. The response type for each Item is
  obtained from the associated SQLResponseItem property.
************************************************************************/

int __ns3__GetSQLResponseItem( struct soap *soap,
    _wsdair__GetSQLResponseItemRequest* req,
    _wsdair__GetSQLResponseItemResponse* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  unsigned int position = req->Position;
  unsigned int* reqCount = req->Count;
  std::string dran = req->wsdai__DataResourceAbstractName;

  if( position != 0 ) {
    return wrongPosition(soap);
  }

  if( reqCount != NULL ) {
    if ( (*reqCount != 0 ) && (*reqCount != 1) ) {
      return wrongCount(soap);
    }
  }


  /* Check If requested dataset format is WebRowset */
  ret = checkDatasetFormatURI( soap, &(req->wsdai__DatasetFormatURI));
  if (ret != SOAP_OK) return ret ;

  res->wsdair__SQLDataset = soap_new_wsdair__SQLDatasetType(soap,-1);

  S_SQLResponse ssr;
  ssr.init( soap, &helper, position, reqCount, 
            dran, res->wsdair__SQLDataset, OP_ITEM, debug);

  /* Handle ResponseItem Request        */
  /* Response is also prepared in the below call  */
  ret = ssr.processITEM();
  if( ret != SOAP_OK ) return ret;
  
  return SOAP_OK;
}

/************************************************************************
SQLResponse::GetSQLRowset
  Get a SQLRowset from the GetSQLRowsetResponse.

 This code is almost same as the above __ns3__GetSQLResponseItem     
 Differences are the last parameter in ssr.init function call        
     , and input parameters in __ns3__GetSQLRowset                    
************************************************************************/

int __ns3__GetSQLRowset(struct soap *soap,
    _wsdair__GetSQLRowsetRequest*       req,
    _wsdair__GetSQLRowsetResponse*      res)
{


  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  unsigned int position = req->Position;
  unsigned int* reqCount = req->Count;
  std::string dran = req->wsdai__DataResourceAbstractName;

  if( position != 0 ) {
    return wrongPosition(soap);
  }

  if( reqCount != NULL ) {
    if ( (*reqCount != 0 ) && (*reqCount != 1) ) {
      return wrongCount(soap);
    }
  }

  ret = checkDatasetFormatURI( soap, & (req->wsdai__DatasetFormatURI) );
  if (ret != SOAP_OK) return ret ;

  res->wsdai__Dataset = soap_new_wsdai__DatasetType(soap,-1);
      
  S_SQLResponse ssr;
  ssr.init(soap, &helper, position, reqCount, 
           dran, res->wsdai__Dataset, OP_ROWSET, debug);
    
  ret = ssr.processITEM();
  if( ret != SOAP_OK ) return ret;
      
  return SOAP_OK;
}

/************************************************************************
SQLResponse::GetSQLUpdateCount
  Get a SQLUpdateCount from the GetSQLUpdateCountResponse.
************************************************************************/

int __ns3__GetSQLUpdateCount(struct soap *soap, 
    _wsdair__GetSQLUpdateCountRequest*  req,
    _wsdair__GetSQLUpdateCountResponse* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

    /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  unsigned int position = req->Position;
  unsigned int* reqCount = req->Count;
  std::string dran = req->wsdai__DataResourceAbstractName;

  if( position != 0 ) {
    return wrongPosition(soap);
  }

  if( reqCount != NULL ) {
    if ( (*reqCount != 0 ) && (*reqCount != 1) ) {
      return wrongCount(soap);
    }
  }

  int num_column;
  std::string table;
  char type;
  
  S_SQLResponse ssr;
  ssr.init(soap, &helper, position, reqCount, dran, NULL, OP_UPDATE, debug);

  /* parse DRAN, and check its validity   */
  ret = ssr.parseDran(table, type, num_column);
  if( ret != SOAP_OK ) return ret;

  /* if the parsed DRAN has type of UPDATE, then it returns error */
  if( type != 'U') return wrongDran(soap, dran);

  /* Returns updated number of rows                                */
  /* The number can be easyly retrieved from the DRAN name         */
  res->UpdateCount.push_back(num_column);

  return SOAP_OK;
}

/************************************************************************
The followings are not implemented because they do not give any merits 
  in AMGA WS-DAIR.
************************************************************************/
int __ns3__GetSQLReturnValue(struct soap *soap,
    _wsdair__GetSQLReturnValueRequest*  wsdair__GetSQLReturnValueRequest,
    _wsdair__GetSQLReturnValueResponse* wsdair__GetSQLReturnValueResponse)
{
  return SOAP_OK;
}

int __ns3__GetSQLOutputParameter(struct soap *soap,
    _wsdair__GetSQLOutputParameterRequest* wsdair__GetSQLOutputParameterRequest,
    _wsdair__GetSQLOutputParameterResponse* wsdair__GetSQLOutputParameterResponse)
{
  return SOAP_OK;
}

int __ns3__GetSQLCommunicationArea(struct soap *soap,
    _wsdair__GetSQLCommunicationAreaRequest* wsdair__GetSQLCommunicationAreaRequest,
    _wsdair__GetSQLCommunicationAreaResponse* wsdair__GetSQLCommunicationAreaResponse)
{
  return SOAP_OK;
}

/************************************************************************
SQLResponseFactory::GetSQLRowsetFactory
  The SQLResponseFactory is used to provide access to a row set as a data resource, 
  by way of a data access service. Again the message exchange patterns specified 
  in the WS-DAI specification are followed. It gets a Reference to a data access 
  service that provides access to a specific rowset.

 This call does not make a new service actually.          
 When the SQLResponse service is created with SQLAccessFactory,
   the SQLRowset service is created at the same time.
 This does not mean that two services are created with SQLAccessFactory.
 It means that SQLResponse service is able to handle SQLRowset requests too.
************************************************************************/

int __ns4__GetSQLRowsetFactory(struct soap *soap,
    _wsdair__GetSQLRowsetFactoryRequest*   req,
    wsdai__DataResourceAddressListType* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  unsigned int position = req->Position;
  unsigned int* reqCount = req->Count;
  std::string dran = req->wsdai__DataResourceAbstractName;

  if( position != 0 ) {
    return wrongPosition(soap);
  }

  if( reqCount != NULL ) {
    if ( (*reqCount != 0 ) && (*reqCount != 1) ) {
      return wrongCount(soap);
    }
  }

  int num_column;
  std::string table;
  char type;

  S_SQLResponseFactory ssrf;
  ssrf.init(soap, &helper, position, reqCount, dran, NULL, OP_UPDATE, debug);

  ret = ssrf.parseDran(table, type, num_column);
  if( ret != SOAP_OK ) return ret;

  /* if the parsed DRAN has type of UPDATE, then it returns error */
  if( type != 'S') return wrongDran(soap, dran);

  /* it just returns input dran to output response */
  ssrf.makeResponse(res);

  return SOAP_OK;
}

/************************************************************************
SQLRowset::GetSQLRowsetPropertyDocument
  Allows a copy of the RowsetPropertyDocument document to be retrieved.
************************************************************************/

int __ns5__GetSQLRowsetPropertyDocument(struct soap *soap,
    _wsdai__GetDataResourcePropertyDocumentRequest* req,
    wsdair__SQLRowsetPropertyDocumentType* res) 
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  int num_column;
  std::string table;
  char type;

  S_SQLRowset ssr;
  ssr.init(soap, &helper, 0, NULL
         , req->wsdai__DataResourceAbstractName, NULL, OP_ROWSET, debug);

  ret = ssr.parseDran(table, type, num_column);
  if( ret != SOAP_OK ) return ret;

  if ( type == 'U') return wrongDran(soap, ssr.dran);

  ssr.makePropertyDocument(num_column,res,table);

  return SOAP_OK;
}



/************************************************************************
SQLRowsetAccess::GetTuples
  Return a specified number of tuples from a service that represents a result set.

 The code is almost same as the above __ns3__GetSQLResponseItem
 Differences are the last parameter in ssr.init function call (OP_TUPLE)
     , and input parameters in __ns5__GetTuples
************************************************************************/

int __ns5__GetTuples( struct soap *soap, 
    _wsdair__GetTuplesRequest* req,
    _wsdair__GetTuplesResponse* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  unsigned int position = req->Position;
  std::string dran = req->wsdai__DataResourceAbstractName;  

/*
  if( position != 0 ) {
    return wrongPosition(soap);
  }
  unsigned int* reqCount = req->Count;
  if( reqCount != NULL ) {
    if ( (*reqCount) < 0 ) {
      return wrongCount(soap);
    }
  }
  DMESG("Count : " << *reqCount << std::endl);
*/


  ret = checkDatasetFormatURI( soap, &(req->wsdai__DatasetFormatURI));
  if (ret != SOAP_OK) return ret ;


  res->wsdai__Dataset = soap_new_wsdai__DatasetType(soap,-1);
 
  S_SQLRowset ssr;
  ssr.init(soap, &helper, position, req->Count
         , dran, res->wsdai__Dataset, OP_TUPLE, debug);

  ret = ssr.processITEM();

  if( ret != SOAP_OK ) return ret;
      
  return SOAP_OK;
}


/************************************************************************
CoreDataAccess::GetDataResourcePropertyDocument
  Returns the core property document values associated with the service 
  implementing this message.
************************************************************************/

int __ns6__GetDataResourcePropertyDocument( struct soap *soap,
    _wsdai__GetDataResourcePropertyDocumentRequest* req,
    wsdai__PropertyDocumentType* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  S_SQLRowset ssr;

  trim2(req->wsdai__DataResourceAbstractName);
  ssr.init(soap, &helper, 0, NULL
         , req->wsdai__DataResourceAbstractName
         , NULL, OP_ROWSET, debug);

  int num_column;
  std::string table;
  char type;

  /* If DRAN is DEFAULT_DRAN, it returns SQLAccess.PropertyDocument */
  if ( ssr.dran.find(DEFAULT_DRAN) != std::string::npos ) {
    DMESG("SQLAccess:GetPropertyDocument >>"  << ssr.dran << "<<\n");
    S_SQLAccess sqla;
    std::string query;
    sqla.init(soap, &helper, req->wsdai__DataResourceAbstractName, query, debug);
    sqla.makePropertyDocument(res, true, DEFAULT_DRAN);
    return SOAP_OK;
  }

  /* Else it returns SQLRowset.PropertyDocument */
  ret = ssr.parseDran(table, type, num_column);
  if( ret != SOAP_OK )  return ret;

  ssr.makePropertyDocument(res);
  return SOAP_OK;
}

/************************************************************************
CoreDataAccess::DestroyDataResource
  Destroy the named data resource; 
************************************************************************/

int __ns6__DestroyDataResource( struct soap *soap,
    _wsdai__DestroyDataResourceRequest* req,
    _wsdai__DestroyDataResourceResponse* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  S_SQLResponse ssr;
  ssr.init(soap, &helper, 0, NULL
         , req->wsdai__DataResourceAbstractName, NULL, OP_ROWSET, debug);

  if ( ssr.dran.find(DEFAULT_DRAN) != std::string::npos ) {
    DMESG("Auth Failed \n" );
    return generateSoapError(soap, "Authentication Failed", NULL
              , SOAP_TYPE_wsdai__NotAuthorizedFaultType );
  }

  int num_column; std::string table; char type;
  ret = ssr.parseDran(table, type, num_column);
  if( ret != SOAP_OK ) { return ret; }

  ret = ssr.DestroyDataResource(table, type);
  if (ret != SOAP_OK) {
    DMESG("Destroy Data Error\n"); 
    return ret ;
  }

  return SOAP_OK;
}



/************************************************************************
CoreDataAccess::GenericQuery
  A general message for passing query documents to a data resource. 

  This function is almost the same as SQLAccess::SQLExecute
************************************************************************/

int __ns6__GenericQuery( struct soap *soap,
    _wsdai__GenericQueryRequest*  req,
    _wsdai__GenericQueryResponse* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  ret = checkDefaultDran(soap, req->wsdai__DataResourceAbstractName );
  if (ret != SOAP_OK) return ret ;


  if ( req->wsdai__GenericExpression == NULL) {
    return generateSoapError(soap, "Invalid Expression Fault" , NULL
            , SOAP_TYPE_wsdai__InvalidExpressionFaultType );
  }

  std::string *lang = req->wsdai__GenericExpression->Language;
  if( lang != NULL ) {
    if (lang->size() != 0) {
      trim2(*lang);
      DMESG("Used Language " << *lang << "\n");
      if( ( lang->compare("http://cern.ch/amga") != 0 ) &&
          ( lang->compare("http://www.sql.org/sql-92") != 0 ) ) {
        return wrongLanguage(soap);
      } 
    }
  }
  
  std::string query = req->wsdai__GenericExpression->Expression ;
  LOG( "SQL Query >>"  << query << "<<" << endl );
  DMESG( "EndPoint >>"  << soap->endpoint << "<<" << endl );

  trim2(query);

  if (query.size() <= 0 ) {
    return generateSoapError(soap, "Invalid Expression Fault" , NULL
            , SOAP_TYPE_wsdai__InvalidExpressionFaultType );
  }



  ret = checkDatasetFormatURI( soap, &(req->wsdai__DatasetFormatURI));
  if (ret != SOAP_OK) return ret ;
  
  S_SQLAccess sqla;
  sqla.init(soap, &helper
          , req->wsdai__DataResourceAbstractName
          , query, debug);
          
  // ret = checkDatasetFormatURI( soap, &(req->wsdai__DatasetFormatURI));
  // if (ret != SOAP_OK) return ret ;
  
  ret = sqla.handleQuery();
  if( ret != SOAP_OK ) return ret;
  
  sqla.makeResponse(res);
  
  
  return SOAP_OK;
}


/************************************************************************
CoreResourceList::GetDataResourceList
  Return a list of abstract names and addresses of data resources 
  that are accessible via the data access service that implements the message.
************************************************************************/

int __ns7__GetResourceList( struct soap *soap,
    _wsdai__GetResourceListRequest* req,
    _wsdai__GetResourceListResponse* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  char file[FILENAME_MAX_SIZE];

  ret = checkDBConnection(soap, &helper);
  if (ret != SOAP_OK ) return ret;

  Statement statement(*(helper.dbConn->dbServer->dbConn), debug);

  /* MetaData for newly created service is stored in /SQLResponse directory */
  std::string query = "select \"directory\" from masterindex where \"directory\" like '/SQLResponse/response%' and \"flags\" != '0'";
  if ( statement.exec(query) ) 
    return servBusy(soap, "DB Error");

  statement.bind(1, file, FILENAME_MAX_SIZE);

  wsdai__DataResourceAddressType *er ;


  while(!statement.fetch()){
    std::string tmp_dir = file ;
    er = makeEndPoint(soap, tmp_dir , "SQLResponse");
    res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
    if ( file[8] == 's' ) { 
      er = makeEndPoint(soap, tmp_dir , "SQLRowset");
      res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
    }
    er = makeEndPoint(soap, tmp_dir , "CoreDataAccess");
    res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
    er = makeEndPoint(soap, tmp_dir , "CoreResourceList");
    res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
  }

  statement.unBind();
  statement.close();


  er = makeEndPoint(soap, DEFAULT_DRAN , "SQLAccess");
  res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
  er = makeEndPoint(soap, DEFAULT_DRAN , "CoreDataAccess");
  res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
  er = makeEndPoint(soap, DEFAULT_DRAN , "CoreResourceList");
  res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);


  return SOAP_OK;
}


/************************************************************************
CoreResourceList::Resolve
  Translate a data resource abstract name into a list of data resource addresses.
************************************************************************/

int __ns7__Resolve( struct soap *soap,
    _wsdai__ResolveRequest* req,
    _wsdai__ResolveResponse* res)
{
  int ret;
  AutoDBConnection helper(soap->user);

  /* Authentication */
  ret = authUser(soap, &helper);
  if (ret != SOAP_OK) return ret ;

  S_SQLResponse ssr;
  ssr.init(soap, &helper, 0, NULL
         , req->wsdai__DataResourceAbstractName, NULL, OP_ROWSET, debug);

  wsdai__DataResourceAddressType *er;

  /* if DRAN is DEFAULT_DRAN name, it return the reference to SQLAccess  */
  /*  else, it return the reference of SQLResponse                         */
  
  if (ssr.dran.find(DEFAULT_DRAN) != std::string::npos ) {
    er = makeEndPoint(soap, DEFAULT_DRAN , "SQLAccess");
    res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
    er = makeEndPoint(soap, DEFAULT_DRAN , "CoreDataAccess");
    res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
    er = makeEndPoint(soap, DEFAULT_DRAN , "CoreResourceList");
    res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
    return SOAP_OK;
  }

  int num_column; std::string table; char type;
  ret = ssr.parseDran(table, type, num_column);
  if( ret != SOAP_OK ) { return ret; }

  er = makeEndPoint(soap, ssr.dran , "SQLResponse");
  res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);

  if ( type == 'S' ) {
    er = makeEndPoint(soap, ssr.dran , "SQLRowset");
    res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);
  }

  er = makeEndPoint(soap, ssr.dran , "CoreDataAccess");
  res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);

  er = makeEndPoint(soap, ssr.dran , "CoreResourceList");
  res->wsdai__DataResourceAddress.push_back((wsdai__DataResourceAddressType*) er);


  return SOAP_OK;
}
