/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "SQLAccess.h"
#include "debug.h"

SQLAccess::SQLAccess(const std::string &host , int port, 
            const std::string &path,
            const std::string &configFile, bool keep_alive, bool dbg)
         : MDSoapClient(host, port, path, configFile, keep_alive, dbg)
{
}

int SQLAccess::GetSQLPropertyDocument(GetSQLPropertyDocumentRequest *req
                                    , GetSQLPropertyDocumentResponse **res)
{
  setUser();
  
  *res = soap_new_wsdair__SQLPropertyDocumentType(&soap,-1);

  int ret = soap_call___ns1__GetSQLPropertyDocument(&soap
                , access_end_point.c_str(), NULL, req, *res);
                     
  if (ret != SOAP_OK ) {
    errorMsg = soap.fault->faultstring;
    return ret;
  }
  return SOAP_OK;
}

int SQLAccess::SQLExecute(SQLExecuteRequest *req
                        , SQLExecuteResponse **res)
{
  setUser();

  *res = soap_new__wsdair__SQLExecuteResponse(&soap, -1);

  int ret =  soap_call___ns1__SQLExecute(&soap
            , access_end_point.c_str(), NULL, req, *res);

  if (ret != SOAP_OK ) {
    DMESG("SOAP Execute Error\n");
    errorMsg = soap.fault->faultstring;
    return ret;
  }
  return SOAP_OK;
}

int SQLAccess::parseType(std::string &command)
{

  size_t pos = command.find_first_not_of(' ',0);
  std::string tmp = command.substr(pos);


  if( tmp.compare(0, 6, "SELECT") == 0 ) 
  { return _SELECT; }

  if ( tmp.compare(0, 6, "UPDATE") == 0 ) 
  { return _UPDATE; }

  if ( tmp.compare(0, 6, "DELETE") == 0 ) 
  { return _DELETE; }

  if ( tmp.compare(0, 6, "INSERT") == 0 ) 
  { return _INSERT; }

  return _NO_OPS;
}


int SQLAccess::FetchWRS(SQLExecuteResponse *seRes
                      , WebRowSet** wrs)
{
  if ( (seRes->wsdair__SQLDataset == NULL) ||
       (seRes->wsdair__SQLDataset->wsdai__DatasetData == NULL) )  {
    errorMsg = "No DatasetData";
    return SOAP_DATA_FAULT;
  }

  *wrs = seRes->wsdair__SQLDataset->wsdai__DatasetData->wrs__webRowSet;

  return SOAP_OK;
}

int SQLAccess::GetUpdateCount( SQLExecuteResponse *seRes
                             , int& updateCount)
{
  if( seRes->wsdair__SQLDataset->wsdair__SQLUpdateCount.size() != 0 ) {
    updateCount = seRes->wsdair__SQLDataset->wsdair__SQLUpdateCount[0];
  } else {
    errorMsg = "No Update Count";
    return SOAP_FAULT;
  }
  return SOAP_OK;
}

int SQLAccess::MakeGetSQLPropertyDocumentRequest(std::string &dran
                     , GetSQLPropertyDocumentRequest **req)
{

  GetSQLPropertyDocumentRequest*  seReq 
     = soap_new__wsdai__GetDataResourcePropertyDocumentRequest(&soap,-1);
  *req = seReq;

  seReq->wsdai__DataResourceAbstractName = dran;

  DMESG("DRAN : " << seReq->wsdai__DataResourceAbstractName << endl);

  return SOAP_OK;
}


int SQLAccess::MakeSQLExecuteRequest(std::string &dran
                     , std::string &uri
                     , std::string &command
                     , SQLExecuteRequest **req) 
{

  SQLExecuteRequest*  seReq = soap_new__wsdair__SQLExecuteRequest(&soap,-1);
  *req = seReq;

  seReq->wsdair__SQLExpression = soap_new_wsdair__SQLExpressionType(&soap, -1) ;
  seReq->wsdair__SQLExpression->Expression = command ;

  seReq->wsdai__DataResourceAbstractName = dran;
  seReq->wsdai__DatasetFormatURI = soap_new_std__string(&soap, -1);

  if ( uri.size() == 0 ) {
    *(seReq->wsdai__DatasetFormatURI) = WRS_FORMAT_URI ;
  } else {
    *(seReq->wsdai__DatasetFormatURI) = uri;
  }

  DMESG("DRAN : " << seReq->wsdai__DataResourceAbstractName << endl);
  DMESG("EXPR : " << seReq->wsdair__SQLExpression->Expression << endl);
  DMESG("URI  : " << *(seReq->wsdai__DatasetFormatURI) << endl);

  return SOAP_OK;
}

int SQLAccess::SimpleExecute(std::string &dran
                     , std::string &command
                     , std::string uri)
{
  SQLExecuteRequest*  seReq ;
  SQLExecuteResponse* seRes ;

  int error = 0 ;
  
  do {
    int ret = MakeSQLExecuteRequest(dran, uri, command, & seReq);
    if (ret != SOAP_OK) { error = 1; break; }

    ret = SQLExecute(seReq, &seRes);
    if( ret != SOAP_OK) { error = 1; break; }

    int type = parseType(command);
    switch ( type ) {
      case _DELETE:
      case _INSERT:
      case _UPDATE:
        int updateCount;
        ret = GetUpdateCount( seRes, updateCount );
        if (ret != SOAP_OK) { 
          error = 1 ; 
          break; 
        }
        // cout << updateCount << "   has been modified" << endl ;
        break;
      case _SELECT:
  
      default:
        WebRowSet *wrs;
        ret =  FetchWRS(seRes, &wrs);
        if (ret != SOAP_OK) { error = 1 ; break; }
        PrintWRS( wrs, type );
        break;
    }
  } while (0) ;

  if ( error == 1 ) {
    cout << "Error : " << errorMsg << endl ;
    return SOAP_FAULT;
  } 

  return SOAP_OK;
}

int SQLAccess::SimplePropertyDocument(std::string &dran)
{ 
  GetSQLPropertyDocumentRequest*  seReq ;
  GetSQLPropertyDocumentResponse* seRes ;

  int error = 0 ;
 
  do { 
    int ret = MakeGetSQLPropertyDocumentRequest(dran, & seReq);
    if (ret != SOAP_OK) { error = 1; break; }
  
    ret = GetSQLPropertyDocument(seReq, &seRes);
    if( ret != SOAP_OK) { error = 1; break; }
    
    PrintPropertyDocument(seRes);
  } while (0) ;

  if ( error == 1 ) {
    cout << "Error : " << errorMsg << endl ;
    return SOAP_FAULT;
  }

  return SOAP_OK;
}


void SQLAccess::PrintDatasetMap(PropertyDocument* res)
{
  int count = res->wsdai__DatasetMap.size();

  cout << "***********************************************\n";
  cout << "                  DatasetMap                   \n";
  cout << "***********************************************\n";

  for(int i=0; i<count; i++){
    cout << "Message Name  : " << res->wsdai__DatasetMap[i]->wsdai__MessageQName << endl ;
    cout << "Dataset Format: " << res->wsdai__DatasetMap[i]->wsdai__DatasetFormatURI << endl;
  }
}

void SQLAccess::PrintDefaultConfiguration(wsdai__ConfigurationDocumentType *cdt)
{
  cout << "**** Default Configuration ******************* \n";
  if ( cdt->wsdai__Readable )
    cout << "Readable   : " << *(cdt->wsdai__Readable) << endl ;
  if ( cdt->wsdai__Writeable )
    cout << "Writeable  : " << *(cdt->wsdai__Writeable) << endl ;
  if ( cdt->wsdai__TransactionInitiation )
    cout << "TransactionInitiation : " << soap__wsdai__TransactionInitiation2s(&soap, *(cdt->wsdai__TransactionInitiation)) << endl ;
  if ( cdt->wsdai__TransactionIsolation )
    cout << "TransactionIsolation  : " << soap__wsdai__TransactionIsolation2s(&soap, *(cdt->wsdai__TransactionIsolation)) << endl ;
  if ( cdt->wsdai__ChildSensitiveToParent )
    cout << "ChildSensitiveToParent: " << soap__wsdai__ChildSensitiveToParent2s(&soap, *(cdt->wsdai__ChildSensitiveToParent)) << endl ;
  if ( cdt->wsdai__ParentSensitiveToChild )
    cout << "ParentSensitiveToChild: " << soap__wsdai__ParentSensitiveToChild2s(&soap, *(cdt->wsdai__ParentSensitiveToChild)) << endl ;
}

void SQLAccess::PrintConfigurationMap(PropertyDocument* res)
{
  int count = res->wsdai__ConfigurationMap.size();

  cout << "***********************************************\n";
  cout << "                  ConfigurationMap             \n";
  cout << "***********************************************\n";

  for(int i=0; i<count; i++){
    cout << "Message Name   : " ;
    cout << res->wsdai__ConfigurationMap[i]->wsdai__MessageQName << endl ;
    cout << "Port Type      : " ;
    cout << res->wsdai__ConfigurationMap[i]->wsdai__PortTypeQName << endl ;
    cout << "Config Document: " ;
    cout << res->wsdai__ConfigurationMap[i]->wsdai__ConfigurationDocumentQName << endl ;
    /* Default Configuration Map */ 
    if( res->wsdai__ConfigurationMap[i]->DefaultConfigurationDocument.wsdai__ConfigurationDocument != NULL ) {
      PrintDefaultConfiguration(res->wsdai__ConfigurationMap[i]->DefaultConfigurationDocument.wsdai__ConfigurationDocument);
    }
  }
}

void SQLAccess::PrintLanguageMap(PropertyDocument* res)
{
  cout << "***********************************************\n";
  cout << "                  Language Map                 \n";
  cout << "***********************************************\n";

  int count = res->wsdai__LanguageMap.size();

  for(int i=0; i<count; i++){
    cout << "Message Name  : " << res->wsdai__LanguageMap[i]->wsdai__MessageQName 
         << endl ;
    cout << "Language Map  : " << res->wsdai__LanguageMap[i]->wsdai__LanguageURI 
         << endl;
  }
}

void SQLAccess::PrintGeneralInfo(PropertyDocument* res)
{
  cout << "***********************************************\n";
  cout << "                  General Info                 \n";
  cout << "***********************************************\n";
  cout << "Data Resource Abstract Name : " ;
  cout << res->wsdai__DataResourceAbstractName << endl ;
  cout << "Data Resource Management    : " ;
  cout << soap__wsdai__DataResourceManagement2s(&soap, res->wsdai__DataResourceManagement) << endl ;
  if ( res->wsdai__DataResourceDescription 
    && res->wsdai__DataResourceDescription->__any.size()) {
    cout << "Data Resource Description   : " ;
    cout << res->wsdai__DataResourceDescription->__any[0] << endl ;
  }
  if ( res->wsdai__Readable )
    cout << "Readable                    : " << res->wsdai__Readable << endl ;
  if ( res->wsdai__Writeable )
    cout << "Writeable                   : " << res->wsdai__Writeable << endl ;
  if ( res->wsdai__ConcurrentAccess )
    cout << "ConcurrentAccess            : " << res->wsdai__ConcurrentAccess << endl ;
  if ( res->wsdai__TransactionInitiation )
    cout << "TransactionInitiation       : " << soap__wsdai__TransactionInitiation2s(&soap, res->wsdai__TransactionInitiation) << endl ;
  if ( res->wsdai__TransactionIsolation )
    cout << "TransactionIsolation        : " << soap__wsdai__TransactionIsolation2s(&soap, res->wsdai__TransactionIsolation ) << endl ;
  if ( res->wsdai__ChildSensitiveToParent )
    cout << "ChildSensitiveToParent      : " << soap__wsdai__ChildSensitiveToParent2s(&soap,res->wsdai__ChildSensitiveToParent) << endl ;
  if ( res->wsdai__ParentSensitiveToChild )
    cout << "ParentSensitiveToChild      : " << soap__wsdai__ParentSensitiveToChild2s(&soap,res->wsdai__ParentSensitiveToChild) << endl ;
}

void SQLAccess::PrintPropertyDocument(PropertyDocument* res)
{
  PrintGeneralInfo(res);
  PrintDatasetMap(res);
  PrintConfigurationMap(res);
  PrintLanguageMap(res);
}


void SQLAccess::PrintWRS(WebRowSet* wrs, int type){
  if ( wrs == NULL ) return ;
  if((type == _SELECT) && (wrs->wrs__metadata != NULL)) {
    // int colNo = wrs->wrs__metadata->union_metadata.__sizecolumn_definition ;
    int colNo = atoi(wrs->wrs__metadata->column_count.c_str());
    DMESG("COLUMN " << colNo << "\n");
    for(int i=0; i< colNo; i++){
      if (wrs->wrs__metadata->column_definition[i].table_name.size() ) {
        cout << wrs->wrs__metadata->column_definition[i].table_name ;
        cout << ".";
      }
      cout << wrs->wrs__metadata->column_definition[i].column_name ;
      cout << endl;
    }
  }
  if( (wrs->wrs__data != NULL) && (wrs->wrs__data->__data_sequence != NULL) 
      && (wrs->wrs__data->__data_sequence->currentRow != NULL) ) {
     int rowsNo = wrs->wrs__data->__data_sequence->__sizecurrentRow;
     int colNo  = wrs->wrs__data->__data_sequence->currentRow->__size_data_currentRow_sequence;
     for(int i=0; i < rowsNo; i++) {
       for(int j=0; j < colNo; j++) {
          xsd__anyType *s = wrs->wrs__data->__data_sequence->currentRow[i].__data_currentRow_sequence[j].wrs__columnValue ;
          std::cout << s->__item << endl;
       }
     } 
  }                         
} 

