/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "SQLRowset.h"
#include "debug.h"

SQLRowset::SQLRowset(const std::string &host , int port, std::string &path,
            const std::string &configFile, bool keep_alive, bool dbg)
         : SQLAccess(host, port, path, configFile, keep_alive, dbg)
{
}

int SQLRowset::GetSQLRowsetPropertyDocument(
        GetSQLRowsetPropertyDocumentRequest *req, 
        GetSQLRowsetPropertyDocumentResponse **res)
{
  setUser();

  *res = soap_new_wsdair__SQLRowsetPropertyDocumentType(&soap,-1);

  int ret = soap_call___ns5__GetSQLRowsetPropertyDocument(&soap
                , access_end_point.c_str(), NULL, req, *res);

  if (ret != SOAP_OK ) {
    errorMsg = soap.fault->faultstring;
    return ret;
  }
  return SOAP_OK;
}

int SQLRowset::MakeGetSQLRowsetPropertyDocumentRequest(std::string &dran
                     , GetSQLRowsetPropertyDocumentRequest **req)
{

  GetSQLRowsetPropertyDocumentRequest*  seReq
     = soap_new__wsdai__GetDataResourcePropertyDocumentRequest(&soap,-1);
  *req = seReq;

  seReq->wsdai__DataResourceAbstractName = dran;

  DMESG("DRAN : " << seReq->wsdai__DataResourceAbstractName << endl);

  return SOAP_OK;
}

int SQLRowset::GetTuples(GetTuplesRequest *req,
                         GetTuplesResponse **res)
{
  setUser();

  *res = soap_new__wsdair__GetTuplesResponse(&soap, -1);

  int ret =  soap_call___ns5__GetTuples(&soap
            , access_end_point.c_str(), NULL, req, *res);

  if (ret != SOAP_OK ) {
    errorMsg = soap.fault->faultstring;
    return ret;
  }
  return SOAP_OK;
}

void SQLRowset::PrintSQLRowsetPropertyDocument(GetSQLRowsetPropertyDocumentResponse* res)
{

  if( res->wsdair__RowSchema == NULL ) return;
  if( res->wsdair__RowSchema->wrs__metadata == NULL ) return;
  cout << "***********************************************\n";
  cout << "               RowSchema Metadata              \n";
  cout << "***********************************************\n";

  cout << "Number of Rows              : " ;
  cout << res->wsdair__NoOfRows << std::endl ; 

  cout << "AccessMode                  : " ;
  cout << res->wsdair__AccessMode << std::endl;

  int colNo = atoi(res->wsdair__RowSchema->wrs__metadata->column_count.c_str());
  cout << "Number of Columns           : " ;
  cout << colNo << std::endl;

  for(int i=0; i< colNo; i++){
    cout << "column " << i+1 << " : " ;
    if (res->wsdair__RowSchema->wrs__metadata->column_definition[i].table_name.size() ) {
      cout << res->wsdair__RowSchema->wrs__metadata->column_definition[i].table_name ;
      cout << ".";
    }
    cout << res->wsdair__RowSchema->wrs__metadata->column_definition[i].column_name ;
    cout << endl;
  }


}


int SQLRowset::SimpleRowsetPropertyDocument(std::string &dran)
{
  GetSQLRowsetPropertyDocumentRequest*  seReq ;
  GetSQLRowsetPropertyDocumentResponse* seRes ;

  int error = 0 ;

  do {
    int ret = MakeGetSQLRowsetPropertyDocumentRequest(dran, & seReq);
    if (ret != SOAP_OK) { error = 1; break; }

    ret = GetSQLRowsetPropertyDocument(seReq, &seRes);
    if( ret != SOAP_OK) { error = 1; break; }

    PrintPropertyDocument(seRes);
    PrintSQLRowsetPropertyDocument(seRes);
  } while (0) ;

  if ( error == 1 ) {
    cout << "Error : " << errorMsg << endl ;
    return SOAP_FAULT;
  }

  return SOAP_OK;
}



int SQLRowset::FetchWRS(GetTuplesResponse *seRes
                      , WebRowSet** wrs)
{
  if ( (seRes->wsdai__Dataset == NULL) ||
       (seRes->wsdai__Dataset->wsdai__DatasetData == NULL) )  {
    errorMsg = "No DatasetData";
    return SOAP_DATA_FAULT;
  }

  *wrs = seRes->wsdai__Dataset->wsdai__DatasetData->wrs__webRowSet;

  return SOAP_OK;
}


int SQLRowset::MakeGetTuplesRequest(std::string &dran
                     , std::string &uri
                     , int position
                     , int count
                     , GetTuplesRequest **req) 
{

  GetTuplesRequest*  seReq = soap_new__wsdair__GetTuplesRequest(&soap,-1);
  *req = seReq;

  seReq->wsdai__DataResourceAbstractName = dran;
  seReq->wsdai__DatasetFormatURI = soap_new_std__string(&soap, -1);

  if ( uri.size() == 0 ) {
    *(seReq->wsdai__DatasetFormatURI) = WRS_FORMAT_URI;
  } else {
    *(seReq->wsdai__DatasetFormatURI) = uri;
  }

  DMESG("DRAN : " << seReq->wsdai__DataResourceAbstractName << endl);
  DMESG("URI  : " << *(seReq->wsdai__DatasetFormatURI) << endl);

  seReq->Position = position;
  seReq->Count = (unsigned int*) soap_new_xsd__unsignedInt(&soap,-1);
  *(seReq->Count) = count;

  return SOAP_OK;
}

int SQLRowset::SimpleExecuteRowset(std::string dran
                     , int position
                     , int count 
                     , std::string uri)
{
  GetTuplesRequest*  seReq ;
  GetTuplesResponse* seRes ;

  int error = 0 ;
  
  do {
    int ret = MakeGetTuplesRequest(dran, uri, position, count, & seReq);
    if (ret != SOAP_OK) { error = 1; break; }

    ret = GetTuples(seReq, &seRes);
    if( ret != SOAP_OK) { error = 1; break; }

    WebRowSet *wrs;
    ret =  FetchWRS(seRes, &wrs);
    if (ret != SOAP_OK) { error = 1 ; break; }

    PrintWRS( wrs, _SELECT);

  } while (0) ;

  if ( error == 1 ) {
    cout << "Error : " << errorMsg << endl ;
    return SOAP_FAULT;
  } 

  return SOAP_OK;
}


