/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "S_SQLAccess.h"

static bool debug = false;
#include "debug.h"

extern std::string DEFAULT_DRAN;


void S_SQLAccess::init(struct soap* soap, AutoDBConnection *helper
          , std::string &dran, std::string &query, bool _debug) {
   this->soap = soap;
   this->dran = dran;
   this->query = query;
   this->helper = helper;
   overFlow = false;
   debug = _debug;
  }


/****************************************************
  handleQuery  
    (1) access DB, process SQL & AMGA query, and retrieve output 
    (2) count the number of rows & columns
*****************************************************/

int S_SQLAccess::handleQuery()
{
  MDInterpreter interpreter;
  cmdToken = interpreter.parseCommand (query);
  
  /* check if the input command is allowed in WS-DAIR interface */
  int ret = checkInvalidCommand(soap, cmdToken, _ACCESS_EXECUTE );
  if (ret != SOAP_OK ) return ret ;
  
  DMESG("Handling Query : << " << query << " \n" );

  /* check if helper maintain a valid DB connection */
  ret = checkDBConnection(soap, helper);
  if (ret != SOAP_OK ) return ret;

  /* execute the requested query */
  helper->dbConn->dbServer->setNextCommand(query);
  interpreter.execLast( *(helper->dbConn->dbServer) );

  DMESG("Execution on : << " << query << " \n" );

  /* retrieve the results of the query */
  helper->dbConn->getResult(dbResult, results);
  if (dbResult.errorCode != 0) {   // error
    return returnErrorDBResult(soap, dbResult);
  }
  DMESG("Successful Execution \n" );

  /* check if there is a overflow at the results                                  */
  /* If the size of results is over certain size,                                 */
  /*     some of results are just skipped                                         */
  /* If there is overflow, it is recommanded to use indirect data access mode     */ 
  /*       and retrieve limited number of rows using SQLRowset::GetTuples         */ 
  /* A user should check the property of WRS returned to see if there is overflow */ 
  /* If the number of max_rows in WRS property is not zero                        */
  /*    then it means there is overflow                                           */

  overFlow = helper->dbConn->isOverFlow();


  /* count the number of rows & columns retrieved */ 
  std::vector<std::string> parts = interpreter.getCommand();

  switch(cmdToken) {
    case MDInterpreter::cSELECT:
    {
      colNum = helper->dbConn->dbServer->asp->numColumn;
      rowsNo = results.size()/colNum;
      break;
    }
    case MDInterpreter::cUPDATE:
    case MDInterpreter::cDELETE:
    case MDInterpreter::cINSERT:
    {
      rowsNo = helper->dbConn->dbServer->getUpdatedRows() ;
      break;
    }
    case MDInterpreter::cGetAttr:
    {
      colNum = parts.size() - 1 ;
      rowsNo = results.size()/colNum;
      break;
    }
    case MDInterpreter::cSelectAttr:
    {
      colNum = parts.size() - 2 ;
      rowsNo = results.size()/colNum;
      break;
    }
    case MDInterpreter::cListAttr:
    case MDInterpreter::cDir:
    {
      colNum = 2 ;
      rowsNo = results.size()/colNum;
      break;
    }
    case MDInterpreter::cCd:
    case MDInterpreter::cSetAttr:
    case MDInterpreter::cClearAttr:
    case MDInterpreter::cRenameAttr:
    case MDInterpreter::cRemoveAttr:
    case MDInterpreter::cAddAttr:
    case MDInterpreter::cCreateDir:
    case MDInterpreter::cAddEntries:
    case MDInterpreter::cRemoveEntries:
    case MDInterpreter::cRm:
    case MDInterpreter::cRmDir:
    case MDInterpreter::cLink:
    case MDInterpreter::cMkDir:
    case MDInterpreter::cAddEntry:
    case MDInterpreter::cInsert:
    case MDInterpreter::cChmod:
    case MDInterpreter::cChown:
    case MDInterpreter::cViewCreate:
    case MDInterpreter::cViewCreateUnion:
    case MDInterpreter::cIndexCreate:
    case MDInterpreter::cIndexRemove:
    case MDInterpreter::cViewRemove:
    case MDInterpreter::cGrpCreate:
    case MDInterpreter::cGrpDelete:
    case MDInterpreter::cGrpAddUser:
    case MDInterpreter::cGrpRemoveUser:
    case MDInterpreter::cAclRemove:
    case MDInterpreter::cAclAdd:
    case MDInterpreter::cUpdateAttr:
    case MDInterpreter::cUpdate:
    case MDInterpreter::cUpdateSingle:
    case MDInterpreter::cCommit:
    case MDInterpreter::cAbort:
    case MDInterpreter::cUserCreate:
    case MDInterpreter::cUserRemove:
    case MDInterpreter::cUserPassChange:
    case MDInterpreter::cCapabilitiesAdd:
    case MDInterpreter::cCapabilitiesRemove:
    case MDInterpreter::cTransaction:
    case MDInterpreter::cImport:
    case MDInterpreter::cRepMount:
    case MDInterpreter::cRepUmount:
    case MDInterpreter::cRepUmountAll:
    case MDInterpreter::cRepSynchronize:
    case MDInterpreter::cRepStartReceive:
    case MDInterpreter::cRepStopReceive:
    case MDInterpreter::cRepAllow:
    case MDInterpreter::cRepDisallow:
    case MDInterpreter::cRepMountUsers:
    case MDInterpreter::cRepUmountUsers:
    case MDInterpreter::cRepDropSubscriber:
    case MDInterpreter::cRepUsersAllow:
    case MDInterpreter::cRepUsersDisallow:
    case MDInterpreter::cSiteAdd:
    case MDInterpreter::cSiteRemove:
    case MDInterpreter::cSiteSetProperties:
    {
      rowsNo = 0;
      break;
    }
    default:
    {
      /*
       case MDInterpreter::cLs
       case MDInterpreter::cPwd
       case MDInterpreter::cFind
       case MDInterpreter::cWhoAmI:
       case MDInterpreter::cUserList:
       case MDInterpreter::cHelp:
       case MDInterpreter::cGrpList:
       case MDInterpreter::cGrpMember:
       case MDInterpreter::cGrpShow:
       case MDInterpreter::cAclShow:
       case MDInterpreter::cCapabilitiesList:
       case MDInterpreter::cBackend
       case MDInterpreter::cRepListMounts:
       case MDInterpreter::cRepShowPermissions:
       case MDInterpreter::cRepListSubscribers:
       case MDInterpreter::cSiteList:
       case MDInterpreter::cSiteDump:
       case MDInterpreter::cSiteGetProperties:
       case MDInterpreter::cSiteListProperties:
      */
      colNum = 1 ;
      rowsNo = results.size();
      break;
    }

  }  // end of switch

  DMESG("Column : << " << colNum << " Rownum : " << rowsNo << " \n" );
  DMESG("Column2 : << " << colNum << " Rownum : " << rowsNo << " \n" );

  return SOAP_OK;
}

int S_SQLAccess::makeResponse( _wsdai__GenericQueryResponse* res )
{
  res->wsdai__Dataset = soap_new_wsdai__DatasetType(soap,-1);
  res->wsdai__Dataset->wsdai__DatasetFormatURI = WRS_FORMAT_URI;
  res->wsdai__Dataset->wsdai__DatasetData
    = soap_new_wsdai__DatasetDataType(soap, -1);
  res->wsdai__Dataset->wsdai__DatasetData->wrs__webRowSet
        = createWRS(soap, colNum, rowsNo, overFlow
                  , dran, query, results, cmdToken);
  return SOAP_OK;
}


/*****************************************************************************

  makeReponse : make SQLExecuteReponse for the output of SQLAccess::SQLExecute 

*****************************************************************************/

int S_SQLAccess::makeResponse( _wsdair__SQLExecuteResponse* res )
{
  DMESG("Making Response" << " \n" );
  res->wsdair__SQLDataset = soap_new_wsdair__SQLDatasetType(soap,-1);
  res->wsdair__SQLDataset->wsdai__DatasetFormatURI = WRS_FORMAT_URI;
  res->wsdair__SQLDataset->wsdai__DatasetData
    = soap_new_wsdai__DatasetDataType(soap, -1);

  switch (cmdToken) {
    case MDInterpreter::cUPDATE:
    case MDInterpreter::cDELETE:
    case MDInterpreter::cINSERT:

    case MDInterpreter::cCd:
    case MDInterpreter::cSetAttr:
    case MDInterpreter::cClearAttr:
    case MDInterpreter::cRenameAttr:
    case MDInterpreter::cRemoveAttr:
    case MDInterpreter::cAddAttr:
    case MDInterpreter::cCreateDir:
    case MDInterpreter::cAddEntries:
    case MDInterpreter::cRemoveEntries:
    case MDInterpreter::cRm:
    case MDInterpreter::cRmDir:
    case MDInterpreter::cLink:
    case MDInterpreter::cMkDir:
    case MDInterpreter::cAddEntry:
    case MDInterpreter::cInsert:
    case MDInterpreter::cChmod:
    case MDInterpreter::cChown:
    case MDInterpreter::cViewCreate:
    case MDInterpreter::cViewCreateUnion:
    case MDInterpreter::cIndexCreate:
    case MDInterpreter::cIndexRemove:
    case MDInterpreter::cViewRemove:
    case MDInterpreter::cGrpCreate:
    case MDInterpreter::cGrpDelete:
    case MDInterpreter::cGrpAddUser:
    case MDInterpreter::cGrpRemoveUser:
    case MDInterpreter::cAclRemove:
    case MDInterpreter::cAclAdd:
    case MDInterpreter::cUpdateAttr:
    case MDInterpreter::cUpdate:
    case MDInterpreter::cUpdateSingle:
    case MDInterpreter::cCommit:
    case MDInterpreter::cAbort:
    case MDInterpreter::cUserCreate:
    case MDInterpreter::cUserRemove:
    case MDInterpreter::cUserPassChange:
    case MDInterpreter::cCapabilitiesAdd:
    case MDInterpreter::cCapabilitiesRemove:
    case MDInterpreter::cTransaction:
    case MDInterpreter::cImport:
    case MDInterpreter::cRepMount:
    case MDInterpreter::cRepUmount:
    case MDInterpreter::cRepUmountAll:
    case MDInterpreter::cRepSynchronize:
    case MDInterpreter::cRepStartReceive:
    case MDInterpreter::cRepStopReceive:
    case MDInterpreter::cRepAllow:
    case MDInterpreter::cRepDisallow:
    case MDInterpreter::cRepMountUsers:
    case MDInterpreter::cRepUmountUsers:
    case MDInterpreter::cRepDropSubscriber:
    case MDInterpreter::cRepUsersAllow:
    case MDInterpreter::cRepUsersDisallow:
    case MDInterpreter::cSiteAdd:
    case MDInterpreter::cSiteRemove:
    case MDInterpreter::cSiteSetProperties:
    {
      /* If the query is to modify something,                */
      /*   the response have the number of modified columns  */
      res->wsdair__SQLDataset->wsdair__SQLUpdateCount.push_back(rowsNo);
      break;
    }
    case MDInterpreter::cSELECT:
    default:
    {
      /* If the query is read-only command,                     */
      /*   the response creates a WRS                            */
      res->wsdair__SQLDataset->wsdai__DatasetData->wrs__webRowSet
        = createWRS(soap, colNum, rowsNo, overFlow
                  , dran, query, results, cmdToken);
      res->wsdair__SQLDataset->wsdair__SQLUpdateCount.push_back(0);
      break;
    }
  }  // end of switch
  DMESG("Making Response Done" << " \n" );
  return SOAP_OK;
}


/*****************************************************************************

  makePropertyDocument : 
    (1) produce PropertyDocument for SQLAccess

*****************************************************************************/
void S_SQLAccess::makePropertyDocument( wsdair__SQLPropertyDocumentType* doc, std::string & dran )
{
  makePropertyDocument( (wsdai__PropertyDocumentType*) doc , true, dran);
  /* schema description */
  doc->wsdair__SchemaDescription = soap_new__wsdair__SchemaDescription(soap,-1);
}

void S_SQLAccess::makePropertyDocument( wsdai__PropertyDocumentType* doc, bool temp, std::string & dran )
{
  std::string endpoint;
  std::string address;
  transEndPoint(soap, address, endpoint, "SQLAccess", DEFAULT_DRAN);

  doc->wsdai__DataResourceAbstractName.append(endpoint);
  doc->wsdai__DataResourceManagement = _wsdai__DataResourceManagement__ExternallyManaged;

  DMESG("Make SQLAccess Property Document \n");

  /* DatasetMap */
  wsdai__DatasetMapType *dmt = soap_new_wsdai__DatasetMapType(soap,-1);
  doc->wsdai__DatasetMap.push_back(dmt);
  dmt->wsdai__MessageQName = "wsdair:SQLExecute" ;
  dmt->wsdai__DatasetFormatURI = WRS_FORMAT_URI;

  
  
  /* Configuration Map */
  
  /*
  wsdai__ConfigurationMapType *cmt = soap_new_wsdai__ConfigurationMapType(soap,-1);
  doc->wsdai__ConfigurationMap.push_back(cmt);
  cmt->wsdai__MessageQName  = "wsdair:SQLExecuteFactory";
  cmt->wsdai__PortTypeQName = "wsdair:SQLResponsePT";
  cmt->wsdai__ConfigurationDocumentQName = "wsdai:ConfigurationDocumentType";

  wsdai__ConfigurationDocumentType *cdt 
       = soap_new_wsdai__ConfigurationDocumentType(soap,-1);
  cmt->DefaultConfigurationDocument.wsdai__ConfigurationDocument = cdt;

  _wsdai__DataResourceDescription *drd
       = soap_new__wsdai__DataResourceDescription(soap,-1);
  drd->__mixed = soap_strdup(soap, "AMGA Metadata Catalog Service");
  cdt->wsdai__DataResourceDescription = drd;

  xsd__boolean* boolTrue = soap_new_xsd__boolean(soap,-1) ;
  boolTrue->__item = true;
  xsd__boolean* boolFalse = soap_new_xsd__boolean(soap,-1) ;
  boolFalse->__item = false;
  
  cdt->wsdai__Readable = (bool*) soap_malloc(soap,1);
  *(cdt->wsdai__Readable) = true;
  cdt->wsdai__Writeable = (bool*) soap_malloc(soap,1);
  *(cdt->wsdai__Writeable) = false;

  cdt->wsdai__TransactionInitiation 
       = (_wsdai__TransactionInitiation *) soap_malloc(soap, sizeof(int));
  *(cdt->wsdai__TransactionInitiation) 
       = _wsdai__TransactionInitiation__Automatic;
  cdt->wsdai__TransactionIsolation
       = (_wsdai__TransactionIsolation *) soap_malloc(soap, sizeof(int)); 
  *(cdt->wsdai__TransactionIsolation) 
       = _wsdai__TransactionIsolation__Serialisable;
  cdt->wsdai__ChildSensitiveToParent
       = (_wsdai__ChildSensitiveToParent *) soap_malloc(soap, sizeof(int)); 
  *(cdt->wsdai__ChildSensitiveToParent) 
       = _wsdai__ChildSensitiveToParent__Sensitive;
  cdt->wsdai__ParentSensitiveToChild
       = (_wsdai__ParentSensitiveToChild *) soap_malloc(soap, sizeof(int)); 
  *(cdt->wsdai__ParentSensitiveToChild) 
       = _wsdai__ParentSensitiveToChild__Sensitive;
  
  */
  

  /* Language Map */
  wsdai__LanguageMapType *lmt = soap_new_wsdai__LanguageMapType(soap,-1);
  lmt->wsdai__MessageQName = "wsdair:SQLExecute";
  lmt->wsdai__LanguageURI = "http://www.sql.org/sql-92";
  doc->wsdai__LanguageMap.push_back(lmt);
  lmt = soap_new_wsdai__LanguageMapType(soap,-1);
  lmt->wsdai__MessageQName = "wsdai:GenericQuery";
  lmt->wsdai__LanguageURI = "http://cern.ch/amga";
  doc->wsdai__LanguageMap.push_back(lmt);
  

  /* Data Resouce Description */
  _wsdai__DataResourceDescription *drd2
       = soap_new__wsdai__DataResourceDescription(soap,-1);
  drd2->__mixed = soap_strdup(soap, "AMGA Metadata Catalog Service");
  doc->wsdai__DataResourceDescription = drd2;

  /* other configuration */
  doc->wsdai__Readable = true;
  doc->wsdai__Writeable = true;
  doc->wsdai__ConcurrentAccess = true;
  doc->wsdai__TransactionInitiation = _wsdai__TransactionInitiation__Automatic;
  doc->wsdai__TransactionIsolation = _wsdai__TransactionIsolation__Serialisable;
  doc->wsdai__ChildSensitiveToParent = _wsdai__ChildSensitiveToParent__Sensitive;
  doc->wsdai__ParentSensitiveToChild = _wsdai__ParentSensitiveToChild__Sensitive;

}

