/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Session.h"

#include <iostream>
#include <cstdio>
#include <assert.h>

#include "server/MDAsyncBuffer.h"

static bool debug = false;
// Activate debug and send it to cout
#define logstream std::cout
#include "debug.h"

using namespace std;

// Static initialization
unsigned Session::_seed = time(NULL);

/**
 * Receives the function that is to be executed on the new thread
 */
Session::Session(DBConnectionPool *dbConnPool, AsynchCommand *cmd, const string& clientIp)
: _busy(false), _dbConn(NULL), _dbConnPool(dbConnPool), _timeStamp((clock_t)0)
{
	DMESG("Preparing to execute asynchronous command" << endl);
	_dbConn = dbConnPool->checkoutConnection();
	assert(_dbConn != NULL);
	_dbConn->startAsynchRequest(cmd);
	_sessionID = generateSessionId();
	_clientIp = clientIp;
}

Session::~Session() {
	DMESG("~Session\n");
}

/**
 * Should be called once after the session is created to get
 * the initial result.
 *
 * @param dbResult
 * @param dataChunk
 */
void Session::getResult(DBResult& dbResult, vector<string> &entries)
throw (BufferCanceledException)
{
  // TODO: Ensure exclusive access
	DMESG("getResult." << endl);
	assert(_dbConn != NULL);
	_dbConn->getResult(dbResult, entries);
	checkForNormalFinish();
}

/**
 * Retrieves the next chunk of results from the DB. Should be called
 * until no more data is available. When that happens, the session is
 * finished automatically.
 *
 * @param dataChunk Where the data is stored.
 */
void Session::getData(vector<string> &entries)
throw (BufferCanceledException) {
	DMESG("Getting data." << endl);
	assert(_dbConn != NULL);
	_dbConn->getData(entries);
	checkForNormalFinish();
}

void Session::toggleBusy() {
	updateTimeStamp();
	_busy = !_busy;
}

void Session::cancel() {
	//assert(_dbConn != NULL);//commented by Ali Javadzadeh 26/2/08
	DMESG("Cancelling a connection.\n");
	if(_dbConn != NULL){//added by  Ali Javadzadeh 26/2/08
	_dbConn->cancelAsynchRequest();
	_dbConnPool->checkinConnection(_dbConn);
	_dbConn = NULL;
	_dbConnPool = NULL;
	}
}

void Session::checkForNormalFinish() {
	
	if(!_dbConn->isAsynchRequestActive()) {
		LOG("Session finished." << endl);
		DMESG("Session finished." << endl);
		_dbConnPool->checkinConnection(_dbConn);
		_dbConn = NULL;
		_dbConnPool = NULL;
	}
}

string& Session::getClientIp() {
	return _clientIp;
}

/*!
	A session ID is a 64 bits integer, consisting of the concatenation of
	a timestamp and random number.
    \fn Session::generateToken()
 */
sessionId_t Session::generateSessionId(){
	ostringstream oss;
	struct tms aux;
	oss << times(&aux) << "-" << rand_r(&_seed);
	return (sessionId_t)oss.str();
}

void Session::updateTimeStamp() {
	struct tms aux;
	_timeStamp = times(&aux);
// 	DMESG("New timestamp: " << _timeStamp << endl);
}
