/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef SESSION_H
#define SESSION_H

#include <string>
#include <sys/times.h>
#include <time.h>
#include <inttypes.h>

#include "DBConnectionPool.h"
#include "DBConnection.h"
#include "DBResult.h"
#include "server/MDServer.h"
// #include "soap/soapStub.h"

typedef std::string sessionId_t;

/**
<p>A session represents a request that may require the client to
make several calls to the server. When a client calls one of these
operations, a cursor is created in the DB and the results are streamed
from the DB to the server, which stores them in a buffer. These results
are returned to the client as it makes sucessive calls to the server.

<p>A session object is used to manage the data structures used to
implement this behavior. This includes a database connection and
a thread used to query the database.

<p>When a worker thread creates a session, the session constructor gets
one connection from the pool of available connections, creates a thread
and executes the request on this newly created thread (DB thread). The
worker thread can then call either getResult or getData to obtain
the data that is being received from the DB. These methods will block
on the MDAsyncBuffer instance used by the DBConnection, until there is
enough information for returning to the client or until the query is
cancelled. At any time, the worker thread can call the cancel() methods
to terminate the request, killing the DB Thread and returning the DB
Connection to the pool.

@author Nuno Santos
*/
class Session {
public:
    /**
     * Gets a connection from the pool and asks it to initiate an asynchronous
	 * request in a new thread using the given command.
     */
    Session(DBConnectionPool *dbConnPool, AsynchCommand *cmd, const std::string& clientIp);

	/** Returns the result from the query. Should be called only once,
	 * right after the session is created. Blocks until some data is
	 * available or until the session is cancelled, in which case it
	 * throws a BufferCancelledException.
	 */
	void getResult(DBResult& dbResult, std::vector<std::string> &entries)
		throw (BufferCanceledException);

	/**
	 * Gets more data. Should be called until no more data is available,
	 * that is, until the isFinished() method returns true. It blocks
	 * until some data is available or until the session is cancelled.
	 */
	void getData(std::vector<std::string> &entries)
		throw (BufferCanceledException);

	/** Terminates the session. Cancels the DB thread, joins with it, and
	 * checks the DB connection back into the DB connection pool. */
	void cancel();
    void closeConnection()
    {
    	//_dbConn->cancelAsynchRequest();
    	_dbConnPool->checkinConnection(_dbConn);
    	_dbConn = NULL;
    }
	/**
	 * If the session is over (cancelled or finished reading from the DB)
	 */
	bool isFinished() const {return _dbConn == NULL;}

	sessionId_t getSessionId() const {return _sessionID;}
	/**
	 * Returns the time when this session last changed state. Whenever
	 * a session changes between IDLE and BUSY a new timestamp is generated.
	 * @return
	 */
	clock_t getTimeStamp() const {return _timeStamp;}
	/** Return the connection being used by this Session */
	DBConnection *getDBConnection() const {return _dbConn;}
	/** If the session is being accessed by a worker thread. */
	bool isBusy() const {return _busy;}

	std::string& getClientIp();
	
private:
	/**The seed used for generating random numbers.*/
	static unsigned _seed;
	friend class SessionManager;
	/**
	 * Only deletes the object, does not terminate the session. This
	 * destructor should be called only by the SessionManager.deleteSession()
	 * method.
	 */
	~Session();

	/** Toggles the busy state and updates the internal timestamp */
	void toggleBusy();

	/** Whether the session is being accessed by a worker thread */
	bool _busy;
	std::string _clientIp;

	sessionId_t _sessionID;

	// The DB connection in use by this session.
	// NULL if the session was canceled
	DBConnection *_dbConn;
	/** Must keep a reference to the pool in order to check in the connection
	 * after using it.
	 */
	DBConnectionPool *_dbConnPool;
	/** Last time this session changed between busy and idle */
	clock_t _timeStamp;

	void checkForNormalFinish();
	void updateTimeStamp();

	static sessionId_t generateSessionId();
};
#endif
