/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <algorithm>
#include <stdexcept>
#include <unistd.h>

#include "SessionManager.h"
#include "util/ThreadUtil.h"

static bool debug = false;
#define logstream std::cout
#include "debug.h"

// In seconds
#define REAPER_THREAD_SLEEP_INTERVAL 2

using namespace std;
using namespace boost;

static const long CLK_SEC = sysconf(_SC_CLK_TCK);

/** Runs periodically to clean expired sessions */
class SessionReaperTask : public Runnable {
	SessionManager *_sessionManager;
public:
	SessionReaperTask(SessionManager *sessionManager)
		: _sessionManager(sessionManager)
	{}
	~SessionReaperTask() {}
	void run() {
		try {
			while(true) {
				// Sleep for a while between checking for expired connections
				sleep(REAPER_THREAD_SLEEP_INTERVAL);
				_sessionManager->reapSessions();
			}
		} catch(...) {
			DMESG("Reaper thread interrupted by Exception\n");
			throw;
		}
	DMESG("Reaper thread dying\n");
	}
};

SessionManager::SessionManager(int sessionTimeout, int queryTimeout, size_t maxSessions, int maxSessionsPerClient)
: _idleSessionTimeout(sessionTimeout), _queryTimeout(queryTimeout) ,_maxSessions(maxSessions), _maxSessionsPerIP(maxSessionsPerClient)
{
	// Start the reaper task
	ec(pthread_create(
			&_reaperThread,
			NULL,
			execute,
			new SessionReaperTask(this)));
}

SessionManager::~SessionManager(){
	DMESG("~SessionManager\n");
	// Try to kill the reaper thread cleanly
	cancelAndJoinThread(_reaperThread, "Reaper");

	DMESG("Deleting sessions\n");
	map<sessionId_t, Session*>::const_iterator iter;
	for(iter=_sessions.begin(); iter!=_sessions.end(); ++iter) {
		delete iter->second;
	}
/*
	map<sessionId_t,_ns__SQLRowsetPropertyDocument*>::const_iterator _iter;
	for(_iter=_sqlRowsetPropertyDocuments.begin(); _iter!=_sqlRowsetPropertyDocuments.end(); ++_iter) {
			delete _iter->second;
		}
*/
}

Session *SessionManager::createSession(DBConnectionPool *dbConnPool, AsynchCommand *cmd, const string& clientIP) {
	boost::mutex::scoped_lock lk(monitor);
	if (_sessions.size() >= _maxSessions) {
		LOG("No more sessions allowed.\n" << endl);
		return NULL;
	}

	// -1 means limits are disabled
	if (_maxSessionsPerIP != -1) {
		// Check if this client is allowed to open more sessions
		// First check if there is an entry for this client on the map
		if (_ipSessionCount.count(clientIP) == 0) {
			// There is none. Create one with the value 1
			_ipSessionCount[clientIP] = 1;
		} else {
			// There is an entry. Therefore, the client has some open sessions.
			// Increment by one the count or fail if the client is already at
			// the maximum allowed limit
			int clientSessions = _ipSessionCount[clientIP];
			if (clientSessions >= _maxSessionsPerIP) {
				LOG("No more sessions allowed from client: " << clientIP << endl);
				return NULL;
			} else {
				_ipSessionCount[clientIP] = clientSessions+1;
			}
		}
	}
// 	DMESG("Number of sessions: " << _ipSessionCount[clientIP] << endl);

	Session *s = new Session(dbConnPool, cmd, clientIP);
	_sessions[s->getSessionId()] = s;
	
	DMESG("Creating session: " << s->getSessionId() << endl);
	// Sessions are created in the not busy state.
	s->toggleBusy();
	return s;

}

void SessionManager::print() {
	map<sessionId_t, Session*>::iterator iter = _sessions.begin();
	while (iter != _sessions.end()) {
		cout << iter->first << " = " << iter->second << endl;
		iter++;
	}
}

void SessionManager::deleteSession(Session *session) throw(logic_error){
	boost::mutex::scoped_lock lk(monitor);
	sessionId_t sId = session->getSessionId();
	DMESG("Deleting session: " << sId << endl);
	string clientIp = session->getClientIp();

	map<sessionId_t, Session*>::iterator it = _sessions.find(sId);
/*
	map<sessionId_t,_ns__SQLRowsetPropertyDocument*>::iterator _iter=_sqlRowsetPropertyDocuments.find(sId);
*/
	if (it == _sessions.end()){
		throw logic_error("Could not find session: " + sId);
	}
	Session* s = it->second;
	if (s == NULL) {
		throw logic_error("Trying to terminate an inexisting session\n");
	} else {
		LOG("------------SessionManager-> deleteSession-----"<<endl);
		// Remove the session from the list of sessions
		_sessions.erase(it);
		// _sqlRowsetPropertyDocuments.erase(_iter);
		// And delete it.
		delete s;
	}
	// -1 means disabled
	if (_maxSessionsPerIP != -1) {
		// Decreases the number of active connections from this client
		assert(_ipSessionCount.count(clientIp) > 0);
		assert(_ipSessionCount[clientIp] > 0);
		_ipSessionCount[clientIp]--;
	// 	DMESG("Client: " << clientIp << ", Sessions: " << _ipSessionCount[clientIp] << endl);
		// Must remove the entry from the map if it reaches 0
		if (_ipSessionCount[clientIp] == 0) {
			_ipSessionCount.erase(clientIp);
		}
	}
}

void SessionManager::reapSessions() {	
	// Keeps track of the sessions that are to be cancelled in the second part
	// of the algorithm
	vector<Session *> expiredSessions;
	{
		//LOG("========SessionManager->reapSessions===="<<endl);
		// This method must lock the object while it is traversing the
		// list of active sessions. During this time, no session can be created,
		// cancelled, checked-in or checked-out. Therefore, this method
		// should release the lock as soon as possible. Cancelling a session
		// might take a while, since it is necessary to cancel the dB thread
		// and to join with it. We avoid this overhead by dividing the algorithm
		// into two phases. First, remove all expired sessions from the
		// list of active sessions. Then cancel the sessions. The internal
		// lock needs to be kept only during the first phase, which should
		// complete very quickly.
		boost::mutex::scoped_lock lk(monitor);
		// Get the current time
		struct tms aux;
		clock_t curTime = times(&aux);
		map<sessionId_t, Session*>::iterator iter = _sessions.begin();
		DMESG("Reaping expired sessions: " << curTime << "\n");
		while (iter != _sessions.end()) {
			Session* session = iter->second;
			// There are two timeout values, one for idle and another for busy sessions.
			int timeout = session->isBusy() ? _queryTimeout : _idleSessionTimeout;
			DMESG("Session: " << session->getSessionId() <<
				", Busy: " << session->isBusy() <<
				", Timeout: " << timeout << endl);
			if (timeout == -1) {
				// Skip this session. -1 means forever.
				iter++;
				continue;
			}
			clock_t expiryTime = timeout*CLK_SEC + session->getTimeStamp();
			DMESG("Expiry: " << expiryTime <<
				", Ticks left: " << (expiryTime - curTime) << endl);
			if (expiryTime < curTime) {
				// Session is expired
				DMESG("Flaging session for cancellation: " << session->getSessionId() << "\n");
				expiredSessions.push_back(session);
				sessionId_t sId=session->getSessionId();
				map<sessionId_t, Session*>::iterator tmp = iter;
				// Must increment the iter before deleting the element, since
				// the standard STL does not guarantee that iterators pointing
				// to an element will be valid after the element is deleted.
				// But it guarantees that iterators pointing to other elements
				// are kept valid.
				iter++;
				_sessions.erase(tmp);
/*
				map<sessionId_t,_ns__SQLRowsetPropertyDocument*>::iterator _iter=_sqlRowsetPropertyDocuments.find(sId);
				_sqlRowsetPropertyDocuments.erase(_iter);
*/
				
			} else {
				iter++;
			}
		}
	}

	// Phase two: Cancel the sessions
	{
		vector<Session *>::iterator iter;
		for(iter = expiredSessions.begin(); iter != expiredSessions.end(); iter++){
			Session *session = *iter;
			LOG("Canceling expired session: " << session->getSessionId()
				<< ". Reason: " <<
				(session->isBusy() ?
					"Maximum query time exceeded" :
					"Maximum client idle time exceeded") << endl);
			// Delete sessions
			 std::string sessionID=session->_sessionID;
			 
			 if (std::remove(sessionID.c_str())) {
			    		 
			    	 }
			session->cancel();
			delete session;
		}
	}
}

Session* SessionManager::checkoutSession(sessionId_t sessionID) {
	boost::mutex::scoped_lock lk(monitor);
	DMESG("[checkoutSession] " << sessionID << endl);
	map<sessionId_t, Session*>::iterator it = _sessions.find(sessionID);
	if (it == _sessions.end()){
		return NULL;
	} else {
		Session* session = it->second;
		assert(!session->isBusy());
		session->toggleBusy();
		return session;
	}
}

void SessionManager::checkinSession(Session &session) {
	boost::mutex::scoped_lock lk(monitor);
	DMESG("[checkinSession] " << session.getSessionId() << endl);
	map<sessionId_t, Session*>::iterator it = _sessions.find(session.getSessionId());
	if (it == _sessions.end()){
		throw logic_error(
			"Trying to checkin an unknown session: " +session.getSessionId() + "\n");
	} else {
		assert(session.isBusy());
		session.toggleBusy();
	}
}
