/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef SESSIONMANAGER_H
#define SESSIONMANAGER_H

#include <map>
#include <boost/thread/condition.hpp>

#include "DBConnectionPool.h"
#include "DBConnection.h"
#include "Session.h"

class SessionManager {
public:
	/**
	*/
  SessionManager(int busySessionTimeout,
		int queryTimeout,
		size_t maxSessions,
		int maxSessionsPerClient);
  ~SessionManager();

  /**
   * Creates a new session. The session is stored internally,
   * so that it can be retrieved later with getSession().
   * A session can be removed in two ways:
   * - By calling deleteSession().
   * - By expiring, if it is left unused for more than sessionTimeout seconds
   */
  Session *createSession(DBConnectionPool *dbConnPool, AsynchCommand *cmd, const std::string& sourceIP);
  /**
   * Destroys a session. It is removed from the internal tables
   * and deleted. The object pointed by session is no longer valid after
   * this method.
   */
  void deleteSession(Session *session) throw(std::logic_error);
  /**
   * Gets a session given its id. Returns null if the id is not found.
   */
  Session* checkoutSession(sessionId_t sessionID);
  /**
   * After using a session, worker threads must check them back in.
   */
  void checkinSession(Session &session);

private:
	friend class SessionReaperTask;
	void reapSessions();
	void print();

	/** How long to keep an unused session alive. If a session exceeds this limit it is cancelled. In seconds. */
	int _idleSessionTimeout;
	/** How long a query can take. Any query exceeding this limit is cancelled.*/
	int _queryTimeout;

	/** How many simultaneous sessions are allowed to exist. */
	size_t _maxSessions;
	/** How many simultaneous sessions are allowed per client. */
	int _maxSessionsPerIP;

	pthread_t _reaperThread;
	boost::mutex monitor;
	
	std::map<sessionId_t, Session*> _sessions;
	
	/** How many sessions are open from a single IP */
	std::map<std::string, int> _ipSessionCount;
};

#endif // SESSIONMANAGER_H
