/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include <iostream>
#include <fstream>
#include "WSDAIRUtil.h"
#include "server/UserContext.h"
#include "util/SecurityUtils.h"
#include "server/protocolHandling.h"

using namespace std;

static bool debug = false;

class AuthContext;
extern AuthContext _authCtx;

extern std::string DEFAULT_DRAN;


std::string *soap_stlstrdup(struct soap *soap, const string &str) {
  string *aux = soap_new_std__string(soap, -1);
  (*aux) = str;
  return aux;
}

void WSDAIRSetDebug(bool _debug)
{
  debug = _debug ;
}

void trim2(string& str)
{
  string::size_type pos = str.find_last_not_of(' ');
  if(pos != string::npos) {
    str.erase(pos + 1);
    pos = str.find_first_not_of(' ');
    if(pos != string::npos) str.erase(0, pos);
  }
  else str.erase(str.begin(), str.end());
}

int wrongDran(struct soap *soap, std::string &dran)
{ 
    DMESG("Wrong DataResourceAbstractName : " << dran <<  endl);
    return generateSoapError(soap, "InvalidResourceNameFault", NULL
              , SOAP_TYPE_wsdai__InvalidResourceNameFaultType);
} 

int wrongPosition(struct soap *soap)
{
    DMESG("Wrong Position " <<  endl);
    return generateSoapError(soap, "InvalidPositionFault", NULL
              , SOAP_TYPE_wsdair__InvalidPositionFaultType);
}

int wrongCount(struct soap *soap)
{
    DMESG("Wrong Count " <<  endl);
    return generateSoapError(soap, "InvalidCountFault", NULL
              , SOAP_TYPE_wsdair__InvalidCountFaultType);
}

int wrongLanguage(struct soap *soap)
{
    DMESG("Wrong Language " <<  endl);
    return generateSoapError(soap, "InvalidLanguageFault", NULL
              , SOAP_TYPE_wsdai__InvalidLanguageFaultType);
}


int servBusy(struct soap *soap, std::string message)
{ 
    return generateSoapError(soap, message.c_str(), NULL
              , SOAP_TYPE_wsdai__ServiceBusyFaultType );
} 


int checkDBConnection(struct soap *soap, AutoDBConnection *helper)
{
  if (helper->dbConn == NULL) {
    DMESG("No DB Connection available \n" );
    return servBusy(soap, "No DB Connection available");
  }
  return SOAP_OK;
}

int generateSoapError(struct soap *soap
                      , const char* faultstring 
                      , const char* detail
                      , int type)
{     
  soap_fault(soap);
  soap_sender_fault(soap, faultstring, NULL);
  soap_faultdetail(soap);

  struct SOAP_ENV__Detail *seDetail ;
  if( soap->version == 2) {
    seDetail = soap->fault->SOAP_ENV__Detail ; 
  } else {
    seDetail = soap->fault->detail ;
  } 
    
  switch(type) {
    case SOAP_TYPE_wsdai__InvalidResourceNameFaultType : 
      seDetail->wsdai__InvalidResourceNameFault
         = soap_new_wsdai__InvalidResourceNameFaultType(soap,-1);
      break;  
    case SOAP_TYPE_wsdai__DataResourceUnavailableFaultType  :
      seDetail->wsdai__DataResourceUnavailableFault
         = soap_new_wsdai__DataResourceUnavailableFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdai__InvalidConfigurationDocumentFaultType  :
      seDetail->wsdai__InvalidConfigurationDocumentFault 
         = soap_new_wsdai__InvalidConfigurationDocumentFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdai__InvalidDatasetFormatFaultType  :
      seDetail->wsdai__InvalidDatasetFormatFault
         = soap_new_wsdai__InvalidDatasetFormatFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdai__InvalidLanguageFaultType  :
      seDetail->wsdai__InvalidLanguageFault
         = soap_new_wsdai__InvalidLanguageFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdai__InvalidExpressionFaultType :
      seDetail->wsdai__InvalidExpressionFault
         = soap_new_wsdai__InvalidExpressionFaultType(soap, -1);
      break;               
    case SOAP_TYPE_wsdai__InvalidPortTypeQNameFaultType  :
      seDetail->wsdai__InvalidPortTypeQNameFault
         = soap_new_wsdai__InvalidPortTypeQNameFaultType(soap,-1);
      break;               
    case SOAP_TYPE_wsdair__InvalidSQLExpressionParameterFaultType  :
      seDetail->wsdair__InvalidSQLExpressionParameterFault
         = soap_new_wsdair__InvalidSQLExpressionParameterFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdai__NotAuthorizedFaultType  :
      seDetail->wsdai__NotAuthorizedFault
         = soap_new_wsdai__NotAuthorizedFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdai__ServiceBusyFaultType  :
      seDetail->wsdai__ServiceBusyFault
         = soap_new_wsdai__ServiceBusyFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdair__InvalidPositionFaultType  :
      seDetail->wsdair__InvalidPositionFault
         = soap_new_wsdair__InvalidPositionFaultType(soap,-1);
      break;
    case SOAP_TYPE_wsdair__InvalidCountFaultType  :
      seDetail->wsdair__InvalidCountFault
         = soap_new_wsdair__InvalidCountFaultType(soap,-1);
      break;

  }
  return SOAP_FAULT;
}


int checkDefaultDran(struct soap *soap, std::string& dran) {
  DMESG("Default DRAN : >>" << DEFAULT_DRAN  << "<<\n");
  DMESG("DRAN : >>" << dran << "<<\n");
  if (dran.find( DEFAULT_DRAN ) != std::string::npos ) {
    DMESG("OK: >>" << dran << "<<\n");
    return SOAP_OK;
  } else {
    DMESG("Wrong DRAN: >>" << dran << "<<\n");
    return wrongDran(soap, dran);
  }
}


int checkDatasetFormatURI(struct soap *soap, std::string **dfURI) {
  // DMESG("My URI : " << WRS_FORMAT_URI << endl);

  if( (*dfURI) == NULL) {
    *dfURI = soap_new_std__string(soap, -1);
    **dfURI = WRS_FORMAT_URI ;
    return SOAP_OK;
  }
  trim2(**dfURI);
  if( (*dfURI)->compare( WRS_FORMAT_URI ) != 0 ) {  
       return generateSoapError(soap, "InvalidDatasetFormatFault", NULL
              , SOAP_TYPE_wsdai__InvalidDatasetFormatFaultType    );
  }

  DMESG("requested URI : " << **dfURI << endl);
  return SOAP_OK;
}

_wrs__webRowSet* createWRS(struct soap *soap,
                           int colNum, 
                           int rowsNo,
                           bool overFlow,
                           std::string& dran,
                           std::string& cmd,
                           vector<string> &results,
                           MDInterpreter::CommandToken cmdToken)
{
  _wrs__webRowSet* wrs  = soap_new__wrs__webRowSet(soap,-1);
  vector<string>::const_iterator it = results.begin();

  if ( cmdToken == MDInterpreter::cSELECT ) {
    rowsNo -= 1;
  }
  
  wrs->wrs__properties = soap_new__wrs__properties(soap,-1);
  wrs->wrs__properties->concurrency = "1";
  wrs->wrs__properties->datasource = dran;
  wrs->wrs__properties->command = cmd;
  wrs->wrs__properties->escape_processing = "false";
  wrs->wrs__properties->fetch_direction = "0";
  wrs->wrs__properties->fetch_size = StringOf(rowsNo);
  wrs->wrs__properties->isolation_level = "0";
    wrs->wrs__properties->max_field_size = "0";
  if ( overFlow ) {
    wrs->wrs__properties->max_rows = StringOf(rowsNo);
  } else {
    wrs->wrs__properties->max_rows = "0";
  }
  wrs->wrs__properties->query_timeout = "0";
  wrs->wrs__properties->read_only = "true";
  wrs->wrs__properties->rowset_type = "TRANSACTION_READ_UNCOMMITED";
  wrs->wrs__properties->show_deleted = "false";
  wrs->wrs__properties->url = soap->endpoint ;
  
  wrs->wrs__metadata     = soap_new__wrs__metadata(soap,-1);
  wrs->wrs__metadata->column_count  = StringOf(colNum) ;

  wrs->wrs__metadata->__sizecolumn_definition = colNum;
  wrs->wrs__metadata->column_definition
    = soap_new__wrs__metadata_column_definition(soap ,colNum);
    
  if ( cmdToken == MDInterpreter::cSELECT ) {
    for(int i=0;i<colNum;i++) {
      wrs->wrs__metadata->column_definition[i].column_index = StringOf(i+1);
      wrs->wrs__metadata->column_definition[i].auto_increment = "false";
      wrs->wrs__metadata->column_definition[i].case_sensitive = "true";
      wrs->wrs__metadata->column_definition[i].currency = "false";
      wrs->wrs__metadata->column_definition[i].nullable = "1";
      wrs->wrs__metadata->column_definition[i].signed_ = "true";
      wrs->wrs__metadata->column_definition[i].searchable = "true";

      std::string tmp = *it;
      size_t pos = tmp.find('.');
      if ( pos == string::npos ) {
        wrs->wrs__metadata->column_definition[i].column_label = tmp ;
        wrs->wrs__metadata->column_definition[i].column_name = tmp ;
      } else {
        wrs->wrs__metadata->column_definition[i].column_label
          = wrs->wrs__metadata->column_definition[i].column_name
          = tmp.substr(pos+1);
        wrs->wrs__metadata->column_definition[i].table_name
          = tmp.substr(0,pos);
      }
      it++;
    }
  }

  wrs->wrs__data = soap_new__wrs__data(soap,-1);
  wrs->wrs__data->__size_data_sequence = 1;
  wrs->wrs__data->__data_sequence = soap_new___wrs__data_sequence(soap,-1);
  wrs->wrs__data->__data_sequence->__sizecurrentRow  = rowsNo;
  wrs->wrs__data->__data_sequence->currentRow 
    = soap_new__wrs__data_currentRow(soap,rowsNo);
  for(int i=0;i<rowsNo;i++) {
    wrs->wrs__data->__data_sequence->currentRow[i].__size_data_currentRow_sequence = colNum;
    wrs->wrs__data->__data_sequence->currentRow[i].__data_currentRow_sequence 
        = soap_new___wrs__data_currentRow_sequence(soap, colNum);
    for(int j=0;j<colNum;j++) {
      xsd__anyType *s = soap_new_xsd__anyType(soap, -1);
      s->__item = soap_strdup(soap, (*it).c_str());
      wrs->wrs__data->__data_sequence->currentRow[i].__data_currentRow_sequence[j].wrs__columnValue = (xsd__anyType*) s ;
      it++;
    }
  }
  return wrs;
}

int checkInvalidCommand(struct soap *soap
                        , MDInterpreter::CommandToken &cmdToken, int type)
{
  switch(cmdToken) {
    case MDInterpreter::cSELECT:
    case MDInterpreter::cUPDATE:
    case MDInterpreter::cDELETE:
    case MDInterpreter::cINSERT:
       break;
    case MDInterpreter::cLS:
    case MDInterpreter::cCd:
    case MDInterpreter::cPwd:
    case MDInterpreter::cWhoAmI:
    case MDInterpreter::cUserList:
    case MDInterpreter::cGetAttr:
    case MDInterpreter::cListAttr:
    case MDInterpreter::cSelectAttr:
    case MDInterpreter::cFind:
    case MDInterpreter::cDir:
    case MDInterpreter::cHelp:
    case MDInterpreter::cGrpList:
    case MDInterpreter::cGrpMember:
    case MDInterpreter::cGrpShow:
    case MDInterpreter::cAclShow:
    case MDInterpreter::cCapabilitiesList:
    case MDInterpreter::cBackend:
    case MDInterpreter::cSetAttr:
    case MDInterpreter::cClearAttr:
    case MDInterpreter::cRenameAttr:
    case MDInterpreter::cRemoveAttr:
    case MDInterpreter::cAddAttr:
    case MDInterpreter::cCreateDir:
    case MDInterpreter::cAddEntries:
    case MDInterpreter::cRemoveEntries:
    case MDInterpreter::cRm:
    case MDInterpreter::cRmDir:
    case MDInterpreter::cLink:
    case MDInterpreter::cMkDir:
    case MDInterpreter::cAddEntry:
    case MDInterpreter::cInsert:
    case MDInterpreter::cChmod:
    case MDInterpreter::cChown:
    case MDInterpreter::cViewCreate:
    case MDInterpreter::cViewCreateUnion:
    case MDInterpreter::cIndexCreate:
    case MDInterpreter::cIndexRemove:
    case MDInterpreter::cViewRemove:
    case MDInterpreter::cGrpCreate:
    case MDInterpreter::cGrpDelete:
    case MDInterpreter::cGrpAddUser:
    case MDInterpreter::cGrpRemoveUser:
    case MDInterpreter::cAclRemove:
    case MDInterpreter::cAclAdd:
    case MDInterpreter::cUpdateAttr:
    case MDInterpreter::cUpdateSingle:
    case MDInterpreter::cUpdate:
    case MDInterpreter::cCommit:
    case MDInterpreter::cAbort:
    case MDInterpreter::cUserCreate:
    case MDInterpreter::cUserRemove:
    case MDInterpreter::cUserPassChange:
    case MDInterpreter::cCapabilitiesAdd:
    case MDInterpreter::cCapabilitiesRemove:
    case MDInterpreter::cTransaction:
    case MDInterpreter::cImport:
    case MDInterpreter::cRepListMounts:
    case MDInterpreter::cRepMount:
    case MDInterpreter::cRepUmount:
    case MDInterpreter::cRepUmountAll:
    case MDInterpreter::cRepSynchronize:
    case MDInterpreter::cRepStartReceive:
    case MDInterpreter::cRepStopReceive:
    case MDInterpreter::cRepAllow:
    case MDInterpreter::cRepDisallow:
    case MDInterpreter::cRepShowPermissions:
    case MDInterpreter::cRepMountUsers:
    case MDInterpreter::cRepUmountUsers:
    case MDInterpreter::cRepListSubscribers:
    case MDInterpreter::cRepDropSubscriber:
    case MDInterpreter::cRepUsersAllow:
    case MDInterpreter::cRepUsersDisallow:
    case MDInterpreter::cSiteAdd:
    case MDInterpreter::cSiteRemove:
    case MDInterpreter::cSiteList:
    case MDInterpreter::cSiteDump:
    case MDInterpreter::cSiteSetProperties:
    case MDInterpreter::cSiteGetProperties:
    case MDInterpreter::cSiteListProperties:
    {
      if ( type == _ACCESS_EXECUTE ) break;
    }
    default:
    {
      DMESG("Invalid Command " << endl);
      return generateSoapError(soap, "Invalide Command" , NULL
              , SOAP_TYPE_wsdai__InvalidLanguageFaultType );
    }
  }
  return SOAP_OK;
}

int returnErrorDBResult(struct soap* soap, DBResult &dbResult )
{
  if (dbResult.errorCode == 4 ) {          // permission denied
    DMESG("Permission Denied : " << dbResult.serverResult << endl);
    return generateSoapError(soap, dbResult.serverResult.c_str() , NULL
              , SOAP_TYPE_wsdai__NotAuthorizedFaultType );
  } else {   // Failed. Abort.
    DMESG("ERROR : " << dbResult.serverResult << endl);
    return generateSoapError(soap, dbResult.serverResult.c_str() , NULL
              , SOAP_TYPE_wsdai__InvalidExpressionFaultType );
  }
}

int authUser(struct soap *soap, AutoDBConnection *helper)
{
  UserContext userCtx;

  if ( soap->userid != NULL ) {
    LOG( "USER   >>"  << soap->userid << "<<" << endl );
    userCtx.login = soap->userid;
  } else {
    userCtx.login = "NULL";
  }

  if ( soap->passwd != NULL ) {
    userCtx.password = soap->passwd;
    // DMESG( "PASSWD >>"  << soap->passwd << "<<" << endl );
  }

  if ( _authCtx.requireAuthentication ) {

    int ret = checkDBConnection(soap, helper);
    if (ret != SOAP_OK ) return ret;

    bool userAuthenticated = authenticateUser(soap->ssl
                      , helper->dbConn->dbServer, _authCtx, userCtx);

    if (!userAuthenticated) {
      DMESG("Auth Failed \n" );
      return generateSoapError(soap, "Authentication Failed", NULL
              , SOAP_TYPE_wsdai__NotAuthorizedFaultType );
    }
    DMESG("Auth Successful : " << userCtx.login << "\n" );
  }

  helper->dbConn->dbServer->setUser(userCtx.login);
  helper->dbConn->dbServer->setGroupRightsMask("r-x");
  helper->dbConn->dbServer->setLogging(true);
  helper->dbConn->dbServer->setDebug(debug);

  return SOAP_OK;
}


void transEndPoint(struct soap* soap, 
                  std::string& address, 
                  std::string& endpoint, 
                  std::string svcType,
                  std::string dran) 
{
  int error = 0 ;
  size_t pos = 0;
  size_t pos2 ;

  std::string tmp_prefix = DIR_PREFIX ;
  std::string tmp_endpoint = soap->endpoint;

  for( int i=0; i<3; i++ ) {
    pos2 = tmp_endpoint.find('/', pos);
    if( pos == string::npos ) { error = 1; break; }
    pos = pos2 + 1;
  }

  if (error == 0) {
    endpoint = tmp_endpoint.substr(0, pos);
    address = endpoint;
    address.append(svcType);
    std::string tmp_table  ;
    if ( dran.compare(DEFAULT_DRAN) == 0 ) {
      endpoint.append("SQLAccess/");
      tmp_table  = DEFAULT_DRAN ;
    } else {
      endpoint.append("SQLResponse/");
      tmp_table  = dran.substr(tmp_prefix.size());
    }
    endpoint.append(tmp_table);
  } else {
    endpoint.append(soap->endpoint);
    address = endpoint ;
    endpoint.append("/");
    endpoint.append(DEFAULT_DRAN);
  }
}


void dran_serialize( wsdai__DataResourceAbstractNameType* dran, 
                    std::string & rp_dran)
{
  struct soap soap ;
  ostringstream *out = new ostringstream();
	
  soap.os = out;
  soap_omode(&soap, SOAP_ENC_XML ); 

  soap_serialize_wsdai__DataResourceAbstractNameType(&soap, dran);
  soap_put_wsdai__DataResourceAbstractNameType(&soap, dran,  
                      "wsdai:DataResourceAbstractNameType", "");

  soap_destroy(&soap);

  rp_dran = out->str();
}

wsdai__DataResourceAddressType* makeEndPoint(struct soap* soap
    , std::string dran, std::string svcType)
{
  wsdai__DataResourceAddressType *er = soap_new_wsdai__DataResourceAddressType(soap,-1);

  std::string endpoint;
  std::string address;
  transEndPoint(soap, address,endpoint, svcType, dran);

  er->Address = soap_new_wsa__AttributedURIType(soap,-1);
  er->Address->__item = soap_strdup(soap, address.c_str());

  er->wsa__ReferenceParameters = soap_new_wsa__ReferenceParametersType(soap,-1);

  std::string rp_dran = "<wsdai:DataResourceAbstractName>";
  rp_dran.append(endpoint);
  rp_dran.append("</wsdai:DataResourceAbstractName>");
  char *rp_dran_char = soap_strdup(soap, rp_dran.c_str());

/*
  wsdai__DataResourceAbstractNameType* wsdai_dran = soap_new_wsdai__DataResourceAbstractNameType(soap,-1);

  std::string rp_dran;
  int ret = dran_serialize(wsdai_dran, rp_dran);
  char *rp_dran_char = soap_strdup(soap, rp_dran.c_str());
*/
  
  er->wsa__ReferenceParameters->__any.push_back(rp_dran_char);

  return er;
}

