/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef WSDAIRUTIL_H
#define WSDAIRUTIL_H

#include "WSDairH.h"
#include "server/MDInterpreter.h"

#include "DBConnection.h"
#include "DBConnectionPool.h"
#include "SessionManager.h"

#define _ACCESS_EXECUTE 1
#define _ACCESS_FACTORY 2

#define URN_PREFIX "urn:dais:"
#define DIR_PREFIX "/SQLResponse/"
#define DIR_PREFIX2 "/SQLRowset/"
#define WRS_FORMAT_URI "http://java.sun.com/xml/ns/jdbc"



/**
 * For single shot operations, it checks out a DBConnection
 * when created and checks it in when destroyed. Should be
 * created on the stack, so the destructor is executed when
 * the function call terminates.
 */


struct ConnectionManager {
  /** The pool with the active network connections */
  DBConnectionPool* dbConnPool;
  /** A map between session IDs and DBConnections */
  SessionManager* sessionManager;
  std::string cwd;
};


class AutoDBConnection {
public:
  // Might be NULL if there was no connection available
  DBConnection *dbConn;
  DBConnectionPool *_dbConnPool;
  ConnectionManager* cMan ;

  AutoDBConnection(void *ptrConnManager) {
    cMan = static_cast<ConnectionManager*>(ptrConnManager);
    _dbConnPool = cMan->dbConnPool;

    tryConnect();
  }

  void tryConnect() {
    dbConn = _dbConnPool->checkoutConnection();
    if ( dbConn == NULL ) return ;

    dbConn->startRequest();
    if( cMan->cwd.size() != 0 ) {
      dbConn->dbServer->setDir(cMan->cwd);
    } else {
      dbConn->dbServer->setDir("/");
    }
  }

  ~AutoDBConnection() {
    if (dbConn != NULL) {
      cMan->cwd = dbConn->dbServer->getDir();
      dbConn->finishRequest();
      _dbConnPool->checkinConnection(dbConn);
    }
  }
};


extern "C" {
  extern std::string* soap_stlstrdup(struct soap *soap, const std::string &str) ;
  extern int generateSoapError(struct soap *soap
                      , const char* faultstring
                      , const char* detail
                      , int type);
  extern void transEndPoint(struct soap* soap, std::string& address, std::string& endpoint, std::string svcType, std::string dran);

  extern int checkDatasetFormatURI(struct soap *soap, std::string **dfURI) ;
  extern int checkDefaultDran(struct soap *soap, std::string& dran) ;

  extern _wrs__webRowSet* createWRS(struct soap *soap,
                           int colNum,
                           int rowsNo,
                           bool overFlow,
                           std::string &dran,
                           std::string &cmd,
                           std::vector<std::string> &results,
                           MDInterpreter::CommandToken cmdToken);
  extern int wrongDran(struct soap *soap, std::string &dran);
  extern int wrongPosition(struct soap *soap);
  extern int wrongCount(struct soap *soap);
  extern int wrongLanguage(struct soap *soap);
  extern int servBusy(struct soap *soap, std::string mess);
  extern int checkInvalidCommand(struct soap *soap
                        , MDInterpreter::CommandToken &cmdToken, int type);
  extern int returnErrorDBResult(struct soap* soap, DBResult &dbResult );
  extern int authUser(struct soap *soap, AutoDBConnection *helper) ;
  extern wsdai__DataResourceAddressType* makeEndPoint(struct soap* soap
    , std::string dran, std::string svcType);
  extern void trim2(std::string &str);
  extern int checkDBConnection(struct soap *soap, AutoDBConnection *helper);
  extern void WSDAIRSetDebug(bool _debug);

}

template <class T> std::string StringOf(T object)
{
  std::ostringstream os;
  os << object;
  return(os.str());
}





#endif
