/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef _BOUNDEDQUEUE_H_
#define _BOUNDEDQUEUE_H_

#include <deque>
#include <boost/utility.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/exceptions.hpp>
#include <boost/thread/xtime.hpp>

  /**
   * @class BoundedQueue
   *
   * @author Eric Crahen <http://www.code-foo.com>
   * @date <2003-07-16T13:54:04-0400>
   * @version 2.3.0
   *
   * A BoundedQueue provides serialized access to a set of values. It differs from other
   * Queues by adding a maximum capacity, giving it the following properties:
   *
   * - Threads calling the empty() methods will be blocked until the BoundedQueue becomes empty. 
   * - Threads calling the next() methods will be blocked until the BoundedQueue has a value to
   *   return. 
   * - Threads calling the add() methods will be blocked until the number of values in the 
   *   Queue drops below the maximum capacity.
   *
   * @see Queue
   */
template <class T>
class BoundedQueue : private boost::noncopyable{

  //! Maximum capacity for the Queue
  size_t _capacity;

  //! Serialize access  
  boost::timed_mutex _lock;  

  //! Signaled if not full
  boost::condition _notFull;

  //! Signaled if not empty
  boost::condition _notEmpty;

  //! Signaled if empty
  boost::condition _isEmpty;
  
  //! Storage backing the queue
  std::deque<T> _queue;

  //! Cancellation flag
  volatile bool _canceled;

  public:


	BoundedQueue(size_t capacity)
	      : _capacity(capacity), _canceled(false) {}
	//        : _notFull(_lock), _notEmpty(_lock), _isEmpty(_lock), 
	
	//! Destroy this Queue
	
	~BoundedQueue() { }
	
	
	size_t capacity() { 
	  return _capacity; 
	}
	
	
	void add(const T& item) {    
	  boost::timed_mutex::scoped_lock lk(_lock);
	  
	  // Wait for the capacity of the Queue to drop 
	  while ((_queue.size() == _capacity) && !_canceled) {
	    _notFull.wait(lk);
	  }
	
	  if(_canceled)
	    throw boost::lock_error();
	
	  _queue.push_back(item);
	  _notEmpty.notify_one(); // Wake any waiters
	}
	
	
	bool add(const T& item, unsigned long timeout) {
	  boost::xtime xt;
	  boost::xtime_get(&xt, boost::TIME_UTC);
	  xt.nsec += timeout*1000;
	  boost::timed_mutex::scoped_timed_lock lk(_lock, xt);
	      
	  // Wait for the capacity of the Queue to drop 
	  while ((_queue.size() == _capacity) && !_canceled)
	    if(!_notFull.timed_wait(lk, xt))
	      return false;
	  
	  if(_canceled)
	    throw boost::lock_error();
	  
	  _queue.push_back(item);
	  _notEmpty.notify_one(); // Wake any waiters
	  
	  return true;
	}
	
	
	T next() {
	  boost::timed_mutex::scoped_lock lk(_lock);
	  
	  while ( _queue.size() == 0 && !_canceled)
	    _notEmpty.wait(lk);
	
	  if( _queue.size() == 0) // Queue canceled
	    throw boost::lock_error();  
	
	  T item = _queue.front();
	  _queue.pop_front();
	  
	  _notFull.notify_one(); // Wake any thread trying to add
	
	  if(_queue.size() == 0) // Wake empty waiters
	      _isEmpty.notify_all();
	
	  return item;
	}
	
	
	T next(unsigned long timeout) {
	  boost::timed_mutex::scoped_lock lk(_lock);
	
	  boost::xtime xt;
	  boost::xtime_get(&xt, boost::TIME_UTC);
	  xt.nsec += timeout*1000;
	  // Wait for items to be added
	  while (_queue.size() == 0 && !_canceled) {
	    if(!_notEmpty.timed_wait(lk, xt))
	      throw boost::lock_error();
	  }
	
	  if(_queue.size() == 0)  // Queue canceled
	    throw boost::lock_error();  
	
	  T item = _queue.front();
	  _queue.pop_front();
	
	  _notFull.notify_one(); // Wake add() waiters
	
	  if(_queue.size() == 0) // Wake empty() waiters
	    _isEmpty.notify_all();
	
	  return item;
	}
	
	
	void cancel() {
	  boost::timed_mutex::scoped_lock lk(_lock);
	
	  _canceled = true;
	  _notEmpty.notify_all(); // Wake next() waiters
	}
	
	
	bool isCanceled() {
	  // Faster check since the Queue will not become un-canceled
	  if(_canceled)
	    return true;
	
	  boost::timed_mutex::scoped_lock lk(_lock);
	  return _canceled;
	}
	
	
	size_t size() {
	  boost::timed_mutex::scoped_lock lk(_lock);  
	  return _queue.size();
	}
	
	
	size_t size(unsigned long timeout) {
	  boost::xtime xt;
	  boost::xtime_get(&xt, boost::TIME_UTC);
	  xt.nsec += timeout*1000;
	  
	  boost::timed_mutex::scoped_timed_lock lk(_lock, xt);
	  return _queue.size();
	}
	
	bool empty() {
	  boost::timed_mutex::scoped_lock lk(_lock);
	  while(_queue.size() > 0) // Wait for an empty signal
	    _isEmpty.wait(lk);
	  return true;
	}
	
	bool empty(unsigned long timeout) {
	  boost::xtime xt;
	  boost::xtime_get(&xt, boost::TIME_UTC);
	  xt.nsec += timeout*1000;
	  boost::timed_mutex::scoped_timed_lock lk(_lock, xt);
	  while(_queue.size() > 0) // Wait for an empty signal
	    _isEmpty.timed_wait(lk, xt);
	  return true;
	}
}; /* BoundedQueue */

#endif //_BOUNDEDQUEUE_H_
