/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "DatabaseConnection.h"

#include "configparser/ConfigParser.h"
#include "Statement.h"
#include "Utils.h"

#include <sqlext.h>
#include <stdexcept>
#include <string>
#include <iostream>
#include <assert.h>

#define SQL_ERR_MSG_SIZE 1024

using namespace std;

bool DatabaseConnection::debug = false;
#include "debug.h"

bool DatabaseConnection::initialized= false;
std::string DatabaseConnection::dataSource="";
std::string DatabaseConnection::dbUser="";
std::string DatabaseConnection::dbPass="";
std::string DatabaseConnection::dbSchema="";  // Schema to search for AMGA tables
std::string DatabaseConnection::defaultSchema="";  // Detected schema
void DatabaseConnection::initialize(ConfigParser &configParser) {
  if (initialized) {
    throw logic_error("Already initialized");
  }
  try {
    dataSource=configParser.getString("DataSource");
  } catch (std::string e) {
    LOG("WARNING: Using default data source PSQL\n");
    dataSource="PSQL";
  }
  
  try {
    dbUser=configParser.getString("DBUser");
    try{
        dbPass=configParser.getString("DBPass");
    } catch (std::string e) {
        LOG("WARNING: Using default password for DB connection with user " << dbUser << "\n");
        dbPass="";
    }
  } catch (std::string e) {
    LOG("WARNING: Using default user/password for ODBC datasource\n");
    dbUser="";
    dbPass="";
  }

  try {
    dbSchema = configParser.getString("DBSchema");
  } catch (std::string e) { }

  initialized = true; 
}


DatabaseConnection *DatabaseConnection::openConnection() throw (SQLException) {
    if (!initialized) {
        throw logic_error("Not initialized");
    }
    return new DatabaseConnection(dataSource, dbUser, dbPass);
}



DatabaseConnection::DatabaseConnection(const std::string &dataSource,
				       const std::string &dbUser,
				       const std::string &dbPass)
  throw (SQLException)
  : _cHandle(NULL), _envHandle(NULL), _statement(NULL),
    _transactionCount(0), _errorsDuringTransaction(0), connAlive(true),
    _manualCommit(false), _clientTransactionID(0), dType(DatabaseConnection::dtUnknown)
{
  DMESG(this <<"Connecting to backend: " << dbUser << "@" << dataSource << endl);
  // Allocate Environment handle and register version
  long res = SQLAllocHandle(SQL_HANDLE_ENV,SQL_NULL_HANDLE, &_envHandle);
  if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)){
    ostringstream os;
    os << "Cannot initialize ODBC: " << res << "\n";
    throw SQLException(os.str());
  }
  try {
    res=SQLSetEnvAttr(_envHandle, SQL_ATTR_ODBC_VERSION, (void*)SQL_OV_ODBC3, 0);
    if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)){
      ostringstream os;
      os << "ODBC version mismatch: " << res << "\n";
      throw SQLException(os.str());
    }
    
    // Allocate connection handle, set timeout
    res = SQLAllocHandle(SQL_HANDLE_DBC, _envHandle, &_cHandle);
    if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
      ostringstream os;
      os << "Cannot allocate connection handle: " << res << "\n";
      throw SQLException(os.str());
    }       
    
    SQLINTEGER  sqlErr;
    char sqlErrMsg[SQL_ERR_MSG_SIZE];
    char sqlStat[10];
    for (int t=1;;) {
      SQLSetConnectAttr(_cHandle, SQL_LOGIN_TIMEOUT, (SQLPOINTER *)5, 0);
      
      // Connect to datasource PSQL
      res = SQLConnect(_cHandle, (SQLCHAR*) dataSource.c_str(), SQL_NTS,
		       (SQLCHAR*)( 0 == dbUser.length() ? NULL : dbUser.c_str()), SQL_NTS,
		       (SQLCHAR*)( 0 == dbPass.length() ? NULL : dbPass.c_str()), SQL_NTS); //was NULL

      if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)){
	SQLSMALLINT len;
	SQLGetDiagRec(SQL_HANDLE_DBC, _cHandle,1,
		      (SQLCHAR *)sqlStat, &sqlErr, (SQLCHAR *)sqlErrMsg,
		      SQL_ERR_MSG_SIZE, &len);
	LOG("MDServer: Cannot connect to DB: " << sqlErrMsg << " "
	    << sqlErr << std::endl);

	if(t>40) {
	  ostringstream os;
	  os << "Could not connect to DB" << sqlErrMsg << " " << sqlErr << "\n";
	  throw SQLException(os.str());
	}
	
	LOG("Retry in " << t << " seconds" << std::endl);
	sleep(t);
	t*=4;
	
      } else {
	DMESG(this << "Connected to database\n");
	// DEBUG: Set serializable mode
	// cout << "Return: " << SQLSetConnectAttr(_cHandle, SQL_ATTR_TXN_ISOLATION, (SQLPOINTER *)SQL_TXN_SERIALIZABLE, 0) << endl;
	return;
      }
    }
  } catch (...) {
    SQLFreeHandle(SQL_HANDLE_ENV, _envHandle);
    throw;
  }
}

SQLHDBC DatabaseConnection::getConnectionHandle() {
    return _cHandle;
}

int DatabaseConnection::getDBVersion() throw (SQLException)
{
   char buff[256];
   SQLSMALLINT len;
 
   long res=SQLGetInfo(_cHandle, SQL_DBMS_VER,
               (SQLCHAR *)buff, 255, &len);
    if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
        throw SQLException("Could not get DB version");     
    }
    buff[len]='\0';
    DMESG(this << "DBMS version >" << buff <<"<" <<std::endl);

    return toInteger(buff);
}

string DatabaseConnection::getODBCDriverVersion() throw (SQLException) {
    char version[21];
    SQLRETURN error = SQLGetInfo(_cHandle, SQL_DRIVER_VER, version, 20, NULL);
    if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
        throw SQLException("Could not determine driver version");
    }
    return version;
}

string DatabaseConnection::getODBCDriverName() throw (SQLException) {
    char buff[81];
    SQLRETURN error = SQLGetInfo(_cHandle, SQL_DRIVER_NAME, buff, 80, NULL);
    if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
        throw SQLException("Could not determine driver name");
    }
    return buff;
}


std::string DatabaseConnection::getDatabaseName()
  throw (SQLException)
{
  char info[256];
  
  SQLRETURN error = SQLGetInfo( _cHandle, SQL_DATABASE_NAME,
				(SQLCHAR *) info, 255, NULL);
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    throw SQLException("Could not get database name");
  } 
  return info;
}


DatabaseConnection::DriverType DatabaseConnection::getDriverType()
{
  if (dType == dtUnknown) {
    char serverName[256];
    char description[256];
    SQLSMALLINT len1, len2;
    bool first=true;

  
    for(;;){
      SQLSMALLINT direction=SQL_FETCH_FIRST;
      if(!first) direction=SQL_FETCH_NEXT;
      first=false;
  
      SQLRETURN res=SQLDataSources(_envHandle, direction,
				   (SQLCHAR *)serverName, 255, &len1,
				   (SQLCHAR *)description, 255, &len2);
      if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO))
        break;
  
      serverName[len1]='\0';
      description[len2]='\0';
  
      DMESG(this << " Data source >" << serverName <<"<" <<std::endl);
      DMESG(this << " >" << description << "<" << std::endl);
  
      if(dataSource==serverName) {
        DMESG(this << " Matching data source" << std::endl);
        std::string desc(description);
        if(desc.find("psqlodbc")!=std::string::npos) {
      dType =  dtPSQL;  // Case where user did not rename Driver in .odbc.ini
      break;
        } else if(desc.find("PostgreSQL")!=std::string::npos) {
      dType =  dtPSQL;  // User has Driver     = PostgreSQL
      break;
        } else if(desc.find("sqliteodbc")!=std::string::npos) {
      dType =  dtSQLite;
      break;
        } else if(desc.find("SQLite")!=std::string::npos){
      dType =  dtSQLite;
      break;
        } else if(desc.find("Oracle")!=std::string::npos){
      dType =  dtOracle;
      break;
        } else if(desc.find("MySQL")!=std::string::npos){
      dType =  dtMySQL;
      break;
        } else if(desc.find("myodbc")!=std::string::npos){
      dType =  dtMySQL;
      break;
        } else  {
          dType =  dtUnknown;
	  break;
	}
      }
    }
    
    if(dType == DatabaseConnection::dtMySQL) {
    	int version = getDBVersion();
    	DMESG("DB version: " << version << std::endl);
    	if (version >=5 ) {
      		dType = DatabaseConnection::dtMySQL5;
    	}
  	}
  }
  return dType;
}


SQLHSTMT DatabaseConnection::getStatement() {
  return _statement;
}

SQLHSTMT DatabaseConnection::createStatement() throw (SQLException) {
  DMESG("Create Statement \n");
  if (_statement != NULL) {
    DMESG("Already Open \n");
    throw SQLException("Statement already open");
  }

  DMESG("Init Statement \n");
  // Init the statement...
  SQLRETURN error=SQLAllocHandle(SQL_HANDLE_STMT, _cHandle, &_statement); 
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    DMESG("Could not create statement \n");
    throw SQLException("FATAL: Could not create statement");
  }
  return _statement;
}

void DatabaseConnection::closeStatement() throw (SQLException) {
    if (_statement != NULL) {
//        SQLRETURN error = SQLFreeHandle(SQL_HANDLE_STMT, _statement);
        SQLFreeHandle(SQL_HANDLE_STMT, _statement);
        _statement = NULL;
    }
}

DatabaseConnection::~DatabaseConnection()
{
  closeStatement();
  SQLDisconnect(_cHandle);
  DMESG(this << " Cleaning up ODBC environment" << std::endl);
  SQLFreeHandle(SQL_HANDLE_DBC, _cHandle);
  SQLFreeHandle(SQL_HANDLE_ENV, _envHandle);
}

void DatabaseConnection::beginTransaction(bool exclusive) throw (SQLException)
{
  DMESG(this << " Statement::beginTransaction ("<< _transactionCount << ")\n");
  if (_transactionCount == 0) {
    DMESG(this << " Opening transaction\n");
    _errorsDuringTransaction = 0;
    
    if(!exclusive) {
      startManualCommit();
    } else {
      DMESG(this << " exclusive");
      Statement st(*this);

      if ( dType ==  dtPSQL ) { 
        if (st.exec("BEGIN ;")) {
          throw SQLException(string("Error opening transaction: ") 
			   + st.getError());
        }
      } else {
        if (st.exec("BEGIN EXCLUSIVE;")) {
          throw SQLException(string("Error opening exclusive transaction: ")
                           + st.getError());
        }
      }
    }
  }
  _transactionCount++;
}


void DatabaseConnection::commitTransaction() throw (SQLException)
{
  DMESG(this << " DatabaseConnection::commitTransaction (" << _transactionCount << ")\n");
  if(_transactionCount == 0) {    
    DMESG(this << " No transaction in progress\n");
    throw SQLException("No transaction in progress");
  }

  assert(_transactionCount >= 1);
  
  if (_transactionCount == 1) {
      _errorsDuringTransaction = 0;
      _transactionCount = 0;
      _clientTransactionID = 0;
      
      if(_manualCommit){
        SQLSMALLINT error=SQLEndTran(SQL_HANDLE_DBC, _cHandle, SQL_COMMIT);
        checkError(error, SQL_HANDLE_DBC, _cHandle);                
        startAutoCommit();
    
      } else {
        Statement st(*this);
        if(!st.exec("COMMIT;")) {
            throw SQLException(st.getError());
        }
      }
  } else {
    DMESG(this << " Decreasing transaction count\n");
    _transactionCount--;
  }
}

void DatabaseConnection::increaseErrorsDuringTransaction() { 
  if(_transactionCount > 0) {
    DMESG(this << " increasing errors during transaction\n");
    _errorsDuringTransaction++;
  }
}

bool DatabaseConnection::checkAlive() 
{
  Statement st(*this);
  return st.checkAlive();
}

void DatabaseConnection::checkError(int error, SQLSMALLINT handleType,
				    SQLHANDLE handle) 
  throw (SQLException)
{
    if ((error == SQL_SUCCESS) || (error == SQL_SUCCESS_WITH_INFO))
        return;

    // Get the error messages
    SQLINTEGER  sqlErr;
    char        sqlErrMsg[256];
    char        sqlStat[10];
    SQLSMALLINT len;
    SQLRETURN ret = SQLGetDiagRec(handleType, handle, 
            1, (SQLCHAR *)sqlStat,
            &sqlErr, (SQLCHAR *)sqlErrMsg,  255, &len);
    if(ret==SQL_SUCCESS){
      sqlErrMsg[len]='\0';
    } else {
      sqlErrMsg[0]='\0';
    }

    // Remove any linefeeds
    int i=0;
    while(sqlErrMsg[i]){
      sqlErrMsg[i] = sqlErrMsg[i]=='\n'? ' ' : sqlErrMsg[i];
      i++;
    }

    DMESG(this << " Error: " << sqlErrMsg << endl);
    throw SQLException(string(sqlErrMsg) + ":" + sqlStat);
}


void DatabaseConnection::rollbackTransaction() throw (SQLException)
{
  _errorsDuringTransaction = 0;
  _clientTransactionID = 0;
  
  DMESG(this << " Statement::ROLLBACK called" << std::endl);
  if(_transactionCount == 0) {
    DMESG(this << "No transaction in progress" << std::endl);
    throw SQLException("No transaction in progress");
  }
    
  // Any client can rollback, regardless of the level of nested transaction.
  // The rationale is that if an error occurred, the global transaction
  // must also be aborted.

  if(_manualCommit){
    SQLSMALLINT error=SQLEndTran(SQL_HANDLE_DBC, _cHandle, SQL_ROLLBACK);    
    // Throws an exception if there was an error.
    checkError(error, SQL_HANDLE_DBC, _cHandle);
    _transactionCount = 0;

    startAutoCommit();

  } else {
    Statement st(*this);
    if(!st.exec("ROLLBACK;")) {
      _transactionCount = 0;
      throw SQLException(st.getError());
    }
  }
}

void DatabaseConnection::startManualCommit() throw (SQLException)
{
  DMESG("startManualCommit() called\n");
  SQLSMALLINT error = SQLSetConnectAttr(_cHandle, SQL_ATTR_AUTOCOMMIT,
              (void *)SQL_AUTOCOMMIT_OFF, 0);
  checkError(error, SQL_HANDLE_DBC, _cHandle);
  _manualCommit=true;
}


void DatabaseConnection::startAutoCommit() throw (SQLException)
{
  DMESG("startAutoCommit() called\n");
  SQLSMALLINT error = SQLSetConnectAttr(_cHandle, SQL_ATTR_AUTOCOMMIT,
              (void *)SQL_AUTOCOMMIT_ON, 0);
  checkError(error, SQL_HANDLE_DBC, _cHandle);
  _manualCommit=false;
}
