/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef DATABASECONNECTION_H_
#define DATABASECONNECTION_H_

#include "DBUtil.h"

#include <sql.h>
#include <string>

#include <stdint.h>

class ConfigParser;


/**
 * Manages a physical connection to the database. Creates and holds 
 * statement objects. The Statement class works in close collaboration 
 * with this one.
 */
class DatabaseConnection
{
public:
    static void initialize(ConfigParser &configParser);
    static DatabaseConnection *openConnection() throw (SQLException);

    DatabaseConnection(const std::string &dataSource,
		       const std::string &dbUser,
		       const std::string &dbPass) throw (SQLException);
    virtual ~DatabaseConnection();
    int getDBVersion() throw (SQLException);
    
    enum DriverType {
        dtUnknown, dtPSQL, dtSQLite, dtMySQL, dtMySQL5, dtOracle
    };

    std::string getODBCDriverVersion() throw (SQLException);
    std::string getODBCDriverName() throw (SQLException);
    
    DriverType getDriverType();

    std::string getDatabaseName()
      throw (SQLException);

    SQLHDBC getConnectionHandle();
    SQLHSTMT getStatement();
    SQLHSTMT createStatement() throw (SQLException);
    void closeStatement() throw (SQLException);

    // Transaction processing
    void commitTransaction() throw (SQLException);
    void beginTransaction(bool exclusive=false) throw (SQLException);
    void rollbackTransaction() throw (SQLException);
    bool transactionInProgress() const { return _transactionCount; };
    
    // If no transaction is in progress, doesn't do anything
    void increaseErrorsDuringTransaction();
    int getErrorsDuringTransaction() {
        return _errorsDuringTransaction;
    };

    bool checkAlive();


    static std::string dataSource;
    static std::string dbUser;
    static std::string dbPass;
    static std::string dbSchema;       // Schema to search for AMGA tables
    static std::string defaultSchema;  // Discovered schema name

    bool connAlive;

 private:
    friend class Statement;
    SQLHDBC          _cHandle;
    SQLHENV          _envHandle;    
    SQLHSTMT         _statement;    
    // Keeps track of the levels of transactions. A transaction
    // is open on the DB only when beginTransaction is called for the first time.
    // It is closed when this counter reaches 0.
    int              _transactionCount;
    int              _errorsDuringTransaction;
    // Transaction handling.
    bool            _manualCommit;
    // If we are on the middle of an explicit transaction requested 
    // by the client, this contains the transaction id. Otherwise, it's 0.
    uint64_t        _clientTransactionID;

    void startAutoCommit() throw (SQLException);
    void startManualCommit()throw (SQLException);

    void checkError(int error, SQLSMALLINT handleType, SQLHANDLE handle) 
        throw (SQLException);
    DriverType dType;

    // Static fields
    static bool debug;
    static bool initialized;
};

#endif /*DATABASECONNECTION_H_*/
