/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: SecurityUtils.cpp,v 1.17 2012/02/01 06:42:03 gpark Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "SecurityUtils.h"
#include "configparser/ConfigParser.h"
#include "server/VOMSInfo.h"
#include "sslHelpers.h"

#include "openssl_mt.h"
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/safestack.h>
#include <stdexcept>
#include <stdlib.h>

#include "debug.h"

using namespace std;

AuthContext::~AuthContext()
{

}
    
void AuthContext::init(ConfigParser &configParser) {
  // Requrire Authentication?
  try {
    requireAuthentication=configParser.getInt("RequireAuthentication");
  } catch (std::string e) {
    requireAuthentication = true;
  }

  // Initilize SSL
  useSSL=false;
  allowGridProxyLogin=false;
  try {
    if(configParser.getInt("UseSSL")){
      useSSL=true;
      if( (allowCertAuth = configParser.getInt("AllowCertificateAuthentication")) ){
        allowGridProxyLogin=configParser.getInt("AllowGridProxyLogin");
        allowMyProxyHack=configParser.getInt("MyProxyHack");
      }
      DMESG("verify client" << allowCertAuth << " allow gp: "
        << allowGridProxyLogin << std::endl);

    }
  } catch(std::string e) { };

  // ignore crl checking ?
  if(useSSL) {
    try {
      ignoreCRLCheck = configParser.getInt("IgnoreCRLCheck");
    } catch (std::string e) {
      ignoreCRLCheck = 1;
    }
    DMESG("IgnoreCRLCheck : " << ignoreCRLCheck << std::endl);
  }

  if(useSSL) {
    if(initCertificates(allowCertAuth, allowGridProxyLogin, configParser)) {
        throw runtime_error("Could not initialize TLS/SSL certificates");
    }
  }

  // Password authentication allowed?
  try {
    allowPasswordAuthentication=configParser.getInt("AllowPasswordAuthentication");
  } catch (std::string e) {
    allowPasswordAuthentication = false;
  }


}


int AuthContext::initCertificates(bool allowCertAuth,
				  bool allowGridProxyLogin,
				  ConfigParser &configParser)
{
  SSLManager::init();
  DMESG("Initializing certificates\n");
  // Creates server instance
  // For SSL 1.0 Higher : SSL 1.0 change SSL_METHOD to const SSL_METHOD
#if OPENSSL_VERSION_NUMBER >= 0x10000000L
    const SSL_METHOD *method= SSLv23_server_method();
#else
    SSL_METHOD *method= SSLv23_server_method();
#endif

  if( !method ) {
    sslPrintErrors();
    abort();
  }
  // Create SSL context
  ctx = SSL_CTX_new(method);
  if( !ctx ) {
    sslPrintErrors();
    abort();
  }
  // Disable buggy v2, enable all compatibility measures
  SSL_CTX_set_options(ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);
  DMESG("Loading security settings\n");
  // Load local certificate from file
  try {
    DMESG("Reading: " << configParser.getString("CertFile") << endl);
    if(SSL_CTX_use_certificate_file(ctx, configParser.getString("CertFile").c_str(),
                    SSL_FILETYPE_PEM)!=1){
      sslPrintErrors();
      return -1;
    }
    DMESG("Reading: " << configParser.getString("KeyFile") << endl);
    // Load private key from local file
    if(SSL_CTX_use_PrivateKey_file(ctx, configParser.getString("KeyFile").c_str(),
                   SSL_FILETYPE_PEM)!=1){
      sslPrintErrors();
      return -1;
    }
  } catch (std::string err){
    ERRLOG("Error initializing server certificate: " << err << std::endl);
    return -1;
  }

  // Verify private key
  if ( !SSL_CTX_check_private_key(ctx) ) {
    sslPrintErrors();
    return -1;
  }

  // Allow client authentication through certificates
  if(allowCertAuth) {

    // Set the certificate verification callback
    if(allowGridProxyLogin) {
      // Use SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT to make sure
      // that certifictes are allways sent
      SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, 
			 certVerifyCallbackAllowProxy);
      LOG("Allowing grid proxy authentication" << std::endl);
    } else {
      SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER,
			 certVerifyCallback);
      DMESG("NOT allowing grid proxy authentication" << std::endl);
    }

    // Load CA certificates
    try {
      DMESG("Loading trusted CA certificates" << std::endl);
      std::string certDir = configParser.getString("TrustedCertDir");
      if( !SSL_CTX_load_verify_locations(ctx, NULL, certDir.c_str()) ){
	sslPrintErrors();
	ERRLOG("Error loading trusted certificates" << std::endl);
	return -1;
      }

      // In openssl 0.9.7 we can use CRLs, we also 
      // need our own issuer check callback, note that even if we have
      // openssl 0.9.8, we cannot use the generic proxy certificates
      // checking the, because they are incompatible to grid proxies:
      // http://www.openssl.org/docs/HOWTO/proxy_certificates.txt


#if OPENSSL_VERSION_NUMBER >= 0x00907000L

      X509_STORE *store = SSL_CTX_get_cert_store(ctx);

      if ( ignoreCRLCheck < 2 ) {
        try { // Additional CRL directory?
          std::string crlDir = configParser.getString("CRLDir");
          // Load CRLs, see code from SSL_X509_STORE_create in ssl_util_ssl.c
          // of mod_ssl, e.g.: at
          // http://www.koders.com/c/fid6EC9E5D7563BBF599E1CF74685C10D05FE52B76A.aspx
          DMESG("Loading CRLs from " << crlDir << '\n');
          X509_LOOKUP *lookup;
          if( !(lookup = X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir()))) {
            ERRLOG("ERROR: Could not add CRL LOOKUP object\n");
            sslPrintErrors();
          }
      
          if (X509_LOOKUP_add_dir(lookup, crlDir.c_str(), X509_FILETYPE_PEM) ){
            ERRLOG("ERROR: Could not add CRL directory : " << crlDir << '\n');
            sslPrintErrors();
          }
        } catch (std::string err) { }
      }

      // We need to override the check_issued function because proxy
      // certificates normally would not be allowed to sign other
      // certificates
      // Instead of the below hack to use a different callback for the
      // issuer checking, we can also use the
      // X509_V_FLAG_CB_ISSUER_CHECK flag to route all issuer checks
      // to the normal verify callback function
      if(allowGridProxyLogin) {
	DMESG("Overrriding cert_store->check_issued of openssl 0.9.7+\n");
	store->check_issued=checkIssuerCallback;
      }
#endif

#ifdef X509_V_FLAG_ALLOW_PROXY_CERTS
      if(allowGridProxyLogin) {
        X509_STORE_set_flags( store, X509_V_FLAG_ALLOW_PROXY_CERTS );
      }
#endif

      // Enable the checking of CRLs, we need to check all CRLs
      // because otherwise we might check only the proxy certificate's
      // CRL, not the user's CRL. If we add
      // X509_V_FLAG_CB_ISSUER_CHECK, we only need to change one CB

      if ( ignoreCRLCheck < 2 ) {
        X509_STORE_set_flags( store, X509_V_FLAG_CRL_CHECK | 
                                 X509_V_FLAG_CRL_CHECK_ALL );
        LOG("Enabled checking of Certificate Revocation Lists\n");
      }


    } catch (std::string err) {
      LOG("WARNING: Not loading any trusted certificates\n"
	  << "     Certificate based authentication will not work!\n");
    }
  } else {
    LOG("No client certificate verification!" << std::endl);
  }
  return 0;
}

    
void AuthContext::cleanupSSL() {
  EVP_cleanup();
}

/*************************
 * SSL class
 *************************/
bool SSLManager::initialized = false;

void SSLManager::init(bool debug) {
    if (!initialized) {
        initialized = true;
        DMESG("Initializing SSL\n");
        CRYPTO_thread_setup();
        // Load error messages
        SSL_load_error_strings();
        // Load crypto algorithms 
        SSL_library_init();
        OpenSSL_add_all_algorithms();        
    }
}

void SSLManager::cleanup() {
    if (initialized) {
        initialized = false;
        CRYPTO_thread_cleanup();
    }
}


/// Callback when certificates are checked
/** This callback is given to SSL_CTX_set_verify and prints out error
    messages when certificate chains are verified.
 */
int serverCertVerifyCallback(int ok, X509_STORE_CTX *store)
{
  // See man SSL_CTX_set_verify

  char buff[256];
  X509 *cert = X509_STORE_CTX_get_current_cert(store);
  int depth = X509_STORE_CTX_get_error_depth(store);
  int err = X509_STORE_CTX_get_error(store);
  int debug = 0;

  if(debug) {
    DMESG("Verifying server certificate:\n");
    X509_NAME_oneline(X509_get_issuer_name(cert), buff, 256);
    DMESG("   issuer = " << buff << std::endl);
    X509_NAME_oneline(X509_get_subject_name(cert), buff, 256);
    DMESG("   subject = " << buff << std::endl);
  }

  if(!ok) {
    ERRLOG("Server certificate error at depth: " << depth
       << " error " << err << std::endl);
    X509_NAME_oneline(X509_get_issuer_name(cert), buff, 256);
    ERRLOG("   issuer = " << buff << std::endl);
    X509_NAME_oneline(X509_get_subject_name(cert), buff, 256);
    ERRLOG("   subject = " << buff << std::endl);
    ERRLOG("   error   = " << X509_verify_cert_error_string(err)
       << std::endl);

  }

  return ok;
}

std::string sslErrors() {
  std::string error;

  char buff[2001];
  BIO *mem=BIO_new(BIO_s_mem());
  ERR_print_errors(mem);
  string sep = "";
  while (!BIO_eof(mem)) {
    BIO_gets(mem, buff, 2000);
    error.append(sep);
    error.append(buff);
    sep = "/";
  }

  BIO_free(mem);

  return error;
}


std::string getProxyCertName()
{
  char *c = getenv("X509_USER_PROXY");
  if( c )
    return c;

  std::stringstream name;
  name << "/tmp/x509up_u";
  name << getuid();

  return name.str();
}
