/*
  Copyright (c) Members of the EGEE Collaboration. 2004.
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
*/

// -*- Mode:c++ -*-
//
// $Id$
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Socket.h"
#include "SecurityUtils.h"

#ifdef WIN32
  #include <winsock.h>         // For socket(), connect(), send(), and recv()
  typedef int socklen_t;
  typedef char raw_type;       // Type used for raw data on this platform
#else
  #include <sys/types.h>       // For data types
  #include <sys/socket.h>      // For socket(), connect(), send(), and recv()
  #include <netdb.h>           // For gethostbyname()
  #include <arpa/inet.h>       // For inet_addr()
  #include <unistd.h>          // For close()
  #include <netinet/in.h>      // For sockaddr_in
  #include <netinet/tcp.h>
  typedef void raw_type;       // Type used for raw data on this platform
#endif

#include <errno.h>             // For errno
#include <string.h>
#include <sstream>
#include <unistd.h> 

#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/safestack.h>

#include "debug.h"
static bool debug = false;

using namespace std;

#ifdef WIN32
static bool initialized = false;
#endif


// SocketException Code

SocketException::SocketException(const string &message, bool inclSysMsg)
  throw() : userMessage(message) {
  if (inclSysMsg) {
    err = errno;
    userMessage.append(": ");
    userMessage.append(strerror(errno));
  }
}

SocketException::~SocketException() throw() {
}

const char *SocketException::what() const throw() {
  return userMessage.c_str();
}

int SocketException::errorNumber() const {
  return err;
}


// Function to fill in address structure given an address and port
static void fillAddr(const string &address, unsigned short port, 
                     struct sockaddr_storage &addr, std::list< std::string > &hostNames)
{
#ifndef WIN32
  // This code does not compile on CYGWIN, but is threadsafe!
  struct addrinfo req, *ans=NULL;
  bzero(&req,sizeof(req));
  // Request only IPv6 sockets
  req.ai_family = PF_UNSPEC;
  req.ai_socktype = SOCK_STREAM;
  req.ai_flags = AI_CANONNAME ; 

  int err;

  if ((err = getaddrinfo(address.c_str(), "0", &req, &ans)) != 0) {
    std::string m = "Lookup error2: ";
    m += gai_strerror(err);
    throw SocketException(m);
  }  
  if ( ans->ai_family == AF_INET ) 
    ((struct sockaddr_in*)ans->ai_addr)->sin_port = htons(port);
  else 
    ((struct sockaddr_in6*)ans->ai_addr)->sin6_port = htons(port);
  memcpy(&addr, ans->ai_addr, ans->ai_addrlen);
  
  struct addrinfo *p = ans;
  while(p) {
    if( p->ai_canonname )
      hostNames.push_back(p->ai_canonname);
    p = p->ai_next;
  }

  freeaddrinfo(ans);
#else
  // Warning: This works on WINDOWS, but is not reentrancy safe!
  struct hostent *serv = gethostbyname(server.c_str());
  if (server == NULL) {
    fprintf(stderr,"ERROR, no such host\n");
    return -1;
  }
  
  serv_addr.sin_family = AF_INET;
  bcopy( (char *)serv->h_addr, (char *)&serv_addr.sin_addr.s_addr,
         serv->h_length );
  serv_addr.sin_port = htons(port);
  err = ::connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr));
  if( err ) {
    DMESG("Connection refused\n")
    throw SocketException("Could not connect", errno);
  }
#endif
}


int Socket::detectIPV6(const std::string &address) {

  struct addrinfo req, *ans=NULL;

  bzero(&req,sizeof(req));

  req.ai_family = PF_UNSPEC;
  req.ai_socktype = SOCK_STREAM;
  int err;

  if ((err = getaddrinfo(address.c_str(), "0", &req, &ans)) != 0) {
    std::string m = "Lookup error: ";
    m += gai_strerror(err);
    throw SocketException(m);
  }

  int family = ans->ai_family ;
  freeaddrinfo(ans);

  return family;
}



// Socket Code
Socket::Socket(int type, int protocol, const string &address) throw(SocketException) {
  #ifdef WIN32
    if (!initialized) {
      WORD wVersionRequested;
      WSADATA wsaData;

      wVersionRequested = MAKEWORD(2, 0);              // Request WinSock v2.0
      if (WSAStartup(wVersionRequested, &wsaData) != 0) {  // Load WinSock DLL
        throw SocketException("Unable to load WinSock DLL");
      }
      initialized = true;
    }
  #endif

 
  int inet_family = AF_INET6; 

  if ( address.size() != 0 ) {
    inet_family = detectIPV6(address); 
    if ((sockDesc = socket(inet_family, type, protocol)) < 0) 
      throw SocketException("Socket creation failed (socket())", true);
  } else {
    if ((sockDesc = socket(AF_INET6, type, protocol)) < 0) 
      if ((sockDesc = socket(AF_INET, type, protocol)) < 0) 
        throw SocketException("Socket creation failed (socket())", true);
  }
  
  // Reuse address, so that we don't have to wait before using the
  // same port as a server.
  int flag=1;
  if (setsockopt(sockDesc, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag)) < 0) {
    throw SocketException("Setting SO_REUSEADDR failed", true);
  }
  flag = 0;
  if (setsockopt(sockDesc, SOL_SOCKET, SO_OOBINLINE, &flag, sizeof(flag)) < 0) {
    throw SocketException("Disabling SO_OOBINLINE failed", true);
  }    
}


Socket::Socket(int sockDesc) 
{
  this->sockDesc = sockDesc;
}


Socket::~Socket() 
{
  #ifdef WIN32
    ::closesocket(sockDesc);
  #else
    if(::close(sockDesc))
      PERRLOG("Closing socket failed:");
  #endif
  sockDesc = -1;
}

string Socket::getLocalAddress() throw(SocketException) 
{
  struct sockaddr_storage addr;
  unsigned int addr_len = sizeof(addr);
  char address[INET6_ADDRSTRLEN+128];

  if (getsockname(sockDesc, (sockaddr *) &addr, (socklen_t *) &addr_len) < 0) {
    throw SocketException("Fetch of local address failed (getsockname())", true);
  }

  return changeAddress(&addr);
}


unsigned short Socket::getLocalPort() throw(SocketException) 
{
  struct sockaddr_storage addr;
  unsigned int addr_len = sizeof(addr);

  if (getsockname(sockDesc, (sockaddr *) &addr, (socklen_t *) &addr_len) < 0) {
    throw SocketException("Fetch of local port failed (getsockname())", true);
  }
  return changePort(&addr) ;
}

void Socket::setLocalPort(unsigned short localPort) throw(SocketException) {
  // Bind the socket to its port
  sockaddr_in6 localAddr;
  memset(&localAddr, 0, sizeof(localAddr));
  // localAddr.sin6_len = sizeof(localAddr);
  localAddr.sin6_family = AF_INET6;
  // localAddr.sin6_addr.s_addr = htonl(in6addr_any);
  localAddr.sin6_addr = in6addr_any;
  localAddr.sin6_port = htons(localPort);
  
  if (bind(sockDesc, (sockaddr *) &localAddr, sizeof(sockaddr_in6)) < 0) {
    throw SocketException("Set of local port failed (bind())", true);
  }
}

void Socket::setLocalAddressAndPort(const string &localAddress,
    unsigned short localPort) throw(SocketException) {
  // Get the address of the requested host
  struct sockaddr_storage localAddr;
  fillAddr(localAddress, localPort, localAddr, localHostnames);

  if (bind(sockDesc, (sockaddr *) &localAddr, sizeof(sockaddr_in6)) < 0) {
    throw SocketException("Set of local address and port failed (bind())", true);
  }
}

void Socket::cleanUp() throw(SocketException) {
  #ifdef WIN32
    if (WSACleanup() != 0) {
      throw SocketException("WSACleanup() failed");
    }
  #endif
}

unsigned short Socket::resolveService(const string &service,
                                      const string &protocol) {
  struct servent *serv;        /* Structure containing service information */

  if ((serv = getservbyname(service.c_str(), protocol.c_str())) == NULL)
    return atoi(service.c_str());  /* Service is port number */
  else 
    return ntohs(serv->s_port);    /* Found port (network byte order) by name */
}

// CommunicatingSocket Code

CommunicatingSocket::CommunicatingSocket(int type, int protocol, const string &address)  
    throw(SocketException) 
  : Socket(type, protocol, address), _bufTop(0), ssl(NULL) {

}


CommunicatingSocket::~CommunicatingSocket() 
{
  if( ssl ) 
    sslShutdown();
}


void CommunicatingSocket::sslShutdown()
{
  if( ssl ) {
    int shutdown=SSL_shutdown(ssl);
    DMESG("SSL shutdown returned: " << shutdown << std::endl);
    
    if(shutdown == 0)
      shutdown = SSL_shutdown(ssl); // Needed on the client

    SSL_free(ssl);
    ssl = NULL;
  }
}


SSL_SESSION *CommunicatingSocket::getSSLSession()
{
  return SSL_get1_session(ssl);
}


bool CommunicatingSocket::isReadyToRead(int seconds, int msecs)
{    
    // First check if there is data on the local buffer
    if (_bufTop != 0) {
        // Yup, so the socket is ready for being read
        return true;
    }
    // No data on the local buffer. Perhaps there is something on the
    // TCP in buffer.
    fd_set rfds;    
    FD_ZERO(&rfds);
    FD_SET(sockDesc, &rfds);
  
    struct timeval tv;
    tv.tv_sec = seconds;
    tv.tv_usec = msecs;
    
    int ret = select(sockDesc+1, &rfds, NULL, NULL, &tv);
    if (ret == -1) {
      // Error
      throw SocketException("Selecting on socket failed.", true);
    }
    // Return true if it there was no interruption
    return ret == 1;
}  
  
CommunicatingSocket::CommunicatingSocket(int newConnSD) 
: Socket(newConnSD), _bufTop(0), ssl(NULL) 
{
  memset(_buffer, 0, sizeof(_buffer));
}

void CommunicatingSocket::connect(const string &foreignAddress,
    unsigned short foreignPort) throw(SocketException) {
  // Get the address of the requested host
  struct sockaddr_storage  destAddr;
  fillAddr(foreignAddress, foreignPort, destAddr, foreignHostnames);

    if (::connect(sockDesc, (sockaddr *) &destAddr, sizeof(struct sockaddr_storage)) < 0) {
    throw SocketException("Connect failed (connect())", true);
  }
}

void CommunicatingSocket::send(const void *buffer, int bufferLen) 
    throw(SocketException) 
{
  // MSG_NOSIGNAL prevents SIGPIPE from being sent to the
  // process when the socket is broken.
  if (ssl == NULL) {
    int totalbytes = 0;
    while(totalbytes < bufferLen) {
      raw_type *b = (raw_type *)((char *)buffer + totalbytes);;
      int bytes = ::send(sockDesc, b, bufferLen-totalbytes, MSG_NOSIGNAL);
      if (bytes < 0) {
        throw SocketException("Send failed (send())", true);
      }
      totalbytes += bytes;
    }
  } else {
    // TODO: SIGPIPE is killing the replication daemon. catch it
      if (::SSL_write(ssl, (raw_type *) buffer, bufferLen) < 0) {
        throw SocketException("Send failed (SSL_send())", true);
      }    
  }
}

void CommunicatingSocket::sendOutOfBand(const void *buffer, int bufferLen) 
    throw(SocketException) 
{
  // MSG_NOSIGNAL prevents SIGPIPE from being sent to the
  // process when the socket is broken
//  if (::send(sockDesc, (raw_type *) buffer, bufferLen, MSG_NOSIGNAL|MSG_OOB) < 0) {
  if (::send(sockDesc, (raw_type *) buffer, bufferLen, MSG_OOB) < 0) {
    throw SocketException("Send failed (send())", true);
  }
}

void CommunicatingSocket::setTCPNoDelay(bool tcpNoDelay) 
    throw(SocketException)
{
  int flag = tcpNoDelay? 1 : 0;
  if (setsockopt(sockDesc, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(flag)) < 0) {
    throw SocketException("Setting TCP_NODELAY failed", true);
  }
}

bool CommunicatingSocket::getTCPNoDelay()
    throw(SocketException)
{
  int flag;
  socklen_t size = sizeof(int);  
  if (getsockopt(sockDesc, IPPROTO_TCP, TCP_NODELAY, &flag, &size) < 0) {
    throw SocketException("Getting TCP_NODELAY failed", true);
  }
  return flag == 1;  
}

int CommunicatingSocket::getTCPReceiveWindow()
    throw(SocketException)
{
    int sockbufsize = 0; 
    socklen_t size = sizeof(int); 
    if (getsockopt(sockDesc, SOL_SOCKET, SO_RCVBUF, (char *)&sockbufsize, &size) < 0) {
        throw SocketException("Getting SO_RCVBUF failed", true);
    }
    return sockbufsize;
}

void CommunicatingSocket::setTCPReceiveWindow(int wndSize)
    throw(SocketException)
{
    if (setsockopt(sockDesc, SOL_SOCKET, SO_RCVBUF, (char *)&wndSize, (int)sizeof(wndSize))) {
        throw SocketException("Setting SO_RCVBUF failed", true);
    }
}

int CommunicatingSocket::getTCPSendWindow()
    throw(SocketException)
{
    int sockbufsize = 0; 
    socklen_t size = sizeof(int); 
    if (getsockopt(sockDesc, SOL_SOCKET, SO_SNDBUF, (char *)&sockbufsize, &size) < 0) {
        throw SocketException("Getting SO_SNDBUF failed", true);
    }
    return sockbufsize;    
}

void CommunicatingSocket::setTCPSendWindow(int wndSize)
    throw(SocketException)
{
    if (setsockopt(sockDesc, SOL_SOCKET, SO_SNDBUF, (char *)&wndSize, (int)sizeof(wndSize))) {
        throw SocketException("Setting SO_SNDBUF failed", true);
    }
}

/**
 * position is the first element of valid data on the buffer,
 * which is copied to the begining of the buffer
 */
void CommunicatingSocket::compactBuffer(int position) {
    // Compact the buffer
    memmove(_buffer, _buffer+position, _bufTop-position);
    _bufTop = _bufTop-position;
}


int CommunicatingSocket::recvOutOfBand(char* buffer, int size)
  throw(SocketException) 
{
    // Out of band messages are independent of whether the socket is SSL or not.
    int read;
    if ((read = ::recv(sockDesc, &buffer, size,  MSG_DONTWAIT|MSG_OOB)) < 0) {
        // Ignore errors, just return 0.
        return 0;        
    } else {
        return read;
    }
}


int CommunicatingSocket::recv(void *buffer, int bufferLen) 
  throw(SocketException) 
{
    if (bufferLen <= _bufTop) {
        // We have enough data on the local buffer
        memcpy(buffer, _buffer, bufferLen);
        compactBuffer(bufferLen);
        return bufferLen;

    } else {
        // Not enough data on the local buffer. Copy what is available...
        memcpy(buffer, _buffer, _bufTop);
        // Save how many bytes we have copied
        int pos = _bufTop;
        // We have used all data available on the local buffer        
        _bufTop = 0;
        int read;
        if (ssl == NULL) {
            if ((read = ::read(sockDesc, (void *) (((char*)buffer)+pos), bufferLen-pos)) < 0) {
                DMESG("Recieve failed with " << errno << std::endl);
                throw SocketException("Received failed (read())", true);
            }
        } else {
            if ((read = ::SSL_read(ssl, (void *) (((char*)buffer)+pos), bufferLen-pos)) < 0) {
                throw SocketException("Received failed (SSL_read())", true);
            }
        }
        return pos + read;
    }
}

string Socket::changeAddress(struct sockaddr_storage *addr) {
  char address[INET6_ADDRSTRLEN+128];

  if (addr->ss_family == AF_INET) {
    struct sockaddr_in *s = (struct sockaddr_in *) addr;
    inet_ntop(AF_INET, &s->sin_addr, address, sizeof address);
  } else { // AF_INET6
    struct sockaddr_in6 *s = (struct sockaddr_in6 *)addr;
    inet_ntop(AF_INET6, &s->sin6_addr, address, sizeof address);
  }

  return address;
}

int Socket::changePort(struct sockaddr_storage *addr) {

  if (addr->ss_family == AF_INET) {
    struct sockaddr_in *s = (struct sockaddr_in *) addr;
    return ntohs(s->sin_port);
  } else { // AF_INET6
    struct sockaddr_in6 *s = (struct sockaddr_in6 *)addr;
    return ntohs(s->sin6_port);
  }
}


 
string CommunicatingSocket::getForeignAddress() 
    throw(SocketException) 
{
  struct sockaddr_storage addr;
  unsigned int addr_len = sizeof(addr);

  if (getpeername(sockDesc, (sockaddr *) &addr,(socklen_t *) &addr_len) < 0) {
    throw SocketException("Fetch of foreign address failed (getpeername())", true);
  }

  return changeAddress(&addr);
}

unsigned short CommunicatingSocket::getForeignPort() throw(SocketException) {
  struct sockaddr_storage addr;
  unsigned int addr_len = sizeof(addr);

  if (getpeername(sockDesc, (sockaddr *) &addr, (socklen_t *) &addr_len) < 0) {
    throw SocketException("Fetch of foreign port failed (getpeername())", true);
  }
  return changePort(&addr);
}


string CommunicatingSocket::getForeignName()
  throw(SocketException)
{
  if( foreignHostnames.size() )
    return *(foreignHostnames.begin());

  return "NO NAME PROVIDED";
}


//! Reads at least a line and removes the training "\n"
/** Reads a line from an SSL or normal network connection using a
    buffer to handle consecutive lines. At least a line is returned
    (blocks!) and the trainling "\n" removed.
*/
string CommunicatingSocket::readLine() throw(SocketException) {
    string line;
        
    while (true) {
        int index = 0;        
        while (index != _bufTop &&  _buffer[index] != '\n') {
            index++;
        }
        // Put a null terminator for the append method
        if (_buffer[index] == '\n' && (index > 0 && _buffer[index-1] == '\r')) {
            // Discard the \r
            _buffer[index-1] = 0;
        } else {
            _buffer[index] = 0;
        }
        line.append(_buffer);
            
        // Mark the end of the buffer
        // Either at the end of the buffer or found a '\n'
        if (index == _bufTop) {
            // At the end of the buffer. Must read again from the socket
            if (ssl == NULL) {
                if ((_bufTop = ::read(sockDesc, (void *) _buffer, BUF_SIZE-1)) < 0) {
                    throw SocketException("Received failed (read())", true);
                }
            } else {
                if ((_bufTop = ::SSL_read(ssl, (void *) _buffer, BUF_SIZE-1)) < 0) {
                    throw SocketException("Received failed (SSL_read())", true);
                }
            }
            if (_bufTop < 0) { // Close connection
                throw SocketException("Error reading", true);
            }
            if (_bufTop == 0) {
                throw SocketException("Connection reset by peer", true);
            }

        } else {
//            DMESG("in >" << line << "<" << std::endl);
            // Discard the null terminating character
            index++;
            // Compact the buffer
            compactBuffer(index);
            return line;
        }
    }
}

void CommunicatingSocket::send(const string &str)
    throw(SocketException) 
{
    CommunicatingSocket::send(str.c_str(), str.size());
}


// TCPSocket Code

TCPSocket::TCPSocket() throw(SocketException) 
    : CommunicatingSocket(SOCK_STREAM, IPPROTO_TCP) {
}

TCPSocket::TCPSocket(const string &foreignAddress, unsigned short foreignPort)
    throw(SocketException) : CommunicatingSocket(SOCK_STREAM, IPPROTO_TCP,foreignAddress) 
{

  connect(foreignAddress, foreignPort);
}


TCPSocket::TCPSocket(int newConnSD) : CommunicatingSocket(newConnSD) 
{

}


void CommunicatingSocket::initSSLServer(SSL_CTX *ctx, char *appData)
  throw (SocketException) 
{
  // see man BIO_f_ssl
  if(!(ssl = SSL_new(ctx))) {
    throw SocketException("Could not init SSL\n");
  }
  // Set application data (needed by cache callbacks)
  SSL_set_app_data(ssl, appData);
  
  if(!SSL_set_fd(ssl, sockDesc))
    throw SocketException( "Could not init SSL: " + sslErrors() );
  
  if(!SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY))
    throw SocketException( "Could not init SSL: " + sslErrors() );
    
  DMESG("SSL_accept" << std::endl);
  int ret=SSL_accept(ssl);
  if( ret<0 )
    throw SocketException( "Could not init SSL: " + sslErrors() );

  if( ret==0 && sslGetError(ret) <0 )
    throw SocketException( "Could not init SSL: " + sslErrors() );
}


void CommunicatingSocket::initSSLClient(SSL_CTX *ctx, SSL_SESSION *sessionData)
  throw (SocketException) 
{
  DMESG("Starting SSL client connection\n");

  // see man BIO_f_ssl
  if(!(ssl = SSL_new(ctx)))
    throw SocketException("Could not create ssl object");

  if(!SSL_set_fd(ssl, sockDesc))
    throw SocketException("Could not set socket in BIO: " + sslErrors() );

  if ( !sessionData ) { 
    if (!SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY)) {
      throw SocketException("Could not set SSL_MODE_AUTO_RETRY");
    }
  } else {
    DMESG("Resuming SSL session\n");
    if(!SSL_set_session(ssl, sessionData)) {
      throw SocketException("Could not set SSL session");
    }
  }

  int res=SSL_connect(ssl);

  DMESG("SSL_connect: " << res << std::endl);
  if(res<0)
    throw SocketException("Could not do SSL handshake: Fatal error" 
			  + sslErrors() );
  if( res==0 && sslGetError(res) <0 )
    throw SocketException("Could not do SSL handshake" + sslErrors() );
}


int CommunicatingSocket::sslGetError(int res)
{
  switch(SSL_get_error(ssl, res)) {
  case SSL_ERROR_NONE:
    DMESG("SSL_ERROR_NONE" << std::endl);
    return 0;
  case SSL_ERROR_ZERO_RETURN:
    ERRLOG("SSL_ERROR_ZERO_RETURN" << std::endl);
    return 0;
  case SSL_ERROR_WANT_READ:
    ERRLOG("SSL_ERROR_WANT_READ" << std::endl);
    return 0;
  case SSL_ERROR_WANT_WRITE:
    ERRLOG("SSL_ERROR_WANT_WRITE" << std::endl);
    return 0;
  case SSL_ERROR_WANT_CONNECT:
    ERRLOG("SSL_ERROR_WANT_CONNECT" << std::endl);
    return 0;
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  case SSL_ERROR_WANT_ACCEPT:
    ERRLOG("SSL_ERROR_WANT_ACCEPT" << std::endl);
    return 0;
#endif
  case SSL_ERROR_WANT_X509_LOOKUP:
    DMESG("SSL_ERROR_WANT_X509_LOOKUP" << std::endl);
    return 0;
  case SSL_ERROR_SYSCALL:
    DMESG("SSL_ERROR_SYSCALL" << std::endl);
    perror("System call failed");
    return 0;
  case SSL_ERROR_SSL:
    DMESG("SSL_ERROR_SSL" << std::endl);
    return -1;
  }
  return 0;
}


// TCPServerSocket Code
TCPServerSocket::TCPServerSocket(unsigned short localPort, int queueLen) 
  throw(SocketException) : Socket(SOCK_STREAM, IPPROTO_TCP) 
{
  setLocalPort(localPort);
  setListen(queueLen);
}


TCPServerSocket::TCPServerSocket(const string &localAddress, 
    unsigned short localPort, int queueLen) 
    throw(SocketException) : Socket(SOCK_STREAM, IPPROTO_TCP)
{
  setLocalAddressAndPort(localAddress, localPort);
  setListen(queueLen);
}


TCPSocket *TCPServerSocket::accept() throw(SocketException)
{
  int newConnSD;
  if ((newConnSD = ::accept(sockDesc, NULL, 0)) < 0) {
    throw SocketException("Accept failed (accept())", true);
  }

  return new TCPSocket(newConnSD);
}


void TCPServerSocket::setListen(int queueLen) throw(SocketException) {
  if (listen(sockDesc, queueLen) < 0) {
    throw SocketException("Set listening socket failed (listen())", true);
  }
}


// UDPSocket Code

UDPSocket::UDPSocket() throw(SocketException) : CommunicatingSocket(SOCK_DGRAM,
    IPPROTO_UDP) {
  setBroadcast();
}


UDPSocket::UDPSocket(unsigned short localPort)  throw(SocketException) : 
    CommunicatingSocket(SOCK_DGRAM, IPPROTO_UDP) 
{
  setLocalPort(localPort);
  setBroadcast();
}


UDPSocket::UDPSocket(const string &localAddress, unsigned short localPort) 
     throw(SocketException) : CommunicatingSocket(SOCK_DGRAM, IPPROTO_UDP) 
{
  setLocalAddressAndPort(localAddress, localPort);
  setBroadcast();
}


void UDPSocket::setBroadcast()
{
  // If this fails, we'll hear about it when we try to send.  This will allow 
  // system that cannot broadcast to continue if they don't plan to broadcast
  int broadcastPermission = 1;
  setsockopt(sockDesc, SOL_SOCKET, SO_BROADCAST, 
             (raw_type *) &broadcastPermission, sizeof(broadcastPermission));
}


void UDPSocket::disconnect() throw(SocketException) 
{
  sockaddr_in6 nullAddr;
  memset(&nullAddr, 0, sizeof(nullAddr));
  nullAddr.sin6_family = AF_UNSPEC;

  // Try to disconnect
  if (::connect(sockDesc, (sockaddr *) &nullAddr, sizeof(nullAddr)) < 0) {
   #ifdef WIN32
    if (errno != WSAEAFNOSUPPORT) {
   #else
    if (errno != EAFNOSUPPORT) {
   #endif
      throw SocketException("Disconnect failed (connect())", true);
    }
  }
}

void UDPSocket::sendTo(const void *buffer, int bufferLen, 
    const string &foreignAddress, unsigned short foreignPort) 
    throw(SocketException) 
{
  struct sockaddr_storage  destAddr;
  std::list< std::string > names;
  fillAddr(foreignAddress, foreignPort, destAddr, names);

  // Write out the whole buffer as a single message.
  if (sendto(sockDesc, (raw_type *) buffer, bufferLen, 0,
             (sockaddr *) &destAddr, sizeof(destAddr)) != bufferLen) {
    throw SocketException("Send failed (sendto())", true);
  }
}

int UDPSocket::recvFrom(void *buffer, int bufferLen, string &sourceAddress,
    unsigned short &sourcePort) throw(SocketException) 
{
  struct sockaddr_storage clntAddr;
  socklen_t addrLen = sizeof(clntAddr);
  int rtn;
  char address[INET6_ADDRSTRLEN+128];

  if ((rtn = recvfrom(sockDesc, (raw_type *) buffer, bufferLen, 0, 
                      (sockaddr *) &clntAddr, (socklen_t *) &addrLen)) < 0) {
    throw SocketException("Send failed (sendto())", true);
  }
  
  sourceAddress = changeAddress(&clntAddr);
  sourcePort = changePort(&clntAddr);

  return rtn;
}

void UDPSocket::setMulticastTTL(unsigned char multicastTTL) throw(SocketException) {
  if (setsockopt(sockDesc, IPPROTO_IP, IP_MULTICAST_TTL, 
                 (raw_type *) &multicastTTL, sizeof(multicastTTL)) < 0) {
    throw SocketException("Multicast TTL set failed (setsockopt())", true);
  }
}

void UDPSocket::joinGroup(const string &multicastGroup) throw(SocketException) 
{
  struct ipv6_mreq multicastRequest;

  if( inet_pton(AF_INET, multicastGroup.c_str(), &multicastRequest.ipv6mr_multiaddr) <= 0)
    throw SocketException("Multicast group join failed (setsockopt())", true);

  multicastRequest.ipv6mr_interface = 0;

  if (setsockopt(sockDesc, IPPROTO_IP, IP_ADD_MEMBERSHIP, 
                 (raw_type *) &multicastRequest, 
                 sizeof(multicastRequest)) < 0) {
    throw SocketException("Multicast group join failed (setsockopt())", true);
  }
}

void UDPSocket::leaveGroup(const string &multicastGroup) throw(SocketException) 
{
  struct ipv6_mreq multicastRequest;
  if( inet_pton(AF_INET, multicastGroup.c_str(), &multicastRequest.ipv6mr_multiaddr) <= 0)
    throw SocketException("Multicast group join failed (setsockopt())", true);

  multicastRequest.ipv6mr_interface = 0;

  if (setsockopt(sockDesc, IPPROTO_IP, IP_DROP_MEMBERSHIP, 
                 (raw_type *) &multicastRequest, 
                 sizeof(multicastRequest)) < 0) {
    throw SocketException("Multicast group leave failed (setsockopt())", true);
  }
}


std::string CommunicatingSocket::getPeerDN() const
{
  
  X509 *peer;
  char peerCN[256];

  peer=SSL_get_peer_certificate(ssl);
  X509_NAME_get_text_by_NID (X509_get_subject_name(peer),
			     NID_commonName, peerCN, 256);

  return peerCN;
}


int CommunicatingSocket::verifyServerCertificate(bool ignoreNameMismatch)
{
  if(SSL_get_verify_result(ssl)!=X509_V_OK)
    return -1;
  
  /*Check the cert chain. The chain length
    is automatically checked by OpenSSL when
    we set the verify depth in the ctx */
  
  if( ignoreNameMismatch )
    return 0;

  /*Check the common name*/
  std::string peerCN = getPeerDN();

  std::list< std::string >::const_iterator I = foreignHostnames.begin();
  while( I!= foreignHostnames.end() ){
    if( !strcasecmp(peerCN.c_str(), I->c_str()) )
      return 0;
    I++;
  }

  return -1;
}

void CommunicatingSocket::setRcvTimeOut(struct timeval tv) 
    throw(SocketException)
{
    if (setsockopt(sockDesc, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv))) {
        throw SocketException("Setting SO_RCVTIMEO failed", true);
    }
}
