/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Statement.h"

#include "DatabaseConnection.h"
#include "debug.h"
#include "util/Utils.h"

#include <sqlext.h>
#include <sqltypes.h>
#include <pthread.h>
#include <ostream>
#include <iostream>
#include <sstream>
#include <boost/algorithm/string.hpp>

using namespace std;

SQLINTEGER Statement::dummy_SQL_NTS = SQL_NTS;

Statement::Statement(DatabaseConnection &dbConnection, bool debugFlag)
  : debug(debugFlag), dbConn(dbConnection), transactionCount(0),
    mySQLFlag(false), isOracle(false), detached(false)
{
  // If the database connection already has a statement,
  // use that one.
  if (dbConn._statement == NULL) {
    dbConn.createStatement();
    ownStatement = true;
  } else {
    DMESG("Reusing statement, transaction count is " << transactionCount << "\n");
    ownStatement = false;
  }
  // Otherwise, create a new statement.

  //  startAutoCommit();
  DatabaseConnection::DriverType dType = dbConn.getDriverType();
  if (dType == DatabaseConnection::dtMySQL || dType == DatabaseConnection::dtMySQL5) {
    DMESG("Activating MySQL quotes\n");
    mySQLFlag = true;
  } else if (dType == DatabaseConnection::dtOracle) {
    DMESG("Detected Oracle\n");
    isOracle = true;
  }
}


Statement::~Statement()
{
  DMESG("~Statement\n");
  // Release bound variables and close the cursor. This does not
  // kill the transaction. 
  close();

  if (detached) {
  	DMESG("Detached\n");
  	// If this statement instance was detached from 
  	// the DatabaseConnection, release it. 
  	return;
  }

  if (ownStatement) {
  	DMESG(" Owns statement\n");
    // If we own the statement, we should close it.
    if(dbConn._transactionCount > 0) 
      rollbackTransaction();
    DMESG(" Deleting statement\n");
    dbConn.closeStatement();
  } else {
    DMESG(" Statement not owned. TransactionCount: " << transactionCount << "\n");
    // We don't own the statement. But if we opened a transaction,
    // we should roll it back.
    if (transactionCount > 0) {
	  rollbackTransaction();
	  dbConn.closeStatement();
	}
  }
}

SQLHSTMT Statement::getStatement() const { 
	return dbConn._statement; 
}

void Statement::quoteDataField(std::string &field) {
  // Must escape the ', " and \ chars or else they are interpreted as
  // escape sequences by the ODBC driver. It works on Postgres and Oracle.
  // Not sure about MySQL.
  if (isOracle) {
    // Only the ' needs quoting. 
    boost::replace_all(field, "'", "''");
  } else {
    boost::replace_all(field, "\\", "\\\\");
    boost::replace_all(field, "\"", "\\\"");  
    boost::replace_all(field, "'", "\\'");
  }
}

// Detach and attach the statement of the underlying
// DatabaseConnection -> Allow external transaction management!
int Statement::detach()
{
//  if( !ownStatement )
//    return -1;
    
  DMESG("Detaching\n");
  detached = true;
  return 0;
}

int Statement::attach()
{
//  // Cannot attached to a statement that we don't own
//  if( !ownStatement )
//    return -1;
  detached = false;
  DMESG("Ataching\n");
  return 0;
}


int Statement::exec(const std::string &query)
{
  if(!dbConn._statement)
    return 1;

  std::string q(query);
  if(mySQLFlag)
    replaceSQLQuotes(q);
  currentQuery = q;
  
  DMESG("Query (" << transactionCount << "): " << query << std::endl);

  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_testcancel();
  error=SQLExecDirect(dbConn._statement, (SQLCHAR *)q.c_str(), SQL_NTS);
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  updateLastQueryResult();

  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    // SQL_NO_DATA means only that the query returned no rows, 
    // it is not really an error, so don't flag it as such
    if (error != SQL_NO_DATA) {      
      dbConn.increaseErrorsDuringTransaction();
    }
    return 2;
  }
  return 0;
}


int Statement::prepare(const std::string &query)
{

  if(!dbConn._statement)
    return 1;

  std::string q(query);
  if(mySQLFlag)
    replaceSQLQuotes(q);

  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  error=SQLPrepare(dbConn._statement, (SQLCHAR *)q.c_str(), SQL_NTS);
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 2;
  }
  statementPrepared = true;
  return 0;
}


int Statement::prepare()
{
  return prepare(currentQuery);
}


int Statement::setParamArraySize(SQLINTEGER bulkSize) {
  // Specify the number of elements in each parameter array.
  error = SQLSetStmtAttr(dbConn._statement, SQL_ATTR_PARAMSET_SIZE,
			 (SQLPOINTER)bulkSize, 0);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    return 1;
  }
  return 0;
}

int Statement::getBulkInsertSize() {
  SQLINTEGER value;
  if (getStmtAttr(SQL_ATTR_PARAMSET_SIZE, value)) {
    return -1;
  } else {
    return value;
  }
}

int Statement::getStmtAttr(SQLINTEGER attr, SQLINTEGER &value) {
  // Warning, in some cases you must provide 64 bit integers for the
  // value passed back, see http://support.microsoft.com/kb/298678
  error = SQLGetStmtAttr (dbConn._statement, attr, &value, SQL_IS_INTEGER, NULL);
  updateLastQueryResult();
  if(error != SQL_SUCCESS && error != SQL_SUCCESS_WITH_INFO) {
    DMESG("Error getting attr " << attr << endl);
    return -1;
  }
  return 0;
}

int Statement::prepareBulkInsert(int bulkSize) {
  // Most of the drivers do not support bulk insertion with parameter arrays.
  // So we test it before.
  if (setParamArraySize(bulkSize)) {
    return -1;
  }
  
  // Set the SQL_ATTR_PARAM_BIND_TYPE statement attribute to use
  // column-wise binding.
  //	DMESG(" SQL_ATTR_PARAM_BIND_TYPE\n");
  error = SQLSetStmtAttr(dbConn._statement, SQL_ATTR_PARAM_BIND_TYPE, SQL_PARAM_BIND_BY_COLUMN, 0);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    return -1;
  }
  
  
  static SQLUSMALLINT paramsProcessed, paramStatusArray[100]; 	
  // Specify an array in which to return the status of each set of
  // parameters.
  //	DMESG(" SQL_ATTR_PARAM_STATUS_PTR\n");
  error = SQLSetStmtAttr(dbConn._statement, SQL_ATTR_PARAM_STATUS_PTR, paramStatusArray, 0);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    return -1;
  }
  
  // Specify an SQLUINTEGER value in which to return the number of sets of
  // parameters processed.
  //	DMESG(" SQL_ATTR_PARAMS_PROCESSED_PTR\n");
  error = SQLSetStmtAttr(dbConn._statement, SQL_ATTR_PARAMS_PROCESSED_PTR, &paramsProcessed, 0);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    return -1;
  }
  return 0;
}


int Statement::execPrepared()
{
  if(!dbConn._statement)
    return 1;

//  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  error=SQLExecute(dbConn._statement);
//  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    if (error != SQL_NO_DATA)
      dbConn.increaseErrorsDuringTransaction();
    return 2;
  }
  return 0;
}


/// Return information about the columns of table
int Statement::columns(const std::string &table)
{
  if(!dbConn._statement)
    return 1;

  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  error=SQLColumns(dbConn._statement,
		   NULL, 0, // All catalogs
		   NULL, 0, // All schemas
		   (SQLCHAR *)table.c_str(), SQL_NTS,
		   NULL, 0); // All columns
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    return 2;
  }
  return 0;
}


/// Return information about tables matching table
int Statement::tables(const std::string &table, const std::string &schema)
{
  if(!dbConn._statement)
    return 1;

  std::string myTable = table;
  if (isOracle) {
    for(size_t i = 0; i < myTable.size(); i++)
      myTable[i] = toupper(myTable[i]);
  }

  std::string mySchema = schema;
  if (isOracle) {
    for(size_t i = 0; i < mySchema.size(); i++)
      mySchema[i] = toupper(mySchema[i]);
  }

  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  /* get table information */
  error=SQLTables(dbConn._statement,
		  NULL, 0, // All catalogs
		  // [nsantos] Using NULL to search for all schemas was failing
		  // to find tables on the _arda_md schema. With % it works fine.
		  (SQLCHAR *)mySchema.c_str(), SQL_NTS,
		  (SQLCHAR *)myTable.c_str(), SQL_NTS,
		  NULL, 0); // All types  
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO))
    return 2;
  return 0;
}

/**
 * @throws SQLException on error. 
 * return 0 on success, 1 when there is no more data to return
 */
int Statement::fetch()
  throw (SQLException)
{
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  error=SQLFetch(dbConn._statement);
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    // SQL_NO_DATA is NOT an error. This was aborting the
    // uploads with replication.
    if (error == SQL_NO_DATA) {
    	return 1;
    } else {
      	LOG("Statement.fetch() error: " << errorMsg << endl);
      	dbConn.increaseErrorsDuringTransaction();
      	throw SQLException(*this);
    }
  }

  std::map<int, BoundColumn *>::iterator I;
  for(I=boundColumns.begin(); I!=boundColumns.end(); I++){
    if(I->second->type == SQL_C_CHAR){
      if(I->second->length == SQL_NULL_DATA){
	DMESG("Setting string to 0\n");
	((char *)(I->second->buffer))[0]='\0';
	continue;
      }
      if(I->second->length == SQL_NO_TOTAL){ // NOT: SQL_NO_DATA !!!!
	DMESG("String too long: Setting string to 0\n");
	((char *)(I->second->buffer))[0]='\0';
	continue;
      }
      ((char *)(I->second->buffer))[I->second->length]='\0';
    }
  }
  return 0;
}


int Statement::getData(int column, uint64_t &data) {
  string tmp;
  int err;
  if ((err = getData(column, tmp)) != 0) {
    return err;
  } else {
    data = toUint64(tmp);
    return 0;
  }
}

int Statement::getData(int column, int64_t &data) {
  string tmp;
  int err;
  if ((err = getData(column, tmp)) != 0) {
    return err;
  } else {
    data = toInt64(tmp);
    return 0;
  }
}

int Statement::getData(int column, std::string &data) 
{
  // The buffer must be larger than any numerical data type read,
  // otherwise the driver complains with: "Numeric value out of range"  
  char buff[512];
  data = "";
  // Either the size or the result
  SQLLEN bufInd;
  while (true) {
    // Returns SQL_NO_DATA when done
    error = SQLGetData(dbConn._statement, column, SQL_C_CHAR, 
        (SQLPOINTER)buff, sizeof(buff), &bufInd);
    updateLastQueryResult();
    if (!(error == SQL_SUCCESS_WITH_INFO || error == SQL_SUCCESS)) {
      break;
    }
    if (bufInd == SQL_NULL_DATA) {
    	DMESG("SQL_NULL_DATA\n");
    	error = SQL_NULL_DATA;
      return 1;
    }    
    data += buff;
  }
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO) && (error != SQL_NO_DATA)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }
  return 0;
}

SQLINTEGER Statement::getColumnLen(int column)
{
  std::map<int, BoundColumn *>::iterator I = boundColumns.find(column);
  if(I==boundColumns.end())
    return SQL_NULL_DATA;
  return I->second->length;
}


int Statement::bind(int column, char *buffer, int size)
{
  DMESG("Binding " << column << " as string of size " << size << std::endl);

  if(!dbConn._statement)
    return 1;

  if(boundColumns.find(column)!=boundColumns.end())
    return 2;

  BoundColumn *b=new BoundColumn();
  b->length = SQL_NTS;
  b->type = SQL_C_CHAR;
  b->buffer = buffer;
  b->bufferSize = size;
  boundColumns[column] = b;

  error=SQLBindCol(dbConn._statement, column, SQL_C_CHAR, buffer, size, &b->length);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }

  return 0;
}


int Statement::bind(int column, unsigned char *buffer, int size)
{
  DMESG("Binding " << column << " as raw of size " << size << std::endl);

  if(!dbConn._statement)
    return 1;

  if(boundColumns.find(column)!=boundColumns.end())
    return 2;

  BoundColumn *b=new BoundColumn();
  b->length = size;
  b->type = SQL_C_BINARY;
  b->buffer = buffer;
  b->bufferSize = size;
  boundColumns[column] = b;

  error=SQLBindCol(dbConn._statement, column, SQL_C_BINARY, buffer, size, &b->length);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }

  return 0;

}


int Statement::bind(int column, int64_t *integer)
{
  if(!dbConn._statement)
    return 1;
  if(boundColumns.find(column)!=boundColumns.end())
    return 2;

  BoundColumn *b=new BoundColumn();
  b->length = SQL_NTS;
  b->type = SQL_C_SBIGINT;
  b->buffer = integer;
  b->bufferSize = 8;
  boundColumns[column] = b;

  error=SQLBindCol(dbConn._statement, column, SQL_C_SBIGINT, integer, 8, &b->length);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }

  return 0;
}


int Statement::bind(int column, uint64_t *integer)
{
  if(!dbConn._statement)
    return 1;
  if(boundColumns.find(column)!=boundColumns.end())
    return 2;

  BoundColumn *b=new BoundColumn();
  b->length = SQL_NTS;
  b->type = SQL_C_UBIGINT;
  b->buffer = integer;
  b->bufferSize = 8;
  boundColumns[column] = b;

  error=SQLBindCol(dbConn._statement, column, SQL_C_UBIGINT, integer, 8, &b->length);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }

  return 0;
}


int Statement::bind(int column, int *integer)
{
  if(!dbConn._statement)
    return 1;
  if(boundColumns.find(column)!=boundColumns.end())
    return 2;

  BoundColumn *b=new BoundColumn();
  b->length = SQL_NTS;
  b->type = (sizeof(int)==4)?SQL_C_SLONG:SQL_C_SBIGINT;
  b->buffer = integer;
  b->bufferSize = sizeof(int);
  boundColumns[column] = b;

  error=SQLBindCol(dbConn._statement, column, b->type, integer, sizeof(int), &b->length);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }

  return 0;
}


int Statement::bind(int column, short int *integer)
{
  if(!dbConn._statement)
    return 1;
  if(boundColumns.find(column)!=boundColumns.end())
    return 2;

  BoundColumn *b=new BoundColumn();
  b->length = SQL_NTS;
  b->type = SQL_C_SSHORT;
  b->buffer = integer;
  b->bufferSize = sizeof(short int);
  boundColumns[column] = b;
  
  error=SQLBindCol(dbConn._statement, column, b->type, integer, sizeof(short int), &b->length);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }
  return 0;
}


int Statement::bindParam(unsigned int param, SQLINTEGER *value, SQLLEN *type)
{
  if(!dbConn._statement)
    return 1;
  SQLBindParameter(dbConn._statement, param, SQL_PARAM_INPUT,
		   SQL_C_ULONG, SQL_INTEGER, 10, 0, value, 0, type);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }

  return 0;
}
// SQL_BINARY

int Statement::bindParam(unsigned int param, const char *buffer,
			 int size, SQLLEN *type)
{
  if(!dbConn._statement)
    return 1;

  error=SQLBindParameter(dbConn._statement, param, SQL_PARAM_INPUT,
			 SQL_C_CHAR, SQL_CHAR, size, 0, (SQLCHAR *) buffer, size, type);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }
  return 0;
}


int Statement::bindBinaryParam(unsigned int param, unsigned char *buffer,
			       SQLINTEGER size, SQLLEN *resultSize)
{
  if(!dbConn._statement)
    return 1;

  error=SQLBindParameter(dbConn._statement, param, SQL_PARAM_INPUT,
			 SQL_C_BINARY, SQL_BINARY, size, 0, (SQLCHAR *) buffer, size,
			 resultSize);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    dbConn.increaseErrorsDuringTransaction();
    return 1;
  }
  return 0;
}


int Statement::unBind()
{
  if(!dbConn._statement)
    return 1;
  // Releases the column parameters bound by SQLBindCol
  error=SQLFreeStmt(dbConn._statement, SQL_UNBIND);
  updateLastQueryResult();
  std::map<int, BoundColumn *>::iterator I;
  for(I=boundColumns.begin(); I!=boundColumns.end(); I++)
    delete I->second;
  boundColumns.clear();
  return 0;
}


int Statement::close()
{
  if(!dbConn._statement)
    return 1;

  // Must call unBind to properly free the statement
  // before calling SQL_CLOSE. The process was crashing when it tried
  // to unbind all the columns after having called only SQL_CLOSE.
  unBind();
  // Releases the input parameters bound by SQLBindParameter
  error=SQLFreeStmt(dbConn._statement, SQL_RESET_PARAMS);
  updateLastQueryResult();
  // Closes the cursor if one is open
  error=SQLFreeStmt(dbConn._statement, SQL_CLOSE);
  updateLastQueryResult();
  return 0;
}

int Statement::closeCursor()
{
  if(!dbConn._statement)
    return 1;
    
  // Closes the cursor if one is open
  error=SQLFreeStmt(dbConn._statement, SQL_CLOSE);
  updateLastQueryResult();
  return 0;
}

std::string Statement::getError() {
    return errorMsg;
}

SQLRETURN Statement::getErrorCode() {
    return error;
}

std::string Statement::getSqlState() {
    return sqlState;
}

bool Statement::checkAlive()
{

  SQLINTEGER ConnectTest;
  int nReturn =0;

  if (dbConn.connAlive == false ) return false;

  SQLRETURN  RetValue = SQLGetConnectAttr(dbConn._cHandle, SQL_ATTR_CONNECTION_DEAD, &ConnectTest, 0, NULL);

  DMESG("Check aliveness\n");
  if ( RetValue == SQL_SUCCESS ) {
     DMESG("Operation Successful \n");
     if(ConnectTest == SQL_CD_TRUE)
       DMESG("It is dead now\n");
       return false;
  }
  DMESG("It is alive " << ConnectTest << "\n");
  return true;
}


void Statement::updateLastQueryResult()
{
  if (error == SQL_SUCCESS) {
//    cout << "SQL_SUCCESS\n";
    // Reset the error messages
    errorMsg = "";
    sqlState = "";
    
  } else if (error == SQL_SUCCESS_WITH_INFO || error == SQL_ERROR) {
//    cout << "SQL_SUCCESS_WITH_INFO || error == SQL_ERROR\n";
  	// Diagnostic messages are generated only when the error is one
  	// of these two.
    SQLINTEGER  sqlErr;
    char        sqlErrMsg[256];
    char        sqlStat[10];
    SQLSMALLINT len;

    SQLRETURN ret = SQLGetDiagRec(SQL_HANDLE_STMT, dbConn._statement, 1, (SQLCHAR *)sqlStat,
                        &sqlErr, (SQLCHAR *)sqlErrMsg, 255, &len);
    if(ret==SQL_SUCCESS){
      sqlErrMsg[len]='\0';
    } else {
      sqlErrMsg[0]='\0';
    }
    sqlState = sqlStat;

    // Remove any linefeeds
    int i=0;
    while(sqlErrMsg[i])
      sqlErrMsg[i++] = sqlErrMsg[i]=='\n'? ' ' : sqlErrMsg[i];
      
    errorMsg=sqlErrMsg;
    errorMsg += " in query: ";
    errorMsg += currentQuery;
  
    DMESG(sqlErrMsg << "  : " << sqlStat << std::endl);    

    if ( (error == SQL_ERROR) && (checkAlive() == false) ) {
      dbConn.connAlive = false;
    } 
    
  } else {
    ostringstream oss;
    oss << "Other error: " << error << "(";
    switch (error) {
    case SQL_NO_DATA:
      oss << "SQL_NO_DATA";
      break;
    case SQL_NEED_DATA:
      oss << "SQL_NEED_DATA";
      break;  		
    default:
      oss << "UNKNOWN";
      break;
    }
    oss << ")";
    errorMsg = oss.str();
    DMESG(errorMsg << endl);
    sqlState = "";
  }
}


/**
 * Returns the number of rows affected by an UPDATE, INSERT, 
 * or DELETE statement; an SQL_ADD, SQL_UPDATE_BY_BOOKMARK, or 
 * SQL_DELETE_BY_BOOKMARK operation in SQLBulkOperations; or an 
 * SQL_UPDATE or SQL_DELETE operation in SQLSetPos.
 */
int Statement::numRows()
{
  SQLLEN rows;
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  error=SQLRowCount(dbConn._statement, &rows);
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
  updateLastQueryResult();
  if ((error != SQL_SUCCESS) && (error != SQL_SUCCESS_WITH_INFO)) {
    return -1;
  }
  return rows;
}


/// Replace SQL92 standard column name quotes by MySQL quotes
void Statement::replaceSQLQuotes(std::string &query)
{
  bool quoted = false;

  for (size_t i=0; i<query.size(); i++) {
    if( query[i] == '\'' ){
      if(!i || (query[i-1] != '\\') || ( (i-1) && ( query[i-2] =='\\') ) )
	quoted = !quoted;
    }
    if( query[i] == '\"' && !quoted ){
      if(!i || (query[i-1] != '\\') || ( (i-1) && ( query[i-2] =='\\') ) )
	query[i] = '`';
    }
  }

  DMESG("Quoted query: >" << query << "<\n");
}

// Transaction handling is delegated to the DBConnection instance.
int Statement::beginTransaction(bool exclusive) 
{
  try {
    dbConn.beginTransaction(exclusive);
    transactionCount++;
    return 0;
  } catch (SQLException &ex) {
    DMESG(ex.what() << "\n");
    errorMsg = ex.what();
    error = 1;
    return 1;
  }
}

int Statement::commitTransaction() {
    try {
        dbConn.commitTransaction();
        transactionCount--;
        return 0;
    } catch (SQLException &ex) {
        DMESG(ex.what() << "\n");
        errorMsg = ex.what();
        error = 1;
        return 1;
    }
}

int Statement::rollbackTransaction() {
    try {
        dbConn.rollbackTransaction();
        return 0;
    } catch (SQLException &ex) {
        DMESG(ex.what() << "\n");
        errorMsg = ex.what();
        error = 1;
        return 1;
    }
}
  
//void Statement::setTransactionId(uint64_t tid) {
//  dbConn._clientTransactionID = tid;
//}
//
//// Use the address of the database 
//// connection as the transaction id.
//uint64_t Statement::getClientTransactionId() const {
//  return dbConn._clientTransactionID;
//}

uint64_t Statement::sequenceGetNextValue(const std::string &sequence) 
throw (SQLException)
{
  string query;
  switch (dbConn.getDriverType()) {
  case DatabaseConnection::dtPSQL :
    query = "SELECT nextval('" + sequence + "');";
    break;
  case DatabaseConnection::dtOracle :
    query = "SELECT " + sequence + ".nextval FROM dual;";
    break;
  case DatabaseConnection::dtMySQL5 :
    query = "SELECT seqnextval('" + sequence +"');";
    break;
  default:
    throw SQLException("Not supported");
  }
  if (exec(query)){
    throw SQLException(*this);
  }  
  if(fetch()){
    throw SQLException(*this);
  }
  uint64_t v;
  getData(1, v);
  return v;
}


uint64_t Statement::sequenceGetCurrentValue(const std::string &sequence) 
throw (SQLException) 
{
  string query;
  switch (dbConn.getDriverType()) {
  case DatabaseConnection::dtPSQL :
    query = "SELECT last_value FROM " + sequence + ";";
    // query = "SELECT currval('" + sequence + "');";
    break;
  case DatabaseConnection::dtOracle :
    query = "SELECT " + sequence + ".currval FROM dual;";
    break;
  case DatabaseConnection::dtMySQL5 :
    query = "SELECT last_val FROM usersequences WHERE name = '" 
      + sequence + "';";
    break;
  default:
    throw SQLException("Not supported");
  }
  if (exec(query)){
    throw SQLException(*this);
  }  
  if(fetch()){
    throw SQLException(*this);
  }  
  uint64_t v;
  getData(1, v);
  closeCursor();
  return v;
}
