/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: Statement.h,v 1.26 2010/05/14 06:19:12 asunil Exp $
//
#ifndef STATEMENT_H
#define STATEMENT_H

#include "util/DBUtil.h"

#include <sql.h>

#include <string>
#include <vector>
#include <map>
#include <stdint.h>

class DatabaseConnection;

class Statement {
 public:
  Statement(DatabaseConnection &dbConn, bool debugFlag = false);
  ~Statement();
  int exec(const std::string &query);
  int prepare(const std::string &query);
  int prepare();

  int getBulkInsertSize();
  int getStmtAttr(SQLINTEGER attr, SQLINTEGER &value);     
  int setParamArraySize(SQLINTEGER bulkSize);
  int prepareBulkInsert(int bulkSize);
  
  int execPrepared();
  int columns(const std::string &table);
  int tables(const std::string &table, const std::string &schema="%");
  int fetch() throw (SQLException);
  /**
   * Reads all the data retrieved by a fetch from a column. Uses
   * a dynamic buffer. Useful for columns with large or unknown 
   * sizes.
   */
  int getData(int column, std::string &data);
  int getData(int column, uint64_t &data);
  int getData(int column, int64_t &data);
  
  int bind(int column, char *buffer, int size);           // String
  int bind(int column, unsigned char *buffer, int size);  // Binary
  int bind(int column, int *integer);
  int bind(int column, int64_t *integer);
  int bind(int column, uint64_t *integer);
  int bind(int column, short int *integer);

  int bindParam(unsigned int param, const char *buffer, int size, SQLLEN *type);
  int bindParam(unsigned int param, SQLINTEGER *value, SQLLEN *type);
  int bindBinaryParam(unsigned int param, unsigned char *buffer,
		      SQLINTEGER size, SQLLEN *resultSize);

//  int bindParam(unsigned int param, char *buffer, int size);

  SQLHSTMT getStatement() const;
  DatabaseConnection &getDatabaseConnection() const { return dbConn; };

  std::string getError();
  SQLRETURN getErrorCode();
  std::string getSqlState();

  int unBind();
  int close();
  int closeCursor();
  
  // Transaction handling. Deletegated to DatabaseConnection
  int commitTransaction();
  int beginTransaction(bool exclusive=false);
  int rollbackTransaction();

//  void setTransactionId(uint64_t tid);
//  uint64_t getClientTransactionId() const;
//  bool isSaveLogsForTransaction() const;

  int numRows();
  SQLINTEGER getColumnLen(int column);

  bool isMySQL() const { return mySQLFlag; };

  // Detach and attach the statement of the underlying
  // DatabaseConnection -> Allow external transaction management!
  int detach();
  int attach();
  
  void quoteDataField(std::string &field);
  
  uint64_t sequenceGetCurrentValue(const std::string &sequence) 
    throw (SQLException);
  uint64_t sequenceGetNextValue(const std::string &sequence) 
    throw (SQLException);
    
  void querySet(const std::string &query) {
    currentQuery = query;
    statementPrepared = false;
  }

  bool debug;

  bool checkAlive();

private:
 
  void updateLastQueryResult();
  
  void replaceSQLQuotes(std::string &query);

  class BoundColumn {
  public:
    SQLINTEGER type;
    SQLLEN     length;
    void       *buffer;
    SQLLEN     bufferSize;
  };

  void resizeColumns(int column);

  DatabaseConnection &dbConn;
  bool ownStatement;
  int transactionCount;

  static SQLINTEGER dummy_SQL_NTS;
  std::map<int, BoundColumn *> boundColumns;
  bool mySQLFlag;
  bool isOracle;
  std::string currentQuery;
  
  // Updated after every operation.
  std::string   errorMsg;
  SQLRETURN     error;
  std::string   sqlState;
  
  bool detached;
  bool statementPrepared;
};

#endif // STATEMENT_H
