/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#include "StatsCollector.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "debug.h"
#include "util/Utils.h"

static bool debug = false;

#include <stdexcept>
#include <sstream>

using namespace std;
using namespace boost;

// Static members definition
map<string, int64_t> StatsCollector::counters;
std::set<StatsSource*> StatsCollector::sources;
boost::mutex StatsCollector::sourcesLock;
boost::mutex StatsCollector::countersLock;


/*=====================================
 * Update methods
 =====================================*/
void StatsCollector::setCounter(const std::string &name, int64_t value) {
  mutex::scoped_lock lk(countersLock);
  DMESG("Setting counter: " << name << ":" << counters[name] << endl);
  counters[name] = value;
}

void StatsCollector::dropCounter(const std::string &name) {    
  mutex::scoped_lock lk(countersLock);
  counters.erase(name);
}

int StatsCollector::incCounter(const std::string &name) 
throw (runtime_error) 
{
  mutex::scoped_lock lk(countersLock);  
  checkCounterExists(name);
  return ++(counters[name]);
}

int StatsCollector::decCounter(const std::string &name) 
throw (std::runtime_error)
{
  mutex::scoped_lock lk(countersLock);
  checkCounterExists(name);
  return --(counters[name]);
}

int StatsCollector::addToCounter(const std::string &name, int64_t value) 
throw (std::runtime_error) 
{
  mutex::scoped_lock lk(countersLock);
  checkCounterExists(name);
  return (counters[name] += value);
}

/*=====================================
 * Query methods
 =====================================*/
bool StatsCollector::hasCounter(const std::string &name) {
  mutex::scoped_lock lk(countersLock);
  return counters.count(name) != 0;
}  
  
int64_t StatsCollector::getCounter(const std::string &name) 
throw (std::runtime_error)
{
  mutex::scoped_lock lk(countersLock);  
  checkCounterExists(name);
  return counters[name];
}

const std::map<std::string, int64_t> &StatsCollector::getAllCounters() {  
  return counters;
}

void StatsCollector::collectAllCounters(map<std::string, string> &auxCounters) {
  {
    mutex::scoped_lock lk(countersLock);
    for (map<string, int64_t>::const_iterator it = counters.begin();
      it != counters.end();
      ++it)
    {
  //    cout << it->first << ":" << it->second << "-" << toString(it->second) << endl;
      auxCounters[it->first] = toString(it->second);
    }
  }

  // Source specific counters
  mutex::scoped_lock lk(sourcesLock);
  for (set<StatsSource*>::const_iterator it = sources.begin();
    it != sources.end();
    ++it) 
  {
    (*it)->collectCounters(auxCounters);
  }    
}

void StatsCollector::dumpAsXml(std::ostringstream &out) {
  map<std::string, string> auxCounters;
  collectAllCounters(auxCounters);
  out << "<Service>\n"
      << "  <Name>AMGA</Name>\n"
      << "  <Version>" << VERSION << "</Version>\n"
      << "  <Data>\n";
  for (map<string, string>::const_iterator it = auxCounters.begin();
    it != auxCounters.end();
    ++it)
  {
//    cout << it->first << ":" << it->second << endl;
    out << "    <Key>" << it->first << "</Key>\n"
        << "    <Value>" << it->second << "</Value>\n";
  }
  out << "  </Data>\n"
      << "</Service>\n";  
}

void StatsCollector::dumpAsPlain(std::ostringstream &out) {
  map<std::string, string> auxCounters;
  collectAllCounters(auxCounters);  
  for (map<string, string>::const_iterator it = auxCounters.begin();
    it != auxCounters.end();
    ++it)
  {
    out << it->first << " = " << it->second << "\n";
  }
}


void StatsCollector::registerStatsSource(StatsSource &source) {
  mutex::scoped_lock lk(sourcesLock);
  sources.insert(&source);
}

void StatsCollector::unregisterStatsSource(StatsSource &source) {
  mutex::scoped_lock lk(sourcesLock);  
  sources.erase(&source);    
}

void StatsCollector::checkCounterExists(const std::string &name) 
throw (runtime_error) 
{
  if (counters.count(name) == 0) {
    throw runtime_error("Counter does not exist");
  }
}
