/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef STATSCOLLECTOR_H_
#define STATSCOLLECTOR_H_

#include <string>
#include <map>
#include <set>
#include <sstream>
#include <boost/thread/mutex.hpp>

class StatsSource {
public:
  virtual void collectCounters(std::map<std::string, std::string> &counters) = 0;
  virtual ~StatsSource() {};
};

class StatsCollector
{
public:
  /** Creates or resets a counter */
  static void setCounter(const std::string &name, int64_t value);
  static void dropCounter(const std::string &name);
  static int incCounter(const std::string &name) 
    throw (std::runtime_error);
  static int decCounter(const std::string &name) 
    throw (std::runtime_error);
  static int addToCounter(const std::string &name, int64_t value) 
    throw (std::runtime_error);

  /*
   * Query methods
   */
  static bool hasCounter(const std::string &name); 
  static int64_t getCounter(const std::string &name) 
    throw (std::runtime_error);
  static const std::map<std::string, int64_t> &getAllCounters();
  static void dumpAsXml(std::ostringstream &out);
  static void dumpAsPlain(std::ostringstream &out);
  
  /*
   * Classes may also manage monitoring variables internally, without
   * storing them on this StatsCollector. In this case, these classes
   * must register with the StatsCollector as a data source. When the
   * StatsCollector receives a request to generate statistics, it will
   * query all registered data sources to collect variables.
   * 
   * These two methods are for managing data sources.
   */   
  static void registerStatsSource(StatsSource &source);
  static void unregisterStatsSource(StatsSource &source);
  
private:
  // The internal counters managed by this class.
  static std::map<std::string, int64_t> counters;
  static boost::mutex countersLock;
  // The list of data sources.
  static std::set<StatsSource*> sources;
  // Protects the sources data structure
  static boost::mutex sourcesLock;
  /** If the given counter does not exist, throws an exception. */
  static void checkCounterExists(const std::string &name) throw (std::runtime_error);
  /** Polls are data sources for available counters */
  static void collectAllCounters(std::map<std::string, std::string> &auxCounters);
  
};

#endif /*STATSCOLLECTOR_H_*/
