// -*- Mode:c++ -*-
//
// $Id: TableEntry.cc,v 1.36 2010/04/14 07:15:30 asunil Exp $
//

/*
  Copyright (c) Members of the EGEE Collaboration. 2004.
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

bool debug = false;
#include "debug.h"
#include <iostream>
#include <vector>
#include <algorithm>



#include "TableEntry.h"
/// Debugging printout for EntryProps
void EntryProps::print(bool debug) const
{
  DMESG("Dumping EntryPropos: " << std::endl);
  DMESG("name :" << name
	<< ", id: " << id
	<< ", indexTable: " << indexTable
	<< ", directoryTable: " << directoryTable
	<< ", flags: " << flags
	<< ", mainColumn: " << mainColumn
	<< ", parent: " << parent
	<< ", parentID: " << parentID
	<< ", parentPerms: " << parentPerms
	<< ", parentACLs: " << parentACLs
	<< ", parentOwner: " << parentOwner
	<< ", parentRepGroups:  " << parentRepGroups << std::endl
	<< ", parentFlags:  " << parentFlags
	<< ", perms: " << perms
	<< ", master: " << master
	<< ", owner: " << owner
	<< ", ACLs:  " << ACLs
	<< ", repGroups:  " << repGroups
	<< std::endl);
}

std::ostream& operator<<(std::ostream& out, const EntryProps &p) {
  out << "name :" << p.name
    << ", id: " << p.id
    << ", indexTable: " << p.indexTable
    << ", directoryTable: " << p.directoryTable
    << ", flags: " << p.flags
    << ", mainColumn: " << p.mainColumn
    << ", parent: " << p.parent
    << ", parentID: " << p.parentID
    << ", parentPerms: " << p.parentPerms
    << ", parentACLs: " << p.parentACLs
    << ", parentOwner: " << p.parentOwner
    << ", parentRepGroups:  " << p.parentRepGroups
    << ", parentFlags:  " << p.parentFlags
    << ", perms: " << p.perms
    << ", master: " << p.master
    << ", owner: " << p.owner
    << ", ACLs:  " << p.ACLs
    << ", repGroups:  " << p.repGroups;
  return out;
}


bool optPart(const char c)
{
  return isalnum(c) || c == '(' || c == ')';
}


void toLower(std::string &s)
{
  for (size_t i=0; i< s.size(); i++)
    s[i] = tolower(s[i]);
}


int decodeTableOptions(const std::string &options, EntryProps &p)
{
  if( options.find("shared") != std::string::npos )
    p.flags |= EP_SHARED;
  
  if( options.find("acl") != std::string::npos ) {
    p.flags |= EP_ACLS;
    p.flags |= EP_PERMS;
  }

  if( options.find("plain") != std::string::npos ) {
    p.flags |= EP_PLAIN;
  } else {
    p.flags |= EP_PERMS;
  }

  if( options.find("file") != std::string::npos )
    p.flags |= EP_FCAT;

  if( options.find("inherit") != std::string::npos )
    p.flags |= EP_INHERIT;

  if( options.find("perms") != std::string::npos )
    p.flags |= EP_PERMS;

  size_t pos;
  if( (pos=options.find("type=")) != std::string::npos ){
    pos += 5;
    size_t npos = pos;
    while( npos < options.size() && optPart(options[npos]) )
      npos++;
  
    p.mainType = options.substr(pos, npos-pos);
    toLower(p.mainType);
    
    if( p.mainType == "int" )
      p.flags |= EP_PRIMARY_INT;
    if( p.mainType == "float" )
      p.flags |= EP_PRIMARY_FLOAT;
    if( p.mainType == "date" )
      p.flags |= EP_PRIMARY_DATE;
  }

  // Decode main column name
  if( (pos=options.find("main=")) != std::string::npos ){
    pos += 5;
    size_t npos = pos;
    while( npos < options.size() && optPart(options[npos]) )
      npos++;
    if( p.flags & EP_PLAIN )
      p.mainColumn = options.substr(pos, npos-pos);
    else
      p.mainColumn = "user:" + options.substr(pos, npos-pos);
  }

  // Decode MySQL table type
  std::string table = "";
  if( (pos=options.find("table=")) != std::string::npos ){
    pos += 5;
    unsigned int npos = pos;
    while(npos < options.size() && isalnum(options[npos]))
      npos++;
    p.tableType = options.substr(pos, npos-pos);
  }

  return p.flags;
}


std::string encodeTableOptions(int options)
{ 
  std::vector<std::string> opts;
  if (options & EP_SHARED) { opts.push_back("shared"); }
  if (options & EP_ACLS) { opts.push_back("acls");  }
  if (options & EP_INHERIT) { opts.push_back("inherit");  }
  if (options & EP_PLAIN) { opts.push_back("plain"); }
  if (options & EP_PERMS) { opts.push_back("perms"); }
  if (options & EP_FCAT) { opts.push_back("file"); }
  if (options & EP_PRIMARY_INT) { opts.push_back("type=int"); }
  if (options & EP_PRIMARY_FLOAT) { opts.push_back("type=float"); }
  if (options & EP_PRIMARY_DATE) { opts.push_back("type=date"); }

  std::string sep;
  std::string codedOptions;
  for (size_t i = 0; i < opts.size(); i++) {
    codedOptions += sep + opts[i];
    sep = ",";
  }  
  return codedOptions;
}



/** Checks whether the permission string is legal, that is contains
    only allowed characters like r, w and x
 */
bool aclPermissionFormatOK(const std::string &perm,
			   std::string &cannonical)
{   
  // Sort and collapse the perm string to remove duplicate elements.
  cannonical = perm;
  sort(cannonical.begin(), cannonical.end());
  std::string::iterator it = unique(cannonical.begin(), cannonical.end());
  cannonical.erase(it, cannonical.end());
  
  // The list of allowed permissions
  std::string allowed = "adrtwx";
  sort(allowed.begin(), allowed.end());
  // Check if the given permissions contain only valid permissions
  return includes(allowed.begin(), allowed.end(), cannonical.begin(), cannonical.end());
}

/** Checks whether the permission string is legal, that is contains
    only allowed characters like r, w, x
 */
bool permissionFormatOK(const std::string &perm, int dir)
{
  if(perm.size() != 3)
    return false;

  if(perm[0] != 'r' && perm[0] != '-')
    return false;

 
  if(perm[1] != 'w' && perm[1] != '-')
    return false;

 
  if( dir ) {
    if(perm[2] != 'x' && perm[2] != '-' && perm[2] != 't')
      return false;
  } else {
    if(perm[2] != 'x' && perm[2] != '-')
      return false;
  }


  return true;
}

/** Converts an ACL string as found in the acls column in the
    directory indices into a map which has the group names as the
    keys, and the respective permissions as values.
*/
int aclStringToMap(const std::string &m,
		   std::map< std::string, std::string > &acls)
{
  size_t pos=0;
  while(pos<m.size()){
    size_t newpos=m.find_first_of("|", pos);
    if(m.size() && newpos == std::string::npos)
      newpos=m.size();
    if(newpos!=std::string::npos && newpos-pos){
      size_t delim=m.find_first_of(" ", pos);
//      DMESG("pos "<< pos << "  newpos " << newpos
//	    << "  delim " << delim << ": >"<<(m.substr(pos, delim-pos))
//	    << "<  >" <<  (m.substr(delim+1, newpos-delim-1)) << "<"
//	    << std::endl);
      acls[m.substr(pos, delim-pos)]=m.substr(delim+1, newpos-delim-1);
      pos=newpos+1;
    } else
      pos=newpos+1;
  }
  return acls.size();
}


// Converts an ACL permission to a normal 3 characterunix permission
std::string aclToPerm(const std::string &acl)
{
  std::string perm("---");
  if( acl.find('r') != std::string::npos)
    perm[0] = 'r';
  if( acl.find('w') != std::string::npos)
    perm[1] = 'w';
  if( acl.find('x') != std::string::npos)
    perm[2] = 'x';

  return perm;
}


int splitGroupString(const std::string m,
		     std::list< std::string > &parts)
{
  size_t pos=0;
  while(pos<m.size()){
    size_t newpos=m.find_first_of("|", pos);
    if(m.size() && newpos == std::string::npos)
      newpos=m.size();
    if(newpos!=std::string::npos && newpos-pos){
      parts.push_back(m.substr(pos, newpos-pos));
      pos=newpos+1;
    } else
      pos=newpos+1;
  }
  return 0;
}


std::string generateTableName(const std::string &type, 
			      const std::string &base,
			      const std::string &name)
{
  std::string r;
  size_t pos = base.find(".");
  if( pos != std::string::npos )
    r = base.substr(0, pos+1);

  if( type == "index" )
    r = "";

  r.append(type).append("_");

  if( pos != std::string::npos )
    r.append(base.substr(pos + 1));
  else
    r.append(base);

  r.append("_").append(name);
  
  return r;
}


std::string stripSchema(const std::string &table)
{
  std::string r(table);

  size_t pos = table.find(".");
  if( pos != std::string::npos )
    r = table.substr(pos + 1);
  return r;
}


std::string getSchema(const std::string &table)
{
  size_t pos = table.find(".");
  if( pos != std::string::npos )
    return table.substr(0, pos);
  return "";
}


std::string qualifySchema(const std::string &table,
			  const std::string &defaultSchema)
{
  DMESG("qualifySchema: got " << table << "\n");

  if( table == "x" ) {
    DMESG("qualifySchema: Returning x");
    return "x";
  }

  std::string name = table;
  if( table.find(".") == std::string::npos 
      && defaultSchema.size() && table.find("dir") ==0 ) 
    name = defaultSchema + "." + table;

  DMESG("qualifySchema: Returning " << name << "\n");
  return name;
}

const char *typeFromFlags(int flags) {
  if( flags & EP_PRIMARY_INT )
    return "int";
  if( flags & EP_PRIMARY_FLOAT )
    return "float";
  if( flags & EP_PRIMARY_DATE )
    return "date"; 
  return "varchar(256)";
}
