/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: TableEntry.h,v 1.24 2010/10/26 07:27:47 asunil Exp $
//
#ifndef TABLE_ENTRY_H
#define TABLE_ENTRY_H

#include <stdexcept>
#include <string>
#include <map>
#include <list>

// EntryProps options
#define EP_SHARED           1
#define EP_ACLS             2
#define EP_PLAIN            4
#define EP_FCAT             8
#define EP_IMPORTED        16
#define EP_INHERIT         32
#define EP_PERMS           64
#define EP_PRIMARY_INT    256
#define EP_PRIMARY_FLOAT  512
#define EP_PRIMARY_DATE  1024
#define EP_PROXY         2048

#define EP_TYPE        0xf000

#define EPT_NONE            0
#define EPT_DIR        0x1000
#define EPT_INDEX      0x2000
#define EPT_VIEW       0x3000
#define EPT_SEQUENCE   0x4000
#define EPT_VOLUME     0x5000
#define EPT_WSDAIR     0x7000

#define DEFAULT_MAIN_COLUMN "file"

struct EntryProps {
  std::string name;             // Only last part of name
  long id;
  long parentID;
  std::string indexTable;
  std::string directoryTable;   // Name of table associated
  std::string parent;           // Full path
  std::string parentPerms;
  std::string parentOwner;
  std::string parentACLs;
  std::string parentRepGroups;
  std::string perms;            // User + Group permissions
  std::string owner;
  std::string ACLs;             // In case of directory
  std::string repGroups;        // In case of directory
  int flags;
  std::string mainColumn;       // The name of the main column
  std::string mainType;         // The type of the main column
  std::string tableType;        // The type of the table (in MySQL)
  std::string inheritsFrom;     // The name of the table this inherits from (PG only)
  std::string master;  // Master site
  void print(bool debug = false) const;
  int parentFlags;
  std::string guid;
  std::string surl;

  int type() const { return flags & EP_TYPE; }
  int parentType() const { return parentFlags & EP_TYPE; }
  int sticky() const { 
    if ( type() && (perms[2] == 't')) return 1; 
    if ( parentPerms[2] == 't') return 1; 
    return 0;
  }

  friend std::ostream& operator<<(std::ostream& out, const EntryProps &p);
};


struct TableEntry {
  TableEntry() {
    // Nothing
  };

  TableEntry(const EntryProps &p, int s)
    : props(p), searchOptions(s) { };

  EntryProps props;
  int searchOptions;
};

typedef std::map< std::string, struct TableEntry > TableMap;


/** Decodes the options provided by the user in the createdir command */
int decodeTableOptions(const std::string &options, EntryProps &p);

/** Recreate the options provided with the createentry command */
std::string encodeTableOptions(int options);

std::string generateTableName(const std::string &type, const std::string &base,
			      const std::string &name);

bool permissionFormatOK(const std::string &perm, int dir);

bool aclPermissionFormatOK(const std::string &perm,
			   std::string &cannonical);

int aclStringToMap(const std::string &m,
		   std::map< std::string, std::string > &acls);

std::string aclToPerm(const std::string &acl);

int splitGroupString(const std::string,
		     std::list< std::string > &parts);

std::string qualifySchema(const std::string &table,
			  const std::string &defaultSchema);

std::string stripSchema(const std::string &table);

std::string getSchema(const std::string &table);

const char *typeFromFlags(int flags);

#endif // TABLE_ENTRY_H
