/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "ThreadUtil.h"
#include "util/Utils.h"
#include <memory>
#include <iostream>
#include <errno.h>
#include <cassert>

using namespace boost;

static bool debug = false;
#define logstream std::cout
#include "debug.h"

using namespace std;

void ec(int retValue) {
	if (retValue) {
		PERRLOG("[ERROR]");
	}
}

void *execute(void *args) {
	// Make sure that the task object is deleted at the end of this function
	auto_ptr<Runnable> task((Runnable*) args);
	try {
		task->run();
	} catch ( ... ) {
		LOG("[execute()]: Interrupted by an unknown exception\n");
		throw;
	}
	return 0;
}

void cancelAndJoinThread(pthread_t &thread, const string &threadName) {
// 	cout << "Cancelling thread " << thread << "\n";
	void *result;
	DMESG("Cancelling\n");
	ec(pthread_cancel(thread));
	DMESG("Joining\n");
	ec(pthread_join(thread, &result));
	if (result == PTHREAD_CANCELED) {
		DMESG(threadName << " thread was cancelled.\n");
	} else if (result == 0) {
		DMESG(threadName << " thread exited normally.\n");
	} else {
		LOG(threadName << " thread exited with unknown result: " << result << endl);
	}
}

Runnable::Runnable(bool deleteOnExit) : 
        _cancelled(false),
        _sleeping(false),
        _deleteOnExit(deleteOnExit)
{
  DMESG("Runnable\n");
}

Runnable::~Runnable() {
  DMESG("~Runnable\n");
  if (!_deleteOnExit) {
    DMESG("Joining with thread\n");
    _thread->join();
  }
}

struct ThreadRunner {
  Runnable *_task;
  ThreadRunner(Runnable* task) : _task(task) { }

  void operator()() {
    try {
      if (_task->_deleteOnExit) {
        // Must delete the object when we are done with it,
        // so wrap it on a auto_ptr.
        auto_ptr<Runnable> ptr(_task);
        ptr->run();
      } else {
        // The object is not to be deleted.
        _task->run();
      }
    } catch ( ... ) {
      LOG("[execute()]: Interrupted by an unknown exception\n");
      throw;
    }
  }
};

/**
 * Starts the thread
 */
void Runnable::start() {
  assert(_cancelled == false);
  try {
    ThreadRunner runner(this);
    if (_deleteOnExit) {
      // Release the thread object as soon as it is created,
      // since we don't need to join with it. If we store it
      // on the stack
      boost::thread thr(runner);
    } else {
      _thread.reset(new boost::thread(runner));
    }
  } catch (boost::thread_resource_error &ex) {
    throw runtime_error(string("Could not create thread: ") + ex.what());
  }
}
// There is no easy way to cancel a thread with boost
void Runnable::cancel() {
    DMESG("Cancelling\n");
    _cancelled = true;
    boost::timed_mutex::scoped_lock lk(_lock);
    if (_sleeping) {
        _sleepCond.notify_all();
    }
//     else {
//      pthread_cancel(_thread);
//    }
    DMESG("Cancelled\n");
}

bool Runnable::isCancelled() {
	return _cancelled;
}

bool Runnable::sleep(unsigned long seconds) {
  if (_cancelled) {
    throw runtime_error("Thread cancelled.");
  }
	// Gets the time that it has to wait
	boost::xtime xt;
	boost::xtime_get(&xt, boost::TIME_UTC);
	xt.sec += seconds;
	
	boost::timed_mutex::scoped_timed_lock lk(_lock, xt);
  _sleeping = true;
  return _sleepCond.timed_wait(lk, xt);
//  _sleeping = false; // <bk> Took this out, why was it here?
}

/**
 * Joins with the thread
 */
void Runnable::join() {
  if (_deleteOnExit) {
    throw runtime_error("Cannot join. Thread is set to be deleted on exit\n");
  }
  try {
    _thread->join();
  } catch (std::exception &ex) {
    PERRLOG("Could not join thread: " << ex.what() << endl);
    throw;
  }
}
