/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


#ifndef PTHREADUTIL_H
#define PTHREADUTIL_H

#include <boost/utility.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/exceptions.hpp>
#include <boost/thread/xtime.hpp>

#include <string>

/**
 * Calling start() creates and starts the thread. When
 * the thread exits, the object deletes itself
 * 
 * @author Nuno Santos
*/
class Runnable {
private:
  //! Serialize access  
  boost::timed_mutex _lock;
  boost::condition _sleepCond;
  volatile bool _cancelled;
  bool _sleeping;
  bool _deleteOnExit;  
  std::auto_ptr<boost::thread> _thread;
  
  friend struct ThreadRunner;
  
protected:
  /** 
   * 
   * Makes the thread sleep for a given time or until 
   * cancel is called. Returns true if thread is cancelled,
   * false if the timer expires.
   */
  bool sleep(unsigned long millis);
  
public:
  /** If deleteOnExit is true, this object will be deleted
   * when the thread exits and this thread will not be joinable. */
	Runnable(bool deleteOnExit = true);
  /**
	 * Runnables should never throw in their destructors
	 */
	virtual ~Runnable();
	
  void cancel();
  bool isCancelled();

  /**
   * Starts the thread
   */
  void start();
  
  /**
   * Joins with the thread
   */
  void join();  
	
	/**
	 * Task to be performed in another thread of execution.
	 * Should honor the cancel request.
	 */
    virtual void run() = 0;
};
  

void ec(int retValue);
void *execute(void *args);
void cancelAndJoinThread(pthread_t &thread, const std::string &threadName = "") ;
#endif
