/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
// $Id: sslHelpers.cc,v 1.15 2012/02/01 06:42:03 gpark Exp $
//
//

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sslHelpers.h"
#include <openssl/x509v3.h>
#include <openssl/evp.h>
#include "SecurityUtils.h"
extern AuthContext::ignoreCRLCheck ;

static int debug = 0;
#include "debug.h"

void sslPrintErrors() {
  // see "man BIO_s_mem"
  
  ERRLOG("SSL error queue: " << std::endl);

  char buff[201];
  BIO *mem=BIO_new(BIO_s_mem());
  ERR_print_errors(mem);
  while (!BIO_eof(mem)) {
    BIO_gets(mem, buff, 200);
    ERRLOG(buff);
  }
  BIO_free(mem);
}


/// Callback when certificates are checked
/** This callback is given to SSL_CTX_set_verify and prints out error
    messages when certificate chains are verified.
 */
int certVerifyCallback(int ok, X509_STORE_CTX *store)
{
  char buff[256];
  
  if(!ok) {
    X509 *cert = X509_STORE_CTX_get_current_cert(store);
    int depth = X509_STORE_CTX_get_error_depth(store);
    int err = X509_STORE_CTX_get_error(store);

    ERRLOG("Client certificate error at depth: " << depth
	   << " error " << err << std::endl);
    X509_NAME_oneline(X509_get_issuer_name(cert), buff, 256);
    ERRLOG("   issuer = " << buff << std::endl);
    X509_NAME_oneline(X509_get_subject_name(cert), buff, 256);
    ERRLOG("   subject = " << buff << std::endl);
    ERRLOG("   error   = " << X509_verify_cert_error_string(err) 
	   << std::endl);

  }

  return ok;
}


/// Callback when certificates are checked, allows proxy certs
/** This callback is given to SSL_CTX_set_verify to allow grid proxy
    certificates to be used.
    In contrast to the Globus solution it does not rely on a
    proprietary mechanism to build the certificate chains but leaves
    this to openssl itself allowing us to profit from the several
    bug-fixes in the 0.9.6, 0.9.7 and 0.9.8 releases. This routine
    only overrides the certificate verification in 2 cases where a user
    certificate is used to create another proxy certificate:
    X509_V_ERR_INVALID_CA and X509_V_ERR_INVALID_PURPOSE
    In both cases it is carefully checked that the certificate signed
    is only a proxy certificate.
    This callback is also needed for openssl 0.9.8 because most admins will
    have proxy certificates disabled in the openssl configuration, anyway.
    This version of the callback works with 0.9.6a-0.9.8a.
    The idea of the routine is taken from the example in the 0.9.7
    man-page of SSL_CTX_set_verify.
*/
int certVerifyCallbackAllowProxy(int ok, X509_STORE_CTX *store)
{
  DMESG("certVerifyCallbackAllowProxy called:\n");

#if OPENSSL_VERSION_NUMBER < 0x00907000L
  store->check_issued=checkIssuerCallback;
#endif

  char buff[256];
 
  // See man SSL_CTX_set_verify 
  if(!ok) {
    X509 *cert = X509_STORE_CTX_get_current_cert(store);
    int depth = X509_STORE_CTX_get_error_depth(store);
    int err = X509_STORE_CTX_get_error(store);

    if( debug ) {
      DMESG("Possible certificate error at depth: " << depth
	    << " error " << err << std::endl);
      X509_NAME_oneline(X509_get_issuer_name(cert), buff, 256);
      DMESG("   issuer = " << buff << std::endl);
      X509_NAME_oneline(X509_get_subject_name(cert), buff, 256);
      DMESG("   subject = " << buff << std::endl);
      DMESG("   error   = " << X509_verify_cert_error_string(err) 
	    << std::endl);
      DMESG("   ignore crl : " << AuthContext::ignoreCRLCheck  << std::endl);
    }

    bool isProxy = isProxyCert(cert);

// In 0.9.6 we cannot override the issuer check, so we need
// to check whether we have a proxy cert at hand an ignore all
// the resulting errors    
#if OPENSSL_VERSION_NUMBER < 0x00907000L
    if( isProxy && err == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY ){
      X509_STORE_CTX_set_error(store, X509_V_OK);
      DMESG("Ignoring X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY for proxy cert\n");
      return 1;
    }
    if( isProxy && err == X509_V_ERR_CERT_UNTRUSTED ){
      X509_STORE_CTX_set_error(store, X509_V_OK);
      DMESG("Ignoring X509_V_ERR_CERT_UNTRUSTED for proxy cert\n");
      return 1;
    }
    if( isProxy && err == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE ){
      X509_STORE_CTX_set_error(store, X509_V_OK);
      DMESG("Ignoring X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE for proxy cert\n");
      return 1;
    }
#endif

    bool isProxyDelegator = false;
    STACK_OF(X509) *chain = X509_STORE_CTX_get_chain(store);
    
    // Openssl 1.0 higher have error in sk_num() and STACK_OF(X509)
    // (cannot conver error)
    // For preventing error must use sk_X509_num()
    
#if defined(OPENSSL_VERSION_NUMBER) && OPENSSL_VERSION_NUMBER >= 0x10000000L
    if(chain && sk_X509_num(chain) ) {
      DMESG("Certificate chain length: " << sk_X509_num(chain) << std::endl);
      if( depth > 0 && depth <= sk_X509_num(chain)) {
	      isProxyDelegator = isProxyOf((X509*)sk_X509_value(chain, depth-1), cert);
      }
    }
#else

    if(chain && sk_num(chain) ) {
      DMESG("Certificate chain length: " << sk_num(chain) << std::endl);
      if( depth > 0 && depth <= sk_num(chain)) {
	      isProxyDelegator = isProxyOf((X509*)sk_value(chain, depth-1), cert);
      }
    }
#endif
    
    // Handle proxy cerificates
    if( isProxyDelegator && err == X509_V_ERR_INVALID_CA ){
      X509_STORE_CTX_set_error(store, X509_V_OK);
      DMESG("Ignoring X509_V_ERR_INVALID_CA for proxy cert\n");
      return 1;
    }

    if( isProxyDelegator && err == X509_V_ERR_INVALID_PURPOSE ){
      X509_STORE_CTX_set_error(store, X509_V_OK);
      DMESG("Ignoring X509_V_ERR_INVALID_PURPOSE, in proxy cert\n");
      return 1;
    }

    DMESG("isProxy: " << isProxy << "err: " << err << std::endl);

    if( (isProxy && err == X509_V_ERR_UNABLE_TO_GET_CRL) || 
        (AuthContext::ignoreCRLCheck > 0)  ){
      X509_STORE_CTX_set_error(store, X509_V_OK);
      DMESG("Ignoring X509_V_ERR_UNABLE_TO_GET_CRL, in proxy cert\n");
      return 1;
    }

#ifdef X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION
    if( err == X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION ) {
      char s[80];
      X509_EXTENSION *ex;

      for (int i = 0; i < X509_get_ext_count(cert); ++i) {
	ex = X509_get_ext(cert, i);
        if (X509_EXTENSION_get_critical(ex) && !X509_supported_extension(ex)) {
	  OBJ_obj2txt(s, sizeof(s), X509_EXTENSION_get_object(ex), 1);

	  DMESG("Found extension " << s << "\n");

	  //	  return 1;
	  if (strcmp(s, "1.3.6.1.4.1.3536.1.222") != 0) return 1; // old
	  if (strcmp(s, "1.3.6.1.5.5.7.1.14") != 0) return 1; // RFC
	}
      }
    }
#endif

    if( !debug ) {
      ERRLOG("Certificate error at depth: " << depth << " error " 
	     << err << "\n");
      X509_NAME_oneline(X509_get_issuer_name(cert), buff, 256);
      ERRLOG("   issuer = " << buff << std::endl);
      X509_NAME_oneline(X509_get_subject_name(cert), buff, 256);
      ERRLOG("   subject = " << buff << std::endl);
      ERRLOG("   error   = " << X509_verify_cert_error_string(err) 
	     << std::endl);
    }
  
  }

  DMESG("  Callback returns: " << ok << std::endl);

  return ok;
}



/** This is a modified version of the issuer check routine
    which overrrides the error that a user certificate is used to sign
    a proxy certificate although it does not have this capability.
    All other checks are done by openssl's standard X509_check_issued
    routine which allows us to profit from any security fix being done
    in openssl.
    This routine is enabled by overriding the
    ctx->cert_store->check_issued function pointer of the certificte
    store of the SSL context in 0.9.7 and above. In 0.9.6 we need to do this
    in the verify routine itself.
    The idea of the code is stolen from openssl's own checking
    routine in x509_vfy.c.
    Returns 1 for OK and 0 for error
*/
int checkIssuerCallback(X509_STORE_CTX *ctx, X509 *cert, X509 *issuer)
{ 
  DMESG("check Issuer called:\n");
  int ret = X509_check_issued(issuer, cert);
  DMESG("   Error " << X509_verify_cert_error_string(ret) 
	<< ": " << ret << std::endl);
  switch(ret) {
  case X509_V_OK:
    return 1; // OK

  case X509_V_ERR_KEYUSAGE_NO_CERTSIGN:
    // Ignore problem with key usage not including cert-signing for proxy certs
    if(isProxyOf(cert, issuer)) {
      DMESG("Ignoring X509_V_ERR_KEYUSAGE_NO_CERTSIGN" << std::endl);
      // Return directly, no need to override our decision in CB
      return 1; // OK
    }
#if OPENSSL_VERSION_NUMBER < 0x00907000L
  case X509_V_ERR_SUBJECT_ISSUER_MISMATCH:
    // Ignore problem with key usage not including cert-signing for proxy certs
    if(isProxyOf(cert, issuer)) {
      DMESG("Ignoring X509_V_ERR_SUBJECT_ISSUER_MISMATCH" << std::endl);
      // Return directly, no need to override our decision in CB
      return 1; // OK
    }
#endif
  }

#if OPENSSL_VERSION_NUMBER >= 0x00908000L
  // If we haven't asked for issuer errors don't set ctx
  if (!(ctx->param->flags & X509_V_FLAG_CB_ISSUER_CHECK))
    return 0;
#else
  // If we haven't asked for issuer errors don't set ctx
  if (!(ctx->flags & X509_V_FLAG_CB_ISSUER_CHECK))
    return 0;
#endif

  DMESG("  returning via verify_cb\n");

  // Allows our callback to override the decision
  ctx->error = ret;
  ctx->current_cert = cert;
  ctx->current_issuer = issuer;
  return ctx->verify_cb(0, ctx);
}


/** Routine which checks whether cert is a proxy certificate
    The check for a proxy certificate is done by verifying that
    the cert and its issuer share the same common name apart from
    an additional proxying in the certificate itself.
*/
int isProxyCert(X509 *cert)
{
  X509_NAME *subject;
  X509_NAME *issuer;

  subject=X509_get_subject_name(cert);
  if (!subject)
    return 0;

  issuer = X509_get_issuer_name(cert);
  if (!issuer)
    return 0;

  std::string realName;
  int proxyLevel = 0;
  
  std::string issuerRealName;
  int issuerProxyLevel = 0;

  int pos = -1;
  while((pos=X509_NAME_get_index_by_NID(subject, NID_commonName,pos)) >-1) {
    X509_NAME_ENTRY *e=X509_NAME_get_entry(subject, pos);
    if(!e)
      break;
    std::string value;
    value=(const char *)(ASN1_STRING_data(X509_NAME_ENTRY_get_data(e)));
    DMESG("Value: >" << value << "<" << std::endl);
    if( value!="proxy" && value != "limited proxy" && !realName.size() )
      realName = value;
    proxyLevel++;
  }

  pos = -1;
  while((pos=X509_NAME_get_index_by_NID(issuer, NID_commonName,pos)) >-1) {
    X509_NAME_ENTRY *e=X509_NAME_get_entry(issuer, pos);
    if(!e)
      break;
    std::string value;
    value=(const char *)(ASN1_STRING_data(X509_NAME_ENTRY_get_data(e)));
    DMESG("Issuer Value: >" << value << "<" << std::endl);
    if( value!="proxy" && value!="limited proxy" && !issuerRealName.size() )
      issuerRealName = value;
    issuerProxyLevel++;
  }

  DMESG("  Real name: >" << realName 
	<< "< and Issuer real name: " << issuerRealName << std::endl);
  DMESG("  Proxy level: " << proxyLevel 
	<< " and " << issuerProxyLevel << std::endl);

  if( (proxyLevel == issuerProxyLevel+1)
      && (issuerRealName == realName) 
      && (realName.size())  ) {
    DMESG("IS a proxy\n");
    return 1;
  }

  DMESG("NOT a proxy!\n");

  return 0;
}


/** Routine which checks whether the child-cert is a proxy of parent
    The check for a proxy certificate is done by verifying that
    the cert and its issuer share the same common name apart from
    an additional proxying in the certificate itself.
*/
int isProxyOf(X509 *child, X509 *parent)
{
  DMESG("Entering isProxyOf\n");

  int childProxyLevel = 0;
  int parentProxyLevel = 0;
  std::string childRealName;
  std::string parentRealName;

  /*
  long serialN;
  serialN = ASN1_INTEGER_get(X509_get_serialNumber(child));
  DMESG("The child's serial number is: " << serialN << "\n");
  */
  
  X509_NAME *childSubject=X509_get_subject_name(child);
  if (!childSubject)
    return 0;

  X509_NAME *parentSubject=X509_get_subject_name(parent);
  if (!parentSubject)
    return 0;

  int pos = -1;
  while((pos=X509_NAME_get_index_by_NID(childSubject, NID_commonName,pos))>-1) {
    X509_NAME_ENTRY *e=X509_NAME_get_entry(childSubject, pos);
    if(!e)
      break;
    std::string value;
    value=(const char *)(ASN1_STRING_data(X509_NAME_ENTRY_get_data(e)));
    DMESG("Value: >" << value << "<" << std::endl);
    childRealName += value;
    childProxyLevel++;
  }

  pos = -1;
  while((pos=X509_NAME_get_index_by_NID(parentSubject, NID_commonName,pos)) >-1) {
    X509_NAME_ENTRY *e=X509_NAME_get_entry(parentSubject, pos);
    if(!e)
      break;
    std::string value;
    value=(const char *)(ASN1_STRING_data(X509_NAME_ENTRY_get_data(e)));
    DMESG("Value: >" << value << "<" << std::endl);
    parentRealName += value;
    parentProxyLevel++;
  }

  DMESG("  Child real name: " << childRealName 
	<< " and parent real name: " << parentRealName << std::endl);
  DMESG("  Proxy level: " << childProxyLevel 
	<< " and " << parentProxyLevel << std::endl);

  if( (strncmp(childRealName.c_str(), parentRealName.c_str(), parentRealName.size()) == 0)
      && (childProxyLevel == parentProxyLevel+1) )
    return 1;

  DMESG("  Not a proxy cert\n");

  return 0;
}


/** Returns the subject name of a certificate (presented by a client)
    for normal non-proxy certificates.
    If the certificate is in fact a proxy certificate, not the subject
    but the issuer is returned (which is the subject of the
    certificate of the original user).
*/
std::string getCertSubjectName(X509 *cert, bool traditional)
{
  DMESG("getCertSubjectName(), traditional=" << traditional << "\n");

  if(!cert)
    return "";

  char buff[256];
  X509_NAME *subject;

  bool isProxy = isProxyCert(cert);
  if(isProxy)
    subject = X509_get_issuer_name(cert);
  else
    subject = X509_get_subject_name(cert);

  if(traditional) {
    X509_NAME_oneline(subject, buff, 255);
  } else {
    BIO *mem=BIO_new(BIO_s_mem());
    X509_NAME_print_ex(mem, subject, 0, XN_FLAG_ONELINE);
    
    BIO_gets(mem, buff, 255);
    BIO_free(mem);
  }

  while (traditional && isProxy && strlen(buff)>9) {
    if(!strcmp(buff+strlen(buff)-9, "/CN=proxy"))
      buff[strlen(buff)-9]='\0';
    else
      break;
  }

  while (traditional && isProxy && strlen(buff)>17) {
    if(!strcmp(buff+strlen(buff)-17, "/CN=limited proxy"))
      buff[strlen(buff)-17]='\0';
    else
      break;
  }

  while (!traditional && isProxy && strlen(buff)>12) {
    if(!strcmp(buff+strlen(buff)-12, ", CN = proxy"))
      buff[strlen(buff)-12]='\0';
    else
      break;
  } 

  while (!traditional && isProxy && strlen(buff)>20) {
    if(!strcmp(buff+strlen(buff)-20, ", CN = limited proxy"))
      buff[strlen(buff)-20]='\0';
    else
      break;
  }

  DMESG("Buff: >" << buff << "<\n");

  return buff;
}


std::string sessionToPrintable(unsigned char *sessionID, unsigned int len)
{
  std::stringstream asc;
  for (size_t i=0; i<len; i++){
    asc.width(2); asc.fill('0');
    asc << std::hex << ((unsigned int)sessionID[i]);
  }
  return asc.str();
}

EVP_MD_CTX *initSignMessage()
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  EVP_MD_CTX *ctx = EVP_MD_CTX_create();

  if(!ctx) {
    ERRLOG("Could not create signing context\n");
    sslPrintErrors();
    return NULL;
  }
  
  const EVP_MD *md = EVP_sha1();

  if(!EVP_SignInit_ex(ctx, md, NULL)){
    ERRLOG("Could not initialize signing context\n");
    sslPrintErrors();
    EVP_MD_CTX_destroy(ctx);
    return NULL;
  }
  
  return ctx;
#else
  return NULL;
#endif
}


int updateSignMessage(EVP_MD_CTX *ctx, const std::string &message)
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  return EVP_DigestUpdate(ctx, message.c_str(), message.size());
#else
  return 0;
#endif
}


std::string finalizeSignMessage(EVP_MD_CTX *ctx, EVP_PKEY *pkey)
{
  unsigned int md_len = EVP_PKEY_size(pkey);

  unsigned char *md_value=new unsigned char[md_len];

  if(!EVP_SignFinal(ctx, md_value, &md_len, pkey)) {
    ERRLOG("Could not finalize message signature\n");
    sslPrintErrors();
    delete [] md_value;
    return "";
  }
  
  DMESG("Signature size: " << md_len << std::endl);

  std::stringstream signature;
  for (size_t i=0; i<md_len; i++){
    signature.width(2); signature.fill('0');
    signature << std::hex << (int)md_value[i];
  }

  delete [] md_value;

  return signature.str();
}


EVP_MD_CTX *initVerifyMessage()
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  EVP_MD_CTX *ctx = EVP_MD_CTX_create();

  if(!ctx) {
    ERRLOG("Could not create message verifying context\n");
    sslPrintErrors();
    return NULL;
  }
  
  const EVP_MD *md = EVP_sha1();

  if(!EVP_VerifyInit_ex(ctx, md, NULL)){
    ERRLOG("Could not initialize message verifying context\n");
    sslPrintErrors();
    EVP_MD_CTX_destroy(ctx);
    return NULL;
  }
  
  return ctx;
#else
  return NULL;
#endif
}


int updateVerifyMessage(EVP_MD_CTX *ctx, const std::string &message)
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  return EVP_VerifyUpdate(ctx, message.c_str(), message.size());
#else
  return 0;
#endif

}


int finalizeVerifyMessage(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const std::string &signature)
{
#if OPENSSL_VERSION_NUMBER >= 0x00907000L

  // size_t md_len = EVP_PKEY_size(pkey);

  unsigned char *md_value=new unsigned char[signature.size()/2];
  for(size_t i=0; i<signature.size()/2; i++){
    std::string byte(signature, i*2, 2);
    md_value[i] = strtoul(byte.c_str(), NULL, 16);
  }

  DMESG("Signature size: " << signature.size()/2 << std::endl);

  int result = EVP_VerifyFinal(ctx, md_value, signature.size()/2, pkey);
  if( result <0) {
    ERRLOG("Error verifying message signature\n");
    sslPrintErrors();
    delete [] md_value;
    return -1;
  }
  
  if (!result && debug) {
    DMESG("Error verifying signature\n");
    sslPrintErrors();
  }

  delete [] md_value;

  return result;
#else
  return 0;
#endif
}
