/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  

  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/


// -*- Mode:c++ -*-
//
// $Id: MDStandalone.cc,v 1.365 2011/04/06 06:29:52 asunil Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "MDStandalone.h"
#include "MDBuffer.h"
#include "DBUserManager.h"
#include "GridMapUserManager.h"
#include "VOMSAttrUserManager.h"
#include "sqlHelpers.h"
#include "DBHelper.h"
#include "SiteManager.h"
#include "protocolHandling.h"
#include "MDServerFactory.h"
#include "MDInterpreter.h"

#include "queryparser/QParser.h"

#include "util/Statement.h"
#include "util/Path.h"
#include "util/Utils.h" 
#include "util/DatabaseConnection.h"
#include "util/guidHelper.h"
#include "util/amgaProtocol.h"

#include "replicationdaemon/MDLogWriter.h"
#include "replicationdaemon/ReplicationDaemonConnection.h"
#include "replicationdaemon/ReplicationManager.h"
#include "replicationdaemon/MountManager.h"
#include "replicationdaemon/SubscriptionManager.h"
#include "replicationdaemon/LogReceiver.h"
 
#include <sys/types.h>
#include <regex.h>
#include <stdlib.h>

#include <sqlext.h>

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string.hpp>


#include <assert.h>
#include <sstream>
#include <iostream>
#include <set>
#include <algorithm>

#include <sys/types.h>
#include <sys/wait.h>

#include "debug.h"

using namespace std;
using namespace boost;

#define SQL_ERR_MSG_SIZE 1024

// Size of the buffers to be used during the upload command
const int UPLOAD_BUFFER_SIZE = 8*1024;
const int UPLOAD_TRANSACTION_SIZE = 1000;
const int UPLOAD_BULK_SIZE = 1;

const std::string MDStandalone::DUMP_NO_PERMS = "noperms";
const std::string MDStandalone::DUMP_NO_SYNC = "nosync";

#define COMMAND_GROUPS 12
#define COMMANDS 68

static const char commandGroups [COMMAND_GROUPS][3][50] = {
  {"10", "entry", "Entry manipulation"},
  {"11", "group", "Group commands"},
  {"12", "acl", "ACLs"},
  {"14", "schema", "Schema management"},
  {"15", "sequence", "Sequence handling"},
  {"16", "user", "User management"},
  {"17", "view", "View handling"},
  {"18", "site", "Site Management"},
  {"19", "replicas", "Replica Management"},
  {"20", "ticket", "Ticket management"},
  {"21", "capabilities", "Capability Managment"},
  {"22", "admin", "Administrative Commands"},
};

static const char commands[COMMANDS][4][100] = {
  {"10", "createdir", "/parentdir/dir [option]",
   "Creates a new directory." },
  {"10", "makedir", "/parentdir/dir [option]",
   "Creates a new directory suitable for storing files" },
  {"10", "rm", "[-rf] pattern [condition]",
   "Removes entries matching pattern/condition" },
  {"10", "link", "file link",
   "Creates a link to another file or to an external URL" },
  {"10", "dir", "[directory]",
   "List a directory" },
  {"10", "ls", "[-Plgm] [entry]",
   "List a directory" },
  {"10", "listentries", "[directory]",
   "List the entries (not the collections!) in a directory" },
  {"10", "stat", "entry",
   "Shows the statistics of an enty or a directory"},
  {"10", "chown", "file owner",
   "Changes the owner of a file or directory" },
  {"10", "chmod", "file access-rights",
   "Changes the access mode of a file or directory"},
  {"10", "rmdir", "directory",
   "Removes a directory"},
  {"10", "dump", "[-noperms] [dir]",
   "Creates a recursive dump starting at dir, default is /"},
  {"10", "insert", "entry [attr value]...",
   "Adds a new entry and initializes attributes which get evaluated first"},
  {"10", "import", "table directory",
   "Imports an existing table in the database as the given directory"},
  {"10", "lfn_lookup", "guid ...",
   "Looks up the lfn which belongs to given guids"},
  {"11", "grp_create", "groupname",
   "Creates a new group"},
  {"11", "grp_delete", "groupname",
   "Deletes a group"},
  {"11", "grp_show", "groupname",
   "Lists the members of a group" },
  {"11", "grp_adduser", "groupname user",
   "Adds a user to a group"},
  {"11", "grp_removeuser", "groupname user",
   "Removes a user from a group"},
  {"11", "grp_member", "[user]",
   "Lists the groups a user belongs to. Only root can investigate other users"},
  {"11", "grp_list", "[-a] [user]",
   "Lists the groups a user (default current user) owns, or all existing groups if -a given."},
  {"11", "grp_dump", "",
   "Dumps all group information for backup"},
  {"11", "grp_voms_add", "groupname voms_group",
     "Allows user to have group permission with a certificate with the given VOMS Group"},
  {"11", "grp_voms_list", "",
     "List all possible VOMS Group"},
  {"11", "grp_voms_remove", "voms_group",
   "Removes VOMS Group allowed"},
  {"12", "acl_add", "entry group rights",
   "Adds a new access control to a file/directory granting a group certain rights"},
  {"12", "acl_remove", "entry group",
   "Removes an access control for a group from a file/directory"},
  {"12", "acl_show", "entry",
   "Shows all access controls for a file/directory"},
  {"14", "schema_cp", "source_dir dest_dir",
   "Copies a schema from one directory to another by adding attributes"},
  {"15", "sequence_create", "name dir [increment] [start value]",
   "Creates a new sequences with given name in the given directory" },
  {"15", "sequence_next", "sequence",
   "Gets the next value from a sequence"},
  {"15", "sequence_remove", "sequence",
   "Deletes a sequence"},
  {"15", "sequence_current", "sequence",
   "Gets the current value from a sequence"},
  {"16", "user_create", "user [password]",
   "Creates a new user and assigns a password if given"},
  {"16", "user_remove", "user",
   "Removes a user"},
  {"16", "user_password_change", "user password",
   "Changes the password. Only root can chage other users' passwords"},
  {"16", "user_subject_add", "user subject",
   "Allows user to log in with a certificate with the given subject"},
  {"16", "user_subject_remove", "user subject",
   "Removes the certificate given by the subject from the list of subject allowed for user"},
  {"16", "user_list", "",
     "Lists all users known to the authentication subsystem"},
  {"16", "user_listcred", "user",
     "Lists all possible credentials of a user. Only for root"},
  {"16", "user_dump", "",
     "Dumps user database for backup"},
  {"16", "user_voms_add", "user voms_role",
     "Allows user to log in with a certificate with the given VOMS Role"},
  {"16", "user_voms_list", "",
     "List all possible VOMS Role"},
  {"16", "user_voms_remove", "voms_role",
   "Removes VOMS Role allowed for user"},
  {"17", "view_create", "view maindir attrs... condition",
   "Creates a view as with maindir as the basis slecting attrs under given condition."},
  {"17", "view_create_union", "view dirs... condition",
   "Creates a view as a union of the entries in the directories under given condition."},
  {"17", "view_remove", "path",
   "Removes a view"},
  {"18", "site_list", "",
   "Lists all sites and their ID"},
  {"18", "site_dump", "[site]*",
   "Outputs the metadata commands to recreate the given sites"},   
  {"18", "site_add", "site_name hostname:port",
   "Registers a site"},
  {"18", "site_remove", "site",
   "Removes a site"},
  {"18", "site_set_properties", "site property value [property value]*",
   "Updates one or more configuration property of a site"},
  {"18", "site_get_properties", "site property [property]*",
   "Gets one or more configuration property of a site"},   
  {"18", "site_list_properties", "site",
   "List all configuration properties of a site"},   
  {"19", "replica_list", "[guid ... | -l lfn ...l]",
   "Lists the sites with replicas of files given as guids or lfns"},
  {"19", "replica_get", "[guid ... | -l lfn ...l]",
   "Lists the SURL replicas of files given as guids or lfns"},
  {"19", "replica_add", "guid SURL",
   "Adds a replica at SURL to the replicas of the file with the given GUID"},
  {"19", "replica_remove", "replica [guid]",
   "Removes a replica for a GUID"},
  {"19", "replica_register", "guid site",
   "Registers a site to a central catalogue that has a certain file"},
  {"19", "replica_unregister", "guid site",
   "Un-Registers a replica on a site at a central catalogue"},
  {"20", "ticket_get", "",
   "Asks for a new ticket to be issued"},
  {"20", "ticket_check", "ticket",
   "Validates a ticket"},
  {"21", "capabilities_add", "user capability",
   "Gives a user the given capability"},
  {"21", "capabilities_remove", "user capability",
   "Removes the capability from the user"},
  {"21", "capabilities_list", "user",
   "Lists all capabilities of the user"},
  {"22", "sudo", "user",
    "Changes root's identity to user"},
  {"22", "version", "",
    "Check AMGA version"}
};

MDStandalone::MDStandalone(MDBuffer &outbuf) : MDSql(outbuf)
{
  setupTableNames();

  // Add help index entries
  unsigned int maxGroup = 0;
  for(unsigned int i = 0; i < COMMANDS; i++ ) {
    struct HelpEntry e;
    e.topic = atol(commands[i][0]);
    maxGroup = maxGroup < e.topic? e.topic : maxGroup;
    e.command = commands[i][1];
    e.args = commands[i][2];
    e.help = commands[i][3];
    helpEntries.push_back(e);
  }

  // Create help topics
  if (maxGroup + 1 > helpTopics.size() )
    helpTopics.resize(maxGroup+1);

  for(int i = 0; i < COMMAND_GROUPS; i++ )
    helpTopics[atol(commandGroups[i][0])]=commandGroups[i][1];
}


MDStandalone::~MDStandalone()
{
  // Removing the connection done by MDServer
}


void MDStandalone::setupTableNames()
{
  MDServer::setupTableNames();

  std::string prefix;
  if(dbSchema.size())
    prefix = dbSchema + ".";

  guidsTable = prefix + "guids";
  replicasTable = prefix + "replicas";
  mountsTable = prefix + "mounts";
  sitesTable = prefix + "sites";
  subscribersTable = prefix + "subscribers";
  transactionsTable = prefix + "transactions";
  masterTable = prefix + "master";
  logsTable = prefix + "logs";
  usersequencesTable = prefix + "usersequences";
}


/// Checks whether a directory has subDirectories
bool MDStandalone::hasSubdirs(const struct EntryProps &p)
{
  Statement statement(*dbConn);
  // Check whether there are entries left in the table
  std::string query=subdirsExist(p.indexTable, p.name);
  DMESG("SQL: >" << query << "<" << std::endl);

  int entries;
  if(statement.exec(query)){
    printError("9 DB error", statement);
    return -1;
  }
  statement.bind(1, &entries);
  if(statement.fetch()){
    printError("9 DB error", statement);
    return -1;
  }
  statement.unBind();
  statement.close();

  DMESG("hasSubdirs: " << entries << std::endl);

  return entries;
}


std::string MDStandalone::permissionClause(char op, int flags,
					   std::list<std::string> &groups,
					   const std::string &table) const 
{

  DMESG("Generating permission clause for '" << op << "' operation on"
	<< " table '" << table << "' with flags " << flags << "\n");

  // Table we mean. Default is "" which means the main one of the query
  std::string t(table);
  if( t.size() )
    t += ".";

  // There are no access restrictions for the root user
  if(user == "root")
    return " (1=1)";

  // There are no access restrictions if an entry has no permission
  if( (flags & EP_PERMS) == 0) 
    return " (1=1)";

  if( flags & EP_ACLS && !groups.size() ) {
    memberOf(user, groups);
    groups.push_back("system:anyuser");
  }

  std::stringstream query;

  if( flags & EP_ACLS ) { // With ACLs for every entry
    switch(op) {
    case 'w':
      query << " ( (" << t << "\"owner\"='" << user << "'"
	    << " AND substr(" << t<< "\"permissions\", 2, 1)='w')";
      break;
    case 'r':
      query << " ( (" << t << "\"owner\"='" << user << "'" 
	    << " AND substr(" << t << "\"permissions\", 1, 1)='r')";
      break;
    case 'x':
      query << " ( (" << t << "\"owner\"='" << user << "')";
      break;
    case 'a':
      query << " ( (" << t << "\"owner\"='" << user << "')";
      break;
    default:
      assert(false);
    }

    query << " OR acl_test(" << t << "acls, '";
    std::list<std::string>::const_iterator I = groups.begin();
    while(I!=groups.end()) {
      if ( I!= groups.begin() )
	query << " ";
      query << (*I++);
    }
    query << "', '" << op << "')=1 ";
    query << ")";

  } else {
    switch(op) {
    case 'w':
      query << " ( (" << t << "\"owner\"='" << user << "'"
	    << " AND substr(" << t << "\"permissions\", 2, 1)='w')"
	    << " OR substr(" << t << "\"group_rights\", 2, 1)='w' )";
      break;
    case 'r':
      query << " ( (" << t << "\"owner\"='" << user << "'" 
	    << " AND substr(" << t << "\"permissions\", 1, 1)='r')"
	    << " OR substr(" << t << "\"group_rights\", 1, 1)='r' )";
      break;
    case 'x':
      query << " ( (" << t << "\"owner\"='" << user << "')"
	    << " OR ( substr(" << t << "\"group_rights\", 3, 3)='x' ) ) ";
      break;
    case 'a':
      query << " ( (" << t << "\"owner\"='" << user << "') ) ";
      break;
    default:
      assert(false);
    }
  }

  return query.str();
}


/** Creates a new directory entry in the \param indexTable.
    To do this, it locks this indexTable and picks the next available
    directory id. When addDirectoryToIndex leaves, it still has a
    transaction open, which can be closed with finalizeDirctoryToIndex().
*/
int MDStandalone::addDirectoryToIndex(const std::string dir,
				      const EntryProps &pInitial,
				      Statement &statement)
{
  std::string initialName( pInitial.directoryTable );
  std::string indexTable( pInitial.indexTable );
  std::string initialACLs( pInitial.ACLs );
  std::string initialRepAllowedUsers( pInitial.repGroups );
  int flags = pInitial.flags;

  // If this connection is being used for replicating, the ACLs
  // are not inherithed by new directories. The dump will set
  // the ACLs explicitly.
  //  string acls = replicating ? "" : initialACLs;
  int num=-1;
  // TODO: Locking unnecessary!
  // New numbers via CREATE SEQUENCE tablename_colname_seq;?????
  // Used ones with SELECT FOR UPDATE
  DMESG("addDirectoryToIndex\n");
  if(lockTable(statement, indexTable))
    return -1;

  std::string query="SELECT \"id\" FROM " + indexTable + " WHERE \"flags\"=0";

  int newID;

  DMESG("SQL: >" << query << "<" << std::endl);

  if(statement.exec(query)){
    printError("9", statement);
    return -2;
  }
  statement.bind(1, &newID);

  std::stringstream query2;

  if(statement.fetch()){
    num=-1;
  } else {
    num=newID;
    DMESG("getRecycledEntry: " << num << std::endl);
    query2 << "UPDATE " << indexTable << " SET "
	   << "\"directory\"='" << dir << "'"
	   << ", \"table_name\"='" << initialName << "'"
	   << ", \"flags\" = " << pInitial.flags
	   << ", \"main_column\" = '" << pInitial.mainColumn << "'"
	   << ", \"master\"='" << master << "'"
	   << ", \"owner_name\"='" << user << "'"
	   << ", \"permissions\"='" << permissionMask << "'"
	   << ", \"acls\"='" << initialACLs << "'"
	   << ", \"rep_groups\"='" << initialRepAllowedUsers << "'"
	   << " WHERE \"id\"=" << num << ";";
  }
  statement.unBind();
  statement.close();

  if(num<0){
    Statement statement(*dbConn);
    query="SELECT MAX(\"id\") FROM " + indexTable + ";";
    DMESG("SQL: >" << query << "<" << std::endl);
    if(statement.exec(query)){
      printError("9", statement);
      return -1;
    }

    statement.bind(1, &newID);
    if(statement.fetch()){
      printError("9 Getting max entry failed", statement);
      return -1;
      DMESG("Could not get maximum ID" << std::endl);
    } else {
      num=newID+1;
      DMESG("getNewEntry: " << num << std::endl);
      query2 << "INSERT INTO " << indexTable << " "
	     << "(\"id\", \"directory\", \"table_name\", \"flags\", \"main_column\", "
	     << "\"master\", \"owner_name\", \"permissions\", \"acls\", \"rep_groups\")"
	     << " VALUES ("
	     << num << ", '" << dir << "'" << ", '" << initialName << "', "
	     << flags << ", '" << pInitial.mainColumn << "', '" << master << "', '" 
	     << user << "', '" << permissionMask << "', '" << initialACLs << "', '" 
	     << initialRepAllowedUsers << "');";
    }
    statement.unBind();
    statement.close();
  }

  // The following is only necessary for MySQL which does not limit
  // locking to transaction like any decent database
  unlockTables(statement);

  // Now insert the change into table
  if(num>-1){
    DMESG("SQL: >" << query2.str() << "<" << std::endl);
    if(statement.exec(query2.str())){
      printError("16 Directory exists", statement);
      return -1;
    }
  }

  return num;
}


//! Finalize the creation of a directory entry
/** Commits or aborts the transaction which is being used to create
    a new directory entry, depending on whether \param num which was
    returned by addDirectoryToIndex() contains a valid ID or not.
*/
int MDStandalone::finalizeDirectoryToIndex(int num, Statement &statement)
{
  if(num <0) {
    if(statement.rollbackTransaction()){
      printError("9 Problem aborting transaction", statement);
      return -1;
    }
    return num;
  }

  // Commit everything
  DMESG("UNLOCK B " << num << std::endl);
  if(statement.commitTransaction()){
    printError("9 Problem unlocking", statement);
    return -1;
  }
  DMESG("UNLOCK SUCCEEDED " << num << std::endl);
  return num;
}


int MDStandalone::getSubDirs(const std::string &directory,
			     vector<DirLine> &dirs)
{
  string dir = absolutePath(directory, currentDir);
  
  std::string query="SELECT \"id\", \"directory\", \"table_name\", "
    "\"master\", \"owner_name\", \"permissions\", \"acls\", \"rep_groups\" FROM "
    + masterindexTable + " WHERE (\"directory\" LIKE '" + dir;    
  if (dir != "/") {
    query += "/";
  }

  std::string style = dbHelper->backend();
  if( style.find("Oracle") == std::string::npos) {
    query += "%' OR \"directory\" = '" + dir + "') AND (\"table_name\" LIKE 'dir%' or \"table_name\" = '');";
  } else {
    query += "%' OR \"directory\" = '" + dir + "') AND (\"table_name\" LIKE 'dir%' or \"table_name\" IS NULL);";
  }

  
  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    ERRLOG("Error reading masterindex: " << statement.getError() << std::endl);
    return 9;
  }
  
  DirLine dL;
  statement.bind(1, &dL.id);
  statement.bind(2, dL.directory, sizeof(dL.directory));
  statement.bind(3, dL.table, sizeof(dL.table));
  statement.bind(4, dL.master, sizeof(dL.master));
  statement.bind(5, dL.owner, sizeof(dL.owner));
  statement.bind(6, dL.permissions, sizeof(dL.permissions));
  statement.bind(7, dL.acls, sizeof(dL.acls));
  statement.bind(8, dL.repGroups, sizeof(dL.repGroups));

  while (!statement.fetch()) {
    dirs.push_back(dL);
  } 
  sort(dirs.begin(), dirs.end());

  DMESG("Fetched " << dirs.size() << " entries\n");
  for (size_t i = 0; i< dirs.size(); i++) {
    DMESG(dirs[i] << endl);
  }
  return 0;
}


void MDStandalone::unlockTables(Statement &statement)
{
  return;
}


//! Creates a table for a directory with a given schema
/** Creates a table for a directory which must already exist in an
    index table. The created table is inserted into the index table.
    Attributes are created according to the given schema
    \param id The id number of the directory
    \param indexTable Name of directory index for table
    \param attrs The list of attributes to create
    \param statement A handle to the statement which already has
    a transaction open
    \return Returns the name of the table which was created
    Any error printout is done by this function

    This implementation is in fact only used for SQLite as it is
    overridden in by all other DB frontends.
*/
std::string MDStandalone::createTable(const EntryProps &p,
				      AttrList &attrs,
				      Statement &statement)
{
  DMESG("In createTable, flags " << p.flags << std::endl);
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return "";
  }

  std::stringstream query;
  std::string mainType = "varchar(256)";
  if( typeFromFlags(p.flags) )
    mainType = typeFromFlags(p.flags);
  if( p.mainType.size() )
    mainType = p.mainType;

  query << "CREATE TABLE dir" << p.id << " (";
  string comma("");
  if( !(p.flags & EP_PLAIN ) ) {
    query << comma << "\"file\" " << dbHelper->mdToDBType(mainType);
    comma = ", ";

    if( p.flags & EP_PERMS ) {
      query << comma << " \"owner\" VARCHAR(64),"
            << " \"permissions\" CHAR(3),"
            << " \"group_rights\" CHAR(3)";
      comma = ", ";
    }
    if( p.flags & EP_FCAT ) {
      query << comma 
        << "\"link\" VARCHAR(1024),"
        << " \"created\" TIMESTAMP,"
        << " \"size\" NUMBER(38),"
        << " \"guid\" " << dbHelper->guidType() << ","
        << " \"md5\" VARCHAR(32)";
      comma = ", ";
    }

    if (p.flags & EP_SHARED) {
      query << comma << "\"dir\" INTEGER REFERENCES " 
            << p.indexTable << "(\"id\")"
            << " ON DELETE CASCADE";
      comma = ", ";
    }
    if (p.flags & EP_ACLS) {
      query << comma << "\"acls\" VARCHAR(2048)";
      comma = ", ";
    }
  }

  AttrList::const_iterator I=attrs.begin();    
  while(I!=attrs.end()) { 
    std::string key;
    if(parser->parse(I->first, key ,p.flags & EP_PLAIN)){
      out.append("7 Illegal Key\n");
      return "";
    }
    query << comma << key << " " << (dbHelper->mdToDBType((*I).second));
    comma = ", ";
    I++;
  }

  if( !(p.flags & EP_PLAIN ) ) {
    if( p.flags & EP_SHARED )
      query << comma << "PRIMARY KEY (\"" << p.mainColumn << "\", \"dir\")";
    else {
      query << comma << "PRIMARY KEY (\"" << p.mainColumn << "\")";
    }    
  }

  query << ");";

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if(statement.exec(query.str())){
    printError("9 Could not alter table", statement);
    return "";
  }
  if(commitDirectoryToIndex(p.id, p.indexTable, statement, p.flags)<0){
    return "";
  }
  if( statement.commitTransaction() ) {
    printError("9 Cannot commit transaction", statement);
    return "";
  }

  query.str("");
  query << "dir" << p.id;

  return query.str();
}


//! Drops a Table
int MDStandalone::dropTable(Statement &statement, const std::string &tableName)
{
  std::string query="DROP TABLE " + tableName + ";";
  DMESG("SQL: >" << query << "<" << std::endl);
  return statement.exec(query);
}


//! Discovers properties of an entry
/** This function fills out the EntryProps structure.
    For directories props.name is the full path of the directory on
    that volume, for entries it contains the entry name. In both cases
    parent is set to the parent directory.
*/
int MDStandalone::getEntryProps(const std::string &entry,
				EntryPropList &entryProps,
				const std::string &indexTable,
				bool enableProxy)
{
  std::string e, aPath, idxTable=indexTable;
  DMESG("Absolute path entry: " << entry  << std::endl);
  DMESG("Absolute path c_dir: " << currentDir << std::endl);
  if(indexTable == ""){
    idxTable = masterindexTable;
    aPath = absolutePath(entry, currentDir);
  } else {
    aPath=entry;
  }
  bool pattern=dirPattern(aPath, e);
  DMESG("Absolute path: " << aPath << ", Pattern: " << pattern << std::endl);
  std::string aPathRegExpText;
  dirRegexpPattern(aPath, aPathRegExpText);
  regex_t aPathRegExp;
  if(regcomp(&aPathRegExp, aPathRegExpText.c_str(), REG_EXTENDED|REG_NEWLINE) ) {
    DMESG("Illegal regexp: " << aPathRegExpText << std::endl);
    return 1;
  }

  std::list< std::string > parts;
  pathPieces(aPath, parts);

  std::ostringstream query;
  query << "SELECT \"id\", \"directory\", \"table_name\", \"flags\", \"main_column\", "
	<< "\"master\", \"owner_name\", \"permissions\", \"acls\", \"rep_groups\" FROM "
	<< idxTable << " WHERE ";
  std::list< std::string >::iterator I=parts.begin();
  std::list< const char * > variables;
  int nParts = 0;
  while( I !=parts.end() ) {
    nParts++;
    if(I == --parts.end() && pattern) {
      // Match directory pattern
      if(opRegExpMatch()=="REGEXP_LIKE")  // Oracle
	query << "REGEXP_LIKE(\"directory\", ?)";
      else
	query << "(\"directory\" " << opRegExpMatch() << " ?)";
      variables.push_back(e.c_str());
    } else {
      query << "(\"directory\" = ?)";
      variables.push_back((*I).c_str());
    }
    
    I++;
    if(I != parts.end())
      query << " OR ";
  }
  query << ";";

  DMESG("Preparing SQL: >" << query.str() << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.prepare(query.str())){
    std::string msg = statement.getError();
    std::string sqlErr = statement.getSqlState();
    ERRLOG("Error reading masterindex: " << msg << std::endl);
    return 9;
  }

  std::list< const char * >::iterator II=variables.begin();
  SQLLEN vTypes[nParts];
  int i=0;
  while( II !=variables.end() ) {
    i++;
    vTypes[i] = SQL_NTS;
    if (statement.bindParam(i, *II, 0, &(vTypes[i]))) {
      printError("9 Could not bind parameters for prepared statement", statement);
      return 9;
    }
    II++;
  }

  if(statement.execPrepared()){
    regfree(&aPathRegExp);
//    int err = statement.getErrorCode();
    std::string msg = statement.getError();
//    std::string sqlErr = statement.getSqlState();
    ERRLOG("Error reading masterindex: " << msg << std::endl);
    return 9;
  }
  DirLine dL;
  statement.bind(1, &dL.id);
  statement.bind(2, dL.directory, sizeof(dL.directory));
  statement.bind(3, dL.table, sizeof(dL.table));
  statement.bind(4, &dL.flags);
  statement.bind(5, dL.mainColumn, sizeof(dL.mainColumn));
  statement.bind(6, dL.master, sizeof(dL.master));
  statement.bind(7, dL.owner, sizeof(dL.owner));
  statement.bind(8, dL.permissions, sizeof(dL.permissions));
  statement.bind(9, dL.acls, sizeof(dL.acls));
  statement.bind(10, dL.repGroups, sizeof(dL.repGroups));
  
  if(statement.fetch()){
    regfree(&aPathRegExp);
    return 1;
  }
  std::list< DirLine > dirLines;
  do {
    dirLines.push_back(dL);
  } while (!statement.fetch());

  statement.close();

  dirLines.sort();

  for(list<DirLine>::iterator K=dirLines.begin(); K!=dirLines.end(); ++K)
    DMESG(">> " << K->directory << std::endl);

  int error=0;
  std::list< DirLine >::iterator J=dirLines.begin();
  bool addPatternEntry = false;
  std::string parentDirectory;

  bool proxyState = false;

  do { // Loops _all_ returned directories: parent and matching pattern
    EntryProps props;

    // Copy parent dir properties from the last entry, this works
    // because we did the sorting above
    if(entryProps.size()) {
      props=entryProps.back();
      props.name = props.parent;
      props.id = props.parentID;
      props.perms = props.parentPerms;
      props.ACLs = props.parentACLs;
      props.repGroups = props.parentRepGroups;
      props.owner = props.parentOwner;
      props.flags = props.parentFlags;
    }

    props.indexTable=idxTable;

    do {  // Descend directory hierarchy
      DMESG("Directory: >" << (*J).directory << "<" << std::endl);

      if(!strncmp((*J).permissions, "del", 3)){
	error=1;
	break;
      }

      // Set directory type
      DMESG("Matching: " << e << " " << (*J).directory << std::endl);
      regmatch_t match[100];
      if(!regexec(&aPathRegExp, (*J).directory, 100, match, 0)){
	DMESG("Match: Is dir" << std::endl;);
	props.flags = EPT_DIR;
	if( !addPatternEntry ) {
	  addPatternEntry = pattern;
	  parentDirectory = props.directoryTable;
	  DMESG("Setting parentDirectory:" << parentDirectory << "\n");
	}
      }

      if(!strcmp((*J).directory, "/")) {
	props.parent = (*J).directory;
	props.parentID = (*J).id;
	props.parentPerms = (*J).permissions;
	props.parentACLs = (*J).acls;
	props.parentRepGroups = (*J).repGroups;
	props.parentOwner = (*J).owner;
	props.parentFlags = (*J).flags;
      } else {
	props.parent = props.name;
	props.parentID = props.id;
	props.parentPerms = props.perms;
	props.parentACLs = props.ACLs;
	props.parentRepGroups = props.repGroups;
	props.parentOwner = props.owner;
	props.parentFlags = props.flags;
      }

      props.name = (*J).directory;
      props.id = (*J).id;
      props.flags = (*J).flags;
      props.mainColumn = (*J).mainColumn;
      if( props.mainColumn.empty() )
	props.mainColumn = DEFAULT_MAIN_COLUMN;
      props.perms = (*J).permissions;
      props.owner = (*J).owner;
      props.ACLs = (*J).acls;
      props.repGroups = (*J).repGroups;
      props.directoryTable = (*J).table;
      props.master=(*J).master;
      
      std::string table = props.name;
      props.directoryTable = qualifySchema(props.directoryTable, dbSchema);

      strcpy((*J).table, table.c_str());
      DMESG("Setting flags to " << props.flags << std::endl);

      if( enableProxy && (props.flags & EP_PROXY ) ) {
        int ret = printProxy(props);
        if ( ret >= 0 ) {
          if (ret == PROXY_CODE) {
            proxyState = true;
          } else {
            error = ret ;
          }
          break;
        }
      }

      J++;
    } while (J != dirLines.end() && isSubdirOf((*J).directory, props.name));


    if(error){
      regfree(&aPathRegExp);
      return error;
    }

    // Check for volumes
    if( (props.flags & EP_TYPE) == EPT_VOLUME ) {
      // Need to continue in next index table?
      if(props.name.size()<aPath.size()) {
	aPath = aPath.substr(props.name.size()+1);
	regfree(&aPathRegExp);
	return getEntryProps(aPath, entryProps, props.directoryTable);
      }
    }

    // Check for entries
    DMESG("Checking for dir >" << props.name << "< vs >" << aPath << "<\n");
    if( dirLevel(props.name)+1 == dirLevel(aPath) ) {
      DMESG("IS AN ENTRY" << std::endl);
      props.parent = props.name;
      props.parentID = props.id;
      props.parentPerms = props.perms;
      props.parentACLs = props.ACLs;
      props.parentRepGroups = props.repGroups;
      props.parentOwner = props.owner;
      props.parentFlags = props.flags;
      props.flags = EPT_NONE;

      if(props.parent.size()==1)
	props.name=aPath.substr(props.name.size());
      else
	props.name=aPath.substr(props.name.size()+1);
      props.print();
    } else {
      DMESG("IS A DIR" << std::endl);
      // Entry below entry or non-exisiting dir
      if( dirLevel(aPath) > dirLevel(props.name)+1 )
        if ( ! proxyState )
	  return 1;
    }

    props.print();
    entryProps.push_back(props);

    if( proxyState ) return 100;

  } while (J != dirLines.end());


  if( addPatternEntry ) {
    DMESG("Adding additional entry to match the pattern\n");
    EntryProps props(entryProps.back());
    size_t pos = aPath.rfind("/");
    if ( pos != std::string::npos ) {
      props.name=aPath.substr(pos+1);
      props.id = props.parentID;
      props.flags = 0;
      props.directoryTable=parentDirectory;
    }
    //    props.print();
    entryProps.push_back(props);
  }

  regfree(&aPathRegExp);


  return error;
}



int MDStandalone::commitDirectoryToIndex(int num, const std::string &indexTable,
					 Statement &statement, int flags)
{
  std::ostringstream query;

  query << "UPDATE " << indexTable 
	<< " SET \"table_name\" = 'dir" << num << "', "
	<< " \"flags\" = " << flags
  	<< " WHERE \"id\"=" << num
	<< ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if(statement.exec(query.str())){
    printError("16 Directory exists", statement);
    return -1;
  }

  return num;
}


//! Removes the link to a table from a directory in the masterindex
/** Changes the table_name field in the directory index of a table to 'x' prior to
    dropping the table itself in rmObject.
*/
int MDStandalone::removeTableFromIndex(const std::string &dir,
				       const std::string &indexTable,
				       Statement &statement)
{
  std::stringstream query;

  query << "UPDATE " << indexTable 
	<< " SET \"table_name\" = '',"
	<< " \"flags\" = 0"
        << " WHERE \"directory\"='" << dir
	<< "';";

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if(statement.exec(query.str())){
    printError("9 DB Error", statement);
    return 1;
  }
  return 0;
}


//! Sets attributes of all file matching pattern
/** Sets a list of attributes to a list of values for a file.
    The dimensions of the input vectors for the attributes and values
    need to be the same.
    \param path The relative or absolute path to the file
    \param attributes The vector of attributes to set
    \param values The vector of values

    ERRORS:
    - 1 File or directory not found
    - 3 Illegal command
      The attributes and values vectors don't match
    - 4 Permission Denied
*/
void MDStandalone::setAttr(const std::string &path,
       std::vector< std::string > &attributes,
       std::vector< std::string > &values)
{
  if(attributes.size() != values.size()) {
    out.append("3 Invalid arguments\n");
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());
  std::string dir=p.parent;
  std::list< std::string > groups;

  int perms = checkPermissions(p, 'w', groups);
 
  int t_flags = p.type() ? p.flags : p.parentFlags;
  if(perms <= 0 && !(t_flags & EP_ACLS)){
    DMESG("Permission denied: " << perms << " " << p.flags << std::endl);
    out.append("4 Permission denied\n");
    return;
  }
  
  // No attributes to set, simply return
  if (attributes.empty()) {
    out.append("0\n");
    return;
  }

  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::stringstream query;
  query << "UPDATE " << p.directoryTable << " SET ";
  for(unsigned int i=0; i<attributes.size(); i++){
    if(parser->parse(attributes[i], keys[i])){
      out.append("7 Illegal Key\n");
      return;
    }
    query << keys[i] << "='" << values[i] << "'";
    if(i<attributes.size()-1)
      query << ", ";
  }
  std::string pattern;
  bool isPattern=sqlPattern(p.name, pattern);
  if( !(isPattern && pattern == "%") && (p.type() == EPT_NONE) ) {
    query << " WHERE \"file\"";
    if(isPattern)
      query << " LIKE '";
    else
      query << " = '";
    query << pattern << "' AND ";
  } else {
    query << " WHERE ";
  }
  if( p.flags & EP_SHARED )
    query << " \"dir\" = " << p.parentID << " AND ";

  query << permissionClause('w', t_flags, groups) << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  Statement statement(*dbConn);
  // [rep] Open transaction for replication
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;  // Transactions are automatically rolled back
  }
  
  if(statement.exec(query.str())){
    int err = statement.getErrorCode();
    std::string msg = statement.getError();
    std::string sqlErr = statement.getSqlState();
    DMESG(err << "msg >" << msg << "< sqlErr: " << sqlErr << std::endl);
    if ( statement.numRows() == 0 ) { 
      out.append("1 No matching entry or none with correct permissions \n");
    } else {
      out.append("10 No such key\n");
    }
    return;
  } 
  if(statement.numRows()==0) {
    out.append("1 No matching entry or none with correct permissions\n");
    return;
  }
  
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}



void MDStandalone::createSchema(const std::string &name,
				AttrList &attrs,
				const std::string &option)
{
  createDir(name, attrs, option);
}


void MDStandalone::removeSchema(const std::string &name)
{
  rmDir(name);
}


bool MDStandalone::completeAddEntryAttrs(std::vector< std::string > &attrs,
					 std::vector< std::string > &values,
					 const std::string &attribute,
					 const std::string &value)
{
  if( findAmongAttributes(attrs, attribute) )
    return false;

  attrs.insert(attrs.begin(), attribute);
  values.insert(values.begin(), value);

  return true;
}


//! Adds an entry with it's attributes to a directory
/** Inserts an entry with a list of attributes and a list of
    correspondin values.
    The dimensions of the input vectors for the attributes and values
    need to be the same.
    \param path The relative or absolute path to the file
    \param attributes The vector of attributes to set
    \param values The vector of values
    \param eval Whether the values are supposed to be evaluated, first.

    This function (for eval = true) also implements the insert command
    which works like the INSERT command in SQL in that the values are
    evaluated prior to insertion.

    ERRORS:
    - 1 File or directory not found
    - 3 Illegal command
      The attributes and values vectors don't match
    - 4 Permission denied
    - 7 Illegal key
*/
void MDStandalone::addEntry(const std::string &path,
			    std::vector< std::string > &attributes,
			    std::vector< std::string > &values, bool eval)
{
  if(attributes.size() != values.size()) {
    out.append("3 Illegal command\n");
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  std::string dir=p.parent;

  std::list< std::string > groups;
  if (!checkPermResult(checkDirPermissions(p, 'w', groups)))
    return;

  DMESG("Permissions OK, flags for dir are " << p.parentFlags << "\n");

  std::vector<bool> evaluate(attributes.size());
  for(unsigned int i=0; i < evaluate.size(); i++)
    evaluate[i]=eval;

  if( p.parentFlags & EP_PERMS ) {
    DMESG("Adding permissions\n");
    if( completeAddEntryAttrs(attributes, values, "OWNER", user) )
      evaluate.insert(evaluate.begin(), false);
    if( completeAddEntryAttrs(attributes, values, "PERMISSIONS", permissionMask) )
      evaluate.insert(evaluate.begin(), false);
    if( completeAddEntryAttrs(attributes, values, "GROUP_RIGHTS", groupRightsMask) )
      evaluate.insert(evaluate.begin(), false);
  }

  std::string surl;
  std::string guid;
  std::string expires;
  // When inserting into FC table, make sure all attributes are there
  if( p.parentFlags & EP_FCAT ) {
    DMESG("Adding to FC table\n");
    if( completeAddEntryAttrs(attributes, values, "SIZE", "-1") )
      evaluate.insert(evaluate.begin(), false);
    if( completeAddEntryAttrs(attributes, values, "GUID", "gen") )
      evaluate.insert(evaluate.begin(), false);
    if( completeAddEntryAttrs(attributes, values, "CREATED", "now") )
      evaluate.insert(evaluate.begin(), false);
    if( completeAddEntryAttrs(attributes, values, "MD5", "NULL") )
      evaluate.insert(evaluate.begin(), false);
  }

  // Validate and expand attributes
  for( size_t i = 0; i < attributes.size(); i++) {
    if( attributes[i] == "GUID" ) {
      if( user != "root" ) {
	if( values[i] != "gen" && values[i].substr(0,3) != "mac" ) {
	  DMESG("Permission denied to set GUID\n");
	  out.append("4 Permission denied to set GUID\n");
	  return;
	}
      }
      try {
	guid = values[i] = AMGA::validateGUID(values[i]);
      } catch (std::runtime_error &e) {
	DMESG("Illegal guid!\n");
	out.append("32 Illegal GUID: ");
	out.append(e.what());
	out.append("\n");
	return;
      }
    }
    if( attributes[i] == "CREATED" ) {
      if( values[i] == "now" ) {
	values[i] = makeTimestamp();
      }
    }
  }

  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::stringstream query;
  query << "INSERT INTO " << p.directoryTable << " (";

  query << "\"" << p.mainColumn << "\"";
  if( p.parentFlags & EP_SHARED )
    query << ", \"dir\"";

  for(unsigned int i=0; i<attributes.size(); i++){
    if( attributes[i] == "SURL" ) {
      surl = values[i];
      continue;
    }
    if( attributes[i] == "EXPIRES" ) {
      expires = values[i];
      continue;
    }
    if(checkAttribute(attributes[i], keys[i], true, p.parentFlags & EP_PLAIN)){
      out.append("7 Illegal Key:" + attributes[i] + "\n");
      return;
    }
    query << ", " << keys[i];
  }

  query << ") VALUES (";

  query << "'" << p.name << "'";

  if( p.parentFlags & EP_SHARED )
    query << ", '" << p.parentID << "'";

  for(unsigned int i=0; i<values.size(); i++) {
    if( attributes[i] == "SURL" ||  attributes[i] == "EXPIRES")
      continue;
    query << ", ";

    if( evaluate[i] ) {
      std::string value;
      if(parser->parse(values[i], value)){
	out.append("9 Illegal value: " + values[i] + "\n");
	return;
      }
      values[i]=value;
      query << values[i];
    }
    else
      query << "?";
  }

  query << ");";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  
  Statement statement(*dbConn);
 // [rep] Open transaction for replication
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;  // Transactions are automatically rolled back
  }

  if(statement.prepare(query.str())){
    out.append("9 Internal error preparing insert\n");
    return;
  }
  
  std::vector <SQLLEN> types(values.size());
  std::vector <SQLINTEGER> sizes(values.size());
  int j = 1;
  for(unsigned int i=0; i<values.size(); i++) {
    if( attributes[i] == "SURL" ||  attributes[i] == "EXPIRES")
      continue;
    
    unescapePreparedString(values[i]);

    DMESG("Value " << i << ": >" << values[i] << "<\n");

    //    sizes[i] = values[i].size();
    sizes[i]=0;
    types[i] = SQL_NTS;
    if (statement.bindParam(j++, values[i].c_str(), sizes[i], &types[i])) {
      printError("9 Could not bind parameters for prepared statement", statement);
      return;
    }
  }

  if(statement.execPrepared()){
    int err = statement.getErrorCode();
    std::string msg = statement.getError();
    std::string sqlErr = statement.getSqlState();
    if(msg.find("unique") != std::string::npos // in postgres/sqlite
       || msg.find("Duplicate") != std::string::npos) { // MySQL
      printError("15 Entry exists", statement);
      return;
    }
    out.append("10 No such key: " + msg + "\n");
    return;
  }

  if( surl.size() ) {
    query.str("");
    query << "INSERT INTO " << replicasTable << " (\"guid\", \"last_access\", "
	  << "\"expires\", \"surl\") VALUES ("
	  << dbHelper->writeGUID(guid) << ", '" << makeTimestamp() << "', ";
    if( expires.size() )
      query << "'" << expires << "', ";
    else
      query << "NULL, ";
    query << "'" << surl << "');";
    if(statement.exec(query.str())){
      printError("9 Internal error", statement);
      return;  // Transactions are automatically rolled back
    }
  }
  
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


/** Adds a list of entries to several directories, the entries must not
    exist before
*/
void MDStandalone::addEntries(std::vector<std::string> &paths)
{
  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;  // Transactions are automatically rolled back
  }
  
  // [Rep] Saves a list of the parent directories of each entry.
  // Needed for the replication logs
  string parents[paths.size()];

  std::list< std::string > groups;
  for(size_t entry=0;  entry<paths.size(); entry++) {

    EntryPropList propList;
    int error=getEntryProps(paths[entry], propList);
    if(handleEntryPropErrors(error, paths[entry]))
      return;

    EntryProps p(*propList.begin());
    DMESG("Entry props: name >" << p.name
	  << "<  parent >"  << p.parent << "<"  << std::endl);
    parents[entry] = p.parent;
    
    if(error && error!=1 && handleEntryPropErrors(error, paths[entry]))
      return;

    if (!checkPermResult(checkPermissions(p, 'w', groups))) {
      return;
    }

    if(inSubdir(p.name) || (p.type() == EPT_DIR) ) {
      out.append("1 No such file or directory: " + paths[entry] + "\n");
      return;
    }


    std::string tableName=p.directoryTable;
    if( !tableName.size() ) {
      AttrList attrs;
      p.flags = p.parentFlags;
      p.id = p.parentID;
      tableName = p.directoryTable = createTable(p, attrs, statement);
    }
    if(!tableName.size())
      return;

    std::stringstream query;
    query << "INSERT INTO " << tableName << " (\"" << p.mainColumn << "\"";
    if( p.parentFlags & EP_SHARED )
      query << ", \"dir\"";
    if( p.parentFlags & EP_PERMS )
      query << ", \"owner\", \"permissions\", \"group_rights\"";

    query << ") VALUES (";
    query << "'" << p.name + "'";

    if( p.parentFlags & EP_SHARED )
      query << ", "<< p.parentID;
    if( p. parentFlags & EP_PERMS )
      query << ", '" << user << "' ,'" << permissionMask << "' ,'" 
	    << groupRightsMask << "'";

    query << ");";

    DMESG("SQL: >" << query.str() << "<" << std::endl);

    if( statement.exec(query.str()) ){
      printError("15 Entry exists", statement);
      return;
    }
  }
  
  // [rep] Sucessful operation. Write the logs
  // For each entry, add a new command
  try {
    for(size_t i=0;  i<paths.size(); i++) {
      string command = "addentries ";
      command += paths[i];
      logWriter->saveLogCommand(statement, command, parents[i], user, currentDir, permissionMask, groupRightsMask);
    }
    statement.commitTransaction();
    out.append("0\n");
  } catch (std::runtime_error &ex) {
      printError(std::string("9 ") + ex.what(), statement);     
  }
}


/** Removes a list of entries from several directories
    This operation is handled as a transaction, so either all entries
    get removed, or none in case something goes wrong with an entry
    \param paths : list of paths
*/
void MDStandalone::removeEntries(std::vector<std::string> &paths)
{
  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;  // Transactions are automatically rolled back
  }
  string parents[paths.size()];
  std::list< std::string > groups;
  for(size_t entry=0;  entry<paths.size(); entry++) {

    EntryPropList propList;
    int error=getEntryProps(paths[entry], propList);
    if(handleEntryPropErrors(error, paths[entry]))
      return;
    EntryProps p(*propList.begin());

    DMESG("Entry props: name >" << p.name
	  << "<  parent >"  << p.parent << "<"  << std::endl);
    parents[entry] = p.parent;

    if(error && error!=1 && handleEntryPropErrors(error, paths[entry]))
      return;

    if (!checkPermResult(checkPermissions(p, 'w', groups))) {
      return;
    }

    if( inSubdir(p.name) ) { /////  TODO  || p.type() == EPT_DIR ) {
      out.append("1 No such file or directory: ");
      out.append(paths[entry]);
      out.append("\n");
      return;
    }

    std::string pattern;
    bool isPattern = sqlPattern(p.name, pattern);
    std::stringstream query;
    query << "DELETE FROM " << opOnly() << " " << p.directoryTable
	  << " WHERE " << "\"" + p.mainColumn + "\"";

    if(isPattern)
      query << " LIKE '";
    else
      query << " = '";
    query << pattern  << "'";
    if( p.parentFlags & EP_SHARED )
      query << " AND \"dir\" = " << p.parentID;

    std::list< std::string > groups; 
    int t_flags = p.type() ? p.flags : p.parentFlags;
    if( p.sticky() ) {
      query << " AND " << permissionClause('a', t_flags, groups) << ";";
    }  else {
      query << " AND " << permissionClause('w', t_flags, groups) << ";";
    }  
    
    DMESG("SQL: >" << query.str() << "<" << std::endl);

    if(statement.exec(query.str())){
      printError("1 Entry does not exist", statement);
      return;
    }
  }

  // [rep] Sucessful operation. Write the logs
  // For each entry, add a new command
  try {
    for(size_t i=0;  i<paths.size(); i++) {
      string command = "removeentries ";
      command += paths[i];
      logWriter->saveLogCommand(statement, command, parents[i], user, currentDir, permissionMask, groupRightsMask);
    }
    statement.commitTransaction();
    out.append("0\n");
  } catch (std::runtime_error &ex) {
      printError(std::string("9 ") + ex.what(), statement);     
  }
}


/** Gets attribute(s) of a number of files
    path is the path to the files where wildcards are only allowed in
    the file part of the path
    attributes is the list of attributes to which need to be returned
    The output including all errors is streamed into the out class
    member. For a description see the protocol description.
*/
void MDStandalone::getAttr(const std::string &path,
			   std::vector< std::string > &attributes)
{
  std::list< std::string > groups; // Caching of groups user is in

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;

  EntryProps p(*propList.begin());
  int perms;
  if( (p.flags & EP_TYPE) == 0 )
    perms = checkPermissions(p, 'x', groups);
  else
    perms = checkDirPermissions(p, 'x', groups);

  if( perms <=0 ){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  if( p.flags & EP_PLAIN ) {
    out.append("29 Operation not permitted on plain table\n");
    return;
  }

  if(!tableExists(p.directoryTable, p.type()))
    return;

  std::string pattern;
  bool isPattern = sqlPattern(p.name, pattern);
  if( p.flags & EP_TYPE ) {
    isPattern = true;
    pattern = "%";
  }

  std::stringstream query;
  query << "SELECT \"file\", ";

  for(size_t i=0; i<attributes.size(); i++){
    std::string key;
    if(parser->parse(attributes[i], key)){
      out.append("7 Illegal Key\n");
      return;
    }
    LOG("GetAttr ATT : " << attributes[i] << " : " << key << "\n");
    query << key;
    if(i<attributes.size()-1)
      query << ", ";
  }
  query << " FROM " << p.directoryTable;
  if( !(isPattern && pattern == "%") ) {
    query << " WHERE \"file\"";
    if(isPattern)
      query << " LIKE '";
    else
      query << " = '";
    query << pattern << "' AND ";
  } else {
    query << " WHERE ";
  }
  if( p.flags & EP_SHARED )
    query << " \"dir\" = " << p.parentID << " AND ";

  int t_flags = p.type() ? p.flags : p.parentFlags;
  query << permissionClause('r', t_flags, groups) << ";";

  // query.append(command[2]).append(" is not null;");

  std::string queryS(query.str());
  DMESG("SQL: >" << queryS << "<" << std::endl);

  Statement statement(*dbConn);
  bool foundSomething=false;
  doBulkRead(statement, queryS, attributes.size()+1, foundSomething);

  if(!foundSomething){
    out.append("1 No matching entry or none with correct permissions\n");
    return;
  }
}


/** Clears an \param attribute of files matching \param path .
    The output is streamed into the out class member. For a
    description see the protocol description.
*/
void MDStandalone::clearAttr(const std::string &path,
			     const std::string &attribute)
{
  std::list< std::string > groups; // Caching of groups user is in

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  if( p.flags & EP_PLAIN ) {
    out.append("29 Operation not permitted on plain table\n");
    return;
  }

  int perms = checkPermissions(p, 'w', groups);

  int t_flags = p.type() ? p.flags : p.parentFlags ;
  if(perms <= 0 && !(t_flags & EP_ACLS)){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  std::string key;
  if(parser->parse(attribute, key)){
    out.append("7 Illegal Key\n");
    return;
  }

  std::string pattern;
  bool isPattern = sqlPattern(p.name, pattern);
  if( p.flags & EP_TYPE) {
    isPattern = true;
    pattern="%";
  }

  std::stringstream query;
  query << "UPDATE " << p.directoryTable << " SET " << key << "=NULL";
  if( !(isPattern && pattern == "%") ) {
    query << " WHERE \"file\"";
    if(isPattern)
      query << " LIKE '";
    else
      query << " = '";
    query << pattern << "' AND ";
  } else {
    query << " WHERE ";
  }
  if( p.flags & EP_SHARED )
    query << " \"dir\" = " << p.parentID << " AND ";

  query << permissionClause('w', t_flags, groups) << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);

  Statement statement(*dbConn);  
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  if(statement.exec(query.str())){
    printError("1 No matching entry or none with correct permissions",
	       statement);
    return;
  }
  
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, getEntryDirectory(p))) {
      statement.commitTransaction();  
      out.append("0\n");
  }
}


/** Finds files matching path and the query which is in the query
    language described in QParser.
    The output including all errors is streamed into the out class
    member. For a description see the protocol description.
*/
void MDStandalone::find(const std::string &path, const std::string &query)
{
  std::list< std::string > groups; // Caching of groups user is in
  char buff[1024];

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  int perms=checkPermissions(p, 'r', groups);

  int t_flags = p.type() ? p.flags : p.parentFlags ;
  if(perms <= 0 && !(t_flags & EP_ACLS)){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }

  if( p.flags & EP_PLAIN ) {
    out.append("29 Operation not permitted on plain table\n");
    return;
  }

  if( p.directoryTable.empty() ) {
    out.append("0\n");
    return;
  }
  
  if(!tableExists(p.directoryTable, p.type() ))
    return;

  if (!checkPermResult(checkDirPermissions(p, 'r', groups))) {
    return;
  }

  if(parser->parse(query)){
    out.append("8 Illegal Query\n");
    return;
  }
  TableMap tableMap;
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap

  std::string q;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q);

  std::string pattern;
  bool isPattern=sqlPattern(p.name, pattern);
  if( p.flags & EP_TYPE) {
    isPattern = true;
    pattern = "%";
  }

  std::stringstream queryString;
  queryString << "SELECT \"file\" FROM " << p.directoryTable;
  if( !(isPattern && pattern == "%") ) {
    queryString << " WHERE \"file\"";
    if(isPattern)
      queryString << " LIKE '";
    else
      queryString << " = '";
    queryString << pattern << "' AND ";
  } else {
    queryString << " WHERE ";
  }
  if(q.size()) {
    queryString << " (" << q << ") AND ";
  }
  if( p.flags & EP_SHARED )
    queryString << " \"dir\" = " << p.parentID << " AND ";

  queryString << permissionClause('r', t_flags, groups) << ";";

  DMESG("SQL: >" << queryString.str() << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(queryString.str())){
    printError("9 Internal error:", statement);
    return;
  }
  statement.bind(1, buff, 1024);

  out.append("0");

  /* Fix 20090728 : Return #column instead of #rows */
  out.append("\n");
  /*
  int rows=statement.numRows();
  if(rows>-1) {
    char buff[20];
    sprintf(buff, " %d\n", rows);
    out.append(buff);
  } else {
    out.append("\n");
  }
  */

  while(1){
    if(statement.fetch())
      break;

    out.append(buff);
    out.append("\n");
  }
}


// Checks whether an entry exists. Returns -1 if it exists, 0 if it
// does not exist, and a positive error code if something goes wrong
int MDStandalone::entryExists(EntryProps &p)
{
  if( !p.directoryTable.size() )
    return 0;

  std::string query = "SELECT \"file\" from ";
  query += p.directoryTable;
  query += " WHERE \"file\" = '";
  query += p.name;
  query += "';";

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    return 9;
  }

  if( statement.fetch() ) {
    return 0;
  }

  return -1;
}


int MDStandalone::prepareDirlikeEntry(const std::string &directory, 
				      EntryProps &p, std::string &name)
{
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return 9;
  p = *propList.begin();

  if( p.flags & EP_TYPE ){
    out.append("16 Directory exists\n");
    return 16;
  }

  // Check the permissions of this dir, creating a directory is 
  // like adding an entry!
  
  std::list< std::string > groups;
  if (!checkPermResult(checkPermissions(p, 'w', groups))) {
    return 4;
  }

  name=p.parent;
  if(name=="/")
    name.append(p.name);
  else
    name.append("/").append(p.name);
  DMESG("Checking if entry exists\n");

  int res = entryExists(p);
  if( res < 0 ) {
    out.append("15 Entry exists\n");
    return 15;
  }

  if( res > 0 ) {
    ERRLOG("Failed to check existence of " << p.name << "\n");
  }

  DMESG("Entry does not exists\n");

  return 0;
}


/** Creates a new directory if it does not yet exist but parentdir
    already exists.
    The directory is created with the current owner and the same ACLs
    as the parent directory.
    \param directory name of the new directory
    \param inherit Whether the schema of the parent directory is to be
    inherited
*/
void MDStandalone::createDir(const std::string &directory,
			     AttrList &attrs, 
			     const std::string &option)
{
  DMESG("createDir()\n");

  if( option.find("file") != std::string::npos 
      && option.find("plain") != std::string::npos ) {
    out.append("1 Cannot create a plain table for files\n");
    return;
  }
  
  EntryProps p;
  std::string name;
  if( prepareDirlikeEntry(directory, p, name) ) return;

  p.mainType = "varchar(256)";
  decodeTableOptions(option, p);

  if( p.flags & EP_PLAIN ) {
    if( p.mainColumn.compare("file") == 0 ) {
      out.append("9 For a plain Directory, main column & type should be given. For example, plain;main=id;type=int\n");
      return;
    }
  }

  if( p.parentFlags & EP_SHARED )
    p.flags = p.parentFlags;
  else
    p.flags |= EPT_DIR;

  std::string parent = p.directoryTable;
  p.directoryTable = "";

  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;
  }

  p.id = addDirectoryToIndex(name, p, statement);
  
  if( (p.flags & EP_INHERIT) && p.id >=0 ) {
    p.print(true);
    p.inheritsFrom = parent;
    if(createTable(p, attrs, statement) == "")
      return;
  }
      
  int id = finalizeDirectoryToIndex(p.id, statement);
  if( id >=0 ) {
    // [rep] Sucessful operation. Write the log
    if (saveLog(statement, p.parent)) {
      statement.commitTransaction();  
      out.append("0\n");
    }
  }
}


/** Creates a new directory if it does not yet exist but parentdir
    already exists and creates a table with the given attribute schema
    The directory is created with the current owner and the same ACLs
    as the parent directory.
    \param directory name of the new directory
    \param inherit Whether the schema of the parent directory is to be
    inherited
*/
void MDStandalone::createDir(const std::string &directory, const std::string &option)
{
  AttrList attrs;
  createDir(directory, attrs, option);
}


/** List all entries in a directory \param dir
    Directory can also be a pattern of directory or a pattern for
    entries in a directory 
*/
void MDStandalone::listEntries(const std::string &dir)
{
  std::list< std::string > groups; // Caching of groups user is in
  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryPropList::iterator I=propList.begin();
  // Whether something was written to the client.
  bool hasPermission = false;
  while(I != propList.end()) {
    EntryProps p=*I;
    int perms = checkPermissions(p, 'x', groups);
    if(  perms <=0 && !(p.flags & EP_ACLS) ) {
      I++;
      continue;
    }
    hasPermission = true;

    if( I == propList.begin() )
      out.append("0\n");

    Statement statement(*dbConn);
    std::string query;

    if( !p.directoryTable.size() ){
      I++;
      continue;
    }

    // Now come the files in the directory
    std::string pattern;
    bool isPattern = sqlPattern(p.name, pattern);
    query="SELECT \"file\" FROM ";
    query.append(opOnly()).append(" ");
    query.append(p.directoryTable);
    if( p.type() != EPT_DIR ) {
      query.append(" WHERE \"file\"");
      if(isPattern)
	query.append(" LIKE '");
      else
	query.append(" = '");
      query.append(pattern).append("'");
    }
    query.append(";");  

    DMESG("SQL: >" << query << "<" << std::endl);
    if(statement.exec(query)){
      printError("9 Internal error", statement);
      return;
    }

    char buff[1001];
    statement.bind(1, buff, 1000);

    while(!statement.fetch()){
      if( p.type() != EPT_DIR )
	out.append(p.parent);
      else
	out.append(p.name);
      if(p.name!="/")
	out.append("/");

      out.append(buff);
      out.append("\n");
    }

    statement.unBind();
    statement.close();

    I++;
  }

  if( I == propList.begin() ) {
    out.append("1 Directory not found\n");
  } else if (!hasPermission) {
    out.append("1 Directory not found\n");
  }
}


/** Parses an option string for the ls command.
    An option string looks like "-lD,attr1,attr2"
    After the - sign options are written as letters, after which a
    comma comes and then a list of attributes which will be output by
    ls, too.
    As a result the option string becomes "lD" for the above example.
*/
int MDStandalone::decodeLSOptions(std::string &options,
				  std::vector< std::string > &attrs)
{
  if( !options.size() )
    return 0;
  
  if( options[0] != '-' )
    return 1; // Illegal option format

  size_t pos = options.find(",");
  if( pos != std::string::npos ) { // We found attributes
    std::string attributes = options.substr(pos+1);
    options = options.substr(1, pos-1);

    
  } else {
    options = options.substr(1);
  }

  return 0;
}

/** Lists all the files in a directory.
    Lists all files in a directory, without checking the permissions.
    If anything fails it returns 1, printing out the error message.
    If found is false at the beginning and an entry is being found, first the 0
    of the protocol is returned.
 */
int MDStandalone::doFilesLS(Statement &statement, EntryProps &p,
			    bool &found,
			    const std::vector< std::string > &attrs,
			    const std::string &options)
{
  DMESG("In MDStandalone::doFilesLS\n");
  std::stringstream query;
  std::list< std::string > groups; // Caching of groups user is in
  
  std::vector< std::string > attributes = attrs;

  char file[FILENAME_MAX_SIZE];

  char owner[USERNAME_MAX_SIZE];
  char permissions[PERMISSION_SIZE];
  char group_rights[PERMISSION_SIZE];

  char created[CREATED_SIZE];
  char size[FILE_SIZE_MAX_SIZE];
  char guid[GUID_SIZE];
  char md5[MD5_SIZE];
  char link[PATH_MAX_SIZE];

  char acls[ACL_MAX_SIZE];
  
  std::map< std::string, std::string > defaultACLs;
  aclStringToMap(p.parentACLs, defaultACLs);
  std::string other_perms = aclToPerm(defaultACLs["system:anyuser"]);

  DMESG("Other: " << other_perms << "\n");

  // p.print();
  std::string pattern;
  bool isPattern = sqlPattern(p.name, pattern);


  // If we have an entry, we copy over the parent's flags:
  DMESG("Entry type: " << p.type() << "\n");
  
  query << "SELECT \"" << p.mainColumn << "\" ";
  
  if( p.flags & EP_PERMS || ((p.type() == EPT_NONE) && (p.parentFlags & EP_PERMS)) )
    query << ", \"owner\", \"permissions\", \"group_rights\"";

  if( p.flags & EP_FCAT || ((p.type() == EPT_NONE) && (p.parentFlags & EP_FCAT)) )
    query << ", \"created\", \"size\", " << dbHelper->readGUID("guid") << ", \"md5\", \"link\"";
  
  if( (p.flags & EP_ACLS) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_ACLS)) ) 
    query << ", \"acls\"";
  
  query << " FROM " << opOnly() << " " << p.directoryTable;

  if( (p.type() != EPT_DIR) && (p.type() != EPT_VIEW) ) {
    query << " WHERE \"file\"";
    if(isPattern)
      query << " LIKE '";
    else
      query << " = '";
    query << pattern << "'";
    
    if( (p.flags & EP_SHARED) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_SHARED)) )
      query << " AND \"dir\"=" << p.id;
    query << ";";
  } else { // Doing a dir on a directory, which exists, but may be empty
    
    if( (p.flags & EP_SHARED) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_SHARED)) )
      query << " WHERE  \"dir\"=" << p.id;


    // It is not necessary to have permissions _on_ the file to list a file
    query << ";";
    
    if(!found) {
      out.append("0\n");
      found=true;
    }
    
  }
DMESG("Other: " << other_perms << "\n");

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return 1;
  }


  int t_count = 1;
  statement.bind(t_count++, file, FILENAME_MAX_SIZE);
    
  if( (p.flags & EP_PERMS) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_PERMS)) ) {
    statement.bind(t_count++, owner, USERNAME_MAX_SIZE);
    statement.bind(t_count++, permissions, PERMISSION_SIZE);
    statement.bind(t_count++, group_rights, PERMISSION_SIZE);
  }

  if( (p.flags & EP_FCAT) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_FCAT)) ) {
    statement.bind(t_count++, created, CREATED_SIZE);
    statement.bind(t_count++, size, FILE_SIZE_MAX_SIZE);
    statement.bind(t_count++, guid, GUID_SIZE);
    statement.bind(t_count++, md5, MD5_SIZE);
    statement.bind(t_count++, link, PATH_MAX_SIZE);
  }

  if( (p.flags & EP_ACLS) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_ACLS)) ) {
    statement.bind(t_count++, acls, ACL_MAX_SIZE);
  }

  DMESG("Other: " << other_perms << "\n");
  

  std::string prefix;
  if( (p.type() != EPT_DIR) && (p.type() != EPT_VIEW) )
    prefix.append(p.parent);
  else
    prefix.append(p.name);
  if(p.name!="/")
    prefix.append("/");

  while(!statement.fetch()){
    if( !strlen(created ) )
      strcpy(created, "unknown            ");
    if( !strlen(size ) )
      strcpy(size, "0");

    std::stringstream line;
    if(!found) {
      out.append("0\n");
      found=true;
    }

    if( options.find("P") != std::string::npos ) {
      line << prefix << file << "\nentry";  // line << prefix
      if( options.find("l") != std::string::npos ) {
	if( (p.flags & EP_PERMS) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_PERMS)) ) {
	  line << "\n" << permissions << "\n" << group_rights 
	       << "\n"<< owner <<"\n" ;
	} else {
	  line << "\n\n\n\n";
	}

	if( (p.flags & EP_ACLS) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_ACLS)) ) {
          DMESG("ACLS : " << acls << "\n");
	  line << acls << "\n" ;
        } else {
          DMESG("ACLS2 : " << acls << "\n");
	  line << "\n" ;
        }

	if( (p.flags & EP_FCAT) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_FCAT)) ) {
	  line << size << "\n" << created << "\n"
	       << guid << "\n" << md5 << "\n" << link << "\n";
	} else {
	  // Make other entry the same length
	  line << "\n\n\n\n\n";
	}
      }
      line << "\n"; // entry seperator
    } else {
      if( options.find("l") != std::string::npos ) {
	if( (p.flags & EP_PERMS) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_PERMS)) ) {
	  line << "-";
	  line << permissions;
	  line << group_rights;
//	  line << other_perms.c_str();
	  line << " ";
	  line.width(8); line.setf(ios::left);
	  line << owner << " ";
	} else {
	  line << "----------         ";
	}
	
	if( (p.flags & EP_FCAT) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_FCAT)) ) {
	  line.width(8); line.setf(ios::right);
	  line << size << " " << created << " ";
	  if( options.find("g") != std::string::npos ) {
	    try {
	      line << AMGA::niceGUID(guid) << " ";
	    } catch (runtime_error) {
	      line << "           <illegal guid>            ";
	    }
	  }
	  if( options.find("m") != std::string::npos )
	    line << md5 << " ";
	} else {
	  // Make other entry the same length
	  line << "                             ";
	}
	line << file;
	if( ( (p.flags & EP_FCAT) || ((p.type() == EPT_NONE) && (p.parentFlags & EP_FCAT)) ) && strlen(link))
	  line << " -> " << link;
	
      } else {
	line << file;
      }
      
      line << "\n";
    }
    out.append(line.str());
  }
  
  statement.unBind();
  statement.close();
  
  return 0;
}


int MDStandalone::statDir(EntryProps &p, const std::string &options,
			  bool &found)
{
  
  if(!found) {
    out.append("0\n");
    found=true;
  }

  std::stringstream line;

  std::map< std::string, std::string > aclMap;
  aclStringToMap(p.ACLs, aclMap);
  std::string other_perms = aclToPerm(aclMap["system:anyuser"]);

  int dt = p.type();

  if( options.find("P") != std::string::npos ) {
    line << p.name;
    
    switch(dt) {
    case EPT_DIR: // dir with table
      if (options.find("o") != std::string::npos ) {
	line << "\ncollection " <<  encodeTableOptions(p.flags)
	     << "\n";
      } else {
	line << "\ncollection\n";
      }
      //     DMESG("Is dir: " << p.directoryTable << "\n");
      break;
    case EPT_NONE:
      if( !p.directoryTable.size() )       // empty dir
	line << "\ncollection\n";
      else
	line << "\ncollection\n";          // imported dir
      break;
    case EPT_VIEW:
      line << "\nview\n";
      break;
    case EPT_INDEX:
      line << "\nindex\n";
      break;
    case EPT_SEQUENCE:
      line << "\nsequence\n";
      break;
    default:
      ERRLOG("Unknown table type for: >" << p.directoryTable << "<\n");
    }
    if( options.find("l") != std::string::npos )
      line << p.perms << "\n" << p.owner <<"\n" << p.ACLs << "\n\n";

  } else {

    if( options.find("l") != std::string::npos ) {
      switch(dt) {
      case EPT_DIR: // Dir with table
	line << "d";
	break;
      case EPT_NONE:
	if( !p.directoryTable.size() )             // empty dir
	  line << "d";
	else
	  line << "d";                              // imported dir
	break;
      case EPT_VIEW:
	line << "v";
	break;
      case EPT_INDEX:
	line << "i";
	break;
      case EPT_SEQUENCE:
	line << "s";
	break;
      default:
      ERRLOG("Unknown table type\n");
      }
      
      line << p.perms ;
//      line << "rwx" ;
      line << other_perms << " ";
      line.width(8); line.setf(ios::left);
      line << p.owner << " " << "                             ";
    }

    line << p.name << "\n";
  }
  out.append(line.str());

  return 0;
}


int MDStandalone::doDirsLS(Statement &statement, EntryProps &p,
			   bool &found, const std::string &options)
{
  std::stringstream query;

  char directory[DIRNAME_MAX_SIZE];
  char table_name[TABLE_NAME_MAX_SIZE];
  char owner[USERNAME_MAX_SIZE];
  char permissions[PERMISSION_SIZE];
  char acls[ACL_MAX_SIZE];
  int flags;

  // First select directory/ies
  query << "SELECT \"directory\", \"table_name\", \"flags\", \"owner_name\", "
	<< "\"permissions\", \"acls\" FROM "
	<< p.indexTable << " WHERE "
	<< matchSubdirs(p.name) << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal error", statement);
    return 1;
  }
  
  statement.bind(1, directory, DIRNAME_MAX_SIZE);
  statement.bind(2, table_name, TABLE_NAME_MAX_SIZE);
  statement.bind(3, &flags);
  statement.bind(4, owner, USERNAME_MAX_SIZE);
  statement.bind(5, permissions, PERMISSION_SIZE);
  statement.bind(6, acls, ACL_MAX_SIZE);
  
  EntryProps e;

  while(!statement.fetch()){
    e.name = directory;
    e.directoryTable = table_name;
    e.owner = owner;
    e.perms = permissions;
    e.ACLs = acls;
    e.flags = flags;
    statDir(e, options, found);
  }
  
  statement.unBind();
  statement.close();
  
  if( !found && !p.directoryTable.size() ){
    out.append("0\n");
    found = true;
  }
  
  return 0;
}


/** List files and directories in \param dir
*/
void MDStandalone::ls(const std::string &options, const std::string &dir)
{
  DMESG("ls called, options >" << options << "<  dir >" << dir << "<\n"); 
  bool found=false;
  std::string directory = dir;
  if( dir == "*" )
    directory = ".";

  std::vector< std::string > attributes;
  std::string opts = options;

  if( decodeLSOptions(opts, attributes) ) {
    out.append("30 Illegal options\n");
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return;

  error = 0;

  std::list< std::string > groups;
  EntryPropList::iterator I=propList.begin();
  while(I != propList.end()) {
    EntryProps p=*I;
    DMESG("Listing entry:\n");
    p.print(debug);

    for(unsigned int i=0; i<attributes.size(); i++){
      if(checkAttribute(attributes[i], attributes[i], true, p.flags & EP_PLAIN)){
	out.append("7 Illegal Key:" + attributes[i] + "\n");
	return;
      }
    }

    Statement statement(*dbConn);

    // List directories
    if( (p.type() == EPT_DIR) || (p.type() == EPT_VIEW) ){
      // List directory itself if -d flag given
      if( options.find("d") != std::string::npos ) {
	statDir(p, options, found);
      } else {
	if( checkPermissions(p, 'x', groups) <=0 ) {
	  I++;
	  error = 4;
	  continue;
	}
	// List directories in directory
	if(doDirsLS(statement, p, found, opts))
	  return; // Error handled by doDirsLS
      }
    }
    
    // List indices and sequences themselves
    if( (p.type() == EPT_INDEX) || (p.type() == EPT_SEQUENCE) ) {
      statDir(p, options, found);
      I++;
      continue;
    }

    // List entries in directory if -d flag not given for dir
    if( ( (p.type() != EPT_DIR) && (p.type() != EPT_VIEW) )
	|| ( (options.find("d") == std::string::npos) 
	     && (options.find("s") == std::string::npos) ) ) {
      // Check whether the table exists
      if( !p.directoryTable.size() ){
	I++;
	continue;
      }
      if( checkPermissions(p, 'x', groups) <=0 ) {
	I++;
	error = 4;
	continue;
      }
      // List files in directory
      if(doFilesLS(statement, p, found, attributes, opts))
	return; // Error handled by doFilesLS
    }

    I++;
  }

  if(!found) {
    switch (error) {
    case 4: 
      out.append("4 Permission denied");
      break;
    default:
      out.append("1 File or directory not found");
      break;
    }
    out.append("\n");
  }
}


void MDStandalone::dir(const std::string &d)
{
  return ls("-P",d);
}


int MDStandalone::removeTableConstraints(Statement &statement, int id)
{
  std::stringstream query;
  query << "DELETE FROM " << constraintsTable 
	<< " WHERE \"table_id\"=" << id << ";";
  DMESG("SQL: " << query.str() << endl);
  int err;
  if ( ( err = statement.exec(query.str()) ) ) {
    if (statement.getErrorCode() == SQL_NO_DATA) {
      return 0;
    }
  }
  return err;
}


//! Removes a directory from an index table, but only if it is empty
//! and there is no user-defined schema attached afterwards the table is dropped
//! You must check that the table itself exists, before calling!
//! Returns 0 if the directory can be removed from the index, -1 on error 
int MDStandalone::removeIfEmpty(const EntryProps &p,
				bool forceAttributeRemoval,
				Statement &statement)
{
  DMESG("In removeIfEmpty\n");
  // Check whether there are entries left in the table
  int id = 0;
  if( p.flags & EP_SHARED )
    id = p.id;

  std::string query = entriesExist(p.directoryTable, id);
  DMESG("SQL entry exists: >" << query << "<" << std::endl);
  
  int entries;
  if( statement.exec(query) ){
    printError("9 DB error", statement);
    return -1;
  }
  statement.bind(1, &entries);
  if(statement.fetch()){
    printError("9 DB error", statement);
    return -1;
  }
  statement.unBind();
  statement.close();
  
  DMESG("removeIfEmpty: Entries in table: " << entries << std::endl);
  
  if(!entries) {  // No entries
    // Check whether there is a user-defined schema
    AttrList keys;
    if(getAttributes(p.directoryTable, keys)) {
      out.append("11 Directory has attributes\n");
      return 11;
    }
    
    // Remove link to table if no attributes or forced to and drop it
    if(!keys.size() || forceAttributeRemoval){
      if(removeTableFromIndex(p.name, p.indexTable, statement)) {
        printError("9 DB error", statement);
	return 9;
      }

      if( removeTableConstraints(statement, p.id) ) {
        printError("9 DB error", statement);
	return 9;
      }

      if( !(p.parentFlags & EP_SHARED) ) {
	if(dropTable(statement, p.directoryTable)) {
	  printError("9 DB error", statement);
	  return 9;
	}
      }
    }
    return 0;
  }

  out.append("11 Directory not empty\n");

  return 11;
}


//! Removes one or more files
/** Remove all files in a directory matching path
    \param path The pattern against which files are matched
*/
void MDStandalone::rm(const std::string &path,
		      const std::string &condition,
		      const std::string &options)
{
  DMESG("rm path: " << path << "  condition: " << condition 
	<< "  options: " << options << std::endl);

  std::string q;
  if( condition.size() ) {
    if(parser->parse(condition)){
      out.append("8 Illegal Query\n");
      return;
    }
  }

  // Start transaction
  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  } 

  // Get all matching directories or entries
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;

  bool didSomething = false;
  bool errorOccured = false;
  EntryPropList::iterator I=propList.begin();
  string logDirectory;
  while(I != propList.end()) {
    EntryProps p=*I;
    if (logDirectory.empty()) {
        logDirectory = getEntryDirectory(p);
    }

    DMESG("Options: >" << options << "<\n");

    q = "";
    if( condition.size() ) {
      if( p.flags & EP_PLAIN )
	parser->toSQL(q, true);
      else
	parser->toSQL(q);
    }

    if( (p.type() == EPT_DIR) && options.find("r")!=std::string::npos ) {
      try{
        deleteDirectoryContents(p.name, true);
      } catch (std::exception &e) {
        printError("9 ", e);
        errorOccured = true;
        break;
      }
      I++;
      didSomething = true;
      continue;
    } else {
      if( (p.type() == EPT_DIR) && !q.size() ){
	DMESG("Not an entry\n");
	out.append("20 Not an entry: " + p.name + "\n");
	errorOccured = true;
	break;
      }
    }

    if( (p.type()!=EPT_DIR) && (p.flags & EP_PLAIN) ) {
      out.append("29 Operation not permitted on plain table\n");
      I++;
      errorOccured = true;
      break;
    }

    if(!tableExists(p.directoryTable, p.type())) {
      I++;
      errorOccured = true;
      break;
    }
    
    std::list< std::string > groups;

    int perm=checkDirPermissions(p, 'w', groups);
    if (!checkPermResult(perm)) {
      I++;
      errorOccured = true;
      break;
    }

    DMESG("Directory permissions OK" << std::endl);
    
    std::string pattern;
    bool isPattern = sqlPattern(p.name, pattern);
    std::stringstream query;
    query << "DELETE FROM " << opOnly() << " " << p.directoryTable
	  << " WHERE ";
    if( (p.type() == EPT_DIR) && q.size() ) {
      query << q;
    } else {
      if( !(p.flags & EP_PLAIN) ) {
	query << "\"" << p.mainColumn <<"\"";
	if(isPattern)
	  query << " LIKE '";
	else
	  query << " = '";
	query << pattern << "'";
	if( p.parentFlags & EP_SHARED )
	  query << " AND \"dir\" = " << p.parentID;
	if( q.size() )
	  query << " AND " << q;
      } else {
	if( q.size() )
	  query << q;
      }
    }

    int t_flags = p.type() ? p.flags : p.parentFlags;
    if( p.sticky() ) {
      query << " AND " << permissionClause('a', t_flags, groups) << ";";
    }  else {
      query << " AND " << permissionClause('w', t_flags, groups) << ";";
    }

    DMESG("SQL: >" << query.str() << "<" << std::endl);

    if(statement.exec(query.str())){
      if( statement.getErrorCode() == 100 ) 
	out.append("1 no matching File/directory or none with correct permissions \n");
      else
	printError("9 Error deleting entries: ", statement);
      I++;
      errorOccured = true;
      break;
    }

    didSomething = true;
    I++;
  }

  if (!didSomething && !errorOccured) {
    out.append("1 no matching File/directory or none with correct permissions \n");
    return;
  }
  
  if( errorOccured )
    return;
    
  if (didSomething) {
    // [rep] Sucessful operation. Write the log
    saveLog(statement, logDirectory);
  }
  statement.commitTransaction();  
  out.append("0\n");
}


int MDStandalone::deleteDirObjects(EntryPropList propList, int objectType,
				   bool force, Statement &statement, 
				   std::string &parent)
{
  EntryPropList::iterator I=propList.begin();
  while(I != propList.end()) {
    EntryProps p=*I;

    if( (p.flags & EP_TYPE) == 0 ){
      out.append("11 Not a directory type\n");
      return -1;
    }

    std::list< std::string > groups;
    if ((p.sticky()) && 
        !checkPermResult(checkPermissions(p, 't', groups)))
      return -1;

    // Make sure only empty directories without schema are removed
    if( p.type() == EPT_DIR ) {
      // Try to remove empty directory tables, force removal of attributes
      if( p.directoryTable.size() ) {
	if( removeIfEmpty(p, true, statement) )
	  return -1; //DBError
      }
      bool empty = !hasSubdirs(p);

      if(!empty){
	out.append("11 Directory not empty\n");
	return -1;
      }
    }

    if( p.type() != objectType ){
      if( objectType & EPT_DIR ) {
	printError("11 Not a directory", statement);
      } else {
	printError("11 Wrong type", statement);
      }
      return -1;
    }

    if(p.name=="/") {
      out.append("11 Cannot delete /\n");
      return -1;
    }

    if (!checkPermResult(checkPermissions(p, 'w', groups)))
      return -1;
    DMESG("Directory permissions OK" << std::endl);

    std::string objectName;
    std::string query;
    std::string style;

    switch(p.type()) {
    case EPT_DIR:
      objectName=p.directoryTable;
      break;
    case EPT_INDEX:
      objectName=p.directoryTable;
      query.append("DROP INDEX ").append(p.directoryTable);

      style = dbHelper->backend();
      if( style.find("MySQL") != std::string::npos) {
        query.append(" ON ");

	std::vector< std::string > result;
        std::string t_str;
	boost::algorithm::split(result, p.directoryTable, boost::is_any_of("_"));
        if( result.size() < 3 ) return -1;
        for(int i=1; i < result.size() - 1; i++) {
          if( i != 1 ) t_str.append("_");
          t_str.append(result.at(i));
        }
        
        query.append(t_str);
      }
      
      query.append(";");
      break;
    case EPT_VIEW:
      objectName=p.directoryTable;
      query.append("DROP VIEW ").append(p.directoryTable).append(";");
      break;
    case EPT_SEQUENCE:
      objectName=p.directoryTable;

      style = dbHelper->backend();
      if( style.find("MySQL") == std::string::npos) {
        query.append("DROP SEQUENCE ").append(p.directoryTable).append(";");
      } else {
        query.append("DELETE FROM ").append(usersequencesTable);
        query.append(" WHERE name like '%").append(p.directoryTable).append("';");
      }
      break;
    case EPT_VOLUME:
    case EPT_NONE:
      LOG("WARNING unhandled directory type. Don't know what to do!");
    }

    if(query.size()) {
      DMESG("SQL: >" << query << "<" << std::endl);
      if(statement.exec(query)){
	LOG("Object already dropped" << p.directoryTable << std::endl);
      }
    }

    query="UPDATE "+p.indexTable+" SET "
      + "\"table_name\"='', \"flags\"=0, ";
    if( !strcmp(dbHelper->backend(), "MySQL") )
      query += "\"directory\"=concat('./', \"id\")";
    else
      query += "\"directory\"='./' || \"id\"";
    query += " WHERE \"directory\" LIKE '" + p.name + "'";
    //    if( !(p.flags & EP_SHARED) )
    //      query += " AND \"table_name\"='" + objectName + "'";
    query += ";";

    DMESG("SQL: >" << query << "<" << std::endl);

    if(statement.exec(query)){
      printError("11 Directory not empty", statement);
      return -1;
    }
    parent = p.parent;    
    I++;
  }

  return 0;
}


//! Removes objects from index tables
/** Removes objects from index tables. Objects can include
    - empty directories
    - indices
    - views
    - volumes
    - mounts
    You cannot delete entries with this function. Basically it only
    operates on index tables, not on directory tables.
*/
void MDStandalone::rmObject(const std::string &path, int objectType)
{
  Statement statement(*dbConn);

  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(path, propList);

  if(handleEntryPropErrors(error, path))
    return;

  // WARNING: This assumes that all the entries returned by
  // getEntryProps have the same parent. If they are different, the
  // log might end up being saved associated with the wrong
  // directory.
  std::string parent;
  // Errors are handled by deleteDirObjects()
  if( !deleteDirObjects(propList, objectType, false, statement, parent) ) {
    if (saveLog(statement, parent)) {
      statement.commitTransaction();
      out.append("0\n");
    }
  }

}


/** Removes an empty directory named path.
    Pattern matching does not work.
*/
void MDStandalone::rmDir(const std::string &path)
{
  return rmObject(path, EPT_DIR);
}


/** Removes an index
*/
void MDStandalone::indexRemove(const std::string &path)
{
  return rmObject(path, EPT_INDEX);
}


/** Removes a view
*/
void MDStandalone::viewRemove(const std::string &path)
{
  return rmObject(path, EPT_VIEW);
}


/** Removes a Sequence
*/
void MDStandalone::sequenceRemove(const std::string &path)
{
  return rmObject(path, EPT_SEQUENCE);
}


/** Sets the link of a file to a certain URL
    link file URL
    URL may be:
    file:/path
    guid:xxxxxxx
*/
void MDStandalone::link(const std::string &path, const std::string &URL)
{
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  if(!p.directoryTable.size()){ // Directory does not exist
    out.append("1 Directory not found\n");
    return;
  }

  std::list< std::string > groups;
  if (!checkPermResult(checkDirPermissions(p, 'w', groups))) {
    return;
  }

  // Is an entry
  if( !p.type() )
    p.id = p.parentID;

  Statement statement(*dbConn);
  if(statement.beginTransaction()){
    printError("9 Internal error", statement);
    return;
  }
  
  if( !p.directoryTable.size() ){
    AttrList attrs;
    p.directoryTable=createTable(p, attrs, statement);
  }
  if(!p.directoryTable.size())
    return;

  std::stringstream query;
  query << "INSERT INTO " << p.directoryTable << " (\"file\", "; 
  if( p.parentFlags & EP_SHARED )
    query << "\"dir\", ";
  if( p.parentFlags & EP_PERMS )
    query << "\"owner\", \"permissions\", \"group_rights\", \"link\"";
  if( p.parentFlags & EP_FCAT )
    query << ", \"created\", \"size\"";

  query << ") VALUES ('" << p.name << "' ,'";

  if( p.parentFlags & EP_SHARED )
    query << p.parentID << "', '";
  if( p.parentFlags & EP_PERMS )
    query << user << "' ,'"	<< permissionMask << "' ,'"
	  << groupRightsMask << "' ,'" << URL  << "'";
  if( p.parentFlags & EP_FCAT )
    query << ", '" << makeTimestamp() <<  "', 0";
  
  query << ");";
  
  DMESG("SQL: >" << query.str() << "<" << std::endl);

  if(statement.exec(query.str())){
    printError("15 Entry exists", statement);
    return;
  }
  
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, p.parent)) {
      statement.commitTransaction();
      out.append("0\n");
  }  
}


/** Returns filename, permissions, group_rights, link
    stat filepattern
*/
void MDStandalone::stat(const std::string &path)
{
  ls("-Pglmdo", path);
}


/// Expands a group name by the user name if necessary
/** Expands a group name by the user name if not already done. A
    qualified group name is built like user:groupname.
    \param grp Group name to be expanded
    \return Expanded group name or "" if group name is illegal.
*/
std::string MDStandalone::expandGroupName(const std::string grp)
{
  if(grp.size() >= GROUPNAME_MAX_SIZE)
    return "";

  bool isQualified = false;

  // Check for group name validity
  for(unsigned int i=0; i<grp.size(); i++){
    if(grp[i] == ':') {
      if( !i || i == grp.size()-1 || isQualified)
	return "";
      isQualified = true;
      continue;
    }

    // SIAHN_20110111 I do not know why the below limitation is necessary
    // So, I remove it
    // if((!isalnum(grp[i]) && !grp[i] == '/') || isupper(grp[i])) {
    //   return "";
    // }
  }

  if( isQualified )
    return grp;

  std::string name=user;
  name.append(":").append(grp);

  if(name.size() >= GROUPNAME_MAX_SIZE)
    return "";

  return name;
}


/// Checks whether a group exists
/** Checks whether a group exists
    \param grp Fully qualified group name
*/
bool MDStandalone::doesGroupExist(const std::string grp,
				  Statement &statement)
{
  if(grp == "system:anyuser")
    return true;

  bool found = true;

  std::string query;
  query.append("SELECT \"members\" FROM " + groupsTable + " WHERE \"name\"=\'");
  query.append(grp).append("';");

  DMESG("SQL: >" << query << "<" << std::endl);

  if(statement.exec(query)){
    return false;
  }
  char buff[1025];
  statement.bind(1, buff, 1024);

  if(statement.fetch())
    found = false;

  statement.close();

  return found;
}


/** Internal function to create a group
    This function is internally called by createGroup. It does
    not produce any output and can be safely called to create
    on the fly group on startup.
*/
int MDStandalone::doCreateGroup(const std::string &grp)
{
  if(grp == "system:anyuser")
    return 5; // Group exists
  
  std::string name=expandGroupName(grp);
  if(!name.size())
    return 21; // Illegal group name;

  DMESG("Group name is fine\n");

  // SIAHN_20110120 the below causes an error 
  //    when a normal user create a group
  // if(!checkIfCapabilityAllowed("create_group_admin")){
  if(user != "root") {
	if(name.size()<user.size()+2
	   || name.substr(0, user.size()) != user
	   || name[user.size()] != ':') {
	  return 4; // Permission denied;
	}
  }

  std::string owner=user;
  size_t pos = grp.find(":");
  if( pos != std::string::npos ) {
    owner = grp.substr(0, pos);
  }

  DMESG("Permissions OK\n");
  
  std::string query;
  query.append("INSERT INTO " + groupsTable);
  query.append(" (\"name\", \"owner\", \"members\") VALUES ");
  query.append("('").append(name).append("', '");
  query.append(owner).append("', '');");

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query))
    return 5; // Group exists
  
  DMESG("Group did not exist\n");
  
  return 0;
}


/** Creates a group owned by the current user:
    grp_create groupname
*/
void MDStandalone::createGroup(const std::string &grp)
{
  Statement st(*dbConn);
  st.beginTransaction();    
  int result = doCreateGroup(grp);
  switch(result) {
  case 21:
    out.append("21 Illegal group name\n");
    return;
  case 4:
    out.append("4 Permission denied\n");
    return;
  case 5:
    out.append("5 Group exists\n");
    return;
  case 0:
    if (saveGroupLog(st, grp)) {    
    	st.commitTransaction();
    	out.append("0\n");
    }
    return;
  default:
    out.append("9 An unknown error occurred\n");
  }
}


/** Deletes a group owned by the current user:
    grp_delete groupname
*/
void MDStandalone::deleteGroup(const std::string &grp)
{
  std::string name=expandGroupName(grp);
  if(!name.size()) {
    out.append("21 Illegal group name\n");
    return;
  }

  std::string query;
  query.append("DELETE FROM " + groupsTable + " WHERE \"name\"=\'");
  query.append(name);

  // SIAHN_20110120 the below causes an error 
  //    when a normal user create a group
  // if(!checkIfCapabilityAllowed("delete_group_admin")){
  if(user != "root") {
    query.append("' AND \"owner\" = '").append(user);
  }
  query.append("';");


  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  statement.beginTransaction();
  if(statement.exec(query)){
    if(doesGroupExist(name, statement)) {
      out.append("4 Permission denied\n");
      return;
    }
    printError("5 Group does not exist", statement);
    return;
  }

  if( !statement.numRows() ) {
    out.append("5 Group does not exist\n");
    return;
  }
  if (saveGroupLog(statement, grp)) {    
   	statement.commitTransaction();
   	out.append("0\n");
  } 
}





/** Fills in the list of members for the group with name @name
 */
int MDStandalone::getMembers(const std::string grp,
			     std::list< std::string > &members)
{
  char buff[GRPMEMBERS_MAX_SIZE+1];

  std::string query;
  query.append("SELECT \"members\" FROM " + groupsTable + " WHERE \"name\"=\'");
  query.append(grp).append("';");

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query))
    return 5; // Group does not exist

  statement.bind(1, buff, GRPMEMBERS_MAX_SIZE);

  if(statement.fetch())
    return 5; // Group does not exist

  splitGroupString(buff, members);
 
  return 0;
}


/** Fills in the list of members for the group with name @name
 */
int MDStandalone::memberOf(const std::string usr,
			   std::list< std::string > &groups) const
{
  if (user.compare(usr) == 0){
    std::list <std::string>::const_iterator I = vomsGroupList.begin();
    while(I != vomsGroupList.end() ) {
      groups.push_back(*I);
      I++;
    }
  }

  char buff[USERNAME_MAX_SIZE+1];

  std::string query;
  query.append("SELECT \"name\" FROM " + groupsTable + " WHERE \"members\" LIKE \'%|");
  query.append(usr).append("|%';");

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("5 Group does not exist", statement);
    return -1;
  }
  statement.bind(1, buff, USERNAME_MAX_SIZE);
  while (1){
    if(statement.fetch())
      return groups.size();
    groups.push_back(buff);
  }

  return groups.size();
}


/** Shows all users of a group, user must be owner or member or root
    grp_show groupname
*/
void MDStandalone::grpShow(const std::string &grp)
{
  std::string name=expandGroupName(grp);
  if(!name.size()){
    out.append("21 Illegal group name\n");
    return;
  }

  std::list< std::string > members;
  std::list< std::string >::const_iterator I;
  int error = getMembers(name, members);
  switch(error) {
  case 5:
    out.append("5 Group does not exists\n");
    return;
  case 0:
    break;
  default:
    out.append("8 Internal error\n");
    return;
  }

  if(user != "root") {
    // Check whether owner
    if(name.size()<user.size()+2
       || name.substr(0, user.size()) != user
       || name[user.size()] != ':') {
      // Check whether member
      I=members.begin();
      while(I!=members.end()){
	if(*I == user)
	  break;
	I++;
      }
      if(I == members.end()) {
	out.append("4 Permission denied: not a group member\n");
	return;
      }
    }
  }

  out.append("0\n");
  I=members.begin();
  while(I!=members.end()){
    out.append(*I++);
    out.append("\n");
  }
}


/** Private function to add/remove users from a list
    Does not produce output and is therefore safe to call
    during authentication phase to create on-the-fly groups
*/
int MDStandalone::grpChangeUser(const std::string &grp,
				const std::string &usr,
				bool add)
{
  DMESG("grpChangeUser: entry >" << grp << "<  user >" << usr
	<< "<  add: " << add << "\n");

  std::string name=expandGroupName(grp);
  if(!name.size())
    return 21; // Illegal group name

  std::list< std::string > members;
  int error = getMembers(name, members);
  if(error)
    return error;

  std::string memberstring="|";
  std::list< std::string >::iterator I=members.begin();
  while(I!=members.end()){
    if(*I != usr){
      memberstring.append(*I).append("|");
    } else {
      if(add)
	return 24; // User already member of that group
    }
    I++;
  }
  if(add)
    memberstring.append(usr).append("|");
  if(memberstring.size()==1) memberstring="";

  std::string query;
  query.append("UPDATE " + groupsTable + " SET \"members\" = '");
  query.append(memberstring);
  query.append("' WHERE \"name\"=\'");
  query.append(name);
  if(user != "root")
    query.append("' AND \"owner\" = '").append(user);
  query.append("';");

  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query)){
    if(doesGroupExist(name, statement))
      return 4; // Permission denied
    return 6; // Group does not exits
  }

  return 0;
}


/** Adds a user to a group
    grp_adduser groupname user
*/
void MDStandalone::grpAddUser(const std::string &group,
			      const std::string &usr)
{
  Statement st(*dbConn);
  st.beginTransaction();
  int result = grpChangeUser(group, usr, true);
  switch(result) {
  case 21:
    out.append("21 Illegal group name\n");
    return;
  case 4:
    out.append("4 Permission denied\n");
    return;
  case 5:
    out.append("5 Group does not exist\n");
    return;
  case 24:
    out.append("24 User already member of that group\n");
    return;
  case 0:
  	// The root user is not replicated.
    if (saveGroupLog(st, group, &usr)) {    
      st.commitTransaction();
      out.append("0\n");
    } 
    return;
  default:
    out.append("9 An unknown error occurred\n");
  }
}


/** Removes a user/group from a group
    grp_removeuser groupname user
*/
void MDStandalone::grpRemoveUser(const std::string &group,
				 const std::string &usr)
{
  Statement st(*dbConn);
  st.beginTransaction();
  int result = grpChangeUser(group, usr, false);
  DMESG("grpChangeUser: Error " << result << std::endl);
  switch(result) {
  case 21:
    out.append("21 Illegal group name\n");
    return;
  case 4:
    out.append("4 Permission denied\n");
    return;
  case 5:
  case 6:
    out.append("5 Group does not exist\n");
    return;
  case 0:
  // The root user is not replicated.
    if (saveGroupLog(st, group, &usr)) {    
   	  st.commitTransaction();
      out.append("0\n");
    }
    return;
  default:
    out.append("9 An unknown error occurred\n");
  }
}


/** Shows the groups of which the user is member of
    grp_member user
*/
void MDStandalone::grpMember(const std::string &usr)
{
  std::string userName = usr;
  if(!usr.size())
    userName = user;

  if(userName != user && user != "root") {
    out.append("4 Permission denied: You can only list your"
	       " own group membership\n");
    return;
  }

  std::list< std::string > groups;
  if(memberOf(userName, groups) <0)
    return;  // Error printed in memberOf()

  out.append("0\n");
  std::list< std::string >::const_iterator I=groups.begin();
  while(I!=groups.end()){
    out.append(*I++);
    out.append("\n");
  }
}


/** Shows the groups of which the user is member of
    grp_member user
*/
void MDStandalone::grpList(const std::string &usr, bool all)
{
  DMESG("grpList called: usr >"<< usr << "< all: " << all << std::endl);
  std::string userName = usr;
  if(!usr.size())
    userName = user;

  // Switched off privacy rule
  //  if(userName != user && user != "root") {
  //    out.append("4 Permission denied: You can only list your"
  //	       " own groups\n");
  //    return;
  //  }

  char buff[GROUPNAME_MAX_SIZE];

  std::string query;
  if( all ) {
    query = "SELECT \"name\" FROM " + groupsTable + ";";
  } else {
    query.append("SELECT \"name\" FROM " + groupsTable + " WHERE \"owner\" = '");
    query.append(userName).append("';");
  }

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  bool found = false;
  if(statement.exec(query)){
    printError("9 Internal error", statement);
    return;
  }

  statement.bind(1, buff, sizeof(buff));

  while( !statement.fetch() ){
    if(!found) {
      found = true;
      out.append("0\n");
    }
    out.append(buff);
    out.append("\n");
  }
  if(!found)
    out.append("0\n");
}




/** Private function to add/remove an ACL
    Only the owner or root can change the ACLs of an entry.
*/
void MDStandalone::aclChange(const std::string &entry,
			     const std::string &grp,
			     const std::string &permissions,
			     bool add)
{
  string cannonicalPerms;
  if( !aclPermissionFormatOK(permissions, cannonicalPerms) ) {
    out.append("25 Illegal access permission format\n");
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(entry, propList);
  if(handleEntryPropErrors(error, entry))
    return;
  
  Statement statement(*dbConn, debug);
  if(statement.beginTransaction()) {
    printError("9 Could not begin transaction", statement);
    return;
  }

  std::string groupName = expandGroupName(grp);
  if(!groupName.size()) {
    out.append("21 Illegal group name\n");
    return;
  }

  /* BUG Fix : 2009. 07. 28
     BIG Name : It is not possible to remove a deleted group from an ACL, where it is still present
  */
  if( !doesGroupExist(groupName, statement) && (grp != "*" && add)) {
    printError("5 Group does not exist: " + grp , statement);
    return;
  }
  std::list< std::string > groups;
  EntryPropList::iterator J=propList.begin();
  while(J != propList.end()) {
    EntryProps p=*J;

    DMESG("Owner: " << p.owner << " user: " << user << "\n");

    if( p.type() && !checkPermResult(checkPermissions(p, 'a', groups)) )
      return;

    // For entries the acls are only read by fillACLMap...
    std::map< std::string, std::string > acls;
    if(fillACLMap(p, acls))
      return;  // Error message done by fillACLMap
    
    DMESG("aclChange: entry" << p.name << "  group " << grp
	  << "  permissions " << cannonicalPerms << "  add "
	  << add << std::endl);

    std::map< std::string, std::string >::iterator I=acls.begin();
    std::string aclstring="|";
    while(I!=acls.end()){
      if((I->first) != groupName){  // Always remove double entries
	aclstring.append(I->first).append(" ").append(I->second).append("|");
      }
      I++;
    }
    
    if(add) {
      aclstring.append(groupName).append(" ").append(cannonicalPerms).append("|");
    }
    if(aclstring.size()==1)
      aclstring="";

    if( aclstring.size() >= ACL_MAX_SIZE-1 ){
      out.append("26 ACL length exceeded\n");
      return;
    }

    if( grp == "*" && !add )
      aclstring = "";

    ostringstream query;
    if( p.type() == EPT_NONE ) {        
      query << "UPDATE " << p.directoryTable << " SET \"acls\" = ";
      if( !aclstring.size() )
	query << "NULL";
      else
	query << "'" << aclstring << "'";

      int t_flags = p.type() ? p.flags : p.parentFlags;
      query << " WHERE \"" << p.mainColumn << "\"=\'" << p.name + "' AND "
	    << permissionClause('a', t_flags, groups);

      if( p.flags & EP_SHARED )
	query << " AND \"dir\" = " << p.parentID << ";";
      else
	query << ";";

    } else {
      query << "UPDATE " << masterindexTable << " SET \"acls\" = ";
      if( !aclstring.size() )
	query << "NULL";
      else
	query << "'" << aclstring << "'";
      query << " WHERE \"directory\"=\'" << p.name << "';";
    }

    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Error updating acl", statement);
      return;
    }

    J++; // Next directory...
  }
  
  // Save log for replication
  if (!propList.empty()) {
    // I'm assuming that they have all the same parent directory.
    if (!saveLog(statement, getEntryDirectory(*propList.begin()))) {
      return;
    }
  }
  if( statement.commitTransaction() ) {
    printError("9 Cannot commit transaction", statement);
    return;
  }

  out.append("0\n");
}


/** Adds an acl to a directory
    acl_add directory group rights
*/
void MDStandalone::aclAdd(const std::string &entry,
			  const std::string &grp,
			  const std::string &permissions)
{
  return aclChange(entry, grp, permissions, true);
}


/** Removes an ACL from a directory
    acl_remove directory group
*/
void MDStandalone::aclRemove(const std::string &entry,
			     const std::string &grp)
{
  return aclChange(entry, grp, "", false);
}


/** Shows the ACLs of a directory
    acl_show directory
*/
void MDStandalone::aclShow(const std::string &directory)
{
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return;

  if(!propList.size()) {
    out.append("1 Directory not found\n");
    return;
  }

  EntryProps p=*propList.begin();
  std::map< std::string, std::string > acls;
  if(fillACLMap(p, acls))
    return;  // Error print in fillACLMap()

  out.append("0\n" + p.owner + " " + p.perms + "\n");
  std::map< std::string, std::string >::const_iterator I=acls.begin();
  while(I!=acls.end()){
    out.append(I->first + " " + I->second + "\n");
    I++;
  }
}


/** Fills in the acls of a into a map for a directory or file
    Only for a file the ACLs are actually read, for a directory they
    are taken out of the EntryProps structure.
    Indices/Sequences are treaded like a directory.
    For an entry the EntryProps structure is updated and the owner,
    permissions and ACLs fields are filled
 */
int MDStandalone::fillACLMap(EntryProps &p,
			     std::map< std::string, std::string > &aclMap)
{
  if( p.type() == EPT_NONE ) {

    char owner[TABLE_NAME_MAX_SIZE];
    char permissions[PERMISSION_SIZE];
    char group_rights[PERMISSION_SIZE];
    char acls[ACL_MAX_SIZE];

    std::stringstream query;

    query << "SELECT \"owner\", \"permissions\", \"group_rights\", \"acls\" FROM "
	  << p.directoryTable << " WHERE \"" << p.mainColumn << "\"=\'" << p.name << "'";
    if( p.flags & EP_SHARED )
      query << " AND \"dir\" = " << p.parentID;

    query << ";";
    
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    Statement statement(*dbConn);
    if(statement.exec(query.str())){
      printError("9 Internal DB error reading acls", statement);
      return 9;
    }
    statement.bind(1, owner, sizeof(owner));
    statement.bind(2, permissions, sizeof(permissions));
    statement.bind(3, group_rights, sizeof(group_rights));
    statement.bind(4, acls, sizeof(acls));
    
    if(statement.fetch()) {
      out.append("1 No such file or directory\n");
      return 1;
    }
    
    p.owner = owner;
    p.perms = (std::string)permissions + (std::string)group_rights;
    p.ACLs = acls;

    statement.unBind();
    statement.close();
  }

  aclStringToMap(p.ACLs, aclMap);

  return 0;
}


/** Creates a new volume in the masterindex
    createvolume /parentdir/dir
    where parentdir must exist.
*/
void MDStandalone::createVolume(const std::string &path)
{

  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;
  EntryProps p(*propList.begin());

  if( p.type() != EPT_NONE ) {
    out.append("16 Directory exists\n");
    return;
  }

  std::list< std::string > groups;
  if (!checkPermResult(checkDirPermissions(p, 'w', groups))) {
    return;
  }

  std::string name;

  Statement statement(*dbConn);
  if(finalizeDirectoryToIndex(addDirectoryToIndex(
			        name, p, statement),
			      statement)<0 ){
    // Error handled by finalizeDirectoryToIndex
    return;
  }

  // Now create new index table


  // And add initial directory


  // And commit the directory


  out.append("0\n");
}


/** Creates a new view
    view_create name /parentdir key1 key2 'query'
    where parentdir must exist. The first key is the primary key with
    the index on it
*/
void MDStandalone::createView(const std::string &name,
			      const std::string &mainDirectory,
			      const std::vector<std::string> &attributes,
			      const std::string &restriction)
{
  std::string fullName;
  EntryProps prop;
  if( prepareDirlikeEntry(name, prop, fullName) ) return;

  std::string shortName = prop.name;
  for(unsigned int i=0; i<shortName.size(); i++){
    if(!isalnum(shortName[i]) || isupper(shortName[i])) {
      out.append("17 Illegal name (for index, view)\n");
      return;
    }
  }

  int subdirSearch = 0;
  std::string mainDir = mainDirectory;
  // If the dirctory name ends in a +, then we search also shared subdirs
  if( mainDir[mainDir.size()-1] == '+' ) {
    mainDir.resize(mainDir.size()-1);
    subdirSearch = 1;
  }

  EntryPropList propList;
  int error=getEntryProps(mainDir, propList);
  if(handleEntryPropErrors(error, mainDir))
    return;
  EntryProps pMain(*propList.begin());  

  if(pMain.type() != EPT_DIR) {
    out.append("1 No such directory ");
    out.append(mainDir);
    out.append("\n");
    return;
  }

  Statement statement(*dbConn);
  std::string table=pMain.directoryTable;
  if( !table.size() ) {
    AttrList attrs;
    pMain.directoryTable=table=createTable(pMain, attrs, statement);
  }
  if(!table.size())
    return;

  TableMap tableMap;
  if(createTableMap(attributes, tableMap))
    return; // Error handled by createTableMap

  if(parser->parse(restriction)){
    out.append("8 Illegal Query\n");
    return;
  }
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap

  std::string q;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q);

  std::string viewname = generateTableName("view", table, shortName);

  std::string ptype = "char(3)";
  if( !strcmp(dbHelper->backend(), "MySQL") )
    ptype = "";
  if( !strcmp(dbHelper->backend(), "Oracle") )
    ptype = "";

  std::stringstream query;
  pMain.print();
  query << "CREATE VIEW " << viewname << " AS SELECT ";

  query << table << ".\"file\"";

  if( pMain.flags & EP_PERMS ) {
    query << ", " << table << ".\"owner\", " << ptype 
     	  << " 'r--' AS \"permissions\", "
     	  << ptype <<  " 'r--' AS \"group_rights\"";
  }
  if ( pMain.flags & EP_FCAT ) {
    query << ", " << table << ".\"link\", " << table << ".\"created\", " 
	  << table << ".\"size\", " << table << ".\"guid\", " 
	  << table << ".\"md5\"";
  }
  if ( pMain.flags & EP_ACLS ) {
    query << ", " << table << ".\"acls\"";
  }
   
  std::vector<std::string>::const_iterator I=attributes.begin();
  while(I!=attributes.end()){
    std::string key;
    parser->parse(*I++);
    parser->replaceTableNames(tableMap);
    parser->toSQL(key);
    query << ", " << key;
  }

  tableMap[pMain.name]=TableEntry(pMain, subdirSearch);
  std::list< std::string > distinctTables;
  getDistinctTables(tableMap, distinctTables);

  std::list< std::string >:: iterator J=distinctTables.begin();
  query << " FROM ";
  while(J != distinctTables.end()) {
    query << (*J++);
    if(J!=distinctTables.end())
      query << ", ";
  }

  if(!q.size())
    q="(1=1)";

  query << " WHERE " << q;

  removeDoubleTables(tableMap);
  if( user != "root" ) {
    std::list< std::string > groups;
    TableMap::iterator K = tableMap.begin();
    while(K != tableMap.end()) {
      int flags = K->second.props.type() ? K->second.props.flags : K->second.props.parentFlags;
      std::string tableName = K->second.props.directoryTable;
      DMESG("Directory >" << K->first << "< (" << tableName << ") :"
          << flags<< "\n");
      query << " AND " <<  permissionClause('r', flags, groups, tableName);
      K++;
    }
  }
  query << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal DB error", statement);
    return;
  }
  
  prop.directoryTable = viewname;

  // KIMNG080619
  prop.flags |= EPT_VIEW;

  int id = addDirectoryToIndex(fullName, prop, statement);

  if(finalizeDirectoryToIndex(id, statement) < 0 ) {
    // Error handled by finalizeDirectoryToIndex
    return;
  }
  
  out.append("0\n");
}


/** Creates a new view
    view_create name /parentdir key1 key2 'query'
    where parentdir must exist. The first key is the primary key with
    the index on it
*/
void MDStandalone::createViewUnion(const std::string &name,
				   const std::vector<std::string> &dirs,
				   const std::string &restriction)
{
  std::string fullName;
  EntryProps prop;
  if( prepareDirlikeEntry(name, prop, fullName) ) return;

  std::string shortName = prop.name;
  for(unsigned int i=0; i<shortName.size(); i++){
    if(!isalnum(shortName[i]) || isupper(shortName[i])) {
      out.append("17 Illegal name (for index, view)\n");
      return;
    }
  }

  std::vector< EntryProps > dirProps(dirs.size());

  for ( size_t i = 0; i < dirs.size(); i++) {
    EntryPropList propList;
    int error=getEntryProps(dirs[i], propList);
    if(handleEntryPropErrors(error, dirs[i]))
      return;
    dirProps[i] = *propList.begin();

    if(dirProps[i].type() != EPT_DIR) {
      out.append("1 No such directory ");
      out.append(dirs[i]);
      out.append("\n");
      return;
    }
  }

  Statement statement(*dbConn);

  TableMap tableMap;
  if(parser->parse(restriction)){
    out.append("8 Illegal Query\n");
    return;
  }
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap

  std::string q;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q);

  AttrList attributes;
  std::string firstTable=dirProps[0].directoryTable;
  if(!firstTable.size() || getAttributes(firstTable, attributes)) {
    out.append("1 Directory does not exist\n");
  }

  std::string viewname;
  viewname = generateTableName("view", prop.directoryTable, shortName);
  std::string ptype = "char(3)";
  if( !strcmp(dbHelper->backend(), "MySQL") )
    ptype = "";
  
  std::stringstream query;
  query << "CREATE VIEW " << viewname << " AS ";
  for ( size_t i = 0; i < dirProps.size(); i++) {
    if(i) query << " UNION ";
    query << "SELECT ";
    std::string table = dirProps[i].directoryTable;

    query << table << ".\"file\"";
    if( dirProps[i].flags & EP_PERMS )
      query << ", " << table << ".\"owner\", " << ptype 
	    << " 'r--' AS \"permissions\", "
	    << ptype <<  " 'r--' AS \"group_rights\"";
    if ( dirProps[i].flags & EP_FCAT )
      query << ", " << table << ".\"link\", " << table << ".\"created\", " 
	    << table << ".\"size\", " << table << ".\"guid\", " 
	    << table << ".\"md5\"";
    if ( dirProps[i].flags & EP_ACLS )
      query << ", " << table << ".\"acls\"";
    
    AttrList::const_iterator I=attributes.begin();
    while(I!=attributes.end()){
      std::string a= "";
      checkAttribute(I++->first, a, false, dirProps[i].flags & EP_PLAIN);
      query << ", " << a;
    }
    query << " FROM " << table;
  }

  if(q.size())
    query << " WHERE " << q;

  query << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal DB error", statement);
    return;
  }

  prop.directoryTable = viewname;

  // AHN 20100312
  prop.flags |= EPT_VIEW;

  if(finalizeDirectoryToIndex(addDirectoryToIndex( fullName, prop, statement),
			      statement) < 0 ) {
    // Error handled by finalizeDirectoryToIndex
    return;
  }
  
  out.append("0\n");
}


/** Creates an index
    createindex name dir 'key1 key2' algo
    Creates an index name on dir for key using the indexing algorithm algo
*/
void MDStandalone::createIndex(const std::string &name,
			       const std::string &directory,
			       const std::vector<std::string> &attributes,
			       const std::string &algo)
{
  std::string algorithm = algo;
  if( !algo.size() ) {
    algorithm = "BTREE";
  }

  for(unsigned int i=0; i<name.size(); i++){
    if(!isalnum(name[i]) || isupper(name[i])) {
      out.append("17 Illegal name (for index, view)\n");
      return;
    }
  }

  // Open the transaction before reading the entry properties 
  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Could not begin transaction", statement);
    return;
  }
    
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return;
  EntryProps p(*propList.begin());

  std::list< std::string >keys;
  std::vector<std::string>::const_iterator I=attributes.begin();
  while(I!=attributes.end()){
    std::string key;
    if(parser->parse(*I, key, p.flags & EP_PLAIN)){
      out.append("7 Illegal Key " + *I + "\n");
      return;
    }
    keys.push_back(key);
    I++;
  }

  if(!keys.size()){
    out.append("7 No key\n");
    return;
  }

  if( p.type() != EPT_DIR ) {
    out.append("17 No such directory " + directory + "\n");
    return;
  }

  std::string table=p.directoryTable;
  if( !table.size() ) {
    AttrList attrs;
    p.directoryTable=table=createTable(p, attrs, statement);
  }
  if(!table.size())
    return;

  std::string entry = generateTableName("index", table, name);

  std::stringstream query;
  if( !strcmp(dbHelper->backend(), "MySQL") ) {
    query << "CREATE INDEX " << entry << " USING " << algorithm << " ON " << table;
  } else {
    if ( !strcmp(dbHelper->backend(), "Oracle") ) {
      query << "CREATE INDEX " << entry << " ON " << table ;
    } else {
      query << "CREATE INDEX " << entry << " ON " << table << " USING " << algorithm;
    }
  }
  query << " (";
  std::list< std::string >::const_iterator J=keys.begin();
  while(J!=keys.end()){
    query << (*J);
    J++;
    if(J!=keys.end())
      query << ", ";
  }
  query << ");";
    
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal DB error", statement);
    return;
  }

  std::string indexDir=p.name;
  if(indexDir != "/")
    indexDir.append("/");
  indexDir.append(name);

  p.directoryTable = entry;
 
  // KIMNG080619
  p.flags = EPT_INDEX;
  
  if(finalizeDirectoryToIndex(addDirectoryToIndex( indexDir, p, statement),
			      statement) < 0 ){
    return;
  }
  // [rep] Sucessful operation. Write the log
  if (saveLog(statement, p.name)) {
      // error handled by saveLog
      statement.commitTransaction();  
      out.append("0\n");
  }
}


/** Changes the owner of a file/directory
    Changes the owner of a file/directory. Note, that only root is allowed to do this.
*/
void MDStandalone::chown(const std::string &path, const std::string &owner, bool recursiveOption)
{
	std::string preCurrentDir = currentDir;

  // If being used for replication, then anything goes.
  if( user != "root" && !replicating){
    DMESG("Permission denied" << std::endl);
    out.append("4 Permission denied\n");
    return;
  }
  
  Statement statement(*dbConn);  
  if(statement.beginTransaction()) {
    printError("9 Could not begin transaction", statement);
    return;
  }
  
  // Change Owner Call - Recursive function
  int retCode = chownChangeOwner(statement, path, owner, recursiveOption);
  
  currentDir = preCurrentDir;
  if(  retCode < 0 ) {
  	if(statement.rollbackTransaction()){
      printError("9 Problem aborting transaction", statement);
      return;
    } else {
      printError("9 Could not set owner", statement);
      return;
    }
  }
  
  statement.commitTransaction();  
  out.append("0\n");  
}

int MDStandalone::chownChangeOwner(Statement &statement, const std::string path, const std::string owner, bool recursiveOption)
{
	std::string preCurrentDir = currentDir;
  
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return-1;

  EntryPropList::iterator I=propList.begin();
  while(I != propList.end()) {
    EntryProps p=*I;
  
    std::stringstream query;
    if( p.type() == EPT_NONE ) { // Entry
    	if(p.directoryTable == "") // No Entry Directory
    			return 0;
    	
      std::string pattern;
      bool isPattern = sqlPattern(p.name, pattern);
      query << "UPDATE " << p.directoryTable << " SET \"owner\"='" << owner
	    << "' WHERE \"file\"";
      if(isPattern)
	query << " LIKE '";
      else
	query << " = '";
      query << pattern << "'";
      if( p.parentFlags & EP_SHARED )
	query << " AND \"dir\" = " << p.parentID;
      query << ";";
    } else { // Directory, index...
 		if(recursiveOption && (p.type() == EPT_DIR)) {  
 					currentDir = p.name + "/"; // Set Current Directory for Recursive call
 					if( chownChangeOwner(statement, "*", owner, recursiveOption) < 0 )
      			return -1;
      		currentDir = preCurrentDir;
      	}
      query << "UPDATE " << p.indexTable << " SET \"owner_name\"='" << owner 
	    << "' WHERE \"directory\" = '" << p.name << "';";
    }
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Could not set owner", statement);
      return -1;
    }
    I++;
  }
  
  if (!propList.empty()) {
    // I'm assuming that they have all the same parent directory.
    if (!saveLog(statement, getEntryDirectory(*propList.begin()))) {
      return-1;
    }
  }
  return 0;
}


/** Changes the access rights of a file or directory
    @param path Filename/Directoryname
    @param mod New mod of style rwxrwx for user and groups of a file
*/
void MDStandalone::chmod(const std::string &path, const std::string &mod)
{
  EntryPropList propList;
  int error=getEntryProps(path, propList);
  if(handleEntryPropErrors(error, path))
    return;

  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Could not begin transaction", statement);
    return;
  }

  std::list< std::string > groups;
  EntryPropList::iterator I=propList.begin();
  while(I != propList.end()) {
    EntryProps p=*I;

    if( p.sticky() ) {
      int perm=checkPermissions(p, 't', groups);
      if( perm<= 0){
	DMESG("Permission denied" << std::endl);
	out.append("4 Permission denied\n");
	return;
      }
    }

    if( p.owner != user ) {
      int perm=checkPermissions(p, 'w', groups);
      if( perm<= 0){
	DMESG("Permission denied" << std::endl);
	out.append("4 Permission denied\n");
	return;
      }
    }

    if(mod.size()<3) {
      DMESG("25 Illegal permission format: " << mod << endl);
      out.append("25 Illegal permission format: " + mod + "\n");
      return;
    }

    std::string perms = mod.substr(0,3);
    int t_dir = p.type() ? 1 : 0 ;
    if( ! permissionFormatOK(perms, t_dir ) ) {
      DMESG("25 Illegal permission format: " << perms << endl);
      out.append("25 Illegal permission format: " + perms + "\n");
      return;
    }

    ostringstream query;
    if( p.type() == EPT_NONE ){ //Entry
      if(mod.size()<4) {
        DMESG("25 Illegal permission format: " << mod << endl);
        out.append("25 Illegal permission format\n");
        return;
      }

      std::string grp_perms=mod.substr(3);
      if( ! permissionFormatOK(grp_perms,0) ) {
        DMESG("25 Illegal permission format: " << grp_perms << endl);
        out.append("25 Illegal permission format\n");
        return;
            }
      
      std::string pattern;

      query << "UPDATE " << p.directoryTable << " SET \"permissions\"='"
	    << perms << "', \"group_rights\"='" << grp_perms 
	    << "' WHERE \"file\"";

	    bool isPattern = sqlPattern(p.name, pattern);
	    
	    if(isPattern)
			query << " LIKE '";
      else
			query << " = '";
      
      query << pattern << "'";
	    
	    
      if( p.parentFlags & EP_SHARED)
	query << " AND \"dir\" = " << p.parentID;
      query << ";";
    } else {
      query << "UPDATE " << p.indexTable << " SET \"permissions\"='" 
	    << perms << "' WHERE \"directory\" = '" << p.name << "';";
    }
    DMESG("SQL: >" << query.str() << "<" << std::endl);

    if(statement.exec(query.str())){
      printError("2", statement);
      return;
    }
    I++;
  }

  if (!propList.empty()) {
    // I'm assuming that they have all the same parent directory.
    if (!saveLog(statement, getEntryDirectory(*propList.begin()))) {
      return;
    }
  }  
  statement.commitTransaction();  
  out.append("0\n");
}



//! Copies a schema from one directory to another
/** Usage:
    schema_cp source_dir dest_dir
*/
void MDStandalone::schemaCp(const std::string &source,
			    const std::string &dest)
{
  DMESG("schemaCp called" << std::endl);

  EntryPropList propListSource;
  int error=getEntryProps(source, propListSource);
  if(handleEntryPropErrors(error, source))
    return;
  EntryProps pSource(*propListSource.begin());

  if(pSource.type() != EPT_DIR){
    out.append("11 Not a directory\n");
    return;
  }

  if(!pSource.directoryTable.size()){
    out.append("1 Directory does not exist\n");
    return;
  }

  std::list< std::string > groups;
  if (!checkPermResult(checkDirPermissions(pSource, 'r', groups))) {
    return;
  }

  EntryPropList propListDest;
  error=getEntryProps(dest, propListDest);
  if(handleEntryPropErrors(error, dest))
    return;
  EntryProps pDest(*propListDest.begin());

  if(pDest.type() != EPT_DIR){
    out.append("11 Not a directory\n");
    return;
  }
  
  if (!checkPermResult(checkPermissions(pDest, 'w', groups))) {
    return;
  }

  Statement statement(*dbConn);
  AttrList sourceKeys;
  int res;
  res=getAttributes(pSource.directoryTable, sourceKeys);
  if(res)
    return;

  if( !pDest.directoryTable.size() ) {
    pDest.directoryTable=createTable(pDest, sourceKeys, statement);
    if(pDest.directoryTable.size()) {
      out.append("0\n");
      return;
    }
  }
  if(!pDest.directoryTable.size())
    return;

  AttrList destKeys;
  res=getAttributes(pDest.directoryTable, destKeys);
  if(res)
    return;

  AttrList::iterator I=destKeys.begin();

  // Copy only attributes not already present
  while(I!=destKeys.end()) {
    sourceKeys.remove(*I); // Remove duplicates
    I++;
  }

  I=sourceKeys.begin();
  while(I!=sourceKeys.end()) {
    std::string key;
    if(parser->parse(I->first, key)){
      out.append("7 Illegal Key: ");
      out.append(I->first); out.append("\n");
      return;
    }
    I->first=key;
    I++;
  }

  if(!alterTableColumns(pDest.directoryTable, sourceKeys, false)){
    // Error output done by alterTableColumns
    return;
  }

  out.append("0\n");
}


std::string MDStandalone::updateAttrQuery(const EntryProps &p,
					  const std::string &entries,
					  bool isPattern,
					  std::vector< std::string > &attributes,
					  std::vector< std::string > &values,
					  const std::string &condition,
					  bool eval)
{
  DMESG("updateAttrQuery called, eval = " << eval << "\n");

  std::list< std::string > groups; // Caching of groups user is in
  int perms = checkDirPermissions(p, 'w', groups);
  std::vector< std::string> keys;
  keys.resize(attributes.size());

  std::stringstream query;
  query << "UPDATE " << p.directoryTable << " SET ";
  for(unsigned int i=0; i<attributes.size(); i++){
    if(parser->parse(attributes[i], keys[i], p.flags & EP_PLAIN)){
      out.append("7 Illegal Key\n");
      return "";
    }
    std::string value;
    // Value is only parsed if eval is true
    if(eval && parser->parse(values[i], value, p.flags & EP_PLAIN)){
      out.append("33 Illegal value: ");
      out.append(values[i]);
      out.append("\n");
      return "";
    }

    if(eval)
      values[i]=value;
    else
      value ="?";

    query << keys[i] << " = " << value;
    if(i<attributes.size()-1)
      query << ", ";
  }
  if( !(isPattern && entries == "%") ) {
    query << " WHERE \"file\"";
    if(isPattern)
      query << " LIKE '";
    else
      query << " = '";
    query << entries << "' AND ";
  } else {
    query << " WHERE ";
  }
  if(condition.size())
    query << "("<< condition << ") AND";
  if( p.parentFlags & EP_SHARED )
    query << " \"dir\" = " << p.parentID << " AND ";

  int t_flags = p.type() ? p.flags : p.parentFlags;
  query << permissionClause('w', t_flags, groups) << ";";

  return query.str();
}


/** Selects attribute(s) of a number of files based on a condition
    path is the path to the files where wildcards are only allowed in
    the file part of the path
    attributes is the list of attributes to which need to be returned
    The output including all errors is streamed into the out class
    member. For a description see the protocol description.
*/
void MDStandalone::selectAttr(std::vector<std::string> &attributes,
			      const std::string &q)
{
  TableMap tableMap;
  if(createTableMap(attributes, tableMap))
    return; // Error handled by createTableMap

  if(parser->parse(q)){
    out.append("8 Illegal Query\n");
    return;
  }
  parser->addToTableMap(tableMap);

  if(setupTableMap(tableMap))
    return; // Error handled by createTableMap

  TableMap::iterator tables = tableMap.begin();
  while(tables != tableMap.end()) {
    if( tables->second.props.directoryTable.empty() ) {
      out.append("0\n");
      return;
    }
    tables++;
  }

  std::string query="SELECT ";

  std::string q2;
  parser->replaceTableNames(tableMap);
  parser->toSQL(q2);
  std::string constraint;
  parser->getConstraint(constraint);
  if ( parser->distinctOnly() )
    query += "DISTINCT ";
  std::string joinCondition;
  parser->getJoin(joinCondition);

  std::vector<std::string>::iterator I=attributes.begin();
  while(I!=attributes.end()){
    std::string key;
    parser->parse(*I++);
    parser->replaceTableNames(tableMap);
    parser->toSQL(key);
    query.append(key);
    if(I!=attributes.end())
      query.append(", ");
  }

  query.append(" FROM ");

  if( joinCondition.size() ) {
    query.append(joinCondition);
  } else {
    std::list< std::string > distinctTables;
    getDistinctTables(tableMap, distinctTables);
    
    std::list< std:: string >:: iterator J=distinctTables.begin();
    while(J != distinctTables.end()) {
      query.append(*J++);
      if(J!=distinctTables.end())
	query.append(", ");
    }
  }

  if(!q2.size())
    q2="(1=1)";

  query.append(" WHERE ").append(q2);
  std::list< std::string > groups;
  TableMap::iterator K = tableMap.begin();
  while(K != tableMap.end()) {
    // K->second.props.print(true);
    int flags = K->second.props.type() ? K->second.props.flags : K->second.props.parentFlags;
    std::string tableName = K->second.props.directoryTable;
    DMESG("Directory >" << K->first << "< (" << tableName << ") :"
	  << flags << "\n");
    query += " AND "
      + permissionClause('r', flags, groups, tableName);
    K++;
  }

  query += constraint + ";";
  
  DMESG("SQL: >" << query << "<" << std::endl);

  Statement statement(*dbConn);
  bool dummy=false;
  doBulkRead(statement, query, attributes.size(), dummy);

  if(!dummy){
    out.append("0\n");
  }
}


bool MDStandalone::findAmongAttributes(std::vector< std::string > &attrs,
				       const std::string &attribute) const
{
  return std::find(attrs.begin(), attrs.end(), attribute) != attrs.end(); 
}


void MDStandalone::completeUploadAttributes(
    std::vector< std::string > &attrs,
    const EntryProps &entryProps) const
{  
  vector<string> sysAttrs;  
  //    sysAttrs.push_back("FILE");

  if( (entryProps.flags & EP_PLAIN) ) {
	  return;
  }

  if( (entryProps.flags & EP_PERMS) ) {  
    sysAttrs.push_back("OWNER");
    sysAttrs.push_back("PERMISSIONS");
    sysAttrs.push_back("GROUP_RIGHTS");
  }

  if (entryProps.flags & EP_FCAT) {
    sysAttrs.push_back("LINK");
    sysAttrs.push_back("CREATED");
    sysAttrs.push_back("SIZE");
    // Handled differently
    //      sysAttrs.push_back("GUID");
      sysAttrs.push_back("MD5");
  }
    
  if (entryProps.flags & EP_ACLS) {
    sysAttrs.push_back("ACL");
  }
    
  // Must iterate backwards so the attributes are inserted in order.
  for (vector<string>::reverse_iterator rit = sysAttrs.rbegin(); 
      rit != sysAttrs.rend(); rit++) 
  {
    if(!findAmongAttributes(attrs, *rit))
      attrs.insert(attrs.begin(), *rit);
  } 
}


map<string, string> MDStandalone::createAttrDefaultValues() const 
{
  map<string, string> m;
  m["OWNER"] = string("'") + user + "'";
  m["PERMISSIONS"] = string("'") + permissionMask + "'";
  m["GROUP_RIGHTS"] = string("'") + groupRightsMask + "'";
  m["LINK"] = "''";
  m["CREATED"] = "'" + makeTimestamp() + "'"; 
  m["SIZE"] = "0";
  m["MD5"] = "''";
  m["ACL"] = "''";
//  m["GUID"] = "gen";
  return m;
}

/** Starts uploading entries
    Error semantics is as follows: If this function fails, then -1 is
    returned and the returned upload handle will have a NULL
    statement. put, commit and abort do a check on this.
    If upload succeeded but there are problems during the transaction
    then they are counted in the statement.
*/
int MDStandalone::upload(UploadHandle &h,
			 const std::string &dir,
			 std::vector< std::string > &attributes)
{
  DMESG("In MDStandalone::upload" << std::endl);
  if (!replicating) { // Everything is allowed when replicating
    if( user!="root" && findAmongAttributes(attributes, "OWNER") ) {
      DMESG("4 Permission denied: You cannot upload the OWNER attribute\n");
      out.append("4 Permission denied: You cannot upload the OWNER attribute\n");
      return -1;
    }
    if( findAmongAttributes(attributes, "FILE") ) {
      DMESG("4 Permission denied: You cannot upload the FILE attribute\n");
      out.append("3 You cannot upload the FILE attribute\n");
      return -1;
    }
  }

  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return -1;
  EntryProps &p = propList.front();

  DMESG("Testing for dir" << std::endl);
  std::string file;
  if( p.type() != EPT_DIR ) {
    out.append("1 File or directory not found\n");
    return -1;
  }

  // Complete the attributes with the system attributes of an entry
  std::vector< std::string > attrs = attributes;
  completeUploadAttributes(attrs, p);

  DMESG("In MDStandalone::upload Preparing insert" << std::endl);  
  auto_ptr<Statement> statement(new Statement(*dbConn, debug));
  // If we are doing the initial synchronization, don't open a transaction.
  // The dump can be very long and we can rollback on our own.    
  DMESG("BEGINNING TRANSACTION"<< std::endl);
  if(statement->beginTransaction()){
    printError("9 Could not start transaction", *statement);
    return -1;
  }
  
  if( !p.directoryTable.size() ) {
    AttrList attrs;
    p.directoryTable = createTable(p, attrs, *statement);
  }
  if(p.directoryTable.empty()) {
    out.append("9 Internal error. Could not create table.\n");
    return -1;
  }

  ostringstream oss;
  oss << "INSERT INTO " << p.directoryTable << " (";  
  string sep;  
  if (p.flags & EP_FCAT && !findAmongAttributes(attributes, "GUID")) {
    // GUID was not provided. Can only happen when this is an upload
    // made by a user. In this case, we generate the GUID.
    assert(!synchronizing);
    h.genGUID = true;       
    oss << sep << "\"guid\"";
    sep = ", ";
  }
  if (!(p.flags & EP_PLAIN)) {
    oss << sep << "\"" << p.mainColumn << "\"";
    sep = ", ";
  }
  if (p.flags & EP_SHARED) {
    oss << sep << "\"dir\"";
    sep = ", ";
  }
    
  for(unsigned int i=0; i<attrs.size(); i++){
    std::string key;
    if(parser->parse(attrs[i], key, p.flags & EP_PLAIN)){
      out.append("7 Illegal Key\n");
      return -1;
    }
    oss << sep << key;
    sep = ", ";
  }

  oss << ") VALUES (";
  sep = "";
  if (h.genGUID) {
    oss << sep << dbHelper->writeGUID("?");
    sep = ", ";
  }
  if (!(p.flags & EP_PLAIN)) {
    oss << sep << "?";
    sep = ", ";
  }
  if (p.flags & EP_SHARED) {
    oss << sep << p.id;
    sep = ", "; 
  }
  map<string, string> attrsDefaults = createAttrDefaultValues();
  for(unsigned int i=0; i<attrs.size(); i++) {
    oss << sep;
    // If a system attribute is not uploaded, we must set it to a default
    //    cout << "Preparing: " << attrs[i] << endl;
    if(!findAmongAttributes(attributes, attrs[i]) ){
      if (synchronizing) {
        PERRLOG("Attribute not provided by synchronization dump: " << attrs[i] << endl);
        throw runtime_error(string("Attribute not provided by dump: ") + attrs[i]);
      }
      oss << attrsDefaults[attrs[i]];
    } else {
      oss << "?";
    }
    sep = ", ";
  }
  oss << ");";
  string query = oss.str();
  
  DMESG("Preparing SQL: >" << query << "<" << std::endl);
  if(statement->prepare(query)){
    printError("10 No such key", *statement);
    return -1;
  }
  
  if(statement->prepareBulkInsert(UPLOAD_BULK_SIZE)) {
    DMESG("Bulk insert not available\n");
    h.bulkSize = 1;
  } else {
    // Must query the result, since some drivers ignore the value
    // requested and revert to a default value if they don't support 
    // this functionality.
    h.bulkSize = statement->getBulkInsertSize(); 
  }
  DMESG("Using bulk size of: " << h.bulkSize << endl); 

  DMESG("MDStandalone::upload preparing buffers" << std::endl);  
  // How many attributes will be provided by user
  h.userProvidedAttrs = (p.flags & EP_PLAIN) ? 
    attributes.size() : 
    attributes.size()+1;
  // If we generate the GUID ourselves, then we must upload to the 
  // database one more attribute
  int insertedAttributes = h.genGUID ? 
    h.userProvidedAttrs+1 : 
    h.userProvidedAttrs;
  
  h.buffers.resize(insertedAttributes);
  h.types.resize(insertedAttributes);
  h.bufferSizes.resize(insertedAttributes);
  h.cachedParams = 0;
  //  cout << "Inserted: " << insertedAttributes << ", provided: " << h.userProvidedAttrs << endl;
  for(unsigned int i=0; i<h.buffers.size(); i++){
    h.buffers[i]=new char[UPLOAD_BUFFER_SIZE*h.bulkSize];
    h.types[i]=new SQLLEN[h.bulkSize];
    h.bufferSizes[i]=UPLOAD_BUFFER_SIZE;
    if (statement->bindParam(i+1, h.buffers[i], h.bufferSizes[i], h.types[i])) {
      printError("9 Could not bind parameters for prepared statement", *statement);
      return -1;
    }
  }

  h.directory = getEntryDirectory(p);
  h.statement = statement;
  
  // If saving logs is not enabled, don't generate a sequence number
  // TODO: Chained replication. Cannot save replication logs until the full
  // synchronization completes, because there is the chance that the 
  // synchronization will fail and the logs would be exposed to the slave.
  if (!synchronizing && logWriter->getSaveLogs()) {
    // Must use a different statement. Create a new connection.
    // This is a bit of an overkill, we might obtain the same result
    // if we used a new statement with the same connection. But this
    // would require support from the Statement and DatabaseConnection classes.
    // Perhaps a TODO item.
    h.logsDBConn.reset(DatabaseConnection::openConnection());
    h.logsStatement.reset(new Statement(*h.logsDBConn, true));
    if(h.logsStatement->beginTransaction()){
      printError("9 Could not start transaction", *h.logsStatement);
      return -1;
    }
    try {    
      logWriter->beginUpload(*h.logsStatement, h.directory, user, 
          currentDir, permissionMask, groupRightsMask);
      h.logWriter = logWriter.get();
    } catch (SQLException &ex) {
      printError("9 Could not write replication logs", *h.logsStatement);
      return -1;
    }
  }
  DMESG("MDStandalone::upload done" << std::endl);
  return 0;
}


void printBuffers(UploadHandle &h) {
	ostringstream oss;
	oss  << "Dumping buffers\n";
	for (int row=0; row<h.cachedParams; row++) {
		int offset = row*UPLOAD_BUFFER_SIZE;
		oss << "Row " << row << ": ";		
		for (size_t col=0; col<h.buffers.size(); col++) {
			cout << (h.buffers[col] + offset) << ", ";			
		}
		oss << endl;
	}
	bool debug = true;
	DMESG(oss.str());
}


int MDStandalone::put(UploadHandle &h,
		      const vector<string>::iterator valuesStart,
		      const vector<string>::iterator valuesEnd)
{
  assert(!h.aborted);
  assert(h.statement.get() != NULL);
  
  if(!h.buffers.size()) {
    out.append("9 Illegal put format\n");        
    return 9;
  }

  //  DMESG("Values: " << values.size() << "  buffers: " 
  //	<< h.buffers.size() << std::endl);
  //  for (vector<string>::iterator it = valuesStart; it != valuesEnd; it++) {
  //    cout << *it << endl;
  //  }

  int nValues = valuesEnd - valuesStart;
  
  if( (nValues != h.userProvidedAttrs)) {
    LOG("Illegal number of arguments. Expected: " << h.userProvidedAttrs << ", Found: " << nValues << endl);   
    out.append("9 Illegal number of arguments\n");    
    return 9;
  }
  int bufferOffset = h.cachedParams*UPLOAD_BUFFER_SIZE;
  
  // Generate the GUID, if one must be generated
  unsigned int i=0;
  if (h.genGUID) {    
    try {
      string guid = AMGA::validateGUID("gen");
      strcpy(h.buffers[0] + bufferOffset, guid.c_str());
      h.types[0][h.cachedParams] = SQL_NTS;
      i++;
    } catch (std::runtime_error &e) {
      out.append("32 Illegal GUID: ");
      out.append(e.what());
      out.append("\n");
    }    
  }

  vector<string>::iterator it = valuesStart; 
  for(; i<h.buffers.size(); ++i, ++it){

    unescapePreparedString(*it);

    if (*it == "\\N") {
        // It's a null value
        h.types[i][h.cachedParams] = SQL_NULL_DATA;
        // As long as types is of SQL_NULL_DATA the value on the
        // buffer is ignored.
    } else {
      //TODO: Check for attributes exceeding the maximum supported size.
      if (it->size() > h.bufferSizes[i]-1) {
	delete [] h.buffers[i];
	delete [] h.types[i];
	h.buffers[i]=new char[2*it->size()*h.bulkSize];
	h.types[i]=new SQLLEN[h.bulkSize];
	h.bufferSizes[i]=2*it->size();
	if (h.statement->bindParam(i+1, h.buffers[i], 2*it->size(), h.types[i])) {
	  printError("9 Could not bind parameters for prepared statement", *h.statement);
	  return 9;
	}
	DMESG("WARNING: Resized attribute buffer << " << i 
	      << " Size now: " << h.bufferSizes[i] << std::endl);
      }
      strcpy(h.buffers[i] + bufferOffset, it->c_str());
      h.types[i][h.cachedParams] = SQL_NTS;
    }
  }  
  h.cachedParams++;
  
  int ret = 0;
  if (h.cachedParams == h.bulkSize) {
    //  	DMESG("Writing upload with bulk size: " << h.bulkSize << "\n");
    h.cachedParams = 0;
    ret=h.statement->execPrepared();
    
    if(ret){
      printError("9 Cannot execute prepared statement", *h.statement);
      return ret;
    }
    if (synchronizing) {
      h.operationsCount+=h.bulkSize;
      if (h.operationsCount ==  UPLOAD_TRANSACTION_SIZE) {
	DMESG("Commiting transaction. Lines inserted: " << h.operationsCount << endl); 
	if (h.statement->commitTransaction()) {
	  DMESG("Error committing transaction\n");      	
	}
	h.operationsCount = 0;
	if (h.statement->beginTransaction()) {
	  DMESG("Error starting transaction\n");
	}
      }
    }
  }
  if (!synchronizing && h.logsStatement.get() != NULL) {
    if (!saveLog(*h.logsStatement, h.directory)) {
      // Failed.
      return 9;
    }
  }
  return ret;
}


void MDStandalone::abort(UploadHandle &h)
{
  DMESG("MDStandalone::abort" << std::endl);
  assert(!h.aborted);
  
  if(h.statement->rollbackTransaction()) {
    printError("9 Cannot abort transaction", *h.statement);
    return;
  }
  
  if (h.logsStatement.get() != NULL) {
    h.logsStatement->rollbackTransaction();
  }

  out.append("0\n");
}


void MDStandalone::commit(UploadHandle &h)
{
  DMESG("MDStandalone::commit" << std::endl);
  assert(!h.aborted);
//  h.c.print("[PERF] st.exec()");
//  h.strcpyClock.print("[PERF] strcpy");
	
  if (h.cachedParams != 0) {
    int ret = h.statement->setParamArraySize(h.cachedParams);
    if (ret) {
      printError("9 Cannot change parameter array size", *h.statement);
      return;
    }
    
    ret=h.statement->execPrepared();
    if(ret){
      printError("9 Cannot execute prepared statement", *h.statement);
      return;
    }
  }
  
  // Write the final log
  if (!synchronizing && h.logsStatement.get() != NULL) {
    if (!saveLog(*h.logsStatement, h.directory)) {
      // Failed.
      return;
    }
    logWriter->commit();
    h.logWriter = NULL;
  }
  
  // Commit the metadata transaction    
  if(h.statement->commitTransaction()){
    printError("9 Cannot commit transaction", *h.statement);
    return;
  }	
  
  // Commit the logs transaction
  if (!synchronizing && h.logsStatement.get() != NULL) {
    if (h.logsStatement->commitTransaction()) {
      LOG("WARNING: Failed to commit logs but operation was committed. Replication logs are inconsistent.\n");
      printError("9 Cannot commit logs transaction. Logs are inconsistent with state of database", *h.statement);
      return;
    }
  }

  out.append("0\n");
}


/** Creates an automatic counting sequence
    createcounter name dir [increment] [start value]
    Creates a counter in directory dir named name with an initial value and
*/
void MDStandalone::createSequence(const std::string &name,
				  const std::string &directory,
				  const std::string &inc,
				  const std::string &start)
{
  for(unsigned int i=0; i<name.size(); i++){
    if(!isalnum(name[i]) || isupper(name[i])) {
      out.append("17 Illegal name for a sequence\n");
      return;
    }
  }

  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return;
  EntryProps p(*propList.begin());

  if( p.type() != EPT_DIR ) {
    out.append("17 No such directory " + directory + "\n");
    return;
  }

  // Need w flag on sequence
  std::list< std::string > groups;
  int perms = checkPermissions(p, 'w', groups);
  if (!checkPermResult(perms))
    return;

  Statement statement(*dbConn);
  statement.beginTransaction();
  
  std::string table=p.directoryTable;
  if( !table.size() ) {
    AttrList attrs;
    p.directoryTable = table = createTable(p, attrs, statement);
  }
  if(!table.size())
    return;

  std::string entry = generateTableName("sequence", table, name);

  std::stringstream query;
  std::string style(dbHelper->backend());

  if( style.find("MySQL") == std::string::npos) {
    // Not MySQL: We can use normal sequences
    query << "CREATE SEQUENCE ";
    
    if( style.find("PostgreSQL")!= std::string::npos ) {
      query << entry << " INCREMENT "; // Syntax of PG <=7.3
      query << inc << " START ";
    } else {
      query << entry << " INCREMENT BY " << inc << " START WITH ";
    }
    query << start << ";";
  } else {
    // MySQL: Emulate sequences
    query << "INSERT INTO " << usersequencesTable 
	  << " (name, increment, last_val)"
	  << " VALUES ('" << entry << "', " << inc << ", " << start << "-" << inc << ");";
  }

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Internal DB error", statement);
    return;
  }

  std::string sequenceDir=p.name;
  if(sequenceDir != "/")
    sequenceDir.append("/");
  sequenceDir.append(name);

  p.directoryTable = entry;
  p.flags = EPT_SEQUENCE;

  if(finalizeDirectoryToIndex(addDirectoryToIndex( sequenceDir, p, statement),
			      statement) < 0 ){
    return;
  }
  
  if (saveLog(statement, getEntryDirectory(p))) {
      // error handled by saveLog
      statement.commitTransaction();  
      out.append("0\n");
  }
}


/** Gets atomically the next value of a sequence
 */
void MDStandalone::sequenceNext(const std::string &name)
{
  EntryPropList propList;
  int error=getEntryProps(name, propList);
  if(handleEntryPropErrors(error, name))
    return;
  EntryProps p(*propList.begin());

  if( p.type() != EPT_SEQUENCE ) {
    out.append("17 Not a sequence " + name + "\n");
    return;
  }

  // Need w flag on sequence
  std::list< std::string > groups;
  int perms = checkPermissions(p, 'w', groups);
  if (!checkPermResult(perms))
    return;

  string reply;
  try {
    Statement statement(*dbConn, debug);
    uint64_t tid = statement.sequenceGetNextValue(p.directoryTable);
    ostringstream oss;
    oss << "0\n" << tid << "\n";
    reply = oss.str();
  } catch (SQLException &ex) {
    reply = string("9 Internal DB error: ") + ex.what() + "\n";
  }
  out.append(reply);
}


/** Gets the current value of a sequence
 */
void MDStandalone::sequenceCurrent(const std::string &name)
{
  EntryPropList propList;
  int error=getEntryProps(name, propList);
  if(handleEntryPropErrors(error, name))
    return;
  EntryProps p(*propList.begin());

  if( p.type() != EPT_SEQUENCE ) {
    out.append("17 Not a sequence " + name + "\n");
    return;
  }

  std::list< std::string > groups;
  int perms = checkPermissions(p, 'r', groups);
  if (!checkPermResult(perms))
    return;

  string reply;
  try {
    Statement statement(*dbConn, debug);      
    uint64_t tid = statement.sequenceGetCurrentValue(p.directoryTable);
    ostringstream oss;
    oss << "0\n" << tid << "\n";
    reply = oss.str();
  } catch (SQLException &ex) {
    reply = string("9 Internal DB error: ") + ex.what() + "\n";
  }
  out.append(reply);
}


std::string MDStandalone::entriesExist(const std::string &tableName, int id) const
{
  // PostgreSQL could also do:
  // select exists (select true from masterindex);

  std::ostringstream query;
  if(hasExists()) {
    query << "SELECT EXISTS (select * FROM " << opOnly() << " "<< tableName;
    if( id )
      query << " WHERE \"dir\" = " << id;
    query << ");";
  } else {
    query << "SELECT count(*) FROM " << opOnly() << " " << tableName;
    if( id )
      query << " WHERE \"dir\" = " << id;
    query << ";";
  }
  return query.str();
}


std::string MDStandalone::subdirsExist(const std::string &index,
				       const std::string &parentDir) const
{
  std::string query;
  if(hasExists())
    query="SELECT EXISTS (select * FROM "
      + index + " WHERE " + matchSubdirs(parentDir)
      + ");";
  else
    query="SELECT count(*) FROM "
      + index + " WHERE " + matchSubdirs(parentDir)
      +";";
  return query;
}


void MDStandalone::getName()
{
  out.append("0\n" + siteName + "\n");
}


void MDStandalone::userCreate(
	const std::string &usr, 
	const std::string &pass, 
	bool passIsHash)
{
  if( !checkIfRoot() )
    return;

  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available. No users manager");
    return;
  }
  Statement st(*dbConn);
  st.beginTransaction();
  MDUserManager *userManager = *(userManagers.begin());
  int res = userManager->addUser(usr, pass, passIsHash);

  switch(res) {
  case 0:
    if (saveUserLog(st, usr)) {    
    	st.commitTransaction();
    	out.append("0\n");
    }
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 101:
    out.append("101 User exists\n");
    return;
  default:
    out.append("9 Internal error\n");
  }

  return;
}


void MDStandalone::userRemove(const std::string &usr)
{
  if( !checkIfRoot() )
    return;

  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available\n");
    return;
  }
  Statement st(*dbConn);
  st.beginTransaction();  
  MDUserManager *userManager = *(userManagers.begin());
  int res = userManager->removeUser(usr);

  switch(res) {
  case 0:
    if (saveUserLog(st, usr)) {    
    	st.commitTransaction();
    	out.append("0\n");
    } 
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 100:
    out.append("100 User does not exist\n");
    return;
  default:
    out.append("9 Internal error\n");
  }

  return;
}


void MDStandalone::userPassChange(const std::string &usr, const std::string &pass)
{
  if(user != "root" && user != usr) {
    out.append("4 Permission denied\n");
    return;
  }

  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }

  Statement st(*dbConn);
  st.beginTransaction();
  MDUserManager *userManager = *(userManagers.begin());
  int res = userManager->changeUserPass(usr, pass);
  
  switch(res) {
  case 0:
    if (saveUserLog(st, usr)) {    
    	st.commitTransaction();
    	out.append("0\n");
    } 
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 101:
    out.append("100 User does not exist\n");
    return;
  default:
    out.append("9 Internal error\n");
  }

  return;
}


void MDStandalone::userSubjectAdd(const std::string &usr, const std::string &subject)
{
  if( !checkIfRoot() )
    return;

  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }
  Statement st(*dbConn);
  st.beginTransaction();
  
  MDUserManager *userManager = *(userManagers.begin());
  int res = userManager->addUserSubject(usr, subject);

  switch(res) {
  case 0:
    if (saveUserLog(st, usr)) {    
    	st.commitTransaction();
    	out.append("0\n");
    }
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 100:
    out.append("100 User does not exist\n");
    return;
  default:
    out.append("9 Internal error\n");
  }

  return;
}

void MDStandalone::userSubjectRemove(const std::string &usr, const std::string &subject)
{
  if( !checkIfRoot() )
    return;

  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }
  Statement st(*dbConn);
  st.beginTransaction();
  MDUserManager *userManager = *(userManagers.begin());
  int res = userManager->removeUserSubject(usr, subject);

  switch(res) {
  case 0:
    if (saveUserLog(st, usr)) {
        st.commitTransaction();
        out.append("0\n");
    }
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 100:
    out.append("100 User does not exist\n");
    return;
  default:
    out.append("9 Internal error\n");
  }

  return;
}



void MDStandalone::userVomsRemove(const std::string &subject, int type)
{
  if( !checkIfRoot() )
    return;

  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }

  Statement st(*dbConn);
  st.beginTransaction();

  std::list< MDUserManager* >::const_iterator J;
  VOMSAttrUserManager *vomsUserManager;
  MDUserManager *uM ;

  J=userManagers.begin();

  bool found = false;
  while( J != userManagers.end() ) {
    uM = *(J);
    if(vomsUserManager = dynamic_cast<VOMSAttrUserManager *>(uM)) {
      DMESG("found VOMSAttrUser\n");
      found = true;
      break;
    }
    J++;
  }

  if (! found) {
    out.append("21 Function not available\n");
    return;
  }
      
  int res = vomsUserManager->removeUserVoms(subject, type);
    
  switch(res) {
  case 0:
    if (saveUserLog(st, "")) {
        st.commitTransaction(); 
        out.append("0\n");
    }
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 100:
    out.append("100 User does not exist\n");
    return;  default:
    out.append("9 Internal error\n");
  }
  
    
  return;
}


void MDStandalone::userVomsList(int type)
{
  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }
  
  Statement st(*dbConn);
  std::list< MDUserManager* >::const_iterator J;
  VOMSAttrUserManager *vomsUserManager;
  MDUserManager *uM ;
  
  J=userManagers.begin();
  bool found = false;
  while( J != userManagers.end() ) {
    uM = *(J);
    if(vomsUserManager = dynamic_cast<VOMSAttrUserManager *>(uM)) {
      DMESG("found VOMSAttrUser\n");
      found = true;
      break;
    }
    J++;
  }

  if (! found) {
    out.append("21 Function not available\n");
    return;
  }

  std::string result;
  int res = vomsUserManager->listUserVoms(result, type);

  switch(res) {
  case 0:
    out.append("0\n");
    out.append(result);
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 100:
    out.append("100 User does not exist\n");
    return;
  default:
    out.append("9 Internal error\n");
  }
  return;
}

void MDStandalone::userVomsAdd(const std::string &usr, const std::string &subject, int type)
{

  if( !checkIfRoot() )
    return;

  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }

  Statement st(*dbConn);
  st.beginTransaction();

  std::list< MDUserManager* >::const_iterator J;
  VOMSAttrUserManager *vomsUserManager;
  MDUserManager *uM ;

  J=userManagers.begin();

  bool found = false;
  while( J != userManagers.end() ) {
    uM = *(J);
    if(vomsUserManager = dynamic_cast<VOMSAttrUserManager *>(uM)) {
      DMESG("found VOMSAttrUser\n");
      found = true;
      break;     
    }
    J++;
  }

  if (! found) { 
    out.append("21 Function not available\n");
    return;
  } 

  int res = vomsUserManager->addUserVoms(usr, subject, type);

  switch(res) {
  case 0:
    if (saveUserLog(st, usr)) {
        st.commitTransaction();
        out.append("0\n");
    }
    return;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 100:
    out.append("100 User does not exist\n");
    return;
  default:
    out.append("9 Internal error\n");
  }

  return;
}


void MDStandalone::userList()
{
  if( !userManagers.size() ) {
    out.append("21 Function not available\n");
    return;
  }

  int res;
  std::list < MDUserManager * >::iterator I;
  std::list < std::string > users;
  I = userManagers.begin();
  while (I != userManagers.end() ) {
    res = (*I++)->listUsers(users);
    DMESG("listUser returned: " << res << std::endl);
    if( res ==9 ) {
      out.append("9 Internal error\n");
      return;
    }
  }
  
  out.append("0\n");
  std::list < std::string >::iterator J = users.begin();
  while ( J != users.end() ) {
    out.append(*J++);
    out.append("\n");
  }
}



void MDStandalone::userListCred(const std::string &usr)
{
  DMESG("userList "<< user << " lists '" << usr << "'\n");
  if( user != "root" && (usr !=user || !usr.size()) ) {
    out.append("4 Permission denied\n");
    return;
  }

  if( !userManagers.size() ) {
    out.append("21 Function not available\n");
    return;
  }

  MDUserManager::UserCred credentials;
  int res = 100;
  std::list < MDUserManager * >::iterator I;
  I = userManagers.begin();
  while (I != userManagers.end() ) {
    int newres = (*I++)->listUserCred(usr, credentials);
    DMESG("listUserCred returned: " << newres << std::endl);
    if( newres ==9 ) {
      out.append("9 Internal error\n");
      return;
    }
    if(!newres)
      res = 0;
  }

  switch(res) {
  case 0:
    out.append("0\n");
    break;
  case 21:
    out.append("21 Function not available\n");
    return;
  case 100:
    out.append("100 No such user\n");
    return;
  default:
    LOG("Illegal state of UserManager::listUserCred: " << res << std::endl);
    out.append("0\n");
    break;        
  }
  ostringstream oss;
  oss << credentials.pass << "\n";

  std::list< std::string >::iterator J = credentials.dbDNs.begin();
  while (J != credentials.dbDNs.end()) {
    oss << "'" << *J++ << "'";
    if(J!= credentials.dbDNs.end())
      oss << ", ";
  }
  oss << "\n";

  J=credentials.fixedDNs.begin();
  while (J != credentials.fixedDNs.end()) {
    oss << "'" << *J++ << "'";
    if(J != credentials.fixedDNs.end())
      oss << ", ";
  }
  oss << "\n";
  oss << credentials.vo << "\n";
  
  out.append(oss.str());
}


void MDStandalone::ticketGet()
{
  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }

  MDUserManager *userManager = *(userManagers.begin());

  std::string ticket;
  std::string groupstring;
  std::list< std::string > groups;
  if(memberOf(user, groups) <0)
    return;  // Error print in memberOf()

  std::list< std::string >::const_iterator I=groups.begin();
  while(I!=groups.end()){
    groupstring.append(*I++);
    if(I!=groups.end())
      groupstring.append(", ");
  }

  int res = userManager->ticketCreate(ticket, user, groupstring);

  switch(res) {
  case 0:
    out.append("0\n" + ticket + "\n");
    break;
  default:
    out.append("9 Invalid ticket: ");
    char buf[10];
    snprintf(buf, 9, "%u", res);
    out.append(buf);
    out.append("\n");
  }

  return;
}


void MDStandalone::ticketCheck(const std::string &ticket)
{
  if(userManagers.begin() == userManagers.end()) {
    out.append("21 Function not available");
    return;
  }

  MDUserManager *userManager = *(userManagers.begin());

  int res = userManager->ticketCheck(ticket);

  switch(res) {
  case 0:
    out.append("0\n");
    break;
   default:
    out.append("9 Invalid ticket: ");
    char buf[10];
    snprintf(buf, 9, "%u", res);
    out.append(buf);
    out.append("\n");
  }

  return;
}


void MDStandalone::setPublicKey(EVP_PKEY *pkey)
{
  std::list < MDUserManager * >::iterator I = userManagers.begin();
  while (I != userManagers.end() )
    (*I++)->setPublicKey(pkey);
}


void MDStandalone::setPrivateKey(EVP_PKEY *pkey)
{
  std::list < MDUserManager * >::iterator I = userManagers.begin();
  while (I != userManagers.end() )
    (*I++)->setPrivateKey(pkey);
}

uint64_t MDStandalone::dump(const std::string &directory, 
	const std::string &secOpt, 
	bool replication)
{
	bool dumpPerms = secOpt != DUMP_NO_PERMS;
	return dump(directory, dumpPerms, replication);
}
	
uint64_t MDStandalone::dump(const std::string &directory, 
	bool dumpPerms, 
	bool replication)
{
//  cout << "***** " << directory << " - " << secOpt << " - " << replication << endl;
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return 0;

  // The operation is denied if the user does not have the replication 
  // capability (R) on the root dump dir
  EntryPropList::iterator it = propList.begin();
  // If we are replicating, ignore any permissions. They were already
  // checked by the ReplicationDaemon
  
  std::list< std::string > groups;
  if (!replication && !checkPermResult(checkPermissions(*it, 'x', groups))) {
    return 0;
  }
  
  if( it->type() != EPT_DIR ) {
    out.append(string("17 No such directory ") + directory + "\n");
    return 0;
  }
  
  // Ok. We have something to dump
  out.append("0\n");

  uint64_t curXid;
  try {
    // This is the xid for the masterindex
    SubscriptionManager subMan(*dbConn);
    ostringstream oss;
    curXid = subMan.getCurrentXid();
    oss << "#xid.masterindex=" << curXid << "\n";
    out.append(oss.str());
  } catch (std::exception &ex) {
    // Ignore
    DMESG("Error getting latest XID: " << ex.what() << endl);
    curXid = 0;
  }

  // Now dump this directory recursevily    
  doDump(it->name, *it, dumpPerms, replication);

  return curXid;
}


int MDStandalone::doDump(const string& dumpRoot,
			 const EntryProps& dirProps, bool dumpPerms,
			 bool replication)
 {
  DMESG("doDump: " << dumpRoot << ", " << dirProps.name << ", "  << endl);
  Statement statement(*dbConn);
  // Needed for isolation against concurrent updates.
  // TODO: Check if this really works.
  statement.beginTransaction();
  
  DMESG("Absolute dir: " << dirProps.name << "<\n");
  ostringstream oss;
  std:string options = encodeTableOptions(dirProps.flags);

  if(dirProps.mainColumn != "file") {
	  if(options != "") {
		  options += ",";
	  }
	  options += "main=" + dirProps.mainColumn;
  }

  if (dirProps.name != "/") {
    oss << "createdir " << dirProps.name 
      << " " << options
      << "\n";
  }

  if (dumpPerms) {
    // Set directory owner and permissions
    oss << "chown " << dirProps.name << " " << dirProps.owner << "\n" 
        << "chmod " << dirProps.name << " " << dirProps.perms << "\n";

    // Remove possibly inherited ACLs
    oss << "acl_remove " << dirProps.name << " *\n";

    // Set directory ACLs
    map<string, string> aclMap;  
    aclStringToMap(dirProps.ACLs, aclMap);
    for (map<string, string>::const_iterator it = aclMap.begin(); it != aclMap.end(); it++) {
      oss << "acl_add " << dirProps.name << " " << it->first << " " << it->second << "\n";
    }
	
    // Set replication permissions
    set<string> values;
    csvStringToSet(dirProps.repGroups, values);
    for (set<string>::const_iterator it = values.begin(); it != values.end(); it++)
      oss << "rep_allow " << dirProps.name << " " << *it << "\n";
  }
  out.append(oss.str());
  
  DMESG("########## table dump\n");
  // dump table including creation of attributes
  if( dumpTable(statement, dumpRoot, dirProps, dumpPerms, replication) ) {
    return 1;
  }
  
  DMESG("######### table dump successful\n");
  // Done
  statement.commitTransaction();

  // Go down. List subdirectories
  DMESG("Dumping the sub-directories of " << dirProps.name << "\n");
  EntryPropList propList;
  int error=getEntryProps(dirProps.name + "/*", propList);
  if(handleEntryPropErrors(error, dirProps.name))
    return 1;

  for (EntryPropList::iterator it = propList.begin(); it != propList.end(); it++) {
    // We are interested only in directories.
    DMESG("Checking >" << it->name << "< for whether it is a directory\n");
    if ( it->type() == EPT_DIR ) {
      DMESG("Recursing: " << it->name << endl);
      // If the user has the replication capability on the top level directory,
      // then it can replicate all the tree
      doDump(dumpRoot, *it, dumpPerms, replication);
    }
  }
  return 0;
}

int MDStandalone::dumpEntriesPG(Statement &statement,
                                const EntryProps &dirProps,
                                const AttrList &keys,
                                bool dumpPerms,
                                bool replication,
                                vector<string> &sysAttrs,
                                std::string &query)
{

  // Now the upload
  std::string uquery = "upload " + dirProps.name;

  // Don't include the file attribute
  for ( size_t i=1; i<sysAttrs.size(); i++ ) {
        uquery += " ";
    // HACK: acls column should be called acl, but we cannot change it
    // to ensure backwards compatibility. So we keep the column as acls,
    // but internally map it to ACL.
    if (sysAttrs[i] == "acls") {
      uquery += "ACL";
    } else {
      to_upper(sysAttrs[i]);
      uquery += sysAttrs[i];
    }
  }
  for (AttrList::const_iterator I = keys.begin(); I!=keys.end(); ++I) {
    uquery += " ";
    uquery += I->first;
  }

  uquery += "\n";
  out.append(uquery);
  uquery = "";

  char buff[1001];

  if( statement.beginTransaction(true) ) {
    printError("9 Internal error", statement);
    return 1;
  }

  if(statement.exec("DECLARE getattr_c CURSOR FOR "+ query)){
    printError("9 Internal error", statement);
    return 1;
  }

  while(1){
    DMESG("GOING to FETCH"<< std::endl);
    uquery="FETCH 100 FROM getattr_c;";
    if(statement.exec(uquery)){
      printError("9 TERRIBLE ERROR fetching from cursor", statement);
      DMESG("Error FETCH" << std::endl);
      break;
    }

    if(statement.fetch()){
      DMESG("Fetch failed: Last data read" << std::endl);
      break;
    }

    SQLHSTMT s=statement.getStatement();
    SQLLEN err;
    int fields = sysAttrs.size() + keys.size();
    do {
      out.append("put");
      for(size_t i=0; i<fields; i++){
        int res = SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err) ;
        if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO) {
          continue;
        }
        out.append(" ");
        if(err==SQL_NULL_DATA) {
          out.append("\\N"); 
        } else {
          out.append("'");
          uquery = buff;
          escapeString(uquery);
          out.append(uquery);   
          while(1) {
            int res = SQLGetData(s, i+1, SQL_C_CHAR, &buff, 1000, &err);
            if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO) 
              break;
            uquery = buff;
            escapeString(uquery);
            out.append(uquery);
          }
          out.append("'");
        }
      }
      out.append("\n");
    } while(!statement.fetch());
    statement.close();
  }
  out.append("commit\n");
  statement.close();
  if(statement.exec("CLOSE getattr_c;")){
    printError("9 CLOSE does not work", statement);
    return 1;
  }

  statement.commitTransaction();
  return 0;
}


int MDStandalone::dumpTable(
    Statement &statement, 
    const string &dumpRoot,
    const EntryProps &dirProps,
    bool dumpPerms, 
    bool replication)
{
  DMESG("dumpTable: " << dirProps.name << " - " << dirProps.parent << ", flags=" << dirProps.flags 
	<<  ", dumpPerms=" << dumpPerms << ", replication=" << replication << "\n");
  
  if( !dirProps.directoryTable.size() )
    return 0;

  ostringstream oss;
  try {
    // Send the XID for this directory
    SubscriptionManager subMan(*dbConn);
    oss << "#xid." << dirProps.name << " = " << subMan.getCurrentXid() << "\n";
  } catch (std::exception &ex) {
    // Ignore
    DMESG("Error getting latest XID: " << ex.what() << endl);
  }
  
  // Dump attributes
  AttrList keys;
  //If PLAIN then we want all attributes
  if(getAttributes(dirProps.directoryTable, keys, dirProps.flags & EP_PLAIN))
    return 1;
  // Do not dump the attributes if the directory is shared, the parent is also 
  // shared and we are not the root of the dump
  if (!
    ((dirProps.flags & EP_SHARED) && (dirProps.parentFlags & EP_SHARED) && 
      dirProps.name != dumpRoot)
     ) 
  { 
    if (!keys.empty()) {
      oss << "addattr " << dirProps.name;
      AttrList::iterator I=keys.begin();

      // In Oracle, there can be only one long column and it must be last
      // So we sort the keys
      int n = keys.size();
      while( n ) {
	DMESG("Variable: " << I->first << endl);
	if(dbHelper->dbToMDType(I->second) == "text") {
	  DMESG("MOVING ORDER!\n");
	  keys.push_back(*I);
	  I = keys.erase(I);
	} else {
	  I++;
	}
	n--;
      }

      I=keys.begin();
      while(I!=keys.end()) {
        oss << " " << I->first << " " << dbHelper->dbToMDType(I->second);
        I++;
      }
      oss << "\n";
      out.append(oss.str());
    }
  }  
  
  // Set table constraints
  string buf;
  if( listTableConstraints(dirProps, statement, buf) )
    return 1;
  out.append(buf);
  if (dumpIndexes(statement, dirProps.name, dirProps.directoryTable))
    return 1;
  if (dumpEntries(statement, dirProps, keys, dumpPerms, replication))
    return 1;
  // TODO: dump sequences and other things...    
  return 0;
}

int MDStandalone::dumpIndexes(Statement &statement, const string &dir, const string &table)
{
  // TODO: dump indexes
  DMESG("dumpIndexes: Standalone\n" );
  return 0;
}

int MDStandalone::dumpEntries(Statement &statement,
			      const EntryProps &dirProps,
			      const AttrList &keys, 
			      bool dumpPerms,
			      bool replication)
{
  // TODO: Extend dump to deal with file catalog tables. Currently
  // it handles only old style tables.  
  DMESG("dumpEntries: " << dirProps.name << ", flags=" << dirProps.flags 
	<<  ", dumpPerms=" << dumpPerms << ", replication=" << replication << "\n");

  vector<string> sysAttrs;
  if(!(dirProps.flags & EP_PLAIN)) {
	  sysAttrs.push_back(dirProps.mainColumn);
  }
  if( dumpPerms && ( dirProps.flags & EP_PERMS ) && ( (dirProps.flags & EP_PLAIN) == 0 )) {
	  sysAttrs.push_back("owner");
	  sysAttrs.push_back("permissions");
	  sysAttrs.push_back("group_rights");
  }

  if ((dirProps.flags & EP_FCAT) && ( (dirProps.flags & EP_PLAIN) == 0 )) {
	  sysAttrs.push_back("link");
	  sysAttrs.push_back("created");
	  sysAttrs.push_back("size");
	  sysAttrs.push_back(dbHelper->readGUID("guid"));
	  sysAttrs.push_back("md5");
  }

  if ( dumpPerms && (dirProps.flags & EP_ACLS) && ( (dirProps.flags & EP_PLAIN) == 0 )){
	  sysAttrs.push_back("acls");
  }
  
  DMESG("Dumping attributes: \n");
  for( size_t i = 0; i < sysAttrs.size(); i++ )
    DMESG("  " << i << ": " << sysAttrs[i] << "\n");

  if (sysAttrs.empty() && keys.empty()) {
    // A plain table with no columns and, therefore, no entries.
    return 0;
  }
  

  // Build the select query
  ostringstream oss;
  oss << "SELECT ";
  string sep;  
  string attr;
  for (size_t i=0; i<sysAttrs.size(); i++) {
    oss << sep << "\"" << sysAttrs[i] << "\"";
    sep = ", ";
  }
  for (AttrList::const_iterator I = keys.begin(); I!=keys.end(); ++I) {
    oss <<sep << "\"" ;
    if ( (dirProps.flags & EP_PLAIN) == 0 )
      oss << parser->prefix() ;
    oss << I->first << "\"";
    sep = ", ";
  }
  oss << " FROM " << dirProps.directoryTable;
  
  sep = " WHERE ";
  if (dirProps.flags & EP_SHARED) {
    oss << sep << "\"dir\" = " << dirProps.id;
    sep = " AND ";
  }
   
  if (!replication) {
    // It's an interactive user. Dump only what is visible to the user.
    list<string> groups; // Group membership cache
    // Only entries for which the user has the r and x permission are dumped.
    oss << sep << permissionClause('x', dirProps.flags, groups)
      << " AND " << permissionClause('r', dirProps.flags, groups);
  }
  oss << ";";
  string query = oss.str();    
  DMESG("SQL: >" << query << "<" << std::endl);

  if( !strcmp(dbHelper->backend(), "PostgreSQL") ) 
    return dumpEntriesPG(statement, dirProps, keys, 
                         dumpPerms, replication, sysAttrs, query);


  if(statement.exec(query)) {
    ERRLOG("Could not dump directory " << dirProps.name << std::endl);
    return 1;
  }
  
  // Now the upload
  query = "upload " + dirProps.name;

  // Don't include the file attribute  
  for ( size_t i=1; i<sysAttrs.size(); i++ ) {
  	query += " ";
    // HACK: acls column should be called acl, but we cannot change it
    // to ensure backwards compatibility. So we keep the column as acls,
    // but internally map it to ACL.
    if (sysAttrs[i] == "acls") {
      query += "ACL";
    } else {
      to_upper(sysAttrs[i]);
      query += sysAttrs[i];
    }
  }
  for (AttrList::const_iterator I = keys.begin(); I!=keys.end(); ++I) {
    query += " ";
    query += I->first;
  }

  query += "\n";
  out.append(query);
  query = "";
  
  // For the system parameters we can use bind variables as we know what is
  // their maximum size. For the others, we have to read the data one by one.  
  char buff[2048];
  try {
    while( !statement.fetch() ) {
      SQLHSTMT s=statement.getStatement();
	SQLLEN err;
      query += "put";
      for(size_t i=1; i<=sysAttrs.size()+keys.size(); i++){
	int res=SQLGetData(s, i, SQL_C_CHAR, &buff, sizeof(buff), &err);
	if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO) {
	  continue;
	}
	query += " ";
	if(err==SQL_NULL_DATA) {
	  query += "\\N";
	} else {
          query += "'";
          query += escapeString(buff);
	  while(1) {
	    int res=SQLGetData(s, i, SQL_C_CHAR, &buff, sizeof(buff), &err);
	    if(res!=SQL_SUCCESS && res!=SQL_SUCCESS_WITH_INFO) {
	      break;
	    }
	    query += escapeString(buff);
	  }
          query += "'";
	}

      }


      AMGA::encodeLine(query);
      query += "\n";
      out.append(query);

      query = "";
    }

    query += "commit\n";
    out.append(query);
  } catch (SQLException &ex) {
    DMESG("Error dumping entries: " << ex.what() << endl);
    return 1;
  }
  statement.close();
  return 0;
}


void MDStandalone::grpDump(bool replication)
{
  if (!checkIfUserRepAllowed() )
    return;

  std::string query="SELECT \"name\", \"owner\", \"members\" FROM " + groupsTable;
  if (replication) {
    query += " WHERE \"owner\" != 'root'";
  }
  query += ";";

  DMESG("SQL: >" << query << "<" << std::endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Cannot read group table", statement);
    return;
  }

  char name[GROUPNAME_MAX_SIZE+1];
  char owner[USERNAME_MAX_SIZE+1];
  char members[GRPMEMBERS_MAX_SIZE+1];

  statement.bind(1, name, GROUPNAME_MAX_SIZE);
  statement.bind(2, owner, USERNAME_MAX_SIZE);
  statement.bind(3, members, GRPMEMBERS_MAX_SIZE);

  ostringstream os;
  os << "0\n";
  while( !statement.fetch() ) {
    // Create group
    os << "grp_create " << name << "\n";
    std::list< std::string > memberList;
    splitGroupString(members, memberList);
    
    // Add members to group
    for (list<string>::iterator I = memberList.begin();
    	I != memberList.end(); ++I) 
    {
    	if (replication && *I == "root") {
	  // Skip the root user
	  continue;
    	}
    	os << "grp_adduser " << name << " " << *I << "\n";    	
    }
  }
  out.append(os.str());
}


// TODO: Put this functionality into the user-managers!
int MDStandalone::userDump(bool replication)
{
  if (!checkIfUserRepAllowed())
    return -1;
  
  string query="SELECT \"name\", \"password\" FROM " +usersTable;
  if (replication)
    query += " WHERE \"name\" != 'root'"; 
  query += ";";

  DMESG("SQL: >" << query << "<" << endl);
  Statement statement(*dbConn);
  if(statement.exec(query)){
    printError("9 Cannot read user table", statement);
    return -1;
  }

  char user[USERNAME_MAX_SIZE+1];
  statement.bind(1, user, USERNAME_MAX_SIZE);  
  char hash[PASSWORD_MAX_SIZE];
  statement.bind(2, hash, PASSWORD_MAX_SIZE);
  string buf = "0\n";
  while( !statement.fetch() ) {
    // Create user
    buf += "user_create ";
    if (replication) {
      buf += "-h "; // The password is an hash, not the password itself
    }
    buf += user;
    if (replication) {
      buf += " ";
      buf += hash;
    }
    buf += "\n";
  }
  out.append(buf);

  statement.close();

  query="SELECT \"subject\", \"name\" FROM " + certificatesTable;
  if (replication)
    query += " WHERE \"name\" != 'root'"; 
  query += ";";

  DMESG("SQL: >" << query << "<" << std::endl);
  if(statement.exec(query)){
    printError("9 Cannot read certificates table", statement);
    return -1;
  }

  char certificate[CERTIFICATE_SUBJECT_MAX_SIZE+1];
  statement.bind(1, certificate, CERTIFICATE_SUBJECT_MAX_SIZE);
  statement.bind(2, user, USERNAME_MAX_SIZE);
  buf = "";
  while( !statement.fetch() ) {
    // Create group
    
    buf = "user_subject_add ";
    buf += escapeString(user);
    buf += " '";
    buf += escapeString(certificate);
    buf += "'";
    AMGA::encodeLine(buf);
    out.append(buf);
    out.append("\n");
    DMESG("Buff: >" << buf << "<" << std::endl);
  }
  return 0;
}


bool MDStandalone::checkCert(const std::string &usr, X509 *cert,
			     std::string &vo,
			     std::list< std::string > &groups,
			     std::string &eUser)
{
  DMESG("In checkCert\n");
  std::list < MDUserManager * >::const_iterator I = userManagers.begin();
  while (I != userManagers.end() ) {
    MDUserManager *um = *I++;
    DMESG("Trying " << um->umName() << " User Manager\n");
    if( um->checkCert(usr, cert, vo, groups, eUser) ) {
      VOMSAttrUserManager *voUM = dynamic_cast<VOMSAttrUserManager *>(um);
      if(vo.size() && groups.size() && voUM) {
	std::list< std::string >::const_iterator I = groups.begin();
	while(I != groups.end()) {
	  std::string groupName(vo);
	  groupName.append(":").append(*I);
	  DMESG("Trying to map group >" << groupName << "<\n");
	  groupName = voUM->getMappedGroup(groupName);
	  if( groupName.size() ) { // Existing group mapping, create group
	    user = "root";  // No problem, this cannot be called during a session!
	    DMESG("Creating on the fly group >" << groupName << "<\n");
	    if(doCreateGroup(groupName))
	      DMESG("  Group existed...\n");
	    if(grpChangeUser(groupName, eUser, true))
	      DMESG("  User was already member of that group\n");
	  }
	  I++;
	}
      }
      return true;
    }
  }
  return false;
}

/**
 * PUBLIC
 */
void MDStandalone::deleteDirectoryContents(const string &dirName,
             bool deleteDirectory)
throw (std::exception)
{
	deleteDirectoryContents(dirName, deleteDirectory, true);
}

/**
 * PRIVATE
 * 'checkperms = false' is only needed for rep_umount if the user is not root
 * but have the capability.
 */
void MDStandalone::deleteDirectoryContents(const string &dirName,
             bool deleteDirectory, bool checkperms)
throw (std::exception)
{
  // Now delete the directory
  EntryPropList propList;
  int error=getEntryProps(dirName, propList);
  if(handleEntryPropErrors(error, dirName))
    return;
  deleteDirectoryContents(propList.front(), deleteDirectory, checkperms);
}    


/**
 * PUBLIC
 */
void MDStandalone::deleteDirectoryContents(const EntryProps &dir,
					   bool deleteDirectory)
  throw (std::exception)
{
	deleteDirectoryContents(dir, deleteDirectory, true);
}
    
/**
 * PRIVATE
 *
 * Recursively deletes the contents of a directory. The flag
 * deleteDirectory controls whether the directory should be deleted
 * or not. In all cases, the directory contents and its attributes
 * are deleted.
 *
 * 'checkperms = false' is only needed for rep_umount if the user is not root
 * but have the capability.
 */
void MDStandalone::deleteDirectoryContents(const EntryProps &dir,
					   bool deleteDirectory, bool checkperms)
  throw (std::exception)
{
  DMESG("Deleting: " << dir.name << endl);
  Statement statement(*dbConn, false);
  statement.beginTransaction();
  // Delete the contents of directory. 
  StatementWrapper stWrapper(statement);
  string query;
  // Read all subdirectories. These will have to be deleted.
  if (dir.name == "/") {
    // Must delete everything
    query = "SELECT \"table_name\", \"id\", \"directory\", \"flags\" FROM "
      + masterindexTable + ";";
  } else {
    query = "SELECT \"table_name\", \"id\", \"directory\", \"flags\" FROM " 
      + masterindexTable
      + " WHERE \"directory\" = '" + dir.name 
      + "' OR \"directory\" LIKE '" + dir.name + "/%';";
  }
  DMESG("SQL: >" << query << "<\n");
  stWrapper.exec(query);
  
  char table[TABLE_NAME_MAX_SIZE];
  char id[16];
  int  flags;
  char directory[DIRNAME_MAX_SIZE];
  statement.bind(1, table, sizeof(table));
  statement.bind(2, id, sizeof(id));
  statement.bind(3, directory, sizeof(directory));
  statement.bind(4, &flags);
  
  // The MySQL odbc driver does not allow sending a request to the DB
  // containing several queries. So we have to store them on a vector
  // and execute one by one
  set< pair<string, int> > tablesToDrop;
  vector<string> dirsToDelete;
  vector<string> dirAll;
  map<string, string> ids;
  query = ""; 
  string sep("DELETE FROM " + masterindexTable + " WHERE ");
  while (!statement.fetch()) {

    dirAll.push_back(directory);

    if ( strlen(table) ) {
      tablesToDrop.insert( pair<string, int> (table, flags) );
    }
    ids[table] = id;

    if (string("/") == directory || (!deleteDirectory && dir.name == directory)) {
      // Should not delete the masterindex entry for the root directory or the
      // main directory of the tree if the used asked for it not to be deleted.
      // Instead, mark it as not having a table present.
      dirsToDelete.push_back(directory);
    } else {
      // Do not delete the entry on the
      query += sep + "\"id\"=" + id;
      sep = " OR ";
    }
  }
  statement.close();

  for( size_t i=0; i < dirAll.size(); i++ ) {
    EntryPropList propList;
    int error=getEntryProps(dirAll[i], propList);
    if(handleEntryPropErrors(error,dirAll[i])) return;
    EntryProps p(*propList.begin());

    std::list<std::string> groups;

    if(checkperms){
		int t_perms;
		if( p.sticky() ) {
		  t_perms=checkPermissions(p, 't', groups);
		  if( t_perms<= 0){
			DMESG("Permission denied" << std::endl);
			out.append("4 Permission denied\n");
			return;
		  }
		}

		t_perms = checkPermissions(p, 'w', groups);
		if(t_perms <= 0){
		  DMESG("Permission denied: " << t_perms << " " << p.flags << std::endl);
		  out.append("4 Permission denied : ");
		  out.append(dirAll[i]) ;
		  out.append("\n");
		  return;
		}
    }
  }
  
  // Update the masterindex to remove the entries 
  // for the directories that are to be deleted.
  if (!query.empty()) {
    query += ";";
    DMESG("SQL: >" << query << "<\n");
    stWrapper.exec(query); 
    statement.close();
  }
  
  // See the comment in the while loop above.
  for( size_t i=0; i<dirsToDelete.size(); i++ ) {
    query = "UPDATE " + masterindexTable + " SET \"table_name\" = '', \"flags\" = 0"
      " WHERE \"directory\"='" + dirsToDelete[i] + "';";
    stWrapper.exec(query);
    statement.close();
  }
  
  // If the top level directory is shared and if its parent directory is 
  // also shared, then this means that the table that is in use by this
  // directory is also used by other directories that are not to be deleted.
  // In this case, we can't delete the table, instead we have to delete
  // the entries for the directories deleted
  if (dir.flags & EP_SHARED && dir.parentFlags & EP_SHARED) {
    DMESG("Deleting entries from shared table: " << dir.directoryTable << "\n");
    // This directory shares the table with the parent directory.
    // Delete only the entries
    query = "DELETE FROM " + dir.directoryTable;
    sep = " WHERE ";
    for (map<string, string>::const_iterator it = ids.begin();
	 it != ids.end();
	 ++it) 
      {
	query.append(sep).append("\"dir\"=").append(it->second);
	sep = " OR "; 
      }
    query += ";";
    DMESG("SQL: >" << query << "<\n");
    stWrapper.exec(query);
    statement.close();
    
  } else {
    vector<string> fcTablesDeleted;
    set< pair<string, int> >::const_iterator it;
    for (it = tablesToDrop.begin(); it != tablesToDrop.end(); ++it) {
      if ( ((*it).second & EP_TYPE) == EPT_SEQUENCE ) {
        std::string style = dbHelper->backend();
        query = "" ;
        if( style.find("MySQL") == std::string::npos) {
          query.append("DROP SEQUENCE ").append((*it).first);
        } else {
          query.append("DELETE FROM ").append(usersequencesTable);
          query.append(" WHERE name like '%").append((*it).first).append("'");
        }
      } else if ( ((*it).second & EP_TYPE) == EPT_INDEX ) {
        continue;
      } else {
        // Shared tables will appear multiple times in the query, we must
        // make sure that we delete them only once.
    	query = "DROP TABLE ";
        std::string tableName = (*it).first;
        int flags = (*it).second;

        if (flags & EP_IMPORTED)
          continue;

        query += tableName;
        if (flags & EP_FCAT) {
          fcTablesDeleted.push_back(tableName);
        }
      }

      query += ";";

      DMESG("SQL: >" << query << "<\n");
      stWrapper.exec(query);
      statement.close();
    }
    
    // If we deleted any FC table, delete the guids from the guids table.
    if (fcTablesDeleted.size() != 0) {
      query = "DELETE FROM " + guidsTable + " WHERE ";
      sep = "";
      for (size_t i=0; i<fcTablesDeleted.size(); i++) {
      query.append(sep).append("\"reference\" LIKE '").append(fcTablesDeleted[i]).append("_%'");
      query.append(" OR \"reference\" = '").append(fcTablesDeleted[i]).append("'");
      sep = " OR ";
      }
      query += ";";
      DMESG("SQL: >" << query << "<\n");
      stWrapper.exec(query);
      statement.close();
    }
  }    

  DMESG("Commiting\n");
  // Local updates are done. Finish the transaction.
  statement.commitTransaction();
}

/*****************************
 * Replication 
 *****************************/
void MDStandalone::listSubscribers() {
  if (!checkIfRoot())
    return;

  try {
    SubscriptionManager subMan(*dbConn);
    map<string, Subscriber> subscribers = subMan.getAllSubscribers();
    vector<Subscription> subscriptions = subMan.getSubscriptions();
    
    ostringstream oss;
    oss << "0\n";
    for (map<string, Subscriber>::const_iterator it=subscribers.begin();
	 it != subscribers.end();
	 ++it) 
      {
	oss << it->second.subscriberID 
	    << ", users and groups: " << (it->second.repUsers? "yes" : "no") 
	    << ", last acked: " << it->second.lastAckedXid << "\n";
	
	for (vector<Subscription>::const_iterator subIt = subscriptions.begin();
	     subIt != subscriptions.end();
	     ++subIt) 
	  {
	    if (subIt->subscriberID == it->second.subscriberID) {
	      // Print subscription
	      oss << "- " << subIt->directory << ", permissions: " 
		  << (subIt->repPermissions? "yes" : "no") << "\n"; 
	    }
	  }
      }
    DMESG("Sending: " << oss.str() << endl);
    out.append(oss.str());    
  } catch (SQLException &ex) {
    printError("9", ex);    
  } catch (std::exception &ex) {    
    printError("9", ex);
  }
} 

void MDStandalone::repDropSubscriber(const std::string &subscriber) {
  if (!checkIfRoot())
    return;
  try {
    SubscriptionManager subMan(*dbConn);
    subMan.dropSubscriber(subscriber);
    out.append("0\n");    
  } catch (SQLException &ex) {
    // Problem with the database
    printError("9", ex);    
  } catch (ReplicationException &ex) {
    // Problem with the subscription. Probably the subscriber does not exist    
    printError("1", ex);
  }
}

void MDStandalone::listMounts() {
  if (!checkIfCapabilityAllowed("rep_list_mounts"))
    return;

  try {
  	// maps master nodes names with the text block to be sent to
  	// the client describing this node and its mounts  	
  	map<string, string> output;
  	map<string, uint64_t> masterXids;
    MountManager mManager(*dbConn);
        
    // First handle the users mounts
    UserMount userMount = mManager.getUsersMount();
    ostringstream os;
    if (userMount.id != "") {
    	string usersMaster = mManager.getUsersMaster();
    	if (output.count(usersMaster) == 0) {
	      	output[usersMaster] = "";	      	
      		masterXids[usersMaster] = mManager.getMaster(usersMaster)->masterXid;
      	}
    	os << "- Users and groups, xid: " << userMount.xid << "\n";
    	output[usersMaster].append(os.str());
    }
    
    vector<Mount> mounts;
    mManager.getAllMounts(mounts);
    DMESG("Got mounts: " << mounts.size() << endl);
    for (vector<Mount>::const_iterator it = mounts.begin(); 
      it != mounts.end(); 
      it++)
    {
      DMESG("Mount: " << it->directory << endl);
      if (output.count(it->master.id) == 0) {
      	output[it->master.id] = "";
      	masterXids[it->master.id] = it->master.masterXid;
      }
      os.str("");
      os << "- " << it->directory << ", permissions: " << it->repPermissions << " - " 
      	<< it->getStateAsString() << "\n";
      output[it->master.id].append(os.str());
    }
    os.str("");
    LogReceiverManager *manager = LogReceiverManager::getInstance();
    os << "0\n";
        
    for (map<string, string>::const_iterator it = output.begin();
	 it != output.end();
	 ++it)
      {
    	os << it->first << ", xid:" << masterXids[it->first] << " - ";
    	if (manager == NULL) {
	  // Server is not set up as a slave
	  os << "Disabled";
    	} else { 
	  switch (manager->getConnectionState(it->first)) {
	  case LogReceiverManager::Inactive: {
	    SiteManager sMan(*dbConn, debug);
	    // Inactive nodes might be unconfigured. 
	    if (sMan.isSiteKnown(it->first)) {
	      os << "Disconnected";
	    } else {
	      // There is no configuration entry for this node
	      os << "Disconnected - Not found in configuration";
	    }
	    break;
	  }
	  case LogReceiverManager::Connecting:
	    os << "Connecting"; break;
	  case LogReceiverManager::Receiving:
	    os << "Receiving"; break;
	  case LogReceiverManager::Reconnecting: 
	    os << "Reconnecting"; break;            
	  default:
	    os << "Unknown";
	  }
    	}
    	// Output all the mounts for this master
    	os << "\n" << it->second;
    }    
    out.append(os.str());
    
  } catch (SQLException &ex) {
    printError("9", ex);    
  } catch (std::exception &ex) {    
    printError("9", ex);
  }
}


void MDStandalone::umountAll() 
{
  if (!checkIfRoot()) { return; }
  if (!checkIfSlaveReplication()) { return; }
  
  try {
    MountManager mManager(*dbConn);
    mManager.umountAll();
  } catch (std::exception &ex) {
    printError("9", ex);
  }
}


void MDStandalone::umountImpl(const Mount &mount, MountManager &mManager, bool erase) 
throw (std::exception) 
{
  Statement st(*dbConn, debug);
  st.beginTransaction();   
  
  // Start by removing the subscription from the internal table
  mManager.umount(mount.directory);

  if (erase)
    deleteDirectoryContents(mount.directory, true, false);

  if (string("/") == mount.directory) {
    // Set the '/' entry on the masterindex to its initial state
    string query = "UPDATE " + masterindexTable + " SET "
      "\"master\" = '', "
      "\"rep_xid\" = NULL, "
      "\"owner_name\" = 'root', "
      "\"permissions\" = 'rwx', "
      "\"acls\" = '|system:anyuser rx|', "
      "\"rep_groups\" = '' "
      " WHERE \"directory\" = '/';";
    DMESG("SQL: >" << query << "<\n");
    if (st.exec(query)) {
      // Can fail either if the \ directory is not found or due to connection errors
      throw SQLException("Could not update root entry", st);
    }
  }
  // commit everything
  st.commitTransaction();
    
    // Now inform the master. This step is not essential, as the
    // subscription would expire by itself.
    try {      
      ReplicationDaemonConnection repConnection(mount.master.id, *dbConn);
      repConnection.unsubscribe(mount.directory);
    } catch (std::exception &ex) {
      // Log the error but otherwise ignore it.
      LOG("Could not unsubscribe from master." << ex.what() << endl);
    }	
}

void MDStandalone::umount(const string &remoteDir, bool erase) 
{
  if (!checkIfCapabilityAllowed("rep_umount")) { return; }
  if (!checkIfSlaveReplication()) { return; }

  string dir = absolutePath(remoteDir, currentDir);
  try {
    // We need the mount details first.
    MountManager mManager(*dbConn);
    auto_ptr<Mount> mount = getMountPoint(mManager, dir);    
    LOG("[Rep Client] Unmounting " << mount->directory << endl);        
    if (!checkIfMasterActive(mount->master)) { return; };    
    umountImpl(*mount, mManager, erase);

    out.append("0\n");
  } catch (std::exception &ex) {
    printError("9", ex);
    return;
  }
}

void MDStandalone::startReceiveLogs(const string &masterID)
{
  if (!checkIfCapabilityAllowed("rep_start_receive")) { return; }
  if (!checkIfSlaveReplication()) { return; }

  try {
    // We need the mount details first.
    MountManager mManager(*dbConn);

	// Are there mounts from this master?    
    vector<Mount> mounts;    
    mManager.getMountsOfMaster(masterID, mounts);    
    if (mounts.empty() && mManager.getUsersMaster() != masterID) {
    	out.append("9 No mounts from master\n");
    	return;
    }

    // Are they all synchronized?
    for (vector<Mount>::const_iterator it = mounts.begin();
    	it != mounts.end();
    	++it)
    {
    	if (it->state != Mount::Synchronized) {
    		out.append("9 Cannot start receiving from master while some mounts are not synchronized\n");
    		return;
    	}
    }

    auto_ptr<Master> master = getMaster(mManager, masterID);
    DMESG("Master " << *master << endl);
    LogReceiverManager::getInstance()->startReceiveLogs(*master, *dbConn);
    mManager.updateMasterState(masterID, true);
    out.append("0\n");
  } catch (std::exception &ex) {
    printError("9", ex);
  }
}


void MDStandalone::stopReceiveLogs(const string &master)
{
  if (!checkIfCapabilityAllowed("rep_stop_receive")) { return; }
  if (!checkIfSlaveReplication()) { return; }

  // Disconnect from master and keep a flag of whether there
  // was a connection active.
  bool wasConnected;
  try {
    LogReceiverManager::getInstance()->closeConnection(master);
    wasConnected = true;
  } catch (ReplicationException &ex) {
    wasConnected = false; 
  }

  try {
    MountManager mManager(*dbConn);
    auto_ptr<Master> mst = mManager.getMaster(master);
    if (mst.get() == NULL) {
      out.append("1 No mount from site\n");
      return;
    }
    
    if (mst->active) {
        if (!wasConnected) { 
            LOG("[ERROR] Inconsistent internal state: connection flagged as active but no connection established to master");
        }
        mManager.updateMasterState(master, false);
        out.append("0\n");  
    } else {
        if (wasConnected) {
            LOG("[ERROR] Inconsistent internal state: connection flagged as inactive but connection established to master");
        }
        out.append("1 Not connected to master\n");
    }
  } catch (std::exception &ex) {
    printError("9", ex);
  } catch (...) {
    DMESG("Unknown exception\n");
    out.append("9 Unknown exception\n");
  }
}


/** 
 * Mounts a remote directory. 
 *
 * Contacts the remote host
 */
void MDStandalone::mount(const string &masterID, 
	const string &remoteDir, 
	const string &secOpt)
{
  DMESG("mount: " << masterID << ", " << remoteDir << ", secOpt: " << secOpt << endl);
  if (!checkIfCapabilityAllowed("rep_mount")) { return; }
  if (!checkIfSlaveReplication()) { return; }
    
//  EntryPropList propList;
//  int error=getEntryProps(remoteDir, propList);
//  if(handleEntryPropErrors(error, remoteDir))
//    return;
//  EntryProps &p = propList.front();
//  for (EntryPropList::const_iterator it = propList.begin(); it != propList.end(); it++) {
//    (*it).print(true);
//  }
//  out.append("0\n");
//  return;
  
  // Values for secOpt sec_none, sec_nousers, sec_all
  bool repPermissions = secOpt != DUMP_NO_PERMS; 
  bool repSync = secOpt != DUMP_NO_SYNC; 
  
  string dir = absolutePath(remoteDir, currentDir);
  /* The success of mount depends on two factors:
      1. The master having the directory available
      2. Being possible to create the mount locally.
      It's not possible to check for these two conditions atomically,
      so one must be performed optimistically and rolledback if the other
      fails. We mount the directory locally first, because we can use
      a DB transaction to try the changes first and then commit them
      only on successful subscription to the remote server. Also, it's
      better to screen out the local erros before consuming resources
      on the master.
    */
  try {
    LOG("[Rep Client] Mounting " << remoteDir << " from " << masterID << endl);
    Statement statement(*dbConn, debug);
    // First get the information about the site        
    SiteManager sMan(*dbConn, debug);
    Site site;
    sMan.getSite(masterID, site);

    MountManager mManager(*dbConn);
    if (!checkIfMasterActive(mManager, site.name)) { return; };
    // Initiate a transaction with the database.
    // We need the mount details first.
    // Must create a new server object to perform the updates, since
    // each server object has a single MDBuffer that cannot be reused
    // accross different producers/consumers.
    // This buffer does not need to be asynchronous.    
    MDBuffer buffer;
    // Reuse the database connection
    auto_ptr<MDStandalone> server(
            dynamic_cast<MDStandalone*>(MDServerFactory::mdServer(buffer, dbConn)));
    auto_ptr<ReplicationDaemonConnection> connection;
    try {
      // The current user must have permissions to create the target            
      // replication directory. 
      // Copy the security context. Depending on the mount options
      // this context might be overwritten. If security settings are 
      // not replicated, the new metadata is created with the local
      // user context.
      server->setUser(user);
      server->setPermissionMask(permissionMask);
      server->setGroupRightsMask(groupRightsMask);
      // Inform the server object that it is being used for saving replication
  	  // commands. This suppresses permission checking
      server->setReplicating(true);
      // Supress opening a transaction during the upload used by the synchronization
      server->setSynchronizing(true);
      // Tell the server who is the master for which we are 
      // going to execute the next commands.
      server->setMaster(site.name);

      statement.beginTransaction();
      // Now create the local mount. This can be rolled back
      // if the remote subscription fails.
      mManager.mount(site.name, dir, repPermissions, *server);

      if (dir == "/") {
        // Set the master on the root directory
        string query = "UPDATE " + masterindexTable + " SET \"master\" = '" + masterID 
          + "' WHERE \"directory\" = '/';";
        DMESG("SLQ: " << query << endl);
        if (statement.exec(query)) {
          // Can fail either if the \ directory is not found or due to connection errors
          throw SQLException("Could not update root entry", statement);
        }
      }
      // Contact the remote host
      connection.reset(new ReplicationDaemonConnection(site));
      connection->subscribe(dir, repPermissions, true);

      // The synchronization is handled in a separate transaction
      statement.commitTransaction();
    } catch (std::exception &ex) {
      statement.rollbackTransaction();
      printError("9", ex);
      return;
    }
    try {
      // Finally, copy the data from the remote host
      mManager.updateMountState(dir, Mount::Synchronizing);
      if (repSync)
        connection->synchronize(dir, mManager, *server, &out);
      mManager.updateMountState(dir, Mount::Synchronized);
      out.append("0\n");
    } catch (std::exception &ex) {
      LOG("Could not create subscription: " << ex.what() << endl);
      // Synchronization failed. Go back on the changes performed on 
      // the first part of this method, when the mount was created.
      try {
          // Remove the subscription from the internal table
          mManager.umount(dir);
      } catch (std::exception &ex) {
          LOG("Error cleaning failed synchronization (unmounting): " << ex.what() << "\n");
      }
      try {
          // Unsubscribe from the server
          connection->unsubscribe(dir);
      } catch (std::exception &ex) {
          LOG("Error cleaning failed synchronization (unsubscribing): " << ex.what() << "\n");
      }
      if (dir == "/") {      
        string query = "UPDATE " + masterindexTable + " SET \"master\" = '' " 
          "WHERE \"directory\" = '/';";
        DMESG("SLQ: " << query << endl);
        if (statement.exec(query)) {
          // Can fail either if the \ directory is not found or due to connection errors
          LOG("Could not update root entry: " << statement.getError() << "\n");
        }
      }
      printError("9", ex);
      return;
    }
  } catch (std::exception &ex) {
      DMESG("Could not create subscription: " << ex.what() << endl);
      printError("9", ex);
      return;
  }
}

void MDStandalone::mountUsers(const string &masterID) {
  if (!checkIfUserRepAllowed()) { return; }
  if (!checkIfSlaveReplication()) { return; }

  /* The success of mounting users depends on:
   * 1. No connection to the master already established
   * 2. No local users or groups defined
   * 3. We are not replicating the users from some other master.
   */
  LOG("[Rep Client] Replicating users from " << masterID << endl);
  
  bool userManagerCopied = false;
  try {
    SiteManager sMan(*dbConn, debug);
    Site site;
    sMan.getSite(masterID, site);
    
  	MountManager mManager(*dbConn);
  	// 1. No connection to the master established
    if (!checkIfMasterActive(mManager, site.name)) { return; };

    // Initiate a transaction with the database.
    Statement statement(*dbConn, false);
    statement.beginTransaction();

    // Check if there are any users or groups apart from 
    // the root user and root owned groups
    // Oracle does not support LIMIT

    string query = "(SELECT \"name\" FROM " + groupsTable + " WHERE \"owner\" != 'root') "
    	" UNION (SELECT \"name\" FROM " + usersTable + " WHERE \"name\" != 'root');";
    DMESG("SQL: " << query << endl);
    if(statement.exec(query)) {
   		printError("9 Could not list users", statement);
    	return;
  	}
  	if (statement.fetch()) {
  		// This error could be because there is no data to return 
  		// or do to something else
  		if (statement.getErrorCode() != SQL_NO_DATA) {
  			printError("9 Error connecting to DB", statement);
  			return;			
    	}
    	// No data, it means there are no local users or groups except from the
    	// root user. 
  	} else {
  		// There is data
  		out.append("9 Cannot mount users while local users and groups exist\n");
  		return;
  	}
  	statement.close();

	// We can proceed with mounting the users/groups    
    // Store in the DB configuration that we are replicating the users
    // from a remote site. This can be rolled back if the remote subscription fails.
    mManager.mountUsers(masterID);

    // Now contact the remote host
    ReplicationDaemonConnection connection(site);
    
    // Must create a new server object to perform the updates, since
    // each server object has a single MDBuffer that cannot be reused
    // accross different producers/consumers.
    // This buffer does not need to be asynchronous.    
    MDBuffer buffer;
    // Reuse the database connection
    auto_ptr<MDStandalone> server(
            dynamic_cast<MDStandalone*>(MDServerFactory::mdServer(buffer, dbConn)));
    // Copy the security context. Depending on the mount options
    // this context might be overwritten. 
    server->setUser(user);
    // These two should be irrelevant for the user/group commands, but even so,
    // let's set them.
    server->setPermissionMask(permissionMask);
    server->setGroupRightsMask(groupRightsMask);
    // Inform the server object that it is being used for saving 
    // replication commands. This suppresses permission checking
    server->setReplicating(true);
    // Tell the server who is the master for which we are 
    // going to execute the next commands.
    server->setMaster(masterID);

    // Initialize User Manager
    userManagerCopied = true;    
  	server->initializeUserManager(userManagers);
    
    connection.subscribeUsers(mManager, *server);
    // The synchronization is handled in a separate transaction
    statement.commitTransaction();
    out.append("0\n");
    
  } catch (std::exception &ex) {
  	if (userManagerCopied) { 
  		initializeUserManager(userManagers);
  	}
    DMESG("Could not replicate users: " << ex.what() << endl);
    // Synchronization failed. Go back on the changes performed on 
    // the first part of this method, when the mount was created.
    printError("9", ex);
    return;
  }
}

void MDStandalone::umountUsers() {
  if (!checkIfUserRepAllowed()) { return; }
  if (!checkIfSlaveReplication()) { return; }
  	
  try {
  	Statement statement(*dbConn, debug);
    statement.beginTransaction();
    // We need the mount details first.
    MountManager mManager(*dbConn);
    
    string usersMaster = mManager.getUsersMaster();
    if (usersMaster == "") {
    	out.append("9 Users are not being replicated\n");
    	return;
    }    
    LOG("[Rep Client] Unmounting users. Master: " << usersMaster <<"\n");
    if (!checkIfMasterActive(mManager, usersMaster)) { return; };     
       
    // Start by removing the subscription from the internal table
    mManager.umountUsers(usersMaster);
    
    // Delete the non-local users. That is, all except root.
    string query = "DELETE FROM " + usersTable + " WHERE \"name\" != 'root';";    
    DMESG("SQL: " << query << endl);
    if(statement.exec(query)) {
      if (statement.getErrorCode() != SQL_NO_DATA) {
	printError("9 Could not delete imported users", statement);   		
	return;
      }
    }
    // Delete the non-local groups. That is, all except groups owned by root  	
    query =  "DELETE FROM " + groupsTable + " WHERE \"owner\" != 'root';";
    DMESG("SQL: " << query << endl);
    if(statement.exec(query)) {
      if (statement.getErrorCode() != SQL_NO_DATA) {
	printError("9 Could not delete imported groups", statement);
	return;
      }
    }
    // Local database is clean.
    // commit everything
    statement.commitTransaction();
    
    // Now inform the master. This step is not essential, as the
    // subscription would expire by itself.
    try {
      ReplicationDaemonConnection repConnection(usersMaster, *dbConn);
      repConnection.unsubscribeUsers();
    } catch (std::exception &ex) {
      // Log the error but otherwise ignore it.
      LOG("Could not unsubscribe from master: " << ex.what() << endl);
    }
    out.append("0\n");
  } catch (std::exception &ex) {
    printError("9", ex);
  }
}

/**
 * Proxies a remote directory.
 *
 */


#define ALLOW_FED_EXEC 1024
#define ALLOW_FED_CONN 2048

#define FED_NONE   0
#define FED_ALL    1
#define FED_RANDOM 2

int MDStandalone::findFedFlags(){
  int flags = MDInterpreter::findType( cmdToken );

  if ( flags < 0 ) return flags;

  if ( MDServer::allowFedExecFail )
    flags |= ALLOW_FED_EXEC;
    
  if ( MDServer::allowFedConnFail )
    flags |= ALLOW_FED_CONN;

  return flags;
}

/**
 * Proxies a remote directory.
 *
 */
void MDStandalone::proxyMount(const string &masterID,
        const string &remoteDir)
{
  DMESG("proxymount: " << masterID << ", " << remoteDir << endl);

  // First get the information about the site
  SiteManager sMan(*dbConn, debug);
  Site site;
  try {
    sMan.getSite(masterID, site);
  } catch (std::exception &ex) {
    printError("9", ex);
    return;
  }

  if(dbConn) {

    Statement statement(*dbConn);
    if(statement.beginTransaction()) {
      printError("9 Internal error", statement);
      return;
    }

    EntryPropList propList;
    int error=getEntryProps(remoteDir, propList, "", false);
    if(handleEntryPropErrors(error, remoteDir))
      return;
    EntryProps p(*propList.begin());

    if( ! (p.flags & EP_TYPE) ){
      out.append("16 Directory does not exist\n");
      return ;
    }

    std::list< std::string > groups;
    if (!checkPermResult(checkPermissions(p, 'w', groups))) {
      return;
    }

    std::stringstream query;
    query << "INSERT INTO proxy ( \"directory\", \"mnt_master\") "
          << " VALUES ('" << p.name 
          << "', '" << masterID << "'); " ;
    
    query << "UPDATE " << masterindexTable 
          << " SET \"flags\" = " << (EPT_DIR | EP_PROXY)
          << " WHERE \"directory\" = '" << p.name << "';";

    DMESG("SQL: >" << query.str() << "<" << std::endl);

    if(statement.exec(query.str())){
      out.append("9 Error executing query: " + statement.getError() + "\n");
      return;
    }

    if (saveLog(statement, p.parent)) {
      statement.commitTransaction();
      out.append("0\n");
    }
  } else {
      out.append("9 Error executing query \n");
      return;
  }
}

/**
 * Unproxies a remote directory.
 *
 */
void MDStandalone::proxyUmount(const string &remoteDir)
{
  DMESG("proxy umount: "  << remoteDir << endl);

  if(dbConn) {

    Statement statement(*dbConn);
    if(statement.beginTransaction()) {
      printError("9 Internal error", statement);
      return;
    }

    EntryPropList propList;
    int error=getEntryProps(remoteDir, propList);
    if(handleEntryPropErrors(error, remoteDir))
      return;
    EntryProps p(*propList.begin());

    if ( (p.flags & EP_PROXY) == 0 ) {
      out.append("4 Error : not proxy type directory\n");
      return;
    }

    std::list< std::string > groups;
    if (!checkPermResult(checkPermissions(p, 'w', groups))) {
      return;
    }
    
    std::stringstream query;
    query << "DELETE FROM proxy "
          << " WHERE \"directory\" = '" 
          << p.name << "';" ;
    
    query << "UPDATE " << masterindexTable 
          << " SET \"flags\" = " << EPT_DIR
          << " WHERE \"directory\" = '" 
          << p.name << "';";
    
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    
    if(statement.exec(query.str())){
      out.append("9 Error executing query: " + statement.getError() + "\n");
      return;
    }
    
    if (saveLog(statement, p.parent)) {
      statement.commitTransaction();
      out.append("0\n");
    }
  } else {
      out.append("9 Error executing query \n");
      return;
  }
          
}

#define PROXY_CODE 1000

int MDStandalone::printProxy(EntryProps& p)
{
  std::stringstream query;
  std::stringstream result;

  int flags = findFedFlags();
  if (flags < 0) return flags;

  if(dbConn) {
    query << "SELECT \"id\",\"name\",\"hostname\",\"port\"," 
          << "\"login\",\"password\",\"use_ssl\","
          << "\"authenticate_with_certificate\",\"cert_file\",\"key_file\","
          << "\"use_grid_proxy\",\"verify_server_cert\",\"trusted_cert_dir\","
          << "\"require_data_encryption\",\"cert_password\" "
          << " FROM proxy, sites WHERE "
          << "proxy.\"directory\" = '" << p.name 
          << "' AND proxy.\"mnt_master\" = sites.\"name\";" ;
    DMESG("SQL: >" << query.str() << "<" << std::endl);

    Statement st(*dbConn);
    if(st.exec(query.str())){
      out.append("9 Error executing query: " + st.getError() + "\n");
      return 9;
    }

    Site s;
    short int useSSL;
    short int authenticateWithCertificate;
    short int useGridProxy;
    short int verifyServerCert;
    short int requireDataEncryption;
    
    st.bind(1, &(s.id));
    st.bind(2, s.name, sizeof(s.name));
    st.bind(3, s.hostname, sizeof(s.hostname));
    st.bind(4, &(s.port));
    st.bind(5, s.login, sizeof(s.login));
    st.bind(6, s.password, sizeof(s.password));
    st.bind(7, &useSSL);
    st.bind(8, &authenticateWithCertificate);
    st.bind(9, s.certFile, sizeof(s.certFile));
    st.bind(10, s.keyFile, sizeof(s.keyFile));
    st.bind(11, &useGridProxy);
    st.bind(12, &verifyServerCert);
    st.bind(13, s.trustedCertDir, sizeof(s.trustedCertDir));
    st.bind(14, &requireDataEncryption);
    st.bind(15, s.certPassword, sizeof(s.certPassword));

    s.useSSL = (useSSL == 1) ? true : false ; 
    s.authenticateWithCertificate = ( authenticateWithCertificate == 1) ? true : false ;
    s.useGridProxy = (useGridProxy==1)? true : false ;
    s.verifyServerCert = (verifyServerCert==1)? true : false ;
    s.requireDataEncryption = (requireDataEncryption==1) ? true : false ;


    if ( MDServer::serverSideFed == 0 ) {
      result << PROXY_CODE << " Federated Directory Accessed\n" ;
      result << flags << "\n";

      while( !st.fetch() ) {
        result << s.hostname << "\n"
               << s.port << "\n" ;
      }
      p.directoryTable = result.str();
      DMESG("Proxy: >" << p.directoryTable << "<" << std::endl);
      out.append(result.str());
    } else {
      if(mdfed == NULL)
        mdfed = new MDFed(debug);
      // mdfed->configF = configFile;
      mdfed->lastCommand = nextCommand;

      std::string errmsg;

      mdfed->clearFedHosts() ;
      while( !st.fetch() ) {
        mdfed->addFedHost(s.hostname, s.port, s);
      }

      mdfed->decodeFedFlags(flags);

      if ( mdfed->initFed(errmsg, true) != 0 ) {
        DMESG("intFed error \n");
        out.append("99 ");
        out.append(errmsg);
        out.append("\n");
        return 9;
      }

      out.append("0 ");

      MDClient* child ;
      child = (MDClient*) mdfed->currentChild->mdc;
      if( child->resultCols > 0 )
        out.append(child->resultCols);

      out.append("\n");

      while ( ! mdfed->eot() ) {
         std::string line;
         mdfed->fetchRow(line, false);
         out.append(line);
         out.append("\n");
      }

    } 
    return PROXY_CODE ;
  } else {
      out.append("9 Error executing query \n");
      return 9;
  }
}

void MDStandalone::proxyPolicy()
{
  int flags = 0;

  if ( MDServer::allowFedExecFail )
    flags |= ALLOW_FED_EXEC;

  if ( MDServer::allowFedConnFail )
    flags |= ALLOW_FED_CONN;

  out.append("0\n");
  out.append(flags);
  out.append("\n");
  return ;
}

void MDStandalone::proxyCheckFailure(std::string code)
{
  out.append("0\n");

  if ( MDServer::serverSideFed == 0 ) {
    out.append(code);
  } else { 
    if( mdfed == NULL ) {
      out.append("0");
    } else {
      out.append( mdfed->failureCode);
    }
  }
  out.append("\n");
}

void MDStandalone::proxyListMounts()
{
  std::stringstream query; 

  if(dbConn) {
    query << "SELECT directory, mnt_master FROM proxy;" ;
    DMESG("SQL: >" << query.str() << "<" << std::endl);

    Statement st(*dbConn);
    if(st.exec(query.str())){ 
      out.append("9 Error executing query: " + st.getError() + "\n");
      return;
    }

    char dire[1024];
    char mast[64];

    st.bind(1, dire, 1024);
    st.bind(2, mast, 64);

    out.append("0\n");
    while( !st.fetch() ) {
      out.append(dire);
      out.append("\n");
      out.append(mast);
      out.append("\n");
    }
  } else {
      out.append("9 Error executing query \n");
      return;
  }

}


void MDStandalone::synchronize(const string &remoteDir) {
  if (!checkIfRoot()) { return; }
  if (!checkIfSlaveReplication()) { return; }

  string dir = absolutePath(remoteDir, currentDir);
  try {
    // We need the mount details first.
    MountManager mManager(*dbConn);
    auto_ptr<Mount> mount = getMountPoint(mManager, dir);
    ReplicationDaemonConnection repConnection(mount->master.id, *dbConn);

    MDBuffer buffer;
    // Reuse the database connection
    auto_ptr<MDStandalone> server(
            dynamic_cast<MDStandalone*>(MDServerFactory::mdServer(buffer, dbConn)));

    repConnection.synchronize(mount->directory, mManager, *server);
    mManager.updateMountState(dir, Mount::Synchronized);
    out.append("0\n");
  } catch (std::exception &ex) {
    DMESG("Could not synchronize: " << ex.what() << endl);
    printError("9", ex);
  }
}

/* 
 * The following operations are for the master side.
 */
void MDStandalone::changeRepGroups(const std::string &directory, const std::string &grp, bool allow) {
  // Add the group to the list
  // Get all subdirectories
  string dir = absolutePath(directory, currentDir);
  
  vector<DirLine> subDirs;
  int ret;
  if ((ret = getSubDirs(dir, subDirs)) != 0) {
    out.append("9 Internal error reading from masterindex\n");
    return;
  }
  if (subDirs.empty()) {
    out.append("4 Directory not found\n");
    return;
  }
  string groupName = expandGroupName(grp);
  std::list< std::string > groups;
  Statement st(*dbConn);
  st.beginTransaction();
  
  // Get all the subscriptions
  SubscriptionManager sMan(*dbConn);
  set<string> subscribedDirs = sMan.getSubscribedDirs();
//  for (set<string>::const_iterator it=subscribedDirs.begin();
//  	it != subscribedDirs.end(); ++it) 
// {
// 	cout << *it << endl;
// }
      
  for (vector<DirLine>::const_iterator it = subDirs.begin(); it != subDirs.end(); it++) {
//    // We need an EntryProps object to call checkPermissions
//    EntryProps aux;
//    // These four fields are used by checkPermissions
//    aux.master = it->master;
//    aux.owner = it->owner;
//    aux.perms = it->permissions;
//    aux.ACLs = it->acls;
//    // Not used, but copy them anyway
//    aux.name = it->directory;
//    aux.directoryTable = it->table;
//    aux.repGroups = it->repGroups;
//    if (!checkPermResult(checkPermissions(aux, 'w', groups))) {
//      return;
//    }

    // The user has permission
    set<string> users;
    csvStringToSet(it->repGroups, users);
    bool updated = false;
    // If the user is already present, ignore.
    if (allow && users.count(groupName) == 0) {
      // The user is not present. Add it
      users.insert(groupName);
      updated = true;

    } else if (!allow && users.count(groupName) != 0) {
      // The user is present. Remove it.
      users.erase(groupName);
      updated = true;
    }
    if (updated) {
      // TODO a prepared statement could help here, depending on
      // the number of subdirectories
      string query = "UPDATE " + masterindexTable
          + " SET \"rep_groups\" = '" + setToCsvString(users)
          + "' WHERE \"directory\"=\'" + it->directory + "'; ";
      DMESG("SQL >" << query << "<\n");
      if (st.exec(query)) {
        out.append("4 Error executing query: " + st.getError() + "\n");
        // Automatic rollback
        return;
      }
      if (subscribedDirs.count(it->directory) != 0) {
      	// There is at least one subscriber interested on this directory.
      	string command = allow ? "rep_allow " : "rep_disallow ";
      	command += it->directory;
      	command += " ";
      	command += grp;
      	logWriter->saveLogCommand(st, command, it->directory, user, currentDir, permissionMask, groupRightsMask);
      }      
    }
  }
  
  if (st.commitTransaction()) {
  	 printError("9 Problem committing transaction", st);  	
  } else {
    out.append("0\n");
  }
  // All done!
}


/** Adds the R capability to the subtree rooted on the given directory.
 * - The user is allowed to update the ACLs of all subdirectories.
 * 
 * - preconditions: 
 *      The current directory does not have the R capability.
 *      There's a transaction open on the current connection
 */
void MDStandalone::repAllow(const std::string &directory, const std::string &grp) 
{
  if (!replicating) {
    if (!checkIfRoot()) { return; }  
  }
  changeRepGroups(directory, grp, true);
}


/** Removes the R capability from the subtree rooted on the given directory.
 * Performs the following checks:
 * - The user is allowed to update the ACLs of all subdirectories.
 * - The parent subdirectory does not have the R capability.
 * 
 * - preconditions: 
 *      The current directory has the R capability.
 *      There's a transaction open on the current connection
*/
void MDStandalone::repDisallow(const std::string &directory, const std::string &grp) {
  // If the user has permission for the parent directory, forbid this operation
  DMESG("repDisallow " << directory << ", " << grp << endl);
  if (!replicating) {
    if (!checkIfRoot()) { return; }
  }
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  if(handleEntryPropErrors(error, directory))
    return;
  EntryProps &p = propList.front();  
    
  string groupName = expandGroupName(grp);
  
  string dir = absolutePath(directory, currentDir);
  if (dir != "/") {
    // We only need to perform this check if the directory has a parent
    set<string> pAllowedGroups;
    csvStringToSet(p.parentRepGroups, pAllowedGroups);
    if (pAllowedGroups.count(groupName) != 0) {
      // Replication capability is set for this group on the parent dir. 
      out.append("4 Parent directory has replication activated for given group. Cannot disallow subdirectory.\n");
      return;
    }  
  }
  
  // Now update
  changeRepGroups(directory, groupName, false);
}


int MDStandalone::getCapabilities(const std::string &u,
				  Capabilities &caps)
{
  std::stringstream query;
  query << "SELECT \"capabilities\" FROM " << usersTable 
	<< " WHERE \"name\" = '" << u << "';";
  DMESG("SQL: >" << query.str() << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query.str())){
    return 9;
  }

  char c[CAPABILITIES_MAX_SIZE];

  statement.bind(1, c, CAPABILITIES_MAX_SIZE);

  if(statement.fetch())
    return 34;

  csvStringToSet(c, caps);
  
  return 0;
}


int MDStandalone::updateCapabilities(const std::string &u,
				     Capabilities &caps)
{
   std::string c = setToCsvString(caps);

  std::stringstream query;
  query << "UPDATE " + usersTable + " SET \"capabilities\" = '" << c 
	<< "' WHERE \"name\" = '" << u << "';";
   DMESG("SQL: >" << query.str() << "<" << std::endl);

  Statement statement(*dbConn);
  if(statement.exec(query.str())){
    return 34;
  }

  return 0;
}


void MDStandalone::repUsersDisallow(const std::string &who)
{  
  capabilitiesRemove(who, "user_rep");
}


void MDStandalone::repUsersAllow(const std::string &who)
{
  capabilitiesAdd(who, "user_rep");
}


void MDStandalone::capabilitiesRemove(const std::string &who,
				      const std::string &cap)
{  
  DMESG("capabilitiesRemove called with: " << who 
	<< ", what: " << cap << "\n");

  if (!replicating) {
    if (!checkIfRoot()) { return; }
  }
  
  Capabilities caps;
  if( getCapabilities(who, caps) ) {
    out.append("34 user unknown\n");
    return;
  }
  
  caps.erase(cap);

  if( updateCapabilities(who, caps) ) {
    out.append("9 Internal error: Could not modify capabilities\n");
    return;
  }
  
  out.append("0\n");
}


void MDStandalone::capabilitiesAdd(const std::string &who,
				   const std::string &cap)
{
  DMESG("capabilitiesAdd called with: " << who 
	<< ", what: " << cap << "\n");
  if (!replicating) {
    if (!checkIfRoot()) { return; }
  }

  Capabilities caps;
  if( getCapabilities(who, caps) ) {
    out.append("34 user unknown\n");
    return;
  }

  caps.insert(cap);
  if( updateCapabilities(who, caps) ) {
    out.append("9 Internal error: Could not modify capabilities\n");
    return;
  }

  out.append("0\n");
}


void MDStandalone::capabilitiesList(const std::string &who)
{
  DMESG("capabilitiesList called with: " << who << "\n");
  std::string userName = who;
  if(!who.size())
    userName = user;

  Capabilities caps;
  if( getCapabilities(userName, caps) ) {
    out.append("34 user unknown\n");
    return;
  }

  out.append("0\n");
  Capabilities::const_iterator I = caps.begin();
  while( I!=caps.end() ) {
    out.append(*I++);
    out.append("\n");
  }

}



void MDStandalone::repShowPermissions() {
  // Only root for the time being.
  if (!checkIfRoot()) { return; }

  int ret;
  vector<DirLine> dirs;
  if ((ret = getSubDirs("/", dirs)) != 0) {
    out.append("9 Internal error reading from masterindex\n");
    return;
  }
  if (dirs.empty()) {
    out.append("4 Directory not found\n");
    return;
  }

  string buf = "0\n";
  for (vector<DirLine>::const_iterator it = dirs.begin(); it != dirs.end(); it++) {
    buf.append(it->directory).append(" [").append(it->repGroups).append("]\n");
  }
  out.append(buf);
}

//! Handle errors of getEntryProps()
/** 
 * Throws an exception if the user does not have permissions. 
 * 
 */
void MDStandalone::checkEntryPropErrors(int error, const std::string &file) 
throw (std::runtime_error)
{
  if(file.size()) {
    switch (error){
    case 9:
      throw runtime_error("9 Internal error processing: " + file);
    case 1:
      throw runtime_error("1 File or directory not found: " + file);
    case 4:
      throw runtime_error("4 Permission denied: "+ file);
    }
    return;
  }

  switch (error){
  case 9:
    throw runtime_error("9 Internal Error");
  case 1:
    throw runtime_error("1 File or directory not found");
  case 4:
    throw runtime_error("4 Permission denied");
  }
  return;
}

/** 
 * Converts a comma separated string to a set
 */
void MDStandalone::csvStringToSet(const std::string &csvString,
                 std::set<string> &values)
{   
//  cout << "Before: >" << usersStr << "<" << endl;
  if (!csvString.empty()) {    
    split(values, csvString, boost::is_any_of(","), boost::token_compress_on);  
//    cout << "Split: >";
//    copy(values.begin(), values.end(), ostream_iterator<string>(cout, ","));
//    cout << "<" << endl; 
  }
}

/** 
 * Converts a set to a csv string
 */
string MDStandalone::setToCsvString(const std::set<string> &values) {
  string str;
  string sep = "";
  for (set<string>::const_iterator it = values.begin(); it != values.end(); it++) {
//    cout << str << " - " << *it << endl;
    str += sep + *it;
    sep = ",";
  }
//  cout << str << endl;
  return str;
}


void MDStandalone::checkReplicatePermission(const std::string &directory)
throw (runtime_error)
{
  DMESG("checkReplicatePermission. User: " << user << ", Dir: " << directory << endl);
  if (user == "root") {
    return; 
  }  
  EntryPropList propList;
  int error=getEntryProps(directory, propList);
  // Throws exception if there is an error
  checkEntryPropErrors(error, directory);
  
  // Use only the first element. Should be the right one
  EntryPropList::iterator it = propList.begin();
  it->print(debug);

  list<string> groups;
  memberOf(user, groups);
  groups.push_back("system:anyuser");
  
  set<string> allowedGroups;
  csvStringToSet(it->repGroups, allowedGroups);
  
  for (list<string>::const_iterator it=groups.begin(); it !=groups.end(); it++) {    
    DMESG("Testing group: " << *it << endl);
    if (allowedGroups.count(*it) != 0) {
      // This group is allowed
      return;
    }
  }
  // None of the users groups is allowed  
  DMESG("Permission denied" << std::endl);
  throw runtime_error("4 Permission denied\n");
}

/**
 * Verifies if the current user is root. If not, returns false 
 * and writes an error message on the out buffer.
 */
bool MDStandalone::checkIfRoot()
{
  if(user == "root") {
    return true;
  } else {
    out.append("4 Permission denied\n");
    return false;
  }
}


bool MDStandalone::checkIfUserRepAllowed() 
{
  if( user == "root" )
    return true;

  Capabilities caps;
  if( getCapabilities(user, caps) ) {
    out.append("34 user unknown\n");
    return false;
  }
  if( caps.find("user_rep") != caps.end() )
    return true;

  out.append("4 Permission denied\n");
    return false;
}

bool MDStandalone::checkIfCapabilityAllowed(const string &capability)
{
  if( user == "root" )
    return true;

  Capabilities caps;
  if( getCapabilities(user, caps) ) {
    out.append("34 user unknown\n");
    return false;
  }
  if( caps.find(capability) != caps.end() )
    return true;

  out.append("4 Permission denied\n");
    return false;
}


bool MDStandalone::checkIfSlaveReplication()
{
  if (ReplicationManager::getInstance()->isSlave()) {
    return true;
  } else {
    out.append("1 Slave replication is disabled\n");
    return false;
  }
}

bool MDStandalone::checkIfMasterActive(MountManager &mManager, const string &master) {
	auto_ptr<Master> mst = mManager.getMaster(master);
	if (mst.get() == NULL) {
		return true;
	} else {
		return checkIfMasterActive(*mst);
	} 	
}

bool MDStandalone::checkIfMasterActive(Master &master) {  
  if (master.active) {
    out.append("1 Operation not allowed. Master in active state.\n");
    return false;
  } else {
    return true;
  }
}

//bool MDStandalone::checkIfConnectedToMaster(const string &master) {
//  if (LogReceiverManager::getInstance()->isConnectionOpen(master)) {
//    out.append("1 Operation not allowed. Connected to master.\n");
//    return false;
//  } else {
//    return true;
//  }
//}

auto_ptr<Mount> MDStandalone::getMountPoint(MountManager &mManager, const string &remoteDir) {
  string dir = absolutePath(remoteDir, currentDir);
  auto_ptr<Mount> mount = mManager.getMountInfo(dir);
  if (mount.get() == NULL) {
    throw ReplicationException("Mount point not found");
  } else {
    return mount;
  }
}

auto_ptr<Master> MDStandalone::getMaster(MountManager &mManager, const string &masterID) {
  auto_ptr<Master> master = mManager.getMaster(masterID);
  if (master.get() == NULL) {
    throw ReplicationException("Node not found");
  } else {
    return master;
  }
}

/**
 * Saves the current log on the database.
 * Returns true if succeeded, false otherwise.
 */
bool MDStandalone::saveLog(Statement &statement, const std::string &parent) {
  DMESG("Save log\n");
  // [rep] Sucessful operation. Write the log
  try {
    logWriter->saveStoredLog(statement, 
        parent, user, currentDir, permissionMask, groupRightsMask);
    return true;
  } catch (std::exception &ex) {
    printError(string("9 ") + ex.what(), statement);
    return false;  // Transactions are automatically rolled back
  }
}

bool MDStandalone::saveUserLog(Statement &statement, const string &user) {
	if (user == "root") {
		return true;
	} else {
		return saveUserOrGroupLog(statement);
	}
}

bool MDStandalone::saveGroupLog(Statement &statement, const string &group, const string *user) {
	// Don't replicate any commands related to the root user.
	if (user != NULL && *user == "root") {
		return true;
	}	
	std::string name = expandGroupName(group);
	if (starts_with(name, "root:")) {
		return true;
	}	
	return saveUserOrGroupLog(statement);	
}

bool MDStandalone::saveUserOrGroupLog(Statement &statement) {
  // [rep] Sucessful operation. Write the log
  try {
  	logWriter->saveStoredLog(statement, 
        USER_CMD_META_DIR, user, currentDir, permissionMask, groupRightsMask);
    return true;
  } catch (std::exception &ex) {
    printError(string("9 ") + ex.what(), statement);
    return false;  // Transactions are automatically rolled back
  }
}

UploadHandle::UploadHandle() : 
	operationsCount(0), cachedParams(0), bulkSize(0), genGUID(false), userProvidedAttrs(-1),
	logWriter(NULL), aborted(false)
{}

UploadHandle::~UploadHandle() {
//  bool debug = true;
//  DMESG("~UploadHandle\n");
  if (logWriter != NULL) {
    logWriter->abort();
  }
  for(size_t i=0; i<buffers.size(); i++) {
    delete [] buffers[i];
    delete [] types[i];
  }
}

ostream &operator<<(std::ostream &out, const MDStandalone::DirLine &dl) {
    out << dl.id << " " << dl.directory << ", " 
        << "table: " << dl.table << ", " 
        << "master: " << dl.master << ", "
        << "owner: " << dl.owner << ", "
        << "permissions: " << dl.permissions << ", "
        << "alcs: " << dl.acls << ", "
        << "repGroups: " << dl.repGroups;
    return out;
}

//
// Replica management
//
/** Adds a new replica location for a given guid
    This function is to be used on a local catalogue
 */
void MDStandalone::replicaAdd(const std::string &guid,
			      const std::string &surl,
			      const std::string &expires)
{
  if( !checkIfRoot() )
    return;

  std::stringstream query;
  std::string now = makeTimestamp();

  Statement statement(*dbConn);
  try {
    query << "INSERT INTO " << replicasTable << " (\"guid\", \"surl\", "
	  << "\"last_access\", \"expires\") "
	  << "VALUES"
	  << "('" << AMGA::validateGUID(guid) 
	  << "', '" << surl << "', '" << now << "', ";
    if( expires.size() )
      query << "'" << expires << "'";
    else
      query << "NULL";
    
    query << ");";
  } catch (std::runtime_error e){
    out.append("32 Illegal GUID\n");
    return;
  }

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Could not add replica", statement);
    return;
  }

  out.append("0\n");
  return;
}


/** Removes a replica location for a given guid
    This function is to be used on a local catalogue
*/
void MDStandalone::replicaRemove(const std::string &surl,
				 const std::string &guid)
{
  DMESG("MDStandalone::replicaRemove called\n");
  if( !checkIfRoot() )
    return;

  std::stringstream query;
  Statement statement(*dbConn);
  query << "DELETE FROM " << replicasTable << " WHERE \"surl\"='"
	<< surl <<"'";
  if( guid.size() )
    query << " AND \"guid\" = " << dbHelper->writeGUID(guid) << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("32 SURL does not exist", statement);
    return;
  }
  out.append("0\n");
  return;
}


/** Fills the idMap map which gives for a site-name the site ID number
    by selecting the entries from the sites table.
*/
int MDStandalone::getSiteIDs(const std::vector< std::string > & sites,
			     std::map< std::string, int > &idMap)
{
  for( size_t i = 0; i < sites.size(); i++ ) {
    if( idMap.find(sites[i]) != idMap.end() )
      continue;
    idMap[sites[i]] = -1;
  }

  std::ostringstream query;
  query << "SELECT \"name\", \"id\" FROM " << sitesTable << " WHERE ";
  bool needQuery = false;

  std::map< std::string, int >::iterator I = idMap.begin();
  while( I != idMap.end() ) {
    std::string site = I->first;
    char *ptr;
    int id = strtol(site.c_str(), &ptr, 10);

    if( site.c_str() != ptr ) { // Site is a number
      I->second = id;
      I++;
      continue;
    }

    if( needQuery )
      query << " OR ";
    query << "\"name\"='" << I->first << "'";
    needQuery = true;
    I++;
  }

  if( !needQuery )
    return 0;

  
  Statement statement(*dbConn);
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Error reading site table", statement);
    return 9;
  }
    
  char name[SITE_NAME_MAX_SIZE+1];
  int id;
  statement.bind(1, name, SITE_NAME_MAX_SIZE);
  statement.bind(2, &id);

  while( !statement.fetch() ) {
    idMap[name] = id;
  }

  statement.unBind();
  statement.close();

  I = idMap.begin();
  while( I != idMap.end() ) {
    if( I->second < 0 ) {
      out.append("31 Unknown site ");
      out.append(I->first);
      out.append("\n");
      return 31;
    }
    I++;
  }

  return 0;
}


int MDStandalone::replicaGetState(const std::vector< std::string > &guids,
				  SiteMap &siteMap,
				  SiteMap &unknownGUIDs)
{
  Statement statement(*dbConn);
  std::ostringstream query;

  query << "SELECT \"si\" FROM " << guidsTable << " WHERE \"guid\"= ?;";
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  statement.prepare(query.str());
  unsigned char si[SI_MAX_SIZE];
  if(statement.bind(1, si , SI_MAX_SIZE)) {
    printError("9 Could not bind columns for prepared statement", statement);
    return 9;
  }

  unsigned char guid[16];
  SQLINTEGER size = 16;
  SQLLEN resultSize = size;

  if (statement.bindBinaryParam(1, guid, size, &resultSize)) {
    printError("9 Could not bind parameters for prepared statement", statement);
    return 9;
  }

  for( size_t i = 0; i < guids.size(); i++ ) {
    DMESG("Reading: " << i << ": " << guids[i] << "\n");
    std::string g;
    try {
      AMGA::GUIDtoRaw(guids[i], guid);
    } catch (std::runtime_error e){
      std::string gu = guids[i];
      out.append("32 Illegal GUID: ");
      out.append(guids[i]);
      out.append("\n");
      return 32;
    }
    DMESG("Executing statement for guid " << guids[i] << "\n")
    if( statement.execPrepared() ) {
      printError("9 Internal error executing prepared select for guid",statement);
      return 9;
    }
    if( statement.fetch() ) {
      unknownGUIDs[guids[i]] = std::pair<std::string, std::string> ("00", "00");
      DMESG("GUID " << guids[i] << " was unknown\n");
      statement.closeCursor();
      continue;
    }
    int siSize = statement.getColumnLen(1);
    std::string oldSI = decodeBinarySI(si, siSize);
    DMESG("Old si = " << oldSI << "\n");
    statement.closeCursor();
    siteMap[guids[i]] = std::pair<std::string, std::string> (oldSI, oldSI);
  }

  return 0;
}


int MDStandalone::updateReplicas(SiteMap &siteMap)
{
  if( siteMap.empty() )
    return 0;

  Statement statement(*dbConn);
  std::ostringstream query;

  query << "UPDATE " + guidsTable + " SET \"si\"=? WHERE \"guid\"=? AND \"si\"=?;";
  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if( statement.prepare(query.str()) ) {
    printError("9 Error preparing statement", statement);
    return 9;
  }

  unsigned char newSI[SI_MAX_SIZE];
  SQLINTEGER newSISize = SI_MAX_SIZE;
  SQLLEN newSIResultSize = SI_MAX_SIZE;

  if (statement.bindBinaryParam(1, newSI, SI_MAX_SIZE, &newSIResultSize)) {
    printError("9 Could not bind parameters for prepared statement", statement);
    return 9;
  }
  
  unsigned char guid[16];
  SQLLEN guidResultSize = 16;


  if (statement.bindBinaryParam(2, guid, 16, &guidResultSize)) {
    printError("9 Could not bind parameters for prepared statement", statement);
    return 9;
  }

  unsigned char oldSI[SI_MAX_SIZE];
  SQLINTEGER oldSISize = SI_MAX_SIZE;
  SQLLEN oldSIResultSize = SI_MAX_SIZE;

  if (statement.bindBinaryParam(3, oldSI, SI_MAX_SIZE, &oldSIResultSize)) {
    printError("9 Could not bind parameters for prepared statement", statement);
    return 9;
  }

  SiteMap::iterator I = siteMap.begin();
  while( I != siteMap.end() ) {
    AMGA::GUIDtoRaw(I->first, guid);
    newSIResultSize = encodeBinarySI(newSI, I->second.second);
    oldSIResultSize = encodeBinarySI(oldSI, I->second.first);
    DMESG("Changing replica " << I->first 
	  << " was " << printBinary(oldSI, oldSIResultSize) << "(" << oldSIResultSize << ")"
	  << " set to " << printBinary(newSI, newSIResultSize) << "(" << newSIResultSize << ")\n");

    if( statement.execPrepared() && statement.getErrorCode() != SQL_NO_DATA) {
      printError("9 Internal error executing prepared guid UPDATE",statement);
      return 9;
    }

    I++;
  }

  return 0;
}


void MDStandalone::replicaUnregister(const std::string &options,
				     const std::vector< std::string > &guids,
				     const std::vector< std::string > &sites)
{
  if( !checkIfRoot() )
    return;

  Statement statement(*dbConn);
  std::ostringstream query;

  if( statement.beginTransaction() ) {
    printError("9 Could not bind parameters for prepared statement", statement);
    return;
  }

  std::map< std::string, int > idMap;
  if( getSiteIDs(sites, idMap) )
    return;

  SiteMap siteMap;
  SiteMap unknownGUIDs;
  if( replicaGetState(guids, siteMap, unknownGUIDs) )
    return;

  if( !unknownGUIDs.empty() && options.find("i") == std::string::npos) {
    out.append("1 Guid not found: ");
    out.append(unknownGUIDs.begin()->first);
    out.append("\n");
    return;
  }


  std::list< std::string > obsoleteGUIDs;
  for( size_t i = 0; i < guids.size(); i++ ) {
    if( siteMap.find(guids[i]) != siteMap.end() ) {
      DMESG("Deleting bit for GUID, was " << siteMap[guids[i]].second << "\n");
      changeSIBit(siteMap[guids[i]].second, idMap[sites[i]], false);
      DMESG("  now is " << siteMap[guids[i]].second << "\n");
      if( !countSIBits(siteMap[guids[i]].second) ) {
	obsoleteGUIDs.push_back(guids[i]);
      }
    }
  }
  
  // Remove guid references in case the last one was deleted and the d flag is set
  if( options.find("d") != std::string::npos ) {
    std::list< std::string >::iterator I = obsoleteGUIDs.begin();
    while( I != obsoleteGUIDs.end() )
      siteMap.erase(*I++);
  }
 
  if( !siteMap.empty() && updateReplicas(siteMap) )
    return;
  
  // Remove guid references in case the last one was deleted and the d flag is set
  if( options.find("d") != std::string::npos && !obsoleteGUIDs.empty() ){
    DMESG("Inserting new replicas\n");
    query << "DELETE FROM " + guidsTable + " WHERE \"guid\" = ?;";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if( statement.prepare(query.str()) ) {
      printError("9 Error preparing statement", statement);
      return;
    }

    unsigned char guid[16];
    SQLLEN guidResultSize = 16;
    
    if (statement.bindBinaryParam(1, guid, 16, &guidResultSize)) {
      printError("9 Could not bind parameters for prepared statement", statement);
      return;
    }
     
    std::list< std::string >::iterator I = obsoleteGUIDs.begin();
    while( I != obsoleteGUIDs.end() ) {
      AMGA::GUIDtoRaw(*I, guid);
      DMESG("Deleteting from replicas: " << *I << "\n");
      if( statement.execPrepared() && statement.getErrorCode() != SQL_NO_DATA) {
	printError("9 Internal error executing prepared guid DELETE",statement);
	return;
      }
      I++;
    }
  }

  DMESG("Commiting new replicas\n");
  if( statement.commitTransaction() ) {
    printError("9 Internal error in commit",statement);
    return;
  }

  out.append("0\n");
}
  

void MDStandalone::replicaRegister(const std::string &options,
				   const std::vector< std::string > &guids,
				   const std::vector< std::string > &sites)
{
  if( !checkIfRoot() )
    return;

  Statement statement(*dbConn);
  std::ostringstream query;

  DMESG("Opening transaction for new replicas\n");
  if( statement.beginTransaction() ){
    printError("9 Could not bind parameters for prepared statement", statement);
    return;
  }

  std::map< std::string, int > idMap;
  if( getSiteIDs(sites, idMap) )
    return;

  SiteMap siteMap;
  SiteMap unknownGUIDs;
  if( replicaGetState(guids, siteMap, unknownGUIDs) )
    return;
  
  for( size_t i = 0; i < guids.size(); i++ ) {
    if( siteMap.find(guids[i]) != siteMap.end() ) {
      DMESG("Setting SI bit for known GUID, was " << siteMap[guids[i]].second << "\n");
      changeSIBit(siteMap[guids[i]].second, idMap[sites[i]], true);
      DMESG("  now is " << siteMap[guids[i]].second << "\n");
    } else {
      DMESG("Setting SI bit for new GUID, was " << unknownGUIDs[guids[i]].second << "\n");
      changeSIBit(unknownGUIDs[guids[i]].second, idMap[sites[i]], true);
      DMESG("  now is " << unknownGUIDs[guids[i]].second << "\n");
    }
  }

  if( updateReplicas(siteMap) )
    return;

  if( options.find("c") == std::string::npos && !unknownGUIDs.empty() ) {
    out.append("1 Guid not found: ");
    out.append(unknownGUIDs.begin()->first);
    out.append("\n");
    return;
  }

  if( !unknownGUIDs.empty() ) {
    DMESG("Inserting new replicas\n");
    query << "INSERT INTO " + guidsTable + " (\"guid\", \"si\") "
	  << "VALUES (?, ?);";
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if( statement.prepare(query.str()) ) {
      printError("9 Error preparing statement", statement);
      return;
    }

    unsigned char guid[16];
    SQLLEN guidResultSize = 16;
    
    if (statement.bindBinaryParam(1, guid, 16, &guidResultSize)) {
      printError("9 Could not bind parameters for prepared statement", statement);
      return;
    }

    unsigned char newSI[SI_MAX_SIZE];
    SQLLEN newSIResultSize = SI_MAX_SIZE;

    if (statement.bindBinaryParam(2, newSI, SI_MAX_SIZE, &newSIResultSize)) {
      printError("9 Could not bind parameters for prepared statement", statement);
      return;
    }
     
    SiteMap::iterator I = unknownGUIDs.begin();
    while( I != unknownGUIDs.end() ) {
      AMGA::GUIDtoRaw(I->first, guid);
      newSIResultSize = encodeBinarySI(newSI, I->second.second);
      DMESG("Inserting guid " << I->first << " set to " << printBinary(newSI, newSIResultSize) << "\n");
      
      if( statement.execPrepared() && statement.getErrorCode() != SQL_NO_DATA) {
	printError("9 Internal error executing prepared guid INSERT",statement);
	return;
      }
      
      I++;
    }
  }

  DMESG("Commiting new replicas\n");
  if( statement.commitTransaction() ) {
    printError("9 Internal error in commit",statement);
    return;
  }

  out.append("0\n");
}


void MDStandalone::replicaListLFNs(std::vector< std::string > &lfns)
{
  DMESG("replicaListLFNs");

  Statement statement(*dbConn);
  std::list< std::string > groups; // Caching of groups user is in
  out.append("0\n");
  size_t i = 0;
  while( i < lfns.size() ) {
    EntryPropList propList;
    int error=getEntryProps(lfns[i], propList);
    if( error ) {
      out.append("\030");
      if(handleEntryPropErrors(error, lfns[i]))
	return;
    }
    EntryProps p(*propList.begin());
    
    
    std::list< std::string > groups;
    int perms = checkDirPermissions(p, 'r', groups);
    if( perms <= 0 && !(p.flags & EP_ACLS) ){
      DMESG("Permission denied" << std::endl);
      out.append("\030");
      out.append("4 Permission denied\n");
      return;
    }

    DMESG("Directory permissions OK" << std::endl);
  
    std::stringstream query;
    if( p.type() ) {
      out.append("\03019 Not an entry: " + lfns[i] + "\n");
      return;
    }

    query << "SELECT \"file\", " << dbHelper->readGUID("si") << " FROM " << guidsTable << ", "
	  << p.directoryTable << " WHERE guids.\"guid\" = "
	  << p.directoryTable << ".\"guid\" AND (";

    query << "\"file\" = '" << p.name << "'";

    i++;
    while( i < lfns.size() ){
      std::string nextFile;
      nextFile = inSameDir(absolutePath(lfns[i-1], currentDir),
			   absolutePath(lfns[i], currentDir));
      if( !nextFile.size() )
	break;
      query << " OR \"file\" = '" << nextFile <<"'";
      i++;
    }

    query << ") AND ";
    
    int t_flags = p.type() ? p.flags : p.parentFlags;
    query << permissionClause('r', t_flags, groups) << ";";

    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      out.append("\030");
      printError("9 Error reading directory", statement);
      return;
    }

    char file[FILENAME_MAX_SIZE];
    char si[FILENAME_MAX_SIZE];

    statement.bind(1, file, FILENAME_MAX_SIZE);
    statement.bind(2, si, FILENAME_MAX_SIZE);

    while( !statement.fetch() ) {
      out.append(p.parent + "/" + file + " ");
      std::list< int > sites;
      decodeSI(si, sites);
      std::list< int >::const_iterator I = sites.begin();
      while( I!= sites.end() ) {
	out.append(" ");
	out.append(*I++);
      }
      out.append("\n");
    }

    statement.unBind();
    statement.close();
  }
}


/** List the sites which have a replica of a given guid.
    Every user can call this method.
 */
void MDStandalone::replicaList(const std::string &options,
			       std::vector< std::string >&guids)
{
  DMESG("replicaList\n");
  
  if( options.find("l") != std::string::npos )
    return replicaListLFNs(guids);

  for( size_t i = 0; i < guids.size(); i++ ){
    try {
      guids[i] = AMGA::validateGUID(guids[i]);
    } catch (std::runtime_error e){
      out.append("32 Illegal GUID: " + guids[i] +"\n");
      return;
    }
  }

  Statement statement(*dbConn);
  std::stringstream query;
  query << "SELECT " << dbHelper->readGUID("guid") 
	<< ", " << dbHelper->readGUID("si") 
	<< " FROM " << guidsTable << " WHERE ";
  for (size_t i = 0; i < guids.size(); i++) {
    if( i ) query << " OR ";
    query << "\"guid\" = " << dbHelper->writeGUID(guids[i]);
  }
  query << ";";

  //  query << "SELECT \"guid\", \"si\" FROM " 
  //	<< guidsTable << " WHERE ";
  //  for (int i = 0; i < guids.size(); i++) {
  //    if( i ) query << " OR ";
  //    query << "\"guid\" = ?";
  //  }


  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Error reading guids table", statement);
    return;
  }

  out.append("0\n");

  char guid[GUID_SIZE];
  char si[SI_MAX_SIZE];
  statement.bind(1, guid, GUID_SIZE);
  statement.bind(2, si , SI_MAX_SIZE);
  while( !statement.fetch() ) {
    out.append(AMGA::niceGUID(guid));
    std::list< int > sites;
    decodeSI(si, sites);
    std::list< int >::const_iterator I = sites.begin();
    while( I!= sites.end() ) {
      out.append(" ");
      out.append(*I++);
    }
    out.append("\n");
  }
}


void MDStandalone::replicaGetLFN(const std::string &options,
				 std::vector< std::string >&lfns)
{
  std::list< std::string > groups; // Caching of groups user is in

  Statement statement(*dbConn);

  out.append("0\n");
  size_t i = 0;
  while( i < lfns.size() ) {
    EntryPropList propList;
    int error=getEntryProps(lfns[i], propList);
    if( error ) {
      out.append("\030");
      if(handleEntryPropErrors(error, lfns[i]))
	return;
    }
    EntryProps p(*propList.begin());
    
    if( checkDirPermissions(p, 'r', groups) <= 0 && !(p.flags & EP_ACLS) ){
      DMESG("Permission denied" << std::endl);
      out.append("\030");
      out.append("4 Permission denied\n");
      return;
    }

    DMESG("Directory permissions OK" << std::endl);
  
    std::stringstream query;
    if( p.type() ) {
      out.append("\03019 Not an entry: " + lfns[i] + "\n");
      return;
    }

    query << "SELECT \"file\", \"surl\" FROM " << replicasTable << ", "
	  << p.directoryTable << " WHERE replicas.\"guid\" = "
	  << p.directoryTable << ".\"guid\" AND (";

    query << "\"file\" = '" << p.name << "'";

    i++;
    while( i < lfns.size() ){
      std::string nextFile;
      nextFile = inSameDir(absolutePath(lfns[i-1], currentDir),
			   absolutePath(lfns[i], currentDir));
      if( !nextFile.size() )
	break;
      query << " OR \"file\" = '" << nextFile <<"'";
      i++;
    }

    query << ") AND ";
    
    int t_flags = p.type() ? p.flags : p.parentFlags;
    query << permissionClause('r', t_flags, groups) << ";";

    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      out.append("\030");
      printError("9 Error reading directory", statement);
      return;
    }

    char file[FILENAME_MAX_SIZE];
    char surl[FILENAME_MAX_SIZE];

    statement.bind(1, file, FILENAME_MAX_SIZE);
    statement.bind(2, surl, FILENAME_MAX_SIZE);

    while( !statement.fetch() ) {
      out.append(p.parent + "/" + file + " "  + surl + "\n");
    }

    statement.unBind();
    statement.close();
  }
}


/** Fast way to get a replica, this is only for root, as it
    does not check any permissions
    Called by replicaGet
*/
void MDStandalone::replicaGetFast(const std::string &options,
				  std::vector< std::string >&guids)
{
  DMESG("replicaGet\n");
  if( options.find("l") != std::string::npos ){
    replicaGetLFN(options, guids);
    return;
  }

  try {
    for( size_t i = 0; i < guids.size(); i++ )
      guids[i] = AMGA::validateGUID(guids[i]);
  } catch (std::runtime_error e){
    out.append("32 Illegal GUID\n");
    return;
  }

  Statement statement(*dbConn);
  std::stringstream query;
  query << "SELECT encode(\"guid\", 'hex'), \"surl\" FROM " << replicasTable << " WHERE ";
  for (size_t i = 0; i < guids.size(); i++) {
    if( i ) query << " OR ";
    query << "\"guid\"='" << guids[i] << "' ";
  }
  query << ";";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Error reading replica table", statement);
    return;
  }

  out.append("0\n");

  char surl[FILENAME_MAX_SIZE];
  char guid[GUID_SIZE];
  statement.bind(1, guid, GUID_SIZE);
  statement.bind(2, surl, FILENAME_MAX_SIZE);

 while( !statement.fetch() ) {
    out.append(AMGA::niceGUID(guid));
    if( options.find("P") != std::string::npos )
      out.append("\n");
    else
      out.append(" ");
    out.append(surl);
    out.append("\n");
  }
}


bool MDStandalone::decodeTableRef(const std::string &tableRef,
				  std::string &table,
				  int &id) const
{
  size_t pos = tableRef.find("_");
  if( pos == std::string::npos ) {
    table = tableRef;
    id = atoi(table.substr(3).c_str());
    return false;
  }
  table = tableRef.substr(0, pos);
  id = atoi(tableRef.substr(pos+1).c_str());
  return true;
}


/** Fills a TableEntry structure with the SURL, LFN and all the other metadata from a GUID
*/
int MDStandalone::entryFromGUID(std::vector< std::string >&guids,
				std::vector< EntryProps > &entries)
{
  entries.resize(guids.size());

  for( size_t i = 0; i < guids.size(); i++ ) {
    try {
      entries[i].guid = AMGA::validateGUID(guids[i]);
    } catch (std::runtime_error e){
      out.append("32 Illegal GUID: " + guids[i] + "\n");
      return -1;
    }
  }
  
  Statement statement(*dbConn);
  std::stringstream query;
  query << "SELECT encode(guids.\"guid\", 'hex')"
	<< ", \"surl\", \"reference\" "
	<< "FROM " << guidsTable << " LEFT OUTER JOIN " << replicasTable 
	<< " ON " << guidsTable << ".\"guid\" = " << replicasTable << ".\"guid\""
	<< " AND " << guidsTable << ".\"guid\" = ?;";

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if( statement.prepare(query.str()) ) {
    printError("9 Error preparing statement", statement);
    return -1;
  }                    

  unsigned char guidp[16];
  SQLLEN guidResultSize = 16;
      
  if (statement.bindBinaryParam(1, guidp, 16, &guidResultSize)) {
    printError("9 Could not bind parameters for prepared statement", statement);
    return -1;
  }

  char surl[FILENAME_MAX_SIZE];
  char guid[GUID_SIZE];
  char table[TABLE_NAME_MAX_SIZE];
  statement.bind(1, guid, GUID_SIZE);
  statement.bind(2, surl, FILENAME_MAX_SIZE);
  statement.bind(3, table, TABLE_NAME_MAX_SIZE);
   
  for (size_t i = 0; i < entries.size(); i++) {
    AMGA::GUIDtoRaw(guids[i], guidp);
    DMESG("Retrieving from replicas: " << entries[i].guid << "\n");
    if( statement.execPrepared() && statement.getErrorCode() != SQL_NO_DATA) {
      printError("9 Internal error executing prepared guid retrieve",statement);
      return -1;
    }
    if( !statement.fetch() ) {
      DMESG("X\n");
      // Instead use ORDER BY and sort entries?
      DMESG("Got " << guid << ":  surl " << surl << "  table: " << table << "\n");
      entries[i].surl = surl;
      entries[i].directoryTable = table;
    }
  }
  
  statement.unBind();
  statement.close();

  DMESG("I am here! 1\n");
  
  if( !entries.size() ) {
     ERRLOG("No guids found!\n");
     out.append("32 No such guid\n");
     return -1;
  }
  
  // Get information from the directory table
  size_t i=0;
  while( i < entries.size() ) {
    std::string tableName;
    int tableID;
    bool shared = decodeTableRef(entries[i].directoryTable,
				 tableName, tableID);
    entries[i].id = tableID;
    std::string oldTable = tableName;

    query.str("");
    query << "SELECT " << dbHelper->readGUID("guid") << ", \"owner\", \"permissions\", "
	  << "\"group_rights\" FROM " << tableName
	  << " WHERE ";

    if( shared )
      query << " (\"guid\" = '" << entries[i].guid << "' AND \"dir\"=" 
	    << tableID << ")";
    else
      query << "\"guid\" = '" << entries[i].guid << "'";
    query << ";";
    
    i++;
        
    while( i < entries.size() ) {
      shared = decodeTableRef(entries[i].directoryTable,
			      tableName, tableID);
      entries[i].id = tableID;
      if (tableName != oldTable )
	break;

      if( shared )
	query << " OR (\"guid\" = " << dbHelper->writeGUID(entries[i].guid) 
	      << " AND \"dir\"=" << tableID << ")";
      else
	query << " OR \"guid\" = " << dbHelper->writeGUID(entries[i].guid);

      query << " OR \"guid\" = " << dbHelper->writeGUID(entries[i].guid);

      i++;
    }
    
    DMESG("SQL: >" << query.str() << "<" << std::endl);
    if(statement.exec(query.str())){
      printError("9 Error reading replica table", statement);
      return -1;
    }

    DMESG("I am here 2a\n");

    char owner[USERNAME_MAX_SIZE];
    char permissions[PERMISSION_SIZE];
    char group_rights[PERMISSION_SIZE];
    
    statement.bind(1, guid, GUID_SIZE);
    statement.bind(2, owner, USERNAME_MAX_SIZE);
    statement.bind(3, permissions, PERMISSION_SIZE);
    statement.bind(4, group_rights, PERMISSION_SIZE);
  
  DMESG("I am here 2b\n");
  
    while( !statement.fetch() ) {
      DMESG("I am here 2c\n");
      
      // Instead use ORDER BY and sort entries?
      for( size_t j = 0; j < entries.size(); j++ ) {
        DMESG("I am here 2d" << j << "\n");
        
	if( entries[j].guid == guid ) {
	  entries[j].owner = owner;
	  entries[j].perms = permissions;
	  entries[j].perms += group_rights;
	  break;
	}
      }
    }
  }

  statement.unBind();
  statement.close();

  // Get information from the masterindex on the directory
  
  DMESG("I am here 3\n");
  
  query.str("");
  query << "SELECT \"directory\", \"owner_name\", \"permissions\", "
	<< "\"acls\", \"id\" FROM " << masterindexTable << " WHERE";

  int oldTableID = -1;
  for( size_t i = 0; i < entries.size(); i++) {
    if(oldTableID == entries[0].id)
      continue;
    if( i )
      query << " OR ";
    oldTableID = entries[0].id;
    query << "\"id\" = " << oldTableID;
  }

  DMESG("SQL: >" << query.str() << "<" << std::endl);
  if(statement.exec(query.str())){
    printError("9 Error reading replica table", statement);
    return -1;
  }
  
  char directory[DIRNAME_MAX_SIZE];
  char owner[USERNAME_MAX_SIZE];
  char permissions[PERMISSION_SIZE];
  char group_rights[PERMISSION_SIZE];
  char acls[ACL_MAX_SIZE];
  int tableID;
  
  statement.bind(1, directory, DIRNAME_MAX_SIZE);
  statement.bind(2, owner, USERNAME_MAX_SIZE);
  statement.bind(3, permissions, PERMISSION_SIZE);
  statement.bind(4, acls, ACL_MAX_SIZE);
  statement.bind(5, &tableID);

  DMESG("I am here 4!");

  while( !statement.fetch() ) {
    // Instead use ORDER BY and sort entries?
    for( size_t i = 0; i < entries.size(); i++ ) {
      if( entries[i].id == tableID ) {
	entries[i].parent = directory;
	entries[i].parentACLs = entries[i].ACLs = acls;
	entries[i].parentOwner = owner;
	entries[i].parentPerms = permissions;
      }
    }
  }

  // TODO: Allow entry ACLs

  return 0;
}


/** Returns a list of SURLs as replicas for the given GUIDs/LFNs
    One SURL is returned per line, if no replica exists, an empty set
    is returned.
*/
void MDStandalone::replicaGet(const std::string &options,
			      std::vector< std::string >&guids)
{
  DMESG("replicaGet\n");
  if( options.find("l") != std::string::npos )
    return replicaGetLFN(options, guids);

  if( user == "root" )
    return replicaGetFast(options, guids);

  std::vector< EntryProps > entries;
  if(entryFromGUID(guids, entries))
    return;

  out.append("0\n");
  std::list< std::string > groups;
  for( size_t i = 0; i < entries.size(); i++) {
    if(checkPermissions(entries[i], 'r', groups)<=0) {
      out.append("\0304 Permission denied: "
		 + AMGA::niceGUID(entries[i].guid) + "\n");
      return;
    }
    out.append(AMGA::niceGUID(entries[i].guid) 
	       + " " + entries[i].surl + "\n");
  }

}


void MDStandalone::lfnLookup(std::vector< std::string > &guids)
{
  DMESG("lfnLookup\n");
  std::vector< EntryProps > entries;
  if(entryFromGUID(guids, entries))
    return;

  out.append("0\n");
  std::list< std::string > groups;
  for( size_t i = 0; i < entries.size(); i++) {
    if(checkPermissions(entries[i], 'r', groups)<=0) {
      out.append("\0304 Permission denied: "
		 + AMGA::niceGUID(entries[i].guid) + "\n");
      return;
    }
    out.append(AMGA::niceGUID(entries[i].guid) 
	       + " " + entries[i].parent + "/" + entries[i].name
	       + "\n");
  }
  out.append("\n");
}


/** Adds a site to the list of sites
*/
void MDStandalone::siteAdd(const std::string &siteName,
			   const std::string &hostAndPort)
{
  if( !checkIfCapabilityAllowed("site_add"))
    return;

  DMESG("siteAdd\n");
  string host;
  int port;

  if( !hostAndPort.empty() ) {
    // Break the host/port
    size_t index = hostAndPort.find(':');
    if (index == string::npos) {
      out.append("9 Invalid site address: " + hostAndPort + "\n");
      return;
    }
    host = hostAndPort.substr(0, index);

    try {
      port = toInteger(hostAndPort.substr(index+1));
    } catch (std::runtime_error &ex) {
      out.append("9 Invalid site address: " + hostAndPort + "\n");
      return;
    }
  } else {
    port = 8822;
  }
  
  try {
    SiteManager sMan(*dbConn, debug);
    sMan.add(siteName, host, port, *this);
    out.append("0\n");
  } catch (SQLException &ex) {
    printError("9", ex);
  }
}


/** Removes a site from the list of sites
*/
void MDStandalone::siteRemove(const std::string &site)
{
  if( !checkIfCapabilityAllowed("site_remove") )
    return;
  try {
    SiteManager sMan(*dbConn, debug);
    sMan.remove(site);
    out.append("0\n");
  } catch (ReplicationException &ex) {
    printError("1", ex);
  } catch (SQLException &ex) {
    printError("9 Internal error", ex);
  }
}


/** Lists all known sites
*/
void MDStandalone::siteList()
{
  try {
    vector<Site> sites;
    SiteManager sMan(*dbConn, debug);
    sMan.getAll(sites);
    ostringstream os;
    os << "0\n";
    for (size_t i=0; i<sites.size(); i++) {
        os << sites[i].id << " " << sites[i].name << " " 
            << sites[i].hostname << ":" << sites[i].port << "\n"; 
    }
    out.append(os.str());
  } catch (SQLException &ex) {
    printError("31", ex);
  }
}

string addQuotes(const string& field) {
    return "'" + field + "'"; 
} 

void MDStandalone::siteDump(const vector<std::string> &siteNames)
{
  if( !checkIfRoot() )
    return;

  try {
    SiteManager sMan(*dbConn, debug);
    vector<Site> sites;
    if (siteNames.empty()) {
        // No site name provided. Retrieve all.
        sMan.getAll(sites);
    } else {
        // Retrieve the sites whose names were provided
        sMan.getSites(siteNames, sites);
        if (sites.size() != siteNames.size()) {
            out.append("9 Site(s) not found\n");
            return;
        }
    }

    ostringstream os;
    os << "0\n";
    for (size_t i=0; i<sites.size(); i++) {
        os << "site_add " << sites[i].name << " " << sites[i].hostname << ":" << sites[i].port << "\n";
        os << "site_set_properties " << addQuotes(sites[i].name)
          << " login " << addQuotes(sites[i].login)
          << " password " << addQuotes(sites[i].password)
          << " use_ssl " << sites[i].useSSL
          << " use_grid_proxy " << sites[i].useGridProxy
          << " authenticate_with_certificate " << sites[i].authenticateWithCertificate
          << " verify_server_cert " << sites[i].verifyServerCert
          << " cert_file " << addQuotes(sites[i].certFile)
          << " key_file " << addQuotes(sites[i].keyFile)
          << " trusted_cert_dir " << addQuotes(sites[i].trustedCertDir)
          << " require_data_encryption " << sites[i].requireDataEncryption
          << "\n";
    }
    out.append(os.str());
  } catch (SQLException &ex) {
    printError("31", ex);
  }
}

void MDStandalone::siteSetProperties(const string &site, 
        const vector<string> &properties, 
        const vector<string> &values)
{
  if(!checkIfCapabilityAllowed("site_set_properties")) return;
  try {
    SiteManager sMan(*dbConn, debug);
    sMan.setProperties(site, properties, values);        
    out.append("0\n");
  } catch (SQLException &ex) {
    printError("9", ex);
  }
}

void MDStandalone::siteGetProperties(const string &site, 
        const vector<string> &properties)
{
    if(!checkIfCapabilityAllowed("site_get_properties")) return;
    try {
        SiteManager sMan(*dbConn, debug);
        vector<string> values;   
        sMan.getProperties(site, properties, values);
        ostringstream os;
        os << "0\n";
        for (size_t i=0; i<values.size(); i++) {
            os << values[i] << "\n"; 
        }
        out.append(os.str());
    } catch (SQLException &ex) {
        printError("9", ex);    
    }
}

void MDStandalone::siteListProperties(const string &siteName) {
    if(!checkIfCapabilityAllowed("site_list_properties")) return;
    
    try {
        SiteManager sMan(*dbConn, debug);
        Site site;
        sMan.getSite(siteName, site);
        ostringstream os;
        os << "0\n" << site << "\n";
        out.append(os.str());
    } catch (SQLException &ex) {
        printError("9", ex);    
    }
}


void MDStandalone::import(const std::string &table, const std::string &dir,
			  const std::string &options)
{
  DMESG("import called with table=" << table << " dir=" << dir<<"\n");

  if(!checkIfRoot()) return; // A root only feature
  
  Statement statement(*dbConn);
  if(statement.beginTransaction()) {
    printError("9 Internal error", statement);
    return;
  }

  EntryPropList propList;
  int error=getEntryProps(dir, propList);
  if(handleEntryPropErrors(error, dir))
    return;
  EntryProps p(*propList.begin());

  if( p.type() ){
    out.append("16 Directory exists\n");
    return;
  }

  // Check the permissions of this dir, creating a directory is 
  // like adding an entry!
  
  std::list< std::string > groups;
  if (!checkPermResult(checkPermissions(p, 'w', groups))) {
    return;
  }

  DMESG("Parent:" << p.parent << "\n");
  std::string name=p.parent;
  if(name=="/")
    name.append(p.name);
  else
    name.append("/").append(p.name);
  DMESG("Checking if entry exists\n");

  DMESG("Name:" << name << "\n");
  if( entryExists(p) != 0 ) {
    out.append("15 Entry exists\n");
    return;
  }
  DMESG("Entry does not exists\n");

  int id;
  p.directoryTable = table;
  p.flags = EPT_DIR | EP_IMPORTED | EP_PLAIN;
  decodeTableOptions(options, p);

  AttrList columns;
  if(getAttributes(table, columns, true))
    return;

  if( columns.empty() ){
    out.append("35 Import error, table empty\n");
    return;
  }

  std::vector< std::string > attrs(columns.size());
  int i = 0;
  AttrList::const_iterator I = columns.begin();
  while( I != columns.end() )
    attrs[i++] = (*I++).first;

  DMESG("Importing table with flags " << p.flags << "\n");
  if(p.flags & EP_PERMS) {
    if(!findAmongAttributes(attrs, "owner") 
       || !findAmongAttributes(attrs, "permissions")
       || !findAmongAttributes(attrs, "group_rights") ) {
      out.append("35 Import error, missing mandatory permission column\n");
      return;
    }
    if( p.flags & EP_ACLS && (!findAmongAttributes(attrs, "acls") ) ) {
      out.append("35 Import error, missing mandatory acls column\n");
      return;
    }
  }
   
  if (options.find("main=") == std::string::npos && !columns.empty() )  {
    p.mainColumn = columns.begin()->first;
    //    p.mainType = columns.begin().second
  }

  id = addDirectoryToIndex(name, p, statement);

  // On error exit
  if (finalizeDirectoryToIndex(id, statement) < 0)
    return;

  if (saveLog(statement, getEntryDirectory(p))) {
    // error handled by saveLog
    statement.commitTransaction();  
    out.append("0\n");
  }

}



void MDStandalone::execute(const std::vector< std::string> &args)
{
  if( !execRoot.size() ){
    out.append("21 Execution not enabled, set ExecRoot on server\n");
    return;
  }

  if( args[1].find_first_of("/") != std::string::npos ) {
    out.append("4 Permission denied: Cannot execute out of ExecRoot\n");
    return;
  }


  std::string executable = execRoot;
  executable.append("/");
  executable.append(args[1]);

  // Setup pipe
  int pfd[2];
  if (pipe(pfd) == -1) {
    PERRLOG("Pipe failed");
    out.append("9 Could not create pipe\n");
    return;
  }

  if ( distinguishedName.size() )
    DMESG("Exec with DN: " << distinguishedName << std::endl);

  DMESG("Executing >" << executable << "<\n");

  // Fork a new process:
  pid_t pid = fork();
  if (pid == -1) {
    PERRLOG("Could not fork child process");
    out.append("9 Could not fork\n");
    return;
  }

  if (pid == 0) { // We are the child
    close(pfd[0]); // Close unused read end

    write(STDOUT_FILENO, "Test\n", 5);

    // Redirect child stdout
    if( pfd[1] != STDOUT_FILENO && dup2(pfd[1], STDOUT_FILENO) < 0) {
      PERRLOG("Could not duplicate pipe fd");
    }
    
    if ( close(pfd[1]) == -1 )
      PERRLOG("Could not close duplicated stdout");

    //Set UID if we are root
    if ( !getuid() &&  execUID >= 0 ) {
      DMESG("Setting UID to " << execUID << " GID: " << execGID << "\n");
      setuid(execUID);
      setgid(execGID);
    }

    // Create command arguments
    char **myargv = (char **) malloc(sizeof(char *) * 12);
    myargv[0] = new char[1025];
    strncpy(myargv[0], executable.c_str(), 1024);

    size_t i;
    assert( args.size() > 0 );
    for (i = 1; i < args.size()-1; i++){
      myargv[i] = new char[1025];
      strncpy(myargv[i], args[i+1].c_str(), 1024);
    }
    myargv[i] = NULL;

    // Set USER enviroment var
    char *myenv[4];
    i = 0;
    myenv[0]= new char [1025];
    strcpy(myenv[0], "USER=");
    strcat(myenv[0], user.c_str());
    i++;

    if( distinguishedName.size() ) {
      myenv[i] = new char [1025];
      strcpy(myenv[i], "DN=");
      strncat(myenv[i], distinguishedName.c_str(), 1024-3);
      i++;
    }
    
    myenv[i] = NULL;

    if( execve(myargv[0], myargv, myenv) < 0) {
      PERRLOG("Could not execute");
      out.append("21 Could not execute.\n");
      return;
    }
  }
  
  // Parent
  close(pfd[1]); // Close unused write end

  out.append("0\n");
  char buf[2];
  buf[1]='\0';
  std::string s;
  DMESG("Reading child response\n");
  while (read(pfd[0], buf, 1) > 0){
    s.append(buf);
  }

  out.append(s);
  DMESG(s);

  close(pfd[0]);

  DMESG("Waiting for program to terminate\n");
  // We are the parent, and need to wait for the child
  int status;
  // waitpid(pid, &status, 0); // Same as below
  wait(&status); 

  DMESG("Execute finished\n");
}


void MDStandalone::sudo(const std::string &new_user)
{
  DMESG("sudo to " << new_user << "\n");
                            
  if(!checkIfRoot()) return; // A root only feature
  
  user = new_user;
  
  out.append("0\n");
}                              


void MDStandalone::getVersion()
{
  char amga_version[VERSION_NAME_MAX_SIZE];

  Statement statement(*dbConn);

  std::string query = "SELECT \"amga_version\" from ";
  query += " amga ";
  query += ";";

  DMESG("SQL: >" << query << "<" << std::endl);
  if(statement.exec(query)){
    printError("9 getVersion Error Exec Statement", statement);
    return;
  }

  statement.bind(1, amga_version, VERSION_NAME_MAX_SIZE);

  if( !statement.fetch() ) {
    out.append("0\n");
    out.append(amga_version);
    out.append("\n");
  } else {
    printError("9 getVersion Error Fetch Statement", statement);
    return;
  }

  statement.unBind();
  statement.close();
}
